# Auto-generated by generate_concept_dicts.py.
# -*- coding: utf-8 -*-

# Dict with cid number as keys; value format is:
#   {scheme designator: <list of keywords for current cid>
#    scheme_designator: ...}

name_for_cid = {}
cid_concepts = {}

name_for_cid[12105] = 'IntracranialCerebralVessels'
cid_concepts[12105] = \
{'SCT': ['InternalCarotidArtery',
         'MiddleCerebralArteryM2Segment',
         'MiddleCerebralArtery',
         'PosteriorCerebralArteryP2Segment',
         'CarotidSiphon',
         'OphthalmicArtery',
         'PosteriorCerebralArtery',
         'PosteriorCommunicatingArtery',
         'AnteriorCerebralArtery',
         'AnteriorPosteriorCerebralArteryBifurcation',
         'MiddleCerebralArteryM1Segment',
         'CentralRetinalVein',
         'InternalCarotidArteryC5Segment',
         'AnteriorCommunicatingArtery',
         'AnteriorMiddleCerebralArteryBifurcation',
         'PosteriorCerebralArteryP1Segment',
         'InternalCarotidArteryC6Segment',
         'CentralRetinalArtery',
         'TerminalInternalCarotidArtery']}

name_for_cid[4051] = 'GeneralDevices'
cid_concepts[4051] = \
{'DCM': ['Phantom',
         'ACRAccreditationPhantomCT',
         'ACRAccreditationPhantomMR',
         'ACRAccreditationPhantomMammography',
         'ACRAccreditationPhantomStereotacticBreastBiopsy',
         'ACRAccreditationPhantomECT',
         'ACRAccreditationPhantomPET',
         'ACRAccreditationPhantomECTPET',
         'ACRAccreditationPhantomPETFaceplate',
         'IECHeadDosimetryPhantom',
         'IECBodyDosimetryPhantom',
         'NEMAXR212000Phantom',
         'Sphere'],
 'SCT': ['Glue',
         'AtherectomyDevice',
         'EmbolizationCoil',
         'EmbolizationParticulate',
         'Laser',
         'Catheter',
         'MeasuringRuler',
         'RotationalAtherectomyDevice',
         'GuidingCatheter',
         'EmbolizationBall',
         'PercutaneousTransluminalAngioplastyBalloon',
         'Syringe',
         'Stent',
         'DetachableBalloon']}

name_for_cid[9233] = 'RequestedReportTypes'
cid_concepts[9233] = \
{'DCM': ['FinalReport', 'PreliminaryReport']}

name_for_cid[6129] = 'ChestSiteInvolvement'
cid_concepts[6129] = \
{'DCM': ['Lobar'],
 'SCT': ['Lung',
         'Bronchial',
         'HilumOfLung',
         'Mediastinum',
         'PleuralStructure',
         'ChestWall',
         'UpperAbdomen',
         'InterstitialTissue',
         'Aorta']}

name_for_cid[6083] = 'ProceduresforBreast'
cid_concepts[6083] = \
{'DCM': ['SurgicalConsult',
         'MammographyCAD',
         'MarkerPlacement',
         'FilmScreenMammography',
         'DigitalMammography',
         'NippleDischargeCytology'],
 'LN': ['BreastBilateralMammogram',
        'WholeBodyPtWRncIv',
        'BreastUnilateralMrWoContrast',
        'BreastLeftMr',
        'BreastLeftMrWoAndWContrastIv',
        'BreastMrWoAndWContrastIv',
        'WholeBodyCt',
        'BreastBilateralMr',
        'BreastLeftMrWContrastIv',
        'BreastRightMammogram',
        'BreastFfdMammogram',
        'BreastUnilateralMrWoAndWContrastIv',
        'BreastUnilateralMammogram',
        'BreastMrWContrastIv',
        'BreastUnilateralMrWContrastIv',
        'BreastBilateralMrWContrastIv',
        'BreastRightMrWoAndWContrastIv',
        'BreastRightMr',
        'BreastBilateralMrWoAndWContrastIv',
        'BreastLeftMammogram',
        'BreastRightMrWContrastIv',
        'BreastUnilateralMr'],
 'SCT': ['ExaminationOfBreast',
         'RemovalOfBreastImplant',
         'FineNeedleAspirationOfBreast',
         'MRIOfBreast',
         'ReductionMammoplasty',
         'MammaryDuctogram',
         'BreastReconstruction',
         'UltrasonographyOfBreast',
         'CoreNeedleBiopsyOfBreast',
         'HormoneTherapy',
         'BoneMarrowTransplant',
         'RadiationTherapy',
         'SpecimenRadiographyOfBreast',
         'Chemotherapy',
         'PreBiopsyLocalizationOfBreastLesion',
         'BreastSurgicalBiopsy',
         'InsertionOfCatheter',
         'SentinelLymphNodeBiopsy',
         'BreastImplantation',
         'CTOfBreast',
         'RadioisotopeScanOfLymphaticSystem',
         'DiagnosticAspirationOfBreastCyst',
         'RadionuclideLocalizationOfTumorLimitedArea']}

name_for_cid[3452] = 'CalibrationMethods'
cid_concepts[3452] = \
{'DCM': ['GeometricNonIsocenter',
         'CalibrationObjectUsed',
         'GeometricIsocenter']}

name_for_cid[6200] = 'ColonOverallAssessment'
cid_concepts[6200] = \
{'DCM': ['C0InadequateStudyAwaitingPriorComparisons',
         'C1NormalColonOrBenignLesion',
         'C2IntermediatePolypOrIndeterminateFinding',
         'C3PolypPossiblyAdvancedAdenoma',
         'C4ColonicMassLikelyMalignant']}

name_for_cid[7041] = 'CalciumScoringPatientSizeCategories'
cid_concepts[7041] = \
{'DCM': ['SmallLesserThan32Point0cmLateralThickness',
         'Medium32Point0To38Point0cmLateralThickness',
         'LargeGreaterThan38Point0cmLateralThickness']}

name_for_cid[3239] = 'PerceivedExertionScales'
cid_concepts[3239] = \
{'DCM': ['BorgCR10Scale', 'BorgRPEScale']}

name_for_cid[12017] = 'GrowthDistributionRank'
cid_concepts[12017] = \
{'DCM': ['GrowthZScore', 'GrowthPercentileRank']}

name_for_cid[12294] = 'CardiacUltrasoundPericardiumandPleuraFindingSites'
cid_concepts[12294] = \
{'SCT': ['Pericardium', 'Pleura', 'LeftPleura', 'RightPleura']}

name_for_cid[7016] = 'LevelofDifficulty'
cid_concepts[7016] = \
{'DCM': ['PrimaryLevel', 'AdvancedLevel', 'IntermediateLevel']}

name_for_cid[3405] = 'ProcedureActionValues'
cid_concepts[3405] = \
{'DCM': ['ArterialConduitAngiography',
         'VascularIntervention',
         'MyocardialBiopsy',
         'AorticIntervention',
         'SeptalDefectIntervention',
         'ValvularIntervention'],
 'SCT': ['CardiacCatheterizationBaselinePhase',
         'TransthoracicEchocardiography',
         'StentPlacement',
         'CardiacVentriculography',
         'TransseptalCatheterization',
         'DrugInfusionChallenge',
         'AtherectomyByRotaryCutter',
         'CardiacCatheterizationInterventionPhase',
         'CardiacCatheterizationTherapyPhase',
         'LeftVentriculography',
         'CatheterizationOfBothLeftAndRightHeartWithoutGraft',
         'IntravascularUltrasound',
         'CardiacCatheterizationPostContrastPhase',
         'Aortography',
         'CatheterizationOfLeftHeart',
         'CardiacCatheterizationPostInterventionPhase',
         'PercutaneousInsertionOfIntravascularFilter',
         'CardiacCatheterizationImageAcquisitionPhase',
         'CatheterizationOfBothLeftAndRightHeartWithGraft',
         'AtherectomyByLaser',
         'RightVentriculography',
         'TranscatheterTherapyForEmbolization',
         'RestingState',
         'CardiacCatheterizationPreInterventionPhase',
         'Atherectomy',
         'SelectiveEmbolizationOfArtery',
         'PercutaneousRetrievalOfIntravascularForeignBody',
         'IntracardiacEchocardiography',
         'CoronaryArteriography',
         'CardiacCatheterizationTestChallengePhase',
         'PercutaneousTransluminalBalloonAngioplasty',
         'EpicardialEchocardiography',
         'BypassGraftAngiography',
         'CardiacCatheterizationBailoutPhase',
         'Angiography',
         'TransesophagealEchocardiography',
         'TranscatheterDeploymentOfDetachableBalloon',
         'CatheterizationOfRightHeart',
         'ExerciseChallenge']}

name_for_cid[3110] = 'NuclearCardiologyProtocols'
cid_concepts[3110] = \
{'DCM': ['StressTechnetiumRestTechnetium1DayProcedure',
         'NMMyocardialViabilityProcedure',
         'RestTechnetiumStressTechnetium1DayProcedure',
         'RestTechnetiumStressTechnetium2DayProcedure',
         'RestThalliumStressTechnetiumProcedure'],
 'SCT': ['StressThalliumProcedure']}

name_for_cid[3813] = 'StentFindings'
cid_concepts[3813] = \
{'DCM': ['StentFracture', 'StentDisintegration', 'Endoleak'],
 'SCT': ['Hypoplasia', 'MigrationOfImplantOrInternalDevice', 'Stenosis']}

name_for_cid[7153] = 'CNSSegmentationTypes'
cid_concepts[7153] = \
{'FMA': ['ArcuateFasciculus'],
 'SCT': ['SpinalCordWhiteMatter',
         'InferiorLongitudinalFasciculus',
         'Midbrain',
         'BrainCerebrospinalFluidPathway',
         'NucleusAccumbens',
         'SubstantiaNigra',
         'MiddleCerebellarPeduncle',
         'PiaMater',
         'CorpusCallosum',
         'FrontalLobe',
         'LimbicLobe',
         'ParietalLobe',
         'ThirdVentricle',
         'CerebralCortex',
         'CerebralWhiteMatter',
         'EntorhinalCortex',
         'SpinalCordGrayMatter',
         'LateralCorticospinalTract',
         'SubarachnoidSpace',
         'BrainStem',
         'OccipitalLobe',
         'BrainVentricle',
         'MedialLemniscus',
         'Pituitary',
         'Diencephalon',
         'InferiorCerebellarPeduncle',
         'CerebrospinalFluid',
         'WhiteMatter',
         'CerebralAqueduct',
         'Insula',
         'TemporalLobe',
         'PinealGland',
         'DuraMater',
         'Thalamus',
         'Amygdala',
         'OpticChiasm',
         'Arachnoid',
         'Neurohypophysis',
         'PosteriorCerebralCommissure',
         'SuperiorCerebellarPeduncle',
         'UncinateFasciculus',
         'SpinalCerebrospinalFluidPathway',
         'Telencephalon',
         'Meninges',
         'SpinalCord',
         'Hippocampus',
         'GrayMatter',
         'CerebellarWhiteMatter',
         'SuperiorLongitudinalFasciculus',
         'CentralNervousSystem',
         'FourthVentricle',
         'CerebralFornix',
         'Cingulum',
         'GlobusPallidus',
         'LateralVentricle',
         'CorpusStriatum',
         'Adenohypophysis',
         'OpticTract',
         'CaudateNucleus',
         'Brain',
         'Putamen',
         'OpticRadiation']}

name_for_cid[4010] = 'DXView'
cid_concepts[4010] = \
{'SCT': ['RightPosteriorOblique',
         'ObliqueCaudoCranial',
         'PosteroAnteriorOblique',
         'Axial',
         'Sagittal',
         'Tangential',
         'CranioCaudal',
         'LateroMedialOblique',
         'Acanthioparietal',
         'Orbitoparietal',
         'OcclusalProjection',
         'ObliqueCranioCaudal',
         'Plantodorsal',
         'LeftAnteriorOblique',
         'LateralOblique',
         'RightOblique',
         'MedioLateralOblique',
         'RightLateral',
         'CaudoCranial',
         'Verticosubmental',
         'TissueSpecimen',
         'Frontal',
         'LeftLateral',
         'Parietoacanthial',
         'PosteroAnterior',
         'RightAnteriorOblique',
         'Submentovertical',
         'AnteroPosteriorOblique',
         'LateralMedial',
         'LeftOblique',
         'Dorsoplantar',
         'ParietoOrbital',
         'Oblique',
         'MedialOblique',
         'FrontalObliqueAxial',
         'SagittalObliqueAxial',
         'AnteroPosterior',
         'MedialLateral',
         'FrontalOblique',
         'ObliqueAxial',
         'Lateral',
         'LeftPosteriorOblique']}

name_for_cid[3690] = 'ECGControlVariablesNumeric'
cid_concepts[3690] = \
{'MDC': ['LowPassFilterCutoffFreq',
         'InterpolatorSNR',
         'NotchFilterFrequency',
         'SampleRate',
         'NotchFilterBandwidth',
         'Sensitivity',
         'ZeroOffset',
         'PadValue',
         'TimeSkew',
         'SampleSkew',
         'TimeOffset']}

name_for_cid[12001] = 'UltrasoundProtocolTypes'
cid_concepts[12001] = \
{'SCT': ['TransthoracicEchocardiography',
         'EllestadProtocol',
         'DobutamineStressProtocol',
         'NaughtonProtocol',
         'StressTestUsingCardiacPacing',
         'PharmacologicAndExerciseStressTest',
         'ModifiedNaughtonProtocol',
         'RampProtocol',
         'IntravascularEchocardiography',
         'BalkeProtocol',
         'BruceProtocol',
         'ExerciseStressEchocardiography',
         'FetalEchocardiography',
         'StressTestUsingBicycleErgometer',
         'IntraoperativeEchocardiography',
         'ContrastEchocardiography',
         'PharmacologicStressProtocol',
         'PepperProtocol',
         'AdenosineStressProtocol',
         'IntracardiacEchocardiography',
         'DipyridamoleStressProtocol',
         'EpicardialEchocardiography',
         'TransesophagealEchocardiography',
         'ExerciseStressECGTest',
         'ModifiedBruceProtocol',
         'Echocardiography',
         'PediatricEchocardiography']}

name_for_cid[3413] = 'AdverseOutcomes'
cid_concepts[3413] = \
{'DCM': ['DeathDuringCatheterization'],
 'SCT': ['LossOfDistalPulse',
         'Pseudoaneurysm',
         'EmergencyPercutaneousCoronaryIntervention',
         'CardiogenicShock',
         'Arrhythmia',
         'ContrastMediaAdverseReaction',
         'ArterialDissection',
         'CardiacTamponade',
         'CardiacArrest',
         'AVFistula',
         'CongestiveHeartFailure',
         'CerebrovascularAccidentOrStroke',
         'OcclusionOfArtery',
         'EmergencyCoronaryArteryBypass',
         'RenalFailure',
         'Bleeding']}

name_for_cid[7000] = 'DiagnosticImagingReportDocumentTitles'
cid_concepts[7000] = \
{'DCM': ['BreastImagingReport'],
 'LN': ['MRIHeadReport',
        'XRayReport',
        'CTAbdomenReport',
        'EchocardiographyReport',
        'ECGReport',
        'CTChestReport',
        'CTHeadReport',
        'CardiacCatheterizationReport',
        'FluoroscopyStudyReport',
        'UltrasoundReport',
        'DEXASkeletalSystemStudyReport',
        'DiagnosticImagingReport',
        'CTReport',
        'DiagnosticInterventionalRadiologyReport',
        'MammographyScreeningReport',
        'CardiacElectrophysiologyReport',
        'UltrasoundObstetricAndGynReport',
        'NuclearMedicineThyroidScanReport',
        'RadiologyReport',
        'HolterStudyReport',
        'ExerciseStressTestReport',
        'PETScanReport',
        'NuclearMedicineReport',
        'MRISpineReport',
        'MRIReport']}

name_for_cid[12282] = 'CardiacUltrasoundVenousReturnSystemicFindingSites'
cid_concepts[12282] = \
{'SCT': ['HepaticVein',
         'LeftSuperiorVenaCava',
         'FontanFenestration',
         'MustardOrSenningSuperiorVenaCavaPathway',
         'RightSuperiorVenaCava',
         'InferiorVenaCava',
         'FontanPathway',
         'FontanBaffleLeak',
         'SuperiorVenaCava',
         'HemiFontanPathway',
         'FontanInferiorVenaCavaPathway',
         'FontanPulmonaryArteryConnection',
         'MustardOrSenningCommonSystemicVenousPathway',
         'GlennPathway',
         'MustardOrSenningInferiorVenaCavaPathway']}

name_for_cid[7450] = 'PersonRoles'
cid_concepts[7450] = \
{'DCM': ['Patient',
         'AdministratorOfRadiologyDepartment',
         'HeadOfCardiology',
         'LeadRadiologicTechnologist',
         'RepresentativeOfEthicsCommittee',
         'RepresentativeOfProtocolCommittee',
         'Fellow',
         'ChairOfProtocolCommittee',
         'HeadOfRadiology',
         'Sonologist'],
 'SCT': ['NaturalGrandFather',
         'MaleFirstCousin',
         'NaturalFather',
         'HealthcareProfessional',
         'HalfSister',
         'NaturalMother',
         'Resident',
         'NaturalSister',
         'Friend',
         'Physician',
         'Attending',
         'NaturalBrother',
         'ScrubNurse',
         'Registrar',
         'RadiologicTechnologist',
         'Aunt',
         'Surgeon',
         'NaturalGrandMother',
         'Uncle',
         'Consultant',
         'Nurse',
         'NaturalDaughter',
         'NaturalSon',
         'FemaleFirstCousin',
         'MedicalPractitioner',
         'HalfBrother',
         'RadiationTherapist'],
 'UMLS': ['Intern',
          'MedicalPhysicist',
          'ConsultingPhysician',
          'RadiationPhysicist',
          'Sonographer']}

name_for_cid[3805] = 'StenosisTypes'
cid_concepts[3805] = \
{'SCT': ['Embolism',
         'Thrombosis',
         'Entrapment',
         'Fibrous',
         'Compression',
         'ArterioscleroticVascularDisease',
         'Vasculitis',
         'CysticAdventitialDisease',
         'RaynaudDisease']}

name_for_cid[3106] = 'PETCardiologyProtocols'
cid_concepts[3106] = \
{'DCM': ['PETMyocardialViabilityRestAndStress',
         'PETMyocardialViabilityRestOnly',
         'PETMyocardialViabilityStressOnly',
         'PETMyocardialPerfusionRestAndStress',
         'PETMyocardialPerfusionRestOnly',
         'PETMyocardialPerfusionStressOnly']}

name_for_cid[3628] = 'CardiacOutputMethods'
cid_concepts[3628] = \
{'SCT': ['ThermalBath', 'ThermalInline', 'DyeDilution']}

name_for_cid[12113] = 'AbdominalVeins(Lateral)'
cid_concepts[12113] = \
{'SCT': ['TesticularVein',
         'FirstLumbarArtery',
         'OvarianVein',
         'FourthLumbarArtery',
         'FifthLumbarArtery',
         'SixthLumbarArtery',
         'CommonIliacVein',
         'GastricVein',
         'ThirdLumbarArtery',
         'SecondLumbarArtery',
         'IlealVein']}

name_for_cid[8121] = 'MicroscopyLensType'
cid_concepts[8121] = \
{'SCT': ['SlideOverviewLens', 'OilImmersionLens', 'HighPowerNonImmersionLens']}

name_for_cid[6095] = 'NumericPropertiesofProcedures'
cid_concepts[6095] = \
{'DCM': ['NeedleGauge', 'NeedleLength']}

name_for_cid[4102] = 'PerfusionMeasurementMethods'
cid_concepts[4102] = \
{'DCM': ['PerfusionAnalysisByStableXenonCTTechnique',
         'PerfusionAnalysisByIVIodinatedContrastCTTechnique',
         'PerfusionAnalysisByArterialSpinLabelingMRTechnique',
         'PerfusionAnalysisBySusceptibilityMRTechnique']}

name_for_cid[3014] = 'CoronaryArterySegments'
cid_concepts[3014] = \
{'BARI': ['_2ndMarginalCoronaryArteryLaterals',
          '_1stDiagonalCoronaryArteryLaterals',
          '_3rdDiagonalCoronaryArteryLaterals',
          '_1stMarginalCoronaryArteryLaterals',
          '_3rdMarginalCoronaryArteryLaterals',
          '_2ndDiagonalCoronaryArteryLaterals',
          'RamusLaterals',
          'PosteriorDescendingSeptalPerforators'],
 'SCT': ['ProximalRightCoronaryArtery',
         '_3rdLeftPosterolateralCoronaryArtery',
         'RightPosteriorAVCoronaryArtery',
         'ProximalCircumflexCoronaryArtery',
         '_2ndMarginalCoronaryArtery',
         'MidCircumflexCoronaryArtery',
         '_1stLeftPosterolateralCoronaryArtery',
         'IntermediateArteryRamus',
         'DistalLeftAnteriorDescendingCoronaryArtery',
         'RightCoronaryArteryOstium',
         'LeftMainCoronaryArtery',
         '_2ndLeftPosterolateralCoronaryArtery',
         'DistalRightCoronaryArtery',
         'MarginalCoronaryArtery',
         '_2ndRightPosterolateralCoronaryArtery',
         'PosteriorDescendingRightCoronaryArtery',
         'LeftPosteriorDescendingCircumflexCoronaryArtery',
         '_1stMarginalCoronaryArtery',
         'ProximalLeftAnteriorDescendingCoronaryArtery',
         '_3rdDiagonalCoronaryArtery',
         '_1stSeptalCoronaryArtery',
         '_3rdMarginalCoronaryArtery',
         '_1stDiagonalCoronaryArtery',
         'MidRightCoronaryArtery',
         '_2ndDiagonalCoronaryArtery',
         'AVGrooveContinuationOfCircumflexArtery',
         'DistalCircumflexCoronaryArtery',
         'MidLeftAnteriorDescendingCoronaryArtery',
         'LeftMainCoronaryArteryOstium',
         '_1stRightPosterolateralCoronaryArtery',
         '_3rdRightPosterolateralCoronaryArtery']}

name_for_cid[7112] = 'RadiotherapyFiducials'
cid_concepts[7112] = \
{'DCM': ['MRMarker',
         'InfraredReflectorMarker',
         'VisibleReflectorMarker',
         'CoilMarker',
         'CylinderMarker',
         'TransponderMarker',
         'WireMarker',
         'Sphere'],
 'SCT': ['RadioactiveImplant',
         'BBShotLeadPellet',
         'Clip',
         'ContrastMedia',
         'Catheter']}

name_for_cid[7305] = 'ImplantPlanningLandmarks'
cid_concepts[7305] = \
{'DCM': ['AnatomicalAxisOfFemur',
         'AcetabularCenterOfRotation',
         'FemurHeadCenterOfRotation']}

name_for_cid[33] = 'Modality'
cid_concepts[33] = \
{'DCM': ['BoneMineralDensitometry',
         'SlideMicroscopy',
         'StructuredReportDocument',
         'SpatialFiducials',
         'OphthalmicPhotography',
         'Other',
         'KeyObjectSelection',
         'Electrocardiography',
         'GeneralMicroscopy',
         'XRayAngiography',
         'IntraocularLensCalculation',
         'ExternalCameraPhotography',
         'ContentAssessmentResult',
         'IntravascularUltrasound',
         'ComputedRadiography',
         'ComputedTomography',
         'OpticalSurfaceScanner',
         'Thermography',
         'Registration',
         'Lensometry',
         'RTDose',
         'OphthalmicTomographyEnFace',
         'HardCopy',
         'RTPlan',
         'HemodynamicWaveform',
         'OpticalCoherenceTomography',
         'UltrasoundBoneDensitometry',
         'ModelFor3DManufacturing',
         'Diaphanography',
         'PresentationState',
         'PositronEmissionTomography',
         'CardiacElectrophysiology',
         'LaserScan',
         'PanoramicXRay',
         'OphthalmicMapping',
         'OphthalmicTomographyBScanVolumeAnalysis',
         'OphthalmicVisualField',
         'DigitalRadiography',
         'OphthalmicTomography',
         'Document',
         'RTTreatmentRecord',
         'Mammography',
         'Ultrasound',
         'IntravascularOpticalCoherenceTomography',
         'RTStructureSet',
         'MagneticResonance',
         'IntraOralRadiography',
         'RTImage',
         'VisualAcuity',
         'RespiratoryWaveform',
         'Endoscopy',
         'RealWorldValueMap',
         'Autorefraction',
         'Segmentation',
         'RadiographicImaging',
         'Radiofluoroscopy',
         'BasicVoiceAudio',
         'Keratometry',
         'StereometricRelationship',
         'CTProtocol',
         'AutomatedSlideStainer',
         'OphthalmicAxialMeasurements',
         'NuclearMedicine',
         'Plan',
         'BiomagneticImaging',
         'SubjectiveRefraction']}

name_for_cid[10010] = 'DoseMeasurementDevices'
cid_concepts[10010] = \
{'SCT': ['Dosimeter']}

name_for_cid[3716] = 'Severity'
cid_concepts[3716] = \
{'SCT': ['Moderate',
         'Mild',
         'MildToModerate',
         'ModerateToSevere',
         'None',
         'Fatal',
         'Severe']}

name_for_cid[6144] = 'RECISTDefinedLesionResponse'
cid_concepts[6144] = \
{'DCM': ['TargetLesionCompleteResponse',
         'TargetLesionPartialResponse',
         'TargetLesionProgressiveDisease',
         'TargetLesionStableDisease',
         'NonTargetLesionCompleteResponse',
         'NonTargetLesionIncompleteResponseOrStableDisease',
         'NonTargetLesionProgressiveDisease']}

name_for_cid[6001] = 'OverallBreastCompositionfromBIRADS®'
cid_concepts[6001] = \
{'SCT': ['ExtremelyDense',
         'AlmostEntirelyFat',
         'HeterogeneouslyDense',
         'ScatteredFibroglandularDensities']}

name_for_cid[3203] = 'ExerciserDevice'
cid_concepts[3203] = \
{'SCT': ['BicycleErgometer', 'Treadmill', 'ArmErgometer']}

name_for_cid[9527] = 'TeletherapyIsotope'
cid_concepts[9527] = \
{'SCT': ['_60Cobalt']}

name_for_cid[12241] = 'TricuspidValveFindingSites'
cid_concepts[12241] = \
{'SCT': ['TricuspidAnnulus']}

name_for_cid[3829] = 'PulmonaryArteries'
cid_concepts[3829] = \
{'SCT': ['TrunkOfPulmonaryArtery',
         'PulmonaryValveSinuses',
         'RightPulmonaryArtery',
         'SuprapulmonicValveArea',
         'LeftPulmonaryArtery']}

name_for_cid[7169] = 'BrainLesionSegmentationTypesWithoutNecrosis'
cid_concepts[7169] = \
{'DCM': ['Background'],
 'NCIt': ['NonEnhancingLesion', 'EnhancingLesion'],
 'SCT': ['Normal', 'Edema']}

name_for_cid[7186] = 'AbstractMultidimensionalImageModelQualitativeDimensionSampleSemantics'
cid_concepts[7186] = \
{'DCM': ['GlutamateAndGlutamine',
         'CholineCreatineRatio',
         'NAcetylaspartateCreatineRatio',
         'NAcetylaspartateCholineRatio',
         'VolumetricDiffusionDxzComponent',
         'RGBRComponent',
         'VolumetricDiffusionDxyComponent',
         'VolumetricDiffusionDxxComponent',
         'YBRFULLCRComponent',
         'YBRFULLCBComponent',
         'VolumetricDiffusionDzzComponent',
         'YBRFULLYComponent',
         'VolumetricDiffusionDyzComponent',
         'RGBBComponent',
         'VolumetricDiffusionDyyComponent',
         'RGBGComponent',
         'CreatineCholineCitrateRatio',
         'CreatineAndCholine',
         'LipidAndLactate',
         'YBRICTCRComponent',
         'YBRICTCBComponent',
         'YBRICTYComponent',
         'YBRPARTIALCRComponent',
         'YBRPARTIALCBComponent',
         'YBRPARTIALYComponent',
         'YBRRCTCRComponent',
         'YBRRCTCBComponent',
         'YBRRCTYComponent'],
 'SCT': ['Inositol',
         'Citrate',
         'Creatine',
         'Lactate',
         'Glutamine',
         'Lipid',
         'Choline',
         'NAcetylaspartate',
         'Tuarine']}

name_for_cid[12304] = 'EchoMeasuredProperties'
cid_concepts[12304] = \
{'DCM': ['PeakBloodPressure',
         'MaximumOrificeArea',
         'ExcursionDistance',
         'EffectiveOrificeArea',
         'DyssynchronyIndex',
         'VenaContractaWidth',
         'RegurgitationJetWidth',
         'RegurgitationJetArea',
         'PISARadius',
         'PeakTissueVelocity'],
 'LN': ['MeanGradientPressureByDoppler',
        'VelocityTimeIntegral',
        'PeakBloodVelocity',
        'VolumeFlowRate',
        'DpDtByUS',
        'Thickness',
        'InternalDimension',
        'PeakGradientPressure',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'EpicardialArea',
        'MyocardialPerformanceIndexTei',
        'MeanBloodVelocity'],
 'SCT': ['Interval',
         'VascularResistance',
         'BloodPressure',
         'Volume',
         'Area',
         'Mass',
         'MeanBloodPressure',
         'RegurgitantFraction',
         'CardiacEjectionFraction',
         'MinorAxis',
         'Angle',
         'Diameter',
         'StrokeVolume',
         'Length']}

name_for_cid[230] = 'YesNo'
cid_concepts[230] = \
{'SCT': ['Yes', 'No', 'Undetermined']}

name_for_cid[4250] = 'VisualFieldStaticPerimetryTestPatterns'
cid_concepts[4250] = \
{'DCM': ['VisualField60To4TestPattern',
         'VisualFieldCentralTestPattern',
         'VisualFieldMaculaTestPattern',
         'VisualFieldCentral40PointTestPattern',
         'VisualFieldCentral76PointTestPattern',
         'VisualFieldPeripheral60PointTestPattern',
         'VisualFieldFullField81PointTestPattern',
         'VisualFieldFullField120PointTestPattern',
         'VisualFieldGTestPattern',
         'VisualField24To2TestPattern',
         'VisualFieldMTestPattern',
         'VisualField10To2TestPattern',
         'VisualField07TestPattern',
         'VisualField30To2TestPattern',
         'VisualFieldLVCTestPattern']}

name_for_cid[6113] = 'MediastinumAnatomyFindingorFeature'
cid_concepts[6113] = \
{'DCM': ['Hiatus',
         'AzygoesophagealRecessInterface',
         'ParaspinalLine',
         'PosteriorTrachealStripe',
         'RightTrachealStripe',
         'Stripe'],
 'SCT': ['PulmonaryValve',
         'FascialLayer',
         'AorticValve',
         'ThoracicDuct',
         'AxillaryFascia',
         'Carina',
         'CarotidBody',
         'TricuspidValve',
         'Heart',
         'ThymusGland',
         'AtrialSeptalDefect',
         'LymphNode',
         'CostalCartilage',
         'Ventricle',
         'TrapeziusMuscle',
         'Atrium',
         'Esophagus',
         'LeftMainBronchus',
         'RightMainBronchus',
         'Thyroid',
         'LigamentumArteriosum',
         'Trachea',
         'MitralValve',
         'AcromioclavicularJoint',
         'EsophagealHiatus']}

name_for_cid[3741] = 'MedicationAdministration,LabVisit'
cid_concepts[3741] = \
{'SCT': ['Contraindicated',
         'AdministeredAfterLabVisit',
         'AdministeredBeforeLabVisit',
         'AdministeredDuringLabVisit']}

name_for_cid[7702] = 'ProjectionandThalamicFibers'
cid_concepts[7702] = \
{'NEU': ['CorticobulbarTract',
         'AnteriorThalamicRadiation',
         'SuperiorOpticRadiationBaumLoop',
         'CorticospinalTract',
         'InferiorOpticRadiationMeyerLoop',
         'SuperiorThalamicRadiation',
         'InferiorThalamicRadiation',
         'PosteriorThalamicRadiation'],
 'SCT': ['AuditoryRadiation',
         'InternalCapsule',
         'ExternalCapsule',
         'OpticRadiation']}

name_for_cid[10047] = 'GFRMeasurementMethods'
cid_concepts[10047] = \
{'DCM': ['CockroftGaultFormulaEstimationOfGFR',
         'CKDEPIFormulaEstimationOfGFR',
         'GlomerularFiltrationRateMDRD',
         'GlomerularFiltrationRateNonBlackMDRD',
         'GlomerularFiltrationRateBlackMDRD',
         'GlomerularFiltrationRateFemaleMDRD',
         'GlomerularFiltrationRateCystatinBasedFormula',
         'GlomerularFiltrationRateCreatinineBasedFormulaSchwartz']}

name_for_cid[64] = 'ImagingAgentAdministrationPatientState'
cid_concepts[64] = \
{'DCM': ['AcuteUnilateralRenalBlockage',
         'LowThyroidUptake',
         'HighThyroidUptake',
         'SeverelyJaundiced',
         'PatientAllergicToMediaContrast'],
 'SCT': ['Cardiomyopathy',
         'MalignantEpithelialNeoplasmOfThyroid',
         'Myeloma',
         'HistoryOfHypertension',
         'HistoryOfDiabetesMellitus',
         'AnginaPectoris',
         'Paraproteinemia',
         'HistoryOfCongestiveHeartFailure',
         'PulmonaryHypertension',
         'Anxiety',
         'Asthma',
         'HistoryOfBetaBlockingAgentsTherapy',
         'HistoryOfRenalFailure',
         'AorticStenosis',
         'AbnormalRenalFunction']}

name_for_cid[3254] = 'ElectrophysiologyProcedurePhase'
cid_concepts[3254] = \
{'SCT': ['AtrialEffectiveRefractoryPeriodEvaluationOf',
         'PostDefibrillationProcedurePhase',
         'VentricularEffectiveRefractoryPeriodEvaluationOf',
         'ElectrophysiologyProcedureBaselinePhase',
         'CarotidSinusMassageProcedurePhase',
         'PostAblationPhase',
         'ElectrophysiologyMappingPhase',
         'SinusNodeRecoveryTimeEvaluationOf',
         'RadiofrequencyAblationProcedurePhase']}

name_for_cid[6056] = 'AssociatedFindingsforBreast'
cid_concepts[6056] = \
{'DCM': ['CooperLigamentChanges', 'MassInTheSkin', 'MassOnTheSkin'],
 'SCT': ['BreastHematoma',
         'BreastComposition',
         'AxillaryAdenopathy',
         'SurgicalScar',
         'NippleRetraction',
         'Edema',
         'AxillaryLymphNode',
         'TrabecularThickeningOfBreast',
         'SkinRetractionOfBreast',
         'ArchitecturalDistortionOfBreast',
         'TubularDensity',
         'MammographyBreastDensity',
         'SkinLesion',
         'IndividualCalcification',
         'SkinThickeningOfBreast',
         'CalcificationCluster',
         'IntramammaryLymphNode']}

name_for_cid[3604] = 'ArterialLesionLocations'
cid_concepts[3604] = \
{'BARI': ['_1stDiagonalCoronaryArteryLaterals',
          '_1stMarginalCoronaryArteryLaterals',
          '_2ndDiagonalCoronaryArteryLaterals',
          'RamusLaterals',
          '_2ndMarginalCoronaryArteryLaterals',
          '_3rdDiagonalCoronaryArteryLaterals',
          '_3rdMarginalCoronaryArteryLaterals',
          'PosteriorDescendingSeptalPerforators'],
 'SCT': ['InternalCarotidArtery',
         'ProximalRightCoronaryArtery',
         'RightPosteriorAVCoronaryArtery',
         'BasilarArtery',
         'PulmonaryArteriovenousFistula',
         'LateralPlantarArtery',
         'MidCircumflexCoronaryArtery',
         'RightPulmonaryArtery',
         'IntermediateArteryRamus',
         'GeniculateArtery',
         'CarotidArtery',
         'UmbilicalArtery',
         'NeoAorta',
         'AorticFistula',
         'RadialArtery',
         'RightFemoralArtery',
         'SuperiorThyroidArtery',
         'IliacArtery',
         'LeftFemoralArtery',
         'DistalRightCoronaryArtery',
         'LeftAnteriorDescendingCoronaryArtery',
         'TruncusArteriosusCommunis',
         'Artery',
         'CommonCarotidArtery',
         '_3rdMarginalCoronaryArtery',
         'Aorta',
         'PatentDuctusArteriosus',
         'SuperficialTemporalArtery',
         'MidRightCoronaryArtery',
         'FistulaCoronaryToLeftVentricle',
         'CoronaryArtery',
         'LeftPosterolateralCircumflexCoronaryArtery',
         'DescendingAorta',
         'SubclavianArtery',
         'TibialArtery',
         '_3rdLeftPosterolateralCoronaryArtery',
         'OphthalmicArtery',
         'MesentericArtery',
         'MedialPlantarArtery',
         'PoplitealArtery',
         'AnteriorSpinalArtery',
         'HepaticArtery',
         'PeronealArtery',
         '_2ndRightPosterolateralCoronaryArtery',
         'BrachiocephalicTrunk',
         'LeftPosteriorDescendingCircumflexCoronaryArtery',
         'RenalArtery',
         '_1stMarginalCoronaryArtery',
         'SystemicCollateralArteryToLung',
         'FacialArtery',
         'LumbarArtery',
         'FemoralArtery',
         'PulmonaryArtery',
         'BrachialArtery',
         '_2ndDiagonalCoronaryArtery',
         'DistalCircumflexCoronaryArtery',
         'MidLeftAnteriorDescendingCoronaryArtery',
         'CoronaryArteryGraft',
         'FistulaCoronaryToRightAtrium',
         'FistulaCoronaryToRightVentricle',
         'CommonFemoralArtery',
         '_1stLeftPosterolateralCoronaryArtery',
         'PulmonaryArteryConduit',
         'CircumflexCoronaryArtery',
         'AnteriorCommunicatingArtery',
         'FistulaCoronaryToLeftAtrium',
         'ThoracicAorta',
         'DistalLeftAnteriorDescendingCoronaryArtery',
         'PulmonaryVeinWedge',
         'RightCoronaryArteryOstium',
         'AxillaryArtery',
         'LeftPulmonaryArtery',
         'MarginalCoronaryArtery',
         'InternalMammaryArtery',
         'ProximalLeftAnteriorDescendingCoronaryArtery',
         'VertebralArtery',
         '_3rdDiagonalCoronaryArtery',
         'AVGrooveContinuationOfCircumflexArtery',
         '_3rdRightPosterolateralCoronaryArtery',
         'NeonatalPulmonaryArteryPrimitivePA',
         'SuperficialFemoralArtery',
         'ProximalCircumflexCoronaryArtery',
         '_2ndMarginalCoronaryArtery',
         'PosteriorCommunicatingArtery',
         'AorticArch',
         'AbdominalAorta',
         'LingualArtery',
         'LeftMainCoronaryArtery',
         '_2ndLeftPosterolateralCoronaryArtery',
         'PosteriorDescendingRightCoronaryArtery',
         'Baffle',
         'LacrimalArtery',
         'OccipitalArtery',
         '_1stSeptalCoronaryArtery',
         'RightCoronaryArtery',
         'AscendingAorta',
         '_1stDiagonalCoronaryArtery',
         'ProfundaFemorisArtery',
         'LeftMainCoronaryArteryOstium',
         'CerebralArtery',
         '_1stRightPosterolateralCoronaryArtery']}

name_for_cid[12216] = 'EchocardiographyHepaticVeins'
cid_concepts[12216] = \
{'LN': ['HepaticVeinAtrialContractionReversalPeakVelocity',
        'HepaticVeinSystolicToDiastolicRatio',
        'HepaticVeinDiastolicPeakVelocity',
        'HeartRate',
        'HepaticVeinSystolicPeakVelocity']}

name_for_cid[3487] = 'IntravascularVolumetricRegions'
cid_concepts[3487] = \
{'DCM': ['DistalStentMargin',
         'EntirePullback',
         'ProximalStentMargin',
         'StentedRegion'],
 'SCT': ['Lesion', 'MorphologicallyAbnormalStructure', 'CulpritLesion']}

name_for_cid[3468] = 'EDVolume'
cid_concepts[3468] = \
{'DCM': ['LeftAtrialEDVolume'],
 'LN': ['RightVentricularEDVolume', 'LeftVentricularEDVolume']}

name_for_cid[4207] = 'OphthalmicImagePosition'
cid_concepts[4207] = \
{'DCM': ['LesionCentered',
         'DiscMaculaCentered',
         'MidPeripheralSuperior',
         'MidPeripheralSuperiorTemporal',
         'MidPeripheralTemporal',
         'MidPeripheralInferiorTemporal',
         'MidPeripheralInferior',
         'MidPeripheralInferiorNasal',
         'Field1ForJoslin3Field',
         'Field2ForJoslin3Field',
         'Field3ForJoslin3Field',
         'MaculaCentered',
         'DiscCentered',
         'PeripheralSuperiorTemporal',
         'PeripheralTemporal',
         'PeripheralInferiorTemporal',
         'PeripheralInferior',
         'PeripheralInferiorNasal',
         'PeripheralNasal',
         'PeripheralSuperiorNasal',
         'MidPeripheralNasal',
         'MidPeripheralSuperiorNasal',
         'PeripheralSuperior'],
 'SCT': ['DiabeticRetinopathyStudyField6',
         'DiabeticRetinopathyStudyField4',
         'DiabeticRetinopathyStudyField2',
         'DiabeticRetinopathyStudyField1',
         'DiabeticRetinopathyStudyField7',
         'DiabeticRetinopathyStudyField5',
         'DiabeticRetinopathyStudyField3']}

name_for_cid[1005] = 'AnatomicalReferenceBasisAbdomen/Pelvis'
cid_concepts[1005] = \
{'SCT': ['FemoralHead',
         'Sacrum',
         'SymphysisPubis',
         'HipJoint',
         'LesserTrochanter',
         'Ischium',
         'CommonIliacArteryBifurcation',
         'Pancreas',
         'Acetabulum',
         'Liver',
         'Kidney',
         'AdrenalGland',
         'IliacCrest']}

name_for_cid[637] = 'ExogenousSubstanceTypes'
cid_concepts[637] = \
{'DCM': ['TumorGraft'],
 'SCT': ['Fibril', 'Virus', 'Cytokine', 'Toxin', 'GraftMaterial']}

name_for_cid[72] = 'ImagingAgentAdministrationStepType'
cid_concepts[72] = \
{'DCM': ['DiagnosticAdministration',
         'TransitTimeTestInjection',
         'PatencyTestInjection',
         'FlushAdministration']}

name_for_cid[3757] = 'CardiacDiagnosticProcedures'
cid_concepts[3757] = \
{'SCT': ['CardiacCT',
         'ExerciseStressEchocardiography',
         'SPECT',
         'TransthoracicEchocardiography',
         'PETHeartStudy',
         'CardiacCTForCalciumScoring',
         'CardiacCathCoronaryAngiogramAndLeftVentriculogram',
         'PharmacologicalStressTest',
         'CardiacBloodPoolImaging',
         'StressTestUsingCardiacPacing',
         'CardiacCatheterization',
         'RadionuclideMyocardialPerfusionStudy',
         'PharmacologicAndExerciseStressTest',
         'CardiacCatheterizationCoronaryAngiogram',
         'PerfusionImaging',
         'CTAngiographyOfCoronaryArteries',
         'CardiacMRIStress',
         'ExerciseToleranceTest',
         'RadionuclideAngiocardiography',
         'NuclearMedicineCardiovascularStudy',
         'MagneticResonanceAngiography',
         'TransesophagealEchocardiography',
         'CardiacMRI',
         'Echocardiography']}

name_for_cid[6105] = 'AbnormalLucencyFindingorFeature'
cid_concepts[6105] = \
{'DCM': ['HaloSign', 'AirBronchiologram', 'AirBronchogram', 'AirCrescent'],
 'SCT': ['Pneumothorax', 'AirTrapping', 'Pneumomediastinum']}

name_for_cid[3612] = 'BloodVelocityMeasurements'
cid_concepts[3612] = \
{'DCM': ['BloodVelocityPeak',
         'BloodVelocityMean',
         'BloodVelocityMinimum',
         'SystolicBloodVelocityMean',
         'SystolicBloodVelocityPeak',
         'DiastolicBloodVelocityMean',
         'DiastolicBloodVelocityPeak']}

name_for_cid[6040] = 'NonlesionObjectType'
cid_concepts[6040] = \
{'DCM': ['OtherMarker', 'Unspecified'],
 'SCT': ['Suture',
         'CardiacPacemaker',
         'CompressionPaddle',
         'JWire',
         'ScarTissue',
         'Bullet',
         'BBShotLeadPellet',
         'OpaqueMarker',
         'Clip',
         'ContrastAgent',
         'Catheter',
         'IDPlate',
         'Collimator',
         'Implant',
         'Staple']}

name_for_cid[7201] = 'ReferencedImagePurposesofReference'
cid_concepts[7201] = \
{'DCM': ['ImagesRelatedToStandaloneObject',
         'AnatomicImage',
         'TemporalSuccessor',
         'Spectroscopy',
         'FunctionalImage',
         'OtherImageOfBiplanePair',
         'VolumeCorrespondingToSpatiallyRelatedAcquisitionFrames',
         'OtherImageOfStereoscopicPair',
         'TemporalPredecessor',
         'BiopsyLocalizer',
         'OtherPartialViews',
         'AcquisitionFramesCorrespondingToVolume',
         'Localizer',
         'ImageUsedForTreatmentPlanning',
         'ImageAcquiredDuringTreatment',
         'ImageUsedForDoseCalculation',
         'ImageUsedAsReferenceImageForTreatment',
         'DeviceLocalizer',
         'SpectralFilteredImage']}

name_for_cid[3491] = 'IVUSLesionMorphologies'
cid_concepts[3491] = \
{'DCM': ['InStentNeointima',
         'SoftPlaque',
         'VulnerablePlaque',
         'ErodedPlaque',
         'NecroticLipidicPlaque',
         'PlaqueRupture',
         'FalseLumen',
         'FibroLipidicPlaque'],
 'SCT': ['VascularCalcification',
         'PseudoAneurysm',
         'Eccentric',
         'PlaqueUlceration',
         'Concentric',
         'ArterialTrueAneurysm',
         'Thrombus',
         'FibrousPlaque']}

name_for_cid[12200] = 'EchocardiographyLeftVentricle'
cid_concepts[12200] = \
{'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'LeftVentricularEndDiastolicVolume',
        'LeftVentricleMass',
        'VolumeFlow',
        'FlowRadius',
        'LeftVentricularEjectionFractionByUS',
        'LeftVentricleInternalSystolicDimension',
        'InterventricularSeptumSystolicThickness',
        'HeartRate',
        'LeftVentriclePosteriorWallDiastolicThickness',
        'EndDiastolicVelocity',
        'LeftVentriclePosteriorWallSystolicThickness',
        'InterventricularSeptumDiastolicThickness',
        'PeakGradient',
        'LeftVentricleInternalEndDiastolicDimension',
        'LeftVentricleDiastolicMajorAxis',
        'InterventricularSeptumPercentThickening',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'LeftVentricularIsovolumicRelaxationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'ACInterval',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'TimeAveragedPeakVelocity',
        'InterventricularSeptumToPosteriorWallThicknessRatio',
        'VelocityOfFlowPropagation',
        'LeftVentriclePosteriorWallPercentThickening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'LeftVentricularEndSystolicVolume',
        'LeftVentricularFractionalShortening',
        'LeftVentricleSystolicMajorAxis',
        'SystolicToDiastolicVelocityRatio',
        'TimeAveragedMeanVelocity'],
 'SCT': ['LeftVentricleSemiMajorAxisDiastolicDimension',
         'LeftVentricularSystolicArea',
         'CardiacOutput',
         'LeftVentricularPeakEarlyDiastolicTissueVelocity',
         'LeftVentricularPeakSystolicTissueVelocity',
         'LeftVentricularIndexOfMyocardialPerformance',
         'LeftVentricularDiastolicArea',
         'CardiovascularOrificeDiameter',
         'StrokeIndex',
         'LeftVentricularFractionalAreaChange',
         'LeftVentricleTruncatedSemiMajorAxisDiastolicDimension',
         'LVPeakDiastolicTissueVelocityDuringAtrialSystole',
         'LeftVentricularIsovolumicContractionTime',
         'CardiacIndex',
         'LeftVentricleEpicardialDiastolicAreaPsaxPapView',
         'RegurgitantFraction',
         'RatioOfMVPeakVelocityToLVPeakTissueVelocityEWave',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[621] = 'TypeofMedicationforSmallAnimalAnesthesia'
cid_concepts[621] = \
{'DCM': ['CarrierGas'],
 'SCT': ['OcularLubricant',
         'CaloricAgent',
         'Vasoconstrictor',
         'CarbapenemAntibiotic',
         'BetaLactamAntibiotic',
         'AminoglycosideAntibiotic',
         'Potassium',
         'LocalAnesthetic',
         'NasalDecongestant',
         'Anticoagulant',
         'PenicillinAntibiotic',
         'ThyroidHormone',
         'MetronidazoleAntibiotic',
         'Calcium',
         'Magnesium',
         'AntidiureticHormone',
         'Oxytocic',
         'InotropicAgent',
         'Estrogen',
         'Antiarrhythmic',
         'Barbiturate',
         'Benzodiazepine',
         'SulfonamideAntibiotic',
         'Diuretic',
         'VancomycinAntibiotic',
         'Antifungal',
         'Hypoglycemic',
         'Steroid',
         'LowMolecularWeightHeparin',
         'RespiratoryStimulant',
         'AntiHeparinAgent',
         'Anticholinergic',
         'NeuromuscularBlockingNMBNonDepolarizing',
         'DrugDiluent',
         'GeneralAnesthetic',
         'HemostaticAgent',
         'MacrolideAntibiotic',
         'Antihypertensive',
         'Antibiotic',
         'Antihistamine',
         'AdrenergicAgent',
         'QuinoloneAntibiotic',
         'NSAID',
         'Vitamin',
         'Antiemetic',
         'BloodProduct',
         'Vasodilator',
         'BenzodiazepineAntagonist',
         'CalciumChannelBlocker',
         'CephalosporinAntibiotic',
         'CholinergicAgent',
         'GastrointestinalProkinetic',
         'PlasmaExpander',
         'Anticonvulsant',
         'BetaBlocker',
         'Bronchodilator',
         'TetracyclineAntibiotic',
         'SkeletalMuscleRelaxant',
         'NarcoticOpiateAntagonist',
         'LincomycinAntibiotic',
         'NeuromuscularBlockingNMBDepolarizing',
         'LinezolidAntibiotic',
         'Analgesic'],
 'UMLS': ['H2Antagonist', 'NarcoticAnalgesic', 'Antihypoglycemic']}

name_for_cid[271] = 'ObservationSubjectClass'
cid_concepts[271] = \
{'DCM': ['Fetus', 'Specimen', 'Patient', 'DeviceSubject']}

name_for_cid[4211] = 'OphthalmicOCTAnatomicStructureImaged'
cid_concepts[4211] = \
{'SCT': ['AnteriorChamberOfEye',
         'CornealEndothelium',
         'Vitreous',
         'OpticNerveHead',
         'ChoroidOfEye',
         'Sclera',
         'Cornea',
         'Lens',
         'CornealEpithelium',
         'Retina',
         'Iris',
         'CiliaryBody',
         'Conjunctiva']}

name_for_cid[6152] = 'Orientation'
cid_concepts[6152] = \
{'DCM': ['Parallel', 'NotParallel']}

name_for_cid[3700] = 'CathDiagnosis'
cid_concepts[3700] = \
{'SCT': ['AorticDissection',
         'Arrhythmia',
         'VentricularFibrillation',
         'RecurrentAnginaPostStent',
         'PericardialTamponade',
         'IdiopathicHypertrophicSubaorticStenosis',
         'AorticValveDisease',
         'UnstableAnginaProgressiveAngina',
         'MitralRegurgitation',
         'PulmonaryHypertension',
         'PulmonaryEdema',
         'MitralStenosis',
         'AtrialSeptalDefect',
         'PericardialDisease',
         'AcuteVentricularSeptalRupture',
         'PostInfarctionAngina',
         'MitralValveProlapse',
         'StableAngina',
         'HeartDiseaseCongenital',
         'AtypicalChestPain',
         'RecurrentAnginaPostRotationalAtherectomy',
         'PulmonicValveDisease',
         'AorticAneurysm',
         'AorticInsufficiency',
         'AtrialFibrillation',
         'CardiogenicShock',
         'TricuspidValveDisease',
         'RecurrentAnginaPostDCA',
         'Myocarditis',
         'VentricularTachycardia',
         'HypertensiveHeartDisease',
         'RecurrentAnginaPostPTCA',
         'CongestiveCardiomyopathy',
         'VentricularSeptalDefect',
         'CoronaryArteryDisease',
         'AcuteMyocardialInfarction',
         'SilentIschemia',
         'RestrictiveCardiomyopathy',
         'SPMIPositiveStressForIschemia',
         'AtypicalAnginaVariantAngina',
         'CongestiveHeartFailure',
         'SubacuteBacterialEndocarditis',
         'AorticStenosis',
         'MitralValveDisease',
         'TricuspidRegurgitation',
         'AcuteMitralRegurgitation',
         'RecurrentAnginaPostCABG',
         'ConstrictivePericarditis']}

name_for_cid[10006] = 'XRayFilterMaterials'
cid_concepts[10006] = \
{'SCT': ['TinOrTinCompound',
         'NiobiumOrNiobiumCompound',
         'TantalumOrTantalumCompound',
         'SilverOrSilverCompound',
         'EuropiumOrEuropiumCompound',
         'MolybdenumOrMolybdenumCompound',
         'LeadOrLeadCompound',
         'CopperOrCopperCompound',
         'RhodiumOrRhodiumCompound',
         'AluminumOrAluminumCompound']}

name_for_cid[25] = 'Radiopharmaceuticals'
cid_concepts[25] = \
{'NCIt': ['Technetium99mTrofolastat'],
 'SCT': ['TechnetiumTc99mTetrofosmin',
         'IodinatedI131Albumin',
         'FerrousChlorideFe59',
         'ColloidalIndium111',
         'SodiumIodideI131',
         'Chromium51Chloride',
         'Indium111RedCellLabel',
         'TechnetiumTc99mAlbuminColloid',
         'IodinatedI125OleicAcidAndTriolein',
         'DiagnosticRadioisotope',
         'SodiumIodideI123',
         'Carbon14DXylose',
         'Iodine131Polyvinylpyrrolidone',
         'TechnetiumTc99mSulfurColloid',
         'TechnetiumTc99mTeboroxime',
         'Technetium99mPhytate',
         'TechnetiumTc99mIronAscorbate',
         'StrontiumNitrateSr87',
         'ColloidalGoldAu198',
         'TechnetiumTc99mPyroAndPolyphosphates',
         'Iodine1233IodobenzylguanidineMIBG',
         'TechnetiumTc99mSestamibi',
         'RoseBengalSodiumI131',
         'Copper64Acetate',
         'Indium113mPentetate',
         'Technetium99mDimercaptosuccinicAcidDMSA',
         'TechnetiumTc99mMedronate',
         'TechnetiumTc99mSuccimer',
         'Technetium99mHydroxymethyleneDiphosphonateHMDP',
         'TechnetiumTc99mSerumAlbumin',
         'IodinatedI131AggregatedAlbumin',
         'SodiumFluorideF18',
         'SodiumPertechnetateTc99m',
         'StrontiumNitrateSr85',
         'TechnetiumTc99mDisofenin',
         'Indium111Pentetreotide',
         'TechnetiumTc99mMebrofenin',
         'IodinatedI131GammaGlobulin',
         'Indium111Chloride',
         'SodiumChromateCr51',
         'TechnetiumTc99mStannousEtidronate',
         'FluorodeoxyglucoseF18',
         'IodinatedI125RoseBengal',
         'Indium111FeOHGreaterThan3LesserThan',
         'TechnetiumTc99mLidofenin',
         'SelenomethionioneSe75',
         'FibrinogenI123',
         'TechnetiumTc99mTaggedRedCells',
         'TechnetiumTc99mOxidronate',
         'IodinatedI125Povidone',
         'IodohippurateI123Sodium',
         'Selenium75HCAT',
         'Iodine1313IodobenzylguanidineMIBG',
         'TechnetiumTc99cAlbuminMicrospheres',
         'StrontiumChlorideSr87',
         'Technetium99mLabeledCarbon',
         'DiiodofluoreceinI131',
         'TechnetiumTc99mDepreotide',
         'StrontiumChlorideSr85',
         'ChromicPhosphateP32',
         'IothalamateSodiumI125',
         'CyanocobalaminCo57',
         'Indium111Transferrin',
         'IodinatedI125Levothyroxine',
         'CyanocobalaminCo58',
         'Technetium99mStannousColloid',
         'Indium111Oxyquinoline',
         'Indium111Pentetate',
         'Copper67Ceruloplasmin',
         'Indium113mOxoquinolineWBCLabel',
         'Technetium99mMercaptoacetylTriglycineMAG3',
         'SodiumChlorideNa24',
         'Carbon14Triolein',
         'Krypton81m',
         'IodohippurateI131Sodium',
         'Copper64Versenate',
         'IodohippurateI125Sodium',
         'PotassiumCarbonateK42',
         'Indium113mChloride',
         'Indium113mOxoquinolineRBCLabel',
         'IofetamineI123Hydrochloride',
         'TechnetiumTc99mPentetate',
         'Iodine131Methylnorcholestenol',
         'IodocholesterolI131',
         'Chromium51Albumin',
         'Iodine123154Iodophenyl3RSMethylpentadecanoicAcid',
         'TechnetiumTc99mMicroaggregatedAlbumin',
         'TechnetiumTc99mExametazine',
         'Indium111CapromabPendetide',
         'CholylCarbon14Glycine',
         'TherapeuticRadioisotope',
         'Technetium99mNPyridoxyl5Methyltryptophan',
         'TechnetiumTc99mSodiumGlucoheptonate',
         'DisodiumIndium111',
         'PotassiumChlorideK42',
         'PentetateCalciumTrisodiumYb169',
         'FerrousCitrateFe59',
         'Technetium99mGalactosylHumanSerumAlbuminDiethylenetriamine',
         'CyanocobalaminCo60',
         'OleicAcidI125',
         'TechnetiumTc99mAggregatedAlbumin',
         'IodinatedI125SodiumIodine',
         'SodiumPhosphateP32',
         'PotassiumChlorideK43',
         'ThallousChlorideTl201',
         'Indium113mOxoquinolinePlateletLabel',
         'IodinatedI125HumanSerumAlbumin',
         'IronFe59LabeledDextran',
         'IodinatedI125Albumin',
         'Indium113mBleomycin',
         'FerrousSulfateFe59',
         'IodinatedI125SealedSource',
         'Gallium67Citrate',
         'TechnetiumTc99mNSubstitutedIminodiacetate',
         'SodiumChlorideNa22']}

name_for_cid[3215] = 'PerfusionFindingMethod'
cid_concepts[3215] = \
{'SCT': ['ECGAnalysis', 'ImageAnalysis']}

name_for_cid[6017] = 'CompositeFeaturefromBIRADS®'
cid_concepts[6017] = \
{'SCT': ['FocalAsymmetricBreastTissue',
         'MammographicBreastMass',
         'AsymmetricBreastTissue']}

name_for_cid[12257] = 'CardiacUltrasoundLeftVentricle'
cid_concepts[12257] = \
{'DCM': ['Tau', 'IndicatorMeanTransitTime', 'RRInterval', 'WallMass'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'TissueVelocityTimeIntegralVTIForTheAreaUnderLeftVentricleAWave',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'EndocardialArea',
        'PeakTissueVelocity',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'EFSlope',
        'TimeToLeftVentricleETissueVelocity',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'PreEjectionPeriod',
        'TissueVelocityTimeIntegralVTIForTheAreaUnderLeftVentricleEWave',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'InterventricularSeptumToPosteriorWallThicknessRatio',
        'IsovolumicRelaxationTime',
        'TimeToLeftVentricleSTissueVelocity',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'LeftVentricleMeridionalWallStress',
        'LeftVentricleEToATissueVelocityRatio',
        'SystolicToDiastolicVelocityRatio',
        'EpicardialArea',
        'MyocardialPerformanceIndexTei',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Volume',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'RatioOfMVPeakVelocityToLVPeakTissueVelocityEWave',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[9531] = 'VolumeandDoseDosimetricObjectives'
cid_concepts[9531] = \
{'DCM': ['MaximumAbsoluteVolumeAtRadiationDose',
         'MinimumAbsoluteVolumeAtRadiationDose']}

name_for_cid[3429] = 'CatheterizationDevices'
cid_concepts[3429] = \
{'DCM': ['Bioptome'],
 'SCT': ['FiberopticCatheter',
         'IntracoronaryPressureGuideWire',
         'IntravascularOpticalCoherenceTomographyDevice',
         'CuttingBalloonAngioplastyCBADevice',
         'Laser',
         'ThermistorCatheter',
         'Sheath',
         'BrachytherapyDevice',
         'FluidFilledCatheter',
         'DirectionalCoronaryAtherectomyDCADevice',
         'SalineThrombectomy',
         'HallCatheter',
         'RotationalAtherectomyDevice',
         'SwannGanzCatheter',
         'GuidingCatheter',
         'PercutaneousTransluminalAngioplastyBalloon',
         'GuideWire',
         'TipManometer',
         'IntracoronaryDopplerGuideWire',
         'IntravascularUltrasoundIVUSDevice',
         'TransseptalCatheter',
         'RadiofrequencyAblationDevice',
         'Stent',
         'TransluminalExtractionCatheterTEC',
         'IntraAorticBalloonPumpIABP']}

name_for_cid[226] = 'PopulationStatisticalDescriptors'
cid_concepts[226] = \
{'DCM': ['StandardDeviationOfPopulation', '_2SigmaDeviationOfPopulation'],
 'SCT': ['_1SigmaLowerValueOfPopulation',
         '_2SigmaLowerValueOfPopulation',
         '_5thPercentileValueOfPopulation',
         '_10thPercentileValueOfPopulation',
         '_2SigmaUpperValueOfPopulation',
         'MeanValueOfPopulation',
         '_95thPercentileValueOfPopulation',
         '_90thPercentileValueOfPopulation',
         '_1SigmaUpperValueOfPopulation',
         'MedianValueOfPopulation']}

name_for_cid[601] = 'BiosafetyLevels'
cid_concepts[601] = \
{'SCT': ['BiosafetyLevel2',
         'BiosafetyLevel1',
         'BiosafetyLevel4',
         'BiosafetyLevel3']}

name_for_cid[251] = 'SeverityofComplication'
cid_concepts[251] = \
{'SCT': ['Minor', 'Major']}

name_for_cid[9403] = 'IEC61217PatientSupportPositionParameters'
cid_concepts[9403] = \
{'DCM': ['IEC61217TableTopLongitudinalPosition',
         'IEC61217TableTopVerticalPosition',
         'IEC61217PatientSupportContinuousAngle',
         'IEC61217TableTopContinuousPitchAngle',
         'IEC61217TableTopContinuousRollAngle',
         'IEC61217TableTopEccentricAxisDistance',
         'IEC61217TableTopContinuousEccentricAngle',
         'IEC61217TableTopLateralPosition']}

name_for_cid[4231] = 'LensStatus'
cid_concepts[4231] = \
{'SCT': ['Pseudophakia', 'Phakic', 'PiggybackIOL', 'Aphakic', 'PhakicIOL']}

name_for_cid[12220] = 'EchocardiographyCommonMeasurements'
cid_concepts[12220] = \
{'LN': ['HeartRate']}

name_for_cid[6060] = 'BreastBiopsyTechniques'
cid_concepts[6060] = \
{'DCM': ['VacuumAssisted', 'Mammographic', 'Mammographic', 'PalpationGuided'],
 'SCT': ['QuadrantectomyOfBreast',
         'Lumpectomy',
         'DiagnosticRadiographyStereotacticLocalization',
         'Mammography',
         'UltrasonicGuidanceProcedure',
         'MagneticResonanceImagingGuidedBiopsy',
         'ComputedTomographyGuidedBiopsy'],
 'UMLS': ['Mastectomy']}

name_for_cid[3262] = 'ECGPatientStateValues'
cid_concepts[3262] = \
{'SCT': ['BaselineState',
         'ExerciseState',
         'SupineBodyPosition',
         'PostExerciseState',
         'RestingState']}

name_for_cid[7221] = 'RTDosePurposeofReference'
cid_concepts[7221] = \
{'DCM': ['SourceDoseForComposingCurrentDose']}

name_for_cid[12109] = 'LowerExtremityArteries'
cid_concepts[12109] = \
{'SCT': ['AnteriorTibialArtery',
         'ExternalIliacArtery',
         'SuperficialFemoralArtery',
         'CommonFemoralArtery',
         'PoplitealArtery',
         'PlantarArterialArch',
         'InternalIliacArtery',
         'CommonIliacArteryBifurcation',
         'CommonIliacArtery',
         'ProfundaFemorisArtery',
         'DorsalisPedisArtery',
         'PosteriorTibialArtery',
         'PeronealArtery']}

name_for_cid[10071] = 'RadiationDoseUnits'
cid_concepts[10071] = \
{'UCUM': ['Sv', 'Gy']}

name_for_cid[3777] = 'ImplantedDevices'
cid_concepts[3777] = \
{'SCT': ['ImplantableDefibrillator',
         'CardiacPacemaker',
         'LeftVentricularAssistDevice',
         'InsulinPump']}

name_for_cid[6125] = 'GeneralChestLocation'
cid_concepts[6125] = \
{'SCT': ['Peripheral', 'Basal', 'Central', 'Apical']}

name_for_cid[3409] = 'AdministrationofDrugs/Contrast'
cid_concepts[3409] = \
{'DCM': ['DrugStart',
         'DrugEnd',
         'InfusateStart',
         'InfusateEnd',
         'ContrastEnd',
         'ContrastAdministered',
         'DrugAdministered',
         'ContrastStart']}

name_for_cid[4266] = 'OphthalmicAnatomicStructureReferencePoint'
cid_concepts[4266] = \
{'DCM': ['DiscFovea'],
 'SCT': ['Lesion',
         'OpticNerveHead',
         'MorphologicallyAbnormalStructure',
         'Cornea',
         'FoveaCentralis']}

name_for_cid[8300] = 'VisualEvaluationMethods'
cid_concepts[8300] = \
{'DCM': ['GrayscaleResolutionEvaluation',
         'OverallImageQualityEvaluation',
         'LuminanceUniformityEvaluation',
         'LuminanceResponseEvaluation',
         'ClinicalEvaluation',
         'AngularViewingEvaluation',
         'PixelFaultsEvaluation',
         'ChromaticityEvaluation',
         'GeometricalImageEvaluation',
         'VeilingGlareEvaluation']}

name_for_cid[12277] = 'CardiacUltrasoundPericardiumandPleuraMeasurements'
cid_concepts[12277] = \
{'LN': ['ROIInternalDimensionByUS', 'ROIThicknessByUS'],
 'SCT': ['Radius',
         'MinorAxis',
         'Length',
         'Diameter',
         'MajorAxis',
         'Circumference']}

name_for_cid[9511] = 'GeneralExternalRadiotherapyProcedureTechniques'
cid_concepts[9511] = \
{'DCM': ['StaticBeam',
         'SlidingWindowBeam',
         'StepAndShootBeam',
         'ConformalArcBeam',
         'ArcBeam',
         'VMAT']}

name_for_cid[6] = 'TransducerOrientation'
cid_concepts[6] = \
{'SCT': ['FiveChamber',
         'LongAxis',
         'Parasagittal',
         'FourChamber',
         'OffAxis',
         'ShortAxis',
         'TwoChamber',
         'Oblique',
         'Sagittal',
         'Coronal',
         'Transverse',
         'Longitudinal']}

name_for_cid[7276] = 'MRDiffusionModelInputs'
cid_concepts[7276] = \
{'DCM': ['SourceImageDiffusionBValue']}

name_for_cid[3235] = 'PerfusionComparisonFindings'
cid_concepts[3235] = \
{'SCT': ['ResolutionOfIschemia',
         'MoreIschemia',
         'NewIschemia',
         'LessIschemia',
         'NoChange',
         'NewInfarction']}

name_for_cid[6037] = 'MammographyQuantitativeTemporalDifferenceType'
cid_concepts[6037] = \
{'SCT': ['DifferenceInSize',
         'DifferenceInLocation',
         'DifferenceInOpacity',
         'DifferenceInSpatialProximity',
         'DifferenceInNumberOfCalcifications']}

name_for_cid[640] = 'Fibrils'
cid_concepts[640] = \
{'DCM': ['HumanAlphaSynucleinPreformedFibrils',
         'MouseAlphaSynucleinPreformedFibrils',
         'HumanTauPreformedFibrils',
         'MouseTauPreformedFibrils']}

name_for_cid[210] = 'QualitativeEvaluationModifierTypes'
cid_concepts[210] = \
{'SCT': ['TopographicalModifier', 'Laterality']}

name_for_cid[4270] = 'OCTAProcessingAlgorithmFamilies'
cid_concepts[4270] = \
{'DCM': ['OCTACorrelationMapping',
         'OCTASpeckleVariance',
         'DopplerOCTA',
         'OCTAOneSidedRatio',
         'OCTAOneSidedRatio',
         'OCTAComplexVariance',
         'OCTAAmplitudeDecorrelation']}

name_for_cid[9507] = 'ExternalBodyModels'
cid_concepts[9507] = \
{'DCM': ['ExtendedPatientAnatomyModel', 'PatientAnatomyModel']}

name_for_cid[12261] = 'CardiacUltrasoundPulmonaryVein'
cid_concepts[12261] = \
{'DCM': ['Tau',
         'IndicatorMeanTransitTime',
         'RRInterval',
         'BloodVelocityPeak',
         'BloodVelocityMean',
         'BloodVelocityMinimum',
         'SystolicBloodVelocityMean',
         'SystolicBloodVelocityPeak',
         'DiastolicBloodVelocityMean',
         'DiastolicBloodVelocityPeak'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'PulmVeinADurationToMVADurationDifference',
        'PeakGradient',
        'EjectionTime',
        'PulmonaryVeinADurationMitralValveADurationRatio',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'ACInterval',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'VelocityOfFlowPropagation',
        'AliasVelocity',
        'PulmonaryVeinAVTIToMitralValveVTIRatio',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'TimeAveragedMeanVelocity'],
 'SCT': ['CardiovascularOrificeDiameter',
         'BloodPressure',
         'RegurgitantFraction',
         'CardiovascularOrificeArea']}

name_for_cid[3809] = 'AssociatedConditions'
cid_concepts[3809] = \
{'SCT': ['MarfanSyndrome', 'TraumaticAbnormality']}

name_for_cid[6021] = 'QuadrantLocationfromBIRADS®'
cid_concepts[6021] = \
{'SCT': ['UpperOuterQuadrantOfBreast',
         'LowerInnerQuadrantOfBreast',
         'LowerOuterQuadrantOfBreast',
         'UpperInnerQuadrantOfBreast']}

name_for_cid[7260] = 'DiffusionAcquisitionValueTypes'
cid_concepts[7260] = \
{'DCM': ['HARDI',
         'DKI',
         'DTI',
         'DSI',
         'LSDI',
         'SingleShotEPI',
         'MultipleShotEPI',
         'ParallelImaging']}

name_for_cid[13] = 'RadiographicContrastAgentIngredient'
cid_concepts[13] = \
{'SCT': ['Gadolinium',
         'Barium',
         'Iron',
         'Iodine',
         'CarbonDioxide',
         'Oxygen',
         'Air',
         'Water',
         'Xenon']}

name_for_cid[10030] = 'DetectorTypes'
cid_concepts[10030] = \
{'DCM': ['IndirectDetector', 'DirectDetector', 'StorageDetector', 'Film']}

name_for_cid[3736] = 'NYHAClassification'
cid_concepts[3736] = \
{'SCT': ['NYHAClassII', 'NYHAClassIII', 'NYHAClassI', 'NYHAClassIV']}

name_for_cid[6164] = 'TimeIntervals'
cid_concepts[6164] = \
{'DCM': ['LesserThan3MonthsAgo',
         '_4MonthsTo1YearAgo',
         'GreaterThan1YearAgo',
         'NotSure']}

name_for_cid[3448] = 'AirwayAssessment'
cid_concepts[3448] = \
{'DCM': ['AirwaySeverelyObstructed',
         'AirwayUnobstructed',
         'AirwayPartiallyObstructed']}

name_for_cid[247] = 'LateralityLeftRightOnly'
cid_concepts[247] = \
{'SCT': ['Right', 'Left']}

name_for_cid[617] = 'AirwayManagementMethodCodeTypeforSmallAnimalAnesthesia'
cid_concepts[617] = \
{'DCM': ['NoseCone', 'NasalCannula'],
 'SCT': ['ViaTracheostomy',
         'AnestheticFaceMask',
         'ArtificialRespiration',
         'ControlledVentilation',
         'OxygenTherapy',
         'IntubationOfRespiratoryTract',
         'LaryngealMaskAirwayLMA']}

name_for_cid[12236] = 'EchoAnatomicSites'
cid_concepts[12236] = \
{'SCT': ['RightVentricleOutflowTract',
         'PericardialEffusion',
         'VentricularSeptalDefect',
         'ThoracicAorticCoarctation',
         'MitralAnnulus',
         'Aorta',
         'MedialMitralAnnulus',
         'LeftVentricle',
         'AtrialSeptalDefect',
         'LateralMitralAnnulus',
         'MitralValve',
         'LeftVentricleOutflowTract',
         'AorticValveRing',
         'TricuspidAnnulus']}

name_for_cid[6133] = 'ChestQuantitativeTemporalDifferenceType'
cid_concepts[6133] = \
{'SCT': ['DifferenceInLocation', 'DifferenceInSize']}

name_for_cid[6099] = 'RacialGroup'
cid_concepts[6099] = \
{'NCIt': ['NativeHawaiianOrOtherPacificIslander'],
 'SCT': ['AmericanIndianOrAlaskaNative',
         'CaucasianRace',
         'AfricanRace',
         'AsianRace']}

name_for_cid[3761] = 'AntilipemicAgents'
cid_concepts[3761] = \
{'SCT': ['FishOils',
         'Statins',
         'Fibrate',
         'AnionExchangeResin',
         'BileAcidSequestrant']}

name_for_cid[10067] = 'RadiationAttenuatorMaterials'
cid_concepts[10067] = \
{'SCT': ['TinOrTinCompound',
         'CarbonFiber',
         'NiobiumOrNiobiumCompound',
         'TantalumOrTantalumCompound',
         'SilverOrSilverCompound',
         'EuropiumOrEuropiumCompound',
         'MolybdenumOrMolybdenumCompound',
         'LeadOrLeadCompound',
         'CopperOrCopperCompound',
         'RhodiumOrRhodiumCompound',
         'AluminumOrAluminumCompound']}

name_for_cid[7165] = 'AbstractSegmentationTypes'
cid_concepts[7165] = \
{'DCM': ['RegistrationInput', 'Background', 'SingleSubjectExtractedFromGroup'],
 'NCIt': ['ReferenceRegion'],
 'SCT': ['Normal', 'WasteMaterial', 'Tissue']}

name_for_cid[7470] = 'LinearMeasurements'
cid_concepts[7470] = \
{'DCM': ['Distance', 'Height', 'PathLength'],
 'SCT': ['DiameterOfCircumscribedCircle',
         'Radius',
         'Perimeter',
         'MinorAxis',
         'Width',
         'Diameter',
         'MajorAxis',
         'PerpendicularAxis',
         'LongAxis',
         'Depth',
         'ShortAxis',
         'Length',
         'Circumference']}

name_for_cid[12021] = 'FetalLongBoneAnatomicSites'
cid_concepts[12021] = \
{'SCT': ['Femur', 'Ulna', 'Radius', 'Clavicle', 'Tibia', 'Fibula']}

name_for_cid[7309] = 'HumanTraumaImplantComponentTypes'
cid_concepts[7309] = \
{'SCT': ['DHSPlate', 'Screw', 'BonePlate', 'BoneNail']}

name_for_cid[400] = 'AuditEventID'
cid_concepts[400] = \
{'DCM': ['ApplicationActivity',
         'ProcedureRecord',
         'PatientRecord',
         'NetworkEntry',
         'Import',
         'Export',
         'DICOMStudyDeleted',
         'DICOMInstancesTransferred',
         'DICOMInstancesAccessed',
         'UserAuthentication',
         'BeginTransferringDICOMInstances',
         'SecurityAlert',
         'AuditLogUsed',
         'Query',
         'OrderRecord']}

name_for_cid[4030] = 'CT,MRandPETAnatomyImaged'
cid_concepts[4030] = \
{'DCM': ['Phantom'],
 'SCT': ['IliacAndOrFemoralArtery',
         'BoneOfLowerLimb',
         'Ureter',
         'Sacrum',
         'SmallIntestine',
         'ParanasalSinus',
         'InternalAuditoryCanal',
         'WristJoint',
         'Thymus',
         'CarotidArtery',
         'Humerus',
         'Ilium',
         'JawRegion',
         'Calcaneus',
         'NasalBone',
         'Esophagus',
         'Mandible',
         'OpticCanal',
         'MastoidBone',
         'NeckChestAbdomenAndPelvis',
         'Thigh',
         'Spleen',
         'ChestAndAbdomen',
         'CoronaryArtery',
         'Trachea',
         'UpperLimb',
         'AnkleJoint',
         'Maxilla',
         'Cerebellum',
         'Mediastinum',
         'Hand',
         'UpperUrinaryTract',
         'Bronchus',
         'Pelvis',
         'SesamoidBonesOfFoot',
         'Eye',
         'Chest',
         'Prostate',
         'ElbowJoint',
         'AbdomenAndPelvis',
         'FacialBones',
         'Colon',
         'Sternum',
         'SubmandibularGland',
         'ThoracicSpine',
         'NeckAndChest',
         'Duodenum',
         'Gallbladder',
         'RenalArtery',
         'BileDuct',
         'UpperArm',
         'Extremity',
         'PancreaticDuctAndBileDuctSystems',
         'Larynx',
         'PulmonaryArtery',
         'Skull',
         'ThoracoLumbarSpine',
         'Liver',
         'Urethra',
         'Thyroid',
         'HipJoint',
         'Rib',
         'BoneOfUpperLimb',
         'Head',
         'AdrenalGland',
         'AcromioclavicularJoint',
         'Clavicle',
         'ParotidGland',
         'SternoclavicularJoint',
         'Breast',
         'EsophagusStomachAndDuodenum',
         'Neck',
         'Finger',
         'CranialVenousSystem',
         'Stomach',
         'ThoracicAorta',
         'Coccyx',
         'Abdomen',
         'HeadAndNeck',
         'Uterus',
         'TemporomandibularJoint',
         'ApexOfLung',
         'Ileum',
         'Testis',
         'PelvisAndLowerExtremities',
         'Bladder',
         'Knee',
         'LargeIntestine',
         'Forearm',
         'EyeRegion',
         'Spine',
         'LumboSacralSpine',
         'Foot',
         'NeckChestAndAbdomen',
         'SacroiliacJoint',
         'Anus',
         'MuscleOfUpperLimb',
         'EntireBody',
         'Thumb',
         'LumbarSpine',
         'Zygoma',
         'SellaTurcica',
         'UterusAndFallopianTubes',
         'LowerLimb',
         'LowerLeg',
         'AorticArch',
         'VertebralColumnAndCranium',
         'Jejunum',
         'ChestAbdomenAndPelvis',
         'AbdominalAorta',
         'Heart',
         'TarsalJoint',
         'Toe',
         'Shoulder',
         'Fibula',
         'OrbitalStructure',
         'CircleOfWillis',
         'Brain',
         'MuscleOfLowerLimb',
         'Parathyroid',
         'CervicoThoracicSpine',
         'Femur',
         'Rectum',
         'Pancreas',
         'Scapula',
         'Kidney',
         'Patella',
         'CervicalSpine']}

name_for_cid[8101] = 'ContainerTypes'
cid_concepts[8101] = \
{'SCT': ['SpecimenContainer',
         'TissueMicroarrayCassette',
         'ElectronMicroscopyGrid',
         'TissueCassette',
         'MicroscopeSlide',
         'SpecimenVial',
         'SpecimenWell']}

name_for_cid[68] = 'ImagingAgentAdministrationPharmaceuticalUnitofPresentation'
cid_concepts[68] = \
{'SCT': ['ParenteralEnteralSolutionBag', 'Syringe', 'Cartridge', 'Bottle']}

name_for_cid[3608] = 'AtrialSourceLocations'
cid_concepts[3608] = \
{'SCT': ['RightAtrium',
         'PulmonaryCapillaryWedge',
         'Baffle',
         'CoronarySinus',
         'LeftAtrium',
         'CommonAtrium',
         'JuxtaposedAppendage',
         'SystemicVenousAtrium',
         'PulmonaryVenousAtrium',
         'PulmonaryArteryWedge',
         'PulmonaryChamberInCorTriatriatum']}

name_for_cid[3472] = 'CardiacContractionPhase'
cid_concepts[3472] = \
{'SCT': ['Diastolic', 'Systolic']}

name_for_cid[7061] = 'ModelDocumentTitles'
cid_concepts[7061] = \
{'DCM': ['US3DCAMModel',
         'SourceImage',
         'MixedModality3DCAMModel',
         'LaserScanning3DCAMModel',
         'PhotogrammetricImaging3DCAMModel'],
 'LN': ['CT3DCAMModel', 'MR3DCAMModel']}

name_for_cid[6109] = 'RadiographicAnatomyFindingorFeature'
cid_concepts[6109] = \
{'BARI': ['_1stDiagonalCoronaryArteryLaterals',
          '_1stMarginalCoronaryArteryLaterals',
          '_2ndDiagonalCoronaryArteryLaterals',
          'RamusLaterals',
          '_2ndMarginalCoronaryArteryLaterals',
          '_3rdDiagonalCoronaryArteryLaterals',
          '_3rdMarginalCoronaryArteryLaterals',
          'PosteriorDescendingSeptalPerforators'],
 'DCM': ['AorticKnob',
         'ArchOfTheAzygosVein',
         'SecondaryPulmonaryLobule',
         'Hiatus',
         'AzygoesophagealRecessInterface',
         'ParaspinalLine',
         'PosteriorTrachealStripe',
         'RightTrachealStripe',
         'Stripe',
         'Lucency',
         'MidlungWindow',
         'CarinaAngle',
         'CentrilobularStructures',
         'AnteriorJunctionLine',
         'PosteriorJunctionLine',
         'Interface',
         'Line'],
 'SCT': ['SubclavianVein',
         'SternocleidomastoidMuscle',
         'PulmonaryValve',
         'ProximalRightCoronaryArtery',
         'RightPosteriorAVCoronaryArtery',
         'FascialLayer',
         'BrachialPlexus',
         'ErectorSpinaeMuscle',
         'AorticValve',
         'MidCircumflexCoronaryArtery',
         'InferiorVenaCava',
         'IntermediateArteryRamus',
         'ThymusGland',
         'SubcostalMuscle',
         'Humerus',
         'SubscapularisMuscle',
         'LymphNode',
         'TrapeziusMuscle',
         'DistalRightCoronaryArtery',
         'IntercostalArtery',
         'Esophagus',
         'LeftAnteriorDescendingCoronaryArtery',
         'CommonCarotidArtery',
         'InnermostIntercostalMuscles',
         'TeresMajorMuscle',
         'LeftMainBronchus',
         '_3rdMarginalCoronaryArtery',
         'PectoralisMinorMuscle',
         'MidRightCoronaryArtery',
         'EsophagealArtery',
         'PulmonaryVein',
         'LatissimusDorsiMuscle',
         'Trachea',
         'SubclavianArtery',
         'LeftPosterolateralCircumflexCoronaryArtery',
         'EsophagealHiatus',
         '_3rdLeftPosterolateralCoronaryArtery',
         'Bronchus',
         'PectoralisMajorMuscle',
         'DorsalScapularArtery',
         'SuperiorVenaCava',
         'Sternum',
         'CostalCartilage',
         'BrachiocephalicTrunk',
         '_2ndRightPosterolateralCoronaryArtery',
         'LeftPosteriorDescendingCircumflexCoronaryArtery',
         '_1stMarginalCoronaryArtery',
         'PulmonaryTrunk',
         'AzygosVein',
         'PulmonaryArtery',
         'BrachialArtery',
         '_2ndDiagonalCoronaryArtery',
         'DistalCircumflexCoronaryArtery',
         'MidLeftAnteriorDescendingCoronaryArtery',
         'Thyroid',
         'Rib',
         'CoronaryArteryGraft',
         'TeresMinorMuscle',
         'MitralValve',
         'AcromioclavicularJoint',
         'LongissimusMuscle',
         'SerratusAnteriorMuscle',
         'SupraspinatusMuscle',
         'Clavicle',
         'AorticIsthmus',
         'AxillaryFascia',
         'Vertebra',
         'ExternalIntercostalMuscle',
         '_1stLeftPosterolateralCoronaryArtery',
         'InfraspinatusMuscle',
         'TricuspidValve',
         'CircumflexCoronaryArtery',
         'Diaphragm',
         'FissureOfLung',
         'LevatoresCostarumMuscles',
         'SuperiorPhrenicArtery',
         'DistalLeftAnteriorDescendingCoronaryArtery',
         'AtrialSeptalDefect',
         'Ventricle',
         'RightCoronaryArteryOstium',
         'AxillaryArtery',
         'DescendingAorta',
         'MarginalCoronaryArtery',
         'Atrium',
         'InternalThoracicArtery',
         'VertebralArtery',
         'ProximalLeftAnteriorDescendingCoronaryArtery',
         '_3rdDiagonalCoronaryArtery',
         'SpinalisMuscle',
         'InternalJugularVein',
         'Spine',
         'BronchialArtery',
         'ChordaeTendineaeCordis',
         'RightMainBronchus',
         'ScalenousAnteriorMuscle',
         'AVGrooveContinuationOfCircumflexArtery',
         'TrabeculaeCarnae',
         '_3rdRightPosterolateralCoronaryArtery',
         'ThyrocervicalTrunk',
         'InferiorPhrenicArtery',
         'BrachiocephalicVein',
         'HilumOfLung',
         'ProximalCircumflexCoronaryArtery',
         '_2ndMarginalCoronaryArtery',
         'ThoracicDuct',
         'Carina',
         'AorticArch',
         'CarotidBody',
         'Heart',
         'AirwayStructure',
         'TransversusThoracis',
         'DeltoidMuscle',
         'InterventricularSeptum',
         'LeftMainCoronaryArtery',
         '_2ndLeftPosterolateralCoronaryArtery',
         'PosteriorDescendingRightCoronaryArtery',
         'InternalIntercostalMuscle',
         '_1stSeptalCoronaryArtery',
         'RightCoronaryArtery',
         'AscendingAorta',
         '_1stDiagonalCoronaryArtery',
         'PericardiophrenicArtery',
         'InteratrialSeptum',
         'CostocervicalTrunk',
         'LeftMainCoronaryArteryOstium',
         'AxillaryVein',
         'LigamentumArteriosum',
         'Scapula',
         'SegmentOfLung',
         'IliocostalisMuscle',
         '_1stRightPosterolateralCoronaryArtery',
         'LobeOfLung']}

name_for_cid[91] = 'FunctionalConditionPresentDuringAcquisition'
cid_concepts[91] = \
{'DCM': ['DuringVoiding', 'PostVoiding', 'PriorToVoiding'],
 'SCT': ['WeightBearing',
         'DrugInfusion',
         'LegUp',
         'ValsalvaManeuver',
         'Phonation',
         'PostVolumeChallenge',
         'HeldVentilation',
         'HeadUp',
         'VagalStimulation',
         'Pacing',
         'NegativeLowerBodyPressure',
         'Handgrip',
         'HeldInspiration',
         'ExerciseChallenge']}

name_for_cid[3560] = 'BloodPressureMethods'
cid_concepts[3560] = \
{'SCT': ['BloodPressureCuffMethod']}

name_for_cid[3833] = 'CardiacEjectionFraction'
cid_concepts[3833] = \
{'LN': ['LeftVentricularEjectionFractionByCT',
        'RightVentricularEjectionFractionByMR',
        'LeftVentricularEjectionFractionByMR',
        'RightVentricularEjectionFractionByCT']}

name_for_cid[7466] = 'PETRegionofInterestMeasurements'
cid_concepts[7466] = \
{'DCM': ['MetabolicVolume',
         'TotalLesionGlycolysis',
         'Glycolysis',
         'TotalLesionProliferation',
         'ProliferativeActivity',
         'StandardizedAddedMetabolicActivity',
         'StandardizedAddedMetabolicActivityBackground',
         'LesionToBackgroundSUVRatio',
         'BackgroundForLesionToBackgroundSUVRatio']}

name_for_cid[7036] = 'ImageQualityAssessment'
cid_concepts[7036] = \
{'RADLEX': ['NonDiagnosticQuality', 'LimitedQuality', 'DiagnosticQuality']}

name_for_cid[3425] = 'InterventionParameters'
cid_concepts[3425] = \
{'SCT': ['AblationPower',
         'AblationFrequency',
         'AngioplastyInflationPressure',
         'AngioplastyInflationDuration',
         'RotationalAtherectomySpeed',
         'DeliveredRadiationDose']}

name_for_cid[29] = 'AcquisitionModality'
cid_concepts[29] = \
{'DCM': ['OphthalmicVisualField',
         'DigitalRadiography',
         'OphthalmicTomography',
         'BoneMineralDensitometry',
         'Mammography',
         'SlideMicroscopy',
         'Ultrasound',
         'OphthalmicPhotography',
         'IntravascularOpticalCoherenceTomography',
         'MagneticResonance',
         'Electrocardiography',
         'GeneralMicroscopy',
         'IntraOralRadiography',
         'XRayAngiography',
         'RTImage',
         'VisualAcuity',
         'ExternalCameraPhotography',
         'RespiratoryWaveform',
         'IntravascularUltrasound',
         'ComputedRadiography',
         'Endoscopy',
         'Autorefraction',
         'ComputedTomography',
         'OpticalSurfaceScanner',
         'Thermography',
         'Lensometry',
         'RadiographicImaging',
         'Radiofluoroscopy',
         'Keratometry',
         'OphthalmicTomographyEnFace',
         'HemodynamicWaveform',
         'OphthalmicAxialMeasurements',
         'NuclearMedicine',
         'OpticalCoherenceTomography',
         'UltrasoundBoneDensitometry',
         'Diaphanography',
         'PositronEmissionTomography',
         'BiomagneticImaging',
         'CardiacElectrophysiology',
         'LaserScan',
         'PanoramicXRay',
         'SubjectiveRefraction',
         'OphthalmicMapping',
         'OphthalmicTomographyBScanVolumeAnalysis']}

name_for_cid[9301] = 'ModalityPPSDiscontinuationReasons'
cid_concepts[9301] = \
{'DCM': ['PatientDied',
         'PatientConditionPreventedContinuing',
         'PatientAllergicToMediaContrast',
         'IncorrectWorklistEntrySelected',
         'IncorrectProcedureOrdered',
         'DiscontinuedForUnspecifiedReason',
         'EquipmentFailure',
         'IncorrectSideOrdered',
         'DoctorCanceledProcedure',
         'NursingUnitCancel',
         'DuplicateOrder',
         'ChangeOfProcedureForCorrectCharging',
         'PatientPregnant',
         'PatientDidNotArrive',
         'PatientTakenForTreatmentOrSurgery',
         'PatientRefusedToContinueProcedure',
         'EquipmentChange'],
 'SCT': ['RadiopharmaceuticalAdverseReaction', 'InjectionSiteExtravasation']}

name_for_cid[800] = 'ProtocolAssertionCodes'
cid_concepts[800] = \
{'DCM': ['ApprovedForUseInTheClinicalTrial',
         'IneligibleForReimbursement',
         'ApprovedForUseAtTheInstitution',
         'EligibleForReimbursementOnPerPatientBasis',
         'DisapprovedForUseOnPregnantPatients',
         'AppropriateForTheDevice',
         'ApprovedForUseOnPregnantPatients',
         'OutsideOperationalLimitsOfTheDevice',
         'OptimizedForTheDeviceInstance',
         'InappropriateForTheDevice',
         'InsideOperationalLimitsOfTheDevice',
         'DisapprovedForAnyUse',
         'NotOptimizedForTheDeviceInstance',
         'InconsistentWithLabelingOfTheDevice',
         'ApprovedForExperimentalUse',
         'InappropriateForTheIndications',
         'DeprecatedProtocol',
         'ConsistentWithLabelingOfTheDevice',
         'DisapprovedForUseInTheClinicalTrial',
         'EligibleForReimbursement',
         'AppropriateForTheIndications',
         'DisapprovedForUseAtTheInstitution',
         'DisapprovedForExperimentalUse']}

name_for_cid[4026] = 'PrimaryAnatomicStructureforIntraoralandCraniofacialRadiographyTeeth'
cid_concepts[4026] = \
{'SCT': ['DeciduousMandibularRightLateralIncisorTooth',
         'SupernumeraryPermanentMandibularLeftThirdMolarTooth',
         'SupernumeraryDeciduousMaxillaryLeftFirstMolarTooth',
         'SupernumeraryPermanentMaxillaryLeftSecondPremolarTooth',
         'SupernumeraryDeciduousMandibularRightCanineTooth',
         'MaxillaryLeftSecondPremolarTooth',
         'MandibularRightLateralIncisorTooth',
         'DeciduousMandibularRightCanineTooth',
         'MandibularRightFirstMolarTooth',
         'MaxillaryLeftSecondMolarTooth',
         'SupernumeraryPermanentMaxillaryRightFirstPremolarTooth',
         'MaxillaryRightLateralIncisorTooth',
         'SupernumeraryPermanentMaxillaryLeftThirdMolarTooth',
         'DeciduousMandibularLeftLateralIncisorTooth',
         'MaxillaryRightFirstMolarTooth',
         'MandibularLeftCentralIncisorTooth',
         'SupernumeraryPermanentMandibularRightFirstMolarTooth',
         'SupernumeraryDeciduousMaxillaryLeftCentralIncisorTooth',
         'MaxillaryRightFirstPremolarTooth',
         'DeciduousMaxillaryRightLateralIncisorTooth',
         'SupernumeraryPermanentMaxillaryRightCentralIncisorTooth',
         'MaxillaryLeftLateralIncisorTooth',
         'DeciduousMaxillaryLeftCentralIncisorTooth',
         'DeciduousMaxillaryRightCanineTooth',
         'SupernumeraryPermanentMandibularRightCentralIncisorTooth',
         'MandibularRightCentralIncisorTooth',
         'SupernumeraryPermanentMandibularLeftFirstMolarTooth',
         'SupernumeraryPermanentMaxillaryLeftSecondMolarTooth',
         'DeciduousMandibularLeftCanineTooth',
         'SupernumeraryDeciduousMandibularRightSecondMolarTooth',
         'DeciduousMandibularRightSecondMolarTooth',
         'SupernumeraryDeciduousMandibularLeftCanineTooth',
         'DeciduousMandibularRightFirstMolarTooth',
         'SupernumeraryPermanentMandibularLeftCentralIncisorTooth',
         'SupernumeraryPermanentMaxillaryRightFirstMolarTooth',
         'SupernumeraryDeciduousMaxillaryRightSecondMolarTooth',
         'MandibularLeftSecondMolarTooth',
         'MaxillaryLeftCanineTooth',
         'MandibularLeftFirstPremolarTooth',
         'SupernumeraryPermanentMandibularLeftSecondMolarTooth',
         'MandibularLeftFirstMolarTooth',
         'MandibularRightFirstPremolarTooth',
         'DeciduousMaxillaryLeftLateralIncisorTooth',
         'MandibularLeftLateralTooth',
         'DeciduousMandibularLeftCentralIncisorTooth',
         'SupernumeraryDeciduousMaxillaryRightLateralIncisorTooth',
         'MaxillaryLeftCentralIncisorTooth',
         'SupernumeraryDeciduousMandibularLeftCentralIncisorTooth',
         'SupernumeraryPermanentMandibularRightThirdMolarTooth',
         'SupernumeraryPermanentMaxillaryLeftLateralIncisorTooth',
         'DeciduousMandibularLeftFirstMolarTooth',
         'DeciduousMandibularRightCentralIncisorTooth',
         'SupernumeraryPermanentMandibularRightCanineTooth',
         'SupernumeraryPermanentMandibularLeftFirstPremolarTooth',
         'DeciduousMaxillaryRightCentralIncisorTooth',
         'SupernumeraryDeciduousMaxillaryLeftLateralIncisorTooth',
         'SupernumeraryPermanentMaxillaryRightThirdMolarTooth',
         'MaxillaryRightCentralIncisorTooth',
         'MandibularRightCanineTooth',
         'DeciduousMaxillaryLeftSecondMolarTooth',
         'SupernumeraryPermanentMaxillaryLeftCentralIncisorTooth',
         'SupernumeraryPermanentMandibularLeftLateralIncisorTooth',
         'MaxillaryRightSecondPremolarTooth',
         'SupernumeraryPermanentMandibularRightLateralIncisorTooth',
         'SupernumeraryPermanentMaxillaryRightSecondMolarTooth',
         'SupernumeraryDeciduousMaxillaryRightFirstMolarTooth',
         'SupernumeraryPermanentMandibularLeftSecondPremolarTooth',
         'SupernumeraryDeciduousMaxillaryLeftSecondMolarTooth',
         'MandibularLeftThirdMolarTooth',
         'SupernumeraryDeciduousMandibularRightLateralIncisorTooth',
         'DeciduousMaxillaryLeftFirstMolarTooth',
         'SupernumeraryPermanentMaxillaryLeftFirstPremolarTooth',
         'SupernumeraryPermanentMaxillaryRightCanineTooth',
         'SupernumeraryDeciduousMandibularLeftFirstMolarTooth',
         'SupernumeraryDeciduousMaxillaryRightCentralIncisorTooth',
         'SupernumeraryPermanentMandibularRightSecondPremolarTooth',
         'MaxillaryLeftFirstPremolarTooth',
         'DeciduousMandibularLeftSecondMolarTooth',
         'MandibularLeftSecondPremolarTooth',
         'SupernumeraryPermanentMaxillaryLeftCanineTooth',
         'MaxillaryLeftThirdMolarTooth',
         'MandibularLeftCanineTooth',
         'MaxillaryRightCanineTooth',
         'MandibularRightSecondPremolarTooth',
         'SupernumeraryDeciduousMandibularLeftSecondMolarTooth',
         'SupernumeraryDeciduousMandibularLeftLateralIncisorTooth',
         'SupernumeraryPermanentMandibularRightFirstPremolarTooth',
         'SupernumeraryPermanentMandibularLeftCanineTooth',
         'DeciduousMaxillaryRightFirstMolarTooth',
         'SupernumeraryDeciduousMaxillaryLeftCanineTooth',
         'SupernumeraryPermanentMaxillaryLeftFirstMolarTooth',
         'MaxillaryRightThirdMolarTooth',
         'SupernumeraryDeciduousMandibularRightFirstMolarTooth',
         'SupernumeraryPermanentMaxillaryRightLateralIncisorTooth',
         'MaxillaryRightSecondMolarTooth',
         'MandibularRightSecondMolarTooth',
         'DeciduousMaxillaryRightSecondMolarTooth',
         'MaxillaryLeftFirstMolarTooth',
         'SupernumeraryPermanentMaxillaryRightSecondPremolarTooth',
         'SupernumeraryDeciduousMaxillaryRightCanineTooth',
         'SupernumeraryPermanentMandibularRightSecondMolarTooth',
         'SupernumeraryDeciduousMandibularRightCentralIncisorTooth',
         'MandibularRightThirdMolarTooth',
         'DeciduousMaxillaryLeftCanineTooth']}

name_for_cid[9300] = 'ProcedureDiscontinuationReasons'
cid_concepts[9300] = \
{'DCM': ['WorkitemExpired',
         'WorkitemAssignmentRejectedByAssignedResource',
         'ExtravasationVisibleInImage',
         'ResourcePreEmpted',
         'MediaFailure',
         'ObjectSetIncomplete',
         'ObjectTypesNotSupported',
         'ObjectsIncorrectlyFormatted',
         'DiscontinuedProcedureStepReschedulingRecommended',
         'DiscontinuedProcedureStepRescheduled',
         'ResourceInadequate',
         'PatientConditionPreventedContinuing',
         'IncorrectWorklistEntrySelected',
         'DiscontinuedForUnspecifiedReason',
         'IncorrectSideOrdered',
         'NursingUnitCancel',
         'DuplicateOrder',
         'EquipmentChange',
         'PatientDied',
         'PatientAllergicToMediaContrast',
         'IncorrectProcedureOrdered',
         'EquipmentFailure',
         'DoctorCanceledProcedure',
         'ChangeOfProcedureForCorrectCharging',
         'PatientPregnant',
         'PatientDidNotArrive',
         'PatientTakenForTreatmentOrSurgery',
         'PatientRefusedToContinueProcedure'],
 'SCT': ['RadiopharmaceuticalAdverseReaction',
         'PallorPaleComplexion',
         'InjectionSiteNerveDamage',
         'InjectionSiteFibrosis',
         'InjectionSiteExtravasation',
         'Cough',
         'ChillsAndFever',
         'InjectionSiteHypersensitivity',
         'InjectionSitePigmentationChange',
         'InjectionSiteScar',
         'InjectionSiteIrritation',
         'Sweating',
         'InjectionSiteAnesthesia',
         'FacialSwelling',
         'InjectionSiteInduration',
         'InjectionSiteUrticaria',
         'Bronchospasm',
         'InjectionSiteMalabsorption',
         'InjectionSiteSterileAbscess',
         'Headache',
         'SensationOfBeingWarm',
         'TachycardiaBradycardia',
         'InjectionSiteNecrosis',
         'Anxiety',
         'InjectionSiteDermatitis',
         'InjectionSiteDisorder',
         'Shaking',
         'InjectionSiteBruising',
         'DrugInducedFlushing',
         'CardiacArrhythmia',
         'LaryngealEdema',
         'InjectionSiteThrombosis',
         'InjectionSiteAbscess',
         'DrugInducedDizziness',
         'InjectionSiteReaction',
         'CardiorespiratoryArrest',
         'InjectionSiteCyst',
         'Hypertension',
         'InjectionSiteEdema',
         'Dyspnea',
         'InjectionSiteAtrophy',
         'DrugInducedHypotension',
         'InjectionSiteBurning',
         'InjectionSiteHemorrhage',
         'InjectionSiteUlcer',
         'EpilepticConvulsions',
         'Itching',
         'DrugRash',
         'InjectionSiteInflammation',
         'InjectionSiteInfection',
         'Hypotension',
         'DiffuseInflammatoryErythema',
         'TasteAndSenseAltered',
         'InjectionSiteParesthesia',
         'NasalCongestion',
         'NoMotorResponseToCommand',
         'InjectionSitePain',
         'DrugInducedNauseaAndVomiting',
         'InjectionSiteMass',
         'InjectionSiteGranuloma']}

name_for_cid[3832] = 'StenosisGrade'
cid_concepts[3832] = \
{'SCT': ['Moderate', 'Mild', 'Severe']}

name_for_cid[7467] = 'GrayLevelCooccurrenceMatrixMeasurements'
cid_concepts[7467] = \
{'DCM': ['RootAngularSecondMomentOfGLCM']}

name_for_cid[6108] = 'ChestAnatomicStructureAbnormalDistribution'
cid_concepts[6108] = \
{'DCM': ['ArchitecturalDistortion', 'MosaicPerfusion', 'Pleonemia', 'Oligemia'],
 'SCT': ['AirTrapping']}

name_for_cid[12124] = 'RenalRatios'
cid_concepts[12124] = \
{'LN': ['RenalArteryAortaVelocityRatio']}

name_for_cid[7060] = 'EncapsulatedDocumentSourcePurposesofReference'
cid_concepts[7060] = \
{'DCM': ['SourceMeasurement',
         'SourceRawData',
         'SourceImage',
         'SourceReport',
         'SourceRealWorldValueMap']}

name_for_cid[69] = 'ImagingAgentAdministrationConsumables'
cid_concepts[69] = \
{'SCT': ['ContrastMediumInjectionSystemManifoldKit',
         'Syringe',
         'Catheter',
         'Needle',
         'TubeDevice',
         'Bottle']}

name_for_cid[3609] = 'VentricularSourceLocations'
cid_concepts[3609] = \
{'SCT': ['RightVentricleOutflowTract',
         'LeftVentricleInflow',
         'RightVentricle',
         'LeftVentricle',
         'RightVentricleInflow',
         'CommonVentricle',
         'LeftVentricleApex',
         'RightVentricleApex',
         'LeftVentricleOutflowTract']}

name_for_cid[3520] = 'BloodSourceType'
cid_concepts[3520] = \
{'SCT': ['SystemicArteryBlood',
         'PulmonaryVeinBlood',
         'MixedVenousBlood',
         'PulmonaryArteryBlood']}

name_for_cid[3465] = 'QAReferenceMethods'
cid_concepts[3465] = \
{'DCM': ['CurveFittedReference',
         'InterpolatedLocalReference',
         'MeanLocalReference']}

name_for_cid[12020] = 'FetalBiometryAnatomicSites'
cid_concepts[12020] = \
{'SCT': ['Foot',
         'Femur',
         'Cerebellum',
         'Skull',
         'Thorax',
         'Abdomen',
         'CisternaMagna',
         'Kidney',
         'Trunk']}

name_for_cid[7308] = 'HumanHipImplantComponentTypes'
cid_concepts[7308] = \
{'DCM': ['FemoralStem',
         'FemoralStemDistalComponent',
         'FemoralHeadConeTaperComponent',
         'FemoralStemProximalComponent',
         'FemoralStemComponent',
         'NeckComponent',
         'DistalCentralizer',
         'MonoblockStem',
         'AcetabularCupShell',
         'ProstheticShaftAugment',
         'AcetabularCupInsert',
         'FemoralHeadResurfacingComponent',
         'AcetabularCupMonoblock',
         'FemoralHeadBallComponent'],
 'SCT': ['FemoralHeadProsthesis']}

name_for_cid[401] = 'AuditEventTypeCode'
cid_concepts[401] = \
{'DCM': ['AuditRecordingStopped',
         'UseOfRestrictedFunction',
         'SoftwareConfiguration',
         'HardwareConfiguration',
         'RemoteServiceOperationStarted',
         'EmergencyOverrideStopped',
         'UserSecurityAttributesChanged',
         'SecurityRolesChanged',
         'ObjectSecurityAttributesChanged',
         'AuditRecordingStarted',
         'Login',
         'AuthorizationDecision',
         'ApplicationStop',
         'AuthenticationDecision',
         'ApplicationStart',
         'LocalServiceOperationStopped',
         'LocalServiceOperationStarted',
         'RemoteServiceOperationStopped',
         'SecurityConfiguration',
         'NetworkConfiguration',
         'EmergencyOverrideStarted',
         'NodeAuthentication',
         'Detach',
         'AccessControlDecision',
         'Attach',
         'SessionStop',
         'Logout',
         'SessionStart']}

name_for_cid[4031] = 'CommonAnatomicRegions'
cid_concepts[4031] = \
{'DCM': ['Phantom'],
 'SCT': ['BoneOfLowerLimb',
         'Ureter',
         'Sacrum',
         'SmallIntestine',
         'ParanasalSinus',
         'InternalAuditoryCanal',
         'WristJoint',
         'Humerus',
         'Ilium',
         'JawRegion',
         'Calcaneus',
         'NasalBone',
         'Esophagus',
         'Mandible',
         'OpticCanal',
         'MastoidBone',
         'NeckChestAbdomenAndPelvis',
         'Thigh',
         'ChestAndAbdomen',
         'Trachea',
         'UpperLimb',
         'AnkleJoint',
         'Maxilla',
         'Mediastinum',
         'Hand',
         'UpperUrinaryTract',
         'Bronchus',
         'Pelvis',
         'SesamoidBonesOfFoot',
         'Eye',
         'Chest',
         'Prostate',
         'ElbowJoint',
         'AbdomenAndPelvis',
         'FacialBones',
         'Colon',
         'Sternum',
         'SubmandibularGland',
         'ThoracicSpine',
         'NeckAndChest',
         'Duodenum',
         'Gallbladder',
         'BileDuct',
         'UpperArm',
         'Extremity',
         'PancreaticDuctAndBileDuctSystems',
         'Larynx',
         'Skull',
         'ThoracoLumbarSpine',
         'Urethra',
         'HipJoint',
         'Rib',
         'BoneOfUpperLimb',
         'Head',
         'AcromioclavicularJoint',
         'Clavicle',
         'ParotidGland',
         'SternoclavicularJoint',
         'Breast',
         'EsophagusStomachAndDuodenum',
         'Neck',
         'Finger',
         'Stomach',
         'Coccyx',
         'Abdomen',
         'HeadAndNeck',
         'TemporomandibularJoint',
         'ApexOfLung',
         'Ileum',
         'PelvisAndLowerExtremities',
         'Bladder',
         'Knee',
         'LargeIntestine',
         'Forearm',
         'EyeRegion',
         'Spine',
         'LumboSacralSpine',
         'Foot',
         'NeckChestAndAbdomen',
         'SacroiliacJoint',
         'Anus',
         'MuscleOfUpperLimb',
         'EntireBody',
         'Thumb',
         'LumbarSpine',
         'Zygoma',
         'SellaTurcica',
         'UterusAndFallopianTubes',
         'LowerLimb',
         'LowerLeg',
         'VertebralColumnAndCranium',
         'Jejunum',
         'ChestAbdomenAndPelvis',
         'Heart',
         'TarsalJoint',
         'Toe',
         'Shoulder',
         'Fibula',
         'OrbitalStructure',
         'MuscleOfLowerLimb',
         'CervicoThoracicSpine',
         'Femur',
         'Rectum',
         'Scapula',
         'Patella',
         'CervicalSpine']}

name_for_cid[7021] = 'MeasurementReportDocumentTitles'
cid_concepts[7021] = \
{'DCM': ['ImagingMeasurementReport',
         'OncologyMeasurementReport',
         'DynamicContrastMRMeasurementReport',
         'PETMeasurementReport']}

name_for_cid[7471] = 'AreaMeasurements'
cid_concepts[7471] = \
{'SCT': ['Area', 'AreaOfDefinedRegion']}

name_for_cid[6132] = 'ChestCalcificationDescriptor'
cid_concepts[6132] = \
{'DCM': ['Target', 'Fibrocalcific', 'Flocculent'],
 'SCT': ['Nodular',
         'Ossification',
         'Laminated',
         'CoarsePopcornLikeCalcification',
         'EggshellCalcification']}

name_for_cid[6098] = 'ClinicalCourseofDisease'
cid_concepts[6098] = \
{'NCIt': ['NoEvidenceOfDisease', 'RecurrentDisease']}

name_for_cid[3760] = 'HypertensionTherapy'
cid_concepts[3760] = \
{'SCT': ['BetaBlocker',
         'Diuretic',
         'AngiotensinIIReceptorAntagonist',
         'ACEInhibitor',
         'CalciumChannelBlocker',
         'NitrateVasodilator']}

name_for_cid[10066] = 'AttenuatorCategory'
cid_concepts[10066] = \
{'DCM': ['BeamBlock',
         'XRayFilters',
         'Table',
         'TableCore',
         'PatientSupport',
         'TablePad',
         'TableOuterLiner'],
 'SCT': ['XRayShield', 'ShieldingBlock']}

name_for_cid[3019] = 'CardiovascularAnatomicLocationModifiers'
cid_concepts[3019] = \
{'DCM': ['GraftToCitedSegmentDistalSection',
         'GraftToCitedSegmentProximalSection',
         'GraftToCitedSegmentMidSection',
         'Median',
         'AneurysmOnCitedVessel'],
 'NCIt': ['Middle'],
 'SCT': ['VenousGraftToCitedSegment',
         'EntireVessel',
         'Central',
         'Right',
         'Posterior',
         'Left',
         'Superior',
         'ArterialGraftToCitedSegment',
         'Distal',
         'Proximal',
         'MidLongitudinal',
         'Lateral',
         'Inferior',
         'GraftToCitedSegmentDistalAnastomosis',
         'Ostium',
         'Anterior',
         'GraftToCitedSegmentBody',
         'GraftToCitedSegmentProximalAnastomosis']}

name_for_cid[246] = 'RelativeLaterality'
cid_concepts[246] = \
{'SCT': ['Contralateral', 'Ipsilateral']}

name_for_cid[616] = 'AnesthesiaMaintenanceCodeTypefromAnesthesiaQualityInitiative(AQI)'
cid_concepts[616] = \
{'SCT': ['InhalationAnesthesiaSystemClosedRebreathingPrimaryAgent',
         'InhalationAnesthesiaSystemClosedNoRebreathingPrimaryAgent']}

name_for_cid[12237] = 'EchocardiographyAnatomicSiteModifiers'
cid_concepts[12237] = \
{'SCT': ['RightLowerSegment',
         'RightUpperSegment',
         'LeftLowerSegment',
         'LeftUpperSegment']}

name_for_cid[6020] = 'QuadrantLocation'
cid_concepts[6020] = \
{'SCT': ['UpperOuterQuadrantOfBreast',
         'LowerInnerQuadrantOfBreast',
         'LowerOuterQuadrantOfBreast',
         'UpperInnerQuadrantOfBreast']}

name_for_cid[7261] = 'DiffusionModelValueTypes'
cid_concepts[7261] = \
{'DCM': ['SingleTensor',
         'MultiTensor',
         'ModelFree',
         'CHARMED',
         'DSI',
         'DOT',
         'PAS',
         'SphericalDeconvolution']}

name_for_cid[12] = 'RadiographicContrastAgent'
cid_concepts[12] = \
{'DCM': ['NonIonicIodinatedContrastAgent'],
 'SCT': ['MeglumineIodipamide',
         'Iophendylate',
         'Iodipamide',
         'Ioxilan',
         'Bunamiodyl',
         'Ioversol',
         'Gadodiamide',
         'Metrizamide',
         'Iodixanol',
         'Ipodate',
         'Oxygen',
         'BariumSulfate',
         'SodiumTyropanoate',
         'Iohexol',
         'Water',
         'GadopentetateDimeglumine',
         'SodiumAcetrizoate',
         'Propyliodone',
         'ContrastAgent',
         'IodoalphionicAcid',
         'Iodopyracet',
         'SodiumDiprotrizoate',
         'Iothalamate',
         'Iodophthalein',
         'Ioxaglate',
         'MeglumineDiatrizoate',
         'MangafodipirTrisodium',
         'IodamideMeglumine',
         'SodiumIodipamide',
         'SodiumDiatrizoate',
         'Diatrizoate',
         'NonRadiopaqueMedium',
         'ChloriodizedOil',
         'IophenoxicAcid',
         'Air',
         'Iopamidol',
         'Iopromide',
         'IopanoicAcid',
         'Gadolinium',
         'Metrizoate',
         'IonicIodinatedContrastAgent',
         'IodizedOil',
         'CarbonDioxide',
         'SodiumIodomethamate']}

name_for_cid[10031] = 'CR/DRMechanicalConfiguration'
cid_concepts[10031] = \
{'DCM': ['TableMount', 'UprightStandMount', 'UnmountedDetector', 'CArmMount']}

name_for_cid[3737] = 'NoninvasiveTestIschemia'
cid_concepts[3737] = \
{}

name_for_cid[6165] = 'BreastLinearMeasurements'
cid_concepts[6165] = \
{'DCM': ['Distance',
         'Height',
         'DistanceFromChestWall',
         'PathLength',
         'DistanceFromNipple',
         'DistanceFromSkin'],
 'SCT': ['DiameterOfCircumscribedCircle',
         'Radius',
         'Perimeter',
         'MinorAxis',
         'Width',
         'Diameter',
         'MajorAxis',
         'PerpendicularAxis',
         'LongAxis',
         'Depth',
         'ShortAxis',
         'Length',
         'Circumference']}

name_for_cid[641] = 'Viruses'
cid_concepts[641] = \
{'SCT': ['TheilerMurineEncephalomyelitisVirus', 'AdenoAssociatedVirusGroup']}

name_for_cid[211] = 'QualitativeEvaluationModifierValues'
cid_concepts[211] = \
{'DCM': ['Median'],
 'NCIt': ['Middle'],
 'SCT': ['Bilateral',
         'Central',
         'Right',
         'Posterior',
         'Left',
         'Unilateral',
         'Superior',
         'Distal',
         'Proximal',
         'MidLongitudinal',
         'Lateral',
         'Inferior',
         'Anterior']}

name_for_cid[4271] = 'EnFaceImageTypes'
cid_concepts[4271] = \
{'DCM': ['SuperficialRetinaStructuralReflectanceMap',
         'WholeEyeVasculatureFlow',
         'SuperficialRetinaVasculatureFlow',
         'ChoroidStructuralReflectanceMap',
         'RetinaDepthEncodedVasculatureFlow',
         'MiddleInnerStructuralReflectanceMap',
         'MiddleInnerRetinaVasculatureFlow',
         'WholeEyeStructuralReflectanceMap',
         'RetinaVasculatureFlow',
         'RetinaDepthEncodedStructuralReflectanceMap',
         'DeepRetinaVasculatureFlow',
         'RetinaStructuralReflectanceMap',
         'OuterRetinaVasculatureFlow',
         'DeepRetinaStructuralReflectanceMap',
         'VitreousStructuralReflectanceMap',
         'ChoriocapillarisVasculatureFlow',
         'VitreousVasculatureFlow',
         'OuterRetinaStructuralReflectanceMap',
         'RadialPeripapillaryStructuralReflectanceMap',
         'ChoroidVasculatureFlow',
         'RadialPeripapillaryVasculatureFlow',
         'ChoriocapillarisStructuralReflectanceMap']}

name_for_cid[9506] = 'BrachytherapyDevices'
cid_concepts[9506] = \
{'DCM': ['BrachytherapyChannelShield',
         'BrachytherapySourceApplicator',
         'BrachytherapyChannel']}

name_for_cid[12260] = 'CardiacUltrasoundPulmonaryArtery'
cid_concepts[12260] = \
{'DCM': ['Tau', 'IndicatorMeanTransitTime', 'RRInterval'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'ACInterval',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'VelocityOfFlowPropagation',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Area',
         'CardiovascularOrificeDiameter',
         'Length',
         'Circumference',
         'BloodPressure',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'CardiovascularOrificeArea']}

name_for_cid[3808] = 'AneurysmTypes'
cid_concepts[3808] = \
{'SCT': ['SaccularAneurysm',
         'RupturedAneurysm',
         'TraumaticAneurysm',
         'DissectingAneurysm',
         'SerpentineAneurysm',
         'RacemoseAneurysm',
         'CalcifiedAneurysm',
         'MiliaryAneurysm',
         'MultipleAneurysm',
         'InflammatoryAbdominalAorticAneurysm',
         'CirsoidAneurysm',
         'FusiformAneurysm',
         'ThrombosedAneurysm',
         'ExpandingAneurysm',
         'CylindroidAneurysm',
         'VaricoseAneurysm',
         'MycoticAneurysm',
         'MixedAneurysm']}

name_for_cid[7] = 'UltrasoundBeamPath'
cid_concepts[7] = \
{'SCT': ['TransGastric',
         'TransMural',
         'TransHepatic',
         'TransAbdominal',
         'TransVaginal',
         'TransThecal',
         'TransVesical',
         'TransSplenic',
         'TransPleural',
         'TransEsophageal',
         'TransPancreatic',
         'TransRenal',
         'TransOrbital',
         'TransTemporal']}

name_for_cid[7277] = 'UnitsofDiffusionRateAreaOverTime'
cid_concepts[7277] = \
{'UCUM': ['Mm2S', 'Um2ms', '_106Mm2S', 'Um2S']}

name_for_cid[3234] = 'EctopicBeatMorphology'
cid_concepts[3234] = \
{'SCT': ['VentricularTachycardiaPolymorphic',
         'VentricularBigeminy',
         'VentricularTachycardia',
         'UnifocalPvcs',
         'MultifocalPvcs']}

name_for_cid[6036] = 'ScopeofFeature'
cid_concepts[6036] = \
{'DCM': ['FeatureDetectedOnTheOnlyImage',
         'FeatureDetectedOnOnlyOneOfTheImages',
         'FeatureDetectedOnMultipleImages',
         'FeatureDetectedOnImagesFromMultipleModalities']}

name_for_cid[3664] = 'OxygenConsumptionEquationsandTables'
cid_concepts[3664] = \
{'DCM': ['VO2EqualsVestpd10FIO2FE02',
         'Vo2maleEqualsBSA138Point111Point49LogeAge0Point378Hrf',
         'Vo2femaleEqualsBSA138Point117Point04LogeAge0Point378Hrf',
         'FleischTable',
         'BoothbyTable',
         'VO2Equals176BSA',
         'RobertsonReidTable',
         'VO2Equals152BSA',
         'VO2Equals175BSA']}

name_for_cid[3721] = 'CardiovascularSurgeries'
cid_concepts[3721] = \
{'SCT': ['AbdominalAorticAneurysmStenting',
         'CorrectionOfCongenitalCardiovascularDeformity',
         'CoronaryArteryBypassGraft',
         'OperationOnHeartValve',
         'HeartTransplant',
         'ImplantationOfCardiacPacemaker',
         'PercutaneousCoronaryIntervention',
         'ImplantationOfAutomaticCardiacDefibrillator',
         'AblationOperationForArrhythmia']}

name_for_cid[3408] = 'ActionsWithConsumables'
cid_concepts[3408] = \
{'DCM': ['RemainingConsumableDisposed',
         'ConsumableUnusable',
         'ConsumableTakenFromInventory',
         'ConsumableReturnedToInventory']}

name_for_cid[4267] = 'CornealTopographyMappingUnitsforRealWorldValueMapping'
cid_concepts[4267] = \
{'UCUM': ['mm', 'Micrometer', 'Diopters']}

name_for_cid[8301] = 'TestPatternCodes'
cid_concepts[8301] = \
{'DCM': ['TG18LPV50Pattern',
         'TG18LPH50Pattern',
         'TG18LPH10Pattern',
         'TG18LPV10Pattern',
         'TG18LPH89Pattern',
         'TG18RV89Pattern',
         'TG18RV50Pattern',
         'TG18CXPattern',
         'TG18PXPattern',
         'TG18RV10Pattern',
         'BN05Pattern',
         'BN04Pattern',
         'BN06Pattern',
         'BN01Pattern',
         'GDPattern',
         'BN03Pattern',
         'BN02Pattern',
         'TG18GQPattern',
         'TG18GVNPattern',
         'TG18GQBPattern',
         'ANGPattern',
         'TG18GQNPattern',
         'OIQPattern',
         'TG18NS50Pattern',
         'TG18NS10Pattern',
         'TG18GVPattern',
         'TG18NS89Pattern',
         'TG18AFCPattern',
         'TG18LPV89Pattern',
         'TG18LN1216Pattern',
         'TG18LN1215Pattern',
         'TG18LN1217Pattern',
         'TG18LN1212Pattern',
         'TG18LN1211Pattern',
         'TG18LN1214Pattern',
         'TG18LN1213Pattern',
         'TG18LN1208Pattern',
         'TG18LN1210Pattern',
         'TG18LN1209Pattern',
         'TG18RH89Pattern',
         'TG18RH50Pattern',
         'TG18ADPattern',
         'TG18UNL80Pattern',
         'TG18RH10Pattern',
         'TG18MPPattern',
         'TG18UN10Pattern',
         'TG18LN1218Pattern',
         'TG18UNL10Pattern',
         'TG18UN80Pattern',
         'TG18LN815Pattern',
         'TG18LN812Pattern',
         'TG18LN811Pattern',
         'TG18LN814Pattern',
         'TG18LN813Pattern',
         'TG18LN808Pattern',
         'DINResolutionPattern',
         'TG18LN807Pattern',
         'DINGrayscalePattern',
         'TG18LN810Pattern',
         'TG18LN809Pattern',
         'TG18LN806Pattern',
         'DINGeometryPattern',
         'TG18LN1205Pattern',
         'TG18LN1204Pattern',
         'TG18LN1207Pattern',
         'TG18LN1206Pattern',
         'TG18LN1201Pattern',
         'TG18LN818Pattern',
         'SMPTEPattern',
         'TG18LN1203Pattern',
         'TG18LN1202Pattern',
         'TG18LN817Pattern',
         'TG18LN816Pattern',
         'WhitePattern',
         'BN16Pattern',
         'BN15Pattern',
         'BN12Pattern',
         'BN11Pattern',
         'BN14Pattern',
         'BN13Pattern',
         'TG18CHImage',
         'BN08Pattern',
         'TG18GA30Pattern',
         'BN07Pattern',
         'BN10Pattern',
         'TG18KNImage',
         'BN09Pattern',
         'TG18GA15Pattern',
         'TG18GA10Pattern',
         'TG18GA25Pattern',
         'TG18GA20Pattern',
         'TG18GA03Pattern',
         'TG18GA08Pattern',
         'TG18GA05Pattern',
         'TG18LN805Pattern',
         'TG18LN804Pattern',
         'TG18LN801Pattern',
         'TG18CTPattern',
         'TG18LN803Pattern',
         'TG18LN802Pattern',
         'TG18QCPattern',
         'BN18Pattern',
         'TG18PQCPattern',
         'TG18BRPattern',
         'BN17Pattern',
         'TG18MM2Image',
         'TG18MM1Image']}

name_for_cid[12276] = 'CardiacUltrasoundAortoPulmonaryConnectionsMeasurements'
cid_concepts[12276] = \
{'DCM': ['BloodVelocityPeak',
         'BloodVelocityMean',
         'BloodVelocityMinimum',
         'SystolicBloodVelocityMean',
         'SystolicBloodVelocityPeak',
         'DiastolicBloodVelocityMean',
         'DiastolicBloodVelocityPeak'],
 'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'PulmonaryToSystemicShuntFlowRatio',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'StenosisPeakVelocity',
        'VolumeFlow',
        'FlowRadius',
        'HeartRate',
        'TimeAveragedPeakVelocity',
        'EndDiastolicVelocity',
        'VelocityOfFlowPropagation',
        'PeakGradient',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeAveragedMeanVelocity'],
 'SCT': ['RegurgitantFraction',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter']}

name_for_cid[9510] = 'PurposeofReferenceforRTTreatmentPlanningInput'
cid_concepts[9510] = \
{'DCM': ['RTTreatmentPlanningInputImages',
         'DiagnosticSourceImages',
         'RegistrationResult',
         'SegmentationResult',
         'RTPrescriptionResult']}

name_for_cid[6061] = 'BreastImagingProcedureModifiers'
cid_concepts[6061] = \
{'DCM': ['Calibration',
         'NeedleLocalizationAndBiopsy',
         'QualityControl',
         'MarkerPlacement',
         'Targeted',
         'Survey'],
 'SCT': ['Diagnostic', 'Screening']}

name_for_cid[3263] = 'ElectrodePlacementValues'
cid_concepts[3263] = \
{'MDC': ['_12LeadFromLimbLeadsIIIAndOneOrMoreVLeads',
         'Unspecified12LeadSystem',
         'Standard12LeadAndXYZ',
         'Standard12LeadPositionsElectrodesPlacedIndividually',
         'Standard12LeadAndNEHB',
         'MasonLikarLeadPositionsElectrodesPlacedIndividually',
         'Standard12LeadAndCC5CM5ML',
         'MasonLikarLeadPositionsV1V6InElectrodePad',
         'Standard12LeadAndCM5CC5CH5',
         '_12LeadElectrodePad',
         '_12LeadDerivedFromFrankXYZLeads',
         'Standard12LeadWithExtraLeadsToTheRightAndOrLeftSides',
         'Standard12LeadExtendedToTheRightByV5RV4RV3R',
         '_12LeadDerivedFromNonStandardLeads',
         'Standard12LeadExtendedToTheLeftByV7V8V9',
         '_12LeadForBicycleExerciseTestingLimbLeadsOnBackOfPatient',
         'Standard12LeadPositionsOneIntercostalSpaceHigher',
         'UnspecifiedXYZLeadSystem',
         'FrankXYZLeadSystem',
         'McfeeParungaoXYZLeadSystem',
         'CubeXYZLeadSystem',
         'BipolarUncorrectedXYZLeadSystem',
         'PseudoOrthogonalXYZLeadSystem',
         'XYZLeadsDerivedFromStandard12Lead',
         'NEHBLeadSystem',
         '_3LeadSystemCC5CM5ML',
         '_3LeadSystemCC5CM5CH5',
         '_12LeadFromFrankLeadsXYZLeadsByDowerTransformation',
         '_12LeadFromEASILeadsESASAIByDowerEASITransformation']}

name_for_cid[7220] = 'RTDoseDerivation'
cid_concepts[7220] = \
{'DCM': ['ComposedFromPriorDoses', 'ComposedFromPriorDosesAndCurrentPlan']}

name_for_cid[12108] = 'UpperExtremityVeins'
cid_concepts[12108] = \
{'SCT': ['InnominateVein',
         'SubclavianVein',
         'UlnarVein',
         'DeepPalmarVenousArch',
         'BasilicVein',
         'SuperficialPalmarVenousArch',
         'InternalJugularVein',
         'BrachialVein',
         'SuperiorVenaCava',
         'CephalicVein',
         'AxillaryVein',
         'MedianCubitalVein',
         'RadialVein']}

name_for_cid[10070] = 'RadiationDoseTypes'
cid_concepts[10070] = \
{'DCM': ['EquivalentDose', 'AbsorbedDose']}

name_for_cid[6124] = 'LocationinChest'
cid_concepts[6124] = \
{'DCM': ['Subpleural'],
 'SCT': ['Peripheral',
         'Basal',
         'LowerLobeOfLung',
         'Central',
         'AnteriorSegmentOfRightUpperLobe',
         'UpperZoneOfLung',
         'LowerZoneOfLung',
         'Apical',
         'PosteriorSegmentOfRightUpperLobe',
         'MiddleZoneOfLung',
         'MiddleLobeOfLung',
         'AnteriorSegmentOfLeftUpperLobe',
         'UpperLobeOfLung']}

name_for_cid[250] = 'PositiveNegative'
cid_concepts[250] = \
{'SCT': ['Negative', 'Positive']}

name_for_cid[9402] = 'IEC61217GantryPositionParameters'
cid_concepts[9402] = \
{'DCM': ['IEC61217GantryContinuousRollAngle',
         'IEC61217GantryContinuousPitchAngle',
         'IEC61217GantryContinuousYawAngle']}

name_for_cid[4230] = 'OphthalmicUltrasoundAxialMeasurementsType'
cid_concepts[4230] = \
{'DCM': ['UltrasoundContact', 'UltrasoundImmersion']}

name_for_cid[12221] = 'FlowDirection'
cid_concepts[12221] = \
{'SCT': ['RetrogradeFlow',
         'AntegradeFlow',
         'RegurgitantFlow',
         'RightToLeftCardiovascularShunt',
         'LeftToRightCardiovascularShunt']}

name_for_cid[7484] = 'DXViewforAnimals'
cid_concepts[7484] = \
{'DCM': ['Caudal10DegreeDistalCranioproximalOblique'],
 'SCT': ['Rostrocaudal',
         'Dorsal40DegreeLateralPlantaromedialOblique',
         'Dorsal45DegreeMedialPlantarolateralOblique',
         'Right20DegreeDorsalLeftVentralOblique',
         'DorsorostralVentrocaudalOblique',
         'Lateral45DegDorsal50DegProximalMedioplantarodistalObliq',
         'DorsoVentral',
         'Dorsal45DegreeMedialPalmarolateral',
         'Ventral30DegreeLeftDorsalRightOblique',
         'LeftVentralRightDorsalOblique',
         'VentrorostralDorsocaudalOblique',
         'RightDorsalLeftVentralOblique',
         'PalmaroproximalDorsodistalOblique',
         'CaudodistalCranioproximalOblique',
         'Dorsopalmar',
         'LeftRightLateral',
         'Left45DegreeDorsalRightVentralOblique',
         'Dorsal20DegreeRostralVentrocaudalOblique',
         'Left20DegreeRostralRightCaudalOblique',
         'Left20DegreeVentralRightDorsalOblique',
         'Lateral45DegDorsal50DegProximalMediopalmarodistalOblique',
         'Dorsal45DegreeLateralPlantaromedialOblique',
         'Dorsal60DegreeLateralPlantaromedialOblique',
         'RostroventralCaudodorsal',
         'VentralLeftDorsalRightOblique',
         'Dorsal40DegreeMedialPlantarolateralOblique',
         'Right45DegreeDorsalLeftVentralOblique',
         'ProximoDistal',
         'Plantar60DegreeLateralDorsomedialOblique',
         'Right30DegreeCaudalLeftRostralOblique',
         'Right20DegreeVentralLeftDorsalOblique',
         'PlantarolateralDorsomedialOblique',
         'Dorsal45DegreeLateralPalmaromedialOblique',
         'PalmaromedialDorsolateral',
         'RostrodorsalCaudoventralOblique',
         'DorsomedialPalmarolateral',
         'Dorsal35DegreeLateralPalmaromedialOblique',
         'Ventral20DegreeRostralDorsocaudalOblique',
         'Rostral20DegreeDorsalCaudoventralOblique',
         'DorsolateralPalmaromedialOblique',
         'Dorsal65DegreeProximalPalmarodistalOblique',
         'Dorsal35DegreeMedialPalmarolateralOblique',
         'RightCaudalLeftRostralOblique',
         'Dorsal60DegreeMedialPalmarolateralOblique',
         'Ventral30DegreeRightDorsalLeftOblique',
         'Dorsal40DegreeMedialPalmarolateralOblique',
         'LeftDorsalRightVentralOblique',
         'Dorsal35DegreeLateralPlantaromedialOblique',
         'VentralRightDorsalLeftOblique',
         'LaterodorsoproximalMediopalmarodistalOblique',
         'VentroDorsal',
         'DorsoproximalPlantarodistalOblique',
         'PlantaroproximalDorsodistalOblique',
         'DorsomedialPlantarolateralOblique',
         'Caudocranial',
         'Left30DegreeCaudalRightRostralOblique',
         'Right20DegreeRostralLeftCaudalOblique',
         'RightRostralLeftCaudalOblique',
         'LeftCaudalRightRostralOblique',
         'Dorsal60DegreeLateralPalmaromedialOblique',
         'Left20DegreeDorsalRightVentralOblique',
         'DorsoproximalPalmarodistalOblique',
         'RightLeftLateral',
         'Dorsal60DegreeMedialPlantarolateralOblique',
         'Lateromedial',
         'Plantar75DegreeProximalDorsodistalOblique',
         'Rostral30DegreeVentralCaudodorsal',
         'Dorsal35DegreeMedialPlantarolateralOblique',
         'Dorsal65DegreeProximalPlantarodistalOblique',
         'Dorsoplantar',
         'LeftRostralRightCaudalOblique',
         'LaterodorsoproximalMedioplantarodistalOblique',
         'Right45DegreeVentralLeftDorsalOblique',
         'Left45DegreeVentralRightDorsalOblique',
         'Mediolateral',
         'Palmar45DegreeMedialDorsolateral',
         'DorsolateralPlantaromedialOblique',
         'Palmar75DegreeProximalDorsodistalOblique',
         'RightVentralLeftDorsalOblique']}

name_for_cid[9530] = 'PercentageandDoseDosimetricObjectives'
cid_concepts[9530] = \
{'DCM': ['MinimumPercentVolumeAtRadiationDose',
         'MaximumPercentVolumeAtRadiationDose']}

name_for_cid[3428] = 'ImagingProcedures'
cid_concepts[3428] = \
{'DCM': ['ArterialConduitAngiography'],
 'SCT': ['TransthoracicEchocardiography',
         'CardiacVentriculography',
         'RightVentriculography',
         'LeftVentriculography',
         'IntravascularUltrasound',
         'IntracardiacEchocardiography',
         'CoronaryArteriography',
         'Aortography',
         'EpicardialEchocardiography',
         'BypassGraftAngiography',
         'Angiography',
         'TransesophagealEchocardiography']}

name_for_cid[227] = 'SampleStatisticalDescriptors'
cid_concepts[227] = \
{'DCM': ['PercentileRankingOfMeasurement', 'ZScoreOfMeasurement']}

name_for_cid[7191] = 'TissueSegmentationPropertyTypes'
cid_concepts[7191] = \
{'SCT': ['Ligament',
         'Nerve',
         'Artery',
         'ConnectiveTissue',
         'LymphaticSystem',
         'Capillary',
         'Meniscus',
         'ScarTissue',
         'BloodVessel',
         'Tissue',
         'Cartilage',
         'Organ',
         'Vein',
         'Muscle',
         'Skin',
         'Blood',
         'Bone',
         'LymphaticVessel',
         'LymphNode',
         'BodyFat',
         'Tendon']}

name_for_cid[6153] = 'LesionBoundary'
cid_concepts[6153] = \
{'DCM': ['AbruptInterface', 'EchogenicHalo']}

name_for_cid[3701] = 'CardiacValvesandTracts'
cid_concepts[3701] = \
{'SCT': ['TricuspidValve',
         'PulmonaryValve',
         'AorticValve',
         'MitralValve',
         'LeftVentricleOutflowTract']}

name_for_cid[10007] = 'XRayFilterTypes'
cid_concepts[10007] = \
{'DCM': ['StripFilter',
         'NoFilter',
         'WedgeFilter',
         'ButterflyFilter',
         'FlatFilter']}

name_for_cid[6016] = 'MammographyCompositeFeature'
cid_concepts[6016] = \
{'DCM': ['BreastGeometry',
         'ImageQuality',
         'NonLesion',
         'CooperLigamentChanges',
         'MassInTheSkin',
         'SelectedRegion',
         'MassOnTheSkin',
         'MassWithCalcifications'],
 'SCT': ['BreastComposition',
         'AxillaryLymphNode',
         'AsymmetricBreastTissue',
         'SkinRetractionOfBreast',
         'ArchitecturalDistortionOfBreast',
         'TubularDensity',
         'MammographyBreastDensity',
         'SkinLesion',
         'SkinThickeningOfBreast',
         'FocalAsymmetricBreastTissue',
         'AxillaryAdenopathy',
         'Edema',
         'TrabecularThickeningOfBreast',
         'IndividualCalcification',
         'Nipple',
         'MammographicBreastMass',
         'CalcificationCluster',
         'IntramammaryLymphNode']}

name_for_cid[12201] = 'LeftVentricleLinear'
cid_concepts[12201] = \
{'LN': ['LeftVentricleInternalEndDiastolicDimension',
        'LeftVentricleDiastolicMajorAxis',
        'InterventricularSeptumPercentThickening',
        'LeftVentriclePosteriorWallPercentThickening',
        'LeftVentricleInternalSystolicDimension',
        'LeftVentricularFractionalShortening',
        'LeftVentricleSystolicMajorAxis',
        'InterventricularSeptumSystolicThickness',
        'LeftVentriclePosteriorWallDiastolicThickness',
        'LeftVentriclePosteriorWallSystolicThickness',
        'InterventricularSeptumToPosteriorWallThicknessRatio',
        'InterventricularSeptumDiastolicThickness'],
 'SCT': ['LeftVentricleTruncatedSemiMajorAxisDiastolicDimension',
         'LeftVentricleSemiMajorAxisDiastolicDimension']}

name_for_cid[620] = 'AirwayManagementSubMethodCodeTypefromAnesthesiaQualityInitiative(AQI)'
cid_concepts[620] = \
{'SCT': ['HighFrequencyVentilation',
         'TranstrachealJetVentilation',
         'ContinuousFlowVentilation'],
 'UMLS': ['InverseRatioVentilation']}

name_for_cid[270] = 'ObserverType'
cid_concepts[270] = \
{'DCM': ['Person', 'Device']}

name_for_cid[4210] = 'OphthalmicTomographyAcquisitionDevice'
cid_concepts[4210] = \
{'DCM': ['ReflectionBasedCornealTopographer',
         'InterferometryBasedCornealTomographer',
         'ScheimpflugCamera',
         'ElevationBasedCornealTomographer'],
 'SCT': ['ScanningLaserPolarimeter',
         'OpticalCoherenceTomographyScanner',
         'RetinalThicknessAnalyzer',
         'ConfocalScanningLaserOphthalmoscope']}

name_for_cid[73] = 'BolusShapingCurves'
cid_concepts[73] = \
{'DCM': ['LinearCurve', 'NegativeExponential']}

name_for_cid[10050] = 'SummaryRadiationExposureQuantities'
cid_concepts[10050] = \
{'DCM': ['EffectiveDose',
         'TotalFluoroTime',
         'TotalNumberOfRadiographicFrames',
         'MeanCtdivol',
         'DoseAreaProductTotal',
         'CTDoseLengthProductTotal',
         'EntranceExposureAtRP',
         'AcquisitionDoseAreaProductTotal',
         'AccumulatedAverageGlandularDose',
         'FluoroDoseAreaProductTotal',
         'AdministeredActivity']}

name_for_cid[3756] = 'CardiacPatientRiskFactors'
cid_concepts[3756] = \
{'SCT': ['Arrhythmia',
         'HistoryOfHypertension',
         'HistoryOfDiabetes',
         'HistoryOfCongestiveHeartFailure',
         'PeripheralVascularDisease',
         'InsulinDependentMotherIDM',
         'HIVPositive',
         'HistoryOfChronicLungDisease',
         'HistoryOfHypercholesterolemia',
         'HistoryOfRenalFailure',
         'HistoryOfCerebrovascularDisease',
         'HistoryOfMyocardialInfarction'],
 'UMLS': ['InfantOfMotherWithGestationalDiabetes']}

name_for_cid[6104] = 'AbnormalOpacityFindingorFeature'
cid_concepts[6104] = \
{'DCM': ['ReticulonodularPattern',
         'GroundGlassOpacity',
         'BeadedSeptumSign',
         'Infiltrate',
         'NodularPattern',
         'Micronodule',
         'Opacity',
         'Pseudoplaque',
         'PhantomTumor',
         'SignetRingSign',
         'Shadow',
         'PrimaryComplex',
         'AbnormalInterstitialPattern',
         'SmallIrregularOpacities',
         'SmallRoundedOpacities',
         'TreeInBudSign',
         'Density',
         'DependentOpacity'],
 'SCT': ['Mass', 'PulmonaryEmbolism', 'Nodule', 'CoinLesion']}

name_for_cid[3613] = 'HemodynamicTimeMeasurements'
cid_concepts[3613] = \
{'DCM': ['Tau',
         'IndicatorMeanTransitTime',
         'RightVentricularFillingTime',
         'LeftVentricularFillingTime',
         'RightVentricularEjectionTime',
         'LeftVentricularEjectionTime',
         'RRInterval']}

name_for_cid[6041] = 'MammographyImageQualityFinding'
cid_concepts[6041] = \
{'DCM': ['ViewAndLateralityMarkerIsPartiallyObscured',
         'ViewAndLateralityMarkerIsIncorrect',
         'ViewAndLateralityMarkerIsOffImage',
         'FlashIsNotNearEdgeOfFilm',
         'FlashIsIllegibleDoesNotFitOrIsLopsided',
         'FlashDoesnTIncludePatientNameAndAdditionalPatientId',
         'OtherFailure',
         'FlashDoesnTIncludeDateOfExamination',
         'UnknownFailure',
         'FlashDoesnTIncludeFacilityNameAndLocation',
         'MLOEvidenceOfMotionBlur',
         'ViewAndLateralityMarkerIsNotNearTheAxilla',
         'ViewAndLateralityMarkerOverlapsBreastTissue',
         'Positioning',
         'MLOInframammaryFoldIsNotOpen',
         'CCNotAllMedialTissueVisualized',
         'CCNippleNotCenteredOnImage',
         'CCPosteriorNippleLineDoesNotMeasureWithin1cmOfMLO',
         'NippleNotInProfile',
         'InsufficientImplantDisplacementIncorrect',
         'GridArtifacts',
         'DateStickerIsMissing',
         'TechnicalFactorsMissing',
         'CollimationTooCloseToBreast',
         'InadequateCompression',
         'MLOInsufficientPectoralMuscle',
         'MLONoFatIsVisualizedPosteriorToFibroglandularTissues',
         'ViewAndLateralityMarkerIsMissing',
         'MLOPoorSeparationOfDeepAndSuperficialBreastTissues',
         'MotionBlur',
         'ViewAndLateralityMarkerDoesNotHaveBothViewAndLaterality',
         'UnderExposed',
         'FlashDoesnTIncludeTechnologistIdentification',
         'FlashDoesnTIncludeCassetteScreenDetectorIdentification',
         'FlashDoesnTIncludeMammographyUnitIdentification',
         'ViewAndLateralityMarkerDoesNotHaveApprovedCodes',
         'OverExposed',
         'NoImage',
         'DetectorArtifacts',
         'ArtifactsOtherThanGridOrDetectorArtifact',
         'MechanicalFailure',
         'ElectricalFailure',
         'SoftwareFailure',
         'InappropriateImageProcessing']}

name_for_cid[12217] = 'EchocardiographyCardiacShunt'
cid_concepts[12217] = \
{'LN': ['PulmonaryToSystemicShuntFlowRatio', 'HeartRate']}

name_for_cid[3486] = 'VascularMeasurementSites'
cid_concepts[3486] = \
{'DCM': ['SiteOfLumenMaximum',
         'SiteOfLumenMinimum',
         'ProximalReference',
         'DistalReference']}

name_for_cid[3469] = 'ESVolume'
cid_concepts[3469] = \
{'DCM': ['LeftAtrialESVolume'],
 'LN': ['RightVentricularESVolume', 'LeftVentricularESVolume']}

name_for_cid[4206] = 'OphthalmicChannelDescription'
cid_concepts[4206] = \
{'SCT': ['Infrared',
         'Ultraviolet',
         'Red',
         'FullSpectrum',
         'Blue',
         'RedFree',
         'Green']}

name_for_cid[1004] = 'AnatomicalReferenceBasisChest'
cid_concepts[1004] = \
{'SCT': ['Lung',
         'ApexOfHeart',
         'SternoclavicularJoint',
         'AorticArch',
         'Carina',
         'SuprasternalNotch',
         'Diaphragm',
         'Heart',
         'XiphoidProcess',
         'CoronaryArteryGraft',
         'ShoulderRegionStructure',
         'Scapula',
         'AcromioclavicularJoint',
         'Sternum',
         'ThoracicInlet']}

name_for_cid[636] = 'TemperatureSensorDeviceComponentTypeforSmallAnimalProcedures'
cid_concepts[636] = \
{'DCM': ['CarrierTemperatureSensor'],
 'SCT': ['RectalTemperature'],
 'UMLS': ['Thermography']}

name_for_cid[6112] = 'PleuraAnatomyFindingorFeature'
cid_concepts[6112] = \
{'DCM': ['AnteriorJunctionLine', 'PosteriorJunctionLine'],
 'SCT': ['FissureOfLung']}

name_for_cid[3740] = 'ThrombolyticAdministration'
cid_concepts[3740] = \
{'SCT': ['Contraindicated',
         'AdministeredLessThan3HoursBeforePCI',
         'AdministeredBetween3And6HoursBeforePCI',
         'AdministeredBetween6HoursAnd7DaysBeforePCI']}

name_for_cid[10046] = 'GFRMeasurements'
cid_concepts[10046] = \
{'LN': ['GlomerularFiltrationRateBlackMDRD',
        'CockroftGaultFormulaEstimationOfGFR',
        'GlomerularFiltrationRateCystatinBasedFormula',
        'CKDEPIFormulaEstimationOfGFR',
        'GlomerularFiltrationRateMDRD',
        'GlomerularFiltrationRateCreatinineBasedFormulaSchwartz',
        'GlomerularFiltrationRateNonBlackMDRD',
        'GlomerularFiltrationRateFemaleMDRD']}

name_for_cid[7703] = 'AssociationFibers'
cid_concepts[7703] = \
{'DCM': ['SuperiorLongitudinalFasciculusIII',
         'SuperiorLongitudinalFasciculusII',
         'SuperiorLongitudinalFasciculusI'],
 'NEU': ['ArcuateFasciculus', 'SuperiorLongitudinalFasciculus'],
 'SCT': ['SuperiorFrontoOccipitalFasciculus',
         'InferiorLongitudinalFasciculus',
         'InferiorFrontoOccipitalFasciculus',
         'UncinateFasciculus',
         'VerticalOccipitalFasciculus']}

name_for_cid[65] = 'PreMedicationforImagingAgentAdministration'
cid_concepts[65] = \
{'SCT': ['Alfentanil',
         'Remifentanil',
         'Diphenhydramine',
         'Papaverine',
         'Propofol',
         'Midazolam',
         'Sufentanil',
         'Methylprednisolone',
         'HydrocortisoneSodiumSuccinate',
         'DexamethasoneSodiumSulfate',
         'MethylprednisoloneSodiumSuccinate',
         'Ephedrine',
         'H1Antihistamine',
         'Prednisone']}

name_for_cid[6057] = 'DuctographyFindingsforBreast'
cid_concepts[6057] = \
{'DCM': ['DuctNarrowing',
         'NormalBreastTissue',
         'CystFill',
         'IntraluminalFillingDefect',
         'MultipleFillingDefect',
         'AbruptDuctTermination',
         'Extravasation'],
 'SCT': ['MammaryDuctEctasia']}

name_for_cid[9526] = 'IonTherapyParticle'
cid_concepts[9526] = \
{'DCM': ['_16OxygenNucleus',
         '_12CarbonNucleus',
         '_4HeliumNucleus',
         '_3HeliumNucleus'],
 'SCT': ['Electron', 'Proton']}

name_for_cid[12240] = 'LeftVentricleArea'
cid_concepts[12240] = \
{'SCT': ['LeftVentricleEpicardialDiastolicAreaPsaxPapView',
         'LeftVentricularDiastolicArea',
         'LeftVentricularFractionalAreaChange',
         'LeftVentricularSystolicArea']}

name_for_cid[7168] = 'BrainLesionSegmentationTypesWithNecrosis'
cid_concepts[7168] = \
{'DCM': ['Background'],
 'NCIt': ['NonEnhancingLesion', 'EnhancingLesion'],
 'SCT': ['Normal', 'Necrosis', 'Edema']}

name_for_cid[12305] = 'BasicEchoAnatomicSites'
cid_concepts[12305] = \
{'DCM': ['RightVentricleAnteriorWall',
         'RightVentricleOutflowTractProximal',
         'RightVentricleOutflowTractDistal'],
 'SCT': ['LeftVentricleMidAnteriorSegment',
         'PulmonicValve',
         'LeftVentricleMyocardium',
         'PulmonaryVein',
         'AorticValve',
         'AorticSinotubularJunction',
         'LeftVentricleBasalAnteriorSegment',
         'LeftVentricleBasalInferiorSegment',
         'InferiorVenaCava',
         'MitralAnnulus',
         'RightPulmonaryArtery',
         'AorticRoot',
         'RightAtrium',
         'TricuspidValve',
         'MedialMitralAnnulus',
         'LeftVentricleMidAnterolateralSegment',
         'RightVentricle',
         'LeftVentricle',
         'RightVentricleMidventricularSegment',
         'LeftVentricleMidInferoseptalSegment',
         'PulmonicRing',
         'LeftVentricleMidAnteroseptalSegment',
         'LeftVentricleOutflowTract',
         'AorticValveRing',
         'InterventricularSeptum',
         'LeftPulmonaryArtery',
         'TricuspidAnnulus',
         'RightVentricleOutflowTract',
         'LeftVentricleBasalAnteroseptalSegment',
         'LeftVentricleBasalInferolateralSegment',
         'LeftVentricleBasalInferoseptalSegment',
         'LeftVentricleMidInferiorSegment',
         'TrunkOfPulmonaryArtery',
         'LeftVentricleBasalAnterolateralSegment',
         'AscendingAorta',
         'PulmonaryArtery',
         'LeftVentricleMidInferolateralSegment',
         'LeftAtrium',
         'RightVentricleBasalSegment',
         'LateralMitralAnnulus',
         'MitralValve']}

name_for_cid[231] = 'YesNoOnly'
cid_concepts[231] = \
{'SCT': ['Yes', 'No']}

name_for_cid[4251] = 'VisualFieldStaticPerimetryTestStrategies'
cid_concepts[4251] = \
{'DCM': ['VisualFieldTOPTestStrategy',
         'VisualFieldCLIPTestStrategy',
         'VisualFieldSITAStandardTestStrategy',
         'VisualFieldDynamicTestStrategy',
         'VisualFieldCLASSStrategy',
         'VisualFieldSITASWAPTestStrategy',
         'VisualFieldNormalTestStrategy',
         'VisualFieldSITAFastTestStrategy',
         'VisualField1LTTestStrategy',
         'VisualFieldFullThresholdTestStrategy',
         'VisualField2LTTestStrategy',
         'VisualFieldFastpacTestStrategy',
         'VisualFieldLVSTestStrategy',
         'VisualFieldFullFromPriorTestStrategy',
         'VisualFieldGATETestStrategy',
         'VisualFieldOptimaTestStrategy',
         'VisualFieldGateiTestStrategy',
         'VisualFieldTwoZoneTestStrategy',
         'VisualField2LTDynamicTestStrategy',
         'VisualFieldThreeZoneTestStrategy',
         'VisualField2LTNormalTestStrategy',
         'VisualFieldQuantifyDefectsTestStrategy',
         'VisualFieldFastThresholdTestStrategy']}

name_for_cid[7304] = 'ImplantTargetAnatomy'
cid_concepts[7304] = \
{'SCT': ['Clavicle',
         'Tibia',
         'DistalFemur',
         'ProximalRadius',
         'ProximalUlna',
         'WristJoint',
         'ThoracoLumbarSpine',
         'Hand',
         'ShaftOfTibia',
         'Shoulder',
         'Pelvis',
         'Humerus',
         'ElbowJoint',
         'FacialBones',
         'DistalRadius',
         'ShaftOfUlna',
         'Fibula',
         'JawRegion',
         'Talus',
         'ThoracicSpine',
         'Calcaneus',
         'Ulna',
         'HeadOfFemur',
         'Radius',
         'Knee',
         'HipJoint',
         'DistalUlna',
         'LumboSacralSpine',
         'Mandible',
         'FingerJoint',
         'CervicoThoracicSpine',
         'ShaftOfRadius',
         'Femur',
         'ShaftOfFemur',
         'Skull',
         'ShaftOfHumerus',
         'DistalHumerus',
         'ProximalHumerus',
         'ProximalFemur',
         'Patella',
         'DistalTibia',
         'AnkleJoint',
         'LumbarSpine',
         'Maxilla',
         'CervicalSpine']}

name_for_cid[32] = 'NonAcquisitionModality'
cid_concepts[32] = \
{'DCM': ['Registration',
         'Segmentation',
         'BasicVoiceAudio',
         'RTDose',
         'Document',
         'StereometricRelationship',
         'RTTreatmentRecord',
         'HardCopy',
         'CTProtocol',
         'RTPlan',
         'AutomatedSlideStainer',
         'StructuredReportDocument',
         'SpatialFiducials',
         'ModelFor3DManufacturing',
         'RTStructureSet',
         'Plan',
         'PresentationState',
         'Other',
         'KeyObjectSelection',
         'IntraocularLensCalculation',
         'ContentAssessmentResult',
         'RealWorldValueMap']}

name_for_cid[10011] = 'EffectiveDoseEvaluationMethod'
cid_concepts[10011] = \
{'DCM': ['DLPToEConversionViaMCComputation',
         'DLPToEConversionViaMeasurement',
         'CtdifreeairToEConversionViaMCComputation',
         'CtdifreeairToEConversionViaMeasurement']}

name_for_cid[3717] = 'MyocardialWallSegments'
cid_concepts[3717] = \
{'SCT': ['LeftVentricleMidAnteriorSegment',
         'LeftVentricleBasalAnteroseptalSegment',
         'LeftVentricleBasalInferolateralSegment',
         'LeftVentricleBasalInferoseptalSegment',
         'LeftVentricleMidInferiorSegment',
         'LeftVentricleApicalInferiorSegment',
         'LeftVentricleApicalLateralSegment',
         'LeftVentricleApicalAnteriorSegment',
         'LeftVentricleBasalAnteriorSegment',
         'LeftVentricleApicalSeptalSegment',
         'ApexOfLeftVentricle',
         'LeftVentricleBasalAnterolateralSegment',
         'LeftVentricleBasalInferiorSegment',
         'LeftVentricleMidInferolateralSegment',
         'LeftVentricleMidAnterolateralSegment',
         'LeftVentricleMidInferoseptalSegment',
         'LeftVentricleMidAnteroseptalSegment']}

name_for_cid[6145] = 'BaselineCategory'
cid_concepts[6145] = \
{'DCM': ['TargetLesionAtBaseline',
         'NonTargetLesionAtBaseline',
         'NonLesionAtBaseline']}

name_for_cid[6000] = 'OverallBreastComposition'
cid_concepts[6000] = \
{'SCT': ['ExtremelyDense',
         'AlmostEntirelyFat',
         'HeterogeneouslyDense',
         'ScatteredFibroglandularDensities']}

name_for_cid[3202] = 'ChestPain'
cid_concepts[3202] = \
{'DCM': ['AnginalEquivalent'],
 'SCT': ['AtypicalAngina',
         'TypicalAngina',
         'NoncardiacChestPain',
         'ChestPainNotPresent',
         'StableAngina',
         'UnstableAngina']}

name_for_cid[3015] = 'CoronaryArteries'
cid_concepts[3015] = \
{'BARI': ['_2ndMarginalCoronaryArteryLaterals',
          '_1stDiagonalCoronaryArteryLaterals',
          '_3rdDiagonalCoronaryArteryLaterals',
          '_1stMarginalCoronaryArteryLaterals',
          '_3rdMarginalCoronaryArteryLaterals',
          '_2ndDiagonalCoronaryArteryLaterals',
          'RamusLaterals',
          'PosteriorDescendingSeptalPerforators'],
 'SCT': ['ProximalRightCoronaryArtery',
         '_3rdLeftPosterolateralCoronaryArtery',
         'RightPosteriorAVCoronaryArtery',
         'ProximalCircumflexCoronaryArtery',
         '_2ndMarginalCoronaryArtery',
         'MidCircumflexCoronaryArtery',
         '_1stLeftPosterolateralCoronaryArtery',
         'IntermediateArteryRamus',
         'CircumflexCoronaryArtery',
         'DistalLeftAnteriorDescendingCoronaryArtery',
         'RightCoronaryArteryOstium',
         'LeftMainCoronaryArtery',
         '_2ndLeftPosterolateralCoronaryArtery',
         'DistalRightCoronaryArtery',
         'MarginalCoronaryArtery',
         '_2ndRightPosterolateralCoronaryArtery',
         'PosteriorDescendingRightCoronaryArtery',
         'LeftAnteriorDescendingCoronaryArtery',
         'LeftPosteriorDescendingCircumflexCoronaryArtery',
         '_1stMarginalCoronaryArtery',
         'ProximalLeftAnteriorDescendingCoronaryArtery',
         '_3rdDiagonalCoronaryArtery',
         '_1stSeptalCoronaryArtery',
         'RightCoronaryArtery',
         '_3rdMarginalCoronaryArtery',
         '_1stDiagonalCoronaryArtery',
         'MidRightCoronaryArtery',
         '_2ndDiagonalCoronaryArtery',
         'AVGrooveContinuationOfCircumflexArtery',
         'DistalCircumflexCoronaryArtery',
         'MidLeftAnteriorDescendingCoronaryArtery',
         'LeftMainCoronaryArteryOstium',
         'CoronaryArteryGraft',
         'LeftPosterolateralCircumflexCoronaryArtery',
         '_1stRightPosterolateralCoronaryArtery',
         '_3rdRightPosterolateralCoronaryArtery']}

name_for_cid[3500] = 'PressureUnits'
cid_concepts[3500] = \
{'UCUM': ['Mmhg', 'Kpa']}

name_for_cid[3629] = 'ProcedureIntent'
cid_concepts[3629] = \
{'DCM': ['QualityControlIntent'],
 'SCT': ['DiagnosticIntent',
         'StagingIntent',
         'TherapeuticIntent',
         'CurativeIntent',
         'ProphylacticIntent',
         'CombinedDiagnosticAndTherapeuticProcedure',
         'AdjunctIntent',
         'ForensicIntent',
         'NeoAdjuvantIntent',
         'SupportiveIntent',
         'GuidanceIntent',
         'PalliativeIntent',
         'AdjuvantIntent',
         'ScreeningIntent',
         'PreventiveIntent']}

name_for_cid[12112] = 'AbdominalArteries(Unilateral)'
cid_concepts[12112] = \
{'SCT': ['InfraRenalAorta',
         'SplenicArtery',
         'CeliacAxis',
         'GastroduodenalArtery',
         'LumbarArtery',
         'SupraRenalAorta',
         'Aorta',
         'LeftBranchOfHepaticArtery',
         'ProperHepaticArtery',
         'CommonIliacArtery',
         'SuperiorMesentericArtery',
         'InferiorMesentericArtery',
         'RightBranchOfHepaticArtery',
         'CommonHepaticArtery']}

name_for_cid[8120] = 'WSIReferencedImagePurposesofReference'
cid_concepts[8120] = \
{'DCM': ['ImagedContainerLabel',
         'Localizer',
         'SameAcquisitionAtDifferentFocalDepth',
         'SameAcquisitionAtDifferentSpectralBand',
         'SameAcquisitionAtLowerResolution',
         'SameAcquisitionAtHigherResolution']}

name_for_cid[6094] = 'QualitativeConceptsforUsage,ExposureFrequency'
cid_concepts[6094] = \
{'DCM': ['RelativeDoseFrequency',
         'RelativeFrequencyOfExposure',
         'RelativeFrequencyOfUse']}

name_for_cid[4103] = 'ArterialInputFunctionMeasurementMethods'
cid_concepts[4103] = \
{'DCM': ['UserDefinedAIFROI',
         'AutomaticallyDetectedAIFROI',
         'BlindEstimationOfAIF',
         'AIFIgnored',
         'PopulationAveragedAIF']}

name_for_cid[3412] = 'InterventionActionsandStatus'
cid_concepts[3412] = \
{'DCM': ['StentReExpanded',
         'ObjectRemoved',
         'AngioplastyBalloonDeflated',
         'DeviceDeployed',
         'GuidewireCrossingLesionSuccessful',
         'InterventionalDeviceWithdrawn',
         'AngioplastyBalloonInflated',
         'InterventionalDevicePlaced',
         'GuidewireCrossingLesionUnsuccessful',
         'InterventionPerformed',
         'RadiationApplied',
         'RadiationRemoved',
         'InterventionalDevicePlacementUnsuccessful']}

name_for_cid[7001] = 'DiagnosticImagingReportHeadings'
cid_concepts[7001] = \
{'LN': ['Addendum',
        'ClinicalInformation',
        'MedicationsAdministered',
        'RadiationExposureAndProtectionInformation',
        'Complications',
        'PatientPresentation',
        'History',
        'Impressions',
        'CommunicationOfCriticalResults',
        'Conclusions',
        'Findings',
        'Recommendations',
        'PreviousFindings',
        'PriorProcedureDescriptions',
        'KeyImages',
        'IndicationsForProcedure',
        'Summary',
        'CurrentProcedureDescriptions',
        'Request']}

name_for_cid[12283] = 'CardiacUltrasoundVenousReturnPulmonaryFindingSites'
cid_concepts[12283] = \
{'SCT': ['PulmonaryVeinGreatVessel',
         'PulmonaryVein',
         'PulmonaryVeinToAtriumConnection',
         'MustardOrSenningPulmonaryVenousPathway',
         'CorTriatriatumOrifice']}

name_for_cid[7451] = 'FamilyMember'
cid_concepts[7451] = \
{'SCT': ['NaturalGrandFather',
         'NaturalBrother',
         'MaleFirstCousin',
         'NaturalFather',
         'Aunt',
         'NaturalGrandMother',
         'Uncle',
         'HalfSister',
         'NaturalMother',
         'NaturalDaughter',
         'NaturalSon',
         'FemaleFirstCousin',
         'NaturalSister',
         'HalfBrother']}

name_for_cid[3804] = 'StenosisMeasurementMethods'
cid_concepts[3804] = \
{'DCM': ['ECST', 'NASCET', 'AreaBasedMethod', 'DiameterBasedMethod']}

name_for_cid[3107] = 'PETCardiologyRadiopharmaceuticals'
cid_concepts[3107] = \
{'SCT': ['_82Rubidium', 'FluorodeoxyglucoseF18', '_13Nitrogen']}

name_for_cid[4011] = 'DXViewModifier'
cid_concepts[4011] = \
{'DCM': ['Crosstable'],
 'SCT': ['Transoral',
         'Transforamenal',
         'Caudad',
         'Cephalad',
         'Transthoracic',
         'Transorbital',
         'MouthClosed']}

name_for_cid[3691] = 'ECGControlVariablesText'
cid_concepts[3691] = \
{'MDC': ['Interpolator',
         'HighPassFilter',
         'InterpolatorDescription',
         'HighPassFilterDescription',
         'NotchFilter',
         'NotchFilterDescription',
         'BaselineDescription',
         'LowPassFilter']}

name_for_cid[7017] = 'CategoryofTeachingMaterialImaging'
cid_concepts[7017] = \
{'DCM': ['CardiovascularImagingSubjectMatter',
         'GenitourinaryImagingSubjectMatter',
         'GastrointestinalImagingSubjectMatter',
         'VascularAndInterventionalImagingSubjectMatter',
         'NeuroimagingSubjectMatter',
         'UltrasoundImagingSubjectMatter',
         'NuclearMedicineImagingSubjectMatter',
         'BreastImagingSubjectMatter',
         'PediatricImagingSubjectMatter',
         'PulmonaryImagingSubjectMatter',
         'MusculoskeletalImagingSubjectMatter']}

name_for_cid[3404] = 'StaffActions'
cid_concepts[3404] = \
{'DCM': ['OfficeCalled',
         'PageSentTo',
         'ConsultationWith',
         'PersonnelArrived',
         'PersonnelDeparted']}

name_for_cid[3111] = 'NuclearCardiologyRadiopharmaceuticals'
cid_concepts[3111] = \
{'SCT': ['Thallium201', 'Tc99mSestamibi', 'Tc99mTetrofosmin']}

name_for_cid[7152] = 'CardiacStructureSegmentationTypes'
cid_concepts[7152] = \
{'SCT': ['Pericardium',
         'PericardialCavity',
         'Heart',
         'RightVentricle',
         'LeftVentricle',
         'Aorta']}

name_for_cid[3668] = 'FraminghamTables'
cid_concepts[3668] = \
{'DCM': ['TotalCholesterolScoreSheetForWomen',
         'LDLCholesterolScoreSheetForWomen',
         'TotalCholesterolScoreSheetForMen',
         'LDLCholesterolScoreSheetForMen']}

name_for_cid[3238] = 'StressScoringScales'
cid_concepts[3238] = \
{'DCM': ['STHRIndex',
         'RatioOfAchievedToPredictedFunctionalCapacity',
         'AerobicIndex',
         'RatioOfAchievedToPredictedMaximalOxygenConsumption'],
 'SCT': ['DukeTreadmillScore']}

name_for_cid[12016] = 'EstimatedFetalWeightPercentileEquationsandTables'
cid_concepts[12016] = \
{'DCM': ['FWPByGACampbell1991', 'FWPByGAHadlock1991'],
 'LN': ['FWPByGAAlexander1996',
        'FWPByGAWilliams1982',
        'MaleTwinsBWPByGAArbuckle1993',
        'FWPByGABrenner1976',
        'FemaleSingletonBWPByGAArbuckle1993',
        'FemaleTwinsBWPByGAArbuckle1993',
        'MaleSingletonBWPByGAArbuckle1993',
        'FWPByGAHadlock1985']}

name_for_cid[3687] = 'ElectrophysiologyWaveformDurations'
cid_concepts[3687] = \
{'MDC': ['R3DurationPerLead',
         'POnsetToP1DurationPerLead',
         'S3DurationPerLead',
         'PQSegmentTimePeriodPerLead',
         'PPTimePeriodPerLead',
         'QRSDurationPerLead',
         'VentricularActivationTimePerLead',
         'S1DurationPerLead',
         'QTUTimePeriodPerLead',
         'PDurationPerLead',
         'POnsetToP2DurationPerLead',
         'R2DurationPerLead',
         'POffsetToQRSOnsetDurationPerLead',
         'S2DurationPerLead',
         'RRTimePeriodPerLead',
         'R1DurationPerLead',
         'POnsetToP3DurationPerLead',
         'QDurationPerLead',
         'PQTimePeriodPerLead',
         'QTDurationPerLead']}

name_for_cid[3453] = 'CardiacVolumeMethods'
cid_concepts[3453] = \
{'DCM': ['AreaLengthDodge',
         'AreaLengthKennedy',
         'Parallelepiped',
         'TSPyramid',
         'TwoChamber',
         'MultipleSlices',
         'Boak',
         'AreaLengthWynne']}

name_for_cid[6201] = 'ColonFindingorFeature'
cid_concepts[6201] = \
{'DCM': ['ImageQuality', 'SelectedRegion', 'NonLesion', 'AnatomicNonColon'],
 'SCT': ['Lipoma',
         'InvertedDiverticulum',
         'IleocecalValve',
         'TumorOfColon',
         'Diverticulum',
         'Feces',
         'ContrastMedia',
         'IntraluminalFluid',
         'RectalMass',
         'OperativeSite',
         'PolypOfColon',
         'ColonicHaustra']}

name_for_cid[3003] = 'HemodynamicWaveformSources'
cid_concepts[3003] = \
{'SCT': ['HemodynamicOxygenSaturationWaveform',
         'LeftAtriumPressureWaveform',
         'RightVentriclePressureWaveform',
         'ThermalCardiacOutputWaveform',
         'HemodynamicPressureWaveform',
         'PulmonaryValvePullbackPressureWaveform',
         'TemperatureWaveform',
         'PulmonaryCapillaryWedgePressureWaveform',
         'ArterialPressureWaveform',
         'TricuspidValvePullbackPressureWaveform',
         'AorticPressureWaveform',
         'FemoralArteryPressureWaveform',
         'HemodynamicFlowWaveform',
         'LeftVentriclePressureWaveform',
         'RightAtriumPressureWaveform',
         'DyeDilutionCardiacOutputWaveform',
         'MitralValvePullbackPressureWaveform',
         'AorticValvePullbackPressureWaveform',
         'HemodynamicWaveform',
         'RespirationImpedanceWaveform',
         'HemodynamicImpedanceWaveform',
         'CentralVenousPressureWaveform',
         'PulmonaryArteryOxygenSaturationWaveform',
         'PulmonaryArteryWedgePressureWaveform',
         'PulmonaryArteryPressureWaveform']}

name_for_cid[7040] = 'BroselowLutenPediatricSizeCategories'
cid_concepts[7040] = \
{'SCT': ['BroselowLutenPurpleZone1011kg',
         'BroselowLutenYellowZone1214kg',
         'BroselowLutenWhiteZone1518kg',
         'BroselowLutenGreenZone3036kg',
         'BroselowLutenBlueZone1923kg',
         'BroselowLutenPinkZone67kg',
         'BroselowLutenOrangeZone2429kg',
         'BroselowLutenRedZone89kg']}

name_for_cid[12104] = 'ExtracranialArteries'
cid_concepts[12104] = \
{'SCT': ['InternalCarotidArtery',
         'ExternalCarotidArtery',
         'VertebralArtery',
         'CommonCarotidArtery',
         'CarotidBulb',
         'SubclavianArtery',
         'CarotidBifurcation']}

name_for_cid[4050] = 'DrugorContrastAgentCharacteristics'
cid_concepts[4050] = \
{'DCM': ['ContrastBolusIngredientOpaque',
         'ActiveIngredientUndilutedConcentration'],
 'SCT': ['ActiveIngredient', 'Volume']}

name_for_cid[6128] = 'ChestDistributionDescriptor'
cid_concepts[6128] = \
{'DCM': ['BatWingDistribution',
         'ButterflyDistribution',
         'Centrilobular',
         'Coalescent',
         'Lobar'],
 'SCT': ['Discoid',
         'Focal',
         'Generalized',
         'Multifocal',
         'Disseminated',
         'Systemic',
         'Diffuse',
         'Segmental']}

name_for_cid[6082] = 'GynecologicalProcedures'
cid_concepts[6082] = \
{'SCT': ['DilationAndCurettage', 'EndometrialBiopsy', 'Hysterectomy']}

name_for_cid[4237] = 'LensConstantType'
cid_concepts[4237] = \
{'DCM': ['HaigisA1',
         'HaigisA2',
         'HofferPacdConstant',
         'SurgeonFactor',
         'BarrettLensFactor',
         'ACDConstant',
         'BarrettDesignFactor',
         'HaigisA0'],
 'SCT': ['AConstant']}

name_for_cid[3458] = 'ContourRealignment'
cid_concepts[3458] = \
{'DCM': ['LongAxisBased', 'NoRealignment', 'CenterOfGravity']}

name_for_cid[607] = 'AnimalFeedTypes'
cid_concepts[607] = \
{'DCM': ['NIH31', 'NIH07', 'AIN76', 'AIN93G', 'AIN93M']}

name_for_cid[12226] = 'EchocardiographyImageView'
cid_concepts[12226] = \
{'SCT': ['ParasternalShortAxisAtTheMitralValveLevel',
         'SubcostalShortAxis',
         'ParasternalLongAxis',
         'SuprasternalShortAxis',
         'SubcostalViewOfCardiacOutletsDirectedAnteriorly',
         'ParasternalLongAxisViewOfTheRVOutflowTract',
         'ParasternalShortAxisAtThePapillaryMuscleLevel',
         'SuprasternalSagittalView',
         'SuprasternalLongAxis',
         'TransesophagealShortAxisView',
         'SuprasternalCoronalView',
         'ParasternalLongAxisViewOfTheRVInflowTract',
         'RightVentricularInflowTractView',
         'ApicalTwoChamber',
         'SuprasternalLongAxisViewOfAorticArch',
         'SubcostalObliqueCoronalView',
         'SubcostalLongAxis',
         'SubcostalShortAxisViewAtMitralValveLevel',
         'ApicalLongAxis',
         'SubcostalShortAxisViewAtAorticValveLevel',
         'ParasternalShortAxisAtTheAorticValveLevel',
         'SubcostalShortAxisViewAtPapillaryMuscleLevel',
         'SubcostalShortAxisViewAtVenousInflowLevel',
         'ApicalFourChamber',
         'ParasternalShortAxisAtTheLevelOfTheMitralChords',
         'RightVentricularOutflowTractView',
         'ParasternalShortAxis']}

name_for_cid[6066] = 'TargetConfirmation'
cid_concepts[6066] = \
{'DCM': ['TargetContainedInTheSpecimen',
         'TargetPartiallyObtainedInTheSpecimen',
         'TargetNotInTheSpecimen',
         'CalcificationsSeenInTheCore',
         'LesionCompletelyRemoved',
         'LesionPartiallyRemoved',
         'FluidObtained']}

name_for_cid[6089] = 'Substances'
cid_concepts[6089] = \
{'SCT': ['LysergicAcidDiethlyamide',
         'Methlyphenidate',
         'Methadone',
         'Phencyclidine',
         'Morphine',
         'Cocaine',
         'Heroin',
         'Amphetamine',
         'Caffeine',
         'Marijuana',
         'EthylAlcohol',
         'ChewingTobacco',
         'CigaretteSmokingTobacco',
         'Mescaline']}

name_for_cid[6123] = 'ChestOpacityDescriptor'
cid_concepts[6123] = \
{'DCM': ['Linear',
         'Profusion',
         'SilhouetteSign',
         'Acinar',
         'AirSpace',
         'Fibronodular',
         'Fluffy']}

name_for_cid[4260] = 'OphthalmicMappingUnitsforRealWorldValueMapping'
cid_concepts[4260] = \
{'UCUM': ['Micrometer']}

name_for_cid[9517] = 'RadiotherapyBlockDeviceTypes'
cid_concepts[9517] = \
{'DCM': ['ApertureBlock'], 'SCT': ['ShieldingBlock']}

name_for_cid[7159] = 'LesionSegmentationTypes'
cid_concepts[7159] = \
{'DCM': ['Background'],
 'NCIt': ['EnhancingLesion', 'NonEnhancingLesion'],
 'SCT': ['Embolus',
         'Mass',
         'Normal',
         'Cyst',
         'Abscess',
         'Edema',
         'NeoplasmPrimary',
         'BloodClot',
         'Nodule',
         'Neoplasm',
         'NeoplasmSecondary',
         'Necrosis',
         'Hemorrhage',
         'Inflammation']}

name_for_cid[12271] = 'CardiacUltrasoundOutflowTractsMeasurements'
cid_concepts[12271] = \
{'DCM': ['Tau', 'IndicatorMeanTransitTime', 'RRInterval', 'WallMass'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'TissueVelocityTimeIntegralVTIForTheAreaUnderLeftVentricleAWave',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'AWaveAmplitude',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'LeafletThickness',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'EndocardialArea',
        'PeakTissueVelocity',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'LeafletSeparation',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeToLeftVentricleETissueVelocity',
        'PercentThickening',
        'ACInterval',
        'BCSlope',
        'PreEjectionPeriod',
        'TissueVelocityTimeIntegralVTIForTheAreaUnderLeftVentricleEWave',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'TimeFromQWaveToPulmonicValveCloses',
        'CEDistance',
        'DpDtByUS',
        'PulmonaryArteryPressureUsingAccelTime',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'InterventricularSeptumToPosteriorWallThicknessRatio',
        'IsovolumicRelaxationTime',
        'TimeToLeftVentricleSTissueVelocity',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'LeftVentricleMeridionalWallStress',
        'LeftVentricleEToATissueVelocityRatio',
        'SystolicToDiastolicVelocityRatio',
        'AorticValveCuspSeparation',
        'EpicardialArea',
        'MyocardialPerformanceIndexTei',
        'TimeAveragedMeanVelocity',
        'DEExcursion'],
 'SCT': ['MajorAxis',
         'Volume',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'RatioOfMVPeakVelocityToLVPeakTissueVelocityEWave',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[6031] = 'BenignPathologyCodesfromBIRADS®'
cid_concepts[6031] = \
{'DCM': ['FibroadenomatoidHyperplasia',
         'Fibroadenolipoma',
         'PseudoangiomatousStromalHyperplasia',
         'NormalAxillaryNode',
         'AxillaryNodeWithCalcifications',
         'SiliconeGranuloma',
         'AxillaryNodeHyperplasia',
         'VirginalHyperplasia',
         'BenignCystWithBlood',
         'OilCyst',
         'PostReductionMammoplasty',
         'BenignCalcifications',
         'DuctalAdenoma',
         'DiabeticFibrousMastopathy',
         'LargeDuctPapilloma',
         'MicroglandularAdenosis',
         'MultipleIntraductalPapillomas',
         'NoAbnormality',
         'NormalBreastTissue',
         'JuvenilePapillomatosis',
         'LactationalChange'],
 'SCT': ['Hemangioma',
         'LactatingAdenoma',
         'CystOfBreast',
         'FatNecrosisOfBreast',
         'InfarctionOfBreast',
         'Adenoma',
         'Leiomyoma',
         'RadialScar',
         'HemangiomaOfSubcutaneousTissue',
         'LymphNode',
         'HemangiomaVenous',
         'BreastLobularHyperplasia',
         'GranularCellTumor',
         'Angiolipoma',
         'EpidermalInclusionCyst',
         'Galactocele',
         'ForeignBody',
         'HyperplasiaUsual',
         'ThrombophlebitisOfBreastMondorDisease',
         'ScarTissue',
         'Angiomatosis',
         'Adenosis',
         'DuctalHyperplasiaUsual',
         'Seroma',
         'Gynecomastia',
         'ExtraAbdominalDesmoid',
         'GiantFibroadenoma',
         'Fibromatosis',
         'Inflammation',
         'IntramammaryLymphNode',
         'ApocrineMetaplasia',
         'TubularAdenoma',
         'IntraductalPapilloma',
         'Hematoma',
         'Amyloid',
         'Neurofibromatosis',
         'PleomorphicAdenoma',
         'Neurofibroma',
         'Adenomyoepithelioma',
         'FocalFibrosis',
         'IntracysticPapilloma',
         'Myofibroblastoma',
         'Lipoma',
         'JuvenileFibroadenoma',
         'EctopicAccessoryBreastTissue',
         'Abscess',
         'SclerosingAdenosis',
         'Edema',
         'Papilloma',
         'Adenolipoma',
         'Hamartoma',
         'AsynchronousInvolutionOfBreast',
         'Fibroadenoma',
         'FibrocysticDiseaseOfBreast',
         'MammaryDuctEctasia',
         'Chondroma',
         'BenignNeoplasmOfNippleOfFemaleBreastNippleAdenoma']}

name_for_cid[3233] = 'STSegmentMorphology'
cid_concepts[3233] = \
{'DCM': ['STDepressionHorizontal',
         'STDepressionUpsloping',
         'STDepressionDownsloping'],
 'SCT': ['STElevation', 'STDepression']}

name_for_cid[3663] = 'BodySurfaceAreaEquations'
cid_concepts[3663] = \
{'DCM': ['BSAEquals0Point038189WT0Point423HT0Point362',
         'BSAEquals13210Point3433WT',
         'BSAEquals0Point010265WT0Point423HT0Point651',
         'BSAEquals0Point0004688WT0Point8168To0Point0154LogWT',
         'BSAEquals0Point008883WT0Point444HT0Point663',
         'BSAEquals0Point024265WT0Point5378Htcm0Point3964',
         'BSAEqualsHTWT360Point5',
         'BSAEquals0Point007358WT0Point425HT0Point725',
         'BSAEquals0Point007184WT0Point425HT0Point725',
         'BSAEquals0Point0235WT0Point51456Htcm0Point42246',
         'BSAEquals0Point003207WT0Point7285To0Point0188LogWTHT0Point3',
         'BSAEquals0Point009568WT0Point473HT0Point655']}

name_for_cid[7270] = 'MRDiffusionComponentSemantics'
cid_concepts[7270] = \
{'DCM': ['VolumetricDiffusionDxzComponent',
         'VolumetricDiffusionDxyComponent',
         'DiffusionWeighted',
         'VolumetricDiffusionDxxComponent',
         'VolumetricDiffusionDzzComponent',
         'VolumetricDiffusionDyzComponent',
         'VolumetricDiffusionDyyComponent']}

name_for_cid[10020] = 'SourceofProjectionXRayDoseInformation'
cid_concepts[10020] = \
{'DCM': ['SystemCalculated',
         'CopiedFromImageAttributes',
         'ManualEntry',
         'DerivedFromHumanReadableReports',
         'AutomatedDataCollection',
         'ComputedFromImageAttributes',
         'MPPSContent'],
 'SCT': ['Dosimeter']}

name_for_cid[3726] = 'IndicationsforCoronaryIntervention'
cid_concepts[3726] = \
{'DCM': ['CoronaryLesionGreaterThanEquals50PercentStenosis'],
 'SCT': ['CardiogenicShock']}

name_for_cid[3419] = 'FindingsTitles'
cid_concepts[3419] = \
{'DCM': ['Recommendation', 'Impression', 'Finding']}

name_for_cid[646] = 'PreclinicalSmallAnimalImagingProcedures'
cid_concepts[646] = \
{'DCM': ['SPECTOfWholeBody', 'SPECTCTOfWholeBody'],
 'LN': ['MRIWholeBody',
        'HeadCT',
        'RadionuclideScanOfWholeBody',
        'BrainMRI',
        'RadionuclideBrainScan',
        'PETWholeBody',
        'WholeBodyCT',
        'BrainPET',
        'BrainSPECT'],
 'SCT': ['UltrasonographyOfTotalBody']}

name_for_cid[12288] = 'CardiacUltrasoundOutflowTractsFindingSites'
cid_concepts[12288] = \
{'SCT': ['RastelliInterventricularTunnel',
         'PulmonicValve',
         'LeftVentricleToPulmonaryArteryConduitAnastomosis',
         'RightVentricle',
         'LeftVentricle',
         'RightVentricleToPulmonaryArteryConduitAnastomosis',
         'AorticValve']}

name_for_cid[12267] = 'CardiacUltrasoundTricuspidValve'
cid_concepts[12267] = \
{'DCM': ['Tau',
         'IndicatorMeanTransitTime',
         'RRInterval',
         'BloodVelocityPeak',
         'BloodVelocityMean',
         'BloodVelocityMinimum',
         'SystolicBloodVelocityMean',
         'SystolicBloodVelocityPeak',
         'DiastolicBloodVelocityMean',
         'DiastolicBloodVelocityPeak'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'TimeFromQWaveToTricuspidValveOpens',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'LeafletThickness',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'LeafletSeparation',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'CEDistance',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'DEExcursion',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'TricuspidDiastolicFillingPeriodDfpt',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[703] = 'BasisofAssessment'
cid_concepts[703] = \
{'DCM': ['AssessmentByRules', 'AssessmentByComparison']}

name_for_cid[9501] = 'PrescriptionAnatomyCategories'
cid_concepts[9501] = \
{'DCM': ['ExternalBodyModel', 'NonSpecificVolume'],
 'SCT': ['AnatomicalStructure', 'PhysicalObject']}

name_for_cid[6027] = 'AssessmentfromBIRADS®'
cid_concepts[6027] = \
{}

name_for_cid[3730] = 'InsurancePayor'
cid_concepts[3730] = \
{}

name_for_cid[6162] = 'AssessmentofMetastasisforBreast'
cid_concepts[6162] = \
{'SCT': ['M1DistantMetastasis',
         'MXDistantMetastasisCannotBeAssessed',
         'M0NoDistantMetastasis']}

name_for_cid[241] = 'PresentAbsentOnly'
cid_concepts[241] = \
{'SCT': ['Absent', 'Present']}

name_for_cid[8202] = 'SurfaceScanModeTypes'
cid_concepts[8202] = \
{'DCM': ['HighResolution', 'FastMode', 'Checkerboard', 'TurntableScanMethod']}

name_for_cid[611] = 'AnesthesiaCategoryCodeTypeforSmallAnimalAnesthesia'
cid_concepts[611] = \
{'SCT': ['MonitoredAnesthesiaCareMAC',
         'GeneralAnesthesia',
         'LocalAnesthesia',
         'Sedation',
         'SpinalAnesthesia',
         'EpiduralAnesthesia',
         'RegionalAnesthesia',
         'TopicalLocalAnesthesia']}

name_for_cid[4221] = 'VisualFixationQualityProblem'
cid_concepts[4221] = \
{'DCM': ['EquipmentFailure', 'OperatorError', 'PatientMovement'],
 'SCT': ['EccentricFixation']}

name_for_cid[12230] = 'GradientMethods'
cid_concepts[12230] = \
{'DCM': ['SimplifiedBernoulli', 'FullBernoulli']}

name_for_cid[6070] = 'BloomRichardsonHistologicGrade'
cid_concepts[6070] = \
{'SCT': ['Grade3PoorlyDifferentiated',
         'Grade1WellDifferentiated',
         'Grade4Undifferentiated',
         'Grade2ModeratelyDifferentiated']}

name_for_cid[12119] = 'VascularUltrasoundProperty'
cid_concepts[12119] = \
{'LN': ['MinimumDiastolicVelocity',
        'ResistivityIndex',
        'PeakGradient',
        'MeanGradient',
        'PeakSystolicVelocity',
        'VolumeFlow',
        'AccelerationIndex',
        'VelocityRatio',
        'DecelerationTime',
        'TimeAveragedPeakVelocity',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'PulsatilityIndex',
        'EndDiastolicVelocity',
        'TimeAveragedMeanVelocity'],
 'SCT': ['VesselLumenDiameter',
         'LengthOfSegment',
         'VesselLumenCrossSectionalArea',
         'VesselDepthFromSurface',
         'VesselOutsideDiameter',
         'VesselIntimalCrossSectionalArea',
         'LumenDiameterStenosis',
         'LumenAreaStenosis',
         'VesselIntimalDiameter']}

name_for_cid[42] = 'NumericValueQualifier'
cid_concepts[42] = \
{'DCM': ['ValueUnknown',
         'NegativeInfinity',
         'NotANumber',
         'ValueIndeterminate',
         'DivideByZero',
         'PositiveInfinity',
         'Overflow',
         'Underflow',
         'MeasurementNotAttempted',
         'MeasurementFailure',
         'ValueOutOfRange',
         'CalculationFailure']}

name_for_cid[10061] = 'AbsorbedRadiationDoseTypes'
cid_concepts[10061] = \
{'DCM': ['MinimumAbsorbedRadiationDose',
         'MaximumAbsorbedRadiationDose',
         'ModeAbsorbedRadiationDose',
         'MeanAbsorbedRadiationDose',
         'MedianAbsorbedRadiationDose']}

name_for_cid[3337] = 'HemodynamicAnnotations'
cid_concepts[3337] = \
{'DCM': ['BeatDetected',
         'AWavePeakPressure',
         '_70PercentOfThermalDyeDilutionCO',
         'ValveOpen',
         'BeatDetected',
         'AWavePressureAverage',
         'VWavePeakPressure',
         '_35PercentOfThermalDyeDilutionCO',
         'ValveClose',
         'VWavePressureAverage',
         'StartOfInspiration',
         'VMaxMyocardial',
         'StartOfThermalCO',
         'EndOfSystole',
         'Tau',
         'IndicatorMeanTransitTime',
         'MaxNegDpDt',
         'StartOfExpiration',
         'PeakOfThermalCardiacOutputBolus',
         'EndOfExpiration',
         'MaxDpDt',
         'EndOfInspiration'],
 'SCT': ['AverageDiastolicBloodPressure',
         'MinimumDiastolicBloodPressure',
         'MaximumSystolicBloodPressure',
         'MeanBloodPressure',
         'EndDiastole',
         'AverageSystolicBloodPressure']}

name_for_cid[6135] = 'ImageQualityFinding'
cid_concepts[6135] = \
{'DCM': ['OtherFailure',
         'MotionBlur',
         'UnknownFailure',
         'UnderExposed',
         'Positioning',
         'OverExposed',
         'NoImage',
         'DetectorArtifacts',
         'ArtifactsOtherThanGridOrDetectorArtifact',
         'MechanicalFailure',
         'ElectricalFailure',
         'SoftwareFailure',
         'GridArtifacts',
         'InappropriateImageProcessing'],
 'RADLEX': ['BeamHardeningArtifact']}

name_for_cid[4108] = 'PerfusionModelParameters'
cid_concepts[4108] = \
{'DCM': ['MeanTransitTime',
         'RelativeRegionalBloodFlow',
         'RelativeRegionalBloodVolume',
         'TimeToPeak',
         'AbsoluteRegionalBloodFlow',
         'AbsoluteRegionalBloodVolume',
         'OxygenExtractionFraction',
         'Tmax']}

name_for_cid[3120] = 'PerfusionFindings'
cid_concepts[3120] = \
{'SCT': ['Abnormal', 'Equivocal', 'MyocardialPerfusionNormal']}

name_for_cid[3823] = 'RespiratoryStatus'
cid_concepts[3823] = \
{'DCM': ['CentralBreathingPosition'],
 'SCT': ['ValsalvaManeuver',
         'Expiration',
         'ShallowBreathing',
         'Inspiration',
         'AutonomousBreathing']}

name_for_cid[7026] = 'RadiotherapeuticDoseMeasurementDevices'
cid_concepts[7026] = \
{'DCM': ['_3DGel',
         'IonChamberArray',
         'DiodeArray',
         'LiquidIonChamber',
         'Diode',
         'IonChamber',
         'OSLD',
         'DiamondDetector'],
 'SCT': ['MOSFETRadiationTherapyDosimetrySystemDosimeter',
         'DigitalImager',
         'MedicalXRayFilm',
         'ThermoluminescentRadiationDosimeter']}

name_for_cid[7476] = 'GrayLevelSizeZoneBasedFeatures'
cid_concepts[7476] = \
{}

name_for_cid[3209] = 'SummaryCodesStressImaging'
cid_concepts[3209] = \
{'SCT': ['ImagingResultEquivocal',
         'ImagingResultNormal',
         'ImagingResultAbnormal',
         'NotPerformed']}

name_for_cid[3527] = 'BloodBaseExcess'
cid_concepts[3527] = \
{'LN': ['BloodBaseExcess', 'VenousBloodBaseExcess', 'ArterialBloodBaseExcess']}

name_for_cid[3462] = 'ChamberIdentification'
cid_concepts[3462] = \
{'SCT': ['RightAtrium', 'RightVentricle', 'LeftVentricle', 'LeftAtrium']}

name_for_cid[6119] = 'ChestBorderShape'
cid_concepts[6119] = \
{'DCM': ['Elliptic', 'Lobulated', 'Spiculated'],
 'SCT': ['RoundShape', 'Irregular']}

name_for_cid[3531] = 'OxygenAdministration'
cid_concepts[3531] = \
{'DCM': ['OxygenAdministrationByVentilator'],
 'SCT': ['OxygenAdministrationByMask', 'OxygenAdministrationByNasalCannula']}

name_for_cid[8111] = 'SpecimenPreparationProcedure'
cid_concepts[8111] = \
{'DCM': ['SpecimenStorage'],
 'SCT': ['Staining',
         'SpecimenProcessing',
         'SpecimenCollection',
         'SamplingOfTissueSpecimen',
         'SpecimenReceiving']}

name_for_cid[502] = 'VolumetricViewModifier'
cid_concepts[502] = \
{'DCM': ['Crosstable'],
 'SCT': ['Transoral',
         'Caudal',
         'Sagittal',
         'Caudad',
         'FiveChamber',
         'LongAxis',
         'Parasagittal',
         'Transforamenal',
         'FourChamber',
         'OffAxis',
         'ShortAxis',
         'TwoChamber',
         'Oblique',
         'Cephalad',
         'Coronal',
         'Transthoracic',
         'Transverse',
         'Transorbital',
         'MouthClosed',
         'Cephalic',
         'Longitudinal']}

name_for_cid[3618] = 'HemodynamicFlows'
cid_concepts[3618] = \
{'DCM': ['SystemicFlow', 'PulmonaryFlow']}

name_for_cid[12123] = 'CarotidRatios'
cid_concepts[12123] = \
{'LN': ['ICACCAVelocityRatio']}

name_for_cid[3835] = 'CardiacVolumeMeasurements'
cid_concepts[3835] = \
{'DCM': ['LeftAtrialEDVolume', 'LeftAtrialESVolume'],
 'LN': ['RightVentricularESVolume',
        'LeftVentricularESVolume',
        'RightVentricularEDVolume',
        'LeftVentricularEDVolume'],
 'SCT': ['StrokeVolume']}

name_for_cid[3423] = 'NumericDeviceCharacteristics'
cid_concepts[3423] = \
{'DCM': ['TransmitFrequency', 'CatheterCurve', 'InterMarkerDistance'],
 'SCT': ['Length', 'Diameter', 'Volume']}

name_for_cid[7460] = 'UnitsofLinearMeasurement'
cid_concepts[7460] = \
{'UCUM': ['Millimeter', 'Centimeter', 'Micrometer']}

name_for_cid[7030] = 'InstitutionalDepartments,UnitsandServices'
cid_concepts[7030] = \
{'DCM': ['PediatricRadiology',
         'ThoracicRadiology',
         'AbdominalRadiology',
         'CardiovascularRadiology',
         'BiomedicalEngineering',
         'MedicalPhysics',
         'InformationTechnology',
         'MusculoskeletalRadiology'],
 'SCT': ['Psychology',
         'Ultrasonography',
         'CardiacSurgery',
         'Dermatology',
         'GeneralMedicine',
         'MagneticResonanceImagingService',
         'CardiothoracicSurgery',
         'IntensiveCare',
         'GeneralSurgery',
         'TropicalMedicine',
         'AllergyAndImmunology',
         'Nephrology',
         'DiagnosticImaging',
         'Neurosurgery',
         'OralSurgery',
         'CardiacIntensiveCare',
         'Otorhinolaryngology',
         'ChildAndAdolescentPsychiatry',
         'PediatricSurgery',
         'Urology',
         'Cytology',
         'Radiotherapy',
         'ThoracicSurgery',
         'ClinicalBiochemistry',
         'GastrointestinalSurgery',
         'AccidentAndEmergency',
         'Endoscopy',
         'Physiotherapy',
         'Rheumatology',
         'NeonatalIntensiveCare',
         'Stroke',
         'MedicalIntensiveCare',
         'PainManagement',
         'PediatricIntensiveCare',
         'Ophthalmology',
         'Audiology',
         'ObstetricsAndGynecology',
         'Pulmonology',
         'MedicalMicrobiology',
         'TraumaSurgery',
         'HepatobiliarySurgery',
         'PalliativeCare',
         'PediatricOncology',
         'SurgicalIntensiveCare',
         'Obstetrics',
         'ClinicalOncology',
         'Endocrinology',
         'Rehabilitation',
         'BurnsIntensiveCare',
         'Surgery',
         'ColorectalSurgery',
         'Optometry',
         'ComputerizedTomographyService',
         'InfectiousDisease',
         'Cardiology',
         'NuclearMedicine',
         'DentalSurgery',
         'SpeechAndLanguageTherapy',
         'TransplantSurgery',
         'Anesthesiology',
         'PrimaryCareDepartment',
         'BreastSurgery',
         'Histopathology',
         'Radiology',
         'HandSurgery',
         'Gynecology',
         'OrthopedicSurgery',
         'InterventionalRadiologyService',
         'PediatricMedicine',
         'Psychiatry',
         'Hematology',
         'Pathology',
         'PlasticSurgery',
         'Neurology',
         'EndocrineSurgery',
         'GeriatricMedicine',
         'VascularSurgery'],
 'UMLS': ['Neuroradiology']}

name_for_cid[6158] = 'CorrelationtoOtherFindings'
cid_concepts[6158] = \
{'DCM': ['CorrelatesToPhysicalExamFindings',
         'CorrelatesToMammographyFindings',
         'CorrelatesToMRIFindings',
         'CorrelatesToUltrasoundFindings',
         'CorrelatesToOtherImagingFindings',
         'NoCorrelationToOtherImagingFindings',
         'NoCorrelationToClinicalFindings']}

name_for_cid[12031] = 'ProtocolIntervalEvents'
cid_concepts[12031] = \
{'DCM': ['CessationOfStimulation',
         'OnsetOfStimulation',
         'CessationOfExercise',
         'OnsetOfExercise',
         'DestructionOfMicrobubbles',
         'StartOfContrastAgentAdministration',
         'StartOfDrugDoseAdministration']}

name_for_cid[9242] = 'RadiotherapyAcquisitionWorkitemDefinition'
cid_concepts[9242] = \
{'DCM': ['RTPatientPositionAcquisitionSinglePlaneMV',
         'RTPatientPositionAcquisitionUltrasound',
         'RTPatientPositionAcquisitionSpatialFiducials',
         'RTPatientPositionAcquisitionOptical',
         'RTPatientPositionAcquisitionCTKv',
         'RTPatientPositionAcquisitionCTMV',
         'RTPatientPositionAcquisitionDualPlaneKv',
         'RTPatientPositionAcquisitionDualPlaneKvMV',
         'RTPatientPositionAcquisitionDualPlaneMV',
         'RTPatientPositionAcquisitionSinglePlaneKv']}

name_for_cid[4020] = 'PETRadionuclide'
cid_concepts[4020] = \
{'DCM': ['_51Manganese',
         '_70Arsenic',
         '_90Niobium',
         '_43Scandium',
         '_152Terbium',
         '_52mManganese',
         '_44Scandium'],
 'SCT': ['_22Sodium',
         '_73Selenium',
         '_60Copper',
         '_14Oxygen',
         '_66Gallium',
         '_76Bromine',
         '_45Titanium',
         '_68Germanium',
         '_89Zirconium',
         '_15Oxygen',
         '_82Rubidium',
         '_11Carbon',
         '_62Copper',
         '_68Gallium',
         '_75Bromine',
         '_18Fluorine',
         '_62Zinc',
         '_64Copper',
         '_77Bromine',
         '_94mTechnetium',
         '_90Yttrium',
         '_124Iodine',
         '_38Potassium',
         '_86Yttrium',
         '_61Copper',
         '_52Iron',
         '_52Manganese',
         '_72Arsenic',
         '_13Nitrogen']}

name_for_cid[12103] = 'VascularUltrasoundAnatomicLocation'
cid_concepts[12103] = \
{'SCT': ['SubclavianVein',
         'ProfundaFemorisVein',
         'InternalCarotidArtery',
         'ExternalIliacArtery',
         'UlnarVein',
         'CeliacAxis',
         'OvarianVein',
         'PosteriorCerebralArtery',
         'BasilarArtery',
         'DeepPalmarVenousArch',
         'LateralCalfPerforator',
         'InferiorVenaCava',
         'CentralRetinalVein',
         'ProperHepaticArtery',
         'InterlobarArteryOfKidney',
         'GastricArtery',
         'FourthLumbarArtery',
         'AnteriorMiddleCerebralArteryBifurcation',
         'ThirdLumbarArtery',
         'PerforatingArteryOfKidney',
         'UmbilicalArtery',
         'InternalCarotidArteryC6Segment',
         'RadialArtery',
         'AnteriorTibialArtery',
         'SolealVein',
         'CommonCarotidArtery',
         'LobarArtery',
         'GreatSaphenousVeinOfCalf',
         'Aorta',
         'ThighPerforator',
         'CommonIliacArteryBifurcation',
         'MiddleHepaticVein',
         'RightMainBranchOfPortalVein',
         'RadialVein',
         'RightBranchOfHepaticArtery',
         'ExternalIliacVein',
         'CommonHepaticArtery',
         'UterineArtery',
         'SubclavianArtery',
         'CentralRetinalArtery',
         'FirstLumbarArtery',
         'OphthalmicArtery',
         'PoplitealArtery',
         'SegmentalArtery',
         'MiddleCerebralArteryM1Segment',
         'BrachialVein',
         'InternalIliacArtery',
         'SuperiorVenaCava',
         'SuperficialPalmarArch',
         'UlnarArtery',
         'OvarianArtery',
         'SixthLumbarArtery',
         'PosteriorArchVein',
         'SplenicVein',
         'PeronealArtery',
         'CarotidBifurcation',
         'SaphenousVein',
         'SuperficialFemoralVein',
         'SplenicArtery',
         'InnominateArtery',
         'RenalArtery',
         'SuperiorMesentericVein',
         'GastroduodenalArtery',
         'LumbarArtery',
         'AccessoryRenalArtery',
         'PeronealVein',
         'BrachialArtery',
         'DorsalisPedisArtery',
         'InfraRenalAorta',
         'CommonFemoralArtery',
         'SuperficialPalmarVenousArch',
         'CarotidBulb',
         'CommonFemoralVein',
         'PoplitealVein',
         'GiacominiVein',
         'InternalCarotidArteryC5Segment',
         'InferiorMesentericVein',
         'AnteriorCommunicatingArtery',
         'MedianCubitalVein',
         'HilarArtery',
         'TerminalInternalCarotidArtery',
         'AxillaryArtery',
         'HepaticVein',
         'VertebralArtery',
         'TransjugularIntrahepaticPortosystemicShunt',
         'SaphenofemoralJunction',
         'SupraRenalAorta',
         'InternalJugularVein',
         'CockettPerforatingVein',
         'ExternalCarotidArtery',
         'SuperiorMesentericArtery',
         'UmbilicalVein',
         'CommonIliacVein',
         'PosteriorTibialArtery',
         'PosteriorCerebralArteryP1Segment',
         'TesticularArtery',
         'GreatSaphenousVeinOfThigh',
         'LeftHepaticVein',
         'InnominateVein',
         'SuperficialFemoralArtery',
         'PortalVein',
         'PosteriorCommunicatingArtery',
         'AnteriorCerebralArtery',
         'LeftBranchOfHepaticArtery',
         'RenalVein',
         'InternalIliacVein',
         'BoydPerforatingVein',
         'IlealVein',
         'ArcuateArteryOfTheKidney',
         'CommonIliacArtery',
         'RightHepaticVein',
         'CephalicVein',
         'GastrocnemiusVein',
         'GastricVein',
         'MiddleCerebralArteryM2Segment',
         'LeftMainBranchOfPortalVein',
         'MiddleCerebralArtery',
         'PosteriorCerebralArteryP2Segment',
         'InferiorRightHepaticVein',
         'TesticularVein',
         'CarotidSiphon',
         'AnteriorTibialVein',
         'PosteriorTibialVein',
         'SaphenopoplitealJunction',
         'BasilicVein',
         'PlantarArterialArch',
         'HunterianPerforatingVein',
         'AnteriorPosteriorCerebralArteryBifurcation',
         'DeepPalmarArchOfRadialArtery',
         'ProfundaFemorisArtery',
         'FifthLumbarArtery',
         'AxillaryVein',
         'LesserSaphenousVein',
         'InferiorMesentericArtery',
         'DigitalArteryOfHand',
         'GreatSaphenousVein',
         'SecondLumbarArtery']}

name_for_cid[8131] = 'PathologyImagingProtocols'
cid_concepts[8131] = \
{'DCM': ['WSI40XRGB',
         'PeriOperativePhotographicImaging',
         'GrossSpecimenImaging',
         'SlideMicroscopy',
         'WholeSlideImaging',
         'WSI20XRGB']}

name_for_cid[6085] = 'TherapiesforBreast'
cid_concepts[6085] = \
{'SCT': ['HormoneTherapy',
         'BoneMarrowTransplant',
         'Chemotherapy',
         'RadiationTherapy']}

name_for_cid[3004] = 'ArterialPulseWaveform'
cid_concepts[3004] = \
{'DCM': ['ArterialPulseWaveform']}

name_for_cid[6206] = 'RecumbentPatientOrientationforColon'
cid_concepts[6206] = \
{'SCT': ['LeftLateralDecubitus', 'Supine', 'RightLateralDecubitus', 'Prone']}

name_for_cid[3680] = 'ECGLeadNoiseDescriptions'
cid_concepts[3680] = \
{'MDC': ['ModerateNoise', 'SevereNoise', 'NoNoise', 'NoSignal']}

name_for_cid[12011] = 'UltrasoundPelvisandUterus'
cid_concepts[12011] = \
{'LN': ['EndometriumThickness', 'CervixLength']}

name_for_cid[3403] = 'PercutaneousEntry'
cid_concepts[3403] = \
{'SCT': ['ViaFemoralVein',
         'ViaFemoralArtery',
         'ViaArtery',
         'ViaBrachialArtery',
         'ViaArmVein',
         'ViaRadialArtery',
         'ViaVein']}

name_for_cid[7010] = 'KeyObjectSelectionDocumentTitle'
cid_concepts[7010] = \
{'DCM': ['OfInterest',
         'RejectedForQualityReasons',
         'ForReferringProvider',
         'ForSurgery',
         'ForTeaching',
         'ContouringInputUsed',
         'DiagnosisInputUsed',
         'RTTreatmentPlanningInputUsed',
         'RTPrescriptionInputUsed',
         'PhysicianReviewInputUsed',
         'RTPlanSummationInputUsed',
         'PlanQualityAssuranceInputUsed',
         'PhysicistReviewInputUsed',
         'PatientSetupVerificationInputUsed',
         'MachineQualityAssuranceInputUsed',
         'ForConference',
         'ForTherapy',
         'ForPatient',
         'ForPeerReview',
         'ForResearch',
         'Manifest',
         'SignedManifest',
         'CompleteStudyContent',
         'SignedCompleteStudyContent',
         'CompleteAcquisitionContent',
         'SignedCompleteAcquisitionContent',
         'GroupOfFramesForDisplay',
         'RejectedForPatientSafetyReasons',
         'ForTumorRegistry',
         'ForClinicalTrialSubmission',
         'RTWorkflowInputUsed',
         'ForRTPlanSummation',
         'ForPhysicistReview',
         'ForPhysicianReview',
         'ForPlanQualityAssurance',
         'ForTumorBoard',
         'ForPatientSetupVerification',
         'ForMachineQualityAssurance',
         'IncorrectModalityWorklistEntry',
         'DataRetentionPolicyExpired',
         'ForReportAttachment',
         'ForLitigation',
         'CollectionOfPresentationStates',
         'ForPublicationExport',
         'ForResearchCollectionExport',
         'ForPlanComparison',
         'ForTeachingFileExport',
         'ForClinicalTrialExport',
         'RTTreatmentSessionResult',
         'PatientSetupVerificationResult',
         'RTTreatmentQAResult',
         'RTTreatmentCourseSummary',
         'ForSegmentation',
         'ForDiagnosis',
         'ForRTTreatmentPlanning',
         'ForRTPrescription',
         'QualityIssue',
         'BestInSet',
         'DosimetricResult',
         'TumorRegistryInputUsed',
         'RTPlanningResult',
         'TumorBoardInputUsed',
         'RTTreatmentSimulationResult',
         'ClinicalTrialSubmissionInputUsed',
         'DiagnosticSourceImages',
         'ForRTWorkflow',
         'ForPrinting',
         'RegistrationResult',
         'SegmentationResult',
         'RTPrescriptionResult',
         'PrePlanningResult',
         'PlanComparisonInputUsed',
         'CoordinateAlignmentImageSeries',
         'DoseCalculationImageSeries']}

name_for_cid[12292] = 'CardiacUltrasoundCoronaryArteriesFindingSites'
cid_concepts[12292] = \
{'SCT': ['CircumflexCoronaryArtery',
         'AnteriorDescendingBranchOfLeftCoronaryArtery',
         'ProximalCircumflexCoronaryArtery',
         'DistalCircumflexCoronaryArtery',
         'PosteriorDescendingCoronaryArtery',
         'SeptalArtery',
         'RightCoronaryArtery',
         'MidCircumflexCoronaryArtery',
         'LeftMainCoronaryArtery']}

name_for_cid[3815] = 'SourceofVascularFinding'
cid_concepts[3815] = \
{'SCT': ['Embolism',
         'AfterProcedure',
         'Trauma',
         'Hyperplasia',
         'Tumor',
         'Thrombosis',
         'Surgical',
         'Vasculitis']}

name_for_cid[7155] = 'ThoracicSegmentationTypes'
cid_concepts[7155] = \
{'SCT': ['Lung',
         'LowerLobeOfLung',
         'Mediastinum',
         'Clavicle',
         'Esophagus',
         'MiddleLobeOfRightLung',
         'BloodVessel',
         'Bronchus',
         'Diaphragm',
         'Pleura',
         'Thymus',
         'BoneOfThorax',
         'Thorax',
         'Rib',
         'ChestWallMuscle',
         'SkinOfChest',
         'Trachea',
         'UpperLobeOfLung',
         'Sternum',
         'ThoracicSpine']}

name_for_cid[3116] = 'NMExtracardiacFindings'
cid_concepts[3116] = \
{'SCT': ['IncreasedLungUptake',
         'AbnormalExtracardiacUptake',
         'NormalExtracardiacUptake']}

name_for_cid[4016] = 'AnatomicRegionforIntraoralRadiography'
cid_concepts[4016] = \
{'SCT': ['TeethGumsAndSupportingStructures',
         'Mandible',
         'JawRegion',
         'Maxilla']}

name_for_cid[3229] = 'ECGAxisMeasurements'
cid_concepts[3229] = \
{'MDC': ['QRSAxis', 'PAxis', 'TAxis']}

name_for_cid[19] = 'PatientOrientation'
cid_concepts[19] = \
{'NCIt': ['Erect'], 'SCT': ['Recumbent', 'SemiErect']}

name_for_cid[12007] = 'FetalCranium'
cid_concepts[12007] = \
{'LN': ['NuchalFoldThickness',
        'InnerOrbitalDiameter',
        'CisternaMagnaLength',
        'PosteriorHornLateralVentricularWidth',
        'OuterOrbitalDiameter',
        'WidthOfHemisphere',
        'LateralVentricleWidth',
        'AnteriorHornLateralVentricularWidth',
        'NuchalTranslucency',
        'TransCerebellarDiameter']}

name_for_cid[12284] = 'CardiacUltrasoundAtriaandAtrialSeptumFindingSites'
cid_concepts[12284] = \
{'SCT': ['RightAtrium',
         'InteratrialSeptumStructure',
         'FontanBaffleLeak',
         'LimbusOfFossaOvalis',
         'LeftAtrium',
         'LeftAuricularAppendage',
         'CommonAtrium',
         'PatentForamenOvale',
         'FontanFenestration',
         'AtrialSeptalDefect',
         'RightAuricularAppendage']}

name_for_cid[7006] = 'SRDocumentPurposesofReference'
cid_concepts[7006] = \
{'DCM': ['ExtendedRenderingForPresentation',
         'SourceDocument',
         'CompleteRenderingForPresentation',
         'PartialRenderingForPresentation',
         'EquivalentCDADocument']}

name_for_cid[7456] = 'UnitsofMeasureforAge'
cid_concepts[7456] = \
{'UCUM': ['Year', 'Minute', 'Hour', 'Week', 'Day', 'Month']}

name_for_cid[3415] = 'CardiacRhythms'
cid_concepts[3415] = \
{'MDC': ['AVDissociation',
         'VentricularCouplet',
         'FirstDegreeAVBlock',
         'NonRespiratorySinusArrhythmia',
         'VentricularParasystole',
         'NonsustainedVentricularTachycardia',
         'AVJunctionalNodalEscapeRhythm',
         'IrregularRhythm',
         'WolfParkinsonWhiteSyndrome',
         'Bradycardia',
         'LownGanongLevineSyndrome',
         'CompleteAVDissociation',
         'AutomaticAtrialTachycardia',
         'SecondDegreeAVBlockTypeIIMobitzTypeII',
         'AtrialEctopicRhythm',
         'TorsadesDePointesVentricularTachycardia',
         'SinusArrhythmia',
         'WanderingSinusPacemakerWithinTheSinusNode',
         'AcceleratedIdioventricularRhythm',
         'SustainedMonomorphicVentricularTachycardia',
         'WolfParkinsonTypeA',
         'AcceleratedAVJunctionalNodalRhythm',
         'VentricularBigeminy',
         'CalibrationSignal',
         'SinusBradycardia',
         'AVReciprocatingTachycardia',
         'SupraventricularAtrialOrJunctionalEctopicRhythm',
         'VentricularRhythm',
         'VentricularFlutter',
         'SupraventricularParoxsysmalTachycardia',
         'VentricularDemandModePacing',
         'SecondDegreeSABlockTypeII',
         'AVDissociationWithInterference',
         'ParoxsysmalAtrialTachycardia',
         'VentricularRun',
         'SecondDegreeAVBlock',
         'LowHeartRateVariability',
         'PreExcitation',
         'RespiratorySinusArrhythmia',
         'IdioventricularVentricularEscapeRhythm',
         'VentricularFibrillation',
         'Asystole',
         'AVJunctionalNodalRhythm',
         'SinusRhythm',
         'FirstDegreeSABlock',
         'TWaveAlternans',
         'AtrialFlutter',
         'ThirdDegreeAVBlockCompleteAVBlock',
         'AtrialBigeminy',
         'AtrialDemandModePacing',
         'VentricularTrigeminy',
         'ReentrantAVNodalTachycardia',
         'JunctionalTachycardia',
         'SinusTachycardia',
         'SupraventricularTachycardia',
         'AntiTachycardiaPacing',
         'IsorhythmicAVDissociation',
         'MultifocalAtrialTachycardia',
         'VentricularTachycardia',
         'SecondDegreeAVBlockTypeIWenckebachOrMobitzTypeI',
         'WanderingPacemakerBetweenTheSinusNodeAndTheAVNode',
         'PolymorphicVentricularTachycardia',
         'WolfParkinsonTypeB',
         'SlowVentricularTachycardiaIdioventricularTachycardia',
         'ThirdDegreeSABlockCompleteSABlock',
         'NormalSinusRhythm',
         'SecondDegreeSABlockTypeIWenckebach',
         'AtrialFibrillation',
         'AtrialTachycardia']}

name_for_cid[3550] = 'CirculatorySupportActions'
cid_concepts[3550] = \
{'DCM': ['EndCirculatorySupport', 'BeginCirculatorySupport']}

name_for_cid[6302] = 'ProstateSectorAnatomyfromEuropeanConcensus16Sector(Minimal)Model'
cid_concepts[6302] = \
{'SCT': ['CentralZoneOfRightHalfProstate',
         'RightPosteromedialMiddlePeripheralZoneOfProstate',
         'CentralZoneOfLeftHalfProstate',
         'LeftBasalPeripheralZoneOfProstate',
         'LeftLateralMiddlePeripheralZoneOfProstate',
         'LeftPosteromedialMiddlePeripheralZoneOfProstate',
         'RightBasalPeripheralZoneOfProstate',
         'LeftApicalTransitionZoneOfProstate',
         'MaleExternalUrethralSphincter',
         'RightApicalTransitionZoneOfProstate',
         'LeftBasalPartTransitionZoneOfProstate',
         'RightBasalTransitionZoneOfProstate',
         'LeftSeminalVesicle',
         'LeftMiddleTransitionZoneOfProstate',
         'LeftApicalPeripheralZoneOfProstate',
         'RightApicalPeripheralZoneOfProstate',
         'RightMiddleTransitionZoneOfProstate',
         'RightSeminalVesicle',
         'RightLateralMiddlePeripheralZoneOfProstate']}

name_for_cid[12115] = 'RenalVessels'
cid_concepts[12115] = \
{'SCT': ['RenalArtery',
         'LobarArtery',
         'PerforatingArteryOfKidney',
         'AccessoryRenalArtery',
         'SegmentalArtery',
         'HilarArtery',
         'RenalVein',
         'ArcuateArteryOfTheKidney',
         'InterlobarArteryOfKidney']}

name_for_cid[6093] = 'QualitativeConceptsforUsage,ExposureAmount'
cid_concepts[6093] = \
{'DCM': ['RelativeDoseAmount',
         'RelativeAmountOfExposure',
         'RelativeAmountOfUse']}

name_for_cid[4104] = 'BolusArrivalTimeDerivationMethods'
cid_concepts[4104] = \
{'DCM': ['TemporalDerivativeExceedsThreshold',
         'TimeOfPeakConcentration',
         'TimeOfLeadingHalfPeakConcentration']}

name_for_cid[6139] = 'NonlesionModifiers'
cid_concepts[6139] = \
{'SCT': ['External', 'Internal']}

name_for_cid[6210] = 'LocationinIntestinalTract'
cid_concepts[6210] = \
{'SCT': ['Rectum',
         'SigmoidColon',
         'DescendingColon',
         'TransverseColon',
         'SplenicFlexureOfColon',
         'Cecum',
         'AscendingColon',
         'HepaticFlexureOfColon']}

name_for_cid[3442] = 'PeripheralPulseMethods'
cid_concepts[3442] = \
{'SCT': ['Doppler', 'Palpation']}

name_for_cid[3710] = 'CoronaryDominance'
cid_concepts[3710] = \
{'SCT': ['BalancedCoronaryDominance',
         'LeftCoronaryDominance',
         'RightCoronaryDominance']}

name_for_cid[6142] = 'CalculatedValue'
cid_concepts[6142] = \
{'DCM': ['MassOfParenchymalTissue',
         'MassOfBreast',
         'AreaOfVascularCalcification',
         'VolumeOfVascularCalcification',
         'PercentageOfVascularCalcification',
         'MassOfVascularCalcification',
         'AverageCalcificationDistanceInACalcificationCluster',
         'CalciumScore',
         'StandardDeviationDistanceOfCalcificationsInACluster',
         'BreastTissueDensity',
         'VolumeOfParenchymalTissue',
         'VolumeOfBreast',
         'CavityExtentAsPercentOfVolume',
         'CalcificationExtentAsPercentOfSurface',
         'CalcificationExtentAsPercentOfVolume']}

name_for_cid[7303] = 'ImplantTemplatesModifiedViewOrientations'
cid_concepts[7303] = \
{'DCM': ['AP45', 'AP45']}

name_for_cid[10016] = 'AnodeTargetMaterial'
cid_concepts[10016] = \
{'SCT': ['TungstenOrTungstenCompound',
         'MolybdenumOrMolybdenumCompound',
         'RhodiumOrRhodiumCompound']}

name_for_cid[3205] = 'IndicationsforPharmacologicalStressTest'
cid_concepts[3205] = \
{'DCM': ['PatientWeightExceedsEquipmentLimit'],
 'SCT': ['RecentMyocardialInfarction',
         'Asthenia',
         'AtaxiaOrIncoordination',
         'OpenWoundOfLowerLimb',
         'Cachexia',
         'TransientLimbParalysis',
         'PatientHasPacemaker',
         'FractureOfLowerLimb',
         'DependenceOnEnablingMachineOrDevice',
         'PeripheralVascularDisease',
         'LeftBundleBranchBlock',
         'ParalyticSyndrome',
         'GaitProblem',
         'RequestByPhysician',
         'PulmonaryDisease',
         'LowerLimbAmputation',
         'CannotReachTargetHeartRate']}

name_for_cid[6007] = 'CharacteristicsofMarginfromBIRADS®'
cid_concepts[6007] = \
{'DCM': ['AngularMargins'],
 'SCT': ['SpiculatedLesion',
         'CircumscribedLesion',
         'ObscuredLesion',
         'IndistinctLesion',
         'MicrolobulatedLesion']}

name_for_cid[12247] = 'Pediatric,FetalandCongenitalCardiacSurgicalInterventions'
cid_concepts[12247] = \
{'SCT': ['RastelliOperation',
         'CreationOfConduitRightAtriumToPulmonaryTrunk',
         'RightSidedBidirectionalGlennShuntProcedure',
         'LeftSidedBidirectionalGlennShuntProcedure',
         'ImplantationOfHeartValveWithTissueGraft',
         'BlalockTaussigShuntPulmonarySubclavianArteryAnastomosis',
         'HemiFontanOperation',
         'NorwoodTypeOperation',
         'PatentDuctusArteriosusCoilOrDeviceClosure',
         'PartialAnomalousPulmonaryVenousConnectionOperation',
         'CentralAortopulmonaryShuntOperation',
         'ImplantationOfHeartValveProsthesisOrSyntheticDevice',
         'RepairOfCoarctationOfAorta',
         'ConstructionOfLVToAortaTunnelWRVToPAValvedConduit',
         'RepairOfTotalAnomalousPulmonaryVenousConnection',
         'PercutaneousProstheticClosureOfAtrialSeptalDefect',
         'CoarctationOfTheAortaAngioplastyWithImplantOfStent',
         'VentricularSeptalDefectDeviceClosure',
         'RadicalAortopulmonaryReconstructWRVToPAValvelessConduit',
         'ArterialSwitchOperation',
         'CoarctationOfTheAortaBalloonAngioplasty',
         'DamusStanselKayeOperation',
         'ClosureOfDuctusArteriosusWithClip',
         'AtrialInversionOperationUsingAtrialWall',
         'RightGlennShuntProcedure',
         'LateralCavalFontanProcedure',
         'SanoProcedure',
         'RepairOfDefectOfTheAtrioventricularSeptum',
         'CorrectionOfVentricularSeptalDefect',
         'AtrialSeptalDefectRepair',
         'ImplantationOfBaffleAtrialOrInteratrial',
         'MustardOperation',
         'LeftGlennShuntProcedure']}

name_for_cid[9521] = 'RadiotherapyTreatmentEnergyUnit'
cid_concepts[9521] = \
{'UCUM': ['Kilovolt', 'Megavolt', 'Megaelectronvolt']}

name_for_cid[4256] = 'VisualFieldProcedureModifier'
cid_concepts[4256] = \
{'SCT': ['Diagnostic', 'Screening']}

name_for_cid[12302] = 'EchoFindingObservationTypes'
cid_concepts[12302] = \
{'DCM': ['BehaviorOfTheFindingSite', 'StructureOfTheFindingSite'],
 'SCT': ['HemodynamicMeasurements']}

name_for_cid[7180] = 'AbstractMultidimensionalImageModelComponentSemantics'
cid_concepts[7180] = \
{'DCM': ['ApparentDiffusionCoefficient',
         'DiffusionWeighted',
         'VolumeRatio',
         'DiffusionCoefficient',
         'WaterFatOutOfPhase',
         'FractionalOccupancySegmentation',
         'WaterFatInPhase',
         'FractionalProbabilisticSegmentation',
         'BinarySegmentation',
         'WaterFraction',
         'MRSignalIntensity',
         'XRayAttenuation',
         'FatFraction',
         'YBRICTCRComponent',
         'R2Coefficient',
         'YBRICTCBComponent',
         'ChiSquare',
         'YBRICTYComponent',
         'DW',
         'YBRPARTIALCRComponent',
         'AIC',
         'YBRPARTIALCBComponent',
         'BIC',
         'YBRPARTIALYComponent',
         'Echogenicity',
         'YBRRCTCRComponent',
         'YBRRCTCBComponent',
         'YBRRCTYComponent',
         'RelativeLinearStoppingPower',
         'RGBRComponent',
         'UltrasoundAttenuation',
         '',
         'SpeedOfSound',
         'Vp',
         'Perfusion',
         'Elasticity',
         'EquivalentDose',
         'AbsorbedDose',
         'YBRFULLCRComponent',
         'YBRFULLCBComponent',
         'YBRFULLYComponent',
         'RGBBComponent',
         'RGBGComponent',
         'CreatineCholineCitrateRatio',
         'MagnetizationTransferRatio',
         'IAUC90BN',
         'IAUC180BN',
         'CreatineAndCholine',
         'LipidAndLactate',
         'SpatialDisplacementYComponent',
         'SpatialDisplacementXComponent',
         'NuclearMedicineTomographicActivity',
         'IAUC',
         'NuclearMedicineProjectionActivity',
         'IAUC60',
         'IAUC90',
         'IAUC180',
         'IAUCBN',
         'IAUC60BN',
         'FlowVariance',
         'FlowVelocity',
         'TissueVelocity',
         'IndexedHemodynamicResistance',
         'HemodynamicResistance',
         'SpatialDisplacementZComponent',
         'AttenuationCoefficient',
         'GlutamateAndGlutamine',
         'CholineCreatineRatio',
         'NAcetylaspartateCreatineRatio',
         'NAcetylaspartateCholineRatio',
         'Tmax',
         'R2',
         'VolumetricDiffusionDxzComponent',
         'MagneticSusceptibility',
         'VolumetricDiffusionDxyComponent',
         'VolumetricDiffusionDxxComponent',
         'RelativeRegionalBloodFlow',
         'RelativeRegionalBloodVolume',
         'Ktrans',
         'Kep',
         'Ve',
         'BloodOxygenationLevel',
         'T2WeightedDynamicContrastEnhancedMRSignalIntensity',
         'T2WeightedDynamicContrastEnhancedMRSignalIntensity',
         'AbsoluteRegionalBloodFlow',
         'T1WeightedDynamicContrastEnhancedMRSignalIntensity',
         'AbsoluteRegionalBloodVolume',
         'VolumetricDiffusionDzzComponent',
         'OxygenExtractionFraction',
         'R1',
         'VolumetricDiffusionDyzComponent',
         'R2',
         'VolumetricDiffusionDyyComponent',
         'VelocityEncoded',
         'ZScore',
         'ContrastAgentAngioMRSignalIntensity',
         'SpinTaggingPerfusionMRSignalIntensity',
         'RelativeAnisotropy',
         'FractionalAnisotropy',
         'FieldMapMRSignalIntensity',
         'T2WeightedMRSignalIntensity',
         'T2WeightedMRSignalIntensity',
         'T1WeightedMRSignalIntensity',
         'ProtonDensityWeightedMRSignalIntensity',
         'TimeOfFlightAngioMRSignalIntensity',
         'T1',
         'T2',
         'T2',
         'TimeCourseOfSignal',
         'TemperatureEncoded',
         'StudentTTest',
         'TimeToPeak',
         'SignalToNoise',
         'TemporalDerivativeThreshold',
         'MaximumSlope',
         'MaximumDifference',
         'TracerConcentration',
         'TimeOfPeakConcentration',
         'BolusArrivalTime',
         'TimeOfLeadingHalfPeakConcentration',
         'MeanTransitTime',
         'KurtosisDiffusionCoefficient',
         'GammaDistributionScaleParameter',
         'NegativeEnhancementIntegral',
         'GammaDistributionShapeParameter',
         'GammaDistributionMode',
         'DistributedDiffusionCoefficient',
         'RCoefficient',
         'AnomalousExponentParameter',
         'ProtonDensity',
         'SignalChange',
         'Suvbsa',
         'Suvibw',
         'SuvlbmJanma',
         'SuvlbmJames128',
         'MonoExponentialApparentDiffusionCoefficient',
         'SlowDiffusionCoefficient',
         'FastDiffusionCoefficient',
         'FastDiffusionCoefficientFraction',
         'StandardizedUptakeValue',
         'Suvbw',
         'Suvlbm'],
 'SCT': ['Fat',
         'Lipid',
         'NAcetylaspartate',
         'PowerDoppler',
         'Creatine',
         'Glutamine',
         'Tuarine',
         'Citrate',
         'Lactate',
         'Inositol',
         'Water',
         'Choline']}

name_for_cid[62] = 'ImagingAgentAdministrationPhaseType'
cid_concepts[62] = \
{'DCM': ['AutomaticAdministrationPhase',
         'AutomaticProgrammedHoldPhase',
         'AutomaticWithManualInjectPhase',
         'AutomaticWithManualHoldPhase']}

name_for_cid[7704] = 'LimbicSystemTracts'
cid_concepts[7704] = \
{'NEU': ['StriaTerminalis'], 'SCT': ['Fornix', 'Cingulum']}

name_for_cid[10041] = 'SourceofRadioisotopeActivityInformation'
cid_concepts[10041] = \
{'DCM': ['DoseCalibrator', 'InfusionSystem', 'RadioisotopeGenerator']}

name_for_cid[6115] = 'OsseousAnatomyModifiers'
cid_concepts[6115] = \
{'DCM': ['RibScaleneTubercle',
         'VertebralIntervertebralNotch',
         'SubscapularFossa',
         'ScapularSpine',
         'ScapularSupraspinatusFossa',
         'ScapularInfraspinatusFossa'],
 'SCT': ['VertebralForamen',
         'InferiorArticularFacetOfAxis',
         'DorsalAspectOfScapula',
         'GlenoidCavityOfScapula',
         'SpinousProcessOfVertebra',
         'TubercleOfRib',
         'SuperiorArticularProcessOfVertebra',
         'BodyOfSternum',
         'AngleOfRib',
         'InferiorArticularProcessOfVertebra',
         'ManubriumOfSternum',
         'SuperiorArticularFacetOfAxis',
         'PedicleOfVertebra',
         'AcromionProcessOfScapula',
         'HeadOfRib',
         'TransverseProcessOrVertebra',
         'LaminaOfVertebra',
         'ArchOfVertebra',
         'NeckOfRib',
         'ShaftOfRib',
         'SuprasternalNotch',
         'CoracoidProcessOfScapula',
         'XiphoidProcessOfSternum',
         'SternalAngle',
         'PectoralGirdle',
         'CostalGroove',
         'VertebralCanal',
         'ClavicularNotchOfSternum']}

name_for_cid[3747] = 'PercutaneousClosure'
cid_concepts[3747] = \
{}

name_for_cid[6050] = 'BreastProcedureReported'
cid_concepts[6050] = \
{'DCM': ['MarkerPlacement',
         'FilmScreenMammography',
         'DigitalMammography',
         'SurgicalConsult',
         'MammographyCAD'],
 'LN': ['BreastRightMammogram',
        'BreastFfdMammogram',
        'BreastUnilateralMrWoAndWContrastIv',
        'BreastBilateralMammogram',
        'WholeBodyPtWRncIv',
        'BreastUnilateralMammogram',
        'BreastMrWContrastIv',
        'BreastUnilateralMrWContrastIv',
        'BreastBilateralMrWContrastIv',
        'BreastRightMrWoAndWContrastIv',
        'BreastUnilateralMrWoContrast',
        'BreastLeftMr',
        'BreastLeftMrWoAndWContrastIv',
        'BreastMrWoAndWContrastIv',
        'BreastRightMr',
        'BreastBilateralMrWoAndWContrastIv',
        'WholeBodyCt',
        'BreastBilateralMr',
        'BreastLeftMammogram',
        'BreastLeftMrWContrastIv',
        'BreastRightMrWContrastIv',
        'BreastUnilateralMr'],
 'SCT': ['ExaminationOfBreast',
         'FineNeedleAspirationOfBreast',
         'MRIOfBreast',
         'SpecimenRadiographyOfBreast',
         'MammaryDuctogram',
         'UltrasonographyOfBreast',
         'PreBiopsyLocalizationOfBreastLesion',
         'BreastSurgicalBiopsy',
         'InsertionOfCatheter',
         'CoreNeedleBiopsyOfBreast',
         'SentinelLymphNodeBiopsy',
         'CTOfBreast',
         'RadioisotopeScanOfLymphaticSystem',
         'DiagnosticAspirationOfBreastCyst',
         'RadionuclideLocalizationOfTumorLimitedArea']}

name_for_cid[3602] = 'HemodynamicPatientState'
cid_concepts[3602] = \
{'SCT': ['BaselineState',
         'ExerciseState',
         'SupineBodyPosition',
         'RestingState',
         'PostExerciseState']}

name_for_cid[3481] = 'IVUSDistanceMeasurements'
cid_concepts[3481] = \
{'DCM': ['PlaquePlusMediaThickness', 'LumenPerimeter', 'EEMDiameter'],
 'SCT': ['StentDiameter', 'VesselLumenDiameter']}

name_for_cid[12210] = 'EchocardiographyPulmonaryArtery'
cid_concepts[12210] = \
{'LN': ['HeartRate',
        'MainPulmonaryArteryDiameter',
        'LeftPulmonaryArteryDiameter',
        'RightPulmonaryArteryDiameter'],
 'SCT': ['MainPulmonaryArteryPeakVelocity']}

name_for_cid[631] = 'PhaseofProcedureRequiringAnesthesia'
cid_concepts[631] = \
{'SCT': ['AfterProcedure',
         'BeforeProcedure',
         'Intraoperative',
         'Preoperative',
         'Postoperative',
         'DuringProcedure']}

name_for_cid[1003] = 'AnatomicalReferenceBasisSpine'
cid_concepts[1003] = \
{'SCT': ['LevelOfT9T10IntervertebralDisc',
         'LevelOfT8T9IntervertebralDisc',
         'LevelOfT5T6IntervertebralDisc',
         'LevelOfL1L2IntervertebralDisc',
         'LevelOfL2L3IntervertebralDisc',
         'LevelOfL3L4IntervertebralDisc',
         'LevelOfT7T8IntervertebralDisc',
         'LevelOfC7T1IntervertebralDisc',
         'LevelOfT10T11IntervertebralDisc',
         'LevelOfT12L1IntervertebralDisc',
         'LevelOfT6T7IntervertebralDisc',
         'LevelOfC2C3IntervertebralDisc',
         'LevelOfC4C5IntervertebralDisc',
         'C1Vertebra',
         'LevelOfC5C6IntervertebralDisc',
         'LevelOfT11T12IntervertebralDisc',
         'LevelOfC6C7IntervertebralDisc',
         'LevelOfC3C4IntervertebralDisc',
         'LevelOfT3T4IntervertebralDisc',
         'LevelOfT4T5IntervertebralDisc',
         'LevelOfT1T2IntervertebralDisc',
         'LevelOfL4L5IntervertebralDisc',
         'LevelOfT2T3IntervertebralDisc',
         'LevelOfL5S1IntervertebralDisc']}

name_for_cid[4201] = 'PatientEyeMovementCommand'
cid_concepts[4201] = \
{'SCT': ['PrimaryGaze',
         'RightGaze',
         'ConvergentGaze',
         'LeftGaze',
         'LeftDowngaze',
         'LeftUpgaze',
         'RightDowngaze',
         'UpwardGaze',
         'RightUpgaze',
         'Downgaze']}

name_for_cid[6103] = 'AbnormalLinesFindingorFeature'
cid_concepts[6103] = \
{'DCM': ['ReticulonodularPattern',
         'KerleyBLine',
         'KerleyCLines',
         'ParenchymalBand',
         'ReticularPattern',
         'SeptalLines',
         'AirFluidLevel',
         'SubpleuralLine',
         'CoronaRadiata',
         'TramlineShadow',
         'HoneycombPattern',
         'TubularShadow',
         'FleischnerLines',
         'IntralobularLines',
         'KerleyALine'],
 'SCT': ['PlateLikeAtelectasis']}

name_for_cid[74] = 'ImagingAgentAdministrationConsumableCatheterType'
cid_concepts[74] = \
{'SCT': ['PeripheralIntravenousCatheter',
         'CentralVenousCatheter',
         'ImplantableVenousAccessPort',
         'RectalCatheter']}

name_for_cid[3614] = 'ValveAreas,Nonmitral'
cid_concepts[3614] = \
{'DCM': ['DerivedNonValveArea'],
 'SCT': ['TricuspidValveArea', 'PulmonicValveArea', 'AorticValveArea']}

name_for_cid[6046] = 'UnitsofFollowupInterval'
cid_concepts[6046] = \
{'UCUM': ['Year', 'Day', 'Week', 'Month']}

name_for_cid[12206] = 'EchocardiographyRightAtrium'
cid_concepts[12206] = \
{'LN': ['RightAtriumSystolicPressure', 'RightAtriumAreaA4CView', 'HeartRate']}

name_for_cid[3497] = 'IVUSArterialMorphology'
cid_concepts[3497] = \
{'DCM': ['NecroticLipidicPlaque', 'FibroLipidicPlaque'],
 'SCT': ['LumenOfArtery',
         'VascularCalcification',
         'Thrombus',
         'ExternalElasticMembrane',
         'FibrousPlaque']}

name_for_cid[1015] = 'PatientAlignmentMethods'
cid_concepts[1015] = \
{'DCM': ['LaserCrossHairs']}

name_for_cid[627] = 'NeuromuscularBlockingAgentsforSmallAnimalAnesthesia'
cid_concepts[627] = \
{'SCT': ['Succinylcholine', 'Pancuronium']}

name_for_cid[23] = 'CranioCaudadAngulation'
cid_concepts[23] = \
{'SCT': ['Caudal', 'Cephalic']}

name_for_cid[10000] = 'ScopeofAccumulation'
cid_concepts[10000] = \
{'DCM': ['ProcedureStepToThisPoint',
         'Study',
         'Series',
         'PerformedProcedureStep',
         'IrradiationEvent']}

name_for_cid[6154] = 'EchoPattern'
cid_concepts[6154] = \
{'DCM': ['Anechoic', 'Hyperechoic', 'Complex', 'Hypoechoic', 'Isoechoic']}

name_for_cid[3706] = 'OverallContractility'
cid_concepts[3706] = \
{'SCT': ['Hypokinesis', 'Hyperkinesis', 'Akinesis', 'NormalWallContractility']}

name_for_cid[6011] = 'CalcificationTypesfromBIRADS®'
cid_concepts[6011] = \
{'DCM': ['FinePleomorphicCalcification', 'Macrocalcifications'],
 'SCT': ['CalcifiedSkinOfBreast',
         'FineLinearCastingCalcification',
         'CalcifiedSutureMaterial',
         'AmorphousCalcification',
         'VascularCalcification',
         'RoundShapedCalcification',
         'LucentCenteredCalcification',
         'PunctateCalcification',
         'MilkOfCalciumCalcification',
         'DystrophicCalcification',
         'EggshellCalcification',
         'LargeRodLikeCalcification',
         'FineLinearBranchingCastingCalcification',
         'MicrocalcificationsOfTheBreast',
         'CoarsePopcornLikeCalcification',
         'HeterogeneousCalcification']}

name_for_cid[3213] = 'StressHemodynamicFindings'
cid_concepts[3213] = \
{'SCT': ['ExertionalHypotension',
         'ChronotropicIncompetence',
         'ExertionalHypertension']}

name_for_cid[7600] = 'LymphNodeAnatomicSites'
cid_concepts[7600] = \
{'SCT': ['DeepLateralCervicalLymphNode',
         'AnteriorAuricularLymphNode',
         'IntermediateExternalIliacLymphNode',
         'LymphNodeOfHead',
         'InferiorInguinalLymphNode',
         'AnteriorCervicalLymphNode',
         'IntrapulmonaryLymphNode',
         'RetropharyngealLymphNode',
         'PectoralAxillaryLymphNode',
         'SuperiorGlutealLymphNode',
         'ParasternalLymphNode',
         'LymphNode',
         'CommonDuctLymphNode',
         'MedialLacunarLymphNode',
         'SegmentalLymphNodeOfTheLung',
         'BuccinatorLymphNode',
         'LateralPericardialLymphNode',
         'HighestMediastinalLymphNode',
         'SubscapularAxillaryLymphNode',
         'PrefemoralLymphNode',
         'ExtrapulmonaryLymphNodeOfLung',
         'ParotidLymphNode',
         'LowerParatrachealIncludingAzygousLymphNode',
         'DelphianLymphNode',
         'PrececalLymphNode',
         'SubiliacLymphNode',
         'SuperiorRectalLymphNode',
         'SuperficialParotidLymphNode',
         'HypogastricLymphNode',
         'IliacLymphNode',
         'LymphNodeOfMesentery',
         'EpitrochlearLymphNode',
         'AxillaryLymphNode',
         'JuxtaintestinalLymphNode',
         'ParatrachealLymphNode',
         'DeepAnteriorCervicalLymphNode',
         'LumbarLymphNode',
         'UpperParatrachealLymphNode',
         'RegionalLymphNode',
         'PyloricLymphNode',
         'LacunarLymphNode',
         'GutAssociatedLymphNode',
         'PosteriorTibialLymphNode',
         'PancreaticosplenicLymphNode',
         'DeepParotidLymphNode',
         'PelvicLymphNode',
         'LymphNodeOfAortopulmonaryWindow',
         'CeliacLymphNode',
         'LateralJugularLymphNode',
         'PosteriorMediastinalLymphNode',
         'LymphNodeOfLesserCurvatureOfStomach',
         'MiddleDeepCervicalLymphNode',
         'InfraclavicularLymphNode',
         'SupraclavicularLymphNode',
         'SubcarinalLymphNode',
         'ParametrialLymphNode',
         'ParavesicularLymphNode',
         'SuperficialPoplitealLymphNode',
         'PrevesicularLymphNode',
         'GlutealLymphNode',
         'SacralLymphNode',
         'RetrotrachealLymphNode',
         'RetroperitonealLymphNode',
         'DeepCervicalLymphNode',
         'DeepPoplitealLymphNode',
         'SuperiorMediastinalLymphNode',
         'VesicularLymphNode',
         'IntermediateCommonIliacLymphNode',
         'SplenicLymphNode',
         'OccipitalLymphNode',
         'InferiorAuricularLymphNode',
         'NasolabialLymphNode',
         'RenalHilarLymphNode',
         'HilarLymphNode',
         'SuperficialInguinalLymphNode',
         'UterineParacervicalLymphNode',
         'InterlobarLymphNodeOfTheLung',
         'LymphNodeOfStomach',
         'LateralExternalIliacLymphNode',
         'EpigastricLymphNode',
         'AbdominalLymphNode',
         'LymphNodeOfGreaterCurvatureOfStomach',
         'PoplitealLymphNode',
         'IntramammaryLymphNode',
         'SubaorticCommonIliacLymphNode',
         'HepaticLymphNode',
         'SuperficialIntraparotidLymphNode',
         'PresymphysialLymphNode',
         'IntercostalLymphNode',
         'LymphNodeOfAorticArch',
         'PararectalLymphNode',
         'PeripancreaticLymphNode',
         'TracheobronchialLymphNodeLocatedNearCarina',
         'AorticLymphNode',
         'SuperficialLymphNode',
         'TibialLymphNode',
         'CentralAxillaryLymphNode',
         'SuperiorMesentericLymphNode',
         'IntraglandularParotidLymphNode',
         'SigmoidLymphNode',
         'PostvesicularLymphNode',
         'MidcolicLymphNode',
         'PosteriorTriangleCervicalLymphNode',
         'InferiorGlutealLymphNode',
         'EsophagealLymphNode',
         'TracheobronchialLymphNode',
         'CardiophrenicAngleLymphNode',
         'LobarLymphNodeOfTheLung',
         'SubclavianLymphNode',
         'SuperiorMedialInguinalLymphNode',
         'PancreaticLymphNode',
         'CervicalLymphNode',
         'IleocolicLymphNode',
         'LateralCommonIliacLymphNode',
         'SuperiorTracheobronchialLymphNode',
         'UpperDeepCervicalLymphNode',
         'RetropyloricLymphNode',
         'SuperficialCervicalLymphNode',
         'MiddleJugularLymphNode',
         'LymphNodeOfEpiploicForamen',
         'AppendicularLymphNode',
         'LateralVesicularLymphNode',
         'PretrachealLymphNode',
         'LymphNodeOfLimb',
         'RetrocecalLymphNode',
         'AnteriorMediastinalLymphNode',
         'PromontoryCommonIliacLymphNode',
         'CysticLymphNode',
         'SubmentalLymphNode',
         'MedialExternalIliacLymphNode',
         'LowerJugularLymphNode',
         'PancreaticoduodenalLymphNode',
         'MandibularLymphNode',
         'PrepericardialLymphNode',
         'MediastinalLymphNode',
         'LowerLimbLymphNode',
         'LungAndTracheobronchialLymphNodes',
         'PrevascularRetrotrachealLymphNode',
         'DiaphragmaticLymphNode',
         'InferiorMesentericLymphNode',
         'FemaleGenitalLymphNode',
         'LymphNodeOfThePulmonaryLigament',
         'PostauricularLymphNode',
         'LymphNodeRingOfCardiaOfStomach',
         'ParavaginalLymphNode',
         'SupramammaryLymphNode',
         'AxillaryVeinLymphNode',
         'DeepInguinalLymphNode',
         'DeepIntraparotidLymphNode',
         'UpperLimbLymphNode',
         'IntestinalLymphNode',
         'SubpyloricLymphNode',
         'InferiorPancreaticoduodenalLymphNode',
         'LymphNodeOfTrunk',
         'ParaesophagealLymphNodeBelowCarina',
         'ParaAorticLymphNodeOfTheAnteriorMediastinum',
         'PerigastricLymphNode',
         'UpperJugularLymphNode',
         'MedialCommonIliacLymphNode',
         'AnteriorTibialLymphNode',
         'InguinalLymphNode',
         'LateralCervicalLymphNode',
         'InterpectoralLymphNode',
         'LowerDeepCervicalLymphNode',
         'LymphNodeOfThorax',
         'ExternalMammaryLymphNode',
         'DeepLymphNode',
         'HemolymphNode',
         'MedianRetropharyngealLymphNode',
         'ColicLymphNode',
         'FacialLymphNode',
         'InteriliacLymphNode',
         'PrelaryngealLymphNode',
         'PosteriorAuricularLymphNode',
         'SubsegmentalLymphNodeOfTheLung',
         'SuperiorPancreaticoduodenalLymphNode',
         'ApicalAxillaryLymphNode',
         'FibularLymphNode',
         'SuprapyloricLymphNode',
         'InnominateLymphNode',
         'SuperficialLateralCervicalLymphNode',
         'PrevertebralLymphNode',
         'MesentericArteryLymphNode',
         'SuperficialAnteriorCervicalLymphNode',
         'LateralRetropharyngealLymphNode',
         'SuperiorPancreaticLymphNode',
         'SubmandibularLymphNode',
         'ExternalIliacLymphNode',
         'SubmaxillaryLymphNode',
         'ParamammaryLymphNode',
         'ObturatorLymphNode',
         'SuperiorLateralInguinalLymphNode',
         'MastoidLymphNode',
         'JugularLymphNode',
         'LymphNodeOfHeadAndNeck',
         'GastroOmentalLymphNode',
         'LateralAxillaryLymphNode',
         'CubitalLymphNode',
         'ScaleneLymphNode',
         'InferiorPancreaticLymphNode']}

name_for_cid[7250] = 'MultiFrameSubsetType'
cid_concepts[7250] = \
{'DCM': ['ReferencedFrames', 'ReferencedSegment']}

name_for_cid[3090] = 'TimeSynchronizationChannelTypes'
cid_concepts[3090] = \
{'DCM': ['XRayOnTrigger',
         'XRayFluoroscopyOnSignal',
         'DigitalTimecodeNOS',
         'IRIGBTimecode',
         'ECGBasedGatingSignalProcessed']}

name_for_cid[9537] = 'EffectiveDoseCalculationMethodCategories'
cid_concepts[9537] = \
{'DCM': ['FractionationBasedOrTemporallyBasedMethods',
         'RadiationTransportBasedMethods']}

name_for_cid[7483] = 'CommonAnatomicRegionsforAnimals'
cid_concepts[7483] = \
{'SCT': ['FetlockOfHindlimb',
         'Sacrum',
         'NasalSinus',
         'Metatarsus',
         'NavicularOfHindfoot',
         'CalcanealTubercle',
         'Humerus',
         'Abdomen',
         'FetlockOfForelimb',
         'Esophagus',
         'Bladder',
         'AtlantalAxialJoint',
         'Mandible',
         'PasternOfHindfoot',
         'Wing',
         'LumboSacralSpine',
         'Metacarpus',
         'MaxillaryIncisorTeeth',
         'UrinaryTract',
         'MandibularDentalArch',
         'Carpus',
         'EntireBody',
         'ChestAndAbdomen',
         'RadiusAndUlna',
         'LumbarSpine',
         'Digit',
         'Hindfoot',
         'MaxillaryDentalArch',
         'TibiaAndFibula',
         'PasternOfForefoot',
         'Forefoot',
         'Pelvis',
         'Tarsus',
         'UpperGastroIntestinalTract',
         'Chest',
         'Shoulder',
         'Stiffle',
         'ElbowJoint',
         'Colon',
         'AllLegs',
         'OrbitalStructure',
         'ThoracicSpine',
         'HipJoint',
         'CoccygealVertrebrae',
         'CervicoThoracicSpine',
         'MandibularIncisorTeeth',
         'Femur',
         'FrontalSinus',
         'Skull',
         'ThoracoLumbarSpine',
         'Urethra',
         'Patella',
         'AtlantoOccipitalJoint',
         'NavicularOfForefoot',
         'CervicalSpine'],
 'UMLS': ['DistalPhalanx']}

name_for_cid[12251] = 'CardiacUltrasoundLinearValveMeasurements'
cid_concepts[12251] = \
{'LN': ['LeafletThickness',
        'LeafletSeparation',
        'ROIInternalDimensionByUS',
        'ROIThicknessByUS',
        'CEDistance',
        'DEExcursion'],
 'SCT': ['Radius',
         'MinorAxis',
         'Length',
         'Diameter',
         'MajorAxis',
         'Circumference']}

name_for_cid[3839] = 'CoronaryVeins'
cid_concepts[3839] = \
{'SCT': ['GreatCardiacVein',
         'AnteriorCardiacVein',
         'VentricularVein',
         'AtrioventricularVein',
         'SmallCardiacVein',
         'SmallestCardiacVein',
         'MiddleCardiacVein',
         'CoronarySinus',
         'AtrialVein',
         'AzygosVein']}

name_for_cid[220] = 'LevelofSignificance'
cid_concepts[220] = \
{'SCT': ['Significant',
         'SignificanceUndetermined',
         'MostSignificant',
         'HighlySignificant',
         'NotSignificant']}

name_for_cid[7196] = 'SpatialandRelationalConceptSegmentationPropertyTypes'
cid_concepts[7196] = \
{'DCM': ['Lucency',
         'RegistrationInput',
         'Background',
         'SingleSubjectExtractedFromGroup',
         'Interface',
         'Line'],
 'NCIt': ['ReferenceRegion'],
 'SCT': ['Normal', 'WasteMaterial', 'Tissue']}

name_for_cid[4240] = 'OphthalmicMeasurementorCalculationDataSource'
cid_concepts[4240] = \
{'DCM': ['MeasurementFromThisDevice',
         'ExternalDataSource',
         'AxialMeasurementsSOPInstance',
         'RefractiveMeasurementsSOPInstance',
         'AutorefractionMeasurementsSOPInstance',
         'ManualEntry',
         'KeratometryMeasurementsSOPInstance']}

name_for_cid[9536] = 'RadiotherapyPrescribingandSegmentingPersonRoles'
cid_concepts[9536] = \
{'DCM': ['LeadRadiologicTechnologist'],
 'SCT': ['RadiologicTechnologist',
         'Resident',
         'MedicalPractitioner',
         'Physician',
         'RadiationTherapist'],
 'UMLS': ['ConsultingPhysician', 'RadiationPhysicist', 'MedicalPhysicist']}

name_for_cid[7482] = 'DXAnatomyImagedforAnimals'
cid_concepts[7482] = \
{'SCT': ['FetlockOfHindlimb',
         'Sacrum',
         'NasalSinus',
         'Metatarsus',
         'NavicularOfHindfoot',
         'CalcanealTubercle',
         'Humerus',
         'Abdomen',
         'FetlockOfForelimb',
         'Esophagus',
         'Bladder',
         'AtlantalAxialJoint',
         'Mandible',
         'PasternOfHindfoot',
         'Wing',
         'LumboSacralSpine',
         'Metacarpus',
         'MaxillaryIncisorTeeth',
         'UrinaryTract',
         'MandibularDentalArch',
         'Carpus',
         'EntireBody',
         'ChestAndAbdomen',
         'RadiusAndUlna',
         'LumbarSpine',
         'Digit',
         'Hindfoot',
         'MaxillaryDentalArch',
         'TibiaAndFibula',
         'PasternOfForefoot',
         'Forefoot',
         'Pelvis',
         'Tarsus',
         'UpperGastroIntestinalTract',
         'Chest',
         'Shoulder',
         'Stiffle',
         'ElbowJoint',
         'Colon',
         'AllLegs',
         'OrbitalStructure',
         'ThoracicSpine',
         'HipJoint',
         'CoccygealVertrebrae',
         'CervicoThoracicSpine',
         'MandibularIncisorTeeth',
         'Femur',
         'FrontalSinus',
         'Skull',
         'ThoracoLumbarSpine',
         'Urethra',
         'Patella',
         'AtlantoOccipitalJoint',
         'NavicularOfForefoot',
         'CervicalSpine'],
 'UMLS': ['DistalPhalanx']}

name_for_cid[12250] = 'CardiacUltrasoundCommonLinearMeasurements'
cid_concepts[12250] = \
{'LN': ['ROIInternalDimensionByUS', 'ROIThicknessByUS'],
 'SCT': ['Radius',
         'MinorAxis',
         'Length',
         'Diameter',
         'MajorAxis',
         'Circumference']}

name_for_cid[3838] = 'DiameterDerivation'
cid_concepts[3838] = \
{'DCM': ['AnteriorPosterior'],
 'SCT': ['Minimum', 'Maximum', 'Mean', 'Transverse']}

name_for_cid[221] = 'MeasurementRangeConcepts'
cid_concepts[221] = \
{'DCM': ['PercentileRankingOfMeasurement',
         'ZScoreOfMeasurement',
         'StandardDeviationOfPopulation',
         '_2SigmaDeviationOfPopulation'],
 'SCT': ['_1SigmaLowerValueOfPopulation',
         '_2SigmaLowerValueOfPopulation',
         '_5thPercentileValueOfPopulation',
         '_10thPercentileValueOfPopulation',
         '_1SigmaUpperValueOfPopulation',
         '_2SigmaUpperValueOfPopulation',
         'MeanValueOfPopulation',
         '_95thPercentileValueOfPopulation',
         '_90thPercentileValueOfPopulation',
         'MedianValueOfPopulation']}

name_for_cid[7197] = 'BodySubstanceSegmentationPropertyTypes'
cid_concepts[7197] = \
{'SCT': ['Feces', 'Urine', 'BodyFluid', 'Bile', 'Gas']}

name_for_cid[4241] = 'OphthalmicAxialLengthSelectionMethod'
cid_concepts[4241] = \
{'DCM': ['MeanValueChosen', 'UserChosenValue']}

name_for_cid[10001] = 'UIDTypes'
cid_concepts[10001] = \
{'DCM': ['StudyInstanceUID',
         'PerformedProcedureStepSOPInstanceUID',
         'SeriesInstanceUID',
         'IrradiationEventUID']}

name_for_cid[6155] = 'PosteriorAcousticFeatures'
cid_concepts[6155] = \
{'DCM': ['CombinedPosteriorEnhancementAndShadowing',
         'NoPosteriorAcousticFeatures',
         'PosteriorEnhancement',
         'PosteriorShadowing']}

name_for_cid[3707] = 'VSDDescription'
cid_concepts[3707] = \
{'SCT': ['NonRestrictive', 'Membranous', 'Restrictive', 'None']}

name_for_cid[6010] = 'MammographyCalcificationTypes'
cid_concepts[6010] = \
{'DCM': ['FinePleomorphicCalcification', 'Macrocalcifications'],
 'SCT': ['CalcifiedSkinOfBreast',
         'FineLinearCastingCalcification',
         'CalcifiedSutureMaterial',
         'AmorphousCalcification',
         'VascularCalcification',
         'RoundShapedCalcification',
         'LucentCenteredCalcification',
         'PunctateCalcification',
         'MilkOfCalciumCalcification',
         'DystrophicCalcification',
         'EggshellCalcification',
         'LargeRodLikeCalcification',
         'FineLinearBranchingCastingCalcification',
         'MicrocalcificationsOfTheBreast',
         'CoarsePopcornLikeCalcification',
         'HeterogeneousCalcification']}

name_for_cid[3212] = 'TreadmillSpeed'
cid_concepts[3212] = \
{'UCUM': ['kmH', 'Mph']}

name_for_cid[3642] = 'DegreeFindings'
cid_concepts[3642] = \
{'SCT': ['SeverelyElevated', 'Decreased', 'Elevated', 'NormalRange']}

name_for_cid[7601] = 'HeadandNeckCancerAnatomicSites'
cid_concepts[7601] = \
{'SCT': ['PharyngealTonsil',
         'ParanasalSinus',
         'Oropharynx',
         'FloorOfMouth',
         'BuccalMucosa',
         'Supraglottis',
         'BaseOfTongue',
         'Lip',
         'Tongue',
         'TonsilAndAdenoid',
         'Hypopharynx',
         'OropharyngealTonsil',
         'SalivaryGland',
         'MaxillarySinus',
         'PyriformSinus',
         'Glottis',
         'LingualTonsil',
         'Larynx',
         'PalatineUvula',
         'OralCavity',
         'PalatineTonsil',
         'NasalCavity',
         'Nasopharynx',
         'TubalTonsil',
         'LowerAlveolarRidge',
         'RetromolarTrigone'],
 'UMLS': ['UnknownPrimaryNeoplasiaSite']}

name_for_cid[12207] = 'EchocardiographyMitralValve'
cid_concepts[12207] = \
{'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'MitralValveAWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'StenosisPeakVelocity',
        'MitralValveEFSlopeByMMode',
        'VolumeFlow',
        'FlowRadius',
        'HeartRate',
        'TimeAveragedPeakVelocity',
        'EndDiastolicVelocity',
        'VelocityOfFlowPropagation',
        'PeakGradient',
        'MitralValveEPSSEWave',
        'MitralValveEWavePeakVelocity',
        'MitralRegurgitationDpDtDerivedFromMitralRegVelocity',
        'AliasVelocity',
        'MitralValveEToARatio',
        'MitralValveDiastolicPeakInstantaneousGradient',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MitralValveAWaveDuration',
         'CardiacIndex',
         'MitralValveClosureToOpeningTime',
         'CardiacOutput',
         'MitralValveATDTRatio',
         'RegurgitantFraction',
         'MitralValveEWaveDecelerationTime',
         'StrokeVolume',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter',
         'StrokeIndex']}

name_for_cid[3496] = 'IVUSFiducialPoints'
cid_concepts[3496] = \
{'SCT': ['VascularCalcification',
         'Vein',
         'VesselOrigin',
         'Stent',
         'FibrousPlaque',
         'CollateralBranchOfVessel']}

name_for_cid[4216] = 'OphthalmicVisualAcuityType'
cid_concepts[4216] = \
{'DCM': ['AutorefractionVisualAcuity',
         'HabitualVisualAcuity',
         'PrescriptionVisualAcuity'],
 'SCT': ['BestCorrectedVisualAcuity',
         'PinholeVisualAcuity',
         'BrightnessAcuityTestingVisualAcuity',
         'PotentialAcuityMeterVisualAcuity',
         'UncorrectedVisualAcuity']}

name_for_cid[626] = 'PremedicationAgentsforSmallAnimalAnesthesia'
cid_concepts[626] = \
{'SCT': ['Chlorpromazine', 'Acepromazine']}

name_for_cid[3750] = 'PreviouslyDilatedLesion'
cid_concepts[3750] = \
{}

name_for_cid[6102] = 'ChestFindingorFeatureModifier'
cid_concepts[6102] = \
{'BARI': ['_1stMarginalCoronaryArteryLaterals',
          '_2ndMarginalCoronaryArteryLaterals',
          '_3rdMarginalCoronaryArteryLaterals',
          'PosteriorDescendingSeptalPerforators',
          '_1stDiagonalCoronaryArteryLaterals',
          '_2ndDiagonalCoronaryArteryLaterals',
          'RamusLaterals',
          '_3rdDiagonalCoronaryArteryLaterals'],
 'DCM': ['KerleyBLine',
         'KerleyCLines',
         'ParenchymalBand',
         'ReticularPattern',
         'SeptalLines',
         'SubpleuralLine',
         'TramlineShadow',
         'TubularShadow',
         'Density',
         'DependentOpacity',
         'AorticKnob',
         'ArchOfTheAzygosVein',
         'AirFluidLevel',
         'CoronaRadiata',
         'HoneycombPattern',
         'FleischnerLines',
         'IntralobularLines',
         'KerleyALine',
         'MosaicPattern',
         'GroundGlassOpacity',
         'Infiltrate',
         'Micronodule',
         'Opacity',
         'PhantomTumor',
         'Shadow',
         'AbnormalInterstitialPattern',
         'SmallIrregularOpacities',
         'SmallRoundedOpacities',
         'TreeInBudSign',
         'GranularPattern',
         'MiliaryPattern',
         'HaloSign',
         'AirBronchiologram',
         'AirBronchogram',
         'AirCrescent',
         'ReticulonodularPattern',
         'BeadedSeptumSign',
         'NodularPattern',
         'Pseudoplaque',
         'SignetRingSign',
         'Portacath',
         'ChestTube',
         'CentralLine',
         'KidneyStent',
         'SecondaryPulmonaryLobule',
         'PancreaticStent',
         'NippleRing',
         'Coin',
         'FiducialMark',
         'PrimaryComplex',
         'Hiatus',
         'AzygoesophagealRecessInterface',
         'ParaspinalLine',
         'PosteriorTrachealStripe',
         'RightTrachealStripe',
         'Stripe',
         'Lucency',
         'MidlungWindow',
         'CarinaAngle',
         'CentrilobularStructures',
         'AnteriorJunctionLine',
         'PosteriorJunctionLine',
         'Interface',
         'Line'],
 'SCT': ['SubclavianVein',
         'SternocleidomastoidMuscle',
         'PulmonaryValve',
         'ProximalRightCoronaryArtery',
         'BrachialPlexus',
         'ErectorSpinaeMuscle',
         'MidCircumflexCoronaryArtery',
         'Prosthesis',
         'IntermediateArteryRamus',
         'SubscapularisMuscle',
         'EndotrachealTube',
         'LymphNode',
         'TrapeziusMuscle',
         'IntercostalArtery',
         'Esophagus',
         'PlateLikeAtelectasis',
         'ScarTissue',
         'TeresMajorMuscle',
         'LeftMainBronchus',
         'MidRightCoronaryArtery',
         'PulmonaryVein',
         'SubclavianArtery',
         'EsophagealHiatus',
         '_3rdLeftPosterolateralCoronaryArtery',
         'HeartValveProsthesis',
         'PectoralisMajorMuscle',
         'Tracheotomy',
         'Sternum',
         'Staple',
         'LeftPosteriorDescendingCircumflexCoronaryArtery',
         '_1stMarginalCoronaryArtery',
         'Pin',
         'PulmonaryArtery',
         'MidLeftAnteriorDescendingCoronaryArtery',
         'Thyroid',
         'CoronaryArteryGraft',
         'MitralValve',
         'AcromioclavicularJoint',
         'SupraspinatusMuscle',
         'AxillaryFascia',
         'ExternalIntercostalMuscle',
         'InfraspinatusMuscle',
         'TricuspidValve',
         'Nodule',
         'FissureOfLung',
         'LevatoresCostarumMuscles',
         'SuperiorPhrenicArtery',
         'DistalLeftAnteriorDescendingCoronaryArtery',
         'AtrialSeptalDefect',
         'Catheter',
         'Ventricle',
         'RightCoronaryArteryOstium',
         'DescendingAorta',
         'MarginalCoronaryArtery',
         'InternalThoracicArtery',
         'ProximalLeftAnteriorDescendingCoronaryArtery',
         'Bullet',
         'Spine',
         'BronchialArtery',
         'BrachiocephalicVein',
         'ProximalCircumflexCoronaryArtery',
         'VenaCavaFilter',
         '_2ndMarginalCoronaryArtery',
         'Needle',
         'Carina',
         'AorticArch',
         'PulmonaryEmbolism',
         'CarotidBody',
         'Heart',
         'AirwayStructure',
         'DeltoidMuscle',
         'LeftMainCoronaryArtery',
         'InternalIntercostalMuscle',
         'UretericStent',
         '_1stDiagonalCoronaryArtery',
         'InteratrialSeptum',
         'LeftMainCoronaryArteryOstium',
         'Scapula',
         'IliocostalisMuscle',
         '_1stRightPosterolateralCoronaryArtery',
         'Suture',
         'RightPosteriorAVCoronaryArtery',
         'FascialLayer',
         'AorticValve',
         'InferiorVenaCava',
         'Jewelry',
         'ThymusGland',
         'SubcostalMuscle',
         'Humerus',
         'DistalRightCoronaryArtery',
         'Mass',
         'LeftAnteriorDescendingCoronaryArtery',
         'CommonCarotidArtery',
         'InnermostIntercostalMuscles',
         '_3rdMarginalCoronaryArtery',
         'PectoralisMinorMuscle',
         'EsophagealArtery',
         'LatissimusDorsiMuscle',
         'CervicalCollar',
         'Trachea',
         'LeftPosterolateralCircumflexCoronaryArtery',
         'Bronchus',
         'DorsalScapularArtery',
         'FeedingTube',
         'SuperiorVenaCava',
         'CostalCartilage',
         'BrachiocephalicTrunk',
         '_2ndRightPosterolateralCoronaryArtery',
         'PulmonaryTrunk',
         'AzygosVein',
         'BrachialArtery',
         '_2ndDiagonalCoronaryArtery',
         'DistalCircumflexCoronaryArtery',
         'Rib',
         'TeresMinorMuscle',
         'LongissimusMuscle',
         'SerratusAnteriorMuscle',
         'Clavicle',
         'AirTrapping',
         'AorticIsthmus',
         'Vertebra',
         '_1stLeftPosterolateralCoronaryArtery',
         'Pneumothorax',
         'CircumflexCoronaryArtery',
         'Diaphragm',
         'AxillaryArtery',
         'Atrium',
         'VertebralArtery',
         '_3rdDiagonalCoronaryArtery',
         'SpinalisMuscle',
         'InternalJugularVein',
         'ChordaeTendineaeCordis',
         'RightMainBronchus',
         'ScalenousAnteriorMuscle',
         'CardiacPacemakerLead',
         'AVGrooveContinuationOfCircumflexArtery',
         'TrabeculaeCarnae',
         '_3rdRightPosterolateralCoronaryArtery',
         'ThyrocervicalTrunk',
         'InferiorPhrenicArtery',
         'JejunostomyTube',
         'HilumOfLung',
         'ThoracicDuct',
         'CoinLesion',
         'TransversusThoracis',
         'InterventricularSeptum',
         '_2ndLeftPosterolateralCoronaryArtery',
         'PosteriorDescendingRightCoronaryArtery',
         '_1stSeptalCoronaryArtery',
         'RightCoronaryArtery',
         'AscendingAorta',
         'PericardiophrenicArtery',
         'CardiacPacemaker',
         'CostocervicalTrunk',
         'AxillaryVein',
         'LigamentumArteriosum',
         'SegmentOfLung',
         'Pneumomediastinum',
         'LobeOfLung']}

name_for_cid[75] = 'LowHighEqual'
cid_concepts[75] = \
{'SCT': ['Low', 'High', 'Equal']}

name_for_cid[3615] = 'ValveAreas'
cid_concepts[3615] = \
{'DCM': ['DerivedNonValveArea'],
 'SCT': ['MitralValveArea',
         'TricuspidValveArea',
         'PulmonicValveArea',
         'AorticValveArea']}

name_for_cid[6047] = 'CADProcessingandFindingsSummary'
cid_concepts[6047] = \
{'DCM': ['AllAlgorithmsSucceededWithoutFindings',
         'AllAlgorithmsSucceededWithFindings',
         'NotAllAlgorithmsSucceededWithoutFindings',
         'NotAllAlgorithmsSucceededWithFindings',
         'NoAlgorithmsSucceededWithoutFindings']}

name_for_cid[3480] = 'IVUSProcedurePhases'
cid_concepts[3480] = \
{'SCT': ['CardiacCatheterizationPostInterventionPhase',
         'CardiacCatheterizationPreInterventionPhase']}

name_for_cid[12211] = 'EchocardiographyAorticValve'
cid_concepts[12211] = \
{'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'StenosisPeakVelocity',
        'VolumeFlow',
        'FlowRadius',
        'AorticValveEjectionTime',
        'HeartRate',
        'TimeAveragedPeakVelocity',
        'EndDiastolicVelocity',
        'VelocityOfFlowPropagation',
        'PeakGradient',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'AorticValveCuspSeparation',
        'TimeAveragedMeanVelocity'],
 'SCT': ['RegurgitantFraction',
         'RatioOfAorticValveAccelerationTimeToEjectionTime',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter']}

name_for_cid[630] = 'LocalAnestheticsforSmallAnimalAnesthesia'
cid_concepts[630] = \
{'SCT': ['Lidocaine', 'Bupivacaine', 'LidocainePrilocaine']}

name_for_cid[1002] = 'AnatomicalReferenceBasisHead'
cid_concepts[1002] = \
{'FMA': ['Nasion'],
 'SCT': ['TemporalBone',
         'MaxillarySinus',
         'PituitaryFossa',
         'Mandible',
         'MastoidCellsAndAntra',
         'MastoidBone',
         'FrontalSinus',
         'Skull',
         'InternalAuditoryMeatus',
         'CarotidArtery',
         'ExternalAuditoryMeatus',
         'VertexOfHead',
         'ForamenMagnum',
         'LateralCanthus',
         'OrbitalStructure']}

name_for_cid[4200] = 'OphthalmicImagingAgent'
cid_concepts[4200] = \
{'SCT': ['IndocyanineGreen',
         'RoseBengal',
         'Fluorescein',
         'TrypanBlue',
         'MethyleneBlue']}

name_for_cid[63] = 'ImagingAgentAdministrationMode'
cid_concepts[63] = \
{'DCM': ['ManualAdministration', 'AutomatedAdministration']}

name_for_cid[10040] = 'RadiopharmaceuticalOrganDoseReferenceAuthority'
cid_concepts[10040] = \
{'DCM': ['InvestigationalNewDrug',
         'MIRDPamphlet1',
         'ICRPPublication53',
         'ICRPPublication80',
         'ICRPPublication106',
         'MIRDOSE',
         'OLINDAEXM',
         'PackageInsert',
         'InstitutionallyApprovedEstimates']}

name_for_cid[7705] = 'CommissuralFibers'
cid_concepts[7705] = \
{'SCT': ['ForcepsMinor',
         'GenuOfCorpusCallosum',
         'SpleniumOfCorpusCallosum',
         'AnteriorCommissure',
         'ForcepsMajor',
         'PosteriorCommissure',
         'CorpusCallosum',
         'HabenularCommissure',
         'TapetumOfCorpusCallosum']}

name_for_cid[6114] = 'OsseousAnatomyFindingorFeature'
cid_concepts[6114] = \
{'SCT': ['Clavicle',
         'Rib',
         'Humerus',
         'Scapula',
         'Vertebra',
         'Spine',
         'Sternum']}

name_for_cid[3746] = 'PercutaneousEntrySite'
cid_concepts[3746] = \
{'SCT': ['ViaFemoralVein',
         'ViaFemoralArtery',
         'ViaArtery',
         'ViaBrachialArtery',
         'ViaArmVein',
         'ViaRadialArtery',
         'ViaVein']}

name_for_cid[6051] = 'BreastProcedureReason'
cid_concepts[6051] = \
{'DCM': ['NeedleLocalizationAndBiopsy',
         'HistoryOfBenignBreastBiopsy',
         'MarkerPlacement',
         'PersonalHistoryOfBreastCancerWithBreastConservationTherapy',
         'PersonalHistoryOfBreastCancerWithMastectomy',
         'KnownBiopsyProvenMalignancy',
         'ClinicalFinding',
         'Targeted',
         'Survey',
         'RecallForTechnicalReasons',
         'Calibration',
         'RecallForImagingFindings',
         'RecallForPatientSymptomsClinicalFindings',
         'ReviewOfAnOutsideStudy',
         'AdditionalEvaluationRequestedFromAbnormalScreeningExam',
         'QualityControl',
         'AdditionalEvaluationRequestedFromPriorStudy',
         'FollowUpAtShortIntervalFromPriorStudy',
         'HistoryOfBreastAugmentationAsymptomatic'],
 'SCT': ['ReductionMammoplasty',
         'Diagnostic',
         'PersonalHistoryOfBreastCancer',
         'AugmentationMammoplasty',
         'Brachytherapy',
         'Screening',
         'RadiationTherapy']}

name_for_cid[6401] = 'NonlesionObjectTypePhysicalObjects'
cid_concepts[6401] = \
{'DCM': ['OtherMarker'],
 'SCT': ['Suture',
         'CardiacPacemaker',
         'CompressionPaddle',
         'JWire',
         'Bullet',
         'BBShotLeadPellet',
         'OpaqueMarker',
         'Clip',
         'Catheter',
         'IDPlate',
         'Collimator',
         'Implant',
         'Staple']}

name_for_cid[7210] = 'RelatedSeriesPurposesofReference'
cid_concepts[7210] = \
{'DCM': ['ForAttenuationCorrection',
         'SameAnatomy',
         'SourceSeriesForImageProcessingOperation',
         'SameIndication',
         'SimultaneouslyAcquired']}

name_for_cid[12246] = 'CardiacUltrasoundIndicationforStudy'
cid_concepts[12246] = \
{'SCT': ['Hemoptysis',
         'Arrhythmia',
         'Palpitations',
         'AnginaPectoris',
         'Syncope',
         'AbnormalECG',
         'CoronaryArteryDisease',
         'SupraventricularTachycardia',
         'ImagingGuidance',
         'Fever',
         'LeftBundleBranchBlock',
         'ValvularHeartDisease',
         'ChestPain',
         'Hypertension',
         'HeartDiseaseRiskFactors',
         'Murmur',
         'CongenitalHeartDisease',
         'Dyspnea']}

name_for_cid[9520] = 'SegmentedRTAccessoryDevices'
cid_concepts[9520] = \
{'DCM': ['BrachytherapyChannel',
         'RectalBalloon',
         'Mold',
         'AccessoryTray',
         'HeadAndNeckMask',
         'BrachytherapyChannelShield',
         'ApertureBlock',
         'HeadMask',
         'BrachytherapySourceApplicator',
         'Headframe',
         'VaginalCylinder',
         'BodyFrame',
         'BreastBoard',
         'Cast',
         'RadiotherapyApplicator',
         'WholeBodyPod',
         'VacuumMold'],
 'SCT': ['ShieldingBlock',
         'Chair',
         'SurfaceBolus',
         'Headrest',
         'BiteBlock',
         'Table']}

name_for_cid[4257] = 'VisualFieldGlobalIndexName'
cid_concepts[4257] = \
{'DCM': ['VisualFieldIndex',
         'VisualFieldLossDueToDiffuseDefect',
         'VisualFieldLossDueToLocalDefect',
         'GlaucomaHemifieldTestAnalysis',
         'OpticalFixationMeasurements']}

name_for_cid[12303] = 'EchoMeasurementTypes'
cid_concepts[12303] = \
{'DCM': ['DirectlyMeasured',
         'Calculated',
         'FractionalChange',
         'Indexed',
         'ManualEntry'],
 'SCT': ['Ratio']}

name_for_cid[7181] = 'AbstractMultidimensionalImageModelComponentUnits'
cid_concepts[7181] = \
{'UCUM': ['Centimeter2Milliliter',
          'Mm2S',
          'CubicCentimeter',
          'SquareCentimeter',
          'Um2ms',
          'StandardizedUptakeValueBodySurfaceArea',
          'StandardizedUptakeValueLeanBodyMassJanma',
          'Millitesla',
          'Sv',
          'SecondPerSquareMillimeter',
          'CubicMillimeter',
          'SquareMillimeter',
          'MilligramsMinuteMilliliter',
          'StandardizedUptakeValueIdealBodyWeight',
          '_106Mm2S',
          'BecquerelsMilliliter',
          'Micrometer',
          'MillimolesPerkgWetWeight',
          'Gy',
          'MicromoleMinuteMilliliter',
          'Decibel',
          'ArbitraryUnit',
          'WoodUM2',
          'DynScm5',
          'MilliliterMinuteGram',
          'Um2S',
          'StandardizedUptakeValueBodyWeight',
          'MilliliterPer100Milliliter',
          'MillimeterSecond',
          'MicromoleMilliliter',
          'ProportionalToCountsPerSecond',
          'Centimeter',
          'MilliliterPerMinute',
          'CountsPerSecond',
          'Ppm',
          'Second',
          'PRUM2',
          'StandardizedUptakeValueLeanBodyMassJames128Multiplier',
          'WoodU',
          'MilliliterGram',
          'ProportionalToCounts',
          'Kpa',
          'NumberParticlesPer100GramOfTissue',
          'MilliliterPer100GramPerMinute',
          'Min',
          'Centimeter',
          'CubicDecimeter',
          'Millimeter',
          'CentimeterSecond',
          'PRU',
          'Ratio',
          'CubicMicrometer',
          'SquareMicrometer',
          'Counts',
          'Millisecond',
          'DynScm5M2',
          'Percent',
          'S',
          'NoUnits',
          'HounsfieldUnit',
          'MilliliterPerSecond',
          'DegreesCelsius',
          'Mmhg',
          'StandardizedUptakeValueLeanBodyMassJames',
          'Hertz']}

name_for_cid[3711] = 'ValvularAbnormalities'
cid_concepts[3711] = \
{'DCM': ['HeartValveFlail'],
 'SCT': ['Regurgitation',
         'CalcifiedHeartValve',
         'Stenosis',
         'ValvularEndocarditis',
         'ImmobileHeartValve']}

name_for_cid[6143] = 'LesionResponse'
cid_concepts[6143] = \
{'DCM': ['TargetLesionCompleteResponse',
         'TargetLesionPartialResponse',
         'TargetLesionProgressiveDisease',
         'TargetLesionStableDisease',
         'NonTargetLesionCompleteResponse',
         'NonTargetLesionIncompleteResponseOrStableDisease',
         'NonTargetLesionProgressiveDisease']}

name_for_cid[7302] = 'ImplantTemplatesViewOrientations'
cid_concepts[7302] = \
{'SCT': ['AnteroPosterior', 'MedioLateral', 'LateralMedial']}

name_for_cid[10017] = 'XRayGrid'
cid_concepts[10017] = \
{'DCM': ['FixedGrid',
         'FocusedGrid',
         'ReciprocatingGrid',
         'ParallelGrid',
         'CrossedGrid',
         'NoGrid']}

name_for_cid[3204] = 'StressAgents'
cid_concepts[3204] = \
{'SCT': ['Dobutamine',
         'Adenosine',
         'Dipyridamole',
         'Atropine',
         'AdenosineA2ReceptorAgonist']}

name_for_cid[6006] = 'MammographyCharacteristicsofMargin'
cid_concepts[6006] = \
{'DCM': ['AngularMargins'],
 'SCT': ['SpiculatedLesion',
         'CircumscribedLesion',
         'ObscuredLesion',
         'IndistinctLesion',
         'MicrolobulatedLesion']}

name_for_cid[6211] = 'ColonCADMaterialDescription'
cid_concepts[6211] = \
{'DCM': ['SoftTissue'], 'SCT': ['BoneMatrix', 'Air', 'Fat']}

name_for_cid[7050] = 'DeidentificationMethod'
cid_concepts[7050] = \
{'DCM': ['RetainUidsOption',
         'BasicApplicationConfidentialityProfile',
         'RetainSafePrivateOption',
         'CleanPixelDataOption',
         'RetainInstitutionIdentityOption',
         'CleanRecognizableVisualFeaturesOption',
         'CleanGraphicsOption',
         'CleanStructuredContentOption',
         'CleanDescriptorsOption',
         'RetainLongitudinalTemporalInformationFullDatesOption',
         'RetainLongitudinalTemporalInformationModifiedDatesOption',
         'RetainPatientCharacteristicsOption',
         'RetainDeviceIdentityOption']}

name_for_cid[12114] = 'AbdominalVeins(Unilateral)'
cid_concepts[12114] = \
{'SCT': ['HepaticVein',
         'LeftMainBranchOfPortalVein',
         'InferiorRightHepaticVein',
         'PortalVein',
         'SuperiorMesentericVein',
         'TransjugularIntrahepaticPortosystemicShunt',
         'InferiorVenaCava',
         'MiddleHepaticVein',
         'InferiorMesentericVein',
         'RightHepaticVein',
         'RightMainBranchOfPortalVein',
         'UmbilicalVein',
         'SplenicVein',
         'LeftHepaticVein']}

name_for_cid[4040] = 'EndoscopyAnatomicRegions'
cid_concepts[4040] = \
{'SCT': ['Mediastinum',
         'Ureter',
         'UterusAndFallopianTubes',
         'ParanasalSinus',
         'EsophagusStomachAndDuodenum',
         'UpperUrinaryTract',
         'Bronchus',
         'SigmoidColon',
         'PharynxAndLarynx',
         'Chest',
         'Shoulder',
         'Abdomen',
         'AnusRectumAndSigmoidColon',
         'LumenOfBloodVessel',
         'TracheaAndBronchus',
         'Gallbladder',
         'Cervix',
         'Bladder',
         'Knee',
         'BileDuct',
         'Joint',
         'LargeIntestine',
         'Larynx',
         'Spine',
         'Rectum',
         'Pharynx',
         'ExternalAuditoryCanal',
         'InguinalRegion',
         'Kidney',
         'Nasopharynx',
         'BladderAndUrethra']}

name_for_cid[6092] = 'QuantitativeConceptsforUsage,Exposure'
cid_concepts[6092] = \
{'DCM': ['VolumeOfUse', 'DoseFrequency', 'RateOfExposure'], 'SCT': ['Dosage']}

name_for_cid[6138] = 'ChestNonlesionObjectType'
cid_concepts[6138] = \
{'DCM': ['Portacath',
         'ChestTube',
         'CentralLine',
         'KidneyStent',
         'PancreaticStent',
         'NippleRing',
         'Coin',
         'FiducialMark'],
 'SCT': ['JejunostomyTube',
         'Suture',
         'VenaCavaFilter',
         'HeartValveProsthesis',
         'Needle',
         'Jewelry',
         'Prosthesis',
         'FeedingTube',
         'Catheter',
         'EndotrachealTube',
         'Tracheotomy',
         'Staple',
         'ScarTissue',
         'UretericStent',
         'Bullet',
         'Pin',
         'CardiacPacemaker',
         'CardiacPacemakerLead',
         'CervicalCollar']}

name_for_cid[4105] = 'PerfusionAnalysisMethods'
cid_concepts[4105] = \
{'DCM': ['LeastMeanSquareLMSDeconvolution',
         'SingularValueDecompositionSVDDeconvolution']}

name_for_cid[12285] = 'CardiacUltrasoundAtrioventricularValvesFindingSites'
cid_concepts[12285] = \
{'SCT': ['TricuspidValve',
         'CommonNonMitralNonTricuspidAtrioventricularValveStructure',
         'MitralValve']}

name_for_cid[7457] = 'SexMaleFemaleorBoth'
cid_concepts[7457] = \
{'DCM': ['MixedSex', 'Female', 'Male']}

name_for_cid[3414] = 'ProcedureUrgency'
cid_concepts[3414] = \
{'SCT': ['EmergentProcedure',
         'UrgentProcedure',
         'ElectiveProcedure',
         'SalvageProcedure']}

name_for_cid[3551] = 'VentilationActions'
cid_concepts[3551] = \
{'DCM': ['EndVentilation', 'BeginVentilation']}

name_for_cid[6303] = 'ProstateSectorAnatomyfromEuropeanConcensus27Sector(Optimal)Model'
cid_concepts[6303] = \
{'SCT': ['CentralZoneOfLeftHalfProstate',
         'LeftPosteromedialMiddlePeripheralZoneOfProstate',
         'MaleExternalUrethralSphincter',
         'LeftPosteromedialApicalPeripheralZoneOfProstate',
         'LeftPosterolateralMiddlePeripheralZoneOfProstate',
         'RightApicalTransitionZoneOfProstate',
         'MiddleAnteriorFibromuscularStromaOfProstate',
         'RightBasalTransitionZoneOfProstate',
         'LeftMiddleTransitionZoneOfProstate',
         'RightPosterolateralBasalPeripheralZoneOfProstate',
         'RightAnteriorBasalPeripheralZoneOfProstate',
         'LeftPosterolateralApicalPeripheralZoneOfProstate',
         'RightSeminalVesicle',
         'LeftAnteriorApicalPeripheralZoneOfProstate',
         'CentralZoneOfRightHalfProstate',
         'RightPosteromedialMiddlePeripheralZoneOfProstate',
         'RightAnteriorMiddlePeripheralZoneOfProstate',
         'RightPosterolateralMiddlePeripheralZoneOfProstate',
         'LeftApicalTransitionZoneOfProstate',
         'RightPosteromedialApicalPeripheralZoneOfProstate',
         'ApicalAnteriorFibromuscularStromaOfProstate',
         'BasalAnteriorFibromuscularStromaOfProstate',
         'LeftBasalPartTransitionZoneOfProstate',
         'LeftSeminalVesicle',
         'RightAnteriorApicalPeripheralZoneOfProstate',
         'LeftAnteriorBasalPeripheralZoneOfProstate',
         'LeftPosterolateralBasalPeripheralZoneOfProstate',
         'RightPosterolateralApicalPeripheralZoneOfProstate',
         'RightMiddleTransitionZoneOfProstate',
         'LeftAnteriorMiddlePeripheralZoneOfProstate']}

name_for_cid[3101] = 'CardiacProceduralStateValues'
cid_concepts[3101] = \
{'DCM': ['ReinjectionState',
         'DelayedRedistributionState',
         'RedistributionState'],
 'SCT': ['CardiacStressState', 'RestingState']}

name_for_cid[3802] = 'PlaqueStructures'
cid_concepts[3802] = \
{'SCT': ['Mixed',
         'NonCalcified',
         'Hyperplasia',
         'Fibrous',
         'PathologicCalcification',
         'FattyDegeneration']}

name_for_cid[4017] = 'AnatomicRegionModifierforIntraoralRadiography'
cid_concepts[4017] = \
{'SCT': ['FirstMolarRegion',
         'FirstPremolarRegion',
         'ThirdMolarRegion',
         'LateralIncisorRegion',
         'SecondMolarRegion',
         'SecondPremolarRegion',
         'CanineRegion',
         'CentralIncisorRegion']}

name_for_cid[3228] = 'ECGTimingMeasurements'
cid_concepts[3228] = \
{'MDC': ['PDurationGlobal',
         'PPIntervalGlobal',
         'PPIntervalPerLead',
         'QTIntervalGlobal',
         'QRSDurationPerLead',
         'QTIntervalPerLead',
         'PDurationPerLead',
         'QRSDurationGlobal',
         'PRIntervalGlobal',
         'PRIntervalPerLead',
         'RRIntervalPerLead',
         'RRIntervalGlobal']}

name_for_cid[3678] = 'QTCorrectionAlgorithms'
cid_concepts[3678] = \
{'DCM': ['FraminghamQtcAlgorithm',
         'HodgesQTCorrectionAlgorithm',
         'FridericiaQtcAlgorithm',
         'BazettQTCorrectionAlgorithm']}

name_for_cid[18] = 'IsotopesinRadiopharmaceuticals'
cid_concepts[18] = \
{'DCM': ['_191mIridium'],
 'SCT': ['_22Sodium',
         '_113mIndium',
         '_42Potassium',
         '_97Ruthenium',
         '_201Thallium',
         '_188Rhenium',
         '_198Gold',
         '_199Gold',
         '_75Selenium',
         '_153Gadolinium',
         '_24Sodium',
         '_153Samarium',
         '_57Cobalt',
         '_59Iron',
         '_67Gallium',
         '_18Fluorine',
         '_64Copper',
         '_32Phosphorus',
         '_111Indium',
         '_186Rhenium',
         '_81mKrypton',
         '_178Tantalum',
         '_43Potassium',
         '_67Copper',
         '_125Iodine',
         '_89Strontium',
         '_90Yttrium',
         '_127Xenon',
         '_87mStrontium',
         '_203Lead',
         '_169Ytterbium',
         '_223Radium',
         '_14Carbon',
         '_133Xenon',
         '_177Lutetium',
         '_51Chromium',
         '_85Strontium',
         '_58Cobalt',
         '_99mTechnetium',
         '_133Barium',
         '_123Iodine',
         '_60Cobalt',
         '_85Krypton',
         '_131Iodine']}

name_for_cid[12006] = 'FetalLongBonesBiometryMeasurements'
cid_concepts[12006] = \
{'LN': ['ClavicleLength',
        'FemurLength',
        'FibulaLength',
        'HumerusLength',
        'RadiusLength',
        'TibiaLength',
        'UlnaLength']}

name_for_cid[3402] = 'PatientStatusandEvents'
cid_concepts[3402] = \
{'DCM': ['PatientPronouncedDead',
         'PatientTransferredToMorgue',
         'PatientGivenPostProcedureInstruction',
         'PatientDischargedFromDepartment',
         'PatientTransferredToCCU',
         'HemostasisNotAchievedOozing',
         'HemostasisNotAchievedActivelyBleeding',
         'PatientTransferredToHoldingArea',
         'PatientDisconnectedFromContinuousMonitoring',
         'PatientTransferredToSurgery',
         'HemostasisAchieved',
         'PatientHasRespiratoryDifficulty',
         'PatientCoughed',
         'PatientConnectedToContinuousMonitoring',
         'PatientAssistedToTable',
         'PatientUnresponsive',
         'PatientPreppedAndDraped',
         'PatientAdvanceDirectiveGiven',
         'PatientSedated',
         'NilPerOsNPOStatusConfirmed',
         'PatientAsleep',
         'PatientGivenPreProcedureInstruction',
         'PatientAlert',
         'PatientInformedConsentGiven',
         'PatientRestless',
         'PatientCalledToProcedureRoom',
         'PatientReportsChestPain',
         'PatientAdmittedToProcedureRoom',
         'PatientReportsNoPain',
         'PatientReportsNausea',
         'PatientReportsDiscomfort',
         'PatientDisoriented']}

name_for_cid[7011] = 'RejectedforQualityReasons'
cid_concepts[7011] = \
{'DCM': ['OtherFailure',
         'MotionBlur',
         'UnknownFailure',
         'UnderExposed',
         'DoubleExposure',
         'Positioning',
         'OverExposed',
         'NoImage',
         'DetectorArtifacts',
         'ArtifactsOtherThanGridOrDetectorArtifact',
         'MechanicalFailure',
         'ElectricalFailure',
         'ImageArtifacts',
         'SoftwareFailure',
         'GridArtifacts',
         'InappropriateImageProcessing']}

name_for_cid[12293] = 'CardiacUltrasoundAortopulmonaryConnectionsFindingSites'
cid_concepts[12293] = \
{'SCT': ['ArteriovenousFistula',
         'AortaToPulmonaryArteryConnection',
         'PatentDuctusArteriosus']}

name_for_cid[3814] = 'StentComposition'
cid_concepts[3814] = \
{'SCT': ['MetalStent', 'PlasticStent']}

name_for_cid[7154] = 'AbdominalSegmentationTypes'
cid_concepts[7154] = \
{'SCT': ['SkinOfAbdomen',
         'PeritonealCavity',
         'SmallIntestine',
         'BileDuct',
         'InferiorVenaCava',
         'UrinarySystem',
         'AbdominalAorta',
         'Peritioneum',
         'RetroperitonealSpace',
         'Liver',
         'AbdominalCavity',
         'Spleen',
         'Abdomen',
         'Kidney',
         'AdrenalGland',
         'Omentum',
         'AbdominalWallMuscle']}

name_for_cid[3117] = 'AttenuationCorrectionMethods'
cid_concepts[3117] = \
{'DCM': ['OSEMAlgorithm', 'ChangMethod']}

name_for_cid[3455] = 'IndexMethods'
cid_concepts[3455] = \
{'DCM': ['BSA1Point219'], 'LN': ['BSA']}

name_for_cid[3005] = 'RespirationWaveform'
cid_concepts[3005] = \
{'DCM': ['RespirationWaveform']}

name_for_cid[6207] = 'ColonQuantitativeTemporalDifferenceType'
cid_concepts[6207] = \
{'SCT': ['DifferenceInLocation', 'DifferenceInAttenuation', 'DifferenceInSize']}

name_for_cid[3510] = 'CatheterSizeUnits'
cid_concepts[3510] = \
{'UCUM': ['French', 'mm']}

name_for_cid[3843] = 'MyocardialSubsegment'
cid_concepts[3843] = \
{'SCT': ['Endocardial', 'Epicardial']}

name_for_cid[12102] = 'TemporalPeriodsRelatingtoProcedureorTherapy'
cid_concepts[12102] = \
{'SCT': ['AfterProcedure', 'BeforeProcedure', 'DuringProcedure']}

name_for_cid[8130] = 'StainingProtocols'
cid_concepts[8130] = \
{'SCT': ['HematoxylinAndEosinStainMethod', 'Staining']}

name_for_cid[6084] = 'MammoplastyProcedures'
cid_concepts[6084] = \
{'SCT': ['BreastImplantation',
         'ReductionMammoplasty',
         'RemovalOfBreastImplant',
         'BreastReconstruction']}

name_for_cid[6159] = 'MalignancyType'
cid_concepts[6159] = \
{'DCM': ['InvasiveBreastCarcinoma', 'OtherMalignancyType'],
 'SCT': ['IntraductalCarcinomaNonInfiltrating']}

name_for_cid[12030] = 'UltrasoundContrast/BolusAgents'
cid_concepts[12030] = \
{'DCM': ['Sonazoid',
         'Sonovue',
         'TargestarB',
         'TargestarP',
         'Cardiosphere',
         'Echovist',
         'Imagify',
         'Levovist'],
 'NDC': ['Definity', 'Optison']}

name_for_cid[9243] = 'RadiotherapyRegistrationWorkitemDefinition'
cid_concepts[9243] = \
{'DCM': ['RTPatientPositionRegistrationSinglePlane',
         'RTPatientPositionRegistrationDualPlane',
         'RTPatientPositionRegistrationSpatialFiducials',
         'RTPatientPositionRegistrationUltrasound',
         'RTPatientPositionRegistration2DOn3DReference',
         'RTPatientPositionRegistrationOptical',
         'RTPatientPositionRegistration3DCTVolumeBased',
         'RTPatientPositionRegistration3DOn2DReference',
         'RTPatientPositionRegistration3DCTGeneral',
         'RTPatientPositionRegistration3DCTMarkerBased']}

name_for_cid[4021] = 'PETRadiopharmaceutical'
cid_concepts[4021] = \
{'DCM': ['Cetuximab89Zr',
         'RO532344189Zr',
         'J59189Zr',
         'RO54290889Zr',
         'Cu3689Zr',
         'RO695894818F',
         'Bevacizumab89Zr',
         'PSMA1007F18',
         'Cg250FAb289Zr',
         'PSMA617Ga68',
         'R150789Zr',
         'E4G1089Zr',
         '_7E1189Zr',
         '_7D1289Zr',
         '_28H189Zr',
         'MonoclonalAntibodyMab64Cu',
         'MonoclonalAntibodyMab89Zr',
         'NanocolloidalAlbumin89Zr',
         'AntiB22089Zr',
         'Trastuzumab89Zr',
         'T807F18',
         'FlubatineF18',
         'CmabU3689Zr',
         'DN3089Zr',
         'Fresolimumab89Zr',
         'TRC10589Zr',
         'Margetuximab89Zr',
         'SAR341989Zr',
         'PittsburghCompoundBC11',
         'Ranibizumab89Zr',
         'Roledumab89Zr',
         'Xmab557489Zr',
         'Brentuximab89Zr',
         'Panitumumab89Zr',
         'Rituximab89Zr',
         'Mogamulizumab89Zr',
         'Ublituximab89Zr',
         'MEDI55189Zr',
         'GA20189Zr',
         'Ecromeximab89Zr',
         'Benralizumab89Zr',
         'Ocaratuzumab89Zr',
         'GlembatumumabVedotin89Zr',
         'PinatuzumabVedotin89Zr',
         'PolatuzumabVedotin89Zr',
         'Blinatumomab89Zr',
         'Pegdinetanib89Zr',
         'AGN15099889Zr',
         'RO693164311C',
         'Obinituzimab89Zr',
         'RO692496311C',
         'GlutamineF18',
         '_2FAF18',
         'NifeneF18',
         'CLR1404I124',
         'CLR1404I131',
         'FlurpiridazF18',
         'GlutamineC14',
         'GlutamineC11',
         'ATSMCu60',
         'ATSMCu61',
         'ATSMCu62',
         'CholineC11',
         'FallyprideC11',
         'FallyprideF18',
         'FLB457C11',
         'FluorotrioprideF18',
         'DfFK89Zr',
         'DfFKPEG389Zr',
         'DfFK289Zr',
         'DfCD4589Zr',
         'DfFK23PEG489Zr'],
 'NCIt': ['DcfpylF18', 'SarcosineC11', 'DCFBCF18', 'PSMA11Ga68'],
 'SCT': ['MespiperoneC11',
         'RacloprideC11',
         'FluciclatideF18',
         'EDTAGa68',
         'CarbonMonoxideO15',
         'CarbonMonoxideC11',
         'FluorobenzothiazoleF18',
         'FluoromethaneF18',
         'SodiumFluorideF18',
         'FlorbetabenF18',
         'GlutamateN13',
         'FlumazenilF18',
         'MethionineC11',
         'FluorodeoxyglucoseF18',
         'FlutemetamolF18',
         'OxygenO15',
         'SodiumIodideI124',
         'AcetateC11',
         'FluorocholineF18',
         'FluorethyltyrosinF18',
         'CarbonDioxideO15',
         'FlorbetapirF18',
         'GermaniumGe68',
         'ThymidineFLTF18',
         'FluoromisonidazoleF18',
         'MonoclonalAntibodyI124',
         'ButanolO15',
         'CarfentanilC11',
         'OxygenWaterO15',
         'ATSMCu64',
         'FluciclovineF18',
         'SpiperoneF18',
         'PalmitateC11',
         'AmmoniaN13',
         'FluoroLDopaF18',
         'FlumazenilC11',
         'PTSMCu62',
         'SodiumNa22',
         'FluorouracilF18',
         'RubidiumChlorideRb82'],
 'UMLS': ['MK6240F18',
          'ISO1F18',
          'FluoropropylDihydrotetrabenazineDTBZF18',
          'FluoroetanidazoleF18',
          'THK5317F18',
          'UCBJC11',
          'FluoroestradiolFESF18',
          'Tyrosine3OctreotateGa68',
          'EdotreotideGa68',
          'THK5351F18',
          'FlortaucipirF18']}

name_for_cid[3422] = 'DeviceUseActions'
cid_concepts[3422] = \
{'DCM': ['DeviceCrossedLesion'],
 'SCT': ['DeviceUsed',
         'DeviceAtSiteOfInterest',
         'DeviceAppliedToPatient',
         'DeviceWithdrawnRemoved',
         'DeviceInsertedIntoSheath',
         'DeviceCrossedSeptum']}

name_for_cid[7461] = 'UnitsofAreaMeasurement'
cid_concepts[7461] = \
{'UCUM': ['SquareMillimeter', 'SquareCentimeter', 'SquareMicrometer']}

name_for_cid[8110] = 'SpecimenSamplingProcedure'
cid_concepts[8110] = \
{'DCM': ['DissectionWithEntireSpecimenSubmission',
         'DissectionWithRepresentativeSectionsSubmission'],
 'SCT': ['CoreSampling',
         'SmearProcedure',
         'BlockSectioning',
         'BlockSurfaceRecut',
         'BlockStepSectioning',
         'Dissection',
         'LaserMicrodissection',
         'TouchPreparation']}

name_for_cid[3619] = 'HemodynamicResistanceMeasurements'
cid_concepts[3619] = \
{'DCM': ['TotalVascularResistance', 'TotalPulmonaryResistance'],
 'SCT': ['PulmonaryVascularResistance', 'SystemicVascularResistance']}

name_for_cid[12122] = 'OtherVascularProperties'
cid_concepts[12122] = \
{'LN': ['DecelerationTime',
        'PeakGradient',
        'AccelerationTime',
        'MeanGradient',
        'VolumeFlow'],
 'SCT': ['VesselLumenCrossSectionalArea',
         'VesselDepthFromSurface',
         'VesselOutsideDiameter',
         'VesselLumenDiameter',
         'VesselIntimalCrossSectionalArea',
         'LengthOfSegment',
         'VesselIntimalDiameter']}

name_for_cid[3530] = 'OxygenAdministrationActions'
cid_concepts[3530] = \
{'DCM': ['BeginOxygenAdministration', 'EndOxygenAdministration']}

name_for_cid[6118] = 'SizeDescriptor'
cid_concepts[6118] = \
{'DCM': ['ExtremelySmall', 'VerySmall', 'TooSmall'],
 'SCT': ['Medium', 'Enlarged', 'Small', 'Large']}

name_for_cid[3526] = 'BloodGasSaturation'
cid_concepts[3526] = \
{'DCM': ['BloodCarbonDioxideSaturation'],
 'LN': ['CapillaryBloodOxygenSaturation',
        'ArterialOxygenSaturation',
        'VenousOxygenSaturation',
        'CapillaryBloodOxygenSaturationByOximetry',
        'BloodOxygenSaturation']}

name_for_cid[3463] = 'VentricleIdentification'
cid_concepts[3463] = \
{'SCT': ['RightVentricle', 'LeftVentricle']}

name_for_cid[3208] = 'SummaryCodesExerciseECG'
cid_concepts[3208] = \
{'SCT': ['ExerciseECGAbnormal',
         'ExerciseECGEquivocal',
         'ExerciseECGNormal',
         'NotPerformed']}

name_for_cid[3121] = 'PerfusionMorphology'
cid_concepts[3121] = \
{'SCT': ['MyocardialIschemia',
         'MyocardialInfarction',
         'MixedIschemiaAndInfarction']}

name_for_cid[7162] = 'SurfaceProcessingAlgorithmFamilies'
cid_concepts[7162] = \
{'DCM': ['ManualProcessing',
         'MultispectralProcessing',
         'ClusterAnalysis',
         'MultiScaleResolutionFiltering',
         'HistogramAnalysis',
         'MorphologicalOperations',
         'EdgeDetection',
         'AdaptiveFiltering',
         'NeighborhoodAnalysis',
         'DeformableModels',
         'ArtificialIntelligence']}

name_for_cid[7027] = 'SegmentedRadiotherapeuticDoseMeasurementDevices'
cid_concepts[7027] = \
{'DCM': ['IonChamberArray',
         'DiodeArray',
         'LiquidIonChamber',
         'Diode',
         'IonChamber',
         'OSLD',
         'DiamondDetector'],
 'SCT': ['MOSFETRadiationTherapyDosimetrySystemDosimeter',
         'DigitalImager',
         'ThermoluminescentRadiationDosimeter']}

name_for_cid[7477] = 'GlobalShapeDescriptors'
cid_concepts[7477] = \
{}

name_for_cid[6071] = 'HistologicGradingMethod'
cid_concepts[6071] = \
{'DCM': ['BloomRichardsonGrade'], 'SCT': ['NottinghamCombinedGrade']}

name_for_cid[12118] = 'MeasurementOrientation'
cid_concepts[12118] = \
{'DCM': ['AnteriorPosterior'], 'SCT': ['Transverse', 'Longitudinal']}

name_for_cid[10060] = 'OrgansforRadiationDoseEstimates'
cid_concepts[10060] = \
{'DCM': ['Phantom'],
 'SCT': ['Lung',
         'SmallIntestine',
         'Breast',
         'Stomach',
         'Heart',
         'Skin',
         'Thymus',
         'Prostate',
         'LymphNode',
         'Colon',
         'Uterus',
         'SalivaryGlands',
         'Gallbladder',
         'Testis',
         'Esophagus',
         'Bladder',
         'Brain',
         'EyeLenses',
         'OralMucosa',
         'Ovary',
         'Muscle',
         'Pancreas',
         'BoneMarrow',
         'Liver',
         'EntireBody',
         'Thyroid',
         'Spleen',
         'Kidney',
         'AdrenalGland',
         'BoneSurface']}

name_for_cid[4109] = 'ModelIndependentDynamicContrastAnalysisParameters'
cid_concepts[4109] = \
{'DCM': ['TimeToPeak',
         'IAUC90BN',
         'IAUC180BN',
         'TemporalDerivativeThreshold',
         'IAUC',
         'MaximumSlope',
         'IAUC60',
         'MaximumDifference',
         'IAUC90',
         'TracerConcentration',
         'IAUC180',
         'IAUCBN',
         'IAUC60BN',
         'TimeOfPeakConcentration',
         'BolusArrivalTime',
         'TimeOfLeadingHalfPeakConcentration']}

name_for_cid[6134] = 'ChestQualitativeTemporalDifferenceType'
cid_concepts[6134] = \
{'SCT': ['DifferenceInBorderDefinition',
         'DifferenceInDistribution',
         'NoSignificantChangesInTheFinding',
         'IncreaseInSize',
         'DifferenceInTexture',
         'DifferenceInBorderShape',
         'DifferenceInSiteInvolvement',
         'DifferenceInSubstance',
         'LessDefined',
         'FindingPartiallyRemoved',
         'MoreDefined',
         'DecreaseInSize']}

name_for_cid[240] = 'PresentAbsent'
cid_concepts[240] = \
{'SCT': ['Absent', 'Present', 'Undetermined']}

name_for_cid[8203] = 'SurfaceScanRegistrationMethodTypes'
cid_concepts[8203] = \
{'DCM': ['IterativeClosestPoint', 'Freehand', 'FiducialAlignment']}

name_for_cid[610] = 'WaterTypes'
cid_concepts[610] = \
{'DCM': ['ReverseOsmosisPurifiedWater',
         'ReverseOsmosisPurifiedHclAcidifiedWater'],
 'SCT': ['TapWater', 'DistilledWater']}

name_for_cid[4220] = 'VisualFixationQualityDuringAcquisition'
cid_concepts[4220] = \
{'SCT': ['Steady', 'Indeterminate', 'NotSteady']}

name_for_cid[12231] = 'VolumeFlowMethods'
cid_concepts[12231] = \
{'DCM': ['DopplerVolumeFlow', 'ProximalIsovelocitySurfaceArea']}

name_for_cid[6026] = 'MammographyAssessment'
cid_concepts[6026] = \
{'DCM': ['PostProcedureMammogramsForMarkerPlacement']}

name_for_cid[6163] = 'MenstrualCyclePhase'
cid_concepts[6163] = \
{'DCM': ['_1stWeek', '_2ndWeek', '_3rdWeek'], 'SCT': ['MenstruationPresent']}

name_for_cid[3418] = 'LesionRisk'
cid_concepts[3418] = \
{}

name_for_cid[647] = 'PositionReferenceIndicatorforFrameofReference'
cid_concepts[647] = \
{'FMA': ['Bregma', 'Lambda']}

name_for_cid[12289] = 'CardiacUltrasoundSemilunarValves,AnnulusandSinusesFindingSites'
cid_concepts[12289] = \
{'SCT': ['PulmonicValve',
         'NeoaorticValve',
         'AorticValve',
         'TruncalValveStructure',
         'NeoaorticRoot',
         'RootOfAorta']}

name_for_cid[12266] = 'CardiacUltrasoundMitralValve'
cid_concepts[12266] = \
{'DCM': ['Tau',
         'IndicatorMeanTransitTime',
         'RRInterval',
         'BloodVelocityPeak',
         'BloodVelocityMean',
         'BloodVelocityMinimum',
         'SystolicBloodVelocityMean',
         'SystolicBloodVelocityPeak',
         'DiastolicBloodVelocityMean',
         'DiastolicBloodVelocityPeak'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'LeafletThickness',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'MitralRegurgitationDpDtDerivedFromMitralRegVelocity',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'LeafletSeparation',
        'AccelerationTime',
        'MitralValveESeptalSeparation',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'CEDistance',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'MitralValveEPSSEWave',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'DEExcursion',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[702] = 'RTContentAssessmentTypes'
cid_concepts[702] = \
{'DCM': ['RTPreTreatmentConsistencyCheck', 'RTPreTreatmentDoseCheck']}

name_for_cid[9500] = 'DosimetricObjectiveTypes'
cid_concepts[9500] = \
{'DCM': ['SpecifiedConformationNumber',
         'SpecifiedHealthyTissueConformityIndex',
         'SpecifiedConformityIndex',
         'MinimumPercentVolumeAtRadiationDose',
         'MinimumRadiationDose',
         'MaximumSurfaceRadiationDose',
         'MinimumSurfaceRadiationDose',
         'SpecifiedHomogeneityIndex',
         'MinimizeMeterset',
         'MinimumEquivalentUniformDose',
         'MaximumAbsoluteVolumeAtRadiationDose',
         'MaximumMeanRadiationDose',
         'MinimumAbsoluteVolumeAtRadiationDose',
         'MinimumMeanRadiationDose',
         'MaximumPercentVolumeAtRadiationDose',
         'MaximumRadiationDose',
         'PrescriptionRadiationDose',
         'MaximumEquivalentUniformDose']}

name_for_cid[6030] = 'MammographyPathologyCodes'
cid_concepts[6030] = \
{'DCM': ['FibroadenomatoidHyperplasia',
         'Fibroadenolipoma',
         'CartilaginousAndOsseousChange',
         'AxillaryNodeWithLymphoma',
         'AxillaryNodalMetastases',
         'BloodVesselVascularInvasion',
         'CarcinomaInChildren',
         'CarcinomaInEctopicBreast',
         'CarcinomaWithEndocrineDifferentiation',
         'BasalCellCarcinomaOfNipple',
         'NormalAxillaryNode',
         'AxillaryNodeWithCalcifications',
         'AxillaryNodeHyperplasia',
         'BenignCystWithBlood',
         'BenignCalcifications',
         'DuctalAdenoma',
         'DiabeticFibrousMastopathy',
         'SquamousCellCarcinomaOfTheNipple',
         'IntraductalCarcinomaHighGrade',
         'PseudoangiomatousStromalHyperplasia',
         'SiliconeGranuloma',
         'MetastaticDiseaseToAxillaryNode',
         'VirginalHyperplasia',
         'PeripheralDuctPapillomas',
         'MultifocalInvasiveDuctalCarcinoma',
         'OilCyst',
         'PostReductionMammoplasty',
         'MetastasisToAnIntramammaryLymphNode',
         'MalignantMelanomaOfNipple',
         'RecurrentMalignancy',
         'LargeDuctPapilloma',
         'MicroglandularAdenosis',
         'MultipleIntraductalPapillomas',
         'NoAbnormality',
         'NormalBreastTissue',
         'LympathicVesselInvasion',
         'OccultCarcinomaPresentingWithAxillaryLymphNodeMetastases',
         'MetastaticCancerToTheBreast',
         'MetastaticCancerToTheBreastFromTheColon',
         'MetastaticCancerToTheBreastFromTheLung',
         'MetastaticMelanomaToTheBreast',
         'MetastaticCancerToTheBreastFromTheOvary',
         'MetastaticSarcomaToTheBreast',
         'MultifocalIntraductalCarcinoma',
         'JuvenilePapillomatosis',
         'CarcinomaInPregnancyAndLactation',
         'LactationalChange',
         'IntraductalComedocarcinomaWithNecrosis',
         'IntraductalCarcinomaLowGrade',
         'IntracysticPapillaryCarcinoma',
         'InvasiveAndInSituCarcinoma',
         'LeukemicInfiltration'],
 'SCT': ['LactatingAdenoma',
         'Leiomyosarcoma',
         'Carcinosarcoma',
         'TubularAdenocarcinoma',
         'CystOfBreast',
         'FatNecrosisOfBreast',
         'Leiomyoma',
         'HodgkinDisease',
         'HemangiomaOfSubcutaneousTissue',
         'LymphNode',
         'BreastLobularHyperplasia',
         'CarcinomaInSituOfMaleBreast',
         'EpidermalInclusionCyst',
         'Galactocele',
         'ThrombophlebitisOfBreastMondorDisease',
         'ScarTissue',
         'LobularCarcinomaInSituOfBreast',
         'Plasmacytoma',
         'DuctalHyperplasiaUsual',
         'Angiosarcoma',
         'Gynecomastia',
         'GiantFibroadenoma',
         'MalignantFibrousHistiocytoma',
         'ApocrineMetaplasia',
         'Fibrosarcoma',
         'TubularAdenoma',
         'SignetRingCellCarcinoma',
         'Lymphoma',
         'Hematoma',
         'Amyloid',
         'PhyllodesTumor',
         'CarcinomaWithMetaplasia',
         'Chondrosarcoma',
         'PleomorphicAdenoma',
         'IntraductalCarcinomaMicroPapillary',
         'Neurofibroma',
         'Hemangiopericytoma',
         'NonHodgkinLymphoma',
         'PagetDiseaseMammary',
         'Adenomyoepithelioma',
         'FocalFibrosis',
         'InfiltratingDuctCarcinoma',
         'Lipoma',
         'SpindleCellNodule',
         'Abscess',
         'SclerosingAdenosis',
         'Liposarcoma',
         'Adenolipoma',
         'Comedocarcinoma',
         'Fibroadenoma',
         'MammaryDuctEctasia',
         'Chondroma',
         'Hemangioma',
         'SquamousCellCarcinoma',
         'InfarctionOfBreast',
         'ApocrineAdenocarcinoma',
         'Adenoma',
         'PapillaryCarcinoma',
         'RadialScar',
         'InflammatoryCarcinoma',
         'HemangiomaVenous',
         'LipidRichLipidSecretingCarcinoma',
         'GranularCellTumor',
         'Angiolipoma',
         'ForeignBody',
         'HyperplasiaUsual',
         'InvasiveCribriformCarcinoma',
         'AdenoidCysticCarcinoma',
         'Angiomatosis',
         'Adenosis',
         'Seroma',
         'MedullaryCarcinoma',
         'AtypicalIntraductalHyperplasia',
         'ExtraAbdominalDesmoid',
         'Fibromatosis',
         'Inflammation',
         'IntramammaryLymphNode',
         'AtypicalLobularHyperplasia',
         'IntraductalPapilloma',
         'NeoplasmOfTheMammarySkin',
         'PhyllodesTumorMalignant',
         'GlycogenRichCarcinoma',
         'SecretoryJuvenileCarcinomaOfTheBreast',
         'Neurofibromatosis',
         'MucinousAdenocarcinomaColloidCarcinoma',
         'IntracysticPapilloma',
         'Myofibroblastoma',
         'PapillaryCarcinomaInSitu',
         'JuvenileFibroadenoma',
         'EctopicAccessoryBreastTissue',
         'Edema',
         'Papilloma',
         'Hamartoma',
         'AsynchronousInvolutionOfBreast',
         'FibrocysticDiseaseOfBreast',
         'BenignNeoplasmOfNippleOfFemaleBreastNippleAdenoma',
         'InvasiveLobularCarcinoma',
         'OsteogenicSarcoma']}

name_for_cid[3232] = 'STSegmentLocation'
cid_concepts[3232] = \
{'SCT': ['LeftVentricleLateralSegment',
         'LeftVentricleInferiorSegment',
         'LeftVentricleApicalSegment',
         'LeftVentricleSeptalSegment',
         'LeftVentricleAnteriorSegment']}

name_for_cid[7271] = 'MRDiffusionAnisotropyIndices'
cid_concepts[7271] = \
{'DCM': ['VolumeRatio', 'RelativeAnisotropy', 'FractionalAnisotropy']}

name_for_cid[10021] = 'SourceofCTDoseInformation'
cid_concepts[10021] = \
{'DCM': ['CopiedFromImageAttributes',
         'ManualEntry',
         'DerivedFromHumanReadableReports',
         'AutomatedDataCollection',
         'ComputedFromImageAttributes']}

name_for_cid[3727] = 'IndicationsforCatheterization'
cid_concepts[3727] = \
{'SCT': ['Cardiomyopathy',
         'HeartDisease',
         'ValvularHeartDisease',
         'CardiogenicShock',
         'Arrhythmia',
         'IschemicHeartDisease',
         'HeartTransplant',
         'CardiacFunctionTestAbnormal',
         'HeartDiseaseCongenital']}

name_for_cid[4261] = 'OphthalmicMappingAcquisitionMethod'
cid_concepts[4261] = \
{'DCM': ['RetinalTopography',
         'TimeDomain',
         'SpectralDomain',
         'NoCornealCompensation',
         'CornealBirefringenceCompensation']}

name_for_cid[9516] = 'RadiotherapyBolusDeviceTypes'
cid_concepts[9516] = \
{'SCT': ['SurfaceBolus']}

name_for_cid[7158] = 'ArtifactSegmentationTypes'
cid_concepts[7158] = \
{'SCT': ['Table', 'ForeignBody', 'Clothing']}

name_for_cid[12270] = 'CardiacUltrasoundAorticValve'
cid_concepts[12270] = \
{'DCM': ['Tau', 'IndicatorMeanTransitTime', 'RRInterval'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'AorticValveCuspSeparation',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[6067] = 'FluidColor'
cid_concepts[6067] = \
{'DCM': ['LightBrownColor',
         'DarkRedColor',
         'DarkBrownColor',
         'BrightRedColor',
         'BloodTingedColor'],
 'SCT': ['BlackColor',
         'Clear',
         'BrownColor',
         'YellowColor',
         'GrayColor',
         'WhiteColor',
         'GreenColor']}

name_for_cid[6088] = 'OBGYNMaternalRiskFactors'
cid_concepts[6088] = \
{'DCM': ['PreviousFetalMalformationSyndrome',
         'PreviousRHNegativeOrBloodDyscrasiaAtBirth',
         'HistoryOfMultipleFetuses',
         'CurrentPregnancyKnownOrSuspectedMalformationsSyndromes',
         'FamilyHistoryFetalMalformationSyndrome',
         'UterineMalformations',
         'SpontaneousAbortion',
         'GynecologicCondition',
         'GynecologicSurgery',
         'PreviousLBWOrIUGRBirth'],
 'SCT': ['HistoryOfObesity',
         'HistoryOfHypertension',
         'HistoryOfSubstanceAbuse',
         'HistoryOfEctopicPregnancy',
         'HistoryOfDiabetesMellitus',
         'HistoryOfEclampsia',
         'HistoryOfInfertility',
         'HistoryOfPrematureDelivery',
         'HistoryOfRegularMedication',
         'HistoryOfSeverePreEclampsia',
         'MultiplePregnancy',
         'HistoryOfCardiovascularDisease']}

name_for_cid[3770] = 'ProblemStatus'
cid_concepts[3770] = \
{'SCT': ['Suspected',
         'ActiveProblem',
         'InactiveProblem',
         'KnownAbsent',
         'Recurrent',
         'WellControlled',
         'Chronic',
         'Intermittent',
         'ProblemResolved']}

name_for_cid[6122] = 'ChestContentDescriptor'
cid_concepts[6122] = \
{'DCM': ['Air', 'SoftTissue', 'Calcium'],
 'SCT': ['ForeignMaterialIodizedOilMercuryTalc', 'Fat']}

name_for_cid[4236] = 'IOLCalculationFormula'
cid_concepts[4236] = \
{'DCM': ['Haigis',
         'HaigisL',
         'HaigisToric',
         'Holladay1',
         'HaigisLToric',
         'Holladay2',
         'BarrettToric',
         'HofferQ',
         'BarrettTrueK',
         'Olsen',
         'BarrettTrueKToric',
         'SRKII',
         'BarrettUniversalII',
         'SRKT']}

name_for_cid[606] = 'AnimalShelterTypes'
cid_concepts[606] = \
{'DCM': ['Igloo', 'RedTranslucentIgloo'], 'SCT': ['None']}

name_for_cid[12227] = 'EchocardiographyMeasurementMethod'
cid_concepts[12227] = \
{'DCM': ['Teichholz',
         'MethodOfDisksSinglePlane',
         'MethodOfDisksBiplane',
         'CubeMethod',
         'BulletMethod',
         'AreaLengthSinglePlane',
         'ModifiedSimpson',
         'AreaLengthBiplane',
         'SinglePlaneEllipse',
         'LeftVentricleMassByTruncatedEllipse',
         'LeftVentricleMassByMMode',
         'Planimetry',
         'DopplerVolumeFlow',
         'SimplifiedBernoulli',
         'DirectlyMeasured',
         'FullBernoulli',
         'ProximalIsovelocitySurfaceArea',
         'ContinuityEquationByVelocityTimeIntegral',
         'LeftVentricleMassByTruncatedEllipseAdjustedByHeight',
         'LeftVentricleMassByMModeAdjustedByHeight',
         'LeftVentricleMassByAreaLength',
         'ContinuityEquationByPeakVelocity',
         'ContinuityEquationByMeanVelocity',
         'ContinuityEquation',
         'BiplaneEllipse',
         'AreaByPressureHalfTime',
         'LeftVentricleMassByAreaLengthAdjustedByHeight']}

name_for_cid[12033] = 'UltrasoundTransducerGeometry'
cid_concepts[12033] = \
{'DCM': ['RingUltrasoundTransducerGeometry',
         'RadialUltrasoundTransducerGeometry',
         'SectorUltrasoundTransducerGeometry',
         'CurvedLinearUltrasoundTransducerGeometry',
         'LinearUltrasoundTransducerGeometry',
         'NonImagingDopplerUltrasoundTransducerGeometry']}

name_for_cid[7462] = 'UnitsofVolumeMeasurement'
cid_concepts[7462] = \
{'UCUM': ['CubicMillimeter',
          'CubicCentimeter',
          'CubicMicrometer',
          'CubicDecimeter']}

name_for_cid[7198] = 'SubstanceSegmentationPropertyTypes'
cid_concepts[7198] = \
{'SCT': ['ContrastAgent']}

name_for_cid[3421] = 'ProcedureAction'
cid_concepts[3421] = \
{'DCM': ['SuspendProcedureAction',
         'ResumeProcedureAction',
         'StartProcedureAction',
         'EndProcedureAction']}

name_for_cid[9539] = 'FractionationBasedEffectiveDoseMethodModifers'
cid_concepts[9539] = \
{'DCM': ['LinearQuadraticModelForLowDoseRateBrachytherapy',
         'LinearQuadraticLinearModel',
         'LinearQuadraticModelWithTimeFactor',
         'LinearQuadraticModel',
         'Equivalent2GrayFractionsModel']}

name_for_cid[3837] = 'FiducialFeature'
cid_concepts[3837] = \
{'SCT': ['CommonIliacBifurcation',
         'RenalArtery',
         'Ostium',
         'AorticBifurcation']}

name_for_cid[12121] = 'VascularIndicesandRatios'
cid_concepts[12121] = \
{'LN': ['AccelerationIndex',
        'ResistivityIndex',
        'VelocityRatio',
        'SystolicToDiastolicVelocityRatio',
        'PulsatilityIndex'],
 'SCT': ['LumenDiameterStenosis', 'LumenAreaStenosis']}

name_for_cid[6048] = 'CADOperatingPointAxisLabel'
cid_concepts[6048] = \
{'DCM': ['ImageSpecificity',
         'CaseSpecificity',
         'ProbabilityOfCancer',
         'CertaintyOfFinding',
         'LesionSensitivity',
         'CaseSensitivity',
         'FalseMarkersPerCase',
         'FalseMarkersPerImage']}

name_for_cid[8113] = 'SpecimenPreparationSteps'
cid_concepts[8113] = \
{'SCT': ['SpecimenClearing',
         'SpecimenSteamHeating',
         'SpecimenFreezing',
         'SpecimenMicrowaveHeating',
         'ProteaseDigestionOfTissueSpecimen',
         'SpecimenDehydration']}

name_for_cid[629] = 'CarrierGasesforSmallAnimalAnesthesia'
cid_concepts[629] = \
{'SCT': ['MedicalAir', 'OxygenGas'], 'UMLS': ['RoomAir']}

name_for_cid[12208] = 'EchocardiographyTricuspidValve'
cid_concepts[12208] = \
{'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'StenosisPeakVelocity',
        'TimeFromQWaveToTricuspidValveOpens',
        'VolumeFlow',
        'FlowRadius',
        'HeartRate',
        'TimeAveragedPeakVelocity',
        'EndDiastolicVelocity',
        'TricuspidValveEToARatio',
        'VelocityOfFlowPropagation',
        'TricuspidRegurgitationDpDt',
        'PeakGradient',
        'TricuspidValveEWavePeakVelocity',
        'AliasVelocity',
        'TricuspidValveAWavePeakVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeAveragedMeanVelocity'],
 'SCT': ['RegurgitantFraction',
         'TricuspidValveClosureToOpeningTime',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter']}

name_for_cid[83] = 'UnitsforRealWorldValueMapping'
cid_concepts[83] = \
{'UCUM': ['MicromoleMinuteMilliliter',
          'Centimeter2Milliliter',
          'Centimeter2',
          'MilliliterGram',
          'StandardizedUptakeValueBodySurfaceArea',
          'StandardizedUptakeValueLeanBodyMassJanma',
          'ProportionalToCounts',
          'Counts',
          'Percent',
          'MilliliterMinuteGram',
          'StandardizedUptakeValueBodyWeight',
          'MicromoleMilliliter',
          'HounsfieldUnit',
          'ProportionalToCountsPerSecond',
          'MilligramsMinuteMilliliter',
          'StandardizedUptakeValueIdealBodyWeight',
          'Centimeter',
          'BecquerelsMilliliter',
          'CountsPerSecond',
          'StandardizedUptakeValueLeanBodyMassJames',
          'StandardizedUptakeValueLeanBodyMassJames128Multiplier']}

name_for_cid[3749] = 'PCIProcedureResult'
cid_concepts[3749] = \
{}

name_for_cid[3460] = 'CircumferentialExtent'
cid_concepts[3460] = \
{'DCM': ['RCARegionInRAOProjection', 'LADRegionInRAOProjection']}

name_for_cid[3525] = 'BloodGasContent'
cid_concepts[3525] = \
{'DCM': ['BloodOxygenContent'],
 'LN': ['ArterialOxygenContent',
        'PulmonaryArteryMainOxygenContent',
        'PulmonaryWedgeOxygenContent',
        'ArterialBloodCarbonDioxideContent',
        'BloodCarbonDioxideContent',
        'VenousBloodCarbonDioxideContent',
        'AorticRootOxygenContent',
        'VenousOxygenContent']}

name_for_cid[280] = 'LongitudinalTemporalEventTypes'
cid_concepts[280] = \
{'DCM': ['Baseline'], 'NCIt': ['Enrollment']}

name_for_cid[404] = 'AuditParticipantObjectIDTypeCode'
cid_concepts[404] = \
{'DCM': ['StudyInstanceUID', 'NodeID', 'SOPClassUID']}

name_for_cid[6009] = 'DensityModifierfromBIRADS®'
cid_concepts[6009] = \
{'SCT': ['LowDensityNotContainingFatLesion',
         'HighDensityLesion',
         'EqualDensityIsodenseLesion',
         'FatContainingRadiolucentLesion']}

name_for_cid[7024] = 'RTProcessInput'
cid_concepts[7024] = \
{'DCM': ['ForRTPlanSummation',
         'ForPhysicistReview',
         'ForPhysicianReview',
         'ForPlanQualityAssurance',
         'ForPatientSetupVerification',
         'ForMachineQualityAssurance',
         'ForRTWorkflow',
         'ForRTTreatmentPlanning',
         'ForRTPrescription']}

name_for_cid[7474] = 'GeneralVolumeCalculationMethods'
cid_concepts[7474] = \
{'DCM': ['IntegrationOfSumOfClosedAreasOnContiguousSlices',
         'SumOfSegmentedVoxelVolumes']}

name_for_cid[7161] = 'PhysiologySegmentationTypes'
cid_concepts[7161] = \
{'SCT': ['Perfusion']}

name_for_cid[12249] = 'CardiacUltrasoundFetalSummaryCodes'
cid_concepts[12249] = \
{'SCT': ['CorrectedTranspositionOfGreatVessels',
         'CoarctationOfTheAorta',
         'PulmonaryAtresiaWithIntactVentricularSeptum',
         'TetralogyOfFallot',
         'CompleteTranspositionOfGreatVessels',
         'Mesocardia',
         'FunctionalSingleVentricle',
         'MitralRegurgitation',
         'HypoplasticLeftHeartSyndrome',
         'CongenitalStenosisOfAorticValve',
         'MitralStenosis',
         'AtrialSeptalDefect',
         'RheumaticFever',
         'PulmonaryAtresiaWithVentricularSeptalDefect',
         'MitralValveProlapse',
         'AorticRegurgitation',
         'EdemaOfFetalChestWall',
         'Dextrocardia',
         'CommonAtrioventricularCanal',
         'VentricularSeptalDefect',
         'EdemaOfFetalScalp',
         'PartialAnomalousPulmonaryVenousConnection',
         'AcuteFebrileMucocutaneousLymphNodeSyndrome',
         'Cardiomegaly',
         'Cyanosis',
         'AtrialSeptalDefectWithEndocardialCushionDefectPartial',
         'PulmonicValveStenosis',
         'AorticStenosis',
         'InterruptedAorticArch']}

name_for_cid[3821] = 'PapillaryMuscleIncluded/Excluded'
cid_concepts[3821] = \
{'DCM': ['PapillaryMuscleIncluded', 'PapillaryMuscleExcluded']}

name_for_cid[3122] = 'VentricularEnlargement'
cid_concepts[3122] = \
{'SCT': ['MarkedlyEnlargedCardiacChamber',
         'MildlyEnlargedCardiacChamber',
         'ModeratelyEnlargedCardiacChamber',
         'NormalSizeCardiacChamber']}

name_for_cid[3335] = 'ECGAnnotations'
cid_concepts[3335] = \
{'MDC': ['CalibrationPulse',
         'AtriumAntitachycardiaPaceSpike',
         'AberratedAtrialPrematureBeatAshmanBeat',
         'PreExcitation',
         'LeftVentricleAntibradycardiaPaceSpike',
         'LeftPosteriorFascicularBlockBeat',
         'AtrialFlutterWave',
         'AbnormalBeat',
         'EntireBeatPonToToffExcludingU',
         'UWave',
         'LeftBundleBranchBlockBeat',
         'TWaveChange',
         'JPoint',
         'TransthoracicCardioversionSpike',
         'NoNoise',
         'VentricleDefibrillationSpike',
         'LeftAtriumAntibradycardiaPaceSpike',
         'NoECGSignalIsAvailable',
         'SWave',
         'QSWave',
         'FusionOfVentricularAndNormalBeat',
         'AtrialPrematureContraction',
         'BilateralBundleBranchBlockBeat',
         'DefibrillationSpike',
         'AtriumCardioversionSpike',
         'ModerateNoiseBeatsCanBeDetectedButCannotBeClassified',
         'IsoelectricRegionFromGlobalQRSOnsetToSpecificLeadOnset',
         'IncompleteRightBundleBranchBlockBeat',
         'VentricularAntibradycardiaPaceSpike',
         'VentricularEscapeBeat',
         'TWaveSecondDeflectionInTWave',
         'PacemakerLearningBeat',
         'IsolatedQRSLikeArtifact',
         'WolfParkinsonTypeB',
         'AtriumAntibradycardiaPaceSpike',
         'VentricleAntitachycardiaPaceSpike',
         'PRSegment',
         'EntireQRSExcludingPTAndU',
         'TransthoracicAntibradycardiaPaceSpike',
         'VentricularFlutterWave',
         'OsborneWave',
         'Notch',
         'QWave',
         'BifascicularBlockBeat',
         'SWaveThirdDeflectionInSWave',
         'DominantBeat',
         'PWaveSecondDeflectionInPWave',
         'RWaveSecondDeflectionInRWave',
         'IncompleteLeftBundleBranchBlockBeat',
         'RightVentricleAntibradycardiaPaceSpike',
         'AtriumDefibrillationSpike',
         'ROnTPrematureVentricularBeat',
         'PacemakerFusionBeat',
         'LownGanongLevineSyndrome',
         'TransthoracicDefibrillationSpike',
         'AtrialEscapeBeat',
         'NonConductedPWave',
         'AntitachycardiaPaceSpike',
         'JunctionalNodalPrematureContraction',
         'WolfParkinsonWhiteSyndrome',
         'IntraventricularConductionDisturbance',
         'IsoelectricPointOrSegment',
         'LeftAnteriorFascicularBlockBeat',
         'EntireBeatQonToToffExcludingPAndU',
         'DeltaWave',
         'STChange',
         'NormalBeatSinusBeatNormalConduction',
         'RightAtriumAntibradycardiaPaceSpike',
         'SWaveSecondDeflectionInSWave',
         'VentricleCardioversionSpike',
         'PWave',
         'RWave',
         'WolfParkinsonTypeA',
         'VentricularPrematureContractionBeat',
         'STMeasurementPoint',
         'AntibradycardiaPaceSpike',
         'SupraventricularPrematureContraction',
         'CardioversionSpike',
         'TrifascicularBlockBeat',
         'RightBundleBranchBlockBeat',
         'STSegment',
         'TUFusedWave',
         'IsoelectricRegionFromSpecificLeadQRSOffsetToGlobalOffset',
         'SevereNoiseBeatsCannotBeDetectedOrClassified',
         'TWave',
         'VentricularActivationTime',
         'TransthoracicAntitachycardiaPaceSpike',
         'JunctionalNodalEscapeBeat',
         'UnclassifiableBeat',
         'RWaveThirdDeflectionInRWave',
         'SupraventricularEscapeBeat',
         'PWaveThirdDeflectionInPWave']}

name_for_cid[6137] = 'TypesofCADAnalysis'
cid_concepts[6137] = \
{'SCT': ['SpatialCollocationAnalysis',
         'SpatialProximityAnalysis',
         'TemporalCorrelation',
         'ImageQualityAnalysis']}

name_for_cid[10063] = 'RadiationDoseEstimateDistributionRepresentation'
cid_concepts[10063] = \
{'DCM': ['SkinDoseMap',
         'DosePointCloud',
         'DoseGradient',
         '_3DDoseMap',
         'DoseImage',
         'Isodose']}

name_for_cid[6072] = 'BreastImplantFindings'
cid_concepts[6072] = \
{'DCM': ['DistortedImplant',
         'SiliconeLadenLymphNodes',
         'FreeSilicone',
         'HerniatedImplant',
         'Explantation',
         'NormalImplants',
         'AsymmetricImplants',
         'CalcifiedImplant'],
 'SCT': ['RuptureOfBreastImplant']}

name_for_cid[3620] = 'HemodynamicRatios'
cid_concepts[3620] = \
{'DCM': ['DiastolicSystolicVelocityRatio',
         'HyperemicRatio',
         'CoronaryFlowReserve',
         'HemodynamicResistanceIndex'],
 'LN': ['TibialBrachialIndex'],
 'SCT': ['FractionalFlowReserve',
         'FractionalFlowReserveUsingIntracoronaryBolus',
         'PulmonarySystemicFlowRatio',
         'PulsatilityIndex',
         'FractionalFlowReserveUsingIntravenousInfusion']}

name_for_cid[12232] = 'MyocardiumMassMethods'
cid_concepts[12232] = \
{'DCM': ['LeftVentricleMassByTruncatedEllipseAdjustedByHeight',
         'LeftVentricleMassByMModeAdjustedByHeight',
         'LeftVentricleMassByAreaLength',
         'LeftVentricleMassByTruncatedEllipse',
         'LeftVentricleMassByMMode',
         'LeftVentricleMassByAreaLengthAdjustedByHeight']}

name_for_cid[613] = 'AnesthesiaInductionCodeTypeforSmallAnimalAnesthesia'
cid_concepts[613] = \
{'SCT': ['IntramuscularRoute',
         'PerRectum',
         'IntravenousRoute',
         'IntraperitonealRoute',
         'ByInhalation']}

name_for_cid[10034] = 'ReasonforRepeatingAcquisition'
cid_concepts[10034] = \
{'DCM': ['SuboptimalContrastTiming', 'PatientMotion']}

name_for_cid[12009] = 'EarlyGestationBiometryMeasurements'
cid_concepts[12009] = \
{'LN': ['GestationalSacDiameter',
        'CrownRumpLength',
        'SpineLength',
        'YolkSacLength',
        'NuchalTranslucency']}

name_for_cid[6160] = 'BreastPrimaryTumorAssessmentFromAJCC'
cid_concepts[6160] = \
{'SCT': ['TXPrimaryTumorCannotBeAssessed',
         'T1bTumorMoreThan0Point5cmButNotMoreThan1cm',
         'T4InflammatoryCarcinoma',
         'T1Tumor2cmOrLessInGreatestDimension',
         'T0NoEvidenceOfPrimaryTumor',
         'TisCarcinomaInSitu',
         'TisDuctalCarcinomaInSitu',
         'T2TumorMoreThan2cmButNotMoreThan5cm',
         'T1cTumorMoreThan1cmButNotMoreThan2cm',
         'TisPagetDiseaseOfTheNippleWithNoTumor',
         'T1aTumorMoreThan0Point1cmButNotMoreThan0Point5cm',
         'T4aTumorOfAnySizeWithExtensionToChestWallNotIncl',
         'T4bTumorOfAnySizeWithEdemaIncludingPeauDOrange',
         'T1micMicroinvasion0Point1cmOrLessInGreatestDimension',
         'T4TumorOfAnySizeWithDirectExtensionToChestWall',
         'T4cTumorOfAnySizeWithDirectExtensionToChestWall',
         'TisLobularCarcinomaInSitu',
         'T3TumorMoreThan5cmInGreatestDimension']}

name_for_cid[3677] = 'SummaryCodesECG'
cid_concepts[3677] = \
{'DCM': ['NonDiagnosticECG'],
 'SCT': ['BorderlineNormalECG', 'ECGEquivocal', 'NormalECG', 'AbnormalECG']}

name_for_cid[3227] = 'QTcMeasurements'
cid_concepts[3227] = \
{'MDC': ['QtcPerLeadUsingFredericiaFormula',
         'QtcIntervalPerLead',
         'QtcIntervalGlobal',
         'QtcGlobalUsingFredericiaFormula',
         'QtcGlobalUsingBazettFormula',
         'QtcPerLeadUsingBazettFormula']}

name_for_cid[4018] = 'PrimaryAnatomicStructureforIntraoralRadiography(PermanentDentitionDesignationofTeeth)'
cid_concepts[4018] = \
{'SCT': ['MaxillaryLeftFirstPremolarTooth',
         'MaxillaryRightCentralIncisorTooth',
         'MandibularLeftSecondPremolarTooth',
         'MandibularRightCanineTooth',
         'MaxillaryLeftThirdMolarTooth',
         'MandibularLeftCanineTooth',
         'MaxillaryLeftSecondPremolarTooth',
         'MandibularRightLateralIncisorTooth',
         'MaxillaryRightSecondPremolarTooth',
         'MandibularRightFirstMolarTooth',
         'MaxillaryRightCanineTooth',
         'MaxillaryLeftSecondMolarTooth',
         'MandibularRightSecondPremolarTooth',
         'MaxillaryRightLateralIncisorTooth',
         'MandibularLeftSecondMolarTooth',
         'MaxillaryLeftCanineTooth',
         'MandibularLeftFirstPremolarTooth',
         'MandibularLeftFirstMolarTooth',
         'MandibularRightFirstPremolarTooth',
         'MaxillaryRightFirstMolarTooth',
         'MandibularLeftLateralTooth',
         'MandibularLeftCentralIncisorTooth',
         'MaxillaryRightThirdMolarTooth',
         'MaxillaryLeftCentralIncisorTooth',
         'MandibularLeftThirdMolarTooth',
         'MaxillaryRightFirstPremolarTooth',
         'MaxillaryLeftLateralIncisorTooth',
         'MaxillaryRightSecondMolarTooth',
         'MandibularRightSecondMolarTooth',
         'MaxillaryLeftFirstMolarTooth',
         'MandibularRightCentralIncisorTooth',
         'MandibularRightThirdMolarTooth']}

name_for_cid[6025] = 'DepthfromBIRADS®'
cid_concepts[6025] = \
{'NCIt': ['Middle'], 'SCT': ['Posterior', 'Anterior']}

name_for_cid[9503] = 'RadiotherapyTherapeuticRoleCategories'
cid_concepts[9503] = \
{'DCM': ['RTDoseCalculationStructure', 'RTTarget']}

name_for_cid[701] = 'ContentAssessmentTypes'
cid_concepts[701] = \
{'DCM': ['RTPreTreatmentConsistencyCheck', 'RTPreTreatmentDoseCheck']}

name_for_cid[12265] = 'CardiacUltrasoundAtriaandAtrialSeptumMeasurements'
cid_concepts[12265] = \
{'DCM': ['WallMass'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'VolumeFlow',
        'FlowRadius',
        'LeftAtriumToAorticRootRatio',
        'HeartRate',
        'FlowArea',
        'EndDiastolicVelocity',
        'EndocardialArea',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'LeftAtriumVolumeToRightAtriumVolumeRatio',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'ACInterval',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'VelocityOfFlowPropagation',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'EpicardialArea',
        'MyocardialPerformanceIndexTei',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Volume',
         'Area',
         'CardiovascularOrificeDiameter',
         'Length',
         'Circumference',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'CardiovascularOrificeArea']}

name_for_cid[644] = 'ExogenousSubstanceAdministrationSites'
cid_concepts[644] = \
{'NCIt': ['MouseMammaryFatPad'],
 'SCT': ['Liver', 'BoneMarrow', 'Tumor', 'Brain', 'Flank']}

name_for_cid[7008] = 'MediaImport'
cid_concepts[7008] = \
{'DCM': ['InstancesImported',
         'StudiesImported',
         'MODImported',
         'DVDImported',
         'CDImported',
         'PaperDigitized',
         'VideoTapeDigitized',
         'CineFilmDigitized',
         'SheetFilmDigitized']}

name_for_cid[3724] = 'SmokingHistory'
cid_concepts[3724] = \
{'SCT': ['CurrentSmoker', 'NoHistoryOfSmoking', 'FormerSmoker']}

name_for_cid[10022] = 'LabelTypes'
cid_concepts[10022] = \
{'DCM': ['SeriesNumber', 'AcquisitionNumber', 'InstanceNumber']}

name_for_cid[2] = 'AnatomicModifier'
cid_concepts[2] = \
{'DCM': ['Median'],
 'SCT': ['Peripheral',
         'Basal',
         'Bilateral',
         'Right',
         'Surface',
         'Caudal',
         'Posterior',
         'Vertical',
         'Sagittal',
         'Left',
         'External',
         'Unilateral',
         'Superior',
         'IntraArticular',
         'Edge',
         'Distal',
         'Proximal',
         'Axial',
         'Marginal',
         'Preaxial',
         'Deep',
         'Subcapsular',
         'Lateral',
         'Inferior',
         'Capsular',
         'Anterior',
         'Intermediate',
         'Transverse',
         'Cephalic',
         'Anterolateral',
         'Central',
         'Hilar',
         'Apical',
         'ExtraArticular',
         'Postaxial',
         'Horizontal',
         'Posterolateral',
         'Afferent',
         'Superficial',
         'Gutter',
         'Coronal',
         'Efferent',
         'Internal',
         'Medial',
         'Longitudinal']}

name_for_cid[7272] = 'MRDiffusionModelParameters'
cid_concepts[7272] = \
{'DCM': ['ApparentDiffusionCoefficient',
         'KurtosisDiffusionCoefficient',
         'GammaDistributionScaleParameter',
         'GammaDistributionShapeParameter',
         'GammaDistributionMode',
         'DistributedDiffusionCoefficient',
         'AnomalousExponentParameter',
         'DiffusionCoefficient',
         'MonoExponentialApparentDiffusionCoefficient',
         'SlowDiffusionCoefficient',
         'FastDiffusionCoefficient',
         'FastDiffusionCoefficientFraction']}

name_for_cid[6033] = 'MalignantPathologyCodesfromBIRADS®'
cid_concepts[6033] = \
{'DCM': ['SquamousCellCarcinomaOfTheNipple',
         'IntraductalCarcinomaHighGrade',
         'CartilaginousAndOsseousChange',
         'AxillaryNodeWithLymphoma',
         'AxillaryNodalMetastases',
         'BloodVesselVascularInvasion',
         'CarcinomaInChildren',
         'CarcinomaInEctopicBreast',
         'CarcinomaWithEndocrineDifferentiation',
         'BasalCellCarcinomaOfNipple',
         'MetastaticDiseaseToAxillaryNode',
         'MultifocalInvasiveDuctalCarcinoma',
         'MetastasisToAnIntramammaryLymphNode',
         'MalignantMelanomaOfNipple',
         'RecurrentMalignancy',
         'LympathicVesselInvasion',
         'OccultCarcinomaPresentingWithAxillaryLymphNodeMetastases',
         'MetastaticCancerToTheBreast',
         'MetastaticCancerToTheBreastFromTheColon',
         'MetastaticCancerToTheBreastFromTheLung',
         'MetastaticMelanomaToTheBreast',
         'MetastaticCancerToTheBreastFromTheOvary',
         'MetastaticSarcomaToTheBreast',
         'MultifocalIntraductalCarcinoma',
         'CarcinomaInPregnancyAndLactation',
         'IntraductalComedocarcinomaWithNecrosis',
         'IntraductalCarcinomaLowGrade',
         'IntracysticPapillaryCarcinoma',
         'InvasiveAndInSituCarcinoma',
         'LeukemicInfiltration'],
 'SCT': ['Leiomyosarcoma',
         'SquamousCellCarcinoma',
         'Carcinosarcoma',
         'TubularAdenocarcinoma',
         'ApocrineAdenocarcinoma',
         'PapillaryCarcinoma',
         'InflammatoryCarcinoma',
         'HodgkinDisease',
         'LipidRichLipidSecretingCarcinoma',
         'CarcinomaInSituOfMaleBreast',
         'InvasiveCribriformCarcinoma',
         'AdenoidCysticCarcinoma',
         'Plasmacytoma',
         'Angiosarcoma',
         'MedullaryCarcinoma',
         'MalignantFibrousHistiocytoma',
         'Fibrosarcoma',
         'SignetRingCellCarcinoma',
         'Lymphoma',
         'NeoplasmOfTheMammarySkin',
         'PhyllodesTumorMalignant',
         'GlycogenRichCarcinoma',
         'CarcinomaWithMetaplasia',
         'Chondrosarcoma',
         'SecretoryJuvenileCarcinomaOfTheBreast',
         'IntraductalCarcinomaMicroPapillary',
         'Hemangiopericytoma',
         'MucinousAdenocarcinomaColloidCarcinoma',
         'NonHodgkinLymphoma',
         'PagetDiseaseMammary',
         'InfiltratingDuctCarcinoma',
         'PapillaryCarcinomaInSitu',
         'SpindleCellNodule',
         'Liposarcoma',
         'Comedocarcinoma',
         'InvasiveLobularCarcinoma',
         'OsteogenicSarcoma']}

name_for_cid[3231] = 'STSegmentFindings'
cid_concepts[3231] = \
{'DCM': ['StronglyPositive',
         'StronglyPositiveSTElevation',
         'NonDiagnosticRestingSTAbnormalities',
         'NonDiagnosticVentricularPacingOrLBBB',
         'NonDiagnosticLowHeartRate'],
 'SCT': ['WeaklyPositive', 'STIntervalNormal', 'Positive']}

name_for_cid[12273] = 'CardiacUltrasoundAorticSinotubularJunction'
cid_concepts[12273] = \
{'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'StenosisPeakVelocity',
        'VolumeFlow',
        'FlowRadius',
        'HeartRate',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'EndDiastolicVelocity',
        'AorticSinotubularJunctionToAorticRootRatio',
        'ROIInternalDimensionByUS',
        'VelocityOfFlowPropagation',
        'PeakGradient',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter',
         'Length',
         'Circumference']}

name_for_cid[9515] = 'RTPatientSupportDevices'
cid_concepts[9515] = \
{'SCT': ['Chair', 'Table']}

name_for_cid[3118] = 'LevelofRisk'
cid_concepts[3118] = \
{'SCT': ['ModerateToHighRisk',
         'HighRisk',
         'LowToModerateRisk',
         'ModerateRisk',
         'NormalRisk',
         'LowRisk',
         'UncertainRisk']}

name_for_cid[9000] = 'PhysicalQuantityDescriptors'
cid_concepts[9000] = \
{'DCM': ['SourceImageDiffusionBValue',
         'ModelFittingMethod',
         'Derivation',
         'Finding',
         'EquivalentMeaningOfConceptName'],
 'NCIt': ['ReferenceRegion'],
 'SCT': ['Quantity', 'FindingSite', 'MeasurementMethod']}

name_for_cid[4262] = 'RetinalThicknessDefinition'
cid_concepts[4262] = \
{'DCM': ['RetinalNerveFiberLayerThickness',
         'GanglionCellComplexThickness',
         'TotalRetinalThicknessILMToISOS',
         'TotalRetinalThicknessILMToRPE',
         'TotalRetinalThicknessILMToBM']}

name_for_cid[3773] = 'UseStatus'
cid_concepts[3773] = \
{'SCT': ['Suspended', 'Ended', 'InProgress']}

name_for_cid[6121] = 'ChestOrientationDescriptor'
cid_concepts[6121] = \
{'SCT': ['Horizontal', 'Vertical', 'Oblique']}

name_for_cid[6064] = 'UltrasoundFindingsforBreast'
cid_concepts[6064] = \
{'DCM': ['ComplexCyst',
         'ComplicatedCyst',
         'IntracysticLesion',
         'SolidMass',
         'ClusteredMicrocysts'],
 'SCT': ['SebaceousCystOfSkinOfBreast',
         'ForeignBody',
         'LymphNode',
         'CystOfBreast',
         'MammaryDuctEctasia',
         'UltrasoundScanNormal']}

name_for_cid[12224] = 'UltrasoundImageModes'
cid_concepts[12224] = \
{'SCT': ['DopplerPulsed',
         'MMode',
         'TissueDopplerImaging',
         '_2DMode',
         '_3DMode',
         'DopplerContinuousWave',
         'PowerDoppler',
         'DopplerColorFlow']}

name_for_cid[605] = 'AnimalBeddingMaterial'
cid_concepts[605] = \
{'DCM': ['AspenShavingBedding',
         'CornCobBedding',
         'PaperBasedBedding',
         'PineChipBedding',
         'PineShavingBedding',
         'AspenChipBedding']}

name_for_cid[4235] = 'KeratometryDescriptors'
cid_concepts[4235] = \
{'DCM': ['ManualKeratometry',
         'AutoKeratometry',
         'SimulatedKeratometry',
         'EquivalentKReading']}

name_for_cid[6208] = 'ColonTypesofQualityControlStandard'
cid_concepts[6208] = \
{'DCM': ['InstitutionallyDefinedQualityControlStandard',
         'ACRPositionStatement',
         'ACRGuidelinePerformanceOfAdultCTColonography',
         'ACRStandardCTMedicalPhysicsPerformanceMonitoring']}

name_for_cid[4242] = 'CorneaMeasurementMethodDescriptors'
cid_concepts[4242] = \
{'DCM': ['PosteriorCorneaSurfaceMeasurementMethod',
         'ManualKeratometry',
         'AutoKeratometry',
         'SimulatedKeratometry',
         'EquivalentKReading',
         'TotalCorneaPowerMeasurementMethod']}

name_for_cid[222] = 'NormalityCodes'
cid_concepts[222] = \
{'SCT': ['AbnormallyHigh',
         'Abnormal',
         'AbnormallyLow',
         'Normal',
         'NormalityUndetermined']}

name_for_cid[7194] = 'MorphologicallyAbnormalStructureSegmentationPropertyTypes'
cid_concepts[7194] = \
{'DCM': ['Background'],
 'NCIt': ['EnhancingLesion', 'NonEnhancingLesion'],
 'SCT': ['Embolus',
         'Mass',
         'Normal',
         'Cyst',
         'Abscess',
         'Edema',
         'NeoplasmPrimary',
         'BloodClot',
         'Nodule',
         'Neoplasm',
         'NeoplasmSecondary',
         'Necrosis',
         'Hemorrhage',
         'Inflammation']}

name_for_cid[7481] = 'BreedRegistry'
cid_concepts[7481] = \
{'DCM': ['CanadianBorderCollieAssociation',
         'UnitedKennelClub',
         'AnimalResearchFoundation',
         'UnitedAllBreedRegistry',
         'CanadianLivestockRecordsAssociation',
         'CanadianKennelClub',
         'UniversalKennelClubInternational',
         'AmericanPurebredRegistry',
         'InternationalProgressiveDogBreedersAlliance',
         'AmericanCanineAssociation',
         'FederationOfInternationalCanines',
         'AnimalRegistryUnlimited',
         'NorthAmericanPurebredDogRegistry',
         'AmericanRareBreedAssociation',
         'NationalKennelClub',
         'CanineFederationOfCanada',
         'WorldKennelClub',
         'WorkingCanineAssociationOfCanada',
         'AmericaPetRegistryInc',
         'DogRegistryOfAmerica',
         'AmericaKennelClub',
         'ContinentalKennelClub',
         'WorldWideKennelClub']}

name_for_cid[12253] = 'CardiacUltrasoundAreaMeasurements'
cid_concepts[12253] = \
{'LN': ['JetArea', 'FlowArea'], 'SCT': ['Area']}

name_for_cid[9535] = 'RadiotherapyDoseCalculationRoles'
cid_concepts[9535] = \
{'DCM': ['AvoidanceVolume',
         'PlanningOrganAtRiskVolume',
         'RadiationInteractionVolume',
         'RadiationDoseShapingVolume',
         'OrganAtRisk',
         'DoseCalculationBoundingVolume',
         'ConformalityShell']}

name_for_cid[6013] = 'CalcificationDistributionModifierfromBIRADS®'
cid_concepts[6013] = \
{'DCM': ['CalcificationsWithinAMass', 'CalcificationsOutsideOfAMass'],
 'SCT': ['SegmentalCalcificationDistribution',
         'LinearCalcificationDistribution',
         'RegionalCalcificationDistribution',
         'DiffuseCalcificationDistribution',
         'GroupedCalcificationDistribution']}

name_for_cid[3211] = 'BPResponse'
cid_concepts[3211] = \
{'SCT': ['Normal', 'Blunted', 'Hypertensive', 'Hypotensive']}

name_for_cid[3641] = 'HemodynamicAssessments'
cid_concepts[3641] = \
{'SCT': ['PulmonaryVascularResistance',
         'LeftVentricularSystolicPressure',
         'RightAtrialPressure',
         'LeftVentricularEndDiastolicPressure',
         'RightVentricularEndDiastolicPressure',
         'PulmonaryCapillaryWedgePressure',
         'RightVentricularSystolicPressure',
         'DiastolicPressure',
         'PulmonaryArteryPressure',
         'VascularResistance']}

name_for_cid[3704] = 'MyocardiumWallMorphologyFindings'
cid_concepts[3704] = \
{'DCM': ['ScarredMyocardium', 'ThinningMyocardium', 'NormalMyocardium'],
 'SCT': ['VentricularAneurysm']}

name_for_cid[21] = 'PatientEquipmentRelationship'
cid_concepts[21] = \
{'DCM': ['LeftFirst', 'RightFirst', 'PosteriorFirst', 'AnteriorFirst'],
 'SCT': ['Headfirst', 'FeetFirst', 'Transverse', 'Oblique']}

name_for_cid[10002] = 'IrradiationEventTypes'
cid_concepts[10002] = \
{'DCM': ['StationaryAcquisition',
         'SteppingAcquisition',
         'RotationalAcquisition'],
 'SCT': ['Fluoroscopy']}

name_for_cid[625] = 'InjectableAnesthesiaAgentsforSmallAnimalAnesthesia'
cid_concepts[625] = \
{'SCT': ['Pentobarbital',
         'Xylazine',
         'Diazepam',
         'Propofol',
         'Urethane',
         'Midazolam',
         'Thiamylal',
         'Etomidate',
         'Alphachloralose',
         'ChloralHydrate',
         'Tribromoethanol',
         'Methohexital',
         'Alphadolone',
         'Butabarbital',
         'Ketamine',
         'Thiopental',
         'Azaperone',
         'Droperidol',
         'Tiletamine',
         'Zolazepam'],
 'UMLS': ['Alphaxalone', 'Fluanisone', 'Metomidate']}

name_for_cid[4215] = 'OphthalmicVerticalDirections'
cid_concepts[4215] = \
{'SCT': ['Up', 'Down']}

name_for_cid[3495] = 'IVUSPlaqueComposition'
cid_concepts[3495] = \
{'DCM': ['NecroticLipidicPlaque', 'FibroLipidicPlaque'],
 'SCT': ['VascularCalcification', 'Thrombus', 'FibrousPlaque']}

name_for_cid[12204] = 'EchocardiographyRightVentricle'
cid_concepts[12204] = \
{'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'StenosisPeakVelocity',
        'VolumeFlow',
        'FlowRadius',
        'RightVentricularAnteriorWallSystolicThickness',
        'HeartRate',
        'TimeAveragedPeakVelocity',
        'RightVentricularAnteriorWallDiastolicThickness',
        'EndDiastolicVelocity',
        'VelocityOfFlowPropagation',
        'RightVentricularInternalDiastolicDimension',
        'PeakGradient',
        'AliasVelocity',
        'RightVentricularInternalSystolicDimension',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeAveragedMeanVelocity'],
 'SCT': ['RVStrokeIndex',
         'RVCardiacOutput',
         'RightVentricularIndexOfMyocardialPerformance',
         'RegurgitantFraction',
         'RVStrokeVolume',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter',
         'RVCardiacIndex',
         'RightVentricularPeakSystolicPressure']}

name_for_cid[3616] = 'HemodynamicPeriodMeasurements'
cid_concepts[3616] = \
{'SCT': ['PulmonarySystolicEjectionPeriodSepp',
         'AorticSystolicEjectionPeriodSepa',
         'DerivedPeriodNonValve',
         'TricuspidDiastolicFillingPeriodDfpt',
         'MitralDiastolicFillingPeriodDfpm']}

name_for_cid[6044] = 'TypesofImageQualityAssessment'
cid_concepts[6044] = \
{'DCM': ['UnusableQualityRendersImageUnusable',
         'UsableDoesNotMeetTheQualityControlStandard',
         'UsableMeetsTheQualityControlStandard']}

name_for_cid[7205] = 'PurposeofReferencetoAlternateRepresentation'
cid_concepts[7205] = \
{'DCM': ['FullFidelityImage',
         'AlternatePhotometricInterpretationImage',
         'LossyCompressedImage',
         'AlternateSOPClassInstance',
         'SourceImage']}

name_for_cid[7710] = 'TractographyAnatomicSites'
cid_concepts[7710] = \
{'DCM': ['Phantom',
         'PerilesionalWhiteMatter',
         'SuperiorLongitudinalFasciculusIII',
         'SuperiorLongitudinalFasciculusII',
         'SuperiorLongitudinalFasciculusI'],
 'NEU': ['SuperiorLongitudinalFasciculus',
         'SuperiorThalamicRadiation',
         'InferiorThalamicRadiation',
         'PosteriorThalamicRadiation',
         'StriaTerminalis',
         'InferiorOpticRadiationMeyerLoop',
         'CorticospinalTract',
         'CorticobulbarTract',
         'ArcuateFasciculus',
         'SuperiorOpticRadiationBaumLoop',
         'AnteriorThalamicRadiation'],
 'SCT': ['SpinalCordWhiteMatter',
         'ForcepsMinor',
         'LateralFuniculus',
         'SuperiorFrontoOccipitalFasciculus',
         'InferiorLongitudinalFasciculus',
         'SkeletalMuscle',
         'MiddleCerebellarPeduncle',
         'AbducensNerve',
         'GenuOfCorpusCallosum',
         'ForcepsMajor',
         'InferiorFrontoOccipitalFasciculus',
         'CorpusCallosum',
         'VagusNerve',
         'VentralFuniculus',
         'OpticNerve',
         'ExternalCapsule',
         'MedialLongitudinalFasciculus',
         'AnteriorCommissure',
         'CerebralWhiteMatter',
         'PeripheralNerve',
         'OlfactoryTract',
         'MedialLemniscus',
         'AccessoryNerve',
         'InferiorCerebellarPeduncle',
         'FacialNerve',
         'GlossopharyngealNerve',
         'WhiteMatterOfBrainAndSpinalCord',
         'OculomotorNerve',
         'LateralLemniscus',
         'TapetumOfCorpusCallosum',
         'CorticospinalTractInBrainstem',
         'PosteriorCommissure',
         'SuperiorCerebellarPeduncle',
         'UncinateFasciculus',
         'HabenularCommissure',
         'CardiacMuscle',
         'AuditoryRadiation',
         'SpleniumOfCorpusCallosum',
         'VestibulocochlearNerve',
         'CerebellarWhiteMatter',
         'Fornix',
         'Cingulum',
         'InternalCapsule',
         'TrigeminalNerve',
         'HypoglossalNerve',
         'VerticalOccipitalFasciculus',
         'TrochlearNerve',
         'DorsalFuniculus',
         'OpticRadiation']}

name_for_cid[76] = 'TypeofPreMedication'
cid_concepts[76] = \
{'DCM': ['ContrastReactionProphylacticAgent', 'PreMedication'],
 'SCT': ['Sedative', 'Antiemetic']}

name_for_cid[6101] = 'ChestFindingorFeature'
cid_concepts[6101] = \
{'DCM': ['AbnormalLucency',
         'AbnormalCalcifications',
         'AbnormalTexture',
         'AbnormalOpacity',
         'ImageQuality',
         'NonLesion',
         'RadiographicAnatomy',
         'SelectedRegion',
         'AbnormalLines1D']}

name_for_cid[1001] = 'AnatomicalReferenceBasis'
cid_concepts[1001] = \
{'FMA': ['Nasion'],
 'SCT': ['KneeJoint',
         'TemporalBone',
         'Sacrum',
         'LevelOfL1L2IntervertebralDisc',
         'ApexOfHeart',
         'SternoclavicularJoint',
         'LevelOfT7T8IntervertebralDisc',
         'C1Vertebra',
         'LevelOfC5C6IntervertebralDisc',
         'WristJoint',
         'LevelOfT11T12IntervertebralDisc',
         'Ischium',
         'Diaphragm',
         'LevelOfT3T4IntervertebralDisc',
         'CarotidArtery',
         'LevelOfT2T3IntervertebralDisc',
         'Talus',
         'FemoralHead',
         'MaxillarySinus',
         'LevelOfT10T11IntervertebralDisc',
         'MalleolarStructureOfTibia',
         'LevelOfT6T7IntervertebralDisc',
         'LevelOfC2C3IntervertebralDisc',
         'Mandible',
         'Metacarpal',
         'Foot',
         'CommonIliacArteryBifurcation',
         'LevelOfC6C7IntervertebralDisc',
         'MastoidBone',
         'XiphoidProcess',
         'Acetabulum',
         'LevelOfL4L5IntervertebralDisc',
         'ForamenMagnum',
         'AnkleJoint',
         'LevelOfL5S1IntervertebralDisc',
         'LevelOfT9T10IntervertebralDisc',
         'Lung',
         'LevelOfT5T6IntervertebralDisc',
         'PituitaryFossa',
         'LevelOfL3L4IntervertebralDisc',
         'SymphysisPubis',
         'LevelOfT12L1IntervertebralDisc',
         'TibialPlateau',
         'LesserTrochanter',
         'AorticArch',
         'Carina',
         'LevelOfC3C4IntervertebralDisc',
         'Heart',
         'Toe',
         'ShoulderRegionStructure',
         'ElbowJoint',
         'LateralCanthus',
         'Sternum',
         'ThoracicInlet',
         'IliacCrest',
         'OrbitalStructure',
         'LevelOfT8T9IntervertebralDisc',
         'LevelOfL2L3IntervertebralDisc',
         'HipJoint',
         'LevelOfC7T1IntervertebralDisc',
         'LevelOfC4C5IntervertebralDisc',
         'MastoidCellsAndAntra',
         'SuprasternalNotch',
         'FrontalSinus',
         'Skull',
         'Pancreas',
         'LevelOfT4T5IntervertebralDisc',
         'InternalAuditoryMeatus',
         'Liver',
         'LevelOfT1T2IntervertebralDisc',
         'CoronaryArteryGraft',
         'ExternalAuditoryMeatus',
         'VertexOfHead',
         'Scapula',
         'Patella',
         'Kidney',
         'AdrenalGland',
         'AcromioclavicularJoint',
         'Scaphoid']}

name_for_cid[4203] = 'OphthalmicPhotographyIllumination'
cid_concepts[4203] = \
{'DCM': ['ScotopicLight',
         'MesopicLight',
         'PhotopicLight',
         'DynamicLight',
         'DiffuseDirectIllumination'],
 'SCT': ['BroadTangentialDirectIllumination',
         'DualDiffuseDirectIllumination',
         'IndirectRetroilluminationFromTheIris',
         'IndirectRetroilluminationFromTheRetina',
         'FineSlitBeamDirectIllumination',
         'IndirectScleroticScatterIllumination',
         'IndirectIrisTransillumination']}

name_for_cid[633] = 'PhaseofImagingProcedureRequiringAnesthesia'
cid_concepts[633] = \
{'SCT': ['AfterProcedure', 'BeforeProcedure', 'DuringProcedure']}

name_for_cid[12212] = 'EchocardiographyAorta'
cid_concepts[12212] = \
{'LN': ['AorticRootDiameter',
        'HeartRate',
        'AscendingAorticDiameter',
        'AorticArchDiameter',
        'AorticIsthmusDiameter',
        'ThoracicAortaCoarctationSystolicPeakVelocity',
        'DescendingAorticDiameter',
        'ThoracicAortaCoarctationSystolicPeakInstantaneousGradient']}

name_for_cid[3529] = 'Arterial/VenousContent'
cid_concepts[3529] = \
{'DCM': ['PulmonaryArterialContentFcpa', 'PulmonaryVenousContentFcpv'],
 'LN': ['ArterialContentFca', 'VenousContentFcv']}

name_for_cid[3483] = 'IVUSLongitudinalMeasurements'
cid_concepts[3483] = \
{'DCM': ['StentGap', 'CalciumLength'],
 'SCT': ['StentLength', 'StenoticLesionLength']}

name_for_cid[6052] = 'BreastImagingReportSectionTitle'
cid_concepts[6052] = \
{'DCM': ['Findings',
         'ProcedureReported',
         'Addendum',
         'PhysicalExaminationResults',
         'ReasonForProcedure',
         'ComparisonToPreviousExams',
         'Conclusions',
         'OverallAssessment',
         'Recommendations'],
 'LN': ['Impressions'],
 'SCT': ['BreastComposition']}

name_for_cid[3250] = 'CatheterizationProcedurePhase'
cid_concepts[3250] = \
{'SCT': ['CardiacCatheterizationBaselinePhase',
         'CardiacCatheterizationImageAcquisitionPhase',
         'CatheterizationOfBothLeftAndRightHeartWithGraft',
         'TransseptalCatheterization',
         'DrugInfusionChallenge',
         'CardiacCatheterizationInterventionPhase',
         'CardiacCatheterizationTherapyPhase',
         'CatheterizationOfBothLeftAndRightHeartWithoutGraft',
         'CardiacCatheterizationPostContrastPhase',
         'RestingState',
         'CardiacCatheterizationPreInterventionPhase',
         'CardiacCatheterizationTestChallengePhase',
         'CatheterizationOfLeftHeart',
         'CardiacCatheterizationPostInterventionPhase',
         'CardiacCatheterizationBailoutPhase',
         'CatheterizationOfRightHeart',
         'ExerciseChallenge']}

name_for_cid[6402] = 'NonlesionObjectTypeSubstances'
cid_concepts[6402] = \
{'SCT': ['ContrastAgent']}

name_for_cid[3600] = 'RelativeTimes'
cid_concepts[3600] = \
{'SCT': ['During', 'After', 'Before']}

name_for_cid[6117] = 'VascularAnatomy'
cid_concepts[6117] = \
{'BARI': ['_2ndMarginalCoronaryArteryLaterals',
          '_3rdDiagonalCoronaryArteryLaterals',
          '_3rdMarginalCoronaryArteryLaterals',
          '_1stDiagonalCoronaryArteryLaterals',
          '_1stMarginalCoronaryArteryLaterals',
          '_2ndDiagonalCoronaryArteryLaterals',
          'RamusLaterals',
          'PosteriorDescendingSeptalPerforators'],
 'DCM': ['AorticKnob', 'ArchOfTheAzygosVein'],
 'SCT': ['SubclavianVein',
         'ProximalRightCoronaryArtery',
         'RightPosteriorAVCoronaryArtery',
         'AorticIsthmus',
         'BrachialPlexus',
         'MidCircumflexCoronaryArtery',
         'InferiorVenaCava',
         '_1stLeftPosterolateralCoronaryArtery',
         'IntermediateArteryRamus',
         'CircumflexCoronaryArtery',
         'SuperiorPhrenicArtery',
         'DistalLeftAnteriorDescendingCoronaryArtery',
         'RightCoronaryArteryOstium',
         'AxillaryArtery',
         'DescendingAorta',
         'DistalRightCoronaryArtery',
         'MarginalCoronaryArtery',
         'IntercostalArtery',
         'InternalThoracicArtery',
         'LeftAnteriorDescendingCoronaryArtery',
         'VertebralArtery',
         'ProximalLeftAnteriorDescendingCoronaryArtery',
         'CommonCarotidArtery',
         '_3rdDiagonalCoronaryArtery',
         'InternalJugularVein',
         '_3rdMarginalCoronaryArtery',
         'BronchialArtery',
         'MidRightCoronaryArtery',
         'EsophagealArtery',
         'AVGrooveContinuationOfCircumflexArtery',
         'PulmonaryVein',
         'SubclavianArtery',
         'LeftPosterolateralCircumflexCoronaryArtery',
         '_3rdRightPosterolateralCoronaryArtery',
         'ThyrocervicalTrunk',
         'InferiorPhrenicArtery',
         'BrachiocephalicVein',
         '_3rdLeftPosterolateralCoronaryArtery',
         'ProximalCircumflexCoronaryArtery',
         '_2ndMarginalCoronaryArtery',
         'AorticArch',
         'DorsalScapularArtery',
         'SuperiorVenaCava',
         'LeftMainCoronaryArtery',
         '_2ndLeftPosterolateralCoronaryArtery',
         'BrachiocephalicTrunk',
         '_2ndRightPosterolateralCoronaryArtery',
         'PosteriorDescendingRightCoronaryArtery',
         'LeftPosteriorDescendingCircumflexCoronaryArtery',
         '_1stMarginalCoronaryArtery',
         'PulmonaryTrunk',
         '_1stSeptalCoronaryArtery',
         'RightCoronaryArtery',
         'AscendingAorta',
         'AzygosVein',
         '_1stDiagonalCoronaryArtery',
         'PericardiophrenicArtery',
         'PulmonaryArtery',
         'BrachialArtery',
         '_2ndDiagonalCoronaryArtery',
         'CostocervicalTrunk',
         'DistalCircumflexCoronaryArtery',
         'MidLeftAnteriorDescendingCoronaryArtery',
         'LeftMainCoronaryArteryOstium',
         'AxillaryVein',
         'CoronaryArteryGraft',
         '_1stRightPosterolateralCoronaryArtery']}

name_for_cid[3745] = 'CalculationMethod'
cid_concepts[3745] = \
{'SCT': ['Estimated', 'Calculated']}

name_for_cid[8109] = 'SpecimenCollectionProcedure'
cid_concepts[8109] = \
{'SCT': ['Scraping',
         'Venipuncture',
         'TakingOfSwab',
         'SpecimenCollection',
         'Excision',
         'Aspiration',
         'RemovalOfDevice',
         'HarvestingOfTissue',
         'Biopsy']}

name_for_cid[60] = 'ImagingAgentAdministrationAdverseEvents'
cid_concepts[60] = \
{'DCM': ['ExtravasationVisibleInImage', 'PatientConditionPreventedContinuing'],
 'SCT': ['PallorPaleComplexion',
         'InjectionSiteNerveDamage',
         'InjectionSiteFibrosis',
         'InjectionSiteExtravasation',
         'Cough',
         'ChillsAndFever',
         'InjectionSiteHypersensitivity',
         'InjectionSitePigmentationChange',
         'InjectionSiteScar',
         'InjectionSiteIrritation',
         'Sweating',
         'InjectionSiteAnesthesia',
         'FacialSwelling',
         'InjectionSiteInduration',
         'InjectionSiteUrticaria',
         'Bronchospasm',
         'InjectionSiteMalabsorption',
         'InjectionSiteSterileAbscess',
         'Headache',
         'SensationOfBeingWarm',
         'TachycardiaBradycardia',
         'InjectionSiteNecrosis',
         'Anxiety',
         'InjectionSiteDermatitis',
         'InjectionSiteDisorder',
         'Shaking',
         'InjectionSiteBruising',
         'DrugInducedFlushing',
         'CardiacArrhythmia',
         'LaryngealEdema',
         'InjectionSiteThrombosis',
         'InjectionSiteAbscess',
         'DrugInducedDizziness',
         'InjectionSiteReaction',
         'CardiorespiratoryArrest',
         'InjectionSiteCyst',
         'Hypertension',
         'InjectionSiteEdema',
         'Dyspnea',
         'InjectionSiteAtrophy',
         'DrugInducedHypotension',
         'InjectionSiteBurning',
         'InjectionSiteHemorrhage',
         'InjectionSiteUlcer',
         'EpilepticConvulsions',
         'Itching',
         'DrugRash',
         'InjectionSiteInflammation',
         'InjectionSiteInfection',
         'Hypotension',
         'DiffuseInflammatoryErythema',
         'TasteAndSenseAltered',
         'InjectionSiteParesthesia',
         'NasalCongestion',
         'NoMotorResponseToCommand',
         'InjectionSitePain',
         'DrugInducedNauseaAndVomiting',
         'InjectionSiteMass',
         'InjectionSiteGranuloma']}

name_for_cid[10043] = 'IntravenousExtravasationSymptoms'
cid_concepts[10043] = \
{'DCM': ['ExtravasationVisibleInImage'],
 'SCT': ['InjectionSiteBruising',
         'InjectionSiteThrombosis',
         'InjectionSiteAbscess',
         'InjectionSiteNerveDamage',
         'InjectionSiteFibrosis',
         'InjectionSiteHypersensitivity',
         'InjectionSitePigmentationChange',
         'InjectionSiteScar',
         'InjectionSiteIrritation',
         'InjectionSiteReaction',
         'InjectionSiteAnesthesia',
         'InjectionSiteCyst',
         'InjectionSiteInduration',
         'InjectionSiteEdema',
         'InjectionSiteUrticaria',
         'InjectionSiteAtrophy',
         'InjectionSiteBurning',
         'InjectionSiteHemorrhage',
         'InjectionSiteUlcer',
         'InjectionSiteMalabsorption',
         'InjectionSiteSterileAbscess',
         'InjectionSiteInflammation',
         'InjectionSiteInfection',
         'InjectionSiteNecrosis',
         'InjectionSiteParesthesia',
         'InjectionSiteDermatitis',
         'InjectionSiteDisorder',
         'InjectionSitePain',
         'InjectionSiteMass',
         'InjectionSiteGranuloma']}

name_for_cid[7706] = 'CranialNerves'
cid_concepts[7706] = \
{'SCT': ['AbducensNerve',
         'AccessoryNerve',
         'TrochlearNerve',
         'FacialNerve',
         'GlossopharyngealNerve',
         'TrigeminalNerve',
         'VestibulocochlearNerve',
         'VagusNerve',
         'HypoglossalNerve',
         'OculomotorNerve',
         'OlfactoryTract',
         'OpticNerve']}

name_for_cid[12300] = 'CoreEchoMeasurements'
cid_concepts[12300] = \
{'LN': ['TricuspidValveDiastolicVTI',
        'MitralValvePeakInstantaneousGradient',
        'LeftAtrialEndSystolicVolumeBiplaneAreaLength',
        'LeftVentricularInternalSystolicDimensionBSA',
        'LeftVentricularPreEjectionPeriod',
        'LVBasalAnteriorTimeToSVmaxTsBasalAnterior',
        'MitralRegurgitationPISARadius',
        'PulmonicValveVmax',
        'PulmonaryVeinSWaveVmax',
        'LeftAtrialEndSystolicDiameterAPMMBSA',
        'RightVentricularMidCavityDimension4C',
        'LeftVentricularEndSystolicVolumeBiplaneMOD',
        'MitralRegurgitationVmax',
        'LeftVentricularEjectionFractionSinglePlane4CMOD',
        'InterventricularSeptumSystolicDimensionMM',
        'LeftVentricularMassDimensionMethodMMHeight2Point7',
        'LVMidSeptalTimeToSVmaxTsMidSeptal',
        'AorticSinotubularJunctionDimension',
        'LeftVentricularInternalDiastolicDimensionBSA',
        'TricuspidRegurgitationPISARadius',
        'InferiorVenaCavaDiameter',
        'AorticRegurgitantJetWidthLVOTWidthPercent',
        'MitralValveMeanGradient',
        'RightVentricularEjectionTime',
        'MitralAnnulusDiastolicDiameterA2C',
        'LeftAtrialEndSystolicVolumeBiplaneAreaLengthBSA',
        'MitralRegurgitationPeakGradient',
        'LeftVentricularInternalSystolicDimensionMM',
        'LeftVentricularStrokeVolume3D',
        'TricuspidRegurgitationVmax',
        'AorticValveAreaContinuityVmax',
        'TricuspidRegurgitationVenaContractaWidth',
        'PulmonicValveVTI',
        'PulmonicRegurgitationEndDiastolicVelocity',
        'MitralRegurgitationVenaContractaWidth',
        'LeftVentricularEndSystolicVolume3D',
        'LeftVentricularMassDimensionMethodMMBSA',
        'MitralValveDecelerationTime',
        'RightVentricularFreeWallThicknessMM',
        'LeftAtrialEndSystolicLength2C',
        'LVMidPosteriorTimeToSVmaxTsMidPosterior',
        'InterventricularSeptumSystolicDimension2D',
        'LeftVentricularFractionalShorteningOfMinorAxisMM',
        'LeftVentricularInternalDiastolicDimensionMM',
        'MitralValveAreaPressureHalfTime',
        'TricuspidValvePressureHalfTime',
        'LeftVentricularEndDiastolicLength4C',
        'InterventricularSeptumDiastolicDimensionMM',
        'MitralValveFlowPropagationVelocityVp',
        'LeftVentricularInternalSystolicDimension2D',
        'AorticValveEffectiveRegurgitantOrificeArea',
        'MitralRegurgitantFlowPISA',
        'LeftVentricularPosteriorWallSystolicThickness',
        'LeftAtrialEndSystolicVolumeBiplaneMOD',
        'LeftVentricularEjectionFractionTeichholzMM',
        'TricuspidValveAPrimeVmax',
        'AorticRegurgitationPISARadius',
        'LeftVentricularMassDimensionMethod2D',
        'AorticValveAreaContinuityVmaxBSA',
        'MitralSeptalEPrimeVmax',
        'PulmonicRegurgitationEndDiastolicPeakGradient',
        'MitralValveAreaPlanimetry',
        'MitralRegurgitationPISAVelocity',
        'PulmonicRegurgitationVmax',
        'LVMidLateralTimeToSVmaxTsMidLateral',
        'LeftVentricularMassDimensionMethodMM',
        'RightVentricularFreeWallThickness2D',
        'LeftVentricularEndDiastolicVolume3D',
        'InterventricularSeptumTimeToPeakDisplacement',
        'LeftVentricularFractionalShorteningOfMinorAxis2D',
        'LeftVentricularInternalDiastolicDimension2D',
        'TricuspidValveAWaveVmax',
        'InterventricularSeptumDiastolicDimension2D',
        'AorticValveMeanBloodVelocity',
        'MitralValveEffectiveRegurgitantOrificeAreaPISA',
        'LeftVentricularInternalDiastolicDimensionBSA',
        'LeftVentricularPosteriorWallSystolicThickness',
        'LeftAtrialEndSystolicVolumeBiplaneMODBSA',
        'MitralRegurgitantFractionPISA',
        'RightPulmonaryArteryDiameter',
        'AorticRegurgitationPISAVelocity',
        'LeftVentricularMassAreaLengthHeight2Point7',
        'AorticValveAreaContinuityVTI',
        'PulmonicValveEjectionTime',
        'PulmonicValvePeakGradient',
        'LVMidInferiorTimeToSVmaxTsMidInferior',
        'RightVentricularEndSystolicArea4C',
        'LeftVentricularEpicardialAreaSAXPMLevel',
        'MitralValveAreaPISA',
        'LeftAtrialEndSystolicArea2C',
        'LeftVentricularMassDimensionMethod2DHeight2Point7',
        'LeftVentricularEndDiastolicVolumeBiplaneMOD',
        'LeftVentricularOutflowTractVmax',
        'MitralValvePressureHalfTime',
        'AorticRegurgitationVolumeContinuityVTI',
        'MitralValveEToARatio',
        'TricuspidValveEPrimeVmax',
        'AorticRegurgitationPressureHalfTime',
        'LeftAtrialEndSystolicVolumeSinglePlane2CMOD',
        'PulmonicAnnulusDiameter',
        'MitralValveEWaveVmax',
        'LeftVentricularPosteriorWallDiastolicThickness',
        'RightAtrialMinorAxisDimension4CBSA',
        'MitralRegurgitantFractionContinuityVTI',
        'AorticValveMeanGradient',
        'LeftVentricularMassAreaLengthBSA',
        'AorticValveAreaContinuityVTIBSA',
        'LeftVentricularEjectionFractionTeichholz2D',
        'RightVentricularFractionalAreaChange',
        'LeftVentricularMassDimensionMethod2DBSA',
        'LVMidAnteroseptalTimeToSVmaxTsMidAnteroseptal',
        'RightVentricularEndDiastolicArea4C',
        'RightVentricularOutflowTractVTI',
        'MitralValveAWaveVmax',
        'LeftAtrialEndSystolicArea4C',
        'LeftVentricularEndocardialAreaSAXPMLevel',
        'LeftVentricularEndDiastolicVolumeBiplaneMODBSA',
        'LeftVentricularOutflowTractDimension2D',
        'MitralValveVmax',
        'AorticRegurgitationVolumePISA',
        'AorticRegurgitationVenaContractaWidth',
        'TricuspidValveEWaveVmax',
        'LeftAtrialEndSystolicVolumeSinglePlane4CMOD',
        'RightAtrialMinorAxisDimension4C',
        'MitralLateralEPrimeVmax',
        'LeftVentricularPosteriorWallDiastolicThickness',
        'LVBasalLateralTimeToSVmaxTsBasalLateral',
        'PulmonaryVeinAWaveDuration',
        'AorticValvePeakInstantaneousGradient',
        'LeftVentricularMassAreaLength',
        'AorticAnnulusDiameter',
        'MitralValveVTI',
        'AorticRegurgitationVTI',
        'LeftAtrialEndSystolicDiameterAP2D',
        'RightVentricularBasalDimension4C',
        'TricuspidRegurgitationPeakGradient',
        'MitralValveAWaveDuration',
        'RightVentricularOutflowTractDiameterAtSubvalvularLevelRVOTProximal',
        'LVMidAnteriorTimeToSVmaxTsMidAnterior',
        'LeftVentricularEjectionFraction3D',
        'LeftVentricularEndDiastolicVolumeSinglePlane2CMOD',
        'LeftVentricularEndSystolicVolumeSinglePlane4CMOD',
        'LeftVentricularMassTruncatedEllipseHeight2Point7',
        'AorticRegurgitationVmax',
        'TricuspidAnnulusDiameter',
        'TricuspidValveSPrimeVmax',
        'LeftVentricularOutflowTractVTI',
        'RightAtrialMajorAxisDimension4C',
        'MitralAnnulusVTI',
        'MainPulmonaryArteryVmax',
        'LeftVentricularIsovolumicRelaxationTimeByTDI',
        'LeftPulmonaryArteryDiameter',
        'LVBasalInferiorTimeToSVmaxTsBasalInferior',
        'AorticValveVmax',
        'AorticRegurgitantFlow',
        'PulmonaryVeinAWaveVmax',
        'AorticRootDiameter',
        'RightVentricularOutflowTractDiameterAtPulmonicValveRVOTDistal',
        'LeftAtrialEndSystolicDiameterAP2DBSA',
        'LeftVentricularEndDiastolicVolumeSinglePlane4CMOD',
        'TricuspidValveMeanGradient',
        'MainPulmonaryArteryDiameter',
        'LeftVentricularEjectionFractionBiplaneMOD',
        'LVBasalSeptalTimeToSVmaxTsBasalSeptal',
        'LeftVentricularEndSystolicVolumeSinglePlane2CMOD',
        'MitralRegurgitationVolumePISA',
        'LeftVentricularMassTruncatedEllipseBSA',
        'TricuspidValveClosureToOpeningTime',
        'TricuspidValveVmax',
        'AorticRegurgitantFraction',
        'LeftVentricularIsovolumicRelaxationTimeByDoppler',
        'MitralAnnulusDiastolicDiameterPLAX',
        'RightVentricularPreEjectionPeriod',
        'RightAtrialEndSystolicArea4C',
        'LVBasalAnteroseptalTimeToSVmaxTSBasalAnteroseptal',
        'LeftAtrialEndSystolicLength4C',
        'LeftVentricularInternalSystolicDimensionBSA',
        'LeftVenticularPosteriorWallTimeToPeakDisplacement',
        'AorticValveVTI',
        'TricuspidAnnularPlaneSystolicExcursionTAPSE',
        'PulmonaryVeinDWaveVmax',
        'TricuspidValvePeakGradient',
        'MitralRegurgitationVolumeContinuityVTI',
        'LeftAtrialEndSystolicDiameterAPMM',
        'RightVentricularMyocardialPerformanceIndex',
        'LVBasalPosteriorTimeToSVmaxTsBasalPosterior',
        'LeftVentricularEjectionFractionSinglePlane2CMOD',
        'LeftVentricularMass',
        'LVTsSDDyssynchronyIndex',
        'AorticRootDiameterBSA',
        'LeftVentricularEndSystolicVolumeBiplaneMODBSA',
        'TricuspidValveDecelerationTime',
        'AscendingAortaDimension',
        'AorticRegurgitantJetAreaLVOTAreaPercent',
        'MitralAnnulusDiastolicDiameterA4C',
        'PulmonicValveAccelerationTime']}

name_for_cid[7478] = 'IntensityHistogramFeatures'
cid_concepts[7478] = \
{}

name_for_cid[7182] = 'AbstractMultidimensionalImageModelDimensionSemantics'
cid_concepts[7182] = \
{'DCM': ['LinearDisplacement', 'Angle', 'Time', 'PhotonEnergy']}

name_for_cid[4254] = 'VisualFieldStaticPerimetryTestAnalysisResults'
cid_concepts[4254] = \
{'DCM': ['OutsideNormalLimits',
         'Borderline',
         'AbnormallyHighSensitivity',
         'GeneralReductionInSensitivity',
         'BorderlineAndGeneralReductionInSensitivity'],
 'SCT': ['WithinNormalLimits']}

name_for_cid[9523] = 'RoboticRadiotherapyProcedureTechniques'
cid_concepts[9523] = \
{'DCM': ['SynchronizedRoboticTreatment', 'NonSynchronizedRoboticTreatment']}

name_for_cid[12245] = 'CardiacUltrasoundReportTitles'
cid_concepts[12245] = \
{'DCM': ['PediatricCardiacUltrasoundReport',
         'AdultCongenitalCardiacUltrasoundReport',
         'FetalCardiacUltrasoundReport']}

name_for_cid[3207] = 'StressTestProcedurePhases'
cid_concepts[3207] = \
{'SCT': ['CardiacStressState',
         'Hyperventilation',
         'PeakCardiacStressState',
         'CardiacStressRecoveryState',
         'RestingState']}

name_for_cid[6005] = 'CharacteristicsofShapefromBIRADS®'
cid_concepts[6005] = \
{'SCT': ['Lobular', 'RoundShape', 'Irregular', 'OvoidShapeOval']}

name_for_cid[7301] = 'InterventionTypes'
cid_concepts[7301] = \
{'SCT': ['HipJointReconstruction',
         'ResurfacingOfTheFemoralHead',
         'InsertionOfHipProsthesis',
         'ResurfacingOfThePatella']}

name_for_cid[10014] = 'ContrastImagingTechnique'
cid_concepts[10014] = \
{'SCT': ['DiagnosticRadiographyWithContrastMedia', 'CTWithoutContrast']}

name_for_cid[3712] = 'VesselDescriptors'
cid_concepts[3712] = \
{'SCT': ['DiffuseDisease',
         'Calcified',
         'Stented',
         'Restenotic',
         'Tortuous',
         'Stenotic',
         'Culprit',
         'Bifurcation',
         'LuminalIrregularities',
         'Aneurysmal',
         'Thrombus',
         'MuscleBridge',
         'Ulcerated',
         'Ectatic']}

name_for_cid[6140] = 'CalculationMethods'
cid_concepts[6140] = \
{'DCM': ['UnspecifiedMethodOfCalculation',
         'AgatstonScoringMethod',
         'VolumeScoringMethod',
         'TwoDimensionalMethod',
         'MassScoringMethod',
         'ThreeDimensionalMethod'],
 'SCT': ['Estimated']}

name_for_cid[6212] = 'CalculatedValueforColonFindings'
cid_concepts[6212] = \
{'DCM': ['PolypStalkWidth', 'DistanceFromAnus'],
 'SCT': ['PolypSizeLargestDimension', 'PolypStalkLength']}

name_for_cid[3010] = 'CardiovascularAnatomicLocations'
cid_concepts[3010] = \
{'SCT': ['InternalCarotidArtery',
         'SubclavianVein',
         'ExternalIliacArtery',
         'TruncusCoeliacus',
         'BasilarArtery',
         'ApexOfRightVentricle',
         'SuperiorRightPulmonaryVein',
         'InferiorVenaCava',
         'RightPulmonaryArtery',
         'GenicularArtery',
         'LeftVentricleInflow',
         'VenousNetwork',
         'CarotidArtery',
         'CommonAtrium',
         'UmbilicalArtery',
         'PrimitiveAorta',
         'AorticFistula',
         'LeftVentricleOutflowTract',
         'RadialArtery',
         'RightFemoralArtery',
         'SuperiorThyroidArtery',
         'IliacArtery',
         'LeftFemoralArtery',
         'AnteriorTibialArtery',
         'TruncusArteriosusCommunis',
         'Artery',
         'AntecubitalVein',
         'CommonCarotidArtery',
         'Aorta',
         'PatentDuctusArteriosus',
         'SuperficialTemporalArtery',
         'MesentericVein',
         'Vein',
         'CongenitalCoronaryArteryFistulaToLeftVentricle',
         'PulmonaryVein',
         'InferiorRightPulmonaryVein',
         'CoronaryArtery',
         'ExternalIliacVein',
         'CongenitalPulmonaryArteriovenousFistula',
         'DescendingAorta',
         'SubclavianArtery',
         'BodyConduit',
         'OphthalmicArtery',
         'RightVentricleInflow',
         'MesentericArtery',
         'PoplitealArtery',
         'PulmonaryVenousAtrium',
         'BrachialVein',
         'InternalIliacArtery',
         'PulmonaryChamberOfCorTriatriatum',
         'SuperiorVenaCava',
         'UlnarArtery',
         'LeftVentricle',
         'AnteriorSpinalArtery',
         'HepaticArtery',
         'SplenicVein',
         'PeronealArtery',
         'RightVentricleOutflowTract',
         'SaphenousVein',
         'InnominateArtery',
         'SplenicArtery',
         'RenalArtery',
         'SuperiorLeftPulmonaryVein',
         'SystemicCollateralArteryToLung',
         'FacialArtery',
         'LumbarArtery',
         'ApexOfLeftVentricle',
         'AzygosVein',
         'FemoralArtery',
         'PulmonaryArtery',
         'BrachialArtery',
         'CongenitalCoronaryArteryFistulaToRightVentricle',
         'PosteriorMedialTributary',
         'CommonFemoralArtery',
         'PulmonaryArteryConduit',
         'RightAtrium',
         'AnteriorCommunicatingArtery',
         'CongenitalCoronaryArteryFistulaToLeftAtrium',
         'RightVentricle',
         'ThoracicAorta',
         'CommonVentricle',
         'CoronarySinus',
         'AxillaryArtery',
         'LeftPulmonaryArtery',
         'HepaticVein',
         'InternalMammaryArtery',
         'VertebralArtery',
         'PulmonaryVeinConfluence',
         'SaphenofemoralJunction',
         'InternalJugularVein',
         'LacrimalArteryOfRightEye',
         'ExternalCarotidArtery',
         'SuperiorMesentericArtery',
         'UmbilicalVein',
         'CommonIliacVein',
         'PosteriorTibialArtery',
         'FemoralVein',
         'RightAuricularAppendage',
         'GreatCardiacVein',
         'InnominateVein',
         'PrimitivePulmonaryArtery',
         'SuperficialFemoralArtery',
         'PortalVein',
         'InferiorLeftPulmonaryVein',
         'PosteriorCommunicationArtery',
         'JuxtaposedAtrialAppendage',
         'AorticArch',
         'RenalVein',
         'BoydPerforatingVein',
         'AbdominalAorta',
         'CommonIliacArtery',
         'CephalicVein',
         'LingualArtery',
         'GastricVein',
         'CongenitalCoronaryArteryFistulaToRightAtrium',
         'InferiorCardiacVein',
         'Baffle',
         'LeftAuricularAppendage',
         'LacrimalArtery',
         'OccipitalArtery',
         'SystemicVenousAtrium',
         'AscendingAorta',
         'HunterianPerforatingVein',
         'AnomalousPulmonaryVein',
         'AnteriorCardiacVein',
         'ProfundaFemorisArtery',
         'LeftAtrium',
         'CerebralArtery',
         'AxillaryVein',
         'InferiorMesentericArtery',
         'GreatSaphenousVein',
         'OccipitalVein',
         'DoddPerforatingVein']}

name_for_cid[3440] = 'PeripheralPulseLocations'
cid_concepts[3440] = \
{'SCT': ['FemoralArtery',
         'BrachialArtery',
         'UlnarArtery',
         'CarotidArtery',
         'DorsalisPedisArtery',
         'PoplitealArtery',
         'PosteriorTibialArtery',
         'RadialArtery']}

name_for_cid[8125] = 'MicroscopyIlluminatorType'
cid_concepts[8125] = \
{'SCT': ['TungstenHalogenLamp',
         'LightEmittingDiode',
         'Laser',
         'MercuryArcLamp',
         'XenonArcLamp']}

name_for_cid[3769] = 'ConcernTypes'
cid_concepts[3769] = \
{'SCT': ['Problem',
         'Complaint',
         'FindingReportedByPatientInformant',
         'Disease',
         'FunctionalPerformanceAndActivity',
         'Finding']}

name_for_cid[6091] = 'RelativeFrequencyofEventValues'
cid_concepts[6091] = \
{'SCT': ['Continuous',
         'SingleEvent',
         'AsRequired',
         'Frequent',
         'MidFrequency',
         'Infrequent']}

name_for_cid[4106] = 'QuantitativeMethodsusedforPerfusionAndTracerKineticModels'
cid_concepts[4106] = \
{'DCM': ['T1ByInversionRecovery',
         'TemporalDerivativeExceedsThreshold',
         'T1ByFixedValue',
         'LeastMeanSquareLMSDeconvolution',
         'SingularValueDecompositionSVDDeconvolution',
         'TimeOfPeakConcentration',
         'T1ByMultipleFlipAngles',
         'TimeOfLeadingHalfPeakConcentration',
         'StandardToftsModel',
         'UserDefinedAIFROI',
         'ExtendedToftsModel',
         'AutomaticallyDetectedAIFROI',
         'ModelFreeConcentrationTimeQuantitification',
         'BlindEstimationOfAIF',
         'FirstPassLeakageProfileFPLPModel',
         'ShutterSpeedModelSSM',
         'PerfusionAnalysisByStableXenonCTTechnique',
         'GammaCapillaryTransitTimeGCTTModel',
         'PerfusionAnalysisByIVIodinatedContrastCTTechnique',
         'AdiabaticTissueHomogeneityATHModel',
         'PerfusionAnalysisByArterialSpinLabelingMRTechnique',
         'TwoCompartmentExchange2CXModel',
         'PerfusionAnalysisBySusceptibilityMRTechnique',
         'AIFIgnored',
         'PopulationAveragedAIF']}

name_for_cid[3339] = 'ElectrophysiologyAnnotations'
cid_concepts[3339] = \
{'DCM': ['StimulationAtRate4Interval',
         'AblationOff',
         'StimulationAtRate3Interval',
         'StimulationAtRate2Interval',
         'AblationOn',
         'StartOfAtrialContraction',
         'SWave',
         'StimulationAtRate1Interval',
         'StartOfAtrialContraction',
         'PWave',
         'VWave',
         'HISBundleWave',
         'TWave',
         'RWave',
         'QWave',
         'VWaveOfNextBeat']}

name_for_cid[12117] = 'VesselBranchModifiers'
cid_concepts[12117] = \
{'NCIt': ['Middle'],
 'SCT': ['Main',
         'Right',
         'Lateral',
         'Inferior',
         'Left',
         'Superior',
         'CollateralBranchOfVessel',
         'Medial']}

name_for_cid[12269] = 'CardiacUltrasoundInterventricularSeptumMeasurements'
cid_concepts[12269] = \
{'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'JetArea',
        'StenosisPeakVelocity',
        'VolumeFlow',
        'FlowRadius',
        'HeartRate',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FlowArea',
        'EndDiastolicVelocity',
        'ROIInternalDimensionByUS',
        'VelocityOfFlowPropagation',
        'PeakGradient',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Area',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter',
         'Length',
         'Circumference']}

name_for_cid[3552] = 'PacingActions'
cid_concepts[3552] = \
{'DCM': ['EndPacing', 'BeginPacing']}

name_for_cid[6300] = 'ProstateSectorAnatomy'
cid_concepts[6300] = \
{'SCT': ['LeftBasalPeripheralZoneOfProstate',
         'MaleExternalUrethralSphincter',
         'RightApicalAnteriorFibromuscularStromaOfProstate',
         'LeftPosteromedialApicalPeripheralZoneOfProstate',
         'LeftPosterolateralMiddlePeripheralZoneOfProstate',
         'RightBasalTransitionZoneOfProstate',
         'LeftAnteriorApicalTransitionZoneOfProstate',
         'LeftMiddleTransitionZoneOfProstate',
         'LeftAnteriorBasalTransitionZoneOfProstate',
         'RightPosterolateralBasalPeripheralZoneOfProstate',
         'LeftApicalPeripheralZoneOfProstate',
         'LeftPosteriorMiddleTransitionZoneOfProstate',
         'LeftBasalAnteriorFibromuscularStromaOfProstate',
         'LeftPosterolateralApicalPeripheralZoneOfProstate',
         'RightSeminalVesicle',
         'RightPosteromedialMiddlePeripheralZoneOfProstate',
         'LeftApicalAnteriorFibromuscularStromaOfProstate',
         'RightBasalPeripheralZoneOfProstate',
         'RightPosteromedialApicalPeripheralZoneOfProstate',
         'BasalAnteriorFibromuscularStromaOfProstate',
         'LeftBasalPartTransitionZoneOfProstate',
         'RightApicalPeripheralZoneOfProstate',
         'LeftPosterolateralBasalPeripheralZoneOfProstate',
         'RightPosterolateralApicalPeripheralZoneOfProstate',
         'RightPosteriorBasalTransitionZoneOfProstate',
         'RightAnteriorApicalTransitionZoneOfProstate',
         'LeftAnteriorMiddlePeripheralZoneOfProstate',
         'RightAnteriorMiddleTransitionZoneOfProstate',
         'RightLateralMiddlePeripheralZoneOfProstate',
         'CentralZoneOfLeftHalfProstate',
         'LeftLateralMiddlePeripheralZoneOfProstate',
         'LeftPosteromedialMiddlePeripheralZoneOfProstate',
         'RightApicalTransitionZoneOfProstate',
         'RightBasalAnteriorFibromuscularStromaOfProstate',
         'MiddleAnteriorFibromuscularStromaOfProstate',
         'RightAnteriorBasalTransitionZoneOfProstate',
         'LeftPosteriorApicalTransitionZoneOfProstate',
         'LeftAnteriorMiddleTransitionZoneOfProstate',
         'LeftPosteriorBasalTransitionZoneOfProstate',
         'RightAnteriorBasalPeripheralZoneOfProstate',
         'LeftMiddleAnteriorFibromuscularStromaOfProstate',
         'LeftAnteriorApicalPeripheralZoneOfProstate',
         'CentralZoneOfRightHalfProstate',
         'RightAnteriorMiddlePeripheralZoneOfProstate',
         'RightPosterolateralMiddlePeripheralZoneOfProstate',
         'LeftApicalTransitionZoneOfProstate',
         'ApicalAnteriorFibromuscularStromaOfProstate',
         'LeftSeminalVesicle',
         'RightAnteriorApicalPeripheralZoneOfProstate',
         'LeftAnteriorBasalPeripheralZoneOfProstate',
         'RightMiddleAnteriorFibromuscularStromaOfProstate',
         'RightPosteriorMiddleTransitionZoneOfProstate',
         'RightPosteriorApicalTransitionZoneOfProstate',
         'RightMiddleTransitionZoneOfProstate']}

name_for_cid[3102] = 'RestStress'
cid_concepts[3102] = \
{'DCM': ['CardiacStressState'], 'SCT': ['RestingState']}

name_for_cid[12286] = 'CardiacUltrasoundInterventricularSeptumFindingSites'
cid_concepts[12286] = \
{'SCT': ['BulboventricularForamen',
         'VentricularSeptalDefect',
         'InterventricularSeptum']}

name_for_cid[218] = 'QuantitativeImageFeatures'
cid_concepts[218] = \
{'DCM': ['WaterFatOutOfPhase',
         'FractionalOccupancySegmentation',
         'WaterFatInPhase',
         'FractionalProbabilisticSegmentation',
         'BinarySegmentation',
         'WaterFraction',
         'MRSignalIntensity',
         'XRayAttenuation',
         'FatFraction',
         'YBRICTCRComponent',
         'YBRICTCBComponent',
         'YBRICTYComponent',
         'YBRPARTIALCRComponent',
         'YBRPARTIALCBComponent',
         'YBRPARTIALYComponent',
         'RootAngularSecondMomentOfGLCM',
         'Echogenicity',
         'YBRRCTCRComponent',
         'YBRRCTCBComponent',
         'YBRRCTYComponent',
         'RGBRComponent',
         'UltrasoundAttenuation',
         'SpeedOfSound',
         'Perfusion',
         'Elasticity',
         'YBRFULLCRComponent',
         'FractalDimension',
         'YBRFULLCBComponent',
         'YBRFULLYComponent',
         'RGBBComponent',
         'RGBGComponent',
         'CreatineCholineCitrateRatio',
         'MagnetizationTransferRatio',
         'CreatineAndCholine',
         'LipidAndLactate',
         'SpatialDisplacementYComponent',
         'SpatialDisplacementXComponent',
         'NuclearMedicineTomographicActivity',
         'NuclearMedicineProjectionActivity',
         'FlowVariance',
         'FlowVelocity',
         'TissueVelocity',
         'IndexedHemodynamicResistance',
         'HemodynamicResistance',
         'SpatialDisplacementZComponent',
         'AttenuationCoefficient',
         'Ktrans',
         'Kep',
         'Ve',
         'Suvbsa',
         'Suvibw',
         'SuvlbmJanma',
         'SuvlbmJames128',
         'StandardizedUptakeValue',
         'Suvbw',
         'Suvlbm',
         'ApparentDiffusionCoefficient',
         'DiffusionWeighted',
         'VolumeRatio',
         'VolumeOfSphere',
         'DiffusionCoefficient',
         'VolumeOfCircumscribedSphere',
         'VolumeOfEllipsoid',
         'VolumeEstimatedFromThreeOrMoreNonCoplanar2DRegions',
         'VolumeEstimatedFromTwoNonCoplanar2DRegions',
         'VolumeEstimatedFromSingle2DRegion',
         'PathLength',
         'VolumeOfBoundingThreeDimensionalRegion',
         'R2Coefficient',
         'ChiSquare',
         'DW',
         'AIC',
         'BIC',
         'Distance',
         'Height',
         'RelativeLinearStoppingPower',
         '',
         'Vp',
         'EquivalentDose',
         'AbsorbedDose',
         'IAUC90BN',
         'IAUC180BN',
         'IAUC',
         'IAUC60',
         'IAUC90',
         'IAUC180',
         'IAUCBN',
         'IAUC60BN',
         'GlutamateAndGlutamine',
         'CholineCreatineRatio',
         'NAcetylaspartateCreatineRatio',
         'NAcetylaspartateCholineRatio',
         'Tmax',
         'R2',
         'VolumetricDiffusionDxzComponent',
         'MagneticSusceptibility',
         'VolumetricDiffusionDxyComponent',
         'VolumetricDiffusionDxxComponent',
         'RelativeRegionalBloodFlow',
         'RelativeRegionalBloodVolume',
         'BloodOxygenationLevel',
         'T2WeightedDynamicContrastEnhancedMRSignalIntensity',
         'T2WeightedDynamicContrastEnhancedMRSignalIntensity',
         'AbsoluteRegionalBloodFlow',
         'T1WeightedDynamicContrastEnhancedMRSignalIntensity',
         'AbsoluteRegionalBloodVolume',
         'VolumetricDiffusionDzzComponent',
         'OxygenExtractionFraction',
         'R1',
         'VolumetricDiffusionDyzComponent',
         'R2',
         'VolumetricDiffusionDyyComponent',
         'VelocityEncoded',
         'ZScore',
         'ContrastAgentAngioMRSignalIntensity',
         'SpinTaggingPerfusionMRSignalIntensity',
         'RelativeAnisotropy',
         'FractionalAnisotropy',
         'FieldMapMRSignalIntensity',
         'T2WeightedMRSignalIntensity',
         'T2WeightedMRSignalIntensity',
         'T1WeightedMRSignalIntensity',
         'ProtonDensityWeightedMRSignalIntensity',
         'TimeOfFlightAngioMRSignalIntensity',
         'T1',
         'T2',
         'T2',
         'TimeCourseOfSignal',
         'TemperatureEncoded',
         'StudentTTest',
         'TimeToPeak',
         'SignalToNoise',
         'TemporalDerivativeThreshold',
         'MaximumSlope',
         'MaximumDifference',
         'TracerConcentration',
         'TimeOfPeakConcentration',
         'BolusArrivalTime',
         'TimeOfLeadingHalfPeakConcentration',
         'MeanTransitTime',
         'KurtosisDiffusionCoefficient',
         'GammaDistributionScaleParameter',
         'NegativeEnhancementIntegral',
         'GammaDistributionShapeParameter',
         'GammaDistributionMode',
         'DistributedDiffusionCoefficient',
         'RCoefficient',
         'AnomalousExponentParameter',
         'ProtonDensity',
         'SignalChange',
         'MonoExponentialApparentDiffusionCoefficient',
         'SlowDiffusionCoefficient',
         'FastDiffusionCoefficient',
         'FastDiffusionCoefficientFraction'],
 'SCT': ['Lipid',
         'DiameterOfCircumscribedCircle',
         'Depth',
         'PowerDoppler',
         'Perimeter',
         'Lactate',
         'Length',
         'Inositol',
         'PerpendicularAxis',
         'Radius',
         'AreaOfDefinedRegion',
         'Width',
         'Volume',
         'Fat',
         'Area',
         'NAcetylaspartate',
         'Creatine',
         'Glutamine',
         'MinorAxis',
         'Diameter',
         'Tuarine',
         'Citrate',
         'MajorAxis',
         'LongAxis',
         'Circumference',
         'Water',
         'Choline',
         'ShortAxis']}

name_for_cid[7004] = 'WaveformPurposesofReference'
cid_concepts[7004] = \
{'DCM': ['SimultaneousRespiratoryWaveform',
         'SimultaneousArterialPulseWaveform',
         'SimultaneousECG',
         'SimultaneousVoiceNarrative',
         'SimultaneousDoppler',
         'SimultaneousHemodynamic',
         'SimultaneousPhonocardiographicWaveform']}

name_for_cid[7454] = 'AnimalTaxonomicRankValues'
cid_concepts[7454] = \
{'ITIS_TSN': ['PeromyscusLeucopusAmericanWhiteFootedMouse',
              'PeromyscusManiculatusDeerMouse',
              'SigmodonGenus'],
 'SCT': ['CanisLupus',
         'RattusNorvegicus',
         'FelisCatus',
         'HomoSapiens',
         'Canis',
         'BosTaurus',
         'OryctolagusCuniculusEuropeanRabbit',
         'Capra',
         'MustelaPutoriusFuro',
         'EquusCaballus',
         'SusScrofa',
         'Equus',
         'Ovis',
         'Bovinae',
         'Bos',
         'CallithrixJacchus',
         'CaviaPorcellus',
         'Felis',
         'Sus',
         'CanisLupusFamiliaris',
         'Rattus',
         'MusGenus',
         'OvisAries',
         'MusMusculusHouseMouse',
         'CapraHircus']}

name_for_cid[12005] = 'FetalBiometryMeasurements'
cid_concepts[12005] = \
{'LN': ['OccipitalFrontalDiameter',
        'BPDAreaCorrected',
        'TranverseAbdominalDiameter',
        'APADTAD',
        'CisternaMagna',
        'AbdominalCircumference',
        'FemurLength',
        'ThoracicCircumference',
        'HeadCircumference',
        'FootLength',
        'BiparietalDiameter',
        'LeftKidneyThickness',
        'TransverseThoracicDiameter',
        'LeftKidneyLength',
        'LeftKidneyWidth',
        'AnteriorPosteriorAbdominalDiameter',
        'ThoracicArea',
        'RightKidneyThickness',
        'AnteriorPosteriorTrunkDiameter',
        'RightKidneyWidth',
        'TransCerebellarDiameter',
        'RightKidneyLength']}

name_for_cid[12140] = 'PelvicVasculatureAnatomicalLocation'
cid_concepts[12140] = \
{'SCT': ['UmbilicalVein',
         'CommonIliacArtery',
         'VitellineVeinOfPlacenta',
         'OvarianArtery',
         'OvarianVein',
         'UterineVein',
         'UmbilicalArtery',
         'UterineArtery',
         'VitellineArteryOfPlacenta']}

name_for_cid[6029] = 'RecommendedFollowupfromBIRADS®'
cid_concepts[6029] = \
{'DCM': ['NormalIntervalFollowUp',
         'AnyDecisionToBiopsyShouldBeBasedOnClinicalAssessment',
         'FollowUpAtShortInterval111Months',
         'BiopsyShouldBeConsidered',
         'NeedleLocalizationAndBiopsy',
         'KnownBiopsyProvenMalignancyTakeAppropriateAction',
         'HistologyUsingCoreBiopsy',
         'AdditionalProjections',
         'SuggestiveOfMalignancyTakeAppropriateAction',
         'SpotMagnificationViews',
         'CytologicAnalysis',
         'BiopsyShouldBeStronglyConsidered',
         'OldFilmsForComparison',
         'HighlySuggestiveOfMalignancyTakeAppropriateAction'],
 'SCT': ['SpotCompression',
         'MRIOfBreast',
         'DiagnosticUltrasonography',
         'MagnificationViews',
         'MammaryDuctogram']}

name_for_cid[4014] = 'ViewforMammography'
cid_concepts[4014] = \
{'SCT': ['SuperolateralToInferomedialOblique',
         'TissueSpecimenFromBreast',
         'MedioLateral',
         'CranioCaudalExaggeratedLaterally',
         'MedioLateralOblique',
         'LateroMedial',
         'LateroMedialOblique',
         'CranioCaudal',
         'InferomedialToSuperolateralOblique',
         'CaudoCranial',
         'CranioCaudalExaggeratedMedially']}

name_for_cid[3114] = 'StudyQuality'
cid_concepts[3114] = \
{'DCM': ['PoorImageQuality',
         'ExcellentImageQuality',
         'GoodImageQuality',
         'UnusableQualityRendersImageUnusable']}

name_for_cid[9519] = 'RadiotherapyAccessorySlotHolderDeviceTypes'
cid_concepts[9519] = \
{'DCM': ['RadiotherapyApplicator']}

name_for_cid[3817] = 'VascularSclerosisTypes'
cid_concepts[3817] = \
{'SCT': ['VascularWallDegeneration',
         'Arteriosclerosis',
         'Phlebosclerosis',
         'UlceratedAtheromatousPlaque',
         'ComplicatedAtheromatousPlaque',
         'AdventitialDegeneration',
         'ElasticVascularSclerosis',
         'Fibroelastosis',
         'Arteriolosclerosis',
         'AtheroscleroticFibrousPlaque',
         'Atheroma',
         'FocalFibroelastosis',
         'CysticMedicalNecrosis',
         'FattyStreaks',
         'CalcifiedAtheromatousPlaque',
         'ArteriosclerosisWithFibrinoidNecrosis',
         'DiffuseFibroelastosis']}

name_for_cid[7157] = 'DeviceSegmentationTypes'
cid_concepts[7157] = \
{'DCM': ['BrachytherapyChannel',
         'RectalBalloon',
         'VaginalCylinder',
         'Mold',
         'BrachytherapyChannelShield',
         'HeadAndNeckMask',
         'BrachytherapySourceApplicator',
         'HeadMask',
         'Headframe',
         'BodyFrame',
         'BreastBoard',
         'Cast',
         'WholeBodyPod',
         'VacuumMold'],
 'SCT': ['Chair',
         'RadioactiveImplant',
         'Needle',
         'InlayDentalRestoration',
         'Table',
         'DentalProsthesis',
         'BonePin',
         'Defibrillator',
         'Headrest',
         'CardiacPacemaker',
         'BiteBlock',
         'BoneScrew',
         'LeftVentricularAssistDevice',
         'Stent']}

name_for_cid[7012] = 'BestinSet'
cid_concepts[7012] = \
{'DCM': ['Study', 'Series', 'PerformedProcedureStep', 'StageView']}

name_for_cid[12290] = 'CardiacUltrasoundPulmonaryArteriesFindingSites'
cid_concepts[12290] = \
{'SCT': ['PulmonaryArtery',
         'PulmonaryTrunk',
         'AortaToPulmonaryArteryConnection']}

name_for_cid[3401] = 'TypesofLogNotes'
cid_concepts[3401] = \
{'DCM': ['PatientStatusOrEvent',
         'ProcedureNote',
         'NursingNote',
         'PhysicianNote',
         'TechNote']}

name_for_cid[12013] = 'GestationalAgeEquationsandTables'
cid_concepts[12013] = \
{'LN': ['ACJeanty1984',
        'ACJeanty1982',
        'OFDASUM2000',
        'CRLHansmann1985',
        'FetalTrunkCrossSectionalAreaOsaka1989',
        'BPDMertz1988',
        'HCJeanty1984',
        'HCMerz1988',
        'RadiusJeanty1983',
        'FLTokyo1986',
        'CRLHansmann1986',
        'BPDTokyo1986',
        'IODMayden1982',
        'TibiaJeanty1984',
        'CRLASUM1991',
        'GestationalAgeByLMP',
        'BDJeanty1982',
        'BPDDoubilet1993',
        'GSTokyo1986',
        'TCDHill1990',
        'CRLOsaka1989',
        'HCDerivedChitty1997',
        'TADTokyo1986',
        'ACLessoway1998',
        'AxtShinozuka1996',
        'FibulaJeanty1983',
        'RadiusMerz1987',
        'OODTrout1994',
        'BPDLessoway1998',
        'FLHansmann1986',
        'FLShinozuka1996',
        'OFDLessoway1998',
        'BPDShinozuka1996',
        'FLJeanty1984',
        'ClavicleLengthYarkoni1985',
        'FLMerz1987',
        'BPDJeanty1984',
        'GSHellman1969',
        'FLHadlock1984',
        'BpdaHadlock1982',
        'GSNyberg1992',
        'LengthOfVertebraTokyo1986',
        'CRLJeanty1982',
        'HCMeasuredChitty1997',
        'TransverseThoracicDiameterLessoway1998',
        'HCLessoway1998',
        'BPDHansmann1986',
        'TCDGoldstein1987',
        'ACHansmann1985',
        'APAbdominalDiameterLessoway1998',
        'CRLTokyo1986',
        'CRLNelson1981',
        'HumerusJeanty1984',
        'BPDJeanty1982',
        'CRLJeanty1984',
        'FLOsaka1989',
        'HCASUM2000',
        'FLMerz1988',
        'TADEriksen1985',
        'BPDRempen1991',
        'OFDHansmann1986',
        'BPDOoChitty1997',
        'BPDKurtz1980',
        'GSRempen1991',
        'FLHansmann1985',
        'CRLDaya1993',
        'TADHansmann1979',
        'BPDHadlock1984',
        'GSHansmann1982',
        'UlnaJeanty1984',
        'HumerusLengthOsaka1989',
        'ACASUM2000',
        'TransverseThoracicDiameterHansmann1985',
        'HCHadlock1984',
        'HCJeanty1982',
        'ACShinozuka1996',
        'CRLShinozuka1996',
        'TCDChitty1994',
        'FLJeanty1982',
        'ACHadlock1984',
        'CRLRobinson1975',
        'HumerusMerz1987',
        'FibulaMerz1987',
        'CRLHadlock1992',
        'HumerusLengthASUM2000',
        'BPDOiChitty1997',
        'FLLessoway1998',
        'HCHansmann1986',
        'SpineLengthTokyo1989',
        'GSDaya1991',
        'IODTrout1994',
        'BPDOsaka1989',
        'ACCampbell1975',
        'BPDSabbagha1978',
        'AverageUltrasoundAge',
        'UlnaMerz1987',
        'FLHohler1982',
        'CRLASUM2000',
        'OFDHansmann1985',
        'BPDHansmann1985',
        'FootLengthMercer1987',
        'GSHansmann1979',
        'OODMayden1982',
        'CRLRempen1991',
        'TCDNimrod1986',
        'ACMertz1988',
        'HCHansmann1985',
        'ThcChitkara1987',
        'FLChitty1997',
        'BPDASUM1989']}

name_for_cid[3728] = 'CathFindings'
cid_concepts[3728] = \
{'SCT': ['AcuteMitralRegurgitationFromChordalRupture',
         'NormalLeftVentricularSystolicFunctionAndWallMotion',
         'MildIntimalCoronaryIrregularitiesNoSignificantStenoses',
         'NormalRightHeartHemodynamics',
         'MultiVesselCoronaryArteryDisease',
         'SignificantCoronaryBypassGraftDisease',
         'PericardialTamponade',
         'MitralRegurgitation',
         'PulmonaryHypertension',
         'MitralStenosis',
         'SingleVesselCoronaryArteryDisease',
         'AtrialSeptalDefect',
         'AcuteMitralRegurgitationFromPapillaryMuscleDysfunction',
         'AcuteVentricularSeptalRupture',
         'TripleVesselCoronaryArteryDisease',
         'MitralValveProlapse',
         'HeartDiseaseCongenital',
         'NormalLeftHeartHemodynamics',
         'AorticInsufficiency',
         'NormalCoronaryArteries',
         'DepressionOfLeftVentricularSystolicFunction',
         'DoubleVesselCoronaryArteryDisease',
         'AcuteMitralRegurgitationFromChordalDysfunction',
         'HypertensiveHeartDisease',
         'CongestiveCardiomyopathy',
         'VentricularSeptalDefect',
         'NormalLeftAndRightHeartHemodynamics',
         'LeftMainCoronaryArteryDisease',
         'RestrictiveCardiomyopathy',
         'HypertrophicCardiomyopathyWithObstruction',
         'HypertrophicCardiomyopathyWithoutObstruction',
         'AcuteMitralRegurgitationFromPapillaryMuscleRupture',
         'AorticStenosis',
         'ConstrictivePericarditis']}

name_for_cid[1200] = 'ContraindicationsForCTImaging'
cid_concepts[1200] = \
{'SCT': ['ContrastMediaAllergy',
         'PatientCurrentlyPregnant',
         'XRayContrastMediaAllergy',
         'ImpairedRenalFunction']}

name_for_cid[7100] = 'RCSRegistrationMethodType'
cid_concepts[7100] = \
{'DCM': ['VisualAlignment',
         'ImageContentBasedAlignment',
         'AcquisitionEquipmentAlignment',
         'FiducialAlignment',
         'FrameOfReferenceIdentity']}

name_for_cid[12228] = 'EchocardiographyVolumeMethods'
cid_concepts[12228] = \
{'DCM': ['Teichholz',
         'MethodOfDisksSinglePlane',
         'MethodOfDisksBiplane',
         'CubeMethod',
         'BulletMethod',
         'AreaLengthSinglePlane',
         'ModifiedSimpson',
         'AreaLengthBiplane',
         'SinglePlaneEllipse',
         'BiplaneEllipse']}

name_for_cid[3840] = 'PulmonaryVeins'
cid_concepts[3840] = \
{'SCT': ['PulmonaryVein',
         'SuperiorLeftPulmonaryVein',
         'InferiorLeftPulmonaryVein',
         'RightPulmonaryVein',
         'InferiorRightPulmonaryVein',
         'LeftPulmonaryVein',
         'SuperiorRightPulmonaryVein']}

name_for_cid[3456] = 'SubsegmentMethods'
cid_concepts[3456] = \
{'DCM': ['UserSelectedMethod', 'EquidistantMethod']}

name_for_cid[6204] = 'AnatomicNoncolonFindings'
cid_concepts[6204] = \
{'SCT': ['Lung',
         'Testis',
         'Cervix',
         'Bladder',
         'Appendix',
         'BloodVessel',
         'Aorta',
         'InferiorVenaCava',
         'SeminalVesicle',
         'Rectum',
         'Ovary',
         'AppendicealStump',
         'Muscle',
         'Bone',
         'Liver',
         'Prostate',
         'Spleen',
         'Kidney',
         'AdrenalGland',
         'Uterus']}

name_for_cid[4239] = 'AnteriorChamberDepthDefinition'
cid_concepts[4239] = \
{'DCM': ['FrontOfCorneaToFrontOfLens', 'BackOfCorneaToFrontOfLens']}

name_for_cid[609] = 'AnimalFeedingMethods'
cid_concepts[609] = \
{'DCM': ['FoodTreat'],
 'NCIt': ['AdLibitum'],
 'SCT': ['Gavage'],
 'UMLS': ['RestrictedDiet']}

name_for_cid[4110] = 'TracerKineticModelingCovariates'
cid_concepts[4110] = \
{'LN': ['Hematocrit']}

name_for_cid[6087] = 'GeneralRiskFactors'
cid_concepts[6087] = \
{'DCM': ['IntermediateFamilyHistoryOfBreastCancer',
         'VeryStrongFamilyHistoryOfBreastCancer',
         'FamilyHistoryOfProstateCancer',
         'FamilyHistoryUnknown',
         'UterineMalformations',
         'SpontaneousAbortion',
         'GynecologicCondition',
         'GynecologicSurgery',
         'PreviousLBWOrIUGRBirth',
         'PreviousFetalMalformationSyndrome',
         'PreviousRHNegativeOrBloodDyscrasiaAtBirth',
         'HistoryOfMultipleFetuses',
         'PersonalBreastCancerHistory',
         'CurrentPregnancyKnownOrSuspectedMalformationsSyndromes',
         'HistoryOfEndometrialCancer',
         'FamilyHistoryFetalMalformationSyndrome',
         'HistoryOfOvarianCancer',
         'HistoryOfHighRiskLesionOnPreviousBiopsy',
         'PostMenopausalPatient',
         'LateChildBearingAfter30',
         'BRCA1BreastCancerGene',
         'BRCA2BreastCancerGene',
         'BRCA3BreastCancerGene',
         'WeakFamilyHistoryOfBreastCancer'],
 'SCT': ['HistoryOfDiabetesMellitus',
         'HistoryOfEclampsia',
         'HistoryOfInfertility',
         'FamilyHistoryOfBreastCancer',
         'HistoryOfCardiovascularDisease',
         'HistoryOfObesity',
         'HistoryOfHypertension',
         'HistoryOfSubstanceAbuse',
         'HistoryOfEctopicPregnancy',
         'Nulliparous',
         'HistoryOfPrematureDelivery',
         'HistoryOfRegularMedication',
         'HistoryOfSeverePreEclampsia',
         'MultiplePregnancy',
         'NoFamilyHistoryOfBreastCarcinoma']}

name_for_cid[8133] = 'TissueSelection'
cid_concepts[8133] = \
{'DCM': ['NominalEmptyTileSuppression',
         'HighThresholdEmptyTileSuppression',
         'NoEmptyTileSuppression']}

name_for_cid[12101] = 'VascularSummary'
cid_concepts[12101] = \
{'DCM': ['Comment']}

name_for_cid[6068] = 'TumorStagesFromAJCC'
cid_concepts[6068] = \
{'DCM': ['Stage0',
         'StageI',
         'StageIIA',
         'StageIIB',
         'StageIIIA',
         'StageIIIB',
         'StageIIIC',
         'StageIV']}

name_for_cid[4111] = 'ContrastCharacteristics'
cid_concepts[4111] = \
{'DCM': ['ContrastLongitudinalRelaxivity']}

name_for_cid[6086] = 'MenopausalPhase'
cid_concepts[6086] = \
{'SCT': ['AfterMenopause',
         'BeforeMenopause',
         'PostsurgicalMenopause',
         'ArtificialMenopauseState',
         'DuringMenopause']}

name_for_cid[8132] = 'MagnificationSelection'
cid_concepts[8132] = \
{'DCM': ['_10X', '_20X', '_40X', '_5X']}

name_for_cid[6069] = 'NottinghamCombinedHistologicGrade'
cid_concepts[6069] = \
{'SCT': ['NottinghamCombinedGradeI3To5Points',
         'NottinghamCombinedGradeII6To7Points',
         'NottinghamCombinedGradeIII8To9Points',
         'NottinghamCombinedGradeCannotBeDetermined']}

name_for_cid[7101] = 'BrainAtlasFiducials'
cid_concepts[7101] = \
{'DCM': ['InterHemisphericPlane',
         'LeftHemisphereMostInferior',
         'LeftHemisphereMostSuperior',
         'LeftHemisphereMostPosterior',
         'LeftHemisphereMostAnterior',
         'RightHemisphereMostInferior',
         'RightHemisphereMostSuperior',
         'RightHemisphereMostPosterior',
         'RightHemisphereMostAnterior'],
 'SCT': ['AnteriorCommissure', 'PosteriorCommissure']}

name_for_cid[12229] = 'EchocardiographyAreaMethods'
cid_concepts[12229] = \
{'DCM': ['ProximalIsovelocitySurfaceArea',
         'ContinuityEquationByVelocityTimeIntegral',
         'ContinuityEquationByPeakVelocity',
         'ContinuityEquationByMeanVelocity',
         'ContinuityEquation',
         'AreaByPressureHalfTime',
         'Planimetry']}

name_for_cid[4238] = 'RefractiveErrorTypes'
cid_concepts[4238] = \
{'SCT': ['Myopia', 'Hyperopia']}

name_for_cid[6205] = 'ClockfaceLocationforColon'
cid_concepts[6205] = \
{'SCT': ['_7OClockPosition',
         '_11OClockPosition',
         '_12OClockPosition',
         '_2OClockPosition',
         '_5OClockPosition',
         '_3OClockPosition',
         '_6OClockPosition',
         '_4OClockPosition',
         '_9OClockPosition',
         '_10OClockPosition',
         '_1OClockPosition',
         '_8OClockPosition']}

name_for_cid[608] = 'AnimalFeedSources'
cid_concepts[608] = \
{'DCM': ['LocallyManufacturedProduct'], 'UMLS': ['CommercialProduct']}

name_for_cid[12012] = 'OBEquationsandTables'
cid_concepts[12012] = \
{'DCM': ['FWPByGACampbell1991', 'FWPByGAHadlock1991'],
 'LN': ['ACJeanty1984',
        'CRLByGARempen1991',
        'ACJeanty1982',
        'OFDASUM2000',
        'FibulaByGAJeanty1983',
        'CRLHansmann1985',
        'FemaleSingletonBWPByGAArbuckle1993',
        'FetalTrunkCrossSectionalAreaOsaka1989',
        'BPDMertz1988',
        'HCJeanty1984',
        'EFWByACFLHCHadlock1985',
        'FLByGAJeanty1982',
        'HCMerz1988',
        'RadiusJeanty1983',
        'CephalicIndexByGAChitty1994',
        'FLTokyo1986',
        'EFWByACBPDFLHadlock1984',
        'CRLHansmann1986',
        'ACByGAASUM2000',
        'BPDTokyo1986',
        'IODMayden1982',
        'TibiaJeanty1984',
        'CRLASUM1991',
        'EFW1ByShinozuka1996',
        'GestationalAgeByLMP',
        'MaleTwinsBWPByGAArbuckle1993',
        'BDJeanty1982',
        'HCByGAASUM2000',
        'BPDByGAJeanty1982',
        'BPDDoubilet1993',
        'GSTokyo1986',
        'EFWByACBPDHadlock1984',
        'HCByGAMerz1988',
        'TCDHill1990',
        'CRLOsaka1989',
        'HCDerivedChitty1997',
        'BPDOuterInnerByGAChitty1994',
        'TADTokyo1986',
        'ACLessoway1998',
        'AxtShinozuka1996',
        'FibulaJeanty1983',
        'RadiusMerz1987',
        'HCACByGACampbell1977',
        'EFWByGAHansmann1986',
        'MaleSingletonBWPByGAArbuckle1993',
        'FLByGAChitty1994',
        'OODTrout1994',
        'EFWByACFLHadlock1984',
        'EFWByBPDAPADTADFLTokyo1987',
        'BPDLessoway1998',
        'FLHansmann1986',
        'FLShinozuka1996',
        'OFDLessoway1998',
        'BPDShinozuka1996',
        'FLJeanty1984',
        'ClavicleLengthYarkoni1985',
        'EFWByBPDFTAFLOsaka1990',
        'FLMerz1987',
        'GSByGARempen1991',
        'BPDJeanty1984',
        'BPDOuterOuterByGAChitty1994',
        'GSHellman1969',
        'HCByGAJeanty1982',
        'BPDByGAShinozuka1996',
        'BPDByGAHadlock1984',
        'OFDByGAChitty1994',
        'EFWByACHCHadlock1984',
        'FLHadlock1984',
        'EFWByACBPDFLHadlock1985',
        'BpdaHadlock1982',
        'ACByGAShinozuka1996',
        'GSNyberg1992',
        'EFWByACAndBPDShepard1982',
        'LengthOfVertebraTokyo1986',
        'CRLJeanty1982',
        'HCMeasuredChitty1997',
        'TransverseThoracicDiameterLessoway1998',
        'HCLessoway1998',
        'BPDHansmann1986',
        'TCDGoldstein1987',
        'ACHansmann1985',
        'APAbdominalDiameterLessoway1998',
        'CRLTokyo1986',
        'FWPByGAHadlock1985',
        'FWPByGAAlexander1996',
        'FLByGAHadlock1984',
        'CRLNelson1981',
        'FWPByGABrenner1976',
        'HumerusJeanty1984',
        'TCDByGAGoldstein1987',
        'EFWByGAHadlock1991',
        'BPDJeanty1982',
        'CRLJeanty1984',
        'FLOsaka1989',
        'HCASUM2000',
        'EFW3ByShinozuka1996',
        'FLMerz1988',
        'TADEriksen1985',
        'EFWByACFLHadlock1985',
        'BPDRempen1991',
        'OFDHansmann1986',
        'ACMeasuredByGAChitty1994',
        'ACDerivedByGAChitty1994',
        'BPDOoChitty1997',
        'FLByGAShinozuka1996',
        'BPDKurtz1980',
        'BPDByGAASUM2000',
        'GSRempen1991',
        'HCDerivedByGAChitty1994',
        'FLHansmann1985',
        'CRLDaya1993',
        'BPDByGARempen1991',
        'TADHansmann1979',
        'BPDHadlock1984',
        'GSHansmann1982',
        'ACByGAMerz1988',
        'OFDByGAASUM2000',
        'EFWByACBPDFLHCHadlock1985',
        'UlnaJeanty1984',
        'CRLByGAASUM2000',
        'HumerusLengthOsaka1989',
        'ACASUM2000',
        'TransverseThoracicDiameterHansmann1985',
        'HCHadlock1984',
        'HCJeanty1982',
        'ACShinozuka1996',
        'CRLShinozuka1996',
        'TCDChitty1994',
        'FLJeanty1982',
        'CRLByGAShinozuka1996',
        'ACHadlock1984',
        'CRLRobinson1975',
        'FWPByGAWilliams1982',
        'FLByGAASUM2000',
        'HumerusMerz1987',
        'FemaleTwinsBWPByGAArbuckle1993',
        'FibulaMerz1987',
        'EFWByBPDTTDHansmann1986',
        'CRLHadlock1992',
        'FLByGAMerz1988',
        'HumerusLengthASUM2000',
        'BPDOiChitty1997',
        'CephalicIndexByGAHadlock1981',
        'FLLessoway1998',
        'HCHansmann1986',
        'SpineLengthTokyo1989',
        'GSDaya1991',
        'ACByGAHadlock1984',
        'IODTrout1994',
        'EFW2ByShinozuka1996',
        'BPDOsaka1989',
        'RadiusByGAJeanty1983',
        'ACCampbell1975',
        'BPDSabbagha1978',
        'AverageUltrasoundAge',
        'AxtByGAShinozuka1996',
        'UlnaMerz1987',
        'FLHohler1982',
        'HCByGAHadlock1984',
        'CRLASUM2000',
        'OFDHansmann1985',
        'BPDHansmann1985',
        'BPDByGAMerz1988',
        'FootLengthMercer1987',
        'HumerusLengthByGAASUM2000',
        'GSHansmann1979',
        'OODMayden1982',
        'EFWByACCampbell1975',
        'CRLRempen1991',
        'TCDNimrod1986',
        'ACMertz1988',
        'HCHansmann1985',
        'ThcChitkara1987',
        'FLChitty1997',
        'BPDASUM1989']}

name_for_cid[3729] = 'AdmissionStatus'
cid_concepts[3729] = \
{'SCT': ['Elective', 'EmergencyDepartment', 'Transfer']}

name_for_cid[9518] = 'RadiotherapyAccessoryNoSlotHolderDeviceTypes'
cid_concepts[9518] = \
{'DCM': ['AccessoryTray']}

name_for_cid[3115] = 'StressImagingQualityIssues'
cid_concepts[3115] = \
{'DCM': ['BreastAttenuation',
         'DiaphragmaticAttenuation',
         'BodyHabitusAttenuation',
         'MotionBlur'],
 'SCT': ['SubdiaphragmaticUptake']}

name_for_cid[7156] = 'VascularSegmentationTypes'
cid_concepts[7156] = \
{'SCT': ['CardiovascularSystem',
         'VascularSclerosis',
         'Intima',
         'Thrombus',
         'SystemicArtery',
         'Adventitia',
         'Media',
         'SystemicVein',
         'Lumen']}

name_for_cid[7013] = 'NonImageSourceInstancePurposesofReference'
cid_concepts[7013] = \
{'DCM': ['SourceMeasurement',
         'SourceRawData',
         'SourceReport',
         'SourceRealWorldValueMap']}

name_for_cid[12291] = 'CardiacUltrasoundAortaFindingSites'
cid_concepts[12291] = \
{'SCT': ['InfraRenalAorta',
         'InnominateArtery',
         'CoarctationOfAorta',
         'LeftSubclavianArtery',
         'NonCoronarySinus',
         'AorticIsthmus',
         'RightSinusOfValsalva',
         'AorticSinotubularJunction',
         'AorticArch',
         'SupraRenalAorta',
         'AscendingAorta',
         'RootOfAorta',
         'LeftCommonCarotidArtery',
         'AbdominalAorta',
         'StructureSinusOfValsalva',
         'ThoracicAorta',
         'RightCommonCarotidArtery',
         'LeftSinusOfValsalva',
         'RightSubclavianArtery']}

name_for_cid[3400] = 'ProcedureLogTitles'
cid_concepts[3400] = \
{'DCM': ['CathLabProcedureLog']}

name_for_cid[12004] = 'FetalBiometryRatios'
cid_concepts[12004] = \
{'LN': ['HCAC', 'FLAC', 'CephalicIndex', 'FLBPD', 'FLHC']}

name_for_cid[12141] = 'FetalVasculatureAnatomicalLocation'
cid_concepts[12141] = \
{'SCT': ['DescendingAorta',
         'MiddleCerebralArtery',
         'PulmonaryArtery',
         'PulmonaryVein',
         'Aorta']}

name_for_cid[4015] = 'ViewModifierforMammography'
cid_concepts[4015] = \
{'SCT': ['RolledMedial',
         'NippleInProfile',
         'AxillaryTissue',
         'AnteriorCompression',
         'RolledSuperior',
         'SpotCompression',
         'Tangential',
         'AxillaryTail',
         'Magnification',
         'ImplantDisplaced',
         'Cleavage',
         'InfraMammaryFold',
         'RolledLateral',
         'RolledInferior']}

name_for_cid[6028] = 'MammographyRecommendedFollowup'
cid_concepts[6028] = \
{'DCM': ['NormalIntervalFollowUp',
         'AnyDecisionToBiopsyShouldBeBasedOnClinicalAssessment',
         'FollowUpAtShortInterval111Months',
         'FollowUpPostBiopsyAsDirectedByClinician',
         'BiopsyShouldBeConsidered',
         'NeedleLocalizationAndBiopsy',
         'KnownBiopsyProvenMalignancyTakeAppropriateAction',
         'HistologyUsingCoreBiopsy',
         'SurgicalConsult',
         'AdditionalProjections',
         'SuggestiveOfMalignancyTakeAppropriateAction',
         'SpotMagnificationViews',
         'CytologicAnalysis',
         'BiopsyShouldBeStronglyConsidered',
         'OldFilmsForComparison',
         'HighlySuggestiveOfMalignancyTakeAppropriateAction'],
 'SCT': ['SpotCompression',
         'NuclearMedicineProcedure',
         'MRIOfBreast',
         'DiagnosticUltrasonography',
         'MagnificationViews',
         'EvaluationProcedure',
         'MammaryDuctogram']}

name_for_cid[7140] = 'BrainStructuresforVolumetricMeasurements'
cid_concepts[7140] = \
{'DCM': ['WhiteMatterT2Hyperintensity',
         'WhiteMatterT1Hypointensity',
         'VentralDiencephalon'],
 'SCT': ['Amygdala',
         'CranialSubarachnoidSpace',
         'FourthVentricle',
         'ThirdVentricle',
         'CerebralGrayMatter',
         'GlobusPallidus',
         'Thalamus',
         'CerebralWhiteMatter',
         'InferiorHornOfLateralVentricle',
         'LateralVentricle',
         'CaudateNucleus',
         'NucleusAccumbens',
         'CerebellarCortex',
         'CerebellarSubarachnoidSpace',
         'BrainStem',
         'Putamen',
         'Hippocampus',
         'FifthVentricle',
         'CerebellarWhiteMatter',
         'CranialCavity',
         'IntracranialStructure']}

name_for_cid[12268] = 'CardiacUltrasoundAtrioventricularValvesMeasurements'
cid_concepts[12268] = \
{'DCM': ['Tau',
         'IndicatorMeanTransitTime',
         'RRInterval',
         'BloodVelocityPeak',
         'BloodVelocityMean',
         'BloodVelocityMinimum',
         'SystolicBloodVelocityMean',
         'SystolicBloodVelocityPeak',
         'DiastolicBloodVelocityMean',
         'DiastolicBloodVelocityPeak'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'TimeFromQWaveToTricuspidValveOpens',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'LeafletThickness',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'MitralRegurgitationDpDtDerivedFromMitralRegVelocity',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'LeafletSeparation',
        'AccelerationTime',
        'MitralValveESeptalSeparation',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'CEDistance',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'MitralValveEPSSEWave',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'DEExcursion',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'TricuspidDiastolicFillingPeriodDfpt',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[3553] = 'CirculatorySupport'
cid_concepts[3553] = \
{'SCT': ['ExternalCounterPulsation',
         'LeftVentricularAssistDevice',
         'CardiopulmonaryBypass',
         'IntraAorticBalloonPump',
         'ExtraCorporealCirculation']}

name_for_cid[6301] = 'ProstateSectorAnatomyfromPIRADSv2'
cid_concepts[6301] = \
{'SCT': ['CentralZoneOfLeftHalfProstate',
         'LeftPosteromedialMiddlePeripheralZoneOfProstate',
         'MaleExternalUrethralSphincter',
         'RightApicalAnteriorFibromuscularStromaOfProstate',
         'LeftPosteromedialApicalPeripheralZoneOfProstate',
         'LeftPosterolateralMiddlePeripheralZoneOfProstate',
         'RightBasalAnteriorFibromuscularStromaOfProstate',
         'RightAnteriorBasalTransitionZoneOfProstate',
         'LeftAnteriorApicalTransitionZoneOfProstate',
         'LeftAnteriorBasalTransitionZoneOfProstate',
         'LeftPosteriorApicalTransitionZoneOfProstate',
         'RightPosterolateralBasalPeripheralZoneOfProstate',
         'LeftPosteriorMiddleTransitionZoneOfProstate',
         'LeftAnteriorMiddleTransitionZoneOfProstate',
         'LeftBasalAnteriorFibromuscularStromaOfProstate',
         'LeftPosteriorBasalTransitionZoneOfProstate',
         'RightAnteriorBasalPeripheralZoneOfProstate',
         'LeftMiddleAnteriorFibromuscularStromaOfProstate',
         'LeftPosterolateralApicalPeripheralZoneOfProstate',
         'RightSeminalVesicle',
         'LeftAnteriorApicalPeripheralZoneOfProstate',
         'CentralZoneOfRightHalfProstate',
         'RightPosteromedialMiddlePeripheralZoneOfProstate',
         'RightAnteriorMiddlePeripheralZoneOfProstate',
         'RightPosterolateralMiddlePeripheralZoneOfProstate',
         'LeftApicalAnteriorFibromuscularStromaOfProstate',
         'RightPosteromedialApicalPeripheralZoneOfProstate',
         'LeftSeminalVesicle',
         'RightAnteriorApicalPeripheralZoneOfProstate',
         'LeftAnteriorBasalPeripheralZoneOfProstate',
         'LeftPosterolateralBasalPeripheralZoneOfProstate',
         'RightMiddleAnteriorFibromuscularStromaOfProstate',
         'RightPosterolateralApicalPeripheralZoneOfProstate',
         'RightPosteriorBasalTransitionZoneOfProstate',
         'RightPosteriorMiddleTransitionZoneOfProstate',
         'RightAnteriorApicalTransitionZoneOfProstate',
         'RightPosteriorApicalTransitionZoneOfProstate',
         'LeftAnteriorMiddlePeripheralZoneOfProstate',
         'RightAnteriorMiddleTransitionZoneOfProstate']}

name_for_cid[3416] = 'RespirationRhythms'
cid_concepts[3416] = \
{'SCT': ['GaspingRespiration',
         'RespirationIntermittent',
         'AbnormalRespiratoryRhythm',
         'NormalRespiratoryRhythm',
         'IrregularBreathing']}

name_for_cid[12287] = 'CardiacUltrasoundVentriclesFindingSites'
cid_concepts[12287] = \
{'SCT': ['RightVentricle', 'LeftVentricle', 'CommonVentricle']}

name_for_cid[7005] = 'ContributingEquipmentPurposesofReference'
cid_concepts[7005] = \
{'DCM': ['PortableMediaImporterEquipment',
         'FilmDigitizer',
         'DeIdentifyingEquipment',
         'ModifyingEquipment',
         'EnhancedMultiFrameConversionEquipment',
         'FrameExtractingEquipment',
         'VideoTapeDigitizerEquipment',
         'ProcessingEquipment',
         'AcquisitionEquipment',
         'DocumentDigitizerEquipment']}

name_for_cid[7455] = 'Sex'
cid_concepts[7455] = \
{'DCM': ['MalePseudohermaphrodite',
         'UnknownSex',
         'UndeterminedSex',
         'Female',
         'MaleChangedToFemale',
         'AmbiguousSex',
         'Hermaphrodite',
         'OtherSex',
         'FemalePseudohermaphrodite',
         'Male',
         'FemaleChangedToMale']}

name_for_cid[8124] = 'MicroscopyFilter'
cid_concepts[8124] = \
{'DCM': ['NoFilter'],
 'SCT': ['BlueOpticalFilter',
         'NomarskiPrism',
         'GreenOpticalFilter',
         'HoffmanModulator',
         'CondenserAnnulus',
         'PolarizingOpticalFilter',
         'VioletOpticalFilter',
         'InfraredOpticalFilter',
         'DarkfieldStop',
         'RheinbergFilter',
         'UltravioletOpticalFilter',
         'DichroicBeamsplitter',
         'PhaseContrastPlate',
         'RedOpticalFilter',
         'DeSénarmontCompensator']}

name_for_cid[6090] = 'RelativeUsage,ExposureAmount'
cid_concepts[6090] = \
{'DCM': ['High', 'Medium', 'NoKnownExposure', 'Low']}

name_for_cid[4107] = 'TracerKineticModelParameters'
cid_concepts[4107] = \
{'DCM': ['', 'Vp', 'Ktrans', 'Kep', 'Ve']}

name_for_cid[4042] = 'XA/XRFAnatomyImaged'
cid_concepts[4042] = \
{'DCM': ['Phantom'],
 'SCT': ['SubclavianVein',
         'ExternalIliacArtery',
         'BoneOfLowerLimb',
         'TruncusCoeliacus',
         'BasilarArtery',
         'ParanasalSinus',
         'ApexOfRightVentricle',
         'WristJoint',
         'GenicularArtery',
         'LeftVentricleInflow',
         'CarotidArtery',
         'UmbilicalArtery',
         'PrimitiveAorta',
         'LeftVentricleOutflowTract',
         'RadialArtery',
         'RightFemoralArtery',
         'JawRegion',
         'LeftFemoralArtery',
         'Calcaneus',
         'Esophagus',
         'AntecubitalVein',
         'Mandible',
         'PatentDuctusArteriosus',
         'MesentericVein',
         'PulmonaryVein',
         'CoronaryArtery',
         'UpperLimb',
         'ExternalIliacVein',
         'CongenitalPulmonaryArteriovenousFistula',
         'DescendingAorta',
         'SubclavianArtery',
         'Maxilla',
         'OphthalmicArtery',
         'Hand',
         'BrachialVein',
         'Eye',
         'Chest',
         'Prostate',
         'ElbowJoint',
         'HepaticArtery',
         'FacialBones',
         'Sternum',
         'PeronealArtery',
         'NeckAndChest',
         'RightVentricleOutflowTract',
         'Duodenum',
         'SuperiorLeftPulmonaryVein',
         'BileDuct',
         'SystemicCollateralArteryToLung',
         'LumbarArtery',
         'PancreaticDuctAndBileDuctSystems',
         'ApexOfLeftVentricle',
         'Larynx',
         'FemoralArtery',
         'PulmonaryArtery',
         'Skull',
         'ThoracoLumbarSpine',
         'Urethra',
         'Head',
         'AcromioclavicularJoint',
         'CongenitalCoronaryArteryFistulaToRightVentricle',
         'PosteriorMedialTributary',
         'SternoclavicularJoint',
         'Breast',
         'EsophagusStomachAndDuodenum',
         'Neck',
         'Finger',
         'PulmonaryArteryConduit',
         'RightAtrium',
         'RightVentricle',
         'CommonVentricle',
         'Coccyx',
         'Abdomen',
         'HeadAndNeck',
         'TemporomandibularJoint',
         'HepaticVein',
         'PelvisAndLowerExtremities',
         'InternalMammaryArtery',
         'Bladder',
         'LargeIntestine',
         'Forearm',
         'Spine',
         'Foot',
         'NeckChestAndAbdomen',
         'ExternalCarotidArtery',
         'SuperiorMesentericArtery',
         'MuscleOfUpperLimb',
         'PosteriorTibialArtery',
         'FemoralVein',
         'RightAuricularAppendage',
         'Zygoma',
         'InnominateVein',
         'SuperficialFemoralArtery',
         'UterusAndFallopianTubes',
         'LowerLimb',
         'PosteriorCommunicationArtery',
         'JuxtaposedAtrialAppendage',
         'AorticArch',
         'VertebralColumnAndCranium',
         'Jejunum',
         'ChestAbdomenAndPelvis',
         'AbdominalAorta',
         'Heart',
         'TarsalJoint',
         'CephalicVein',
         'LingualArtery',
         'GastricVein',
         'CongenitalCoronaryArteryFistulaToRightAtrium',
         'Fibula',
         'OrbitalStructure',
         'InferiorCardiacVein',
         'Baffle',
         'LacrimalArtery',
         'OccipitalArtery',
         'SystemicVenousAtrium',
         'AnomalousPulmonaryVein',
         'CervicoThoracicSpine',
         'Femur',
         'Rectum',
         'AnteriorCardiacVein',
         'LeftAtrium',
         'CerebralArtery',
         'Scapula',
         'InferiorMesentericArtery',
         'GreatSaphenousVein',
         'CervicalSpine',
         'InternalCarotidArtery',
         'Ureter',
         'Sacrum',
         'SmallIntestine',
         'InternalAuditoryCanal',
         'SuperiorRightPulmonaryVein',
         'InferiorVenaCava',
         'RightPulmonaryArtery',
         'VenousNetwork',
         'CommonAtrium',
         'Humerus',
         'Ilium',
         'AorticFistula',
         'SuperiorThyroidArtery',
         'IliacArtery',
         'NasalBone',
         'AnteriorTibialArtery',
         'TruncusArteriosusCommunis',
         'Artery',
         'CommonCarotidArtery',
         'OpticCanal',
         'Aorta',
         'SuperficialTemporalArtery',
         'MastoidBone',
         'Vein',
         'CongenitalCoronaryArteryFistulaToLeftVentricle',
         'NeckChestAbdomenAndPelvis',
         'Thigh',
         'ChestAndAbdomen',
         'InferiorRightPulmonaryVein',
         'Trachea',
         'BodyConduit',
         'AnkleJoint',
         'Mediastinum',
         'RightVentricleInflow',
         'MesentericArtery',
         'PoplitealArtery',
         'UpperUrinaryTract',
         'PulmonaryVenousAtrium',
         'Bronchus',
         'Pelvis',
         'SesamoidBonesOfFoot',
         'InternalIliacArtery',
         'PulmonaryChamberOfCorTriatriatum',
         'SuperiorVenaCava',
         'UlnarArtery',
         'LeftVentricle',
         'AnteriorSpinalArtery',
         'AbdomenAndPelvis',
         'SplenicVein',
         'Colon',
         'SubmandibularGland',
         'ThoracicSpine',
         'SaphenousVein',
         'Gallbladder',
         'InnominateArtery',
         'SplenicArtery',
         'RenalArtery',
         'UpperArm',
         'Extremity',
         'FacialArtery',
         'AzygosVein',
         'BrachialArtery',
         'HipJoint',
         'Rib',
         'BoneOfUpperLimb',
         'Clavicle',
         'ParotidGland',
         'CommonFemoralArtery',
         'Stomach',
         'AnteriorCommunicatingArtery',
         'CongenitalCoronaryArteryFistulaToLeftAtrium',
         'ThoracicAorta',
         'CoronarySinus',
         'AxillaryArtery',
         'LeftPulmonaryArtery',
         'ApexOfLung',
         'Ileum',
         'Knee',
         'VertebralArtery',
         'PulmonaryVeinConfluence',
         'SaphenofemoralJunction',
         'InternalJugularVein',
         'EyeRegion',
         'LacrimalArteryOfRightEye',
         'LumboSacralSpine',
         'SacroiliacJoint',
         'Anus',
         'UmbilicalVein',
         'EntireBody',
         'CommonIliacVein',
         'Thumb',
         'LumbarSpine',
         'GreatCardiacVein',
         'PrimitivePulmonaryArtery',
         'SellaTurcica',
         'PortalVein',
         'InferiorLeftPulmonaryVein',
         'LowerLeg',
         'RenalVein',
         'BoydPerforatingVein',
         'CommonIliacArtery',
         'Toe',
         'Shoulder',
         'LeftAuricularAppendage',
         'MuscleOfLowerLimb',
         'AscendingAorta',
         'HunterianPerforatingVein',
         'ProfundaFemorisArtery',
         'AxillaryVein',
         'Patella',
         'OccipitalVein',
         'DoddPerforatingVein']}

name_for_cid[12116] = 'VesselSegmentModifiers'
cid_concepts[12116] = \
{'NCIt': ['Middle'],
 'SCT': ['Distal', 'Proximal', 'DilatedPortionOfSegment', 'OriginOfVessel']}

name_for_cid[3011] = 'ElectrophysiologyAnatomicLocations'
cid_concepts[3011] = \
{'SCT': ['GreatCardiacVein',
         'LateralHighRightAtrium',
         'LowRightAtrium',
         'RightVentricleInflow',
         'InferiorLeftPulmonaryVein',
         'ApexOfRightVentricle',
         'SuperiorRightPulmonaryVein',
         'PurkinjeFibers',
         'LeftPosteriorDivisionOfLeftBranchAtrioventricularBundle',
         'MitralRing',
         'RightAtrium',
         'LeftVentricleInflow',
         'RightVentricle',
         'LeftVentricle',
         'CommonVentricle',
         'MiddleCardiacVein',
         'PulmonicRing',
         'CoronarySinus',
         'HighRightAtrium',
         'AtrioventricularNode',
         'LeftVentricleOutflowTract',
         'MidRightAtrium',
         'TricuspidRing',
         'RightVentricleOutflowTract',
         'TendonOfTodaro',
         'SuperiorLeftPulmonaryVein',
         'LeftAuricularAppendage',
         'Epicardium',
         'RightBranchOfAtrioventricularBundle',
         'AccessoryAtrioventricularBundle',
         'ApexOfLeftVentricle',
         'OstiumOfCoronarySinus',
         'RightAtrioventricularOstium',
         'LeftAnteriorDivisionOfLeftBranchAtrioventricularBundle',
         'LeftAtrium',
         'PulmonaryVein',
         'InferiorRightPulmonaryVein',
         'LeftBranchOfAtrioventricularBundle',
         'RightAuricularAppendage',
         'AtrioventricularBundle',
         'SinoAtrialNode']}

name_for_cid[3441] = 'PatientAssessments'
cid_concepts[3441] = \
{'LN': ['RespirationRhythm', 'CardiacRhythm'],
 'SCT': ['RespirationAssessment',
         'PatientMentalStateAssessment',
         'SkinConditionAssessment']}

name_for_cid[3206] = 'NoninvasiveCardiacImagingProcedures'
cid_concepts[3206] = \
{'SCT': ['SPECT',
         'PETHeartStudy',
         'CardiacBloodPoolImaging',
         'NuclearMedicineCardiovascularStudy',
         'CardiacMRI',
         'Echocardiography']}

name_for_cid[6004] = 'MammographyCharacteristicsofShape'
cid_concepts[6004] = \
{'SCT': ['Lobular', 'RoundShape', 'Irregular', 'OvoidShapeOval']}

name_for_cid[7300] = 'ImplantMaterials'
cid_concepts[7300] = \
{'SCT': ['Polymer',
         'CarbonFiber',
         'StainlessSteelMaterial',
         'GoldAlloy',
         'NickelTitanium']}

name_for_cid[10015] = 'CTDoseReferenceAuthorities'
cid_concepts[10015] = \
{'DCM': ['ICRPPub60', 'ICRPPub103']}

name_for_cid[3713] = 'TIMIFlowCharacteristics'
cid_concepts[3713] = \
{'SCT': ['_1PenetrationWithoutPerfusion',
         '_2PartialPerfusion',
         '_3CompletePerfusion',
         '_0NoPerfusion']}

name_for_cid[6141] = 'AttenuationCoefficientMeasurements'
cid_concepts[6141] = \
{'DCM': ['StandardDeviationOfAttenuationCoefficient',
         'AttenuationCoefficient',
         'MinimumAttenuationCoefficient',
         'MaximumAttenuationCoefficient',
         'MeanAttenuationCoefficient',
         'MedianAttenuationCoefficient']}

name_for_cid[12301] = 'MeasurementSelectionReasons'
cid_concepts[12301] = \
{'DCM': ['MostRecentValueChosen', 'MeanValueChosen', 'UserChosenValue'],
 'SCT': ['Minimum', 'Maximum']}

name_for_cid[7479] = 'GreyLevelDistanceZoneBasedFeatures'
cid_concepts[7479] = \
{}

name_for_cid[7183] = 'AbstractMultidimensionalImageModelDimensionUnits'
cid_concepts[7183] = \
{'UCUM': ['Millimeter',
          'Radian',
          'Degree',
          'Centimeter',
          'Micrometer',
          'Millisecond',
          'Second']}

name_for_cid[4255] = 'VisualFieldIlluminationColor'
cid_concepts[4255] = \
{'SCT': ['Yellow', 'Red', 'White', 'Blue', 'Green']}

name_for_cid[9522] = 'MultiSourceRadiotherapyProcedureTechniques'
cid_concepts[9522] = \
{'DCM': ['MultipleFixedSources']}

name_for_cid[12244] = 'CongenitalFindingSites'
cid_concepts[12244] = \
{'SCT': ['AtrialSeptalDefect', 'VentricularSeptalDefect']}

name_for_cid[6053] = 'BreastImagingReportElements'
cid_concepts[6053] = \
{'DCM': ['ProcedureReported',
         'Conclusion',
         'PhysicalExaminationResults',
         'Recommendation',
         'ReasonForProcedure',
         'ComparisonToPreviousExams',
         'OverallAssessment',
         'Impression',
         'Finding'],
 'SCT': ['BreastComposition']}

name_for_cid[6403] = 'NonlesionObjectTypeTissues'
cid_concepts[6403] = \
{'SCT': ['ScarTissue']}

name_for_cid[6116] = 'MuscularAnatomy'
cid_concepts[6116] = \
{'SCT': ['SternocleidomastoidMuscle',
         'ErectorSpinaeMuscle',
         'ExternalIntercostalMuscle',
         'PectoralisMajorMuscle',
         'InfraspinatusMuscle',
         'Diaphragm',
         'LevatoresCostarumMuscles',
         'SubcostalMuscle',
         'SubscapularisMuscle',
         'TransversusThoracis',
         'DeltoidMuscle',
         'InterventricularSeptum',
         'TrapeziusMuscle',
         'InternalIntercostalMuscle',
         'InnermostIntercostalMuscles',
         'TeresMajorMuscle',
         'SpinalisMuscle',
         'PectoralisMinorMuscle',
         'ChordaeTendineaeCordis',
         'InteratrialSeptum',
         'ScalenousAnteriorMuscle',
         'TeresMinorMuscle',
         'LatissimusDorsiMuscle',
         'TrabeculaeCarnae',
         'IliocostalisMuscle',
         'LongissimusMuscle',
         'SerratusAnteriorMuscle',
         'SupraspinatusMuscle']}

name_for_cid[3744] = 'EFTestingMethod'
cid_concepts[3744] = \
{'SCT': ['CardiacVentriculography',
         'RadionuclideVentriculography',
         'Echocardiography']}

name_for_cid[61] = 'TimeRelativetoProcedure'
cid_concepts[61] = \
{'SCT': ['AfterProcedure', 'BeforeProcedure', 'DuringProcedure']}

name_for_cid[7707] = 'SpinalCordFibers'
cid_concepts[7707] = \
{'SCT': ['LateralFuniculus', 'DorsalFuniculus', 'VentralFuniculus']}

name_for_cid[1000] = 'CTTransversePlaneReferenceBasis'
cid_concepts[1000] = \
{'DCM': ['AcquiredVolume'],
 'FMA': ['Nasion'],
 'SCT': ['KneeJoint',
         'TemporalBone',
         'Sacrum',
         'LevelOfL1L2IntervertebralDisc',
         'ApexOfHeart',
         'SternoclavicularJoint',
         'LevelOfT7T8IntervertebralDisc',
         'OrthopedicDevice',
         'C1Vertebra',
         'LevelOfC5C6IntervertebralDisc',
         'WristJoint',
         'LevelOfT11T12IntervertebralDisc',
         'Ischium',
         'Diaphragm',
         'LevelOfT3T4IntervertebralDisc',
         'CarotidArtery',
         'LevelOfT2T3IntervertebralDisc',
         'Talus',
         'FemoralHead',
         'MaxillarySinus',
         'LevelOfT10T11IntervertebralDisc',
         'MalleolarStructureOfTibia',
         'LevelOfT6T7IntervertebralDisc',
         'LevelOfC2C3IntervertebralDisc',
         'Mandible',
         'Metacarpal',
         'Foot',
         'CommonIliacArteryBifurcation',
         'LevelOfC6C7IntervertebralDisc',
         'MastoidBone',
         'XiphoidProcess',
         'Acetabulum',
         'LevelOfL4L5IntervertebralDisc',
         'ForamenMagnum',
         'AnkleJoint',
         'LevelOfL5S1IntervertebralDisc',
         'LevelOfT9T10IntervertebralDisc',
         'Lung',
         'LevelOfT5T6IntervertebralDisc',
         'PituitaryFossa',
         'LevelOfL3L4IntervertebralDisc',
         'SymphysisPubis',
         'LevelOfT12L1IntervertebralDisc',
         'TibialPlateau',
         'LesserTrochanter',
         'AorticArch',
         'Carina',
         'LevelOfC3C4IntervertebralDisc',
         'Heart',
         'Toe',
         'ShoulderRegionStructure',
         'ElbowJoint',
         'MorphologicallyAbnormalStructure',
         'LateralCanthus',
         'Sternum',
         'ThoracicInlet',
         'IliacCrest',
         'OrbitalStructure',
         'LevelOfT8T9IntervertebralDisc',
         'LevelOfL2L3IntervertebralDisc',
         'HipJoint',
         'LevelOfC7T1IntervertebralDisc',
         'LevelOfC4C5IntervertebralDisc',
         'MastoidCellsAndAntra',
         'SuprasternalNotch',
         'FrontalSinus',
         'CardiacPacemaker',
         'Skull',
         'Pancreas',
         'LevelOfT4T5IntervertebralDisc',
         'InternalAuditoryMeatus',
         'Liver',
         'LevelOfT1T2IntervertebralDisc',
         'CoronaryArteryGraft',
         'ExternalAuditoryMeatus',
         'VertexOfHead',
         'Scapula',
         'Patella',
         'Kidney',
         'AdrenalGland',
         'StentDevice',
         'AcromioclavicularJoint',
         'Scaphoid',
         'ImplantDevice']}

name_for_cid[4202] = 'OphthalmicPhotographyAcquisitionDevice'
cid_concepts[4202] = \
{'SCT': ['FundusCamera',
         'SlitLampBiomicroscope',
         'DirectOphthalmoscope',
         'OphthalmicEndoscope',
         'ScanningLaserOphthalmoscope',
         'Keratoscope',
         'ExternalCamera',
         'Pupillograph',
         'IndirectOphthalmoscope',
         'SpecularMicroscope',
         'OperatingMicroscope']}

name_for_cid[632] = 'PhaseofSurgicalProcedureRequiringAnesthesia'
cid_concepts[632] = \
{'SCT': ['Intraoperative', 'Preoperative', 'Postoperative']}

name_for_cid[3528] = 'BloodpH'
cid_concepts[3528] = \
{'LN': ['BloodPh', 'VenousBloodPh', 'ArterialBloodPh']}

name_for_cid[3482] = 'IVUSAreaMeasurements'
cid_concepts[3482] = \
{'DCM': ['InStentNeointimalCrossSectionalArea',
         'EEMCrossSectionalArea',
         'PlaquePlusMediaCrossSectionalArea'],
 'SCT': ['VesselLumenCrossSectionalArea', 'StentCrossSectionalArea']}

name_for_cid[3617] = 'ValveFlows'
cid_concepts[3617] = \
{'SCT': ['DerivedFlowNonValve',
         'AorticValveFlow',
         'MitralValveFlow',
         'TricuspidValveFlow',
         'PulmonaryValveFlow']}

name_for_cid[6045] = 'MammographyTypesofQualityControlStandard'
cid_concepts[6045] = \
{'DCM': ['InstitutionallyDefinedQualityControlStandard',
         'MammographyQualityControlManual1999ACR',
         'Title21CFRSection900SubpartB']}

name_for_cid[3752] = 'GuidewireCrossing'
cid_concepts[3752] = \
{'DCM': ['GuidewireCrossingLesionSuccessful',
         'GuidewireCrossingLesionUnsuccessful']}

name_for_cid[6100] = 'ChestComponentCategories'
cid_concepts[6100] = \
{'DCM': ['Bronchovascular', 'Osseous'],
 'SCT': ['Lung',
         'Mediastinum',
         'Heart',
         'PleuralStructure',
         'SystemicVascularStructure',
         'Muscular']}

name_for_cid[624] = 'InhalationalAnesthesiaAgentsforSmallAnimalAnesthesia'
cid_concepts[624] = \
{'SCT': ['Chloroform',
         'Isoflurane',
         'Methoxyflurane',
         'Enflurane',
         'DiethylEther',
         'CarbonDioxide',
         'Desflurane',
         'Halothane',
         'Sevoflurane',
         'NitrousOxide']}

name_for_cid[4214] = 'OphthalmicHorizontalDirections'
cid_concepts[4214] = \
{'SCT': ['Inward', 'Outward']}

name_for_cid[3494] = 'IVUSNonMorphologicalFindings'
cid_concepts[3494] = \
{'DCM': ['TrueLumen'],
 'SCT': ['ArterialBloodStasis',
         'AcquiredIncompleteStentApposition',
         'IncompleteStentApposition']}

name_for_cid[12205] = 'EchocardiographyLeftAtrium'
cid_concepts[12205] = \
{'LN': ['LeftAtriumAreaA4CView',
        'LeftAtriumToAorticRootRatio',
        'HeartRate',
        'LeftAtrialAppendagePeakVelocity',
        'LeftAtriumAnteroPosteriorSystolicDimension'],
 'SCT': ['LeftAtriumSystolicVolume']}

name_for_cid[6012] = 'CalcificationDistributionModifier'
cid_concepts[6012] = \
{'DCM': ['CalcificationsWithinAMass', 'CalcificationsOutsideOfAMass'],
 'SCT': ['SegmentalCalcificationDistribution',
         'LinearCalcificationDistribution',
         'RegionalCalcificationDistribution',
         'DiffuseCalcificationDistribution',
         'GroupedCalcificationDistribution']}

name_for_cid[3210] = 'SpeedofResponse'
cid_concepts[3210] = \
{'SCT': ['Normal', 'Blunted', 'Accentuated']}

name_for_cid[3640] = 'Hypertension'
cid_concepts[3640] = \
{'SCT': ['PulmonaryHypertension', 'SystemicArterialHypertension']}

name_for_cid[6157] = 'Vascularity'
cid_concepts[6157] = \
{'DCM': ['VascularityNotPresent',
         'VascularityNotAssessed',
         'VascularityPresentInLesion',
         'VascularityPresentImmediatelyAdjacentToLesion',
         'DiffuselyIncreasedVascularityInSurroundingTissue']}

name_for_cid[3705] = 'ChamberSize'
cid_concepts[3705] = \
{'SCT': ['MarkedlyEnlargedCardiacChamber',
         'MildlyEnlargedCardiacChamber',
         'AbnormallySmallCardiacChamber',
         'ModeratelyEnlargedCardiacChamber',
         'NormalSizeCardiacChamber']}

name_for_cid[20] = 'PatientOrientationModifier'
cid_concepts[20] = \
{'SCT': ['CurledUp',
         'SemiProne',
         'Anatomical',
         'Supine',
         'LateralDecubitus',
         'KneeChest',
         'Kneeling',
         'Lithotomy',
         'Prone',
         'Standing',
         'LeftLateralDecubitus',
         'StoopedOver',
         'RightLateralDecubitus',
         'InverseTrendelenburg',
         'Sitting',
         'Frog',
         'Trendelenburg',
         'Lordotic']}

name_for_cid[10003] = 'EquipmentPlaneIdentification'
cid_concepts[10003] = \
{'DCM': ['AllPlanes', 'PlaneA', 'PlaneB', 'SinglePlane']}

name_for_cid[4243] = 'OphthalmicQualityMetricType'
cid_concepts[4243] = \
{'DCM': ['StandardDeviationOfMeasurementsUsed', 'SignalToNoiseRatio']}

name_for_cid[223] = 'NormalRangeValues'
cid_concepts[223] = \
{'SCT': ['NormalRangeLowerLimit', 'NormalRangeUpperLimit']}

name_for_cid[7195] = 'FunctionSegmentationPropertyTypes'
cid_concepts[7195] = \
{'SCT': ['Perfusion']}

name_for_cid[7480] = 'Breed'
cid_concepts[7480] = \
{'SCT': ['CanastrãoPigBreed',
         'FrenchArdennaisHorseBreed',
         'CanadianDurocPigBreed',
         'CriouloHorseBreed',
         'AvetonouCattleBreed',
         'Spaniel',
         'KigeziCattleBreed',
         'HuazhongTwoEndBlackSatzelingPigBreed',
         'CoopworthSheepBreed',
         'RatTerrierDogBreed',
         'MixedBreedDog',
         'AustralianBrafordXZebuCattleBreed',
         'EskimoDog',
         'SouthMalawiZebuCattleBreed',
         'TropicalDairyCattleCattleBreed',
         'CorsicanGoatBreed',
         'ShropshireSheepBreed',
         'KirgizHorseBreed',
         'LaoshanPigBreed',
         'SanMartineroCattleBreed',
         'ClunForestSheepBreed',
         'PapillonDog',
         'HrbineckyCattleBreed',
         'SeshagaCattleBreed',
         'WirehairStyrianMountainDogBreed',
         'PerianganHorseBreed',
         'ScottishDeerhound',
         'MurgeseHorseBreed',
         'UmblacheryCattleBreed',
         'ToggenburgGoatBreed',
         'SarabiCattleBreed',
         'CalabrianHorseBreed',
         'Dalland020PigBreed',
         'GeorgianMountainCattleBreed',
         'GasaraCattleBreed',
         'StandardDachshundDogBreed',
         'AfghanHound',
         'BelgianHalfbloodHorseBreed',
         'MonarchPigBreed',
         'IstrianPramenkaSheepBreed',
         'ShilohShepherdDogBreed',
         'BaluchiSheepBreed',
         'RostovPigBreed',
         'NormanzuXZebuCattleBreed',
         'TaishuhHorseBreed',
         'CubanPasoHorseBreed',
         'SchwyzZeboidXZebuCattleBreed',
         'UralBlackPiedCattleBreed',
         'LonghornCowBreed',
         'ConnemaraPonyHorseBreed',
         'GermanSpanielDogBreed',
         'OuessantSheepBreed',
         'AustralianBrumbyHorseBreed',
         'NorthSwedishHorseHorseBreed',
         'DanishRedPiedCattleBreed',
         'EnglishHunterHorseBreed',
         'AmericanBuckskinHorseBreed',
         'ChinPigBreed',
         'SouthAfricanBrownSwissCattleBreed',
         'SmallEastAfricanZebuCattleBreed',
         'BraccoItalianoDogBreed',
         'NurasCattleBreed',
         'IlocosCattleBreed',
         'CriolloSheepBreed',
         'AmericanBerkshirePigBreed',
         'NorthSwedishTrotterHorseBreed',
         'AmericanLandracePigBreed',
         'LantangPigBreed',
         'MangalargaPaulistaHorseBreed',
         'JiaoxiPigBreed',
         'ItalianRedPiedCattleBreed',
         'WakwaCattleBreed',
         'YugoslavianHoundDogBreed',
         'MiniatureZebuCattleBreed',
         'IrishWolfhound',
         'SmoothFoxTerrierDogBreed',
         'GermanBerkshirePigBreed',
         'VyatkaHorseBreed',
         'BielleseSheepBreed',
         'OldFormatWelshCorgiDogBreed',
         'HinisHorseBreed',
         'PolledLincolnRedCattleBreed',
         'BeijingBlackPiedCattleBreed',
         'WirehairedPointingGriffonDog',
         'ChaparGoatBreed',
         'LöwchenDogBreed',
         'DairyZebuOfUberabaCattleBreed',
         'AmericanCockerSpanielDogBreed',
         'PhilippineNativeIlocosPigBreed',
         'TouabireSheepBreed',
         'KhurasaniZebuCattleBreed',
         'YujiangPigBreed',
         'HampshireSheepBreed',
         'RhoenschafSheepBreed',
         'LeopardCurDogBreed',
         'HeilongjiangSpottedPigBreed',
         'MeishanPigBreed',
         'IcelandicHorseBreed',
         'SiberianHuskie',
         'CachenaCattleBreed',
         'ThraceCattleBreed',
         'BritishWhiteCattleBreed',
         'SlovakWarmbloodHorseBreed',
         'DachshundMiniatureDogBreed',
         'BoarPowerPig59PigBreed',
         'OrobicaGoatBreed',
         'GlenOfImaalTerrierDogBreed',
         'AlpineChamoiseeGoatBreed',
         'SwaledaleSheepBreed',
         'KirhizDogBreed',
         'GentileDiPugliaSheepBreed',
         'BelgianLaekenDog',
         'TunisSheepBreed',
         'HunterHorseBreed',
         'SanPierrePigBreed',
         'KarakulSheepBreed',
         'AmericanPaintHorseBreed',
         'DamaraCattleBreed',
         'COOPELSO93CattleBreed',
         'KaghaniGoatBreed',
         'LipizzanerHorseBreed',
         'AzerbaijanZebuCattleBreed',
         'CasertanaPigBreed',
         'DobrogeaBlackPigBreed',
         'GermanWhiteheadedMuttonSheepBreed',
         'Line32PigBreed',
         'LucernaCattleBreed',
         'BassetArtésianNormandDogBreed',
         'LaguneCattleBreed',
         'YonaguniHorseBreed',
         'AostaRedPiedCattleBreed',
         'WenshanCattleBreed',
         'JapanesePointerDogBreed',
         'BritishBlueCatBreed',
         'NiloPigBreed',
         'PonwarCattleBreed',
         'SL96PigBreed',
         'DutchBeltedCattleBreed',
         'PomeranianDog',
         'VestlandRedPolledCattleBreed',
         'BrafordXZebuCattleBreed',
         'FineMerinoSheepBreed',
         'CavallinoPigBreed',
         'FinnhorseDraftHorseBreed',
         'PyreneanMastiffDogBreed',
         'LaManchaGoatBreed',
         'GasconCattleBreed',
         'AustralianTerrier',
         'BedlingtonTerrier',
         'NamibDesertHorseHorseBreed',
         'DutchDraftHorseBreed',
         'PartiColorCockerSpanielDogBreed',
         'BelgianGriffonRoughDogBreed',
         'CasanarenoCattleBreed',
         'JapaneseChinDog',
         'KemerovoPigBreed',
         'RussianHeavyDraftHorseBreed',
         'WestHighlandWhiteTerrier',
         'PolledSussexCattleBreed',
         'ShakhansurriCattleBreed',
         'GordonSetterDogBreed',
         'AuxoisHorseBreed',
         'CanadianLandracePigBreed',
         'PoitouMuleProducerHorseBreed',
         'IrishTerrier',
         'DingoDogBreed',
         'EastAndSoutheastAnadoluHorseBreed',
         'WhitebredShorthornCattleBreed',
         'LesserCaucasusCattleBreed',
         'ToroCattleBreed',
         'BasutoCattleBreed',
         'BrownSwissCowBreed',
         'BeagleElizabethanDogBreed',
         'BalkhiSheepBreed',
         'SuksunCattleBreed',
         'EnglishShepherdDogBreed',
         'IsraeliHolsteinCattleBreed',
         'SaintBernardLongHairedDogBreed',
         'YorkshireTerrier',
         'EpagneulPontAudemerDogBreed',
         'BalinesePigBreed',
         'MirgorodPigBreed',
         'BergerShetlandDogBreed',
         'IndoBrazilianCattleBreed',
         'CholistaniSheepBreed',
         'YananPigBreed',
         'CroatianRedCattleBreed',
         'KangayamCattleBreed',
         'CharolaisCattleBreed',
         'Terrier',
         'SarPlaninaSheepBreed',
         'TurkomanHorseBreed',
         'GoleCattleBreed',
         'KaranSwissXZebuCattleBreed',
         'SardinianCattleBreed',
         'EurolinePigBreed',
         'JutlandHorseBreed',
         'VerataGoatBreed',
         'KoreanImprovedPigBreed',
         'EdelschweinPigBreed',
         'PenArLan77PigBreed',
         'LimpurgerCattleBreed',
         'MurcianCattleBreed',
         'LokaiHorseBreed',
         'ColoredBullTerrierDogBreed',
         'EasternNuerCattleBreed',
         'BrahmaCowBreed',
         'GriffonFauveDeBretegneDogBreed',
         'MurcianaGranadinaGoatBreed',
         'AlistanaSanabresaCattleBreed',
         'MalteseCatBreed',
         'HerefordPigBreed',
         'EvolèneCattleBreed',
         'TropicalXZebuCattleBreed',
         'VerzascaGoatBreed',
         'CerbatHorseBreed',
         'CharnequeriaGoatBreed',
         'MandiPigBreed',
         'MagalCattleBreed',
         'PolesianPigBreed',
         'WelshSpringerSpanielDogBreed',
         'TerskHorseBreed',
         'AustralianMilkingZebuXZebuCattleBreed',
         'AriégeoisPonyHorseBreed',
         'MixedBreedChicken',
         'FleuveHorseBreed',
         'GermanLonghairedPointerDogBreed',
         'OropaCattleBreed',
         'PigghamPigBreed',
         'BrazilianSomaliSheepBreed',
         'AnadoluHorseBreed',
         'KoreanBlackCattleBreed',
         'CatalinaIslandPigBreed',
         'BergerDePyreneesDogBreed',
         'LonghairedManx',
         'EnderbyIslandCattleBreed',
         'AnatolianBlackCattleBreed',
         'ZnamenskCattleBreed',
         'SumbaHorseBreed',
         'XinjiangWhitePigBreed',
         'ToronkeCattleBreed',
         'EnglishToyTerrierDogBreed',
         'EastFriesianSheepBreed',
         'JemJemZebuCattleBreed',
         'BoarPowerPig747PigBreed',
         'BakosiPigBreed',
         'NamaCattleBreed',
         'HungarianGrayCattleBreed',
         'HissarCattleBreed',
         'EstonianHoundDogBreed',
         'CalabrianCattleBreed',
         'RedPollCowBreed',
         'Camborough15PigBreed',
         'PresticePigBreed',
         'PriobHorseBreed',
         'LohiSheepBreed',
         'ChartPolskiDogBreed',
         'BrazilianDairyHybridXZebuCattleBreed',
         'KleenLeenWhitePigBreed',
         'CabanninaCattleBreed',
         'TurkishBrownCattleBreed',
         'BritishAppaloosaHorseBreed',
         'KabyleCattleBreed',
         'DøleHorseBreed',
         'AmericanBobtailCatBreed',
         'VizslaSuperbreedDogBreed',
         'AztecaHorseBreed',
         'TransylvanianHoundDogBreed',
         'BlackForestHoundDogBreed',
         'SouthSuffolkSheepBreed',
         'HorroCattleBreed',
         'SprecaCattleBreed',
         'BragadoDoSorraiaCattleBreed',
         'LithuanianRedCattleBreed',
         'BelgianTervurenDog',
         'BreitovPigBreed',
         'RahvanHorseBreed',
         'CantonesePigBreed',
         'VietnamesePigBreed',
         'PhilippineNativePigBreed',
         'SwedishLapphundDogBreed',
         'AbyssinianShorthornedZebuCattleBreed',
         'JidduCattleBreed',
         'SouthRussianOvcharkaDogBreed',
         'CharaHorseHorseBreed',
         'SovietMeatPigBreed',
         'ChantillyTiffanyCatBreed',
         'MerinolandschafSheepBreed',
         'CaldeanoCattleBreed',
         'AmericanSaddlebredHorseBreed',
         'BardigianoHorseBreed',
         'ThuringianGoatBreed',
         'SpitiHorseBreed',
         'GrayAlpineCattleBreed',
         'AndalusianBlackCattleBreed',
         'BritishLopPigBreed',
         'AmericanEskimoToyDogBreed',
         'MarchigianaCowBreed',
         'YorkshireCoachHorseBreed',
         'AradoCattleBreed',
         'LleynSheepBreed',
         'GolpayeganiCattleBreed',
         'PinzhouCattleBreed',
         'ArmantDogBreed',
         'SmallWhitePigBreed',
         'PonyHorseBreed',
         'AngloFrancaisDePetitVenerieDogBreed',
         'RottalHorseBreed',
         'PixieBobCatBreed',
         'GuineaHogPigBreed',
         'CalasparrenaCattleBreed',
         'WestAfricanDwarfSheepBreed',
         'RätienGrayCattleBreed',
         'SinoVietnamesePigBreed',
         'RomanianLargeWhitePigBreed',
         'KhorsanCattleBreed',
         'CornishRexCatBreed',
         'KabardaHorseBreed',
         'KisantuCattleBreed',
         'JabeHorseBreed',
         'SelkirkRexCatBreed',
         'EnglishCoonhoundDogBreed',
         'BritishAlpineGoatBreed',
         'MazandaraniCattleBreed',
         'PagliarolaSheepBreed',
         'OldDanishBirdDogBreed',
         'BelgianBlueCattleBreed',
         'HattonXZebuCattleBreed',
         'NewfoundlandPonyHorseBreed',
         'RomanianBrownCattleBreed',
         'FinnishAyrshireCattleBreed',
         'LiangGuangSmallSpottedLuchuanPigBreed',
         'YemeniHorsesHorseBreed',
         'WallisCountrySheepBreed',
         'CotswoldGoldPigBreed',
         'MudiDogBreed',
         'BorneoZebuCattleBreed',
         'DBIPigBreed',
         'SchwerfurtMeatPigBreed',
         'ErlunchunHorseBreed',
         'BelgianWolfhoundDogBreed',
         'MukotaPigBreed',
         'GuzeratCattleBreed',
         'DerbyshireGritstoneSheepBreed',
         'CampanianBarbarySheepBreed',
         'SykiaCattleBreed',
         'ChifordCattleBreed',
         'KhevsurianCattleBreed',
         'ZanzibarZebuCattleBreed',
         'FloridaCrackerHorseBreed',
         'ModicanaLowlandCattleBreed',
         'BorderTerrier',
         'MixedBreedPig',
         'SwissWarmbloodHorseBreed',
         'GiantSchnauzerDogBreed',
         'NavajoChurroSheepBreed',
         'NorwichTerrier',
         'JamaicaHopeXZebuCattleBreed',
         'NewForestPonyHorseBreed',
         'ItalianGreyhound',
         'BlondeDAquitaineCowBreed',
         'SlovakianWhitePigBreed',
         'BelgiumLandracePigBreed',
         'CalascibettaPigBreed',
         'AdaevHorseBreed',
         'MiniatureSchnauzerDogBreed',
         'ChinoSantandereanoCattleBreed',
         'AmericanBulldogBreed',
         'RussianWhiteGoatBreed',
         'MixedBreedCattle',
         'BapediCattleBreed',
         'BeltsvillePigBreed',
         'CaspianHorseBreed',
         'PolishMarshPigBreed',
         'PoodleSuperbreedDogBreed',
         'ZlotnikiSpottedPigBreed',
         'RomanianNativeStocliPigBreed',
         'NelloreCowBreed',
         'GermanBlackheadedMuttonSheepBreed',
         'JindoDogBreed',
         'HainanTunchangPigBreed',
         'BelgianStreetDogBreed',
         'AustralianMistCatBreed',
         'AsturianValleyCattleBreed',
         'CãoDeCastroLaboreiroDogBreed',
         'ArouquesaCattleBreed',
         'DesFossesCommunesDeLOuestGoatBreed',
         'BoarPowerPig616PigBreed',
         'CiritHorseBreed',
         'CamargueCattleBreed',
         'LinebackCattleBreed',
         'DKPig61PigBreed',
         'BameiPigBreed',
         'CoPigBreed',
         'GambianNDamaXZebuCattleBreed',
         'YanbianCattleBreed',
         'TajmaCattleBreed',
         'DnepropetrovskPigBreed',
         'AbigarCattleBreed',
         'HeilongkaingHorseBreed',
         'LargePolishLongEaredPigBreed',
         'MurrayGrayCattleBreed',
         'HanoverianHoundDogBreed',
         'ResavaPigBreed',
         'AsturianTarpanHorseXDomesticHorseBreed',
         'SomaliCat',
         'PuertoRicanCriolloCattleBreed',
         'PerdigueiroPortugueseDogBreed',
         'NorwegianForestCatBreed',
         'AidiDogBreed',
         'DalmatianDog',
         'BassetFauveDeBretagneDogBreed',
         'RackaSheepBreed',
         'BlacksidedTrondheimAndNorlandCattleBreed',
         'BolivianPigBreed',
         'AmericanTrotterHorseBreed',
         'NorwegianLandracePigBreed',
         'UjumqinCattleBreed',
         'DamaniGoatBreed',
         'EnglishHackHorseBreed',
         'RyelandSheepBreed',
         'NorthSomaliZebuCattleBreed',
         'Cotswold90PigBreed',
         'BelgianMalinoisDog',
         'KushumHorseBreed',
         'PeacockGoatBreed',
         'ThaiBinhPigBreed',
         'TharparkarCattleBreed',
         'ChamoiseeGoatBreed',
         'PomeranianPigBreed',
         'KholmogoryCattleBreed',
         'DhanniCattleBreed',
         'EastMacedonianCattleBreed',
         'SwedishVallhundDogBreed',
         'BombayCat',
         'BlueTickCoonhoundDogBreed',
         'TanzanianZebuCattleBreed',
         'PindosPonyHorseBreed',
         'TawleedHorseBreed',
         'GuangxiWhitePigBreed',
         'ChienFrançaisBlancEtOrangeDogBreed',
         'PankotaRedCattleBreed',
         'SwedishFurSheepBreed',
         'TaraiCattleBreed',
         'BarbariGoatBreed',
         'KarachaiHorseBreed',
         'RaadPigBreed',
         'BretonBlackPiedCattleBreed',
         'CanastrãoVermelhoPigBreed',
         'MustangHorseBreed',
         'ItalianLandracePigBreed',
         'CzechoslovakianSmallRidingHorseHorseBreed',
         'PesterCattleBreed',
         'DanubianHorseBreed',
         'IbanPigBreed',
         'PalmeraCattleBreed',
         'LaviniaXZebuCattleBreed',
         'BurmeseCat',
         'CanadianSportHorseHorseBreed',
         'MalviCattleBreed',
         'FrenchGuardDogBreed',
         'HampshirePigBreed',
         'DevonRexCatBreed',
         'MaryutiCattleBreed',
         'NarymHorseBreed',
         'AlaTauCattleBreed',
         'GalicenoHorseBreed',
         'GüssingForestPigPigBreed',
         'KangarooIslandPigBreed',
         'KyogaCattleBreed',
         'OwczarekPodhalandskiDogBreed',
         'LundyPonyHorseBreed',
         'BosTaurusIndicusCowBreed',
         'NoniusHorseBreed',
         'WessexSaddlebackPigBreed',
         'MezenHorseBreed',
         'QinchuanCattleBreed',
         'MampatiCattleBreed',
         'BullTerrier',
         'SpanishAmericanHorseHorseBreed',
         'FrijolilloCattleBreed',
         'ZlotnikiWhitePigBreed',
         'WeisseHornloseHeidschnuckeSheepBreed',
         'XinjinJilinBlackPigBreed',
         'ChiTerrierDogBreed',
         'SonValleyCattleBreed',
         'ScottishHighlandCowBreed',
         'BenadirGoatBreed',
         'BoarPowerPig474PigBreed',
         'ChesterWhitePigBreed',
         'RazorBackPigBreed',
         'PhilippineNativeCattleBreed',
         'CanadianRoyalBluePigBreed',
         'HornedHerefordCowBreed',
         'CarpathianPonyHorseBreed',
         'DKPig77PigBreed',
         'NiloticSangaCattleBreed',
         'BoulonnaisHorseBreed',
         'ReniteloXZebuCattleBreed',
         'NorwegianRedCattleBreed',
         'CaucasianMountainDogBreed',
         'ChenghuaPigBreed',
         'BlancDeBouscatRabbitBreed',
         'BarkaZebuCattleBreed',
         'SicilianBarbarySheepBreed',
         'BeagleStandardDogBreed',
         'AdamawaCattleBreed',
         'BorderLeicesterSheepBreed',
         'TibetanSpaniel',
         'LincolnRedCowBreed',
         'AnkoleWatusiCattleBreed',
         'PekingeseDog',
         'CréolePigBreed',
         'StCroixSheepBreed',
         'KikoGoatBreed',
         'SouthRussianSteppeHoundDogBreed',
         'SpanishNormanHorseBreed',
         'FortCrossCattleBreed',
         'BorgouXZebuCattleBreed',
         'SpanishMastiffDogBreed',
         'BurlinaCattleBreed',
         'HillRadnorSheepBreed',
         'MastergiltPigBreed',
         'DRUTerminalsPigBreed',
         'ChineseCrestedDogBreed',
         'SchwyzCattleBreed',
         'SouthwestSpanishMustangHorseBreed',
         'CzechWarmbloodHorseBreed',
         'BulgarianNativePigBreed',
         'AmericanCashmereGoatBreed',
         'GriffonDog',
         'HangPigBreed',
         'SphynxCatBreed',
         'DongolaCattleBreed',
         'PolledNeloreCattleBreed',
         'DebouilletSheepBreed',
         'MerinoSheepSuperbreed',
         'HawaiianWildCattleBreed',
         'SaanenGoatBreed',
         'LonkSheepBreed',
         'BantuPigBreed',
         'HuazhongTwoEndBlackPigBreed',
         'KalakheriCattleBreed',
         'BlackBaldyCattleBreed',
         'ThanhHoaCattleBreed',
         'PindosHorseBreed',
         'HavanaBrownCat',
         'ChianinaCowBreed',
         'MixedDairyCattleBreed',
         'BadenWurttemburgHorseBreed',
         'BostonTerrier',
         'BlackFacedHighlandSheepBreed',
         'JersindXZebuCattleBreed',
         'SchnauzerSuperbreedDogBreed',
         'GroningenWhiteheadedCattleBreed',
         'BraqueDAriégeDogBreed',
         'DidingaCattleBreed',
         'KatahdinSheepBreed',
         'XinjiangFinewoolSheepBreed',
         'PembrokeWelshCorgiDogBreed',
         'HungarianDunHorseBreed',
         'PolishBlackAndWhiteLowlandCattleBreed',
         'NewZealandHeadingDogBreed',
         'SaidiCattleBreed',
         'BalikunHorseBreed',
         'PechoraCattleBreed',
         'MessaoriaCattleBreed',
         'AliabDinkaCattleBreed',
         'SkyrosPonyHorseBreed',
         'RoughAndSmoothDogBreed',
         'OliventinaPigBreed',
         'KrskopoljePigBreed',
         'AmericanLamaleseDogBreed',
         'EnglishSetterDogBreed',
         'PirapetingaPigBreed',
         'MinisibPigBreed',
         'FalabellaHorseBreed',
         'PatibarcinaHorseBreed',
         'SarvarHorseBreed',
         'VorderwaldCattleBreed',
         'AlpineDachsbrackeDogBreed',
         'MissouriFoxTrottingHorseBreed',
         'EnglishCobHorseBreed',
         'QingpingPigBreed',
         'LewitzerHorseBreed',
         'RendenaCattleBreed',
         'TeeswaterSheepBreed',
         'DomesticMediumHairedCat',
         'BashkirCurlyHorseBreed',
         'OldNorwegianSheepBreed',
         'PitangueirasXZebuCattleBreed',
         'DonGoatBreed',
         'DachshundSuperbreedOfDog',
         'FoxTerrierSuperbreedDogBreed',
         'PutianPigBreed',
         'GuanlingPigBreed',
         'PorcelaineDogBreed',
         'MiniaturePinscherDog',
         'JapaneseSpaniel',
         'SardoModicanaCattleBreed',
         'KamdhinoCattleBreed',
         'MongolianCattleBreed',
         'SinoGasconyPigBreed',
         'JerseyCowBreed',
         'ShkodraRedCattleBreed',
         'ToyPoodleDogBreed',
         'LatukaCattleBreed',
         'BulgarianWhitePigBreed',
         'SomaliSheepBreed',
         'KalmykCattleBreed',
         'AlentejanaCattlebreed',
         'MegezhHorseBreed',
         'KaiKenDogBreed',
         'DamascusCattleBreed',
         'BamiCattleBreed',
         'WalachenschafSheepBreed',
         'AngelnSaddlebackPigBreed',
         'ShetlandPonyHorseBreed',
         'FauveDeBretagneDogBreed',
         'CambridgeSheepBreed',
         'ModeneseCattleBreed',
         'RussianBlueCat',
         'GulfCoastNativeSheepBreed',
         'SouthDevonCattleBreed',
         'HolsteinerHorseBreed',
         'KenyaBoranCattleBreed',
         'EastFinnishCattleBreed',
         'TreeingCurDogBreed',
         'RyaSheepBreed',
         'MonchinaCattleBreed',
         'NigerianDwarfGoatBreed',
         'KunekunePigBreed',
         'GermanRedPiedPigBreed',
         'MarylandPigBreed',
         'ReggianaCattleBreed',
         'LurcherHoundDogBreed',
         'RomneletSheepBreed',
         'PantaneiroHorseBreed',
         'MarwariHorseBreed',
         'GuadarramaGoatBreed',
         'FonthillMerinoSheepBreed',
         'BordeauxDogBreed',
         'SlovakianPinzgauCattleBreed',
         'BoarPowerPig141PigBreed',
         'TielingHorseBreed',
         'BlackHampshirePigBreed',
         'SouthernCriouloCattleBreed',
         'QuarterPonyHorseBreed',
         'UkrainianGrayCattleBreed',
         'BraqueSaintGermainDogBreed',
         'SunandiniXZebuCattleBreed',
         'TarentaiseCowBreed',
         'NewHuaiPigBreed',
         'BritishShorthairedCat',
         'BlancaCeltibericaGoatBreed',
         'GratiCattleBreed',
         'GermanPinscherDogBreed',
         'NantaisCattleBreed',
         'CanadianCowBreed',
         'CuinoPigBreed',
         'AmericanTunisSheepBreed',
         'HokkaidoHorseBreed',
         'FellPonyHorseBreed',
         'MiKiDogBreed',
         'AnatolianShepherdDogBreed',
         'GrandGasconSaintongeoisDogBreed',
         'DutchKooikerDogBreed',
         'KoratCat',
         'PetitBleuDeGascogneDogBreed',
         'RedSindhiCattleBreed',
         'HungarofriesCattleBreed',
         'ToriHorseBreed',
         'FuriosoHorseBreed',
         'NepaleseHillZebuCattleBreed',
         'CyprusCattleBreed',
         'AmericanCurlCatBreed',
         'BarrosoCattleBreed',
         'TonkineseCat',
         'CharfordXZebuCattleBreed',
         'MuongKhuongPigBreed',
         'BelgianLandraceBNPigBreed',
         'KirdiPonyHorseBreed',
         'SomaliGoatBreed',
         'HogIslandSheepBreed',
         'SiberianBlackPiedPigBreed',
         'OldFormatMinToyPoodleDogBreed',
         'TyrolMountainSheepBreed',
         'MezzalinaCattleBreed',
         'GermanMountainSheepBreed',
         'GermanShorthornCattleBreed',
         'AinuDogBreed',
         'DashtiariCattleBreed',
         'MaduraWildJavanXZebuCattleBreed',
         'DutchTuigpaardHorseBreed',
         'WanzheSpottedChunanSpottedPigBreed',
         'MontafonCattleBreed',
         'FlatCoatedRetrieverDogBreed',
         'NorthernBlueCattleBreed',
         'LiberianDwarfCattleBreed',
         'MhaswadCattleBreed',
         'KyiApsoDogBreed',
         'SardaGoatBreed',
         'KuiMlkDogBreed',
         'VladimirHeavyDraftHorseBreed',
         'VestlandFjordCattleBreed',
         'DKPig30PigBreed',
         'LipetskPigBreed',
         'AostaChestnutCattleBreed',
         'ZhemaichuHorseBreed',
         'IberianBlackIberianPigBreed',
         'MagyarAgårDogBreed',
         'ShireHorseBreed',
         'RoyaVesubieGoatBreed',
         'AmericanCreamDraftHorseBreed',
         'NorfolkTerrier',
         'AustralianFriesianSahiwalXZebuCattleBreed',
         'AustralianStockHorseHorseBreed',
         'FHCElitePig7PigBreed',
         'ByelorussianSyntheticCattleBreed',
         'CanaryIslandCattleBreed',
         'NyoroCattleBreed',
         'SokolkaHorseBreed',
         'ShanghaiWhitePigBreed',
         'WanzheSpottedPigBreed',
         'KaninganCattleBreed',
         'TibetanMastiffDogBreed',
         'PuliDog',
         'StreiPigBreed',
         'SouthAnatolianRedCattleBreed',
         'LithuanianWhitePigBreed',
         'DatongHorseBreed',
         'EgyptianHorseBreed',
         'VolynskCattleBreed',
         'PolesianHorseBreed',
         'BriardDog',
         'MontanaraCattleBreed',
         'BuryatHorseBreed',
         'CamarronGoatBreed',
         'AmericanHairlessTerrierDogBreed',
         'StandardbredHorseBreed',
         'DanishLargeWhitePigBreed',
         'BaggaraCattleBreed',
         'NakaliCattleBreed',
         'PlevenHorseBreed',
         'SahfordXZebuCattleBreed',
         'GarganicaGoatBreed',
         'PoltavaPigBreed',
         'MurleCattleBreed',
         'CostenoHorseBreed',
         'CentralAsianZebuXZebuCattleBreed',
         'ManorMeishanPigBreed',
         'SpottedPigBreed',
         'SpanishGoatBreed',
         'DKPigBreed',
         'GallowayCowBreed',
         'DonPigBreed',
         'TurkishVanCatBreed',
         'KhamalaCattleBreed',
         'VosgesCattleBreed',
         'MeusseRhineIjsselCowBreed',
         'DanishOldenborgHorseBreed',
         'GermanRedPiedCattleBreed',
         'KedahKelantanCattleBreed',
         'FinnhorseHorseBreed',
         'DutchFriesianCattleBreed',
         'SouthAfricanMerinoSheepBreed',
         'BerkshirePigBreed',
         'BosnianHoundDogBreed',
         'RastreadorBrasileiroDogBreed',
         'MiniatureBullTerrierDogBreed',
         'AltaiHorseBreed',
         'DomesticLonghairedCat',
         'SkuddeSheepBreed',
         'MurcianPigBreed',
         'DamiettaCattleBreed',
         'NejdiCattleBreed',
         'AmericanBisonXCattleBreed',
         'SiberianPigBreed',
         'CanadianHampshirePigBreed',
         'BassetGriffonVendéenDogBreed',
         'BaXuyenPigBreed',
         'RockyMountainHorseHorseBreed',
         'PolledWelshBlackCattleBreed',
         'BeulahSpeckledFaceSheepBreed',
         'BelgianArdennaisHorseBreed',
         'GermanRidingPonyHorseBreed',
         'KoreanNativePigBreed',
         'YorkshireBlueAndWhitePigBreed',
         'AustralianKelpieDogBreed',
         'TuniCattleBreed',
         'KigerMustangHorseBreed',
         'LapermShorthairCatBreed',
         'EastFriesianWarmbloodModernTypeHorseBreed',
         'ValdresCattleBreed',
         'DrentsePatrijshondDogBreed',
         'WindsnyerPigBreed',
         'BoarPowerPig282PigBreed',
         'PiauSorocabaPigBreed',
         'IrishMoiledCattleBreed',
         'RioLimonDairyCriolloCattleBreed',
         'OxfordDownSheepBreed',
         'OldeEnglishBulldoggeDogBreed',
         'CanindéGoatBreed',
         'KarabakhHorseBreed',
         'AustrianYellowCattleBreed',
         'DairySyntheticCattleBreed',
         'TropicanaCattleBreed',
         'KomondorDog',
         'NiloCanastraPigBreed',
         'Keeshond',
         'GuernseyCowBreed',
         'CaiuaXZebuCattleBreed',
         'BuohaiHorseBreed',
         'DKPig63PigBreed',
         'HetaoLopEarPigBreed',
         'HolmongerCattleBreed',
         'RathCattleBreed',
         'BarzonaCattleBreed',
         'AmericanShetlandPonyHorseBreed',
         'SwissAlpineGoatBreed',
         'SokokeCatBreed',
         'ChineseKazakhHorseBreed',
         'JerdiXZebuCattleBreed',
         'LakenvelderCattleBreed',
         'IrishGoatBreed',
         'MirandesaCattleBreed',
         'DabieshanCattleBreed',
         'SwabianHallPigBreed',
         'PolledShorthornUSCattleBreed',
         'GalicianPigBreed',
         'GraueGehoernteHeidschnuckeSheepBreed',
         'SwedishWarmbloodHorseBreed',
         'OrkneySheepBreed',
         'UzunyaylaHorseBreed',
         'SlovakCuvakDogBreed',
         'GermanSpitzDogBreed',
         'PittIslandSheepBreed',
         'BeefSyntheticCattleBreed',
         'SiberianBlackPiedCattleBreed',
         'SalersCowBreed',
         'VoronezhCoachHorseHorseBreed',
         'EurasianDogBreed',
         'SwallowBeliedMangalitzaPigBreed',
         'KankrejCattleBreed',
         'HarbinWhitePigBreed',
         'AmericanYorkshirePigBreed',
         'LiangGuangSmallSpottedGuangdongSmallEarPigBreed',
         'SistaniCattleBreed',
         'LuciePigBreed',
         'Dalland030PigBreed',
         'GermanAngusCattleBreed',
         'BichonTeneriffeDogBreed',
         'YanqiHorseBreed',
         'ShagyaHorseBreed',
         'MadonieSicilianPigBreed',
         'BavarianForestSheepBreed',
         'NomaHorseBreed',
         'LeiqiongCattleBreed',
         'ColourpointShorthairedCat',
         'BlackMangalitsaPigBreed',
         'IstrianCattleBreed',
         'LuxiCattleBreed',
         'HinterwaldCattleBreed',
         'RussianBrownCattleBreed',
         'FHCElitePig2PigBreed',
         'AndalusianGrayCattleBreed',
         'CaliforniaSpangledCatBreed',
         'MinusinHorseBreed',
         'BSIPigBreed',
         'TriggHoundDogBreed',
         'JinhuaPigBreed',
         'BalearicHorseBreed',
         'IleDeFranceSheepBreed',
         'MandalongXZebuCattleBreed',
         'CarpatheGoatBreed',
         'DoranCattleBreed',
         'SpanishPointerDogBreed',
         'BakosiCattleBreed',
         'RedEngadineSheepBreed',
         'MorganHorseBreed',
         'BoXuPigBreed',
         'TurkishAngoraCat',
         'SouthdownSheepBreed',
         'AmritMahalZebuCattleBreed',
         'PolishDraftHorseBreed',
         'CeskyTerrierDogBreed',
         'KalmykHorseBreed',
         'JenubiCattleBreed',
         'RedKandhariCattleBreed',
         'YakutCattleBreed',
         'BelgianRedCattleBreed',
         'KarabairHorseBreed',
         'LincolnRedShorthornCattleBreed',
         'DoayoCattleBreed',
         'SardinianBrownCattleBreed',
         'AustralianPonyHorseBreed',
         'RagamuffinCatBreed',
         'WhiteGallowayCattleBreed',
         'AcehCattleBreed',
         'DexterCowBreed',
         'KachhiSheepBreed',
         'BaluchiHorseBreed',
         'GuzerandoXZebuCattleBreed',
         'LucanianPigBreed',
         'RideauArcottSheepBreed',
         'LonghairedMiniatureDachshundDogBreed',
         'AkitaDog',
         'BelgianRedPiedCattleBreed',
         'WhiteWelshCattleBreed',
         'TargheeSheepBreed',
         'XinjinXinjinPigBreed',
         'TroteEnGallopeHorseBreed',
         'SwissBlackPiedCattleBreed',
         'SulawesiHorseBreed',
         'MeixinPigBreed',
         'DianiPigBreed',
         'AngloFrancaisDeMoyenVenerieDogBreed',
         'LincolnSheepBreed',
         'AustralianGreyhoundDogBreed',
         'RomanianSaddleHorseHorseBreed',
         'EquusCaballusGmeliniHorseBreed',
         'MangalargaMarchadorHorseBreed',
         'BibrikSheepBreed',
         'CretanMountainCattleBreed',
         'BelgianGriffonSmoothDogBreed',
         'WensleydaleSheepBreed',
         'RedDesertCattleBreed',
         'PersianCat',
         'MassifCentralGoatBreed',
         'MurakozHorseBreed',
         'WestrainPigBreed',
         'NdaguXZebuCattleBreed',
         'HamplinePigBreed',
         'NanyangCattleBreed',
         'CharSwissCattleBreed',
         'KareloFinnishLaikaDogBreed',
         'NorwegianDunkerhoundDogBreed',
         'SilesianHorseBreed',
         'MurnauWerdenfelsCattleBreed',
         'MaureCattleBreed',
         'ErhulianPigBreed',
         'SemirechenskPigBreed',
         'FujianSmallPigBreed',
         'BejaCattleBreed',
         'ButanaCattleBreed',
         'ItalianHeavyDraftHorseBreed',
         'MinPigBreed',
         'LeeSungPigBreed',
         'WallisBlacknosedSheepBreed',
         'SaintBernardShortHairedDogBreed',
         'MalteseGoatBreed',
         'CastlemilkMooritSheepBreed',
         'FerrandaisCattleBreed',
         'ApulianPodolianCattleBreed',
         'SmallPolishPrickEaredPigBreed',
         'GorbatovRedCattleBreed',
         'DandieDinmontTerrier',
         'PortoAmboimCattleBreed',
         'DahePigBreed',
         'AmericanBlueGasconHoundDogBreed',
         'JapaneseBlackCattleBreed',
         'IranianZebuCattleBreed',
         'PolishRedCattleBreed',
         'OICPigBreed',
         'DamaraSheepBreed',
         'BorderCollieDogBreed',
         'KoninkTarpanHorseXDomesticHorseBreed',
         'BooroolaMerinoSheepBreed',
         'NorwegianGoatBreed',
         'MiyakoHorseBreed',
         'ArapawaIslandGoatBreed',
         'KaranFriesXZebuCattleBreed',
         'NewYorkRedPigBreed',
         'BündnerOberlandSheepBreed',
         'AlbanianHorseBreed',
         'BoarPowerPig161PigBreed',
         'GadoDaTerraCattleBreed',
         'PineywoodsPigBreed',
         'HessenHorseBreed',
         'PodengoCanarioDogBreed',
         'UzbekBlackGoatBreed',
         'Bulldog',
         'SabreXZebuCattleBreed',
         'SmoothCollieDogBreed',
         'ToposaCattleBreed',
         'GermanFleckViehCowBreed',
         'PrisheksninskPigBreed',
         'FHCElitePig6PigBreed',
         'SmoothHairedVizslaDogBreed',
         'PR1050PigBreed',
         'RomneyMarshSheepBreed',
         'CoburgerFuchsschafSheepBreed',
         'GoldenAmericanSaddlebredHorseBreed',
         'OulmesBlondCattleBreed',
         'DutchLandracePigBreed',
         'CarunchinhoPigBreed',
         'KonariCattleBreed',
         'SuboticaWhitePigBreed',
         'ArsiCattleBreed',
         'MurbodenCattleBreed',
         'KapsikiCattleBreed',
         'JagdterrierDogBreed',
         'DjermaHorseBreed',
         'Pudelpointer',
         'EriskayPonyHorseBreed',
         'BoarPowerPig414PigBreed',
         'LargeWhitePigBreed',
         'KeErQinHorseBreed',
         'MergellandSheepBreed',
         'SouthWalesMountainSheepBreed',
         'CapeHorseHorseBreed',
         'FreibergHorseBreed',
         'BoreraySheepBreed',
         'PunganurCattleBreed',
         'FrenchSpanielDogBreed',
         'PeruvianIncaOrchidDogBreed',
         'JielinHorseBreed',
         'PyreneanCattleBreed',
         'TribredPigBreed',
         'AmericanIndianDogBreed',
         'Greyhound',
         'CatahoulaLeopardDogBreed',
         'QuarterHorseBreed',
         'VenezuelanBlackPigBreed',
         'PalominoHorseBreed',
         'DorsetSheepSuperbreed',
         'SinclairMiniaturePigBreed',
         'EastAnatolianRedCattleBreed',
         'VichPigBreed',
         'MewatiCattleBreed',
         'AkhaPigBreed',
         'BarbadosSheepBreed',
         'ASCOBCockerSpanielDogBreed',
         'PicCambourghPigBreed',
         'KetekuXZebuCattleBreed',
         'BritishSaanenGoatBreed',
         'MixedBreedHorse',
         'Ocicat',
         'KerryCattleBreed',
         'CestrumParqui',
         'PharaohHound',
         'RamoGrandeCattleBreed',
         'IrishCobHorseBreed',
         'SmoothStandardDachshundDogBreed',
         'BraMaineXZebuCattleBreed',
         'BloodhoundSouthernHoundDogBreed',
         'LimousinCowBreed',
         'FinnishLandraceGoatBreed',
         'MarinhoaCattleBreed',
         'NorwegianElkhound',
         'BritishSaddlebackPigBreed',
         'DKPig31PigBreed',
         'BoarPowerPig292PigBreed',
         'AmericanWaterSpanielDogBreed',
         'YaroslavlCattleBreed',
         'GaolaoCattleBreed',
         'JinhuaDongyangPigBreed',
         'GarfagninaCattleBreed',
         'MBayarHorseBreed',
         'AustralianCattleDog',
         'JianchangHorseBreed',
         'TamworthPigBreed',
         'KookaSheepBreed',
         'ChernigovCattleBreed',
         'VillardDeLansCattleBreed',
         'FrenchBrownCattleBreed',
         'BarbetDogBreed',
         'GuadianaSpottedCattleBreed',
         'BengaliCattleBreed',
         'NDamaCattleBreed',
         'MarianasCattleBreed',
         'RangerXZebuCattleBreed',
         'NorfolkHornSheepBreed',
         'CatalanSheepdogBreed',
         'BaouleCattleBreed',
         'RomanianNativeBaltaretPigBreed',
         'BarrosaCattleBreed',
         'BerneseHoundDogBreed',
         'FHCElitePig8PigBreed',
         'EstrelaMountainDogBreed',
         'NovaScotiaDuckTollingRetrieverDogBreed',
         'CaptainCookerPigBreed',
         'CurraleiroCattleBreed',
         'MinhotaCattleBreed',
         'NooitgedachtHorseBreed',
         'QuarabHorseBreed',
         'SunkumaCattleBreed',
         'AmericanEssexPigBreed',
         'IrishWaterSpanielDogBreed',
         'GuoxiaHorseBreed',
         'MoravianLargeYorkshirePigBreed',
         'SpanishAngloArabHorseBreed',
         'HanzhongWhitePigBreed',
         'AraappaloosaHorseBreed',
         'SilkyTerrier',
         'ChianinaPigBreed',
         'KilisCattleBreed',
         'HezuoPigBreed',
         'HighlandPonyHorseBreed',
         'BelgianGroenendaelDog',
         'KleenLeenBlackPigBreed',
         'AwassiSheepBreed',
         'AmericanMerinoSheepBreed',
         'RussianBlackPiedCattleBreed',
         'UrzhumPigBreed',
         'SchapendoesDogBreed',
         'PanjeHorseBreed',
         'JapaneseRetrieverDogBreed',
         'MoxotóGoatBreed',
         'GermanShepherdDog',
         'StandardPoodleDogBreed',
         'BahimaCattleBreed',
         'MessaraHorseBreed',
         'NavajoSheepBreed',
         'BeijingBlackPigBreed',
         'AfricanaSheepBreed',
         'NewKirgizHorseBreed',
         'SavinjaGrayCattleBreed',
         'JellicutCattleBreed',
         'TadzhikZebuCattleBreed',
         'CarthusianHorseBreed',
         'ShetlandSheepdog',
         'JamaicaBlackXZebuCattleBreed',
         'FjordHorseBreed',
         'OtterHound',
         'CheurfaCattleBreed',
         'NormandieCowBreed',
         'SiskaPigBreed',
         'LobiCattleBreed',
         'TaiwanYellowCattleBreed',
         'KharkovPigBreed',
         'GuangxiHorseBreed',
         'ChineseBlackAndWhiteCattleBreed',
         'ShorthornCowBreed',
         'BlackSlavonianPigBreed',
         'AltaiMountainGoatBreed',
         'ComtoisHorseBreed',
         'TambovRedCattleBreed',
         'BlackCockerSpanielDogBreed',
         'RambouilletSheepBreed',
         'DomesticShorthairedCat',
         'MinzhuPigBreed',
         'HerdwickSheepBreed',
         'StyrianMountainDogBreed',
         'SingleFootingHorseHorseBreed',
         'SwedishRedPolledCattleBreed',
         'ChamDocCattleBreed',
         'XinjinNingAnPigBreed',
         'BraqueDAuvergneDogBreed',
         'LombokHorseBreed',
         'PhilippineGoatBreed',
         'SchleswigHorseBreed',
         'FellataCattleBreed',
         'CãoDeFilaMiguelDogBreed',
         'DutchWarmbloodHorseBreed',
         'LangheSheepBreed',
         'IcelandicSheepdogBreed',
         'CanadianRusticPonyHorseBreed',
         'SantaInêsSheepBreed',
         'PeruvianPasoHorseBreed',
         'KachchaSiriCattleBreed',
         'Cotswold29PigBreed',
         'HainanLingaoPigBreed',
         'NorthBangladeshGrayCattleBreed',
         'EnglishGoatBreed',
         'BirmanCat',
         'FightingBullCattleBreed',
         'HartónCattleBreed',
         'GermanKoolieDogBreed',
         'ManxameseCatBreed',
         'MertolengaCattleBreed',
         'TelomainDogBreed',
         'XiangxiBlackPigBreed',
         'ChinookDogBreed',
         'BritishMilkSheepBreed',
         'AfricangnusCattleBreed',
         'CalvanaCattleBreed',
         'BritishLandracePigBreed',
         'WhitefaceWoodlandSheepBreed',
         'CanastrãoCabanoPigBreed',
         'ToubouXZebuCattleBreed',
         'MinnesotaPigBreed',
         'KenanaCattleBreed',
         'PercheronHorseBreed',
         'DevonCattleBreed',
         'FuzhouBlackPigBreed',
         'OrlovTrotterHorseBreed',
         'AlbanianIllyrianCattleBreed',
         'BosnianHorseBreed',
         'TaiwanSmallRedPigBreed',
         'BashiCattleBreed',
         'AmericanTunisHorseBreed',
         'SlovenianBrownCattleBreed',
         'SouthAfricanMuttonMerinoSheepBreed',
         'CildirCattleBreed',
         'MaremmanaHorseBreed',
         'BhirumPonyHorseBreed',
         'NkoneCattleBreed',
         'ShetlandSheepBreed',
         'MiniatureAustralianShepherdDogBreed',
         'FriesianHorseBreed',
         'DnieperPigBreed',
         'LebaneseCattleBreed',
         'HaitiCreolePigBreed',
         'PolishSimmentalCattleBreed',
         'BanyoCattleBreed',
         'BraqueFrancaisDePetiteTailleDogBreed',
         'HuaiPigBreed',
         'TuliCattleBreed',
         'SopravissanaSheepBreed',
         'StrongWoolMerinoSheepBreed',
         'MacedonianPigBreed',
         'PaphosCattleBreed',
         'TamankaduwaCattleBreed',
         'BaliHorseBreed',
         'PoitouGoatBreed',
         'WestAfricanBarbHorseBreed',
         'DanishRedCattleBreed',
         'LeineschafSheepBreed',
         'SimbrahCattleBreed',
         'SealyhamTerrier',
         'GirCowBreed',
         'PerendaleSheepBreed',
         'AltaySheepBreed',
         'NoTailSheepBreed',
         'Line500DurocPigBreed',
         'GermanPasturePigBreed',
         'MoghaniSheepBreed',
         'MinPeiDogBreed',
         'LepingPigBreed',
         'RojhanCattleBreed',
         'SandanHorseBreed',
         'ManorHybridPigBreed',
         'GermanCornwallPigBreed',
         'ToyHavaneseTerrierDogBreed',
         'AbyssinianCat',
         'KisoHorseBreed',
         'BraqueDupuyDogBreed',
         'AmericanCoonhoundDogBreed',
         'SharPeiDogBreed',
         'BritishWarmbloodHorseBreed',
         'MaineAnjouCowBreed',
         'CreoleAntillesGoatBreed',
         'GalwaySheepBreed',
         'DorperSheepBreed',
         'SloughiDogBreed',
         'PolishLandracePigBreed',
         'CarpathianBrownCattleBreed',
         'HamrocPigBreed',
         'SpinoneItalianoDogBreed',
         'DutchShepherdDogBreed',
         'GrandBleuDeGascogneDogBreed',
         'DroughtmasterXZebuCattleBreed',
         'FHCPigBreed',
         'RedSteppeCattleBreed',
         'ChienFrançaisTricoloreDogBreed',
         'SteigarSheepBreed',
         'CorsicanPigBreed',
         'ToyFoxTerrierDogBreed',
         'PieRougeDeLEstCattleBreed',
         'SuisbuXZebuCattleBreed',
         'HondurasSwitchTailPigBreed',
         'SonkheriCattleBreed',
         'EastBalkanPigBreed',
         'KochiCattleBreed',
         'CanastraPigBreed',
         'SanfratelloHorseBreed',
         'MeoPigBreed',
         'OldSwedishSpottedPigBreed',
         'KoronadalPigBreed',
         'BetizuakCattleBreed',
         'AmericanBobtailLonghairCatBreed',
         'TibetanCattleBreed',
         'WelshMountainPonyHorseBreed',
         'BambaraXZebuCattleBreed',
         'TexelSheepBreed',
         'TokaraHorseBreed',
         'FinnishHoundDogBreed',
         'TibetanPigBreed',
         'AlentejanaPigBreed',
         'RedWattlePigBreed',
         'CardiganWelshCorgiDogBreed',
         'KhillariCattleBreed',
         'ImprovedRodopiCattleBreed',
         'ArmoricanCattleBreed',
         'MurgheseHorseBreed',
         'AubracCattleBreed',
         'OhminiPigBreed',
         'DajalCattleBreed',
         'BraviaGoatBreed',
         'BlackheadPersianSheepBreed',
         'HavaneseDogBreed',
         'GoldenRetrieverDogBreed',
         'RexCatBreed',
         'CheviotSheepBreed',
         'MetohijaRedCattleBreed',
         'BalwenWelshMountainSheepBreed',
         'PortugueseGuardDogBreed',
         'AlpineGoatBreed',
         'SpanishBarbHorseBreed',
         'BusaCattleBreed',
         'AnkoleCattleBreed',
         'CosentinaPigBreed',
         'BulgarianRedCattleBreed',
         'BatanesBlackXZebuCattleBreed',
         'BoarPowerPig84PigBreed',
         'CharbrayXZebuCattleBreed',
         'ThalliSheepBreed',
         'SengologaCattleBreed',
         'TswanaCattleBreed',
         'HuazhongTwoEndBlackJianliPigBreed',
         'GreatDaneDog',
         'MouflonSheepBreed',
         'WestAfricanDwarfShorthornCattleBreed',
         'MontdaleSheepBreed',
         'LargeBlackWhitePigBreed',
         'BashkirHorseBreed',
         'ElliottdaleSheepBreed',
         'NebelungCatBreed',
         'LundehundDogBreed',
         'SienaBeltedPigBreed',
         'KrolevetsPigBreed',
         'BouvierDesFlandres',
         'AustrianSmoothHairedBrackeDogBreed',
         'BlancoOrejinegroCattleBreed',
         'HanfordMiniaturePigBreed',
         'BulgarianLandracePigBreed',
         'AzoresHorseBreed',
         'TibetanHorseBreed',
         'HerelandCattleBreed',
         'XinjiangBrownCattleBreed',
         'ColoradoRangerHorseBreed',
         'AmericanIndianHorseHorseBreed',
         'ManchesterTerrierSuperbreedDogBreed',
         'ExmoorHornSheepBreed',
         'AustralianShepherdDogBreed',
         'NanjingBlackPigBreed',
         'JinnanCattleBreed',
         'NortheasternHorseBreed',
         'PortuguesePointerDogBreed',
         'OutaouaisArcottSheepBreed',
         'TavdaHorseBreed',
         'BoarPowerPig48PigBreed',
         'ChumyshHorseBreed',
         'HellenicHoundDogBreed',
         'DjakoreXZebuCattleBreed',
         'KinderGoatBreed',
         'AustralianShorthornCattleBreed',
         'LadakhiCattleBreed',
         'RagdollCatBreed',
         'AnkinaCattleBreed',
         'GidranHorseBreed',
         'HackneyHorseBreed',
         'AriégeoisDogBreed',
         'RathiCattleBreed',
         'WireFoxTerrierDogBreed',
         'BazadaisCattleBreed',
         'YimengBlackPigBreed',
         'BatardDogBreed',
         'CzechMiniaturePigBreed',
         'BlackGrisonneGoatBreed',
         'MingrelianRedCattleBreed',
         'GreatMünsterländerDogBreed',
         'PrianganSheepBreed',
         'GlanCattleBreed',
         'ShortCoatChihuahuaDogBreed',
         'BeagleSmoothDogBreed',
         'BeltedGallowayCowBreed',
         'ChaouiaCattleBreed',
         'SardinianAngloArabHorseBreed',
         'SuffolkSheepBreed',
         'DartmoorSheepBreed',
         'CharollandaisCattleBreed',
         'AlanoEspañolDogBreed',
         'AmericanWhiteParkCattleBreed',
         'Weimaraner',
         'DanishDurocPigBreed',
         'TrakehnerHorseBreed',
         'SheepBreedUndeterminedSheepBreed',
         'BeefmakerCattleBreed',
         'GromarkSheepBreed',
         'DunkerDogBreed',
         'LhasaApso',
         'NormanCoachHorseBreed',
         'PelibüeySheepBreed',
         'IllawarraCattleBreed',
         'BavendaCattleBreed',
         'BelgianGriffonDogBreed',
         'KolbroekPigBreed',
         'RodopiCattleBreed',
         'ManorRangerPigBreed',
         'SouthernUkranianCattleBreed',
         'YorkshirePigBreed',
         'FHCElitePig4PigBreed',
         'MastiffDog',
         'SlovenianWhitePigBreed',
         'JavaneseCat',
         'FellTerrierDogBreed',
         'ChiangusCattleBreed',
         'MiddleWhitePigBreed',
         'BeagleHarrierDogBreed',
         'KenkathaCattleBreed',
         'CorriedaleSheepBreed',
         'ByelorussianRedCattleBreed',
         'LimousinPigBreed',
         'ArgentineDogoDogBreed',
         'Bullmastiff',
         'PolledHerefordCowBreed',
         'AmericanWirehairedCat',
         'CarinthianBlondCattleBreed',
         'EnglishSpringerSpanielDogBreed',
         'WujinPigBreed',
         'SpælsauSheepBreed',
         'PygoraGoatBreed',
         'TuvaHorseBreed',
         'UdaSheepBreed',
         'SiboneyXZebuCattleBreed',
         'MiddleAsianOvtcharkaDogBreed',
         'LatvianBlueRoanCattleBreed',
         'KakhetianPigBreed',
         'SerpentinaGoatBreed',
         'GermanRexCatBreed',
         'HaflingerHorseBreed',
         'BentheimBlackPiedPigBreed',
         'NorthAmericanShepherdDogBreed',
         'MontbeliardCattleBreed',
         'ChiloteHorseBreed',
         'SouthAfricanLandracePigBreed',
         'MilkingDevonCattleBreed',
         'PonyOfTheAmericasHorseBreed',
         'HanSheepBreed',
         'LandracePigBreed',
         'GreekLocalPigBreed',
         'MeatlinePigBreed',
         'NationalShowHorseHorseBreed',
         'AostaBlackPiedCattleBreed',
         'VietnameseYellowCattleBreed',
         'NorwegianLundehundDogBreed',
         'LonglinPigBreed',
         'BasutoPonyHorseBreed',
         'BraqueDuBourbonnaisDogBreed',
         'MilkingShorthornCowBreed',
         'AustralianTerrierSilkyDogBreed',
         'WelshPonyHorseBreed',
         'DobermanPinscher',
         'AustralianTerrierRoughCoatedDogBreed',
         'RhinelandHeavyDraftHorseBreed',
         'NGabouCattleBreed',
         'OvamboCattleBreed',
         'ThuocNhieuPigBreed',
         'RomagnolaPigBreed',
         'INRA95CattleBreed',
         'LongCoatChihuahuaDogBreed',
         'RedAngusCattleBreed',
         'KleenLeenPigBreed',
         'HungairanColdbloodHorseBreed',
         'BoarPowerPig454PigBreed',
         'ModicanaCattleBreed',
         'AmericanQuarterHorseHorseBreed',
         'Manx',
         'VendéenSheepBreed',
         'FumatiPigBreed',
         'OldFormatManchesterTerrierDogBreed',
         'LonghairedStandardDachshundDogBreed',
         'AmericanDraftPonyHorseBreed',
         'VlaamperdHorseBreed',
         'IraqiXZebuCattleBreed',
         'CarazebuXZebuCattleBreed',
         'MongCaiPigBreed',
         'PetitBassetGriffonVendeenDog',
         'NooitgedachtPonyHorseBreed',
         'PolledSimmentalCattleBreed',
         'BeefShorthornCattleBreed',
         'SnowshoeCatBreed',
         'GermanShorthairedPointerDogBreed',
         'RedFlemishCattleBreed',
         'RoveGoatBreed',
         'GujamavuCattleBreed',
         'SovereignPigBreed',
         'NorthEastChinaSpottedPigBreed',
         'PiauPigBreed',
         'KazakhHorseBreed',
         'KazakhWhiteheadCattleBreed',
         'BushuevXZebuCattleBreed',
         'HashtNagriSheepBreed',
         'KamanPigBreed',
         'JapaneseShorthornCattleBreed',
         'BritishSpottedPonyHorseBreed',
         'TudancaCattleBreed',
         'FHCElitePig5PigBreed',
         'LandaisHorseBreed',
         'BonsmaraCattleBreed',
         'AkhalTekeHorseBreed',
         'ColumbiaSheepBreed',
         'BKB1PigBreed',
         'CanaanDogBreed',
         'RedPolledØstlandCattleBreed',
         'MixedPerijaneroXZebuCattleBreed',
         'HarrierDog',
         'GermanBlackPiedDairyCattleBreed',
         'FinnishLandraceSheepBreed',
         'ArapawaIslandPigBreed',
         'AlmanzorenaCattleBreed',
         'ChanniCattleBreed',
         'WaiChowPigBreed',
         'EnglishFoxhound',
         'PantaneiroCattleBreed',
         'JezerskosolcavskaSheepBreed',
         'AchhamCattleBreed',
         'GirgentanaGoatBreed',
         'SRDGoatBreed',
         'McnabDogBreed',
         'DairyShorthornCattleBreed',
         'MpwapwaCattleBreed',
         'BlackIberianCattleBreed',
         'GhanaShorthornCattleBreed',
         'YunnanHorseBreed',
         'StaffordshireBullTerrierDogBreed',
         'CriolloHorseBreed',
         'SteppeMeatPigBreed',
         'IrishHunterHorseBreed',
         'KameroenSheepBreed',
         'SinhalaCattleBreed',
         'ParthenaisCattleBreed',
         'FrenchSaddlePonyHorseBreed',
         'HackneyPonyHorseBreed',
         'MongolianHorseBreed',
         'DamaniSheepBreed',
         'BologneseDogBreed',
         'TungchangPigBreed',
         'GarranoTarpanHorseXDomesticHorseBreed',
         'DalesbredSheepBreed',
         'GurktalPigBreed',
         'JapaneseBrownCattleBreed',
         'DKPig33PigBreed',
         'FengjingPigBreed',
         'YemeniZebuCattleBreed',
         'JianliPigBreed',
         'PontremoleseCattleBreed',
         'WesternSudanPonyHorseBreed',
         'ScottishFoldCat',
         'KurganCattleBreed',
         'SussexSpanielDogBreed',
         'BelgianBlackPiedHolsteinCattleBreed',
         'AksaîBlackPiedPigBreed',
         'GirolandoXZebuCattleBreed',
         'PodengoPequenoDogBreed',
         'BraqueFrançaisDeGrandTailleDogBreed',
         'TurkishGraySteppeCattleBreed',
         'BrazilianPolledCattleBreed',
         'CanaryIslandGoatBreed',
         'ArgentinePoloPonyHorseBreed',
         'LoenbergerDogBreed',
         'EstonianBlackPiedCattleBreed',
         'CubanTrotterHorseBreed',
         'MunichMiniaturePigBreed',
         'ChittagongRedCattleBreed',
         'SocotraCattleBreed',
         'DelibozHorseBreed',
         'SalornCattleBreed',
         'CharollaisSheepBreed',
         'NingxiangPigBreed',
         'DanishLandraceSheepBreed',
         'CeskyFousekDogBreed',
         'FloresHorseBreed',
         'PetisoArgentinoHorseBreed',
         'ProvençaleGoatBreed',
         'KathiawariHorseBreed',
         'AmericanPonyHorseBreed',
         'PoitevinDogBreed',
         'PinziritaSheepBreed',
         'KateriniCattleBreed',
         'KinniyaCattleBreed',
         'SiniHorseBreed',
         'PalousePigBreed',
         'CanastrãoZabumbaPigBreed',
         'YuedongBlackPigBreed',
         'TyroleanHoundDogBreed',
         'NguniCattleBreed',
         'MantiqueiraXZebuCattleBreed',
         'Retriever',
         'RussianShortEaredWhitePigBreed',
         'BrandroodLjsselveeCattleBreed',
         'GackoCattleBreed',
         'ChihuahuaSuperbreedDogBreed',
         'BizanianHoundDogBreed',
         'JiaxingBlackPigBreed',
         'MacedonianBusaCattleBreed',
         'GrisonsStripedGoatBreed',
         'AngelnCattleBreed',
         'AtpadiMahalCattleBreed',
         'MasaiCowBreed',
         'IskarCattleBreed',
         'SmallMünsterländerDogBreed',
         'MenufiCattleBreed',
         'BavarianWarmbloodHorseBreed',
         'CampolinaHorseBreed',
         'GreekShorthornCattleBreed',
         'VaynolCattleBreed',
         'HamiltonstövareDogBreed',
         'EnglishLonghornCattleBreed',
         'WestAfricanPigBreed',
         'DexterKerryCattleBreed',
         'NeloreCattleBreed',
         'DanishWarmbloodHorseBreed',
         'PernaCurtaPigBreed',
         'MalnadGiddaCattleBreed',
         'InkukuCattleBreed',
         'IrishRedAndWhiteSetterDogBreed',
         'ColNoirDuValaisGoatBreed',
         'Coonhound',
         'SuperfineMerinoSheepBreed',
         'JackRusselTerrier',
         'PolledGuzeratCattleBreed',
         'Camborough12PigBreed',
         'AmericanPitBullTerrierDogBreed',
         'AlaskanKleeKaiDogBreed',
         'BraqueBelgeDogBreed',
         'SmoothMiniatureDachshundDogBreed',
         'Setter',
         'MalopolskiHorseBreed',
         'AberdeenAngusCowBreed',
         'FinnishYorkshirePigBreed',
         'WokalupXZebuCattleBreed',
         'MexicanHairlessDog',
         'DreverDog',
         'Minnesota4PigBreed',
         'AlterRealHorseBreed',
         'HuangHuaiHaiBlackPigBreed',
         'HanoverianHorseBreed',
         'SelleFrancaisHorseBreed',
         'LibyanCattleBreed',
         'IsraeliRedCattleBreed',
         'RustaqiXZebuCattleBreed',
         'TexonCattleBreed',
         'SwedishArdennesHorseBreed',
         'BarbadosBlackbellySheepBreed',
         'CaucasianShepherdDogBreed',
         'AmericanShorthairCat',
         'PolishHoundDogBreed',
         'SychevkaCattleBreed',
         'BovskaSheepBreed',
         'DelaineMerinoSheepBreed',
         'PugDog',
         'ZambiaAngoniCattleBreed',
         'RedBororoCattleBreed',
         'IberianJabugoSpottedPigBreed',
         'LanYuPigBreed',
         'MalagueñaGoatBreed',
         'DraftPonySuperbreedHorseBreed',
         'RussoLaikaDogBreed',
         'GayalCattleBreed',
         'SchoonebekerSheepBreed',
         'WestSiberianLaikaDogBreed',
         'AvilenaBlackIberianCattleBreed',
         'RasaAragonesaSheepBreed',
         'TennesseeWalkingHorseBreed',
         'SwissMountainDogBreed',
         'CretanLowlandCattleBreed',
         'KustanaiHorseBreed',
         'BagotGoatBreed',
         'FHCElitePig9PigBreed',
         'CukurovaHorseBreed',
         'SuffolkHorseBreed',
         'GrandGriffonVendeenDogBreed',
         'SenepolCattleBreed',
         'StBernardDog',
         'FriesianMilkSheepBreed',
         'KingShepherdDogBreed',
         'SwissImprovedLandracePigBreed',
         'LaconiePigBreed',
         'TexasLonghornCattleBreed',
         'PortugueseWaterDog',
         'TagilCattleBreed',
         'BanatWhitePigBreed',
         'SardinianPigBreed',
         'CotonDeTuléarDogBreed',
         'NagoriCattleBreed',
         'DnieperCattleBreed',
         'CostaRicanSaddleHorseHorseBreed',
         'ChileanCorraleroHorseBreed',
         'YunnanZebuCattleBreed',
         'BlackHairlessPigBreed',
         'WelshCorgiSuperbreedDogBreed',
         'WitrikCattleBreed',
         'KromfohrländerDogBreed',
         'ÍPigBreed',
         'BlackMouthCurDogBreed',
         'BraSwissXZebuCattleBreed',
         'BankerHorseHorseBreed',
         'SmithfieldDogBreed',
         'GorynCattleBreed',
         'BhujGoatBreed',
         'FjallCattleBreed',
         'ThibarXZebuCattleBreed',
         'RussianSaddleHorseHorseBreed',
         'PosavinaCattleBreed',
         'MaremmanaCattleBreed',
         'HebrideanSheepBreed',
         'BisonBisonXSimmentalHybrid',
         'BoarPowerPig27PigBreed',
         'TahltanBearDog',
         'MontmorillonPigBreed',
         'TuyHoaHairlessPigBreed',
         'LivnyPigBreed',
         'LlaneroHorseBreed',
         'TsivilskPigBreed',
         'DongolaHorseBreed',
         'VogelsbergCattleBreed',
         'NDamaSangaCattleBreed',
         'BravonXZebuCattleBreed',
         'RanaXZebuCattleBreed',
         'SpanishColonialHorseHorseBreed',
         'NorthCountyCheviotSheepBreed',
         'LijiangHorseBreed',
         'MouraPigBreed',
         'TuxZillertalCattleBreed',
         'MerensHorseBreed',
         'PR1075PigBreed',
         'WirehairedMiniatureDachshundDogBreed',
         'NorthFujianBlackAndWhitePigBreed',
         'ErzgebirgGoatBreed',
         'BrazilianSportHorseHorseBreed',
         'KurdiCattleBreed',
         'SableIslandHorseHorseBreed',
         'EpagneulPicardDogBreed',
         'JamaicaBrahmanCattleBreed',
         'AngloKabardaHorseBreed',
         'WelshTerrier',
         'WelshMountainBadgerFacedSheepBreed',
         'DutchYorkshirePigBreed',
         'JavaneseZebuCattleBreed',
         'UkrainianSteppeWhitePigBreed',
         'SouthChinaPigBreed',
         'AfricanderCattleBreed',
         'SouthernBlackmouthCurDogBreed',
         'BrillenschafSheepBreed',
         'BamaxiangPigBreed',
         'HallikarCattleBreed',
         'ChineseImperialChInDogBreed',
         'LusitanoHorseBreed',
         'BralersXZebuCattleBreed',
         'HungarianWhitePigBreed',
         'HeidschnuckeSheepBreed',
         'SovietHeavyDraftHorseBreed',
         'PeeWeeCattleBreed',
         'PyreneanGoatBreed',
         'RomosinuanoCattleBreed',
         'SahelTypeSheepBreed',
         'DortyolCattleBreed',
         'AmericanCreamHorseBreed',
         'GreaterSwissMountainHoundDogBreed',
         'ManxLoaghtanSheepBreed',
         'HovawartDogBreed',
         'TurkmenZebuCattleBreed',
         'KabotaCattleBreed',
         'SaddlebackPigBreed',
         'LargeSpanishHoundDogBreed',
         'BestuzhevCattleBreed',
         'SterlingPigBreed',
         'TyrolGrayCattleBreed',
         'VitaVetLabMinipigPigBreed',
         'LourdaisCattleBreed',
         'Foxhound',
         'SenneHorseBreed',
         'CanadianEskimoDogBreed',
         'Wolf',
         'HormelMiniaturePigBreed',
         'MixedBreedGoat',
         'RedFriesianCattleBreed',
         'HarlequinCatBreed',
         'ZavotCattleBreed',
         'BündnerGoatBreed',
         'KappiliyanCattleBreed',
         'CPFPigBreed',
         'DingPigBreed',
         'MaronesaCattleBreed',
         'DKPig51PigBreed',
         'BlackWelshCowBreed',
         'ArapawaIslandSheepBreed',
         'WhippetDog',
         'WhiteParkCowBreed',
         'ArabiSheepBreed',
         'GreenlandDogBreed',
         'AngloNubianGoatBreed',
         'FogeraCattleBreed',
         'GroningenHorseBreed',
         'LincolnLongwoolSheepBreed',
         'RussianTrotterHorseBreed',
         'AgeroleseCattleBreed',
         'WhitefaceDartmoorSheepBreed',
         'OssabawIslandPigBreed',
         'CavalierKingCharlesSpaniel',
         'BelmontAdaptaurCattleBreed',
         'MountainCurDogBreed',
         'BlackAndTanCoonhoundDogBreed',
         'LacombePigBreed',
         'AustrianPinscherDogBreed',
         'OrehlaDeColherPigBreed',
         'MadaripurCattleBreed',
         'KostromaCattleBreed',
         'SchwarzwalderFuchseHorseBreed',
         'HuertanaCattleBreed',
         'RoughFellSheepBreed',
         'NkediCattleBreed',
         'XilingolHorseBreed',
         'MorochucoHorseBreed',
         'TukidaleSheepBreed',
         'BariCattleBreed',
         'Cotswold16PigBreed',
         'CatalanaChickenBreed',
         'PerdigueroNavarroDogBreed',
         'CukurovaCattleBreed',
         'HyperLargeWhitePigBreed',
         'MixedBreedSheep',
         'NeijiangPigBreed',
         'BergerDePicardDogBreed',
         'StandardManchesterTerrier',
         'SayaguesaCattleBreed',
         'BeardedCollieDogBreed',
         'BlondeDuCapBonCattleBreed',
         'CzechImprovedWhitePigBreed',
         'ChesiCattleBreed',
         'SiameseCat',
         'DonHorseBreed',
         'ArdennesHorseBreed',
         'CaliforniaRedSheepBreed',
         'HornDorsetSheepBreed',
         'CorrienteCattleBreed',
         'HuculHorseBreed',
         'UjumqinSheepBreed',
         'SussexCattleBreed',
         'YurinoCattleBreed',
         'BargurCattleBreed',
         'KaramajongCattleBreed',
         'BasqueBlackPiedPigBreed',
         'BergerDuLanguedocDogBreed',
         'DevonClosewoolSheepBreed',
         'BorghigianaPigBreed',
         'MinorcanCattleBreed',
         'TaiwanZebuCattleBreed',
         'ColombianCriolloHorseBreed',
         'PremierPigBreed',
         'LongyouBlackPigBreed',
         'JinhuaYongkangPigBreed',
         'PulMBorCattleBreed',
         'OrientalShorthairedCat',
         'ThaiCattleBreed',
         'WhiteBullTerrierDogBreed',
         'SandalwoodHorseBreed',
         'GujaratiCowBreed',
         'VikingHorseBreed',
         'GöttingenMiniaturePigBreed',
         'TiaMeslanPigBreed',
         'EurasierDogBreed',
         'BrazilianShorthairCatBreed',
         'MyotonicGoatBreed',
         'BalticBlackPiedCattleBreed',
         'RoughCollieDogBreed',
         'GuelmaCattleBreed',
         'OldenburgHorseBreed',
         'NorthernArdennaisHorseBreed',
         'GermanMuttonMerinoSheepBreed',
         'BritishToggenburgGoatBreed',
         'TrakyaHorseBreed',
         'ClawnPigBreed',
         'CanchimXZebuCattleBreed',
         'WhiteCáceresCattleBreed',
         'SumadijaPigBreed',
         'KisberFelverHorseBreed',
         'HighConformationWhitePigBreed',
         'KumamotoCattleBreed',
         'AndalusianHorseBreed',
         'GrandBassettGriffonVendeenDogBreed',
         'ThessalyCattleBreed',
         'HalhinGolCattleBreed',
         'GotlandHorseBreed',
         'BrownWelshCowBreed',
         'EastSiberianLaikaDogBreed',
         'SyntheticSL98PigBreed',
         'LoashanGoatBreed',
         'CubanZebuXZebuCattleBreed',
         'KaiDogBreed',
         'MytileneHorseBreed',
         'BraunviehCattleBreed',
         'LeicesterSheepBreed',
         'WelshCobHorseBreed',
         'TongSheepBreed',
         'TarinaCattleBreed',
         'HolsteinFriesianCowBreed',
         'PatterdaleTerrierDogBreed',
         'AmericanBobtailShorthairCatBreed',
         'ToyManchesterTerrier',
         'KladrubyHorseBreed',
         'DagestanMountainCattleBreed',
         'PetitGriffonBleuDeGascogneDogBreed',
         'TongConPigBreed',
         'MangalistaPigBreed',
         'SahiwalCattleBreed',
         'GelbrayXZebuCattleBreed',
         'WaziriSheepBreed',
         'PietrainPigBreed',
         'ThaiPonyHorseBreed',
         'AzawakhDogBreed',
         'KumauniCattleBreed',
         'ManagraPigBreed',
         'PresaCanarioDogBreed',
         'TreeingWalkerCoonhoundDogBreed',
         'LebedinCattleBreed',
         'MengguCattleBreed',
         'Dalland080PigBreed',
         'HollandShepherdDogBreed',
         'ChinampoCattleBreed',
         'TabapuaCattleBreed',
         'FrederiksborgHorseBreed',
         'ShanxiBlackPigBreed',
         'EastBulgarianHorseBreed',
         'TinosCattleBreed',
         'RepartidaGoatBreed',
         'AustrianSimmentalCattleBreed',
         'OssimiSheepBreed',
         'MunchkinShorthairCatBreed',
         'NitraHybridPigBreed',
         'ShetlandCattleBreed',
         'CupremHybridCattleBreed',
         'MulefootPigBreed',
         'ApulianPigBreed',
         'BoarPowerPig656PigBreed',
         'CanadianCuttingHorseHorseBreed',
         'BeauceronDogBreed',
         'KrishnaValleyCattleBreed',
         'FinnishLapphundDogBreed',
         'GreatPyreneeDog',
         'EnglishGuernseyGoatBreed',
         'DanubeWhitePigBreed',
         'DrysdaleSheepBreed',
         'SarnyPigBreed',
         'FrenchBulldog',
         'ThessalonianHorseBreed',
         'ChumbivilcasHorseBreed',
         'SchipperkeDog',
         'OldFarmCollieDogBreed',
         'DalesPonyHorseBreed',
         'VictorianBulldoggeDogBreed',
         'CaracuCattleBreed',
         'ForestMountainPigBreed',
         'PolledJerseyCattleBreed',
         'DishtyXZebuCattleBreed',
         'PortlandSheepBreed',
         'RandallLinebackCattleBreed',
         'DzumaliaPigBreed',
         'InternationalStripedHorseHorseBreed',
         'EstonianNativeHorseBreed',
         'SomaliCattleBreed',
         'HubeiWhitePigBreed',
         'BeetalGoatBreed',
         'SanheCattleBreed',
         'NorthSiberianPigBreed',
         'AfghanCattleBreed',
         'ComisanaSheepBreed',
         'CarolinaDogBreed',
         'PeneiaHorseBreed',
         'SerereCattleBreed',
         'AzerbaijanHorseBreed',
         'WestFinnishCattleBreed',
         'WielkopolskiHorseBreed',
         'WanzheSpottedWannanSpottedPigBreed',
         'BarbHorseBreed',
         'BambeyXZebuCattleBreed',
         'NorwegianFurSheepBreed',
         'MunchkinLonghairCatBreed',
         'NorthMalawiAngoniCattleBreed',
         'WirehairedStandardDachshundDogBreed',
         'PicPigBreed',
         'BhagnariCattleBreed',
         'TaiwanSmallBlackPigBreed',
         'SkyeTerrier',
         'MinbeiSpottedPigBreed',
         'Bodmin950PigBreed',
         'BlackSeaHorseHorseBreed',
         'CriolloCowBreed',
         'SiamesePigBreed',
         'YucaPigBreed',
         'BarraDoCuanzoCattleBreed',
         'BalkanHoundDogBreed',
         'JaulanCattleBreed',
         'KazakhCattleBreed',
         'AlambadiCattleBreed',
         'AmericanMiniatureHorseBreed',
         'EstonianBaconPigBreed',
         'PisanaCattleBreed',
         'RetintaExtremenaGoatBreed',
         'BleuDuMaineSheepBreed',
         'GobraCattleBreed',
         'IrishSetterDogBreed',
         'SteinschafSheepBreed',
         'WiltshirePigBreed',
         'SouthAfricanMiniatureHorseBreed',
         'StabyhounDogBreed',
         'EstonianNativeCattleBreed',
         'HarnaiSheepBreed',
         'KulaPigBreed',
         'BulgarianBrownCattleBreed',
         'DangiCattleBreed',
         'GuizhouHorseBreed',
         'WirehairedVizslaDogBreed',
         'MorabHorseBreed',
         'KoreanNativeCattleBreed',
         'JinzhouHorseBreed',
         'CanastrinhoPigBreed',
         'GoldenGuernseyGoatBreed',
         'FaeroesSheepBreed',
         'KanemXZebuCattleBreed',
         'LiangGuangSmallSpottedPigBreed',
         'FaeroesCattleBreed',
         'TuropoljePigBreed',
         'CormoSheepBreed',
         'NigerianHorseBreed',
         'ChejuHorseBreed',
         'NorthCaucasusPigBreed',
         'KarelianBearDog',
         'AlapahaBluebloodBulldogBreed',
         'SpanishMustangHorseBreed',
         'NuerCattleBreed',
         'BrunoDeJuraDogBreed',
         'WelshSheepdogBreed',
         'ShihTzuDog',
         'ArgentineFriesianCattleBreed',
         'AlbanianDwarfCattleBreed',
         'AulieAtaCattleBreed',
         'WestMacedonianCattleBreed',
         'WatusiUSACattleBreed',
         'SiberianCatBreed',
         'DanakilCattleBreed',
         'AmericanMustangHorseBreed',
         'MashonaCattleBreed',
         'AveligneseHorseBreed',
         'CanadianYorkshirePigBreed',
         'DanishLandraceGoatBreed',
         'BaldingerSpottedPigBreed',
         'IbageXZebuCattleBreed',
         'HirzaiHorseBreed',
         'ShekoCattleBreed',
         'BassetHound',
         'BelgianHorseBreed',
         'BoarPowerPigBreed',
         'PygmyGoatBreed',
         'ChickasawHorseBreed',
         'CholistaniCattleBreed',
         'BelgianSheepdog',
         'DanishSportPonyHorseBreed',
         'BerneseMountainDog',
         'JarmelistaCattleBreed',
         'NewfoundlandDog',
         'AshantiDwarfPigBreed',
         'BahiaPigBreed',
         'FilaBrasileiroDogBreed',
         'WelshPonyOfCobTypeHorseBreed',
         'OrientalHorseHorseBreed',
         'NoricHorseBreed',
         'MecklenburgHorseBreed',
         'BlackWelshMountainSheepBreed',
         'CanadianArcottSheepBreed',
         'WoodenLegGoatBreed',
         'KherigarhCattleBreed',
         'AmericanEskimoDogBreed',
         'CorsicanCattleBreed',
         'ChienDArtoisDogBreed',
         'CamboroughBluePigBreed',
         'NghiaBinhPigBreed',
         'SokotoGudaliCattleBreed',
         'VietnamesePotBelliedPigPigBreed',
         'SanheHorseBreed',
         'FabrianeseSheepBreed',
         'KentuckyRedBerkshirePigBreed',
         'GypsyVannerHorseBreed',
         'DanishLandracePigBreed',
         'YiwuHorseBreed',
         'YiliHorseBreed',
         'BatangasCattleBreed',
         'KerryBlueTerrier',
         'BelgianWarmbloodHorseBreed',
         'RomanianRedCattleBreed',
         'MoradaNovaSheepBreed',
         'LangHongPigBreed',
         'NubaMountainXZebuCattleBreed',
         'JavaHorseBreed',
         'BisaroPigBreed',
         'FlemishCattleBreed',
         'UkrainianWhiteheadedCattleBreed',
         'GoomsurCattleBreed',
         'LithuanianNativePigBreed',
         'HazaragieSheepBreed',
         'MangalargaHorseBreed',
         'HolgusCattleBreed',
         'PanamaSheepBreed',
         'FHCElitePig3PigBreed',
         'CzechMeatPigBreed',
         'HuazhongTwoEndBlackTongchengPigBreed',
         'TongaCattleBreed',
         'BichonsFriseDog',
         'AvilenaCattleBreed',
         'MashenPigBreed',
         'RedAndWhiteCattleBreed',
         'DurocPigBreed',
         'AustrianHoundDogBreed',
         'LorquinaCattleBreed',
         'VenezuelanZebuCattleBreed',
         'Basenji',
         'GalicianPonyHorseBreed',
         'SuiaXZebuCattleBreed',
         'RussianLongEaredWhitePigBreed',
         'GarreCattleBreed',
         'DanishHampshirePigBreed',
         'BoerboelDogBreed',
         'JacobSheepBreed',
         'GermanBlackPiedCattleBreed',
         'HamlinePigBreed',
         'JonicaGoatBreed',
         'MozambiqueAngoniCattleBreed',
         'SpanishBrownAlpineCattleBreed',
         'CentralRussianBlackPiedCattleBreed',
         'KuznetHorseBreed',
         'FrenchLandracePigBreed',
         'ChesapeakeBayRetrieverDogBreed',
         'BassetBleuDeGascogneDogBreed',
         'HequHorseBreed',
         'PumiDogBreed',
         'DartmoorPonyHorseBreed',
         'KrishnagariCattleBreed',
         'FannongSpottedPigBreed',
         'BentheimerLandschafSheepBreed',
         'GrandAngloFrançaisDogBreed',
         'WagyuCattleBreed',
         'DølafeCattleBreed',
         'AbyssinianHorseBreed',
         'PollMerinoSheepBreed',
         'BelmontRedCattleBreed',
         'DalaSheepBreed',
         'PressaMallorquinDogBreed',
         'SpiegelSheepBreed',
         'PabnaXZebuCattleBreed',
         'ShkodraPigBreed',
         'BeagleRoughDogBreed',
         'GelderlandHorseBreed',
         'KleenLeenRedPigBreed',
         'BavarianMountainHoundDogBreed',
         'SerranaGoatBreed',
         'RomanovSheepBreed',
         'CutchiCattleBreed',
         'IberianPigBreed',
         'EgyptianMauCat',
         'MoiCattleBreed',
         'SantaCruzXZebuCattleBreed',
         'MorenasDelNoroesteCattleBreed',
         'MiPigBreed',
         'DanishJerseyCattleBreed',
         'HuchuanMountainPigBreed',
         'IcelandicPigBreed',
         'AmerifaxcattleBreed',
         'Line21PigBreed',
         'AlbanianNativePigBreed',
         'WelshWalkingHorseBreed',
         'NoordhollanderSheepBreed',
         'BarbadoSheepBreed',
         'ChiosSheepBreed',
         'NordlandHorseBreed',
         'PhilippineNativeJalajalaPigBreed',
         'MiniaturePoodleDogBreed',
         'NegraSerranaGoatBreed',
         'SantaCruzSheepBreed',
         'RougeDeLQuestSheepBreed',
         'RedBrangusXZebuCattleBreed',
         'ChamoisOfTheAlpsGoatBreed',
         'AlurCattleBreed',
         'AlentejoHerderDogBreed',
         'SalukiDog',
         'RongchangPigBreed',
         'KangalDogBreed',
         'ItalianBrownCattleBreed',
         'HallingdalCattleBreed',
         'SkyrosHorseBreed',
         'BurwashCattleBreed',
         'BurmesePigBreed',
         'SombaCattleBreed',
         'HerefordCattleSuperbreed',
         'RaboLargoSheepBreed',
         'IsraeliHorseBreed',
         'SharabiCattleBreed',
         'ShahutouPigBreed',
         'CaneCorsoDogBreed',
         'DulmenPonyHorseBreed',
         'BoarPowerPig72PigBreed',
         'WhiteShepherdDogBreed',
         'PersianArabHorseBreed',
         'TosaInuDogBreed',
         'WhiteFulaniCattleBreed',
         'CotswoldSheepBreed',
         'HuangHuaiHaiBlackShenxianPigBreed',
         'FRSCattleBreed',
         'AiredaleTerrier',
         'Camborough22PigBreed',
         'GreekSteppeCattleBreed',
         'GansuAlpineFineWoolSheepBreed',
         'TelemarkCattleBreed',
         'ToupouriCattleBreed',
         'Bloodhound',
         'SwedishLandraceGoatBreed',
         'CotswoldPlatinumPigBreed',
         'CirnecoDellEtnaDogBreed',
         'FinnishSpitzDog',
         'MauritiusCreoleCattleBreed',
         'CaldelanaCattleBreed',
         'RiopardenzeXZebuCattleBreed',
         'OregonRexCatBreed',
         'AzaouakCattleBreed',
         'BaznaPigBreed',
         'BoarPowerPig565PigBreed',
         'BosTaurusXBisonBisonHybridCowBreed',
         'CorsicanHorseBreed',
         'SmallSpanishHoundDogBreed',
         'SouthernTswanaCattleBreed',
         'TSSH1XZebuCattleBreed',
         'EstonianRedCattleBreed',
         'GuteSheepBreed',
         'KilaraCattleBreed',
         'KuriCattleBreed',
         'DanishBroholmerDogBreed',
         'SarplaninacDogBreed',
         'Pointer',
         'FrenchAlpineGoatBreed',
         'SlovakianBlackPiedPigBreed',
         'AmericanEskimoMiniatureDogBreed',
         'IncaHairlessDogBreed',
         'RX3CattleBreed',
         'LiaoningBlackPigBreed',
         'KarakacanHorseBreed',
         'AmericanBashkirCurlyHorseBreed',
         'DeogirCattleBreed',
         'CubanPintoHorseBreed',
         'MontanaPigBreed',
         'ChakouyiHorseBreed',
         'GalicianBlondCattleBreed',
         'SoaySheepBreed',
         'NorthFinncattleCattleBreed',
         'GlanDonnersbergCattleBreed',
         'DaeraDinPanahGoatBreed',
         'WuzhishanPigBreed',
         'AmericanEskimoStandardDogBreed',
         'KishuDogBreed',
         'BorzoiDog',
         'GriffonNivernaisDogBreed',
         'LuingCowBreed',
         'MeraukeCattleBreed',
         'KM1PigBreed',
         'NorlinePigBreed',
         'OldFormatDachsundDogBreed',
         'LimianaCattleBreed',
         'LithuanianHeavyDraftHorseBreed',
         'VitoriaPigBreed',
         'TreeingTennesseeBrindleDogBreed',
         'IberianExtremaduraRedPigBreed',
         'RijnlamASheepBreed',
         'CzechColdbloodHorseBreed',
         'NimariCattleBreed',
         'MezohegyesSportHorseHorseBreed',
         'TurinoCattleBreed',
         'PasoFinoHorseBreed',
         'CentralRussianPigBreed',
         'PantelleriaCattleBreed',
         'AmericanWalkingPonyHorseBreed',
         'BurmeseXZebuCattleBreed',
         'SouthYunnanShortEaredPigBreed',
         'BritishRidingPonyHorseBreed',
         'WelshPigBreed',
         'ThaiRidgebackDogBreed',
         'KuvaszDog',
         'AweilDinkaCattleBreed',
         'BelorusBlackPiedPigBreed',
         'RomanianMeatPigPigBreed',
         'MalakanHorseBreed',
         'JapanesePollCattleBreed',
         'LabradorRetrieverDogBreed',
         'WalerHorseBreed',
         'VietnameseYorkshirePigBreed',
         'MaineCoonCat',
         'GhoriPigBreed',
         'KolubaraCattleBreed',
         'BelarusRedCattleBreed',
         'HainanPigBreed',
         'CanastrãoCapitãoChicoPigBreed',
         'AndeanHorseBreed',
         'IomudHorseBreed',
         'BillyDogBreed',
         'LaVelasquezCattleBreed',
         'PomeranianCoarsewoolSheepBreed',
         'JiningGrayGoatBreed',
         'ShillukCattleBreed',
         'SingapuraCat',
         'GermanLandraceBPigBreed',
         'BretonHorseBreed',
         'NewfoundlandSheepBreed',
         'BeiroaCattleBreed',
         'ShendiCattleBreed',
         'YacumentoCattleBreed',
         'TushinHorseBreed',
         'EpagneulBleuDePicardieDogBreed',
         'ArgentineCriolloHorseBreed',
         'LeicomaPigBreed',
         'MunsterLanderPointerDogBreed',
         'SpanishGreyhoundDogBreed',
         'MixedBreedCat',
         'AppenzellerDogBreed',
         'BrecknockHillCheviotSheepBreed',
         'RoyalCanadianMountedPoliceHorseHorseBreed',
         'SouthChinaBlackPigBreed',
         'FrenchCobHorseBreed',
         'PenshirePigBreed',
         'BaltaretPigBreed',
         'FinnishLandracePigBreed',
         'ShuwaCattleBreed',
         'DomesticLeopardCat',
         'GloucesterOldSpotPigBreed',
         'AlgarviaGoatBreed',
         'LimCattleBreed',
         'AngoraGoatBreed',
         'LatvianHorseBreed',
         'NewZealandRabbitBreed',
         'RamgarhiCattleBreed',
         'LakelandTerrier',
         'SaddlebredHorseSuperbreedHorseBreed',
         'PenbukPigBreed',
         'ByelorussianHarnessHorseBreed',
         'BraunesBergschafSheepBreed',
         'BatakHorseBreed',
         'VerinesaCattleBreed',
         'PretaCattleBreed',
         'LutaiWhitePigBreed',
         'TuyHoaCattleBreed',
         'WeleraPonyHorseBreed',
         'BloodhoundStHubertDogBreed',
         'VictoriaXZebuCattleBreed',
         'SumbawaHorseBreed',
         'RomneySheepBreed',
         'YiliWhitePigBreed',
         'MakaweliCattleBreed',
         'ShorthairCatBreed',
         'AmericanAlbinoHorseBreed',
         'WestFrenchWhitePigBreed',
         'IcelandicCattleBreed',
         'PicLinePig26PigBreed',
         'AmericanFoxhound',
         'BragancaCattleBreed',
         'MalteseDog',
         'PolskiOwczarekNizinnyDogBreed',
         'NigerianPigBreed',
         'IngessanaCattleBreed',
         'ScottishTerrier',
         'MiniatureHerefordCattleBreed',
         'AndalusianBlondCattleBreed',
         'DengchuanCattleBreed',
         'Beagle',
         'ClevelandBayHorseBreed',
         'MoraRomagnolaPigBreed',
         'SardinianSheepBreed',
         'BudyonnyHorseBreed',
         'CanastrãoJunqueiraPigBreed',
         'VianesaCattleBreed',
         'IloiloCattleBreed',
         'CaucasianCattleBreed',
         'GermanLandracePigBreed',
         'HarianaCattleBreed',
         'MalawiZebuCattleBreed',
         'FieldSpanielDogBreed',
         'XiangPigBreed',
         'BerrendasCattleBreed',
         'TimorHorseBreed',
         'RygjaSheepBreed',
         'BanEiHorseBreed',
         'SalernoHorseBreed',
         'AbyssinianHighlandZebuCattleBreed',
         'MadagascarZebuCattleBreed',
         'OberhasliGoatBreed',
         'CanikHorseBreed',
         'WuyiBlackPigBreed',
         'MoscowLonghairedToyTerrierDogBreed',
         'KuchinoshimaCattleBreed',
         'JijjigaZebuCattleBreed',
         'MasaiSheepBreed',
         'SwedishLandracePigBreed',
         'PolwarthSheepBreed',
         'ExoticShorthairedCat',
         'ShanHorseBreed',
         'TreeingFeistDogBreed',
         'RomagnolaCattleBreed',
         'MubanPigBreed',
         'PajunaCattleBreed',
         'NorbottenspetsDogBreed',
         'LugwareCattleBreed',
         'PulawyPigBreed',
         'StumpyTailCattleDogBreed',
         'ChartreuxCat',
         'ZaobeiCattleBreed',
         'CambodianCattleBreed',
         'AustralianGoatBreed',
         'WelshMountainSheepBreed',
         'RottweilerDog',
         'MegrelHorseBreed',
         'TaihuPigBreed',
         'ThillariCattleBreed',
         'PeloponnesusCattleBreed',
         'DLSSheepBreed',
         'ManapariCattleBreed',
         'WhiteSuffolkSheepBreed',
         'SanjiangWhitePigBreed',
         'PiauCarunchoPiauPigBreed',
         'MacauPigBreed',
         'NgandaCattleBreed',
         'BariaVietnamMadagascarCattleBreed',
         'PintoHorseBreed',
         'FrenchLargeWhitePigBreed',
         'MisakiHorseBreed',
         'CayuseHorseBreed',
         'SicilianPigBreed',
         'TaoyuanPigBreed',
         'BimalCattleBreed',
         'BrandenburgHorseBreed',
         'AbtenauerHorseBreed',
         'MediumWoolMerinoSheepBreed',
         'BengalCatBreed',
         'FrenchTrotterHorseBreed',
         'DrakensbergerCattleBreed',
         'JamaicaRedXZebuCattleBreed',
         'KraskyOvcarDogBreed',
         'PlottHound',
         'LucanianCattleBreed',
         'Affenpinscher',
         'BoarPowerPig828PigBreed',
         'SeferihisarCattleBreed',
         'CaliforniaVariegatedMutantSheepBreed',
         'ArabianHorseBreed',
         'GreaterSwissMountainDogBreed',
         'SuffolkCattleBreed',
         'ZhongweiGoatBreed',
         'LlanwenogSheepBreed',
         'CosteñoConCuernosCattleBreed',
         'SolognoteSheepBreed',
         'BrownAtlasCattleBreed',
         'TeddyRooseveltTerrierDogBreed',
         'BriquetBassetGriffonVendeenDogBreed',
         'PolledGirCattleBreed',
         'SoftCoatedWheatenTerrier',
         'DaweiziPigBreed',
         'HarzRedCattleBreed',
         'PicLinePig24PigBreed',
         'DorsetDownSheepBreed',
         'OrientalLonghairCatBreed',
         'KyiLeoDogBreed',
         'BakwiriCattleBreed',
         'OxfordSandyBlockPigBreed',
         'KamoriGoatBreed',
         'PieRougeDesPlainesCattleBreed',
         'DermantsiPiedPigBreed',
         'RasNLansaPigBreed',
         'PitmanMooreMiniaturePigBreed',
         'KnabstrupHorseBreed',
         'BlackRussianTerrierDogBreed',
         'ChryakPICPigBreed',
         'SiriZebuCattleBreed',
         'InobutaInterSpeciesHybridPigBreed',
         'BachaurCattleBreed',
         'BosTaurusTaurusSubspeciesDomesticEuropeanCowBreed',
         'SouthChinaZebuCattleBreed',
         'ManipuriHorseBreed',
         'BoerGoatBreed',
         'DeutschesBlaukoepfigesFleischschafSheepBreed',
         'BergamascoDogBreed',
         'AngloArabHorseBreed',
         'LargeBlackPigBreed',
         'BrabantHorseBreed',
         'JavaneseOngoleCattleBreed',
         'IrishDraftHorseBreed',
         'JapaneseSpitzDogBreed',
         'GhanaSangaXZebuCattleBreed',
         'EstonianDraftHorseBreed',
         'DeutscheBrackeDogBreed',
         'DrentsHeideschaapSheepBreed',
         'MiamiPigBreed',
         'HerensCattleBreed',
         'KenranCattleBreed',
         'HalfSaddlebredHorseBreed',
         'HungahybPigBreed',
         'EntlebucherDogBreed',
         'LocalIndianDairyXZebuCattleBreed',
         'AlaskanMalamute',
         'LohaniCattleBreed',
         'BeltedWelshCattleBreed',
         'SomaliPonyHorseBreed',
         'ShropshireTerrierDogBreed',
         'DornaCattleBreed',
         'YucatanMinaturePigBreed',
         'AureEtSaintGironsCattleBreed',
         'AbondanceCattleBreed',
         'WurttembergHorseBreed',
         'FuanSpottedPigBreed',
         'JämthundDogBreed',
         'KravarskyCattleBreed',
         'EpirusCattleBreed',
         'AsturianMountainCattleBreed',
         'HaBacPigBreed',
         'PerdigueroDeBurgosDogBreed',
         'Hound',
         'RomanianSteppeCattleBreed',
         'YakutHorseBreed',
         'LatvianBrownCattleBreed',
         'WaziriHorseBreed',
         'HanjiangBlackPigBreed',
         'BelgianWhiteAndRedCattleBreed',
         'GreaterCaucasusCattleBreed',
         'MaremmanaPigBreed',
         'BrandlbrackeDogBreed',
         'MazuryCattleBreed',
         'BoykinSpanielDogBreed',
         'BrahornXZebuCattleBreed',
         'EgyptianCattleBreed',
         'PottokTarpanHorseXDomesticHorseBreed',
         'BoarPowerPig929PigBreed',
         'SorraiaHorseBreed',
         'HeiheHorseBreed',
         'FriuliBlackPigBreed',
         'FaeroesIslandHorseHorseBreed',
         'RegusCattleBreed',
         'EnglishToySpaniel',
         'ClydesdaleHorseBreed',
         'OldFormatSheepBreed',
         'CalabrianPigBreed',
         'MaremmaSheepdogsDogBreed',
         'RedHampracePigBreed',
         'BoranCattleBreed',
         'SwissEdelschweinPigBreed',
         'NadbuzanskaPigBreed',
         'AppaloosaHorseBreed',
         'HailumPigBreed',
         'BanjaraGreyhoundDogBreed',
         'NationalSpottedSaddleHorseHorseBreed',
         'NGaoundereCattleBreed',
         'GermanColoredGoatBreed',
         'BearnaisCattleBreed',
         'AkbashDogBreed',
         'EnglishBulldogBreed',
         'GuadeloupeCreolePigBreed',
         'OldGermanShepherdDogBreed',
         'KaokoveldCattleBreed',
         'SenegusCattleBreed',
         'LapermLonghairCatBreed',
         'RackingHorseHorseBreed',
         'SimmentalCowBreed',
         'GanzhouWhitePigBreed',
         'ChineseMongolianHorseBreed',
         'EastFriesianOldTypeHorseBreed',
         'JapaneseBobtailCat',
         'KwaiPigBreed',
         'WarmbloodHorseBreed',
         'AlbeseCattleBreed',
         'GuanzhongHorseBreed',
         'KelePigBreed',
         'ParmensePigBreed',
         'ChowChow',
         'NormandPigBreed',
         'MishimaCattleBreed',
         'ComebackSheepBreed',
         'BergamascaSheepBreed',
         'BlackForrestCattleBreed',
         'PrewakwaCattleBreed',
         'HungarianPiedCattleBreed',
         'NachiGoatBreed',
         'UkrainianSaddleHorseHorseBreed',
         'ThailandFightingZebuCattleBreed',
         'ElegantWarmbloodHorseBreed',
         'BalineseCat',
         'TibetanTerrier',
         'AustralianLowlineCattleBreed',
         'PetitGasconSaintongeoisDogBreed',
         'IstobenCattleBreed',
         'AdenZebuCattleBreed',
         'JotkoXZebuCattleBreed',
         'AracenaCattleBreed',
         'BondSheepBreed',
         'PolypaySheepBreed',
         'OldEnglishSheepdog',
         'BoarPowerPig545PigBreed',
         'BlancaAndaluzaGoatBreed',
         'DeoniCattleBreed',
         'BluefacedLeicesterSheepBreed',
         'BrittanySpanielDogBreed',
         'AppenzellGoatBreed',
         'CairnTerrier',
         'ApennineSheepBreed',
         'IbizanHound',
         'CurlyCoatedRetrieverDogBreed',
         'SudanCountryBredHorseBreed',
         'AleppoCattleBreed',
         'LongAndShortCoatChihuahuaDogBreed',
         'SouthGermanColdbloodHorseBreed',
         'BaladicattleCattleBreed',
         'MoyleHorseBreed',
         'IcelandicSheepBreed',
         'BrahmousinXZebuCattleBreed',
         'FratiCattleBreed',
         'XinjinPigBreed',
         'RhaetianGrayCattleBreed',
         'GaunlingCattleBreed',
         'HaysConverterCowBreed',
         'JinhongHorseBreed',
         'SardinianPonyHorseBreed',
         'LichuanHorseBreed',
         'GanzhongnanSpottedPigBreed',
         'StandardSchnauzerDogBreed',
         'HainanWenchangPigBreed',
         'WestphalianHorseBreed',
         'CãoDaSerraDeAiresDogBreed',
         'RhodesianRidgebackDog',
         'CamargueHorseBreed',
         'CinisaraCattleBreed',
         'SabuesosEspañolesDogBreed',
         'MoruchaCattleBreed',
         'HampenPigBreed',
         'QuasahXZebuCattleBreed',
         'BlackAngusCowBreed',
         'LatiSheepBreed',
         'SamoyedDog',
         'FHCElitePig1PigBreed',
         'ConnorPrairiePigBreed',
         'EnglishCockerSpanielDogBreed',
         'RockAlpineGoatBreed',
         'CzechPiedCattleBreed',
         'SanClementeGoatBreed',
         'CubanCriolloXZebuCattleBreed',
         'LatvianWhitePigBreed',
         'ShibaInuDogBreed',
         'KandahariCattleBreed',
         'ClumberSpanielDogBreed',
         'FloridaCrackerPineywoodsCattleBreed',
         'GermanBrownCattleBreed',
         'Slovhyb1PigBreed',
         'PolandChinaPigBreed',
         'TainoXZebuCattleBreed',
         'GloucesterCattleBreed',
         'KajliSheepBreed',
         'PiquiraPonyHorseBreed',
         'GelbveihCowBreed',
         'ChineseFooDogBreed',
         'GasconyPigBreed',
         'RomedaleSheepBreed',
         'TinimaCattleBreed',
         'ChincoteaguePonyHorseBreed',
         'CanadianHorseBreed',
         'DialiCattleBreed',
         'SakhalinWhitePigBreed',
         'BoxerDog',
         'MongallaCattleBreed',
         'PelónPigBreed',
         'PechoraHorseBreed',
         'ChienFrançaisBlancEtNoirDogBreed',
         'NeopolitanMastiffDogBreed',
         'RetintaCattleBreed',
         'EleiaHorseBreed',
         'ShengxianSpottedPigBreed',
         'KenyaZebuCattleBreed',
         'PintabianHorseBreed',
         'PetitBrabaçonDogBreed',
         'SlovakianPiedCattleBreed',
         'MexicanWattledPigBreed',
         'Collie',
         'KerryHillSheepBreed',
         'BarotseCattleBreed',
         'BhutiaHorseBreed',
         'GermanWirehairedPointerDogBreed',
         'MorangCattleBreed',
         'RussianLargeWhitePigBreed',
         'ChimaineCattleBreed',
         'LuzeinSheepBreed',
         'FoutaHorseBreed',
         'RedGallowayCattleBreed',
         'TransylvanianPinzguaCattleBreed',
         'FroxfieldPygmyPigBreed',
         'AyrshireCowBreed',
         'MoravaPigBreed',
         'HampshireDownSheepBreed',
         'BambawaCattleBreed',
         'AustralianWhiteCattleBreed',
         'SudaneseFulaniCattleBreed',
         'WelshHillSpeckledFaceSheepBreed',
         'SchillerHoundDogBreed',
         'AlbèresCattlebreed',
         'RedboneCoonhoundDogBreed',
         'BootedGoatBreed',
         'HungarianSportHorseHorseBreed',
         'SPMPigBreed',
         'ExmoorPonyHorseBreed',
         'PinzgauerCowBreed',
         'WiltshireHornSheepBreed']}

name_for_cid[12252] = 'CardiacUltrasoundCardiacFunction'
cid_concepts[12252] = \
{'LN': ['HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'FractionalShortening',
        'PercentThickening'],
 'SCT': ['CardiacEjectionFraction']}

name_for_cid[9534] = 'RadiotherapyTargets'
cid_concepts[9534] = \
{'DCM': ['CTVPrimary',
         'PTVPrimary',
         'PTVNodal',
         'RadiationDoseReferencePoint',
         'GTVPrimary',
         'RadiationDoseNormalizationPoint',
         'GTVNodal',
         'ITV',
         'EntireBodyTargetVolume',
         'CTVNodal',
         'TreatedVolume'],
 'SCT': ['IrradiatedVolume', 'GTV', 'PTV', 'CTV']}

name_for_cid[604] = 'DeviceReuse'
cid_concepts[604] = \
{'DCM': ['Unused', 'Reused']}

name_for_cid[6209] = 'ColonMorphologyDescriptor'
cid_concepts[6209] = \
{'SCT': ['Circumferential', 'Sessile', 'Flat', 'Ulcer', 'Pedunculated']}

name_for_cid[4234] = 'RefractiveSurgeryTypes'
cid_concepts[4234] = \
{'DCM': ['SMILE'], 'SCT': ['RK', 'LASIK', 'LASEK', 'PRK']}

name_for_cid[3772] = 'HealthStatus'
cid_concepts[3772] = \
{'SCT': ['LostToFollowUp',
         'Disabled',
         'Deceased',
         'SymptomFree',
         'ChronicallyIll',
         'Alive',
         'AliveAndWell',
         'SeverelyDisabled',
         'InRemission',
         'SeverelyIll']}

name_for_cid[6120] = 'ChestBorderDefinition'
cid_concepts[6120] = \
{'DCM': ['SharplyDemarcated',
         'PoorlyDemarcated',
         'SharplyDefined',
         'DistinctlyDefined',
         'WellDemarcated'],
 'SCT': ['PoorlyDefined', 'WellDefined', 'Circumscribed']}

name_for_cid[6065] = 'InstrumentApproach'
cid_concepts[6065] = \
{'DCM': ['InferolateralToSuperomedial',
         'InferomedialToSuperolateral',
         'SuperolateralToInferomedial',
         'SuperomedialToInferolateral'],
 'SCT': ['Lateral', 'Inferior', 'Superior', 'Medial']}

name_for_cid[12272] = 'CardiacUltrasoundSemilunarValves,AnnulateandSinusesMeasurements'
cid_concepts[12272] = \
{'DCM': ['Tau', 'IndicatorMeanTransitTime', 'RRInterval'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'AWaveAmplitude',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'LeafletThickness',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'LeafletSeparation',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'BCSlope',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'TimeFromQWaveToPulmonicValveCloses',
        'CEDistance',
        'DpDtByUS',
        'PulmonaryArteryPressureUsingAccelTime',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'AorticValveCuspSeparation',
        'TimeAveragedMeanVelocity',
        'DEExcursion'],
 'SCT': ['MajorAxis',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[3119] = 'LVFunction'
cid_concepts[3119] = \
{'SCT': ['Normal', 'ImpairedLeftVentricularFunction']}

name_for_cid[9514] = 'AnatomicalStructuresforRadiotherapy'
cid_concepts[9514] = \
{'BARI': ['_1stMarginalCoronaryArteryLaterals',
          '_2ndMarginalCoronaryArteryLaterals',
          '_3rdMarginalCoronaryArteryLaterals',
          '_1stDiagonalCoronaryArteryLaterals',
          '_2ndDiagonalCoronaryArteryLaterals',
          'RamusLaterals',
          '_3rdDiagonalCoronaryArteryLaterals',
          'PosteriorDescendingSeptalPerforators'],
 'DCM': ['Phantom',
         'AorticKnob',
         'ArchOfTheAzygosVein',
         'SecondaryPulmonaryLobule',
         'Hiatus',
         'AzygoesophagealRecessInterface',
         'ParaspinalLine',
         'PosteriorTrachealStripe',
         'RightTrachealStripe',
         'Stripe',
         'MidlungWindow',
         'CarinaAngle',
         'CentrilobularStructures',
         'AnteriorJunctionLine',
         'PosteriorJunctionLine',
         'SurfaceOfTheChoroidScleraInterface',
         'OuterSurfaceOfTheBM',
         'OuterSurfaceOfTheCC',
         'PosteriorSurfaceOfTheRPE',
         'SurfaceOfTheCenterOfTheRPE',
         'OuterSurfaceOfIPL',
         'OuterSurfaceOfGCL',
         'OuterSurfaceOfOPL',
         'OuterSurfaceOfINL',
         'SurfaceBetweenInnerAndOuterSegmentsOfThePhotoreceptors',
         'OuterSurfaceOfHFL',
         'AnteriorSurfaceOfTheRPE',
         'SurfaceOfTheInterdigitatingZoneBetweenRetinaAndRPE',
         'OuterSurfaceOfRNFL'],
 'FMA': ['ArcuateFasciculus'],
 'SCT': ['ProfundaFemorisVein',
         'LeftHypochondriacRegion',
         'SternocleidomastoidMuscle',
         'TruncusCoeliacus',
         'InferiorLongitudinalFasciculus',
         'BasilarArtery',
         'ParanasalSinus',
         'BrainCerebrospinalFluidPathway',
         'WristJoint',
         'MidCircumflexCoronaryArtery',
         'CentralRetinalVein',
         'IntermediateArteryRamus',
         'GenicularArtery',
         'LeftVentricleInflow',
         'FontanelOfSkull',
         'SigmoidColon',
         'GastricArtery',
         'SubscapularisMuscle',
         'ExternalEar',
         'LeftVentricleOutflowTract',
         'RightFemoralArtery',
         'Intracranial',
         'JawRegion',
         'Calcaneus',
         'SolealVein',
         'IntercostalArtery',
         'PeripheralNerve',
         'LobarArtery',
         'TeresMajorMuscle',
         'PatentDuctusArteriosus',
         'ThighPerforator',
         'MesentericVein',
         'Pituitary',
         'MaleInternalGenitalia',
         'MidRightCoronaryArtery',
         'PalatineBone',
         'CoronaryArtery',
         'RightBranchOfHepaticArtery',
         'UpperLimb',
         'DuraMater',
         'ExternalIliacVein',
         'CongenitalPulmonaryArteriovenousFistula',
         'DescendingAorta',
         'ParietalBone',
         'Flank',
         'Maxilla',
         'EsophagealHiatus',
         'CentralRetinalArtery',
         'Neurohypophysis',
         'SegmentalArtery',
         'BrachialVein',
         'RightUpperQuadrantOfAbdomen',
         'Eye',
         'SuperficialPalmarArch',
         'Prostate',
         'HepaticArtery',
         'FacialBones',
         'Back',
         'PeronealArtery',
         'Duodenum',
         'RightHypochondriacRegion',
         '_1stMarginalCoronaryArtery',
         'SuperiorMesentericVein',
         'LumbarArtery',
         'Cornea',
         'Larynx',
         'PeronealVein',
         'FemoralArtery',
         'Ovary',
         'ThoracoLumbarSpine',
         'Urethra',
         'CoronaryArteryGraft',
         'Head',
         'AdrenalGland',
         'MitralValve',
         'AcromioclavicularJoint',
         'UmbilicalRegion',
         'SupraspinatusMuscle',
         'CongenitalCoronaryArteryFistulaToRightVentricle',
         'Calyx',
         'PosteriorMedialTributary',
         'PulmonaryVein',
         'SoftPalate',
         'Submental',
         'ExternalIntercostalMuscle',
         'CranialVenousSystem',
         'NucleusAccumbens',
         'GiacominiVein',
         'PoplitealFossa',
         'MiddleCerebellarPeduncle',
         'PulmonaryArteryConduit',
         'InternalCarotidArteryC5Segment',
         'FissureOfLung',
         'RightVentricle',
         'LevatoresCostarumMuscles',
         'FrontalLobe',
         'VagusNerve',
         'Adventitia',
         'Abdomen',
         'HeadAndNeck',
         'Hypopharynx',
         'RightCoronaryArteryOstium',
         'HepaticVein',
         'MarginalCoronaryArtery',
         'PelvisAndLowerExtremities',
         'InternalMammaryArtery',
         'Bladder',
         'Vulva',
         'LargeIntestine',
         'EntorhinalCortex',
         'Forearm',
         'BronchialArtery',
         'NeckChestAndAbdomen',
         'ExternalCarotidArtery',
         'SuperiorMesentericArtery',
         'TemporalLobe',
         'SystemicArtery',
         'PosteriorTibialArtery',
         'FemoralVein',
         'RightAuricularAppendage',
         'GreatSaphenousVeinOfThigh',
         'LeftHepaticVein',
         'AbdominalWallMuscle',
         'SuperficialFemoralArtery',
         'VomerBone',
         'UterusAndFallopianTubes',
         'Arachnoid',
         '_2ndMarginalCoronaryArtery',
         'SuperiorCerebellarPeduncle',
         'AorticArch',
         'Carina',
         'Telencephalon',
         'LeftBranchOfHepaticArtery',
         'ArcuateArteryOfTheKidney',
         'Heart',
         'TarsalJoint',
         'CephalicVein',
         'LingualArtery',
         'Lip',
         'GastricVein',
         'AirwayStructure',
         'CongenitalCoronaryArteryFistulaToRightAtrium',
         'DeltoidMuscle',
         'LeftMainCoronaryArtery',
         'CerebralFornix',
         'Penis',
         'Baffle',
         'InferiorRightHepaticVein',
         'VentricularVein',
         'CarotidSiphon',
         'SystemicVenousAtrium',
         'PlantarArterialArch',
         'Parathyroid',
         'AnomalousPulmonaryVein',
         'CervicoThoracicSpine',
         'LeftMainCoronaryArteryOstium',
         'Scapula',
         'LesserSaphenousVein',
         'Kidney',
         'AutonomicNerve',
         'GreatSaphenousVein',
         'OmentalBursa',
         'InternalCarotidArtery',
         'Ureter',
         'RightPosteriorAVCoronaryArtery',
         'UlnarVein',
         'FascialLayer',
         'AorticValve',
         'MiddleEar',
         'InternalAuditoryCanal',
         'RightPulmonaryArtery',
         'RenalPelvis',
         'ProperHepaticArtery',
         'InterlobarArteryOfKidney',
         'Thrombus',
         'VenousNetwork',
         'SubcostalMuscle',
         'CorpusCallosum',
         'CommonAtrium',
         'PerforatingArteryOfKidney',
         'Humerus',
         'Ilium',
         'AorticFistula',
         'InternalCarotidArteryC6Segment',
         'ILMInternalLimitingMembrane',
         'NasalBone',
         'AnteriorTibialArtery',
         'CranialNerve',
         'Cervix',
         'SkinOfAbdomen',
         'LeftAnteriorDescendingCoronaryArtery',
         'Artery',
         'CommonCarotidArtery',
         'Buttock',
         'InnermostIntercostalMuscles',
         '_3rdMarginalCoronaryArtery',
         'FallopianTube',
         'MastoidBone',
         'Vein',
         'CommonIliacArteryBifurcation',
         'MiddleHepaticVein',
         'CerebrospinalFluid',
         'EsophagealArtery',
         'Thigh',
         'RadialVein',
         'InferiorRightPulmonaryVein',
         'Trachea',
         'PulmonaryValveSinuses',
         'LeftPosterolateralCircumflexCoronaryArtery',
         'UterineArtery',
         'BodyConduit',
         'Thalamus',
         'EndoVesical',
         'RightVentricleInflow',
         'MesentericArtery',
         'LeftLowerQuadrantOfAbdomen',
         'MuscleOfHead',
         'SpinalCerebrospinalFluidPathway',
         'MiddleCerebralArteryM1Segment',
         'SesamoidBonesOfFoot',
         'InternalIliacArtery',
         'SuperiorVenaCava',
         'SixthLumbarArtery',
         'PosteriorArchVein',
         'AnteriorSpinalArtery',
         'AnusRectumAndSigmoidColon',
         'Omentum',
         'SplenicVein',
         'Colon',
         'SubmandibularGland',
         'CentralNervousSystem',
         'CarotidBifurcation',
         'LeftLumbarRegion',
         'InnominateArtery',
         'SplenicArtery',
         'FemaleInternalGenitalia',
         'Extremity',
         'FacialArtery',
         'AccessoryRenalArtery',
         'BoneStructureOfHeadAndOrNeck',
         'EpigastricRegion',
         'Media',
         'SeminalVesicle',
         'DistalCircumflexCoronaryArtery',
         'AbdominalCavity',
         'TeresMinorMuscle',
         'BoneOfUpperLimb',
         'DorsalisPedisArtery',
         'Vagina',
         'LongissimusMuscle',
         'LowerInnerQuadrantOfBreast',
         'Clavicle',
         'ParotidGland',
         'AorticIsthmus',
         'CommonFemoralVein',
         'SubstantiaNigra',
         '_1stLeftPosterolateralCoronaryArtery',
         'Subxiphoid',
         'UpperInnerQuadrantOfBreast',
         'Diaphragm',
         'ThoracicAorta',
         'RightInguinalRegion',
         'VasDeferens',
         'AxillaryArtery',
         'LeftPulmonaryArtery',
         'Ileum',
         'SalivaryGland',
         'VertebralArtery',
         'TransjugularIntrahepaticPortosystemicShunt',
         'RightLowerQuadrantOfAbdomen',
         'SupraRenalAorta',
         'InternalJugularVein',
         'LacrimalArteryOfRightEye',
         'CockettPerforatingVein',
         'ChordaeTendineaeCordis',
         'LumboSacralSpine',
         'SacroiliacJoint',
         'RightMainBronchus',
         'ScalenousAnteriorMuscle',
         'AVGrooveContinuationOfCircumflexArtery',
         'EntireBody',
         'TesticularArtery',
         'Thumb',
         'HypogastricRegion',
         'RightLumbarRegion',
         'GreatCardiacVein',
         'HyoidBone',
         'ThoracicDuct',
         'MiddleLobeOfRightLung',
         'BoydPerforatingVein',
         'Peritioneum',
         'CommonIliacArtery',
         'SupraclavicularRegionOfNeck',
         'Meninges',
         'Shoulder',
         'TransversusThoracis',
         'VascularGraft',
         'CerebellarWhiteMatter',
         'InterventricularSeptum',
         '_2ndLeftPosterolateralCoronaryArtery',
         'MiddleCerebralArteryM2Segment',
         'Cingulum',
         'LowerLobeOfLung',
         'TesticularVein',
         'AnteriorTibialVein',
         'LacrimalBone',
         'BasilicVein',
         'LeftInguinalRegion',
         'RightCoronaryArtery',
         'HunterianPerforatingVein',
         'IntraPelvic',
         'Putamen',
         'ProfundaFemorisArtery',
         'CostocervicalTrunk',
         'Pharynx',
         'FifthLumbarArtery',
         'AxillaryVein',
         'LigamentumArteriosum',
         'Nasopharynx',
         'SuprapulmonicValveArea',
         'SympatheticTrunk',
         'LobeOfLung',
         'SubclavianVein',
         'PulmonaryValve',
         'ExternalIliacArtery',
         'ProximalRightCoronaryArtery',
         'TemporalBone',
         'BoneOfLowerLimb',
         'Midbrain',
         'OvarianVein',
         'PosteriorCerebralArtery',
         'BrachialPlexus',
         'ErectorSpinaeMuscle',
         'DeepPalmarVenousArch',
         'ApexOfRightVentricle',
         'LateralCalfPerforator',
         'Mouth',
         'PharynxAndLarynx',
         'CarotidArtery',
         'FourthLumbarArtery',
         'SmallestCardiacVein',
         'ThirdLumbarArtery',
         'LimbicLobe',
         'SkinOfChest',
         'UmbilicalArtery',
         'BuccalRegionOfFace',
         'LymphNode',
         'PrimitiveAorta',
         'RadialArtery',
         'LeftFemoralArtery',
         'LumenOfBloodVessel',
         'TrapeziusMuscle',
         'Esophagus',
         'AntecubitalVein',
         'SpinalCordGrayMatter',
         'Mandible',
         'LateralCorticospinalTract',
         'Parasternal',
         'LeftMainBronchus',
         'BroadLigament',
         'SystemicVein',
         'OccipitalLobe',
         'CerebralAqueduct',
         'Insula',
         'PulmonaryVein',
         'MuscleOfNeck',
         'SubclavianArtery',
         'Amygdala',
         'Lung',
         '_3rdLeftPosterolateralCoronaryArtery',
         'OpticChiasm',
         'OphthalmicArtery',
         'Nose',
         'PancreaticDuct',
         'Hand',
         'PectoralisMajorMuscle',
         'CardiovascularSystem',
         'FemaleExternalGenitalia',
         'OvarianArtery',
         'Chest',
         'ElbowJoint',
         'Hippocampus',
         'Sternum',
         'NeckAndChest',
         'RightVentricleOutflowTract',
         'GlobusPallidus',
         'SuperficialFemoralVein',
         'SuperiorLeftPulmonaryVein',
         'LeftPosteriorDescendingCircumflexCoronaryArtery',
         'BileDuct',
         'SystemicCollateralArteryToLung',
         'UpperOuterQuadrantOfBreast',
         'PancreaticDuctAndBileDuctSystems',
         'ApexOfLeftVentricle',
         'PericardialCavity',
         'PulmonaryArtery',
         'Skull',
         'MidLeftAnteriorDescendingCoronaryArtery',
         'AtrioventricularVein',
         'Thyroid',
         'EthmoidBone',
         'OpticRadiation',
         'InfraRenalAorta',
         'SternoclavicularJoint',
         'EndoNasopharyngeal',
         'Breast',
         'EsophagusStomachAndDuodenum',
         'Neck',
         'AxillaryFascia',
         'CarotidBulb',
         'Finger',
         'PoplitealVein',
         'InfraspinatusMuscle',
         'PiaMater',
         'TricuspidValve',
         'RightAtrium',
         'VascularSclerosis',
         'SuperiorPhrenicArtery',
         'CommonVentricle',
         'DistalLeftAnteriorDescendingCoronaryArtery',
         'InnerEar',
         'Coccyx',
         'AtrialSeptalDefect',
         'RightPulmonaryVein',
         'TemporomandibularJoint',
         'Uterus',
         'Ventricle',
         'DescendingAorta',
         'TracheaAndBronchus',
         'Testis',
         'ProximalLeftAnteriorDescendingCoronaryArtery',
         'OccipitalBone',
         'Eyeball',
         'Face',
         'Spine',
         'UrinarySystem',
         'BrainStem',
         'Foot',
         'WhiteMatter',
         'MuscleOfUpperLimb',
         'PeripheralNervousSystem',
         'PinealGland',
         'UpperLobeOfLung',
         'Zygoma',
         'InnominateVein',
         'ProximalCircumflexCoronaryArtery',
         'LowerLimb',
         'PosteriorCommunicationArtery',
         'JuxtaposedAtrialAppendage',
         'VertebralColumnAndCranium',
         'Jejunum',
         'CarotidBody',
         'ChestAbdomenAndPelvis',
         'AbdominalAorta',
         'RightHepaticVein',
         'GastrocnemiusVein',
         'Scalp',
         'Tongue',
         'ChestWallMuscle',
         'GrayMatter',
         'Fibula',
         'SuperiorLongitudinalFasciculus',
         'OrbitalStructure',
         'InferiorCardiacVein',
         'LeftMainBranchOfPortalVein',
         'InternalIntercostalMuscle',
         'LateralVentricle',
         'CaudateNucleus',
         'LacrimalArtery',
         'OccipitalArtery',
         'Brain',
         'BloodVessel',
         'Sclera',
         'AnteriorPosteriorCerebralArteryBifurcation',
         'AtrialVein',
         '_1stDiagonalCoronaryArtery',
         'SuprasternalNotch',
         'Perineum',
         'InteratrialSeptum',
         'Femur',
         'Rectum',
         'AnteriorCardiacVein',
         'Pancreas',
         'BoneOfThorax',
         'LeftAtrium',
         'CerebralArtery',
         'InferiorMesentericArtery',
         'DigitalArteryOfHand',
         'SecondLumbarArtery',
         'IliocostalisMuscle',
         '_1stRightPosterolateralCoronaryArtery',
         'BladderAndUrethra',
         'CervicalSpine',
         'IliacAndOrFemoralArtery',
         'SpinalCordWhiteMatter',
         'Sacrum',
         'SmallIntestine',
         'SuperiorRightPulmonaryVein',
         'InferiorVenaCava',
         'MaleExternalGenitalia',
         'IntraArticular',
         'Axilla',
         'RetroperitonealSpace',
         'Thymus',
         'ELMExternalLimitingMembrane',
         'AnteriorMiddleCerebralArteryBifurcation',
         'ParietalLobe',
         'SuperiorThyroidArtery',
         'IliacArtery',
         'DistalRightCoronaryArtery',
         'CerebralCortex',
         'TruncusArteriosusCommunis',
         'Tooth',
         'Joint',
         'SubmandibularArea',
         'GreatSaphenousVeinOfCalf',
         'LeftPulmonaryVein',
         'PectoralisMinorMuscle',
         'OpticCanal',
         'Aorta',
         'SuperficialTemporalArtery',
         'MedialLemniscus',
         'Diencephalon',
         'CongenitalCoronaryArteryFistulaToLeftVentricle',
         'NeckChestAbdomenAndPelvis',
         'InferiorCerebellarPeduncle',
         'RightMainBranchOfPortalVein',
         'SphenoidBone',
         'LatissimusDorsiMuscle',
         'ChestAndAbdomen',
         'Spleen',
         'CommonHepaticArtery',
         'AnkleJoint',
         'LowerOuterQuadrantOfBreast',
         'Mediastinum',
         'Cerebellum',
         'FirstLumbarArtery',
         'PoplitealArtery',
         'UpperUrinaryTract',
         'PulmonaryVenousAtrium',
         'Bronchus',
         'Pelvis',
         'PulmonaryChamberOfCorTriatriatum',
         'DorsalScapularArtery',
         'UlnarArtery',
         'LeftVentricle',
         'AbdomenAndPelvis',
         'LumbarRegion',
         'CostalCartilage',
         'ThoracicSpine',
         'FourthVentricle',
         'SaphenousVein',
         'Gallbladder',
         '_2ndRightPosterolateralCoronaryArtery',
         'RenalArtery',
         'PeritonealCavity',
         'UpperArm',
         'OpticTract',
         'HipJoint',
         'TrunkOfPulmonaryArtery',
         'GastroduodenalArtery',
         'AzygosVein',
         'BrachialArtery',
         '_2ndDiagonalCoronaryArtery',
         'Liver',
         'HipJoint',
         'Rib',
         'SerratusAnteriorMuscle',
         'CommonFemoralArtery',
         'SmallCardiacVein',
         'SuperficialPalmarVenousArch',
         'Vertebra',
         'Stomach',
         'CircumflexCoronaryArtery',
         'AnteriorCommunicatingArtery',
         'CongenitalCoronaryArteryFistulaToLeftAtrium',
         'InferiorMesentericVein',
         'MedianCubitalVein',
         'CoronarySinus',
         'HilarArtery',
         'TerminalInternalCarotidArtery',
         'ApexOfLung',
         'ThirdVentricle',
         'Atrium',
         'CerebralWhiteMatter',
         'Knee',
         '_3rdDiagonalCoronaryArtery',
         'SpinalisMuscle',
         'SubarachnoidSpace',
         'PulmonaryVeinConfluence',
         'SaphenofemoralJunction',
         'EyeRegion',
         'BrainVentricle',
         'Anus',
         'UmbilicalVein',
         'FrontalBone',
         'CommonIliacVein',
         'TrabeculaeCarnae',
         'InguinalRegion',
         'PosteriorCerebralArteryP1Segment',
         '_3rdRightPosterolateralCoronaryArtery',
         'EndoNasal',
         'LumbarSpine',
         'ThyrocervicalTrunk',
         'InferiorPhrenicArtery',
         'PrimitivePulmonaryArtery',
         'HilumOfLung',
         'SellaTurcica',
         'PortalVein',
         'Scrotum',
         'PosteriorCerebralCommissure',
         'InferiorLeftPulmonaryVein',
         'UncinateFasciculus',
         'AnteriorCerebralArtery',
         'LowerLeg',
         'RenalVein',
         'InternalIliacVein',
         'IlealVein',
         'Ear',
         'Toe',
         'SpinalCord',
         'MiddleCardiacVein',
         'Pericardium',
         'MiddleCerebralArtery',
         'PosteriorCerebralArteryP2Segment',
         'PosteriorDescendingRightCoronaryArtery',
         'CorpusStriatum',
         'LeftAuricularAppendage',
         'CircleOfWillis',
         'Adenohypophysis',
         'PosteriorTibialVein',
         'SaphenopoplitealJunction',
         '_1stSeptalCoronaryArtery',
         'MuscleOfLowerLimb',
         'Eyelid',
         'AscendingAorta',
         'LeftUpperQuadrantOfAbdomen',
         'DeepPalmarArchOfRadialArtery',
         'PericardiophrenicArtery',
         'Pleura',
         'Subcostal',
         'Intima',
         'ExternalAuditoryCanal',
         'Patella',
         'SegmentOfLung',
         'OccipitalVein',
         'DoddPerforatingVein']}

name_for_cid[4263] = 'OphthalmicThicknessMapValueType'
cid_concepts[4263] = \
{'DCM': ['AbsoluteOphthalmicThickness',
         'ThicknessDeviationCategoryFromNormativeData',
         'ThicknessDeviationFromNormativeData']}

name_for_cid[10023] = 'SizeSpecificDoseEstimationMethodforCT'
cid_concepts[10023] = \
{'DCM': ['AAPM204EffectiveDiameterEstimatedFromPatientAge',
         'WaterEquivalentDiameterRepresentativeValue',
         'WaterEquivalentDiameterFromRawData',
         'WaterEquivalentDiameterIntegratedAcrossScanRange',
         'WaterEquivalentDiameterFromLocalizer',
         'AAPM204LateralDimension',
         'AAPM204SumOfLateralAndAPDimension',
         'AAPM204APDimension']}

name_for_cid[7273] = 'MRDiffusionModels'
cid_concepts[7273] = \
{'DCM': ['MonoExponentialDiffusionModel',
         'BiExponentialIVIMDiffusionModel',
         'KurtosisDiffusionModel',
         'GammaDistributionModel',
         'StretchedExponentialDiffusionModel',
         'TruncatedGaussianDiffusionModel']}

name_for_cid[6032] = 'HighRiskLesionsPathologyCodesfromBIRADS®'
cid_concepts[6032] = \
{'DCM': ['PeripheralDuctPapillomas'],
 'SCT': ['AtypicalLobularHyperplasia',
         'AtypicalIntraductalHyperplasia',
         'PhyllodesTumor',
         'LobularCarcinomaInSituOfBreast']}

name_for_cid[3230] = 'ECGFindings'
cid_concepts[3230] = \
{'SCT': ['VentricularPrematureContraction',
         'VentricularFibrillation',
         'VentricularPreExcitation',
         'AtrialPrematureContraction',
         'STDepression',
         'RightBundleBranchBlock',
         'SecondDegreeAtrioventricularBlock',
         'ThirdDegreeAtrioventricularBlock',
         'SecondarySTTAbnormality',
         'Normal',
         'EarlyRepolarization',
         'NonSustainedVentricularTachycardia',
         'IntraventricularConductionDisturbance',
         'AtrialFibrillation',
         'BifascicularBlock',
         'VentricularTachycardia',
         'STElevation',
         'SupraventricularTachycardia',
         'LeftAnteriorFascicularBlock',
         'IncompleteRightBundleBranchBlock',
         'LeftBundleBranchBlock',
         'IncompleteLeftBundleBranchBlock',
         'FirstDegreeAtrioventricularBlock',
         'NonspecificSTTAbnormality',
         'LeftPosteriorFascicularBlock']}

name_for_cid[9502] = 'RTSegmentAnnotationCategories'
cid_concepts[9502] = \
{'DCM': ['ExternalBodyModel',
         'NonSpecificVolume',
         'BrachytherapyDevice',
         'FixationOrPositioningDevice',
         'RTGeometricInformation',
         'RTDoseCalculationStructure',
         'RTTarget'],
 'SCT': ['PhysicalObject']}

name_for_cid[12264] = 'CardiacUltrasoundVenousReturnSystemicMeasurements'
cid_concepts[12264] = \
{'DCM': ['Tau',
         'IndicatorMeanTransitTime',
         'RRInterval',
         'BloodVelocityPeak',
         'BloodVelocityMean',
         'BloodVelocityMinimum',
         'SystolicBloodVelocityMean',
         'SystolicBloodVelocityPeak',
         'DiastolicBloodVelocityMean',
         'DiastolicBloodVelocityPeak'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[645] = 'ExogenousSubstanceTissueofOrigin'
cid_concepts[645] = \
{'DCM': ['EmbryonicKidney'],
 'SCT': ['Lung',
         'PulmonaryMetastasis',
         'Breast',
         'Brain',
         'AsciticFluid',
         'PleuralEffusion',
         'Lymph',
         'Ovary',
         'Bone',
         'LymphaticTissue',
         'Prostate',
         'Kidney',
         'LymphNode',
         'Colon',
         'Metastasis',
         'CentralNervousSystem']}

name_for_cid[7009] = 'PurposeofReferencetoPredecessorReport'
cid_concepts[7009] = \
{'DCM': ['PartialReport',
         'AddendedReport',
         'PreliminaryReport',
         'ReplacedReport',
         'CurrentProcedureEvidence']}

name_for_cid[7320] = 'PlanningMethods'
cid_concepts[7320] = \
{'DCM': ['Generic2DPlanning',
         'Generic3DPlanning',
         'GenericPlanningForHipReplacement',
         'GenericPlanningForKneeReplacement',
         'MüllerMethodPlanningForHipReplacement']}

name_for_cid[12008] = 'OBGYNAmnioticSac'
cid_concepts[12008] = \
{'LN': ['ThirdQuadrantDiameter',
        'SecondQuadrantDiameter',
        'FourthQuadrantDiameter',
        'FirstQuadrantDiameter'],
 'SCT': ['Diameter']}

name_for_cid[3733] = 'PrimaryCauseofDeath'
cid_concepts[3733] = \
{}

name_for_cid[6161] = 'ClinicalRegionalLymphNodeAssessmentforBreast'
cid_concepts[6161] = \
{'SCT': ['N1MetastasisIn1To3AxillaryLymphNodes',
         'N2bMetastasisInClinicallyApparentInternalNodes',
         'N3MetastasisToIpsilateralInternalMammaryLymphNodes',
         'N3cMetastasisInIpsilateralSupraclavicularLymphNodes',
         'N0NoRegionalLymphNodeMetastasisHistologically',
         'N3aMetastasisIn10OrMoreAxillaryLymphNodes',
         'NXRegionalLymphNodesCannotBeAssessed',
         'N2MetastasisIn4To9AxillaryLymphNodes',
         'N3bMetastasisInClinicallyApparentIpsilateralInternal',
         'N2aMetastasisIn4To9AxillaryLymphNodes2Point0mm']}

name_for_cid[3676] = 'LeadMeasurementTechnique'
cid_concepts[3676] = \
{'SCT': ['Representative', 'SingleBeats', 'Averaged', 'Routine', 'Median']}

name_for_cid[6024] = 'Depth'
cid_concepts[6024] = \
{'NCIt': ['Middle'], 'SCT': ['Posterior', 'Anterior']}

name_for_cid[4019] = 'PrimaryAnatomicStructureforIntraoralRadiography(DeciduousDentitionDesignationofTeeth)'
cid_concepts[4019] = \
{'SCT': ['DeciduousMandibularRightLateralIncisorTooth',
         'DeciduousMandibularLeftSecondMolarTooth',
         'DeciduousMandibularLeftCanineTooth',
         'DeciduousMaxillaryRightLateralIncisorTooth',
         'DeciduousMandibularRightSecondMolarTooth',
         'DeciduousMaxillaryLeftCentralIncisorTooth',
         'DeciduousMaxillaryLeftSecondMolarTooth',
         'DeciduousMandibularRightFirstMolarTooth',
         'DeciduousMaxillaryLeftFirstMolarTooth',
         'DeciduousMaxillaryRightSecondMolarTooth',
         'DeciduousMandibularRightCanineTooth',
         'DeciduousMandibularLeftFirstMolarTooth',
         'DeciduousMaxillaryRightCanineTooth',
         'DeciduousMandibularRightCentralIncisorTooth',
         'DeciduousMaxillaryRightCentralIncisorTooth',
         'DeciduousMaxillaryRightFirstMolarTooth',
         'DeciduousMaxillaryLeftLateralIncisorTooth',
         'DeciduousMandibularLeftLateralIncisorTooth',
         'DeciduousMandibularLeftCentralIncisorTooth',
         'DeciduousMaxillaryLeftCanineTooth']}

name_for_cid[12233] = 'CardiacPhase'
cid_concepts[12233] = \
{'SCT': ['Diastole',
         'EndDiastole',
         'VentricularIsovolumicContraction',
         'VentricularSystole',
         'EarlyDiastole',
         'Diastasis',
         'EndSystole',
         'AtrialSystole',
         'PeakSystolic',
         'VentricularEjection',
         'DiastolicRapidInflow',
         'VentricularIsovolumicRelaxation',
         'Systole']}

name_for_cid[4222] = 'OphthalmicMacularGridProblem'
cid_concepts[4222] = \
{'DCM': ['InterferingTearsOrDrops',
         'PatientPositioningProblem',
         'EquipmentFailure',
         'OperatorError',
         'PatientMovement'],
 'SCT': ['DryEyesProblem',
         'PoorVisualFixation',
         'EccentricFixation',
         'CornealOpacity',
         'EyelidDisease',
         'VitreousOpacity',
         'LensOpacity',
         'RefractiveError',
         'ConstrictedPupil']}

name_for_cid[242] = 'NormalAbnormal'
cid_concepts[242] = \
{'SCT': ['Abnormal', 'Normal', 'NormalityUndetermined']}

name_for_cid[8201] = 'SurfaceScanAcquisitionTypes'
cid_concepts[8201] = \
{'DCM': ['TimeOfFlight',
         'LaserScanning',
         'Interferometry',
         'ShapeFromShading',
         'PatternProjection',
         'ConfocalImaging',
         'ShapeFromMotion',
         'PointCloudAlgorithmic']}

name_for_cid[612] = 'AnesthesiaCategoryCodeTypefromAnesthesiaQualityInitiative(AQI)'
cid_concepts[612] = \
{'SCT': ['MonitoredAnesthesiaCareMAC',
         'GeneralAnesthesia',
         'LocalAnesthesia',
         'Sedation',
         'SpinalAnesthesia',
         'EpiduralAnesthesia',
         'RegionalAnesthesia',
         'TopicalLocalAnesthesia']}

name_for_cid[6136] = 'ChestTypesofQualityControlStandard'
cid_concepts[6136] = \
{'DCM': ['PerformanceOfPediatricAndAdultThoracicCT',
         'PerformanceOfCTForDetectionOfPulmonaryEmbolismInAdults',
         'InstitutionallyDefinedQualityControlStandard',
         'PerformanceOfHighResolutionCTOfTheLungsInAdults',
         'PerformanceOfPediatricAndAdultChestRadiographyACR',
         'ACRPositionStatement']}

name_for_cid[3764] = 'MyocardialInfarctionTherapies'
cid_concepts[3764] = \
{'SCT': ['InsertionOfCoronaryArteryStent',
         'ThrombolyticTherapy',
         'CoronaryArteryBypassGraft',
         'PercutaneousCoronaryIntervention']}

name_for_cid[10062] = 'EquivalentRadiationDoseTypes'
cid_concepts[10062] = \
{'DCM': ['MedianEquivalentRadiationDose',
         'MinimumEquivalentRadiationDose',
         'MaximumEquivalentRadiationDose',
         'ModeEquivalentRadiationDose',
         'MeanEquivalentRadiationDose']}

name_for_cid[3271] = 'HemodynamicPhysiologicalChallenges'
cid_concepts[3271] = \
{'SCT': ['LegUp',
         'ValsalvaManeuver',
         'PostVolumeChallenge',
         'HeldVentilation',
         'DrugInfusion',
         'HeadUp',
         'VagalStimulation',
         'Pacing',
         'NegativeLowerBodyPressure',
         'Handgrip',
         'HeldInspiration',
         'ExerciseChallenge']}

name_for_cid[3621] = 'FractionalFlowReserve'
cid_concepts[3621] = \
{'SCT': ['FractionalFlowReserve',
         'FractionalFlowReserveUsingIntracoronaryBolus',
         'FractionalFlowReserveUsingIntravenousInfusion']}

name_for_cid[7025] = 'RTProcessInputUsed'
cid_concepts[7025] = \
{'DCM': ['RTTreatmentPlanningInputUsed',
         'RTPrescriptionInputUsed',
         'PhysicianReviewInputUsed',
         'RTPlanSummationInputUsed',
         'PlanQualityAssuranceInputUsed',
         'PhysicistReviewInputUsed',
         'PatientSetupVerificationInputUsed',
         'MachineQualityAssuranceInputUsed',
         'RTWorkflowInputUsed']}

name_for_cid[7475] = 'GrayLevelRunLengthBasedFeatures'
cid_concepts[7475] = \
{}

name_for_cid[7160] = 'PelvicOrganSegmentationTypes'
cid_concepts[7160] = \
{'SCT': ['Testis',
         'Cervix',
         'Bladder',
         'FemaleInternalGenitalia',
         'MaleExternalGenitalia',
         'FallopianTube',
         'MaleInternalGenitalia',
         'Rectum',
         'SeminalVesicle',
         'Ovary',
         'FemaleExternalGenitalia',
         'Prostate',
         'VasDeferens',
         'Vagina',
         'Uterus']}

name_for_cid[12248] = 'CardiacUltrasoundSummaryCodes'
cid_concepts[12248] = \
{'SCT': ['CorrectedTranspositionOfGreatVessels',
         'CoarctationOfTheAorta',
         'PulmonaryAtresiaWithIntactVentricularSeptum',
         'TetralogyOfFallot',
         'CompleteTranspositionOfGreatVessels',
         'Mesocardia',
         'FunctionalSingleVentricle',
         'MitralRegurgitation',
         'HypoplasticLeftHeartSyndrome',
         'CongenitalStenosisOfAorticValve',
         'MitralStenosis',
         'AtrialSeptalDefect',
         'RheumaticFever',
         'PulmonaryAtresiaWithVentricularSeptalDefect',
         'MitralValveProlapse',
         'AorticRegurgitation',
         'Dextrocardia',
         'CommonAtrioventricularCanal',
         'VentricularSeptalDefect',
         'PartialAnomalousPulmonaryVenousConnection',
         'AcuteFebrileMucocutaneousLymphNodeSyndrome',
         'Cardiomegaly',
         'Cyanosis',
         'AtrialSeptalDefectWithEndocardialCushionDefectPartial',
         'PulmonicValveStenosis',
         'AorticStenosis',
         'InterruptedAorticArch']}

name_for_cid[3820] = 'NoninvasiveVascularProcedures'
cid_concepts[3820] = \
{'SCT': ['CardiovascularCT',
         'PeripheralAngiography',
         'CTOfAbdominalAorta',
         'VascularMRI',
         'TrunkAngiography',
         'CardiacMRI']}

name_for_cid[405] = 'MediaTypeCode'
cid_concepts[405] = \
{'DCM': ['PaperDocument',
         'URI',
         'SecureDigitalCard',
         'MultiMediaCard',
         'CompactFlash',
         'DVD',
         'CD',
         'Film',
         'Email',
         'USBDiskEmulation']}

name_for_cid[6008] = 'DensityModifier'
cid_concepts[6008] = \
{'SCT': ['LowDensityNotContainingFatLesion',
         'HighDensityLesion',
         'EqualDensityIsodenseLesion',
         'FatContainingRadiolucentLesion']}

name_for_cid[3461] = 'RegionalExtent'
cid_concepts[3461] = \
{'DCM': ['MultipleRCARegionInRAOProjection',
         'SingleRCARegionInRAOProjection',
         'MultipleLADRegionInRAOProjection',
         'SingleLADRegionInRAOProjection',
         'CFXRegionInLAOProjection',
         'LADRegionInLAOProjection',
         'RCARegionInLAOProjection']}

name_for_cid[3524] = 'BloodGasPressures'
cid_concepts[3524] = \
{'LN': ['BloodCarbonDioxidePartialPressure',
        'BloodOxygenPartialPressure',
        'VenousOxygenPartialPressure',
        'OxygenPartialPressureAt50PercentSaturationP50',
        'ArterialOxygenPartialPressureAt50PercentSaturation',
        'VenousBloodCarbonDioxidePartialPressure',
        'VenousOxygenPartialPressureAt50PercentSaturation',
        'ArterialOxygenPartialPressure',
        'ArterialBloodCarbonDioxidePartialPressure']}

name_for_cid[8104] = 'BreastTissueSpecimenTypes'
cid_concepts[8104] = \
{'SCT': ['SpecimenFromBreastObtainedByExcision',
         'SpecimenFromBreastObtainedByTotalMastectomy',
         'SpecimenFromBreastObtainedByCoreNeedleBiopsy',
         'SegmentectomyBreastSample',
         'LumpectomyBreastSample',
         'BreastTruCutBiopsySample',
         'SpecimenFromBreastByIncisionalBiopsyOfBreastMass',
         'BreastDuctSample',
         'FrozenSectionBreastSample',
         'SpecimenFromBreastStereotacticallyGuidedCoreNeedleBiopsy',
         'SpecimenFromBreastObtainedByImageGuidedCoreBiopsy']}

name_for_cid[3748] = 'AngiographicEFTestingMethod'
cid_concepts[3748] = \
{'DCM': ['BiPlaneAngiography', 'SinglePlaneAngiography']}

name_for_cid[7064] = 'ModelUsage'
cid_concepts[7064] = \
{'DCM': ['ToolFabrication',
         'PlanningIntent',
         'QualityControlIntent',
         'ImplantFabrication',
         'ProstheticFabrication',
         'SimulationIntent',
         'EducationalIntent'],
 'SCT': ['DiagnosticIntent']}

name_for_cid[628] = 'AncillaryMedicationsforSmallAnimalAnesthesia'
cid_concepts[628] = \
{}

name_for_cid[12209] = 'EchocardiographyPulmonicValve'
cid_concepts[12209] = \
{'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'StenosisPeakVelocity',
        'TimeFromQWaveToPulmonicValveCloses',
        'VolumeFlow',
        'FlowRadius',
        'PulmonicValveEjectionTime',
        'HeartRate',
        'TimeAveragedPeakVelocity',
        'EndDiastolicVelocity',
        'VelocityOfFlowPropagation',
        'PeakGradient',
        'PulmonicValveAreaByContinuity',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeAveragedMeanVelocity'],
 'SCT': ['RatioOfPulmonicValveAccelerationTimeToEjectionTime',
         'RegurgitantFraction',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter']}

name_for_cid[12120] = 'BloodVelocityMeasurementsbyUltrasound'
cid_concepts[12120] = \
{'LN': ['MinimumDiastolicVelocity',
        'TimeAveragedPeakVelocity',
        'PeakSystolicVelocity',
        'EndDiastolicVelocity',
        'TimeAveragedMeanVelocity']}

name_for_cid[94] = 'PhysicalForceAppliedDuringAcquisition'
cid_concepts[94] = \
{'SCT': ['TractionAction', 'RotationAction', 'CompressionAction']}

name_for_cid[8112] = 'SpecimenStains'
cid_concepts[8112] = \
{'SCT': ['MayGrunwaldGiemsaStain',
         'FastBlueBBSaltStain',
         'EriochromeBlueBlackSEStain',
         'CarmoisineAStain',
         'FieldStain',
         'BrazilinStain',
         'ThioflavineTStain',
         'AlkaliBlue5B4BStain',
         'GramStain',
         'AlizarinYellowRStain',
         'NaphtholASDChloracetateEsteraseStain',
         'SudanIIStain',
         'MethylGreenStain',
         'FastRedITRStain',
         'WaxolineBlueStain',
         'WaterSolubleEosinStain',
         'PonceauSStain',
         'SilverNitrateStain',
         'Chromotrope2RStain',
         'RoseBengalStain',
         'SudanStain',
         'VictoriaBlueBStain',
         'PeriodicAcidSchiffStain',
         'OilRedOStain',
         'ButyrateEsteraseStain',
         'FouchetStain',
         'MethylRedStain',
         'AzocarmineGGXStain',
         'AtebrinFSStain',
         'PeroxidaseStain',
         'FluorexonStain',
         'NeutralRedStain',
         'MeldolaBlueStain',
         'QuinacrineFluorescentStain',
         'FluoresceinSodiumStain',
         'ErythrosinBStain',
         'BismarkBrownRStain',
         'AlbertStain',
         'CarminicAcidStain',
         'AlizarinYellowGGStain',
         'FatRed7BStain',
         'AlizarinRedSStain',
         'ModifiedTrichromeStain',
         'ThionineStain',
         'VictoriaBlue4RStain',
         'NaphthaleneBlack12BStain',
         'FastGreenFCFStain',
         'MethyleneVioletStainBernthsen',
         'SaffronStain',
         'SpiritSolubleEosinStain',
         'TrypanBlueStain',
         'LeishmanStain',
         'BrilliantCresylBlueStain',
         'OrangeGStain',
         'PontamineSkyBlue6BXStain',
         'SpiritSolubleAnilineBlueStain',
         'KenacidBlueRStain',
         'VerhoeffHematoxylinStain',
         'PyronineBStain',
         'HemateinStain',
         'PontamineSkyBlue5BXStain',
         'MethylVioletStain',
         'AuramineGStain',
         'NonspecificEsteraseStain',
         'FluorescentStain',
         'PhloxinBStain',
         'SunsetYellowFCFStain',
         'BismarkBrownYStain',
         'TerminalDeoxynucleotidylTransferaseStain',
         'SporeStain',
         'ThioninStain',
         'ChromicAcidStain',
         'SolochromeAzurineBSStain',
         'FluoresceinStain',
         'IndophenolFromNaphtholStain',
         'MethylOrangeStain',
         'JennerGiemsaStain',
         'NitrazineYellowStain',
         'FastBlueBSaltStain',
         'AcidPhosphataseStain',
         'CresylVioletStain',
         'ProtargolSStain',
         'MethylViolet6BStain',
         'InsolubleBerlinBlueStain',
         'AlizarinCyanineGreenStain',
         'NaphtholGreenBStain',
         'VitalNewRedStain',
         'MetanilYellowStain',
         'MucicarmineStain',
         'ProcionBrilliantBlueMRSStain',
         'AldehydeFuchsinStain',
         'PageBlue83Stain',
         'BrilliantOrangeStain',
         'FeulgenReactionStain',
         'DiamondBlackStain',
         'JanusGreenBStain',
         'GiemsaStain',
         'AlcoholSolubleNigrosineStain',
         'MethylBlueStain',
         'HematoxylinStain',
         'WaterSolubleAnthraceneBrownStain',
         'FastGarnetGBCSaltStain',
         'BenzoFastScarletStain',
         'NeutrophilAlkalinePhosphataseStain',
         'AcriflavineStain',
         'AcridineOrangeStain',
         'AcidFastStain',
         'EthylVioletStain',
         'CurcuminStain',
         'SolubleBerlinBlueStain',
         'CochinealStain',
         'MalloryBleachStain',
         'FastSulfonBlackFStain',
         'ErythrosinYStain',
         'FastVioletBSaltStain',
         'FuchsinAcidStain',
         'LuxolFastBlueStain',
         'LissamineRhodamineStain',
         'ImmunofluorescentStain',
         'SolochromeCyanineRStain',
         'BetaGlucuronidaseStain',
         'AlizarinBlueSStain',
         'CarmineStain',
         'GallocyanineStain',
         'NightBlueStain',
         'BrilliantCroceinStain',
         'WaysonStain',
         'CrystalPonceauStain',
         'XyleneCyanolFFStain',
         'AzureAStain',
         'SolochromeDarkBlueStain',
         'WaterSolubleAnilineBlueStain',
         'BauerChromicAcidLeucofuchsinStain',
         'TropaeolinOStain',
         'ChrysoidineRStain',
         'ChloroacetateEsteraseStain',
         'AlkalinePhosphataseStain',
         'SiriusRedF3BStain',
         'SudanBlueStain',
         'ColloidalIronStain',
         'MartiusYellowStain',
         'MyeloperoxidaseStain',
         'ChromeAzurolSStain',
         'LissamineFastRedBStain',
         'AlcianBlue8GXStain',
         'LissamineFastYellowStain',
         'CongoRedStain',
         'ToluidineBlueStain',
         'CresylEchtVioletStain',
         'SudanBlackBStain',
         'ZiehlNeelsenStain',
         'AzoBlackStain',
         'RhodamineStain',
         'NileBlueStain',
         'CelestineBlueBStain',
         'AzureBStain',
         'LissamineGreenBStain',
         'TitanYellowStain',
         'CoriphosphineStain',
         'MethylViolet2BStain',
         'SafraninStain',
         'SudanIIIStain',
         'LacmoidStain',
         'RhodamineBStain',
         'AlkaliBlue6BStain',
         'AuramineStain',
         'SudanIVStain',
         'Ponceau3RStain',
         'FastBlueRRSaltStain',
         'IndiaInkStain',
         'RosolicAcidSodiumSaltStain',
         'VanGiesonStain',
         'TropaeolinOOStain',
         'AzophloxinStain',
         'BlueShadeEosinStain',
         'WaterSolubleNigrosineStain',
         'OrceinStain',
         'TrichromeStain',
         'WrightStain',
         'AzorubinSStain',
         'HanselStain',
         'AzureCStain',
         'NewFuchsinStain',
         'FlagellarStain',
         'PageBlueG90Stain',
         'RomanowskyStain',
         'LightGreenSFStain',
         'ErieGarnetStain',
         'MethyleneBlueStain',
         'MethylGreenPyroninStain',
         'DurazolRedStain',
         'CarbolFuchsinStain',
         'MethyleneVioletStain',
         'FastRedBSaltStain',
         'NaphtholYellowSStain',
         'ChrysoidineYStain',
         'SilverStain',
         'AnilineBlueStain',
         'TartrazineStain',
         'PyronineGStain',
         'BrilliantYellowStain',
         'PatentBlueVSodiumSaltStain',
         'MalachiteGreenStain',
         'AuramineOStain',
         'PermethrinStain',
         'OrangeIIStain',
         'DibromofluoresceinStain',
         'EvansBlueStain',
         'PrussianBlueStain',
         'SudanBlackStain',
         'NuclearFastRedStain',
         'FuchsinBasicStain',
         'IndigoCarmineStain',
         'BiebrichScarletStain',
         'PonceauXylidineStain',
         'ThioflavineSStain',
         'PotassiumHydroxideStain',
         'PhosphotungsticAcidHematoxylinStain',
         'FastRedTRSaltStain',
         'SolochromeBlack6BStain']}

name_for_cid[501] = 'VolumetricViewDescription'
cid_concepts[501] = \
{'SCT': ['SubcostalShortAxis',
         'RightPosteriorOblique',
         'LeftLateralProjection',
         'ParasternalLongAxisViewOfTheRVOutflowTract',
         'ParasternalShortAxisAtThePapillaryMuscleLevel',
         'Tangential',
         'CranioCaudal',
         'LateroMedialOblique',
         'Acanthioparietal',
         'SuprasternalSagittalView',
         'Orbitoparietal',
         'PosteriorProjection',
         'FiveChamber',
         'ObliqueCranioCaudal',
         'LongAxis',
         'RightAnteriorOblique',
         'RightVentricularInflowTractView',
         'LateralOblique',
         'RightOblique',
         'OffAxis',
         'TwoChamber',
         'Lateral',
         'MedioLateralOblique',
         'RightLateral',
         'Verticosubmental',
         'ObliqueAxial',
         'MedialLateral',
         'SubcostalObliqueCoronalView',
         'Frontal',
         'Parietoacanthial',
         'PosteroAnterior',
         'Submentovertical',
         'HorizontalLongAxis',
         'AnteroPosteriorOblique',
         'RightPosteriorOblique',
         'LeftOblique',
         'SubcostalShortAxisViewAtMitralValveLevel',
         'ParietoOrbital',
         'MedialOblique',
         'FrontalObliqueAxial',
         'ParasternalShortAxisAtTheAorticValveLevel',
         'SubcostalShortAxisViewAtPapillaryMuscleLevel',
         'AnteroPosterior',
         'SubcostalShortAxisViewAtVenousInflowLevel',
         'FrontalOblique',
         'LeftPosteriorOblique',
         'FourChamber',
         'ParasternalShortAxisAtTheLevelOfTheMitralChords',
         'LeftPosteriorOblique',
         'RightVentricularOutflowTractView',
         'ParasternalShortAxis',
         'ParasternalShortAxisAtTheMitralValveLevel',
         'ParasternalLongAxis',
         'ObliqueCaudoCranial',
         'SuprasternalShortAxis',
         'SubcostalViewOfCardiacOutletsDirectedAnteriorly',
         'SagittalObliqueAxial',
         'PosteroAnteriorOblique',
         'VerticalLongAxis',
         'Axial',
         'Sagittal',
         'LeftAnteriorOblique',
         'OcclusalProjection',
         'SuprasternalLongAxis',
         'TransesophagealShortAxisView',
         'Axial',
         'Plantodorsal',
         'SuprasternalCoronalView',
         'Parasagittal',
         'LeftAnteriorOblique',
         'ParasternalLongAxisViewOfTheRVInflowTract',
         'Oblique',
         'RightLateralProjection',
         'SuprasternalLongAxisViewOfAorticArch',
         'CaudoCranial',
         'Transverse',
         'TissueSpecimen',
         'SubcostalLongAxis',
         'LeftLateral',
         'RightAnteriorOblique',
         'LateralMedial',
         'Dorsoplantar',
         'ApicalLongAxis',
         'SubcostalShortAxisViewAtAorticValveLevel',
         'Oblique',
         'AnteriorProjection',
         'SagittalObliqueAxial',
         'LateralMedial',
         'MedialLateral',
         'ShortAxis',
         'MedioLateralOblique',
         'LateroMedialOblique',
         'ObliqueAxial',
         'Lateral',
         'Coronal',
         'Longitudinal']}

name_for_cid[9538] = 'RadiationTransportBasedEffectiveDoseMethodModifiers'
cid_concepts[9538] = \
{'DCM': ['LocalEffectModel', 'MicrodosimetricKineticModel']}

name_for_cid[3836] = 'TimebasedPerfusionMeasurements'
cid_concepts[3836] = \
{'DCM': ['SignalTimeToPeak',
         'SignalIncreaseStartTime',
         'SignalEarliestPeakTime',
         'SignalBaselineStart',
         'SignalBaselineEnd']}

name_for_cid[9241] = 'RadiotherapyGeneralWorkitemDefinition'
cid_concepts[9241] = \
{'DCM': ['RTPatientSetup',
         'RTPatientPositionInTreatmentSessionReview',
         'RTTreatmentQAByRTIonPlanDifferenceCheck',
         'RTTreatmentSimulationWithInternalVerification',
         'RTBrachyTreatment',
         'RTTreatmentQAByRTPlanDifferenceCheck',
         'RTPatientPositionAdjustment',
         'RTTreatmentQAByRTIonPlanDoseCheck',
         'RTMachineQA',
         'RTTreatmentQAByRTPlanDoseCheck',
         'RTTreatmentQAWithExternalVerification',
         'RTTreatmentWithExternalVerification',
         'RTTreatmentQAWithInternalVerification',
         'RTTreatmentSimulationWithExternalVerification',
         'RTTreatmentWithInternalVerification']}

name_for_cid[12032] = 'TransducerScanPattern'
cid_concepts[12032] = \
{'DCM': ['VolumeScanPattern', 'PlaneScanPattern', 'LineScanPattern']}

name_for_cid[3709] = 'AorticRootDescription'
cid_concepts[3709] = \
{'SCT': ['EnlargedAorticRoot',
         'RupturedSinusOfValsalva',
         'PostStenoticDilation',
         'NormalAorticRoot',
         'AneurysmOfAorticRoot',
         'AnnularAbscessOfAorticRoot']}

name_for_cid[224] = 'SelectionMethod'
cid_concepts[224] = \
{'DCM': ['MostRecentValueChosen', 'MeanValueChosen', 'UserChosenValue']}

name_for_cid[7468] = 'TextureMeasurements'
cid_concepts[7468] = \
{'DCM': ['RootAngularSecondMomentOfGLCM', 'FractalDimension']}

name_for_cid[7192] = 'AnatomicalStructureSegmentationPropertyTypes'
cid_concepts[7192] = \
{'BARI': ['_1stMarginalCoronaryArteryLaterals',
          '_2ndMarginalCoronaryArteryLaterals',
          '_3rdMarginalCoronaryArteryLaterals',
          '_1stDiagonalCoronaryArteryLaterals',
          '_2ndDiagonalCoronaryArteryLaterals',
          'RamusLaterals',
          '_3rdDiagonalCoronaryArteryLaterals',
          'PosteriorDescendingSeptalPerforators'],
 'DCM': ['Phantom',
         'AorticKnob',
         'ArchOfTheAzygosVein',
         'SecondaryPulmonaryLobule',
         'Hiatus',
         'AzygoesophagealRecessInterface',
         'ParaspinalLine',
         'PosteriorTrachealStripe',
         'RightTrachealStripe',
         'Stripe',
         'MidlungWindow',
         'CarinaAngle',
         'CentrilobularStructures',
         'AnteriorJunctionLine',
         'PosteriorJunctionLine',
         'SurfaceOfTheChoroidScleraInterface',
         'OuterSurfaceOfTheBM',
         'OuterSurfaceOfTheCC',
         'PosteriorSurfaceOfTheRPE',
         'SurfaceOfTheCenterOfTheRPE',
         'OuterSurfaceOfIPL',
         'OuterSurfaceOfGCL',
         'OuterSurfaceOfOPL',
         'OuterSurfaceOfINL',
         'SurfaceBetweenInnerAndOuterSegmentsOfThePhotoreceptors',
         'OuterSurfaceOfHFL',
         'AnteriorSurfaceOfTheRPE',
         'SurfaceOfTheInterdigitatingZoneBetweenRetinaAndRPE',
         'OuterSurfaceOfRNFL'],
 'FMA': ['ArcuateFasciculus'],
 'SCT': ['ProfundaFemorisVein',
         'LeftHypochondriacRegion',
         'SternocleidomastoidMuscle',
         'TruncusCoeliacus',
         'InferiorLongitudinalFasciculus',
         'BasilarArtery',
         'ParanasalSinus',
         'BrainCerebrospinalFluidPathway',
         'WristJoint',
         'MidCircumflexCoronaryArtery',
         'CentralRetinalVein',
         'IntermediateArteryRamus',
         'GenicularArtery',
         'LeftVentricleInflow',
         'FontanelOfSkull',
         'SigmoidColon',
         'GastricArtery',
         'SubscapularisMuscle',
         'ExternalEar',
         'LeftVentricleOutflowTract',
         'RightFemoralArtery',
         'Intracranial',
         'JawRegion',
         'Calcaneus',
         'SolealVein',
         'IntercostalArtery',
         'PeripheralNerve',
         'LobarArtery',
         'TeresMajorMuscle',
         'PatentDuctusArteriosus',
         'ThighPerforator',
         'MesentericVein',
         'Pituitary',
         'MaleInternalGenitalia',
         'MidRightCoronaryArtery',
         'PalatineBone',
         'CoronaryArtery',
         'RightBranchOfHepaticArtery',
         'UpperLimb',
         'DuraMater',
         'ExternalIliacVein',
         'CongenitalPulmonaryArteriovenousFistula',
         'DescendingAorta',
         'ParietalBone',
         'Flank',
         'Maxilla',
         'EsophagealHiatus',
         'CentralRetinalArtery',
         'Neurohypophysis',
         'SegmentalArtery',
         'BrachialVein',
         'RightUpperQuadrantOfAbdomen',
         'Eye',
         'SuperficialPalmarArch',
         'Prostate',
         'HepaticArtery',
         'FacialBones',
         'Back',
         'PeronealArtery',
         'Duodenum',
         'RightHypochondriacRegion',
         '_1stMarginalCoronaryArtery',
         'SuperiorMesentericVein',
         'LumbarArtery',
         'Cornea',
         'Larynx',
         'PeronealVein',
         'FemoralArtery',
         'Ovary',
         'ThoracoLumbarSpine',
         'Urethra',
         'CoronaryArteryGraft',
         'Head',
         'AdrenalGland',
         'MitralValve',
         'AcromioclavicularJoint',
         'UmbilicalRegion',
         'SupraspinatusMuscle',
         'CongenitalCoronaryArteryFistulaToRightVentricle',
         'Calyx',
         'PosteriorMedialTributary',
         'PulmonaryVein',
         'SoftPalate',
         'Submental',
         'ExternalIntercostalMuscle',
         'CranialVenousSystem',
         'NucleusAccumbens',
         'GiacominiVein',
         'PoplitealFossa',
         'MiddleCerebellarPeduncle',
         'PulmonaryArteryConduit',
         'InternalCarotidArteryC5Segment',
         'FissureOfLung',
         'RightVentricle',
         'LevatoresCostarumMuscles',
         'FrontalLobe',
         'VagusNerve',
         'Adventitia',
         'Abdomen',
         'HeadAndNeck',
         'Hypopharynx',
         'RightCoronaryArteryOstium',
         'HepaticVein',
         'MarginalCoronaryArtery',
         'PelvisAndLowerExtremities',
         'InternalMammaryArtery',
         'Bladder',
         'Vulva',
         'LargeIntestine',
         'EntorhinalCortex',
         'Forearm',
         'BronchialArtery',
         'NeckChestAndAbdomen',
         'ExternalCarotidArtery',
         'SuperiorMesentericArtery',
         'TemporalLobe',
         'SystemicArtery',
         'PosteriorTibialArtery',
         'FemoralVein',
         'RightAuricularAppendage',
         'GreatSaphenousVeinOfThigh',
         'LeftHepaticVein',
         'AbdominalWallMuscle',
         'SuperficialFemoralArtery',
         'VomerBone',
         'UterusAndFallopianTubes',
         'Arachnoid',
         '_2ndMarginalCoronaryArtery',
         'SuperiorCerebellarPeduncle',
         'AorticArch',
         'Carina',
         'Telencephalon',
         'LeftBranchOfHepaticArtery',
         'ArcuateArteryOfTheKidney',
         'Heart',
         'TarsalJoint',
         'CephalicVein',
         'LingualArtery',
         'Lip',
         'GastricVein',
         'AirwayStructure',
         'CongenitalCoronaryArteryFistulaToRightAtrium',
         'DeltoidMuscle',
         'LeftMainCoronaryArtery',
         'CerebralFornix',
         'Penis',
         'Baffle',
         'InferiorRightHepaticVein',
         'VentricularVein',
         'CarotidSiphon',
         'SystemicVenousAtrium',
         'PlantarArterialArch',
         'Parathyroid',
         'AnomalousPulmonaryVein',
         'CervicoThoracicSpine',
         'LeftMainCoronaryArteryOstium',
         'Scapula',
         'LesserSaphenousVein',
         'Kidney',
         'AutonomicNerve',
         'GreatSaphenousVein',
         'OmentalBursa',
         'InternalCarotidArtery',
         'Ureter',
         'RightPosteriorAVCoronaryArtery',
         'UlnarVein',
         'FascialLayer',
         'AorticValve',
         'MiddleEar',
         'InternalAuditoryCanal',
         'RightPulmonaryArtery',
         'RenalPelvis',
         'ProperHepaticArtery',
         'InterlobarArteryOfKidney',
         'Thrombus',
         'VenousNetwork',
         'SubcostalMuscle',
         'CorpusCallosum',
         'CommonAtrium',
         'PerforatingArteryOfKidney',
         'Humerus',
         'Ilium',
         'AorticFistula',
         'InternalCarotidArteryC6Segment',
         'ILMInternalLimitingMembrane',
         'NasalBone',
         'AnteriorTibialArtery',
         'CranialNerve',
         'Cervix',
         'SkinOfAbdomen',
         'LeftAnteriorDescendingCoronaryArtery',
         'Artery',
         'CommonCarotidArtery',
         'Buttock',
         'InnermostIntercostalMuscles',
         '_3rdMarginalCoronaryArtery',
         'FallopianTube',
         'MastoidBone',
         'Vein',
         'CommonIliacArteryBifurcation',
         'MiddleHepaticVein',
         'CerebrospinalFluid',
         'EsophagealArtery',
         'Thigh',
         'RadialVein',
         'InferiorRightPulmonaryVein',
         'Trachea',
         'PulmonaryValveSinuses',
         'LeftPosterolateralCircumflexCoronaryArtery',
         'UterineArtery',
         'BodyConduit',
         'Thalamus',
         'EndoVesical',
         'RightVentricleInflow',
         'MesentericArtery',
         'LeftLowerQuadrantOfAbdomen',
         'MuscleOfHead',
         'SpinalCerebrospinalFluidPathway',
         'MiddleCerebralArteryM1Segment',
         'SesamoidBonesOfFoot',
         'InternalIliacArtery',
         'SuperiorVenaCava',
         'SixthLumbarArtery',
         'PosteriorArchVein',
         'AnteriorSpinalArtery',
         'AnusRectumAndSigmoidColon',
         'Omentum',
         'SplenicVein',
         'Colon',
         'SubmandibularGland',
         'CentralNervousSystem',
         'CarotidBifurcation',
         'LeftLumbarRegion',
         'InnominateArtery',
         'SplenicArtery',
         'FemaleInternalGenitalia',
         'Extremity',
         'FacialArtery',
         'AccessoryRenalArtery',
         'BoneStructureOfHeadAndOrNeck',
         'EpigastricRegion',
         'Media',
         'SeminalVesicle',
         'DistalCircumflexCoronaryArtery',
         'AbdominalCavity',
         'TeresMinorMuscle',
         'BoneOfUpperLimb',
         'DorsalisPedisArtery',
         'Vagina',
         'LongissimusMuscle',
         'LowerInnerQuadrantOfBreast',
         'Clavicle',
         'ParotidGland',
         'AorticIsthmus',
         'CommonFemoralVein',
         'SubstantiaNigra',
         '_1stLeftPosterolateralCoronaryArtery',
         'Subxiphoid',
         'UpperInnerQuadrantOfBreast',
         'Diaphragm',
         'ThoracicAorta',
         'RightInguinalRegion',
         'VasDeferens',
         'AxillaryArtery',
         'LeftPulmonaryArtery',
         'Ileum',
         'SalivaryGland',
         'VertebralArtery',
         'TransjugularIntrahepaticPortosystemicShunt',
         'RightLowerQuadrantOfAbdomen',
         'SupraRenalAorta',
         'InternalJugularVein',
         'LacrimalArteryOfRightEye',
         'CockettPerforatingVein',
         'ChordaeTendineaeCordis',
         'LumboSacralSpine',
         'SacroiliacJoint',
         'RightMainBronchus',
         'ScalenousAnteriorMuscle',
         'AVGrooveContinuationOfCircumflexArtery',
         'EntireBody',
         'TesticularArtery',
         'Thumb',
         'HypogastricRegion',
         'RightLumbarRegion',
         'GreatCardiacVein',
         'HyoidBone',
         'ThoracicDuct',
         'MiddleLobeOfRightLung',
         'BoydPerforatingVein',
         'Peritioneum',
         'CommonIliacArtery',
         'SupraclavicularRegionOfNeck',
         'Meninges',
         'Shoulder',
         'TransversusThoracis',
         'VascularGraft',
         'CerebellarWhiteMatter',
         'InterventricularSeptum',
         '_2ndLeftPosterolateralCoronaryArtery',
         'MiddleCerebralArteryM2Segment',
         'Cingulum',
         'LowerLobeOfLung',
         'TesticularVein',
         'AnteriorTibialVein',
         'LacrimalBone',
         'BasilicVein',
         'LeftInguinalRegion',
         'RightCoronaryArtery',
         'HunterianPerforatingVein',
         'IntraPelvic',
         'Putamen',
         'ProfundaFemorisArtery',
         'CostocervicalTrunk',
         'Pharynx',
         'FifthLumbarArtery',
         'AxillaryVein',
         'LigamentumArteriosum',
         'Nasopharynx',
         'SuprapulmonicValveArea',
         'SympatheticTrunk',
         'LobeOfLung',
         'SubclavianVein',
         'PulmonaryValve',
         'ExternalIliacArtery',
         'ProximalRightCoronaryArtery',
         'TemporalBone',
         'BoneOfLowerLimb',
         'Midbrain',
         'OvarianVein',
         'PosteriorCerebralArtery',
         'BrachialPlexus',
         'ErectorSpinaeMuscle',
         'DeepPalmarVenousArch',
         'ApexOfRightVentricle',
         'LateralCalfPerforator',
         'Mouth',
         'PharynxAndLarynx',
         'CarotidArtery',
         'FourthLumbarArtery',
         'SmallestCardiacVein',
         'ThirdLumbarArtery',
         'LimbicLobe',
         'SkinOfChest',
         'UmbilicalArtery',
         'BuccalRegionOfFace',
         'LymphNode',
         'PrimitiveAorta',
         'RadialArtery',
         'LeftFemoralArtery',
         'LumenOfBloodVessel',
         'TrapeziusMuscle',
         'Esophagus',
         'AntecubitalVein',
         'SpinalCordGrayMatter',
         'Mandible',
         'LateralCorticospinalTract',
         'Parasternal',
         'LeftMainBronchus',
         'BroadLigament',
         'SystemicVein',
         'OccipitalLobe',
         'CerebralAqueduct',
         'Insula',
         'PulmonaryVein',
         'MuscleOfNeck',
         'SubclavianArtery',
         'Amygdala',
         'Lung',
         '_3rdLeftPosterolateralCoronaryArtery',
         'OpticChiasm',
         'OphthalmicArtery',
         'Nose',
         'PancreaticDuct',
         'Hand',
         'PectoralisMajorMuscle',
         'CardiovascularSystem',
         'FemaleExternalGenitalia',
         'OvarianArtery',
         'Chest',
         'ElbowJoint',
         'Hippocampus',
         'Sternum',
         'NeckAndChest',
         'RightVentricleOutflowTract',
         'GlobusPallidus',
         'SuperficialFemoralVein',
         'SuperiorLeftPulmonaryVein',
         'LeftPosteriorDescendingCircumflexCoronaryArtery',
         'BileDuct',
         'SystemicCollateralArteryToLung',
         'UpperOuterQuadrantOfBreast',
         'PancreaticDuctAndBileDuctSystems',
         'ApexOfLeftVentricle',
         'PericardialCavity',
         'PulmonaryArtery',
         'Skull',
         'MidLeftAnteriorDescendingCoronaryArtery',
         'AtrioventricularVein',
         'Thyroid',
         'EthmoidBone',
         'OpticRadiation',
         'InfraRenalAorta',
         'SternoclavicularJoint',
         'EndoNasopharyngeal',
         'EsophagusStomachAndDuodenum',
         'Breast',
         'Neck',
         'AxillaryFascia',
         'CarotidBulb',
         'Finger',
         'PoplitealVein',
         'InfraspinatusMuscle',
         'PiaMater',
         'TricuspidValve',
         'RightAtrium',
         'VascularSclerosis',
         'SuperiorPhrenicArtery',
         'CommonVentricle',
         'DistalLeftAnteriorDescendingCoronaryArtery',
         'InnerEar',
         'Coccyx',
         'AtrialSeptalDefect',
         'RightPulmonaryVein',
         'Uterus',
         'TemporomandibularJoint',
         'Ventricle',
         'DescendingAorta',
         'TracheaAndBronchus',
         'Testis',
         'ProximalLeftAnteriorDescendingCoronaryArtery',
         'OccipitalBone',
         'Eyeball',
         'Face',
         'Spine',
         'UrinarySystem',
         'BrainStem',
         'Foot',
         'WhiteMatter',
         'MuscleOfUpperLimb',
         'PeripheralNervousSystem',
         'PinealGland',
         'UpperLobeOfLung',
         'Zygoma',
         'InnominateVein',
         'ProximalCircumflexCoronaryArtery',
         'LowerLimb',
         'PosteriorCommunicationArtery',
         'JuxtaposedAtrialAppendage',
         'VertebralColumnAndCranium',
         'Jejunum',
         'CarotidBody',
         'ChestAbdomenAndPelvis',
         'AbdominalAorta',
         'RightHepaticVein',
         'GastrocnemiusVein',
         'Scalp',
         'Tongue',
         'ChestWallMuscle',
         'GrayMatter',
         'Fibula',
         'SuperiorLongitudinalFasciculus',
         'OrbitalStructure',
         'InferiorCardiacVein',
         'LeftMainBranchOfPortalVein',
         'InternalIntercostalMuscle',
         'LateralVentricle',
         'CaudateNucleus',
         'LacrimalArtery',
         'OccipitalArtery',
         'Brain',
         'BloodVessel',
         'Sclera',
         'AnteriorPosteriorCerebralArteryBifurcation',
         'AtrialVein',
         '_1stDiagonalCoronaryArtery',
         'SuprasternalNotch',
         'Perineum',
         'InteratrialSeptum',
         'Rectum',
         'Femur',
         'AnteriorCardiacVein',
         'Pancreas',
         'BoneOfThorax',
         'LeftAtrium',
         'CerebralArtery',
         'InferiorMesentericArtery',
         'DigitalArteryOfHand',
         'SecondLumbarArtery',
         'IliocostalisMuscle',
         '_1stRightPosterolateralCoronaryArtery',
         'BladderAndUrethra',
         'CervicalSpine',
         'IliacAndOrFemoralArtery',
         'SpinalCordWhiteMatter',
         'Sacrum',
         'SmallIntestine',
         'SuperiorRightPulmonaryVein',
         'InferiorVenaCava',
         'MaleExternalGenitalia',
         'IntraArticular',
         'Axilla',
         'RetroperitonealSpace',
         'Thymus',
         'ELMExternalLimitingMembrane',
         'AnteriorMiddleCerebralArteryBifurcation',
         'ParietalLobe',
         'SuperiorThyroidArtery',
         'IliacArtery',
         'DistalRightCoronaryArtery',
         'CerebralCortex',
         'TruncusArteriosusCommunis',
         'Tooth',
         'Joint',
         'SubmandibularArea',
         'GreatSaphenousVeinOfCalf',
         'LeftPulmonaryVein',
         'PectoralisMinorMuscle',
         'OpticCanal',
         'Aorta',
         'SuperficialTemporalArtery',
         'MedialLemniscus',
         'Diencephalon',
         'CongenitalCoronaryArteryFistulaToLeftVentricle',
         'NeckChestAbdomenAndPelvis',
         'InferiorCerebellarPeduncle',
         'RightMainBranchOfPortalVein',
         'SphenoidBone',
         'LatissimusDorsiMuscle',
         'Spleen',
         'ChestAndAbdomen',
         'CommonHepaticArtery',
         'AnkleJoint',
         'LowerOuterQuadrantOfBreast',
         'Mediastinum',
         'Cerebellum',
         'FirstLumbarArtery',
         'PoplitealArtery',
         'UpperUrinaryTract',
         'PulmonaryVenousAtrium',
         'Bronchus',
         'Pelvis',
         'PulmonaryChamberOfCorTriatriatum',
         'DorsalScapularArtery',
         'UlnarArtery',
         'LeftVentricle',
         'AbdomenAndPelvis',
         'LumbarRegion',
         'CostalCartilage',
         'ThoracicSpine',
         'FourthVentricle',
         'SaphenousVein',
         'Gallbladder',
         '_2ndRightPosterolateralCoronaryArtery',
         'RenalArtery',
         'PeritonealCavity',
         'UpperArm',
         'OpticTract',
         'HipJoint',
         'TrunkOfPulmonaryArtery',
         'GastroduodenalArtery',
         'AzygosVein',
         'BrachialArtery',
         '_2ndDiagonalCoronaryArtery',
         'Liver',
         'HipJoint',
         'Rib',
         'SerratusAnteriorMuscle',
         'CommonFemoralArtery',
         'SmallCardiacVein',
         'SuperficialPalmarVenousArch',
         'Vertebra',
         'Stomach',
         'CircumflexCoronaryArtery',
         'AnteriorCommunicatingArtery',
         'CongenitalCoronaryArteryFistulaToLeftAtrium',
         'InferiorMesentericVein',
         'MedianCubitalVein',
         'CoronarySinus',
         'HilarArtery',
         'TerminalInternalCarotidArtery',
         'ApexOfLung',
         'ThirdVentricle',
         'Atrium',
         'CerebralWhiteMatter',
         'Knee',
         '_3rdDiagonalCoronaryArtery',
         'SpinalisMuscle',
         'SubarachnoidSpace',
         'PulmonaryVeinConfluence',
         'SaphenofemoralJunction',
         'EyeRegion',
         'BrainVentricle',
         'Anus',
         'UmbilicalVein',
         'FrontalBone',
         'CommonIliacVein',
         'TrabeculaeCarnae',
         'InguinalRegion',
         'PosteriorCerebralArteryP1Segment',
         '_3rdRightPosterolateralCoronaryArtery',
         'EndoNasal',
         'LumbarSpine',
         'ThyrocervicalTrunk',
         'InferiorPhrenicArtery',
         'PrimitivePulmonaryArtery',
         'HilumOfLung',
         'SellaTurcica',
         'PortalVein',
         'Scrotum',
         'PosteriorCerebralCommissure',
         'InferiorLeftPulmonaryVein',
         'UncinateFasciculus',
         'AnteriorCerebralArtery',
         'LowerLeg',
         'RenalVein',
         'InternalIliacVein',
         'IlealVein',
         'Ear',
         'Toe',
         'SpinalCord',
         'MiddleCardiacVein',
         'Pericardium',
         'MiddleCerebralArtery',
         'PosteriorCerebralArteryP2Segment',
         'PosteriorDescendingRightCoronaryArtery',
         'CorpusStriatum',
         'LeftAuricularAppendage',
         'CircleOfWillis',
         'Adenohypophysis',
         'PosteriorTibialVein',
         'SaphenopoplitealJunction',
         '_1stSeptalCoronaryArtery',
         'MuscleOfLowerLimb',
         'Eyelid',
         'AscendingAorta',
         'LeftUpperQuadrantOfAbdomen',
         'DeepPalmarArchOfRadialArtery',
         'PericardiophrenicArtery',
         'Pleura',
         'Subcostal',
         'Intima',
         'ExternalAuditoryCanal',
         'Patella',
         'SegmentOfLung',
         'OccipitalVein',
         'DoddPerforatingVein']}

name_for_cid[4244] = 'OphthalmicAgentConcentrationUnits'
cid_concepts[4244] = \
{'UCUM': ['Percent', 'mgml']}

name_for_cid[9533] = 'DeliveryTimeStructure'
cid_concepts[9533] = \
{'DCM': ['ContinuousPermanent',
         'ContinuousTemporary',
         'HyperFractionation',
         'HypoFractionation',
         'StandardFractionation',
         'SingleFraction']}

name_for_cid[12255] = 'CardiacUltrasoundMyocardiumMeasurements'
cid_concepts[12255] = \
{'DCM': ['WallMass'],
 'LN': ['EndocardialArea', 'EpicardialArea', 'MyocardialPerformanceIndexTei']}

name_for_cid[3217] = 'ComparisonFinding'
cid_concepts[3217] = \
{'DCM': ['AgreementWithPriorFindings', 'DisagreementWithPriorFindings']}

name_for_cid[6015] = 'SingleImageFindingfromBIRADS®'
cid_concepts[6015] = \
{'DCM': ['CooperLigamentChanges', 'MassInTheSkin', 'MassOnTheSkin'],
 'SCT': ['BreastComposition',
         'AxillaryAdenopathy',
         'Edema',
         'AxillaryLymphNode',
         'TrabecularThickeningOfBreast',
         'SkinRetractionOfBreast',
         'ArchitecturalDistortionOfBreast',
         'TubularDensity',
         'MammographyBreastDensity',
         'SkinLesion',
         'IndividualCalcification',
         'SkinThickeningOfBreast',
         'CalcificationCluster',
         'IntramammaryLymphNode']}

name_for_cid[4028] = 'CraniofacialAnatomicRegions'
cid_concepts[4028] = \
{'SCT': ['TemporalBone',
         'VomerBone',
         'HyoidBone',
         'ParanasalSinus',
         'SoftPalate',
         'MiddleEar',
         'InternalAuditoryCanal',
         'Neck',
         'MuscleOfHead',
         'VertebralColumnAndCranium',
         'Lip',
         'Tongue',
         'InnerEar',
         'ExternalEar',
         'FacialBones',
         'HeadAndNeck',
         'SubmandibularGland',
         'JawRegion',
         'TemporomandibularJoint',
         'NasalBone',
         'OrbitalStructure',
         'SalivaryGland',
         'OccipitalBone',
         'Tooth',
         'LacrimalBone',
         'BoneStructureOfHeadAndOrNeck',
         'Mandible',
         'Eyeball',
         'EyeRegion',
         'Larynx',
         'OpticCanal',
         'MastoidBone',
         'Skull',
         'FrontalBone',
         'SphenoidBone',
         'EthmoidBone',
         'Head',
         'PalatineBone',
         'Trachea',
         'MuscleOfNeck',
         'ParietalBone',
         'Maxilla',
         'CervicalSpine',
         'Zygoma']}

name_for_cid[10004] = 'FluoroModes'
cid_concepts[10004] = \
{'DCM': ['Continuous', 'Pulsed']}

name_for_cid[27] = 'BasicCardiacViews'
cid_concepts[27] = \
{'SCT': ['VerticalLongAxis', 'ShortAxis', 'HorizontalLongAxis']}

name_for_cid[1011] = 'ReferenceGeometryPoints'
cid_concepts[1011] = \
{'DCM': ['CenterOfMass', 'GeometricCenterpoint']}

name_for_cid[623] = 'MedicationforSmallAnimalAnesthesia'
cid_concepts[623] = \
{'SCT': ['Chloroform',
         'Succinylcholine',
         'Pancuronium',
         'Lidocaine',
         'Pentobarbital',
         'Xylazine',
         'Chlorpromazine',
         'Diazepam',
         'Enflurane',
         'Urethane',
         'Acepromazine',
         'Thiamylal',
         'Sevoflurane',
         'Etomidate',
         'Tribromoethanol',
         'Methohexital',
         'Bupivacaine',
         'LidocainePrilocaine',
         'Azaperone',
         'Halothane',
         'Tiletamine',
         'Zolazepam',
         'Isoflurane',
         'Propofol',
         'Midazolam',
         'Desflurane',
         'Alphachloralose',
         'ChloralHydrate',
         'Methoxyflurane',
         'Alphadolone',
         'Butabarbital',
         'Ketamine',
         'DiethylEther',
         'MedicalAir',
         'Thiopental',
         'CarbonDioxide',
         'Droperidol',
         'NitrousOxide',
         'OxygenGas'],
 'UMLS': ['RoomAir', 'Alphaxalone', 'Fluanisone', 'Metomidate']}

name_for_cid[12202] = 'LeftVentricleVolume'
cid_concepts[12202] = \
{'LN': ['LeftVentricularEjectionFractionByUS',
        'LeftVentricularEndSystolicVolume',
        'LeftVentricularEndDiastolicVolume']}

name_for_cid[3493] = 'IVUSRelativeStenosisSeverities'
cid_concepts[3493] = \
{'DCM': ['T2Secondary', 'T3Secondary', 'T1Worst', 'T4Secondary']}

name_for_cid[7203] = 'ImageDerivation'
cid_concepts[7203] = \
{'DCM': ['SpatialResampling',
         'PixelByPixelAddition',
         'EdgeEnhancement',
         'Smoothing',
         'GaussianBlur',
         'UnsharpMask',
         'PixelByPixelDivision',
         'PixelByPixelMask',
         'PixelByPixelMaximum',
         'LossyCompression',
         'PixelByPixelSubtraction',
         'PerfusionImageAnalysis',
         'DiffusionTractography',
         'DiffusionImageAnalysis',
         'PixelByPixelMean',
         'VolumeRendering',
         'PixelByPixelMultiplication',
         'SurfaceRendering',
         'MultiEnergyProportionalWeighting',
         'Segmentation',
         'ExtractionOfIndividualSubjectFromGroup',
         'VolumeEditing',
         'MaximumIntensityProjection',
         'MinimumIntensityProjection',
         'ImageStitching',
         'SpatiallyRelatedFramesExtractedFromTheVolume',
         'TemporallyRelatedFramesExtractedFromTheSetOfVolumes',
         'PolarToRectangularScanConversion',
         'MetaboliteMapsFromSpectroscopyData',
         'MultiplanarReformatting',
         'PixelByPixelMinimum',
         'CurvedMultiplanarReformatting',
         'OCTBScanAnalysis']}

name_for_cid[3610] = 'GradientSourceLocations'
cid_concepts[3610] = \
{'SCT': ['TricuspidValve',
         'RightVentricleOutflowTract',
         'PulmonaryValve',
         'PulmonaryArtery',
         'CoarctationOfAorta',
         'AorticValve',
         'AtrialSeptalDefect',
         'MitralValve',
         'VentricularSeptalDefect',
         'LeftVentricleOutflowTract']}

name_for_cid[6042] = 'StatusofResults'
cid_concepts[6042] = \
{'DCM': ['Succeeded', 'PartiallySucceeded', 'Failed', 'NotAttempted']}

name_for_cid[3240] = 'ElectrophysiologyMeasurementFunctionsandTechniques'
cid_concepts[3240] = \
{'DCM': ['PacingElectricalStimulusVoltage',
         'MonopoleSignal',
         'HisBundleElectrogram',
         'DifferentialSignal',
         'VoltageMeasurementByMappingCatheter',
         'VoltageMeasurementByBasketCatheter',
         'VoltageMeasurement',
         'RadioFrequencyAblationPower']}

name_for_cid[3755] = 'CathComplications'
cid_concepts[3755] = \
{'SCT': ['EmergencyPercutaneousCoronaryIntervention',
         'CardiogenicShock',
         'Arrhythmia',
         'ContrastMediaAdverseReaction',
         'CongestiveHeartFailure',
         'CerebrovascularAccidentOrStroke',
         'CardiacTamponade',
         'EmergencyCoronaryArteryBypass',
         'RenalFailure',
         'CardiacArrest']}

name_for_cid[6107] = 'WidthDescriptor'
cid_concepts[6107] = \
{'DCM': ['Vasoconstriction', 'Vasodilation'], 'SCT': ['Narrow', 'Enlarged']}

name_for_cid[70] = 'Flush'
cid_concepts[70] = \
{'SCT': ['Dextran', 'Saline', 'LactatedRingerS']}

name_for_cid[635] = 'HeatingMethod'
cid_concepts[635] = \
{'DCM': ['HeatedWaterBlanket',
         'PreHeatedPad',
         'Unheated',
         'ForcedAirHeater',
         'HeatedImagingDevice',
         'HeatedPatientSupport'],
 'SCT': ['NonElectricHeatingPad',
         'ForcedAirWarmingBlanket',
         'ElectricHeatingPad',
         'WarmerDevice',
         'AirHeatingPad',
         'ElectricBlanket',
         'WarmingBlanket'],
 'UMLS': ['HeatLamp']}

name_for_cid[4205] = 'OphthalmicLens'
cid_concepts[4205] = \
{'SCT': ['IndirectOphthalmoscopyLens',
         'ContactFundusLens',
         'ConcaveNoncontactFundusLens',
         'ConcaveContactFundusLens',
         'Goniolens',
         'NoncontactFundusLens',
         'ConvexContactFundusLens',
         'ConvexNoncontactFundusLens']}

name_for_cid[3485] = 'IVUSVolumeMeasurements'
cid_concepts[3485] = \
{'DCM': ['NativePlaqueVolume',
         'TotalPlaqueVolume',
         'InStentNeointimalVolume',
         'EEMVolume',
         'LumenVolume'],
 'SCT': ['StentVolume']}

name_for_cid[12214] = 'EchocardiographyPulmonaryVeins'
cid_concepts[12214] = \
{'LN': ['PulmonaryVeinSystolicPeakVelocity',
        'PulmonaryVeinDiastolicPeakVelocity',
        'HeartRate',
        'PulmonaryVeinAtrialContractionReversalPeakVelocity',
        'PulmonaryVeinSystolicToDiastolicRatio'],
 'SCT': ['PulmonaryVeinAWaveDuration',
         'PulmonaryVeinSWaveVelocityTimeIntegral',
         'PulmonaryVeinDWaveVelocityTimeIntegral']}

name_for_cid[6054] = 'BreastImagingFindings'
cid_concepts[6054] = \
{'DCM': ['ComplexCyst',
         'IntracysticLesion',
         'SolidMass',
         'NormalBreastTissue',
         'BreastGeometry',
         'ImageQuality',
         'DuctNarrowing',
         'NonLesion',
         'IntraluminalFillingDefect',
         'ClusteredMicrocysts',
         'MultipleFillingDefect',
         'AbruptDuctTermination',
         'Extravasation',
         'ComplicatedCyst',
         'CystFill',
         'CooperLigamentChanges',
         'MassInTheSkin',
         'SelectedRegion',
         'MassOnTheSkin',
         'MassWithCalcifications'],
 'SCT': ['CalcificationOfBreast',
         'BreastComposition',
         'ForeignBody',
         'CystOfBreast',
         'UltrasoundScanNormal',
         'AxillaryLymphNode',
         'AsymmetricBreastTissue',
         'SkinRetractionOfBreast',
         'ArchitecturalDistortionOfBreast',
         'TubularDensity',
         'MammographyBreastDensity',
         'SkinLesion',
         'BreastNormal',
         'SkinThickeningOfBreast',
         'FocalAsymmetricBreastTissue',
         'LymphNode',
         'AxillaryAdenopathy',
         'Edema',
         'TrabecularThickeningOfBreast',
         'SebaceousCystOfSkinOfBreast',
         'IndividualCalcification',
         'Nipple',
         'MammaryDuctEctasia',
         'MammographicBreastMass',
         'CalcificationCluster',
         'Implant',
         'IntramammaryLymphNode']}

name_for_cid[3606] = 'ArterialSourceLocations'
cid_concepts[3606] = \
{'SCT': ['InternalCarotidArtery',
         'BasilarArtery',
         'CommonFemoralArtery',
         'PulmonaryArteriovenousFistula',
         'LateralPlantarArtery',
         'RightPulmonaryArtery',
         'PulmonaryArteryConduit',
         'AnteriorCommunicatingArtery',
         'FistulaCoronaryToLeftAtrium',
         'GeniculateArtery',
         'ThoracicAorta',
         'CarotidArtery',
         'UmbilicalArtery',
         'NeoAorta',
         'PulmonaryVeinWedge',
         'AorticFistula',
         'RadialArtery',
         'RightFemoralArtery',
         'SuperiorThyroidArtery',
         'IliacArtery',
         'LeftFemoralArtery',
         'AxillaryArtery',
         'LeftPulmonaryArtery',
         'InternalMammaryArtery',
         'TruncusArteriosusCommunis',
         'Artery',
         'VertebralArtery',
         'CommonCarotidArtery',
         'Aorta',
         'PatentDuctusArteriosus',
         'SuperficialTemporalArtery',
         'FistulaCoronaryToLeftVentricle',
         'CoronaryArtery',
         'DescendingAorta',
         'SubclavianArtery',
         'NeonatalPulmonaryArteryPrimitivePA',
         'TibialArtery',
         'SuperficialFemoralArtery',
         'OphthalmicArtery',
         'MesentericArtery',
         'MedialPlantarArtery',
         'PosteriorCommunicatingArtery',
         'PoplitealArtery',
         'AorticArch',
         'AbdominalAorta',
         'LingualArtery',
         'AnteriorSpinalArtery',
         'HepaticArtery',
         'PeronealArtery',
         'Baffle',
         'BrachiocephalicTrunk',
         'RenalArtery',
         'SystemicCollateralArteryToLung',
         'FacialArtery',
         'LacrimalArtery',
         'LumbarArtery',
         'OccipitalArtery',
         'AscendingAorta',
         'FemoralArtery',
         'PulmonaryArtery',
         'BrachialArtery',
         'ProfundaFemorisArtery',
         'CerebralArtery',
         'FistulaCoronaryToRightAtrium',
         'FistulaCoronaryToRightVentricle']}

name_for_cid[6404] = 'ChestNonlesionObjectTypePhysicalObjects'
cid_concepts[6404] = \
{'DCM': ['Portacath',
         'ChestTube',
         'CentralLine',
         'KidneyStent',
         'PancreaticStent',
         'NippleRing',
         'Coin',
         'FiducialMark'],
 'SCT': ['JejunostomyTube',
         'Suture',
         'VenaCavaFilter',
         'HeartValveProsthesis',
         'Needle',
         'Jewelry',
         'Prosthesis',
         'FeedingTube',
         'Catheter',
         'EndotrachealTube',
         'Tracheotomy',
         'Staple',
         'UretericStent',
         'Bullet',
         'Pin',
         'CardiacPacemaker',
         'CardiacPacemakerLead',
         'CervicalCollar']}

name_for_cid[7215] = 'SpectroscopyPurposeofReference'
cid_concepts[7215] = \
{'DCM': ['SpectroscopyDataForWaterPhaseCorrection']}

name_for_cid[10045] = 'RadiopharmaceuticalPatientState'
cid_concepts[10045] = \
{'DCM': ['CardiacStressState',
         'AcuteUnilateralRenalBlockage',
         'LowThyroidUptake',
         'HighThyroidUptake',
         'SeverelyJaundiced'],
 'SCT': ['AbnormalRenalFunction', 'RestingState']}

name_for_cid[66] = 'MedicationforImagingAgentAdministration'
cid_concepts[66] = \
{'SCT': ['Alfentanil', 'Remifentanil', 'Propofol', 'Midazolam', 'Sufentanil']}

name_for_cid[6111] = 'BronchovascularAnatomyFindingorFeature'
cid_concepts[6111] = \
{'DCM': ['CarinaAngle', 'CentrilobularStructures'],
 'SCT': ['HilumOfLung', 'AirwayStructure', 'Carina', 'Bronchus']}

name_for_cid[3743] = 'Clopidogrel/TiclopidineAdministration'
cid_concepts[3743] = \
{'SCT': ['Contraindicated',
         'AdministeredAfterPercutaneousCoronaryIntervention',
         'NotAdministered',
         'AdministeredLessThan72HoursBeforePCI']}

name_for_cid[4252] = 'VisualFieldStaticPerimetryScreeningTestModes'
cid_concepts[4252] = \
{'DCM': ['AgeCorrected',
         'ThresholdRelated',
         'UserChosenValue',
         'SingleLuminance',
         'FovealSensitivityRelated',
         'RelatedToNonMacularSensitivity']}

name_for_cid[7184] = 'AbstractMultidimensionalImageModelAxisDirection'
cid_concepts[7184] = \
{'DCM': ['SeptumWallAxis',
         'AnteriorInferiorAxis',
         'ApexBaseAxis',
         'AnteriorPosteriorAxis',
         'HeadFootAxis',
         'LeftRightAxis']}

name_for_cid[12306] = 'EchoFlowDirections'
cid_concepts[12306] = \
{'SCT': ['RetrogradeDirection', 'AntegradeDirection']}

name_for_cid[12243] = 'LeftVentricleFindingSites'
cid_concepts[12243] = \
{'SCT': ['LeftVentricleOutflowTract']}

name_for_cid[9525] = 'RadiationTherapyParticle'
cid_concepts[9525] = \
{'SCT': ['Electron', 'Photon']}

name_for_cid[3651] = 'HemodynamicMeasurementPhase'
cid_concepts[3651] = \
{'SCT': ['CardiacCatheterizationBaselinePhase',
         'CardiacCatheterizationTestChallengePhase',
         'CardiacCatheterizationGradientAssessmentPhase',
         'DrugInfusionChallenge',
         'CardiacCatheterizationPostInterventionPhase',
         'CardiacCatheterizationPostContrastPhase',
         'RestingState',
         'ExerciseChallenge']}

name_for_cid[6003] = 'ChangeSinceLastMammogramorPriorSurgeryfromBIRADS®'
cid_concepts[6003] = \
{'SCT': ['RemovalOfImplantSincePreviousMammogram',
         'ImplantRevisedSincePreviousMammogram',
         'NoSignificantChangesInTheFinding',
         'IncreaseInSize',
         'NewFinding',
         'IncreaseInNumberOfCalcifications',
         'DecreaseInNumberOfCalcifications',
         'LessDefined',
         'FindingPartiallyRemoved',
         'MoreDefined',
         'DecreaseInSize']}

name_for_cid[3201] = 'IndicationsforStressTest'
cid_concepts[3201] = \
{'SCT': ['AbnormalExerciseToleranceTest',
         'Arrhythmia',
         'Palpitations',
         'PreOperative',
         'AnginaPectoris',
         'Syncope',
         'HeartFailure',
         'AbnormalECG',
         'CoronaryArteryDisease',
         'SupraventricularTachycardia',
         'HistoryOfCABG',
         'LeftBundleBranchBlock',
         'ValvularHeartDisease',
         'PostPTCA',
         'ChestPain',
         'Hypertension',
         'OccupationalRequirement',
         'HistoryOfMyocardialInfarction',
         'HeartDiseaseRiskFactors',
         'Dyspnea']}

name_for_cid[3714] = 'Thrombus'
cid_concepts[3714] = \
{'SCT': ['PossibleThrombus',
         'SevereThrombus',
         'NoThrombus',
         'DefiniteThrombus']}

name_for_cid[6146] = 'TimePointTypes'
cid_concepts[6146] = \
{'DCM': ['Posttreatment', 'Eligibility'],
 'UMLS': ['Unscheduled', 'Baseline', 'Pretreatment', 'Nadir']}

name_for_cid[7307] = 'ImplantComponentTypes'
cid_concepts[7307] = \
{'DCM': ['FemoralStem',
         'FemoralStemDistalComponent',
         'FemoralHeadConeTaperComponent',
         'FemoralStemProximalComponent',
         'FemoralStemComponent',
         'NeckComponent',
         'DistalCentralizer',
         'MonoblockStem',
         'AcetabularCupShell',
         'ProstheticShaftAugment',
         'AcetabularCupInsert',
         'FemoralHeadResurfacingComponent',
         'AcetabularCupMonoblock',
         'FemoralHeadBallComponent'],
 'SCT': ['DHSPlate', 'FemoralHeadProsthesis', 'Screw', 'BonePlate', 'BoneNail']}

name_for_cid[31] = 'AbstractPriors'
cid_concepts[31] = \
{'DCM': ['AtLastAppointment',
         'OnDischargeFromUnit',
         'OnDischarge',
         'PostIntervention',
         'PreIntervention',
         'OnAdmissionToUnit'],
 'SCT': ['Intraoperative',
         'PreOperative',
         'OnAdmission',
         'PreAdmission',
         'PreDose',
         'PostOperative',
         'PostDose']}

name_for_cid[12238] = 'WallMotionScoringSchemes'
cid_concepts[12238] = \
{'DCM': ['_5PointSegmentFindingScaleWithGradedHypokinesis',
         '_5PointSegmentFindingScale',
         '_4PointSegmentFindingScale']}

name_for_cid[3850] = 'ContrastBolusSubstance'
cid_concepts[3850] = \
{'DCM': ['NonIonicIodinatedContrastAgent'],
 'SCT': ['MeglumineIodipamide',
         'Iophendylate',
         'Iodipamide',
         'Ioxilan',
         'Bunamiodyl',
         'Ioversol',
         'Gadodiamide',
         'Metrizamide',
         'Iodixanol',
         'Ipodate',
         'Oxygen',
         'BariumSulfate',
         'SodiumTyropanoate',
         'Iohexol',
         'Water',
         'GadopentetateDimeglumine',
         'SodiumAcetrizoate',
         'Propyliodone',
         'ContrastAgent',
         'IodoalphionicAcid',
         'Iodopyracet',
         'SodiumDiprotrizoate',
         'Iothalamate',
         'Iodophthalein',
         'Ioxaglate',
         'MeglumineDiatrizoate',
         'MangafodipirTrisodium',
         'IodamideMeglumine',
         'SodiumIodipamide',
         'SodiumDiatrizoate',
         'Diatrizoate',
         'Saline',
         'NonRadiopaqueMedium',
         'ChloriodizedOil',
         'IophenoxicAcid',
         'Air',
         'Iopamidol',
         'Iopromide',
         'IopanoicAcid',
         'Dextran',
         'Gadolinium',
         'Metrizoate',
         'IonicIodinatedContrastAgent',
         'IodizedOil',
         'CarbonDioxide',
         'SodiumIodomethamate',
         'LactatedRingerS']}

name_for_cid[7110] = 'FiducialsCategories'
cid_concepts[7110] = \
{'DCM': ['FiducialMark'],
 'SCT': ['BodyReferencePointMarker', 'BodySurfacePoint', 'AnatomicalPoint']}

name_for_cid[3503] = 'IndexedHemodynamicResistanceUnits'
cid_concepts[3503] = \
{'UCUM': ['WoodUM2', 'DynScm5M2', 'PRUM2']}

name_for_cid[3016] = 'MajorCoronaryArteries'
cid_concepts[3016] = \
{'SCT': ['CircumflexCoronaryArtery',
         'LeftAnteriorDescendingCoronaryArtery',
         'RightCoronaryArtery',
         'LeftMainCoronaryArtery']}

name_for_cid[3446] = 'SkinCondition'
cid_concepts[3446] = \
{'DCM': ['SkinConditionMottled',
         'SkinConditionPale',
         'SkinConditionDiaphoretic',
         'SkinConditionFlush',
         'SkinConditionDry',
         'SkinConditionClammy',
         'SkinConditionCool',
         'SkinConditionCold',
         'SkinConditionWarm']}

name_for_cid[619] = 'AirwayManagementSubMethodCodeTypeforSmallAnimalAnesthesia'
cid_concepts[619] = \
{'SCT': ['HighFrequencyVentilation',
         'TranstrachealJetVentilation',
         'ContinuousFlowVentilation'],
 'UMLS': ['InverseRatioVentilation']}

name_for_cid[6097] = 'SideofFamily'
cid_concepts[6097] = \
{'DCM': ['Maternal'], 'SCT': ['Paternal']}

name_for_cid[4100] = 'T1MeasurementMethods'
cid_concepts[4100] = \
{'DCM': ['T1ByInversionRecovery', 'T1ByFixedValue', 'T1ByMultipleFlipAngles']}

name_for_cid[8123] = 'MicroscopyIlluminationMethod'
cid_concepts[8123] = \
{'DCM': ['DifferentialInterferenceContrast',
         'TotalInternalReflectionFluorescence',
         'TransmissionIllumination',
         'ReflectionIllumination',
         'EpifluorescenceIllumination',
         'BrightfieldIllumination',
         'DarkfieldIllumination',
         'ObliqueIllumination',
         'PhaseContrastIllumination']}

name_for_cid[10069] = 'RadiationDoseEstimationParameter'
cid_concepts[10069] = \
{'DCM': ['Distance',
         'AttenuationCoefficient',
         'WaterEquivalentDiameterRepresentativeValue',
         'WaterEquivalentDiameterFromRawData',
         'WaterEquivalentDiameterIntegratedAcrossScanRange',
         'PercentFibroglandularTissue',
         'DistanceCorrection',
         'BreastThickness',
         'TissueFraction',
         'PatientAPDimension',
         'Dgn',
         'PatientLateralDimension',
         'CurveFitParameter',
         'CorrectionFactor',
         'Backscatter',
         'TissueAirRatio',
         'HomogeneityFactor',
         'SSDEConversionFactor',
         'OffsetFactor',
         'NormalizationFactor',
         'WaterEquivalentDiameterFromLocalizer',
         'HalfValueLayer'],
 'NCIt': ['UnitConversionFactor']}

name_for_cid[12111] = 'AbdominalArteries(Lateral)'
cid_concepts[12111] = \
{'SCT': ['GastricArtery',
         'OvarianArtery',
         'AccessoryRenalArtery',
         'UmbilicalArtery',
         'TesticularArtery',
         'UterineArtery']}

name_for_cid[3554] = 'Ventilation'
cid_concepts[3554] = \
{'SCT': ['PressureSupportVentilator', 'AmbuBag', 'VolumeSupportVentilator']}

name_for_cid[9509] = 'PurposeofReferenceforRTPhysicianIntentInput'
cid_concepts[9509] = \
{'DCM': ['HistoricalRTPrescription',
         'RTPrescriptionInputImages',
         'DiagnosticSourceImages',
         'RegistrationResult',
         'SegmentationResult'],
 'LN': ['RelevantDiagnosticTestsAndOrLaboratoryData']}

name_for_cid[3104] = 'CardiacSynchronizationTechnique'
cid_concepts[3104] = \
{'DCM': ['ProspectiveGating',
         'RealTimeAcquisition',
         'Paced',
         'RetrospectiveGating']}

name_for_cid[3807] = 'VolumeMeasurementMethods'
cid_concepts[3807] = \
{'DCM': ['VolumeBasedMethod', 'AreaBasedMethod', 'DiameterBasedMethod']}

name_for_cid[7002] = 'DiagnosticImagingReportElements'
cid_concepts[7002] = \
{'DCM': ['PreviousFinding',
         'Conclusion',
         'PatientPresentation',
         'Summary',
         'Recommendation',
         'ProcedureDescription',
         'Impression',
         'Finding'],
 'LN': ['History', 'Request'],
 'SCT': ['ComplicationOfProcedure']}

name_for_cid[12280] = 'CardiacUltrasoundTargetSites'
cid_concepts[12280] = \
{'SCT': ['RightVentricleOutflowTract',
         'LeftSuperiorPulmonaryVein',
         'TransverseAorticArch',
         'LeftInferiorPulmonaryVein',
         'Effusion',
         'RightSuperiorPulmonaryVein',
         'RightPulmonaryArtery',
         'PosteriorWall',
         'PulmonaryVeinCommonRightSegment',
         'PreductalRegionOfAorticArch',
         'VenaContracta',
         'PulmonaryVeinCommonLeftSegment',
         'RightInferiorPulmonaryVein',
         'PulmonaryVeinRightMiddleSegment',
         'CardiacValveLeaflet',
         'LeftVentricleOutflowTract',
         'AnteriorWall',
         'CardiacValveAnnulus',
         'PulmonaryVeinConfluenceToAtriumConnection',
         'PostductalRegionOfAorticArch',
         'LeftPulmonaryArtery']}

name_for_cid[7452] = 'OrganizationalRoles'
cid_concepts[7452] = \
{'DCM': ['AdministratorOfRadiologyDepartment',
         'HeadOfCardiology',
         'LeadRadiologicTechnologist',
         'RepresentativeOfEthicsCommittee',
         'RepresentativeOfProtocolCommittee',
         'Fellow',
         'ChairOfProtocolCommittee',
         'HeadOfRadiology',
         'Sonologist'],
 'SCT': ['Resident',
         'Physician',
         'Attending',
         'ScrubNurse',
         'Registrar',
         'RadiologicTechnologist',
         'Surgeon',
         'Consultant',
         'Nurse',
         'MedicalPractitioner',
         'RadiationTherapist'],
 'UMLS': ['Intern',
          'MedicalPhysicist',
          'ConsultingPhysician',
          'RadiationPhysicist',
          'Sonographer']}

name_for_cid[3411] = 'IntracoronaryDevices'
cid_concepts[3411] = \
{'SCT': ['SalineThrombectomy',
         'RotationalAtherectomyDevice',
         'GuidingCatheter',
         'PercutaneousTransluminalAngioplastyBalloon',
         'GuideWire',
         'IntracoronaryPressureGuideWire',
         'IntravascularOpticalCoherenceTomographyDevice',
         'CuttingBalloonAngioplastyCBADevice',
         'Laser',
         'IntracoronaryDopplerGuideWire',
         'IntravascularUltrasoundIVUSDevice',
         'RadiofrequencyAblationDevice',
         'Stent',
         'BrachytherapyDevice',
         'TransluminalExtractionCatheterTEC',
         'DirectionalCoronaryAtherectomyDCADevice']}

name_for_cid[12003] = 'OBGYNDates'
cid_concepts[12003] = \
{'LN': ['ConceptionDate',
        'EstimatedLMPByEDD',
        'EDDFromAverageUltrasoundAge',
        'EDDFromOvulationDate',
        'EDD',
        'LMP',
        'EDDFromLMP',
        'OvulationDate']}

name_for_cid[3738] = 'PreCathAnginaType'
cid_concepts[3738] = \
{}

name_for_cid[4012] = 'ProjectionEponymousName'
cid_concepts[4012] = \
{'SCT': ['KempHarper',
         'Holmblad',
         'Swanson',
         'Lawrence',
         'Fuchs',
         'LaquerrierePierquin',
         'MorticeProjection',
         'Teufel',
         'May',
         'Ottonello',
         'AlbersSchonberg',
         'LowBeer',
         'Cleaves',
         'DaneliusMiller',
         'Kasabach',
         'Stryker',
         'BrewertonProjection',
         'Arcelin',
         'Causton',
         'GaynorHart',
         'Henschen',
         'Pearson',
         'ColcherSussman',
         'Stecher',
         'ChassardLapine',
         'Valdini',
         'Hough',
         'Twining',
         'Zanelli',
         'Cahoon',
         'Chamberlain',
         'Tarrant',
         'Rafert',
         'Towne',
         'Rosenberg',
         'Kovacs',
         'Waters',
         'Kuchendorf',
         'Friedman',
         'Chausse',
         'Garth',
         'BlackettHealy',
         'Fleischner',
         'Wolf',
         'Kurzbauer',
         'Lysholm',
         'Kandel',
         'HarrisBeathAxialProjection',
         'Kite',
         'Miller',
         'Hsieh',
         'Stenvers',
         'Dunlap',
         'Clements',
         'Caldwell',
         'Apple',
         'Burman',
         'Fisk',
         'Grashey',
         'Haas',
         'Merchant',
         'RafertLong',
         'JudetProjection',
         'Ferguson',
         'Lewis',
         'Broden',
         'Schuller',
         'Titterington',
         'Pawlow',
         'Nolke',
         'Norgaard',
         'Taylor',
         'Grandy',
         'Hickey',
         'Moore',
         'Neer',
         'Law',
         'LeonardGeorge',
         'Pirie',
         'Folio',
         'Holly',
         'Beclere',
         'Rhese',
         'WigbyTaylor',
         'Hughston',
         'ErasoModification',
         'Bertel',
         'Lorenz',
         'ClementsNakayama',
         'Judd',
         'Lauenstein',
         'Settegast',
         'Mayer',
         'Staunig',
         'CampCoventry',
         'Isherwood',
         'Penner',
         'VanRosenProjection',
         'WestPoint',
         'Alexander',
         'HirtzModification',
         'Lilienfeld',
         'Robert',
         'Lindblom']}

name_for_cid[12279] = 'CardiacUltrasoundFetalGeneralMeasurements'
cid_concepts[12279] = \
{'LN': ['OccipitalFrontalDiameter',
        'FLAC',
        'AbdominalCircumference',
        'HeadCircumference',
        'BiparietalDiameter',
        'CephalicIndex',
        'NuchalTranslucency',
        'TransverseThoracicDiameter',
        'ThoracicArea',
        'TransCerebellarDiameter',
        'HCAC',
        'BPDAreaCorrected',
        'TranverseAbdominalDiameter',
        'CisternaMagna',
        'FemurLength',
        'ThoracicCircumference',
        'LeftVentricleRightVentricleDiameterRatio',
        'ForamenOvaleDiameterAorticRootDiameter',
        'CardiacCircumferenceTransverseByUS',
        'CardiothoracicAreaRatio',
        'FLBPD',
        'CardiacCrossSectionalAreaTransverseByUS',
        'AnteriorPosteriorAbdominalDiameter',
        'FLHC',
        'AnteriorPosteriorTrunkDiameter',
        'CardiothoracicCircumferenceRatio'],
 'SCT': ['NumberOfUmbilicalArteries']}

name_for_cid[7501] = 'NeighbouringGreyLevelDependenceBasedFeatures'
cid_concepts[7501] = \
{}

name_for_cid[7151] = 'SegmentationPropertyTypes'
cid_concepts[7151] = \
{'BARI': ['_1stMarginalCoronaryArteryLaterals',
          '_2ndMarginalCoronaryArteryLaterals',
          '_3rdMarginalCoronaryArteryLaterals',
          'RamusLaterals',
          '_3rdDiagonalCoronaryArteryLaterals',
          '_1stDiagonalCoronaryArteryLaterals',
          '_2ndDiagonalCoronaryArteryLaterals',
          'PosteriorDescendingSeptalPerforators'],
 'DCM': ['Phantom',
         'AorticKnob',
         'ArchOfTheAzygosVein',
         'IonChamberArray',
         'DiodeArray',
         'LiquidIonChamber',
         'Diode',
         'IonChamber',
         'OSLD',
         'DiamondDetector',
         'Portacath',
         'ChestTube',
         'CentralLine',
         'KidneyStent',
         'PancreaticStent',
         'NippleRing',
         'Coin',
         'FiducialMark',
         'RegistrationInput',
         'Background',
         'SurfaceOfTheChoroidScleraInterface',
         'OuterSurfaceOfTheBM',
         'OuterSurfaceOfTheCC',
         'Mold',
         'HeadAndNeckMask',
         'HeadMask',
         'Headframe',
         'BodyFrame',
         'BreastBoard',
         'Cast',
         'WholeBodyPod',
         'VacuumMold',
         'RectalBalloon',
         'VaginalCylinder',
         'PosteriorSurfaceOfTheRPE',
         'SurfaceOfTheCenterOfTheRPE',
         'OuterSurfaceOfIPL',
         'OuterSurfaceOfGCL',
         'OuterSurfaceOfOPL',
         'OuterSurfaceOfINL',
         'SurfaceBetweenInnerAndOuterSegmentsOfThePhotoreceptors',
         'OuterSurfaceOfHFL',
         'AnteriorSurfaceOfTheRPE',
         'SurfaceOfTheInterdigitatingZoneBetweenRetinaAndRPE',
         'OuterSurfaceOfRNFL',
         'Sphere',
         'OtherMarker',
         'SecondaryPulmonaryLobule',
         'SingleSubjectExtractedFromGroup',
         'Hiatus',
         'AzygoesophagealRecessInterface',
         'ParaspinalLine',
         'PosteriorTrachealStripe',
         'RightTrachealStripe',
         'Stripe',
         'Lucency',
         'MidlungWindow',
         'CarinaAngle',
         'CentrilobularStructures',
         'AnteriorJunctionLine',
         'PosteriorJunctionLine',
         'Interface',
         'Line',
         'BrachytherapyChannelShield',
         'BrachytherapySourceApplicator',
         'BrachytherapyChannel'],
 'FMA': ['ArcuateFasciculus'],
 'NCIt': ['NonEnhancingLesion', 'ReferenceRegion', 'EnhancingLesion'],
 'SCT': ['ProfundaFemorisVein',
         'LeftHypochondriacRegion',
         'SternocleidomastoidMuscle',
         'Ligament',
         'TruncusCoeliacus',
         'InferiorLongitudinalFasciculus',
         'WristJoint',
         'IntermediateArteryRamus',
         'GenicularArtery',
         'LeftVentricleInflow',
         'FontanelOfSkull',
         'SigmoidColon',
         'RightFemoralArtery',
         'Intracranial',
         'JawRegion',
         'LobarArtery',
         'DentalProsthesis',
         'ThighPerforator',
         'Pituitary',
         'MidRightCoronaryArtery',
         'UpperLimb',
         'ExternalIliacVein',
         'CongenitalPulmonaryArteriovenousFistula',
         'DescendingAorta',
         'ParietalBone',
         'Cartilage',
         'RightUpperQuadrantOfAbdomen',
         'Eye',
         'SuperficialPalmarArch',
         'Prostate',
         'FacialBones',
         'Back',
         'Duodenum',
         '_1stMarginalCoronaryArtery',
         'Larynx',
         'BonePin',
         'PeronealVein',
         'FemoralArtery',
         'Ovary',
         'Urethra',
         'Head',
         'AdrenalGland',
         'MitralValve',
         'Calyx',
         'PosteriorMedialTributary',
         'PulmonaryVein',
         'ForeignBody',
         'Submental',
         'NucleusAccumbens',
         'PoplitealFossa',
         'PulmonaryArteryConduit',
         'InternalCarotidArteryC5Segment',
         'Nodule',
         'LevatoresCostarumMuscles',
         'FrontalLobe',
         'Adventitia',
         'Hypopharynx',
         'RightCoronaryArteryOstium',
         'HepaticVein',
         'MarginalCoronaryArtery',
         'InternalMammaryArtery',
         'BronchialArtery',
         'ExternalCarotidArtery',
         'BiteBlock',
         'Neoplasm',
         'SystemicArtery',
         'VomerBone',
         'UterusAndFallopianTubes',
         'CardiacPacemaker',
         'Arachnoid',
         'VenaCavaFilter',
         'TarsalJoint',
         'LingualArtery',
         'NeoplasmSecondary',
         'AirwayStructure',
         'CongenitalCoronaryArteryFistulaToRightAtrium',
         'DeltoidMuscle',
         'CerebralFornix',
         'Penis',
         'Baffle',
         'VentricularVein',
         'UretericStent',
         'Parathyroid',
         'CervicoThoracicSpine',
         'Scapula',
         'Kidney',
         'OmentalBursa',
         'InternalCarotidArtery',
         'RightPosteriorAVCoronaryArtery',
         'FascialLayer',
         'RenalPelvis',
         'Thrombus',
         'VenousNetwork',
         'SubcostalMuscle',
         'CommonAtrium',
         'Ilium',
         'InternalCarotidArteryC6Segment',
         'AnteriorTibialArtery',
         'CranialNerve',
         'SkinOfAbdomen',
         'ConnectiveTissue',
         'CommonCarotidArtery',
         'InnermostIntercostalMuscles',
         '_3rdMarginalCoronaryArtery',
         'MastoidBone',
         'CommonIliacArteryBifurcation',
         'MiddleHepaticVein',
         'Thigh',
         'Trachea',
         'LeftPosterolateralCircumflexCoronaryArtery',
         'BodyConduit',
         'Thalamus',
         'RightVentricleInflow',
         'LeftLowerQuadrantOfAbdomen',
         'SpinalCerebrospinalFluidPathway',
         'SesamoidBonesOfFoot',
         'SixthLumbarArtery',
         'BodyFluid',
         'Omentum',
         'SubmandibularGland',
         'CarotidBifurcation',
         'LeftLumbarRegion',
         'InnominateArtery',
         'FemaleInternalGenitalia',
         'Extremity',
         'FacialArtery',
         'BoneStructureOfHeadAndOrNeck',
         'EpigastricRegion',
         'NeoplasmPrimary',
         'DigitalImager',
         'AbdominalCavity',
         'TeresMinorMuscle',
         'CommonFemoralVein',
         'SubstantiaNigra',
         'UpperInnerQuadrantOfBreast',
         'ThoracicAorta',
         'RightInguinalRegion',
         'RightLowerQuadrantOfAbdomen',
         'CockettPerforatingVein',
         'LumboSacralSpine',
         'RightMainBronchus',
         'AVGrooveContinuationOfCircumflexArtery',
         'Thumb',
         'GreatCardiacVein',
         'MiddleLobeOfRightLung',
         '_2ndLeftPosterolateralCoronaryArtery',
         'MiddleCerebralArteryM2Segment',
         'LacrimalBone',
         'BasilicVein',
         'HunterianPerforatingVein',
         'ProfundaFemorisArtery',
         'Pharynx',
         'LigamentumArteriosum',
         'Clothing',
         'Nasopharynx',
         'SuprapulmonicValveArea',
         'SympatheticTrunk',
         'LobeOfLung',
         'OvarianVein',
         'BrachialPlexus',
         'ApexOfRightVentricle',
         'LateralCalfPerforator',
         'FourthLumbarArtery',
         'ThirdLumbarArtery',
         'LimbicLobe',
         'SkinOfChest',
         'PrimitiveAorta',
         'LumenOfBloodVessel',
         'Hemorrhage',
         'AntecubitalVein',
         'SpinalCordGrayMatter',
         'Mandible',
         'Parasternal',
         'LeftMainBronchus',
         'SystemicVein',
         'OccipitalLobe',
         'SubclavianArtery',
         'Amygdala',
         '_3rdLeftPosterolateralCoronaryArtery',
         'OpticChiasm',
         'OphthalmicArtery',
         'Nose',
         'PancreaticDuct',
         'Hand',
         'PectoralisMajorMuscle',
         'FemaleExternalGenitalia',
         'OvarianArtery',
         'Hippocampus',
         'SuperficialFemoralVein',
         'SuperiorLeftPulmonaryVein',
         'LeftPosteriorDescendingCircumflexCoronaryArtery',
         'BileDuct',
         'SystemicCollateralArteryToLung',
         'Skull',
         'AtrioventricularVein',
         'IDPlate',
         'OpticRadiation',
         'EndoNasopharyngeal',
         'EsophagusStomachAndDuodenum',
         'Breast',
         'Meniscus',
         'AxillaryFascia',
         'PoplitealVein',
         'EmbolizationCoil',
         'PiaMater',
         'TricuspidValve',
         'RightAtrium',
         'VascularSclerosis',
         'DistalLeftAnteriorDescendingCoronaryArtery',
         'InnerEar',
         'Coccyx',
         'AtrialSeptalDefect',
         'DescendingAorta',
         'ProximalLeftAnteriorDescendingCoronaryArtery',
         'OccipitalBone',
         'Bullet',
         'Spine',
         'Muscle',
         'WhiteMatter',
         'Zygoma',
         'InnominateVein',
         'RadioactiveImplant',
         'Needle',
         'JuxtaposedAtrialAppendage',
         'Jejunum',
         'CarotidBody',
         'ChestAbdomenAndPelvis',
         'RightHepaticVein',
         'GastrocnemiusVein',
         'Scalp',
         'OrbitalStructure',
         'LeftMainBranchOfPortalVein',
         'BloodVessel',
         'Sclera',
         'AnteriorPosteriorCerebralArteryBifurcation',
         '_1stDiagonalCoronaryArtery',
         'Perineum',
         'InteratrialSeptum',
         'Rectum',
         'Femur',
         'LeftAtrium',
         'CerebralArtery',
         'InferiorMesentericArtery',
         'IliocostalisMuscle',
         'LeftVentricularAssistDevice',
         'CervicalSpine',
         'IliacAndOrFemoralArtery',
         'SpinalCordWhiteMatter',
         'SmallIntestine',
         'SuperiorRightPulmonaryVein',
         'InferiorVenaCava',
         'IntraArticular',
         'Jewelry',
         'Axilla',
         'Thymus',
         'ELMExternalLimitingMembrane',
         'MeasuringRuler',
         'ParietalLobe',
         'IliacArtery',
         'DistalRightCoronaryArtery',
         'SubmandibularArea',
         'PectoralisMinorMuscle',
         'OpticCanal',
         'SuperficialTemporalArtery',
         'MedialLemniscus',
         'InferiorCerebellarPeduncle',
         'RightMainBranchOfPortalVein',
         'SphenoidBone',
         'Spleen',
         'ChestAndAbdomen',
         'LowerOuterQuadrantOfBreast',
         'Mediastinum',
         'Cerebellum',
         'FirstLumbarArtery',
         'PoplitealArtery',
         'UpperUrinaryTract',
         'PulmonaryVenousAtrium',
         'EmbolizationParticulate',
         'DorsalScapularArtery',
         'LeftVentricle',
         'AbdomenAndPelvis',
         'CostalCartilage',
         'SaphenousVein',
         'Gallbladder',
         '_2ndRightPosterolateralCoronaryArtery',
         'RenalArtery',
         'PeritonealCavity',
         'UpperArm',
         'GastroduodenalArtery',
         'Bile',
         'AzygosVein',
         'Liver',
         'Rib',
         'Stent',
         'Implant',
         'CommonFemoralArtery',
         'SmallCardiacVein',
         'SuperficialPalmarVenousArch',
         'Vertebra',
         'CongenitalCoronaryArteryFistulaToLeftAtrium',
         'HilarArtery',
         'Embolus',
         'ApexOfLung',
         'ThirdVentricle',
         '_3rdDiagonalCoronaryArtery',
         'SpinalisMuscle',
         'SaphenofemoralJunction',
         'Anus',
         'FrontalBone',
         'Clip',
         'CommonIliacVein',
         'TrabeculaeCarnae',
         'BoneScrew',
         'PosteriorCerebralArteryP1Segment',
         'LumbarSpine',
         'ThyrocervicalTrunk',
         'PrimitivePulmonaryArtery',
         'HilumOfLung',
         'SellaTurcica',
         'Scrotum',
         'InferiorLeftPulmonaryVein',
         'LowerLeg',
         'RenalVein',
         'InternalIliacVein',
         'IlealVein',
         'Toe',
         'Skin',
         'Laser',
         'MiddleCardiacVein',
         'MiddleCerebralArtery',
         'PosteriorDescendingRightCoronaryArtery',
         'CompressionPaddle',
         'CorpusStriatum',
         'Adenohypophysis',
         'PosteriorTibialVein',
         'MuscleOfLowerLimb',
         'AscendingAorta',
         'LeftUpperQuadrantOfAbdomen',
         'PericardiophrenicArtery',
         'CardiacPacemaker',
         'ExternalAuditoryCanal',
         'Patella',
         'OccipitalVein',
         'BasilarArtery',
         'ParanasalSinus',
         'BrainCerebrospinalFluidPathway',
         'MidCircumflexCoronaryArtery',
         'CentralRetinalVein',
         'Prosthesis',
         'GastricArtery',
         'Blood',
         'SubscapularisMuscle',
         'Perfusion',
         'ExternalEar',
         'LeftVentricleOutflowTract',
         'Calcaneus',
         'SolealVein',
         'IntercostalArtery',
         'PeripheralNerve',
         'TeresMajorMuscle',
         'ScarTissue',
         'PatentDuctusArteriosus',
         'MesentericVein',
         'MaleInternalGenitalia',
         'PalatineBone',
         'CoronaryArtery',
         'RightBranchOfHepaticArtery',
         'DuraMater',
         'Flank',
         'Maxilla',
         'EsophagealHiatus',
         'CentralRetinalArtery',
         'Normal',
         'Neurohypophysis',
         'SegmentalArtery',
         'BrachialVein',
         'HepaticArtery',
         'Tracheotomy',
         'PeronealArtery',
         'Staple',
         'RightHypochondriacRegion',
         'SuperiorMesentericVein',
         'LumbarArtery',
         'Cornea',
         'PercutaneousTransluminalAngioplastyBalloon',
         'ThoracoLumbarSpine',
         'CoronaryArteryGraft',
         'AcromioclavicularJoint',
         'UmbilicalRegion',
         'SupraspinatusMuscle',
         'CongenitalCoronaryArteryFistulaToRightVentricle',
         'SoftPalate',
         'ExternalIntercostalMuscle',
         'CranialVenousSystem',
         'GiacominiVein',
         'Glue',
         'MiddleCerebellarPeduncle',
         'BloodClot',
         'MOSFETRadiationTherapyDosimetrySystemDosimeter',
         'FissureOfLung',
         'RightVentricle',
         'VagusNerve',
         'Abdomen',
         'HeadAndNeck',
         'PelvisAndLowerExtremities',
         'Bladder',
         'Vulva',
         'LargeIntestine',
         'EntorhinalCortex',
         'Forearm',
         'NeckChestAndAbdomen',
         'SuperiorMesentericArtery',
         'TemporalLobe',
         'PosteriorTibialArtery',
         'FemoralVein',
         'RightAuricularAppendage',
         'GreatSaphenousVeinOfThigh',
         'LeftHepaticVein',
         'DetachableBalloon',
         'AbdominalWallMuscle',
         'SuperficialFemoralArtery',
         'JWire',
         '_2ndMarginalCoronaryArtery',
         'SuperiorCerebellarPeduncle',
         'AorticArch',
         'Carina',
         'Telencephalon',
         'LeftBranchOfHepaticArtery',
         'ArcuateArteryOfTheKidney',
         'Heart',
         'CephalicVein',
         'Lip',
         'GastricVein',
         'LeftMainCoronaryArtery',
         'InferiorRightHepaticVein',
         'CarotidSiphon',
         'SystemicVenousAtrium',
         'PlantarArterialArch',
         'AnomalousPulmonaryVein',
         'LeftMainCoronaryArteryOstium',
         'LesserSaphenousVein',
         'AutonomicNerve',
         'GreatSaphenousVein',
         'Necrosis',
         'Ureter',
         'UlnarVein',
         'AorticValve',
         'MiddleEar',
         'InternalAuditoryCanal',
         'RightPulmonaryArtery',
         'ProperHepaticArtery',
         'InterlobarArteryOfKidney',
         'CorpusCallosum',
         'PerforatingArteryOfKidney',
         'Humerus',
         'AorticFistula',
         'ILMInternalLimitingMembrane',
         'NasalBone',
         'Collimator',
         'Mass',
         'Cervix',
         'LeftAnteriorDescendingCoronaryArtery',
         'Artery',
         'Buttock',
         'FallopianTube',
         'Vein',
         'CerebrospinalFluid',
         'EsophagealArtery',
         'RadialVein',
         'InferiorRightPulmonaryVein',
         'PulmonaryValveSinuses',
         'UterineArtery',
         'Gas',
         'EndoVesical',
         'MesentericArtery',
         'MuscleOfHead',
         'MiddleCerebralArteryM1Segment',
         'InternalIliacArtery',
         'SuperiorVenaCava',
         'FeedingTube',
         'Urine',
         'PosteriorArchVein',
         'AnteriorSpinalArtery',
         'AnusRectumAndSigmoidColon',
         'SplenicVein',
         'Colon',
         'CentralNervousSystem',
         'SplenicArtery',
         'AccessoryRenalArtery',
         'Media',
         'BBShotLeadPellet',
         'SeminalVesicle',
         'Headrest',
         'DistalCircumflexCoronaryArtery',
         'Feces',
         'BoneOfUpperLimb',
         'DorsalisPedisArtery',
         'Vagina',
         'LongissimusMuscle',
         'LowerInnerQuadrantOfBreast',
         'Clavicle',
         'ParotidGland',
         'AorticIsthmus',
         'Capillary',
         '_1stLeftPosterolateralCoronaryArtery',
         'Subxiphoid',
         'Diaphragm',
         'VasDeferens',
         'AxillaryArtery',
         'LeftPulmonaryArtery',
         'Ileum',
         'SalivaryGland',
         'VertebralArtery',
         'TransjugularIntrahepaticPortosystemicShunt',
         'SupraRenalAorta',
         'InternalJugularVein',
         'LacrimalArteryOfRightEye',
         'ChordaeTendineaeCordis',
         'OpaqueMarker',
         'SacroiliacJoint',
         'ScalenousAnteriorMuscle',
         'CardiacPacemakerLead',
         'EntireBody',
         'TesticularArtery',
         'HypogastricRegion',
         'RightLumbarRegion',
         'HyoidBone',
         'ThoracicDuct',
         'AtherectomyDevice',
         'BoydPerforatingVein',
         'Organ',
         'Peritioneum',
         'CommonIliacArtery',
         'SupraclavicularRegionOfNeck',
         'Meninges',
         'Shoulder',
         'TransversusThoracis',
         'VascularGraft',
         'CerebellarWhiteMatter',
         'InterventricularSeptum',
         'Cingulum',
         'LowerLobeOfLung',
         'TesticularVein',
         'AnteriorTibialVein',
         'LeftInguinalRegion',
         'RightCoronaryArtery',
         'Edema',
         'IntraPelvic',
         'Putamen',
         'CostocervicalTrunk',
         'FifthLumbarArtery',
         'AxillaryVein',
         'SubclavianVein',
         'PulmonaryValve',
         'ExternalIliacArtery',
         'ProximalRightCoronaryArtery',
         'TemporalBone',
         'BoneOfLowerLimb',
         'Midbrain',
         'PosteriorCerebralArtery',
         'ErectorSpinaeMuscle',
         'DeepPalmarVenousArch',
         'Mouth',
         'PharynxAndLarynx',
         'CarotidArtery',
         'SmallestCardiacVein',
         'UmbilicalArtery',
         'BuccalRegionOfFace',
         'EndotrachealTube',
         'LymphNode',
         'RadialArtery',
         'LeftFemoralArtery',
         'TrapeziusMuscle',
         'Esophagus',
         'RotationalAtherectomyDevice',
         'LateralCorticospinalTract',
         'BroadLigament',
         'CerebralAqueduct',
         'Insula',
         'PulmonaryVein',
         'MuscleOfNeck',
         'Lung',
         'HeartValveProsthesis',
         'CardiovascularSystem',
         'Chest',
         'ElbowJoint',
         'Sternum',
         'NeckAndChest',
         'RightVentricleOutflowTract',
         'GlobusPallidus',
         'UpperOuterQuadrantOfBreast',
         'PancreaticDuctAndBileDuctSystems',
         'ApexOfLeftVentricle',
         'PericardialCavity',
         'PulmonaryArtery',
         'Defibrillator',
         'MidLeftAnteriorDescendingCoronaryArtery',
         'Thyroid',
         'EthmoidBone',
         'InfraRenalAorta',
         'SternoclavicularJoint',
         'Neck',
         'CarotidBulb',
         'Finger',
         'InlayDentalRestoration',
         'InfraspinatusMuscle',
         'SuperiorPhrenicArtery',
         'CommonVentricle',
         'RightPulmonaryVein',
         'Catheter',
         'BodyFat',
         'Uterus',
         'TemporomandibularJoint',
         'Ventricle',
         'Nerve',
         'TracheaAndBronchus',
         'Testis',
         'Eyeball',
         'Face',
         'Tissue',
         'UrinarySystem',
         'BrainStem',
         'Foot',
         'MuscleOfUpperLimb',
         'PeripheralNervousSystem',
         'PinealGland',
         'UpperLobeOfLung',
         'Inflammation',
         'ProximalCircumflexCoronaryArtery',
         'LowerLimb',
         'PosteriorCommunicationArtery',
         'VertebralColumnAndCranium',
         'AbdominalAorta',
         'Bone',
         'Tongue',
         'Table',
         'ChestWallMuscle',
         'GrayMatter',
         'Fibula',
         'SuperiorLongitudinalFasciculus',
         'InferiorCardiacVein',
         'InternalIntercostalMuscle',
         'LateralVentricle',
         'CaudateNucleus',
         'LacrimalArtery',
         'OccipitalArtery',
         'Brain',
         'WasteMaterial',
         'EmbolizationBall',
         'AtrialVein',
         'SuprasternalNotch',
         'AnteriorCardiacVein',
         'Pancreas',
         'BoneOfThorax',
         'DigitalArteryOfHand',
         'SecondLumbarArtery',
         '_1stRightPosterolateralCoronaryArtery',
         'BladderAndUrethra',
         'Suture',
         'Sacrum',
         'MaleExternalGenitalia',
         'RetroperitonealSpace',
         'AnteriorMiddleCerebralArteryBifurcation',
         'SuperiorThyroidArtery',
         'CerebralCortex',
         'TruncusArteriosusCommunis',
         'Tooth',
         'Joint',
         'GreatSaphenousVeinOfCalf',
         'LeftPulmonaryVein',
         'Aorta',
         'Diencephalon',
         'CongenitalCoronaryArteryFistulaToLeftVentricle',
         'NeckChestAbdomenAndPelvis',
         'LatissimusDorsiMuscle',
         'CervicalCollar',
         'CommonHepaticArtery',
         'AnkleJoint',
         'Bronchus',
         'Pelvis',
         'PulmonaryChamberOfCorTriatriatum',
         'UlnarArtery',
         'LumbarRegion',
         'ThoracicSpine',
         'FourthVentricle',
         'OpticTract',
         'HipJoint',
         'TrunkOfPulmonaryArtery',
         'Abscess',
         'BrachialArtery',
         '_2ndDiagonalCoronaryArtery',
         'HipJoint',
         'SerratusAnteriorMuscle',
         'Chair',
         'LymphaticSystem',
         'Stomach',
         'CircumflexCoronaryArtery',
         'AnteriorCommunicatingArtery',
         'InferiorMesentericVein',
         'MedianCubitalVein',
         'CoronarySinus',
         'Tendon',
         'TerminalInternalCarotidArtery',
         'Atrium',
         'CerebralWhiteMatter',
         'Knee',
         'Cyst',
         'SubarachnoidSpace',
         'PulmonaryVeinConfluence',
         'GuidingCatheter',
         'EyeRegion',
         'ThermoluminescentRadiationDosimeter',
         'BrainVentricle',
         'UmbilicalVein',
         'InguinalRegion',
         '_3rdRightPosterolateralCoronaryArtery',
         'EndoNasal',
         'InferiorPhrenicArtery',
         'JejunostomyTube',
         'PortalVein',
         'PosteriorCerebralCommissure',
         'UncinateFasciculus',
         'AnteriorCerebralArtery',
         'Ear',
         'LymphaticVessel',
         'SpinalCord',
         'Pericardium',
         'PosteriorCerebralArteryP2Segment',
         'LeftAuricularAppendage',
         'CircleOfWillis',
         'SaphenopoplitealJunction',
         '_1stSeptalCoronaryArtery',
         'Eyelid',
         'DeepPalmarArchOfRadialArtery',
         'Pleura',
         'Subcostal',
         'Intima',
         'SegmentOfLung',
         'DoddPerforatingVein']}

name_for_cid[3112] = 'AttenuationCorrection'
cid_concepts[3112] = \
{'DCM': ['CTBasedAttenuationCorrection',
         'NoAttenuationCorrection',
         'AlgorithmicAttenuationCorrection',
         'NMTransmissionAttenuationCorrection']}

name_for_cid[4268] = 'CornealTopographyMapValueType'
cid_concepts[4268] = \
{'DCM': ['CornealAxialPowerMap',
         'CornealInstantaneousPowerMap',
         'CornealRefractivePowerMap',
         'CornealElevationMap',
         'CornealWavefrontMap']}

name_for_cid[3407] = 'PurposeofReferencetoObject'
cid_concepts[3407] = \
{'DCM': ['PriorReportForCurrentPatient',
         'PreProcedureLog',
         'AnalysisOrMeasurementsForCurrentProcedure']}

name_for_cid[7014] = 'ExportAdditionalInformationDocumentTitles'
cid_concepts[7014] = \
{'DCM': ['ForTeachingFileExport',
         'ForClinicalTrialExport',
         'ForPublicationExport',
         'ForResearchCollectionExport']}

name_for_cid[12015] = 'FetalGrowthEquationsandTables'
cid_concepts[12015] = \
{'LN': ['CRLByGAShinozuka1996',
        'CRLByGARempen1991',
        'FLByGAHadlock1984',
        'HCACByGACampbell1977',
        'FLByGAASUM2000',
        'FibulaByGAJeanty1983',
        'TCDByGAGoldstein1987',
        'EFWByGAHansmann1986',
        'EFWByGAHadlock1991',
        'FLByGAMerz1988',
        'FLByGAJeanty1982',
        'FLByGAChitty1994',
        'CephalicIndexByGAHadlock1981',
        'CephalicIndexByGAChitty1994',
        'ACByGAASUM2000',
        'ACByGAHadlock1984',
        'ACMeasuredByGAChitty1994',
        'ACDerivedByGAChitty1994',
        'RadiusByGAJeanty1983',
        'FLByGAShinozuka1996',
        'GSByGARempen1991',
        'BPDOuterOuterByGAChitty1994',
        'BPDByGAASUM2000',
        'HCByGAJeanty1982',
        'AxtByGAShinozuka1996',
        'HCDerivedByGAChitty1994',
        'BPDByGAShinozuka1996',
        'HCByGAASUM2000',
        'BPDByGARempen1991',
        'HCByGAHadlock1984',
        'BPDByGAHadlock1984',
        'OFDByGAChitty1994',
        'BPDByGAMerz1988',
        'BPDByGAJeanty1982',
        'HumerusLengthByGAASUM2000',
        'ACByGAMerz1988',
        'OFDByGAASUM2000',
        'ACByGAShinozuka1996',
        'CRLByGAASUM2000',
        'HCByGAMerz1988',
        'BPDOuterInnerByGAChitty1994']}

name_for_cid[8] = 'AngiographicInterventionalDevices'
cid_concepts[8] = \
{'DCM': ['Sphere'],
 'SCT': ['RotationalAtherectomyDevice',
         'GuidingCatheter',
         'Glue',
         'EmbolizationBall',
         'PercutaneousTransluminalAngioplastyBalloon',
         'AtherectomyDevice',
         'EmbolizationCoil',
         'EmbolizationParticulate',
         'Laser',
         'Catheter',
         'MeasuringRuler',
         'Stent',
         'DetachableBalloon']}

name_for_cid[6039] = 'NippleCharacteristic'
cid_concepts[6039] = \
{'SCT': ['NormalShape', 'NippleRetraction']}

name_for_cid[3515] = 'SpecimenCollection'
cid_concepts[3515] = \
{'SCT': ['BloodSamplingFromExtracorporealBloodCircuit',
         'SpecimenCollection',
         'CollectionOfBloodSpecimenForLaboratory']}

name_for_cid[6202] = 'ColonFindingorFeatureModifier'
cid_concepts[6202] = \
{'DCM': ['ChestTube', 'KidneyStent', 'PancreaticStent', 'Coin', 'FiducialMark'],
 'SCT': ['JejunostomyTube',
         'Suture',
         'Lung',
         'IleostomyBag',
         'ColostomyBag',
         'VenaCavaFilter',
         'UrostomyBag',
         'UrostomySet',
         'Appendix',
         'Needle',
         'InferiorVenaCava',
         'Prosthesis',
         'Jewelry',
         'RectalTube',
         'Bone',
         'Prostate',
         'Catheter',
         'Uterus',
         'Staple',
         'Testis',
         'UrethralCatheter',
         'Cervix',
         'Bladder',
         'ScarTissue',
         'UreteralStent',
         'Bullet',
         'BloodVessel',
         'Pin',
         'Aorta',
         'SeminalVesicle',
         'Rectum',
         'Ovary',
         'AppendicealStump',
         'Muscle',
         'ColostomySet',
         'Liver',
         'Spleen',
         'Kidney',
         'AdrenalGland',
         'IleostomySet']}

name_for_cid[3000] = 'AudioChannelSource'
cid_concepts[3000] = \
{'DCM': ['PhysiologicalAudioSignal',
         'Phonocardiogram',
         'OperatorNarrative',
         'Voice',
         'DopplerAudio',
         'AmbientRoomEnvironment']}

name_for_cid[6081] = 'BreastCancerRiskFactors'
cid_concepts[6081] = \
{'DCM': ['IntermediateFamilyHistoryOfBreastCancer',
         'PersonalBreastCancerHistory',
         'VeryStrongFamilyHistoryOfBreastCancer',
         'HistoryOfEndometrialCancer',
         'FamilyHistoryOfProstateCancer',
         'HistoryOfOvarianCancer',
         'FamilyHistoryUnknown',
         'HistoryOfHighRiskLesionOnPreviousBiopsy',
         'PostMenopausalPatient',
         'LateChildBearingAfter30',
         'BRCA1BreastCancerGene',
         'BRCA2BreastCancerGene',
         'BRCA3BreastCancerGene',
         'WeakFamilyHistoryOfBreastCancer'],
 'SCT': ['Nulliparous',
         'FamilyHistoryOfBreastCancer',
         'NoFamilyHistoryOfBreastCarcinoma']}

name_for_cid[9231] = 'WorkitemDefinition'
cid_concepts[9231] = \
{'DCM': ['NoSubsequentWorkitems',
         'AddAddendumToReport',
         'Print',
         'ReportVerification',
         'Transcription',
         'Interpretation',
         'ComputerAidedDetection',
         'ComputerAidedDiagnosis',
         'QualityControl',
         'MediaImport',
         'ImageProcessing']}

name_for_cid[12107] = 'UpperExtremityArteries'
cid_concepts[12107] = \
{'SCT': ['BrachialArtery',
         'InnominateArtery',
         'SuperficialPalmarArch',
         'UlnarArtery',
         'DigitalArteryOfHand',
         'RadialArtery',
         'SubclavianArtery',
         'DeepPalmarArchOfRadialArtery',
         'AxillaryArtery']}

name_for_cid[6019] = 'ClockfaceLocationorRegionfromBIRADS®'
cid_concepts[6019] = \
{'SCT': ['_7OClockPosition',
         '_11OClockPosition',
         '_12OClockPosition',
         '_5OClockPosition',
         '_3OClockPosition',
         '_6OClockPosition',
         '_4OClockPosition',
         'AxillaRegion',
         '_1OClockPosition',
         '_8OClockPosition',
         '_2OClockPosition',
         'CentralRegionOfBreast',
         'AxillaryTailRegion',
         'SubareolarRegion',
         '_9OClockPosition',
         '_10OClockPosition']}

name_for_cid[9303] = 'InterpretationRequestDiscontinuationReasons'
cid_concepts[9303] = \
{'DCM': ['ResourcePreEmpted',
         'PatientDied',
         'IncorrectProcedureOrdered',
         'DiscontinuedForUnspecifiedReason',
         'ObjectSetIncomplete',
         'DoctorCanceledProcedure',
         'DuplicateOrder',
         'InterpretationRequiresSpecialistExpertise',
         'InsufficientQualityForInterpretation',
         'WorkitemAssignmentRejectedByAssignedResource',
         'ChangeOfProcedureForCorrectCharging']}

name_for_cid[10008] = 'DoseRelatedDistanceMeasurements'
cid_concepts[10008] = \
{'DCM': ['DistanceSourceToTablePlane',
         'DistanceSourceToDetector',
         'TableHeightEndPosition',
         'TableLateralEndPosition',
         'TableLateralPosition',
         'TableLongitudinalPosition',
         'TableXEndPositionToIsocenter',
         'TableZPositionToIsocenter',
         'TableZEndPositionToIsocenter',
         'TableYPositionToIsocenter',
         'TableXPositionToIsocenter',
         'TableHeightPosition',
         'TableYEndPositionToIsocenter',
         'DistanceSourceToIsocenter',
         'DistanceSourceToReferencePoint',
         'TableLongitudinalEndPosition']}

name_for_cid[12035] = 'UltrasoundTransducerApplication'
cid_concepts[12035] = \
{'DCM': ['ExternalTransducer',
         'IntravascularTransducer',
         'EndorectalTransducer',
         'EndovaginalTransducer',
         'TransesophagealTransducer']}

name_for_cid[3427] = 'EquipmentEvents'
cid_concepts[3427] = \
{'DCM': ['EquipmentFailure',
         'EquipmentRemoved',
         'EquipmentBroughtToProcedureRoom',
         'EquipmentReady']}

name_for_cid[7464] = 'GeneralRegionofInterestMeasurementModifiers'
cid_concepts[7464] = \
{'DCM': ['Skewness', 'Kurtosis', 'PeakValueWithinROI'],
 'SCT': ['StandardDeviation',
         'Mode',
         'Minimum',
         'Maximum',
         'Total',
         'Mean',
         'Median'],
 'UMLS': ['CoefficientOfVariance', 'Variance', 'RootMeanSquare']}

name_for_cid[228] = 'EquationorTable'
cid_concepts[228] = \
{'DCM': ['TableOfValues',
         'TableOfValuesCitation',
         'MethodCitation',
         'Equation',
         'EquationCitation']}

name_for_cid[12259] = 'CardiacUltrasoundVentriclesMeasurements'
cid_concepts[12259] = \
{'DCM': ['Tau', 'IndicatorMeanTransitTime', 'RRInterval', 'WallMass'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'TissueVelocityTimeIntegralVTIForTheAreaUnderLeftVentricleAWave',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'EndocardialArea',
        'PeakTissueVelocity',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'EFSlope',
        'TimeToLeftVentricleETissueVelocity',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'PreEjectionPeriod',
        'TissueVelocityTimeIntegralVTIForTheAreaUnderLeftVentricleEWave',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'InterventricularSeptumToPosteriorWallThicknessRatio',
        'IsovolumicRelaxationTime',
        'TimeToLeftVentricleSTissueVelocity',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'LeftVentricleMeridionalWallStress',
        'LeftVentricleEToATissueVelocityRatio',
        'SystolicToDiastolicVelocityRatio',
        'EpicardialArea',
        'MyocardialPerformanceIndexTei',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Volume',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'RatioOfMVPeakVelocityToLVPeakTissueVelocityEWave',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[3831] = 'StenosisLength'
cid_concepts[3831] = \
{'SCT': ['Long', 'Short']}

name_for_cid[93] = 'JointPositioningMethod'
cid_concepts[93] = \
{'SCT': ['PassiveMovement', 'ManipulationOfJoint']}

name_for_cid[8115] = 'SpecimenEmbeddingMedia'
cid_concepts[8115] = \
{'SCT': ['EpoxyResin',
         'AcrylicResin',
         'TissueFreezingMedium',
         'ParaffinWax',
         'Plastic',
         'Agar']}

name_for_cid[7063] = 'ModelScaleUnits'
cid_concepts[7063] = \
{'UCUM': ['mm', 'M', 'cm', 'um']}

name_for_cid[3470] = 'VesselLumenCrosssectionalAreaCalculationMethods'
cid_concepts[3470] = \
{'DCM': ['DensitometricMethod', 'CircularMethod']}

name_for_cid[85] = 'SUVUnits'
cid_concepts[85] = \
{'UCUM': ['StandardizedUptakeValueIdealBodyWeight',
          'StandardizedUptakeValueBodySurfaceArea',
          'StandardizedUptakeValueLeanBodyMassJanma',
          'StandardizedUptakeValueBodyWeight',
          'StandardizedUptakeValueLeanBodyMassJames',
          'StandardizedUptakeValueLeanBodyMassJames128Multiplier']}

name_for_cid[6058] = 'ProcedureModifiersforBreast'
cid_concepts[6058] = \
{'DCM': ['SalineImplant',
         'PolyurethaneImplant',
         'PercutaneousSiliconeInjection',
         'CombinationImplant',
         'PrePectoralImplant',
         'NeedleLocalizationAndBiopsy',
         'RetroPectoralImplant',
         'MarkerPlacement',
         'Mammographic',
         'Mammographic',
         'PalpationGuided',
         'Targeted',
         'Survey',
         'VacuumAssisted',
         'Calibration',
         'QualityControl'],
 'SCT': ['Lumpectomy',
         'TissueDopplerImaging',
         'BreastImplantTypeNotSpecified',
         '_3DMode',
         '_2DMode',
         'Mammography',
         'DopplerContinuousWave',
         'MagneticResonanceImagingGuidedBiopsy',
         'Diagnostic',
         'QuadrantectomyOfBreast',
         'SiliconeGelImplant',
         'DopplerPulsed',
         'MMode',
         'DiagnosticRadiographyStereotacticLocalization',
         'Screening',
         'PowerDoppler',
         'UltrasonicGuidanceProcedure',
         'DopplerColorFlow',
         'ComputedTomographyGuidedBiopsy'],
 'UMLS': ['Mastectomy']}

name_for_cid[8103] = 'AnatomicPathologySpecimenTypes'
cid_concepts[8103] = \
{'SCT': ['SpecimenFromBreastObtainedByExcision',
         'SmearSample',
         'Aspirate',
         'SpecimenFromBreastObtainedByTotalMastectomy',
         'SegmentectomyBreastSample',
         'BreastTruCutBiopsySample',
         'BodySubstanceSample',
         'TissueSection',
         'SpecimenFromBreastObtainedByImageGuidedCoreBiopsy',
         'TissueSpot',
         'LumpectomyBreastSample',
         'LiquidBasedCytologicMaterial',
         'BreastDuctSample',
         'SpecimenFromBreastStereotacticallyGuidedCoreNeedleBiopsy',
         'TissueSpecimen',
         'GeneticSample',
         'SpecimenFromBreastObtainedByCoreNeedleBiopsy',
         'CoreSampleOfTissueBlock',
         'BodyFluidSample',
         'FrozenSectionBreastSample',
         'TouchPreparationCytologicMaterial',
         'Slide',
         'EntireBody',
         'SpecimenFromBreastByIncisionalBiopsyOfBreastMass',
         'GrossSpecimen']}

name_for_cid[4209] = 'OphthalmicAnatomicStructureImaged'
cid_concepts[4209] = \
{'SCT': ['BothEyes',
         'AnteriorChamberOfEye',
         'OphthalmicArtery',
         'LowerEyelid',
         'OpticNerveHead',
         'ChoroidOfEye',
         'Sclera',
         'Cornea',
         'Eyelid',
         'Lens',
         'Eye',
         'LacrimalSac',
         'Retina',
         'LacrimalCaruncle',
         'Iris',
         'UpperEyelid',
         'CiliaryBody',
         'Conjunctiva',
         'LacrimalGland',
         'FoveaCentralis']}

name_for_cid[3466] = 'PlaneIdentification'
cid_concepts[3466] = \
{'SCT': ['AnteroPosterior',
         'LeftAnteriorOblique',
         'LeftLateral',
         'CaudalRAO',
         'RightAnteriorOblique',
         'CranialRAO',
         'CaudalLAO',
         'CranialLAO']}

name_for_cid[639] = 'TumorGraftHistologicType'
cid_concepts[639] = \
{'NCIt': ['BronchioloalveolarAdenocarcinoma'],
 'SCT': ['NonSmallCellCarcinoma',
         'RenalCellCarcinoma',
         'SquamousCellCarcinoma',
         'AmelanoticMelanoma',
         'SpindleCellCarcinoma',
         'Glioblastoma',
         'Adenocarcinoma',
         'Carcinosarcoma',
         'MixedSmallCellCarcinoma',
         'Carcinoma',
         'Leukemia',
         'InfiltratingDuctalCarcinomaOfBreast',
         'Mesothelioma',
         'Cystadenocarcinoma',
         'LargeCellCarcinoma',
         'Astrocytoma',
         'Melanoma',
         'Osteosarcoma',
         'Sarcoma',
         'SmallCellCarcinoma']}

name_for_cid[12218] = 'EchocardiographyCongenital'
cid_concepts[12218] = \
{'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'StenosisPeakVelocity',
        'VolumeFlow',
        'FlowRadius',
        'HeartRate',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'EndDiastolicVelocity',
        'ROIInternalDimensionByUS',
        'VelocityOfFlowPropagation',
        'PeakGradient',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter',
         'Length',
         'Circumference']}

name_for_cid[3489] = 'CalciumDistribution'
cid_concepts[3489] = \
{'SCT': ['Deep', 'Superficial']}

name_for_cid[4032] = 'MRSpectroscopyMetabolites'
cid_concepts[4032] = \
{'DCM': ['CreatineCholineCitrateRatio',
         'GlutamateAndGlutamine',
         'CholineCreatineRatio',
         'NAcetylaspartateCreatineRatio',
         'CreatineAndCholine',
         'NAcetylaspartateCholineRatio',
         'LipidAndLactate'],
 'SCT': ['Inositol',
         'Citrate',
         'Choline',
         'Creatine',
         'Lactate',
         'Glutamine',
         'NAcetylaspartate',
         'Lipid',
         'Tuarine']}

name_for_cid[9250] = 'ScheduledProcessingParameterConceptCodesforRTTreatment'
cid_concepts[9250] = \
{'DCM': ['TreatmentDeliveryType']}

name_for_cid[402] = 'AuditActiveParticipantRoleIDCode'
cid_concepts[402] = \
{'DCM': ['SourceMedia',
         'DestinationMedia',
         'SourceRoleID',
         'DestinationRoleID',
         'ApplicationLauncher',
         'Application']}

name_for_cid[12023] = 'PelvisandUterusAnatomicSites'
cid_concepts[12023] = \
{'SCT': ['Endometrium', 'Cervix', 'Uterus']}

name_for_cid[3718] = 'MyocardialWallSegmentsinProjection'
cid_concepts[3718] = \
{'SCT': ['MyocardiumOfPosterolateralRegion',
         'MyocardiumOfApexOfHeart',
         'MyocardiumOfInferolateralRegion',
         'LeftVentricleBasalLateralSegment',
         'LeftVentricularBasalSeptalSegment',
         'LeftVentricleBasalAnteriorSegment',
         'LeftVentricleApicalSeptalSegment',
         'MyocardiumOfDiaphragmaticRegion',
         'LeftVentricleBasalInferiorSegment',
         'LeftVentricularPosterobasalSegment',
         'MyocardiumOfAnterolateralRegion']}

name_for_cid[7022] = 'RadiotherapyPurposesofReference'
cid_concepts[7022] = \
{'DCM': ['RegistrationCreatedDuringTreatment',
         'RTTreatmentPlanForThePositionBeingVerified',
         'RegistrationUsedInPlanning']}

name_for_cid[7472] = 'VolumeMeasurements'
cid_concepts[7472] = \
{'DCM': ['VolumeEstimatedFromThreeOrMoreNonCoplanar2DRegions',
         'VolumeEstimatedFromTwoNonCoplanar2DRegions',
         'VolumeEstimatedFromSingle2DRegion',
         'VolumeOfSphere',
         'VolumeOfCircumscribedSphere',
         'VolumeOfEllipsoid',
         'VolumeOfBoundingThreeDimensionalRegion'],
 'SCT': ['Volume']}

name_for_cid[9529] = 'SingleDoseDosimetricObjectives'
cid_concepts[9529] = \
{'DCM': ['MinimumRadiationDose',
         'MaximumSurfaceRadiationDose',
         'MinimumSurfaceRadiationDose',
         'MinimumEquivalentUniformDose',
         'MaximumMeanRadiationDose',
         'MinimumMeanRadiationDose',
         'MaximumRadiationDose',
         'PrescriptionRadiationDose',
         'MaximumEquivalentUniformDose']}

name_for_cid[7167] = 'PeripheralNervousSystemSegmentationTypes'
cid_concepts[7167] = \
{'SCT': ['CranialNerve',
         'PeripheralNervousSystem',
         'PeripheralNerve',
         'VagusNerve',
         'AutonomicNerve',
         'SympatheticTrunk']}

name_for_cid[3827] = 'VesselSegments'
cid_concepts[3827] = \
{'BARI': ['_1stMarginalCoronaryArteryLaterals',
          '_2ndMarginalCoronaryArteryLaterals',
          '_3rdMarginalCoronaryArteryLaterals',
          'PosteriorDescendingSeptalPerforators',
          '_1stDiagonalCoronaryArteryLaterals',
          '_2ndDiagonalCoronaryArteryLaterals',
          'RamusLaterals',
          '_3rdDiagonalCoronaryArteryLaterals'],
 'SCT': ['ProfundaFemorisVein',
         'SubclavianVein',
         'ExternalIliacArtery',
         'ProximalRightCoronaryArtery',
         'CeliacAxis',
         'OvarianVein',
         'PosteriorCerebralArtery',
         'BasilarArtery',
         'DeepPalmarVenousArch',
         'LateralCalfPerforator',
         'MidCircumflexCoronaryArtery',
         'CentralRetinalVein',
         'IntermediateArteryRamus',
         'GastricArtery',
         'FourthLumbarArtery',
         'SmallestCardiacVein',
         'ThirdLumbarArtery',
         'UmbilicalArtery',
         'RadialArtery',
         'SolealVein',
         'LobarArtery',
         'ThighPerforator',
         'MidRightCoronaryArtery',
         'RightBranchOfHepaticArtery',
         'ExternalIliacVein',
         'SubclavianArtery',
         'CentralRetinalArtery',
         '_3rdLeftPosterolateralCoronaryArtery',
         'OphthalmicArtery',
         'SegmentalArtery',
         'BrachialVein',
         'SuperficialPalmarArch',
         'OvarianArtery',
         'PeronealArtery',
         'SuperficialFemoralVein',
         'LeftPosteriorDescendingCircumflexCoronaryArtery',
         'SuperiorLeftPulmonaryVein',
         '_1stMarginalCoronaryArtery',
         'SuperiorMesentericVein',
         'LumbarArtery',
         'PeronealVein',
         'MidLeftAnteriorDescendingCoronaryArtery',
         'AtrioventricularVein',
         'CoronaryArteryGraft',
         'InfraRenalAorta',
         'PulmonaryVein',
         'CarotidBulb',
         'PoplitealVein',
         'GiacominiVein',
         'InternalCarotidArteryC5Segment',
         'DistalLeftAnteriorDescendingCoronaryArtery',
         'RightPulmonaryVein',
         'RightCoronaryArteryOstium',
         'MarginalCoronaryArtery',
         'HepaticVein',
         'ProximalLeftAnteriorDescendingCoronaryArtery',
         'ExternalCarotidArtery',
         'SuperiorMesentericArtery',
         'PosteriorTibialArtery',
         'GreatSaphenousVeinOfThigh',
         'LeftHepaticVein',
         'InnominateVein',
         'SuperficialFemoralArtery',
         'ProximalCircumflexCoronaryArtery',
         '_2ndMarginalCoronaryArtery',
         'PosteriorCommunicatingArtery',
         'LeftBranchOfHepaticArtery',
         'ArcuateArteryOfTheKidney',
         'RightHepaticVein',
         'GastrocnemiusVein',
         'CephalicVein',
         'GastricVein',
         'LeftMainCoronaryArtery',
         'LeftMainBranchOfPortalVein',
         'InferiorRightHepaticVein',
         'VentricularVein',
         'CarotidSiphon',
         'PlantarArterialArch',
         'AnteriorPosteriorCerebralArteryBifurcation',
         'AtrialVein',
         '_1stDiagonalCoronaryArtery',
         'AnteriorCardiacVein',
         'LeftMainCoronaryArteryOstium',
         'LesserSaphenousVein',
         'InferiorMesentericArtery',
         'DigitalArteryOfHand',
         'GreatSaphenousVein',
         'SecondLumbarArtery',
         '_1stRightPosterolateralCoronaryArtery',
         'InternalCarotidArtery',
         'RightPosteriorAVCoronaryArtery',
         'UlnarVein',
         'SuperiorRightPulmonaryVein',
         'InferiorVenaCava',
         'ProperHepaticArtery',
         'InterlobarArteryOfKidney',
         'AnteriorMiddleCerebralArteryBifurcation',
         'PerforatingArteryOfKidney',
         'InternalCarotidArteryC6Segment',
         'DistalRightCoronaryArtery',
         'AnteriorTibialArtery',
         'LeftAnteriorDescendingCoronaryArtery',
         'CommonCarotidArtery',
         'GreatSaphenousVeinOfCalf',
         'LeftPulmonaryVein',
         '_3rdMarginalCoronaryArtery',
         'Aorta',
         'CommonIliacArteryBifurcation',
         'MiddleHepaticVein',
         'RightMainBranchOfPortalVein',
         'RadialVein',
         'InferiorRightPulmonaryVein',
         'LeftPosterolateralCircumflexCoronaryArtery',
         'CommonHepaticArtery',
         'UterineArtery',
         'FirstLumbarArtery',
         'PoplitealArtery',
         'MiddleCerebralArteryM1Segment',
         'InternalIliacArtery',
         'SuperiorVenaCava',
         'UlnarArtery',
         'SixthLumbarArtery',
         'PosteriorArchVein',
         'SplenicVein',
         'CarotidBifurcation',
         'SaphenousVein',
         '_2ndRightPosterolateralCoronaryArtery',
         'SplenicArtery',
         'InnominateArtery',
         'RenalArtery',
         'GastroduodenalArtery',
         'AccessoryRenalArtery',
         'AzygosVein',
         'BrachialArtery',
         '_2ndDiagonalCoronaryArtery',
         'DistalCircumflexCoronaryArtery',
         'DorsalisPedisArtery',
         'CommonFemoralArtery',
         'SmallCardiacVein',
         'SuperficialPalmarVenousArch',
         'CommonFemoralVein',
         '_1stLeftPosterolateralCoronaryArtery',
         'CircumflexCoronaryArtery',
         'InferiorMesentericVein',
         'AnteriorCommunicatingArtery',
         'MedianCubitalVein',
         'CoronarySinus',
         'HilarArtery',
         'TerminalInternalCarotidArtery',
         'AxillaryArtery',
         'VertebralArtery',
         '_3rdDiagonalCoronaryArtery',
         'TransjugularIntrahepaticPortosystemicShunt',
         'SaphenofemoralJunction',
         'SupraRenalAorta',
         'InternalJugularVein',
         'CockettPerforatingVein',
         'AVGrooveContinuationOfCircumflexArtery',
         'UmbilicalVein',
         'CommonIliacVein',
         'PosteriorCerebralArteryP1Segment',
         'TesticularArtery',
         '_3rdRightPosterolateralCoronaryArtery',
         'GreatCardiacVein',
         'PortalVein',
         'InferiorLeftPulmonaryVein',
         'AnteriorCerebralArtery',
         'RenalVein',
         'InternalIliacVein',
         'BoydPerforatingVein',
         'IlealVein',
         'CommonIliacArtery',
         'MiddleCardiacVein',
         '_2ndLeftPosterolateralCoronaryArtery',
         'MiddleCerebralArteryM2Segment',
         'MiddleCerebralArtery',
         'PosteriorCerebralArteryP2Segment',
         'PosteriorDescendingRightCoronaryArtery',
         'TesticularVein',
         'AnteriorTibialVein',
         'PosteriorTibialVein',
         'SaphenopoplitealJunction',
         'BasilicVein',
         '_1stSeptalCoronaryArtery',
         'RightCoronaryArtery',
         'HunterianPerforatingVein',
         'DeepPalmarArchOfRadialArtery',
         'ProfundaFemorisArtery',
         'FifthLumbarArtery',
         'AxillaryVein']}

name_for_cid[10065] = 'RadiationTransportModelType'
cid_concepts[10065] = \
{'DCM': ['MeasuredRadiationDose',
         'VoxelizedRadiationTransportModel',
         'GeometricRadiationTransportModel',
         'NURBSRadiationTransportModel',
         'MeshRadiationTransportModel',
         'BREPRadiationTransportModel']}

name_for_cid[6131] = 'ChestTextureDescriptor'
cid_concepts[6131] = \
{'DCM': ['Inhomogeneous', 'Homogeneous']}

name_for_cid[300] = 'MultienergyRelevantMaterials'
cid_concepts[300] = \
{'SCT': ['Silicon',
         'Yttrium',
         'Iron',
         'Iodine',
         'UricAcid',
         'CobaltChromiumAlloy',
         'Silicone',
         'Hafnium',
         'Tantalum',
         'Water',
         'Fat',
         'Gold',
         'Calcium',
         'Barium',
         'Gadolinium',
         'RenalStone',
         'NickelCobaltChromium',
         'Bismuth',
         'CalciumHydroxyapatite',
         'Titanium',
         'Platinum',
         'Silver',
         'AluminumOxide']}

name_for_cid[12234] = 'RespirationState'
cid_concepts[12234] = \
{'SCT': ['Expiration', 'Inspiration']}

name_for_cid[245] = 'LateralitywithMedian'
cid_concepts[245] = \
{'DCM': ['Median'], 'SCT': ['Bilateral', 'Right', 'Left', 'Unilateral']}

name_for_cid[615] = 'AnesthesiaMaintenanceCodeTypeforSmallAnimalAnesthesia'
cid_concepts[615] = \
{'SCT': ['InhalationAnesthesiaSystemClosedRebreathingPrimaryAgent',
         'InhalationAnesthesiaSystemClosedNoRebreathingPrimaryAgent']}

name_for_cid[6166] = 'CADGeometrySecondaryGraphicalRepresentation'
cid_concepts[6166] = \
{'DCM': ['OrthogonalLocationArcInnerMargin',
         'OrthogonalLocationArcOuterMargin',
         'OutlineOfLobulations',
         'InnerLimitsOfFuzzyMargin',
         'OuterLimitsOfFuzzyMargin',
         'OutlineOfSpiculations',
         'LinearSpiculation',
         'PixelatedSpiculations',
         'OrthogonalLocationArc'],
 'SCT': ['LongAxis', 'ShortAxis']}

name_for_cid[11] = 'RouteofAdministration'
cid_concepts[11] = \
{'DCM': ['RetroOrbitalRoute'],
 'NCIt': ['IntraepithelialRoute', 'ExtraluminalRoute', 'TransluminalRoute'],
 'SCT': ['IntramedullaryRoute',
         'IntraluminalRoute',
         'TopicalRoute',
         'IntrathecalRoute',
         'OralRoute',
         'IntramuscularRoute',
         'NasalRoute',
         'PerRectum',
         'IntratumorRoute',
         'IntravenousRoute',
         'SubcutaneousRoute',
         'IntraperitonealRoute',
         'ByInhalation',
         'VaginalRoute',
         'IntraArterialRoute',
         'IntraArticularRoute',
         'IntracardiacRoute',
         'IntraventricularRouteCardiac',
         'IntracutaneousRoute',
         'IntracoronaryRoute']}

name_for_cid[10032] = 'ProjectionXRayAcquisitionDeviceTypes'
cid_concepts[10032] = \
{'DCM': ['CassetteBasedProjectionRadiographySystem',
         'IntegratedProjectionRadiographySystem',
         'FluoroscopyGuidedProjectionRadiographySystem']}

name_for_cid[7262] = 'DiffusionTractographyAlgorithmFamilies'
cid_concepts[7262] = \
{'DCM': ['Deterministic',
         'Probabilistic',
         'Global',
         'FACT',
         'Streamline',
         'TEND',
         'Bootstrap',
         'Euler',
         'RungeKutta']}

name_for_cid[3671] = 'ReasonforECGExam'
cid_concepts[3671] = \
{'SCT': ['PreSurgeryTesting',
         'RoutineProcedure',
         'EmergencyDepartmentPatientVisit',
         'EmergencyProcedure',
         'ProcedureInCardiacCareUnit',
         'EvaluationOfMurmur',
         'OutpatientProcedure']}

name_for_cid[6023] = 'SidefromBIRADS®'
cid_concepts[6023] = \
{'SCT': ['LeftBreast', 'RightBreast', 'BothBreasts']}

name_for_cid[3221] = 'StressTestTerminationReasons'
cid_concepts[3221] = \
{'SCT': ['HypotensiveEpisode',
         'EndOfProtocol',
         'Arrhythmia',
         'TargetHeartRateAchieved',
         'Claudication',
         'ChestPain',
         'Syncope',
         'AbnormalECG',
         'Fatigue',
         'PatientRefusedExerciseTest',
         'Dyspnea',
         'HypertensiveEpisode']}

name_for_cid[12263] = 'CardiacUltrasoundVenousReturnPulmonaryMeasurements'
cid_concepts[12263] = \
{'DCM': ['Tau',
         'IndicatorMeanTransitTime',
         'RRInterval',
         'BloodVelocityPeak',
         'BloodVelocityMean',
         'BloodVelocityMinimum',
         'SystolicBloodVelocityMean',
         'SystolicBloodVelocityPeak',
         'DiastolicBloodVelocityMean',
         'DiastolicBloodVelocityPeak'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'AWaveAmplitude',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'LeafletThickness',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'PulmVeinADurationToMVADurationDifference',
        'EjectionTime',
        'PulmonaryVeinADurationMitralValveADurationRatio',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'LeafletSeparation',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'BCSlope',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'TimeFromQWaveToPulmonicValveCloses',
        'CEDistance',
        'PulmonaryArteryPressureUsingAccelTime',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'FractionalShortening',
        'AliasVelocity',
        'PulmonaryVeinAVTIToMitralValveVTIRatio',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'DEExcursion',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[3108] = 'NM/PETProcedures'
cid_concepts[3108] = \
{'SCT': ['PETBrainStudy',
         'PETHeartStudy',
         'RadioisotopeStudyOfRespiratorySystem',
         'PETStudyForLocalizationOfTumor',
         'RadioisotopeStudyOfGenitourinarySystem',
         'PETBreastStudy',
         'RadioisotopeStudyOfMusculoskeletalSystem',
         'RadioisotopeStudyOfHematopoieticSystem',
         'NuclearMedicineDiagnosticProcedureOnNervousSystem',
         'RadioisotopeStudyOfGastrointestinalSystem',
         'NuclearMedicineCardiovascularStudy',
         'RadioisotopeStudyOfEndocrineSystem',
         'RadionuclideStudyForLocalizationOfInflammatoryDisease',
         'RadionuclideLocalizationOfTumor']}

name_for_cid[9505] = 'FixationorPositioningDevices'
cid_concepts[9505] = \
{'DCM': ['RectalBalloon',
         'Mold',
         'HeadAndNeckMask',
         'HeadMask',
         'Headframe',
         'VaginalCylinder',
         'BodyFrame',
         'BreastBoard',
         'Cast',
         'WholeBodyPod',
         'VacuumMold'],
 'SCT': ['Chair', 'Headrest', 'BiteBlock', 'Table']}

name_for_cid[4272] = 'OPTScanPatternTypes'
cid_concepts[4272] = \
{'DCM': ['GridBScanPattern',
         'CircleRadialBScanPattern',
         'CubeBScanPattern',
         'RasterBScanPattern',
         'RadialBScanPattern',
         'LineBScanPattern',
         'CircleBScanPattern',
         'CrossBScanPattern',
         'CircleRasterBScanPattern',
         'ConcentricCircleBScanPattern']}

name_for_cid[642] = 'Cytokines'
cid_concepts[642] = \
{'SCT': ['InterferonGamma',
         'VascularEndothelialGrowthFactor',
         'TumorNecrosisFactorAlpha']}

name_for_cid[212] = 'GenericAnatomicLocationModifiers'
cid_concepts[212] = \
{'DCM': ['Median'],
 'NCIt': ['Middle'],
 'SCT': ['Distal',
         'Proximal',
         'MidLongitudinal',
         'Central',
         'Lateral',
         'Posterior',
         'Inferior',
         'Anterior',
         'Superior']}

name_for_cid[10024] = 'WaterEquivalentDiameterMethod'
cid_concepts[10024] = \
{'DCM': ['AAPM220']}

name_for_cid[12019] = 'OBGYNFetusSummary'
cid_concepts[12019] = \
{'LN': ['EFWPercentileRank',
        'FetalHeartRate',
        'GestationalAgeByLMP',
        'GestationalAge',
        'EstimatedWeight',
        'CompositeUltrasoundAge']}

name_for_cid[3688] = 'ElectrophysiologyWaveformVoltages'
cid_concepts[3688] = \
{'MDC': ['PMinimumAmplitudePerLead',
         'R1AmplitudePerLead',
         'S2AmplitudePerLead',
         'AmplitudeAt20msIntoSTSegmentPerLead',
         'STAmplitudePerLead',
         'TMinimumAmplitudePerLead',
         'SAmplitudePerLead',
         'S1AmplitudePerLead',
         'AmplitudeAt40msIntoSTSegmentPerLead',
         'QAmplitudePerLead',
         'RAmplitudePerLead',
         'R3AmplitudePerLead',
         'AmplitudeAt60msIntoSTSegmentPerLead',
         'TMaximumAmplitudePerLead',
         'JPointAmplitudePerLead',
         'R2AmplitudePerLead',
         'P3AmplitudePerLead',
         'AmplitudeAt80msIntoSTSegmentPerLead',
         'S3AmplitudePerLead',
         'PMaximumAmplitudePerLead']}

name_for_cid[3722] = 'DiabeticTherapy'
cid_concepts[3722] = \
{'SCT': ['DiabeticOnOralTreatment',
         'DiabeticOnInsulin',
         'DiabeticOnDietaryTreatment']}

name_for_cid[3237] = 'WallMotionComparisonFindings'
cid_concepts[3237] = \
{'SCT': ['NewWallMotionAbnormality', 'NoChange', 'ImprovementOfWallMotion']}

name_for_cid[6035] = 'CompositeFeatureRelations'
cid_concepts[6035] = \
{'DCM': ['TargetContentItemsAreRelatedTemporally',
         'TargetContentItemsAreRelatedSpatially',
         'TargetContentItemsAreRelatedContraLaterally']}

name_for_cid[3667] = 'FraminghamScores'
cid_concepts[3667] = \
{'DCM': ['ComparativeLow10YearCHDRisk',
         '_10YearCHDRisk',
         'ComparativeAverage10YearCHDRisk']}

name_for_cid[4] = 'AnatomicRegion'
cid_concepts[4] = \
{'DCM': ['Phantom'],
 'SCT': ['LeftHypochondriacRegion',
         'SubclavianVein',
         'ExternalIliacArtery',
         'BoneOfLowerLimb',
         'TruncusCoeliacus',
         'BasilarArtery',
         'ParanasalSinus',
         'ApexOfRightVentricle',
         'WristJoint',
         'FontanelOfSkull',
         'Mouth',
         'GenicularArtery',
         'LeftVentricleInflow',
         'SigmoidColon',
         'PharynxAndLarynx',
         'CarotidArtery',
         'BuccalRegionOfFace',
         'UmbilicalArtery',
         'PrimitiveAorta',
         'LeftVentricleOutflowTract',
         'RadialArtery',
         'Intracranial',
         'RightFemoralArtery',
         'JawRegion',
         'LeftFemoralArtery',
         'LumenOfBloodVessel',
         'Calcaneus',
         'EndoEsophageal',
         'AntecubitalVein',
         'Mandible',
         'Parasternal',
         'BroadLigament',
         'PatentDuctusArteriosus',
         'MesentericVein',
         'PulmonaryVein',
         'CoronaryArtery',
         'UpperLimb',
         'ExternalIliacVein',
         'CongenitalPulmonaryArteriovenousFistula',
         'DescendingAorta',
         'SubclavianArtery',
         'Flank',
         'Maxilla',
         'Lung',
         'OphthalmicArtery',
         'Nose',
         'PancreaticDuct',
         'Hand',
         'BrachialVein',
         'RightUpperQuadrantOfAbdomen',
         'Eye',
         'IntraThoracic',
         'Prostate',
         'ElbowJoint',
         'HepaticArtery',
         'Back',
         'FacialBones',
         'Sternum',
         'PeronealArtery',
         'NeckAndChest',
         'RightVentricleOutflowTract',
         'Duodenum',
         'RightHypochondriacRegion',
         'SuperiorLeftPulmonaryVein',
         'BileDuct',
         'SystemicCollateralArteryToLung',
         'UpperOuterQuadrantOfBreast',
         'LumbarArtery',
         'PancreaticDuctAndBileDuctSystems',
         'ApexOfLeftVentricle',
         'Cornea',
         'Larynx',
         'PulmonaryArtery',
         'FemoralArtery',
         'Ovary',
         'Skull',
         'ThoracoLumbarSpine',
         'EndoUrethral',
         'Thyroid',
         'Head',
         'AdrenalGland',
         'AcromioclavicularJoint',
         'UmbilicalRegion',
         'Calyx',
         'CongenitalCoronaryArteryFistulaToRightVentricle',
         'PosteriorMedialTributary',
         'SternoclavicularJoint',
         'EndoNasopharyngeal',
         'EsophagusStomachAndDuodenum',
         'Breast',
         'Neck',
         'Finger',
         'Submental',
         'CranialVenousSystem',
         'PoplitealFossa',
         'PulmonaryArteryConduit',
         'RightAtrium',
         'RightVentricle',
         'CommonVentricle',
         'Coccyx',
         'Abdomen',
         'HeadAndNeck',
         'Hypopharynx',
         'Ventricle',
         'Uterus',
         'TemporomandibularJoint',
         'TracheaAndBronchus',
         'HepaticVein',
         'Testis',
         'PelvisAndLowerExtremities',
         'InternalMammaryArtery',
         'Bladder',
         'Vulva',
         'LargeIntestine',
         'Forearm',
         'Face',
         'Spine',
         'Foot',
         'NeckChestAndAbdomen',
         'ExternalCarotidArtery',
         'SuperiorMesentericArtery',
         'MuscleOfUpperLimb',
         'PosteriorTibialArtery',
         'FemoralVein',
         'RightAuricularAppendage',
         'Zygoma',
         'InnominateVein',
         'SuperficialFemoralArtery',
         'UterusAndFallopianTubes',
         'LowerLimb',
         'PosteriorCommunicationArtery',
         'JuxtaposedAtrialAppendage',
         'AorticArch',
         'VertebralColumnAndCranium',
         'Jejunum',
         'ChestAbdomenAndPelvis',
         'AbdominalAorta',
         'EndoCardiac',
         'TarsalJoint',
         'Scalp',
         'CephalicVein',
         'Tongue',
         'LingualArtery',
         'GastricVein',
         'CongenitalCoronaryArteryFistulaToRightAtrium',
         'Fibula',
         'OrbitalStructure',
         'Penis',
         'InferiorCardiacVein',
         'Baffle',
         'LacrimalArtery',
         'OccipitalArtery',
         'EndoVascular',
         'Sclera',
         'Brain',
         'SystemicVenousAtrium',
         'Parathyroid',
         'AnomalousPulmonaryVein',
         'SuprasternalNotch',
         'CervicoThoracicSpine',
         'Perineum',
         'EndoRectal',
         'Femur',
         'AnteriorCardiacVein',
         'Pancreas',
         'LeftAtrium',
         'CerebralArtery',
         'Scapula',
         'InferiorMesentericArtery',
         'EndoRenal',
         'GreatSaphenousVein',
         'OmentalBursa',
         'BladderAndUrethra',
         'CervicalSpine',
         'IliacAndOrFemoralArtery',
         'InternalCarotidArtery',
         'EndoUreteric',
         'Sacrum',
         'SmallIntestine',
         'InternalAuditoryCanal',
         'SuperiorRightPulmonaryVein',
         'IntraArticular',
         'RenalPelvis',
         'InferiorVenaCava',
         'RightPulmonaryArtery',
         'Axilla',
         'Retroperitoneum',
         'Thymus',
         'VenousNetwork',
         'CommonAtrium',
         'Humerus',
         'Ilium',
         'AorticFistula',
         'SuperiorThyroidArtery',
         'IliacArtery',
         'NasalBone',
         'AnteriorTibialArtery',
         'Cervix',
         'TruncusArteriosusCommunis',
         'EndoArterial',
         'CommonCarotidArtery',
         'Joint',
         'Buttock',
         'SubmandibularArea',
         'OpticCanal',
         'Aorta',
         'SuperficialTemporalArtery',
         'EndoVenous',
         'MastoidBone',
         'CongenitalCoronaryArteryFistulaToLeftVentricle',
         'NeckChestAbdomenAndPelvis',
         'Thigh',
         'Spleen',
         'ChestAndAbdomen',
         'InferiorRightPulmonaryVein',
         'Trachea',
         'BodyConduit',
         'LowerOuterQuadrantOfBreast',
         'AnkleJoint',
         'EndoVesical',
         'Mediastinum',
         'Cerebellum',
         'RightVentricleInflow',
         'MesentericArtery',
         'PoplitealArtery',
         'LeftLowerQuadrantOfAbdomen',
         'UpperUrinaryTract',
         'PulmonaryVenousAtrium',
         'Bronchus',
         'Pelvis',
         'SesamoidBonesOfFoot',
         'InternalIliacArtery',
         'PulmonaryChamberOfCorTriatriatum',
         'SuperiorVenaCava',
         'UlnarArtery',
         'LeftVentricle',
         'AnteriorSpinalArtery',
         'AbdomenAndPelvis',
         'LumbarRegion',
         'Omentum',
         'AnusRectumAndSigmoidColon',
         'SplenicVein',
         'Colon',
         'SubmandibularGland',
         'LeftLumbarRegion',
         'ThoracicSpine',
         'SaphenousVein',
         'Gallbladder',
         'InnominateArtery',
         'SplenicArtery',
         'RenalArtery',
         'HipJoint',
         'UpperArm',
         'Extremity',
         'FacialArtery',
         'EpigastricRegion',
         'AzygosVein',
         'BrachialArtery',
         'Liver',
         'IntraAbdominal',
         'HipJoint',
         'Rib',
         'BoneOfUpperLimb',
         'EndoVaginal',
         'LowerInnerQuadrantOfBreast',
         'Clavicle',
         'ParotidGland',
         'CommonFemoralArtery',
         'Stomach',
         'Subxiphoid',
         'UpperInnerQuadrantOfBreast',
         'AnteriorCommunicatingArtery',
         'CongenitalCoronaryArteryFistulaToLeftAtrium',
         'RightInguinalRegion',
         'ThoracicAorta',
         'CoronarySinus',
         'AxillaryArtery',
         'LeftPulmonaryArtery',
         'ApexOfLung',
         'Ileum',
         'Atrium',
         'Knee',
         'VertebralArtery',
         'RightLowerQuadrantOfAbdomen',
         'PulmonaryVeinConfluence',
         'SaphenofemoralJunction',
         'InternalJugularVein',
         'EyeRegion',
         'LacrimalArteryOfRightEye',
         'LumboSacralSpine',
         'SacroiliacJoint',
         'Anus',
         'UmbilicalVein',
         'EntireBody',
         'CommonIliacVein',
         'InguinalRegion',
         'EndoNasal',
         'HypogastricRegion',
         'RightLumbarRegion',
         'Thumb',
         'LumbarSpine',
         'GreatCardiacVein',
         'PrimitivePulmonaryArtery',
         'SellaTurcica',
         'PortalVein',
         'Scrotum',
         'InferiorLeftPulmonaryVein',
         'LowerLeg',
         'RenalVein',
         'BoydPerforatingVein',
         'Ear',
         'SupraclavicularRegionOfNeck',
         'CommonIliacArtery',
         'Toe',
         'Shoulder',
         'SpinalCord',
         'VascularGraft',
         'LeftAuricularAppendage',
         'CircleOfWillis',
         'LeftInguinalRegion',
         'Eyelid',
         'MuscleOfLowerLimb',
         'LeftUpperQuadrantOfAbdomen',
         'AscendingAorta',
         'HunterianPerforatingVein',
         'IntraPelvic',
         'Subcostal',
         'ProfundaFemorisArtery',
         'Pharynx',
         'ExternalAuditoryCanal',
         'AxillaryVein',
         'Patella',
         'Nasopharynx',
         'OccipitalVein',
         'DoddPerforatingVein']}

name_for_cid[7274] = 'MRDiffusionModelFittingMethods'
cid_concepts[7274] = \
{'DCM': ['LeastSquaresFitOfMultipleSamples',
         'LevenbergMarquardt',
         'TrustRegion',
         'FixedDp',
         'SegmentedUnconstrained',
         'SegmentedConstrained',
         'BayesianProbability',
         'LogOfRatioOfTwoSamples']}

name_for_cid[9513] = 'FixationDevices'
cid_concepts[9513] = \
{'DCM': ['RectalBalloon',
         'Mold',
         'HeadAndNeckMask',
         'HeadMask',
         'Headframe',
         'VaginalCylinder',
         'BodyFrame',
         'BreastBoard',
         'Cast',
         'WholeBodyPod',
         'VacuumMold'],
 'SCT': ['Headrest', 'BiteBlock']}

name_for_cid[8302] = 'MeasurementPatternCodes'
cid_concepts[8302] = \
{'DCM': ['TG18UNL80Pattern', 'TG18UNL10Pattern']}

name_for_cid[12275] = 'CardiacUltrasoundCoronaryArteriesMeasurements'
cid_concepts[12275] = \
{'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'StenosisPeakVelocity',
        'VolumeFlow',
        'FlowRadius',
        'HeartRate',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'EndDiastolicVelocity',
        'ROIInternalDimensionByUS',
        'VelocityOfFlowPropagation',
        'PeakGradient',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter',
         'Length',
         'Circumference']}

name_for_cid[7018] = 'MiscellaneousDocumentTitles'
cid_concepts[7018] = \
{'DCM': ['AdditionalTeachingFileInformation']}

name_for_cid[4264] = 'OphthalmicMapPurposesofReference'
cid_concepts[4264] = \
{'DCM': ['Localizer',
         'SourceImageForImageProcessingOperation',
         'RelatedOphthalmicThicknessMap']}

name_for_cid[6127] = 'SegmentLocationinLung'
cid_concepts[6127] = \
{'SCT': ['AnteriorSegmentOfLeftUpperLobe',
         'AnteriorSegmentOfRightUpperLobe',
         'PosteriorSegmentOfRightUpperLobe']}

name_for_cid[50] = 'InstanceAvailabilityStatus'
cid_concepts[50] = \
{'DCM': ['Unavailable', 'Nearline', 'Offline', 'Online']}

name_for_cid[3630] = 'CardiovascularAnatomicLocations'
cid_concepts[3630] = \
{'SCT': ['SubclavianVein',
         'InternalCarotidArtery',
         'PulmonaryValve',
         'BasilarArtery',
         'PulmonaryArteriovenousFistula',
         'AorticValve',
         'LateralPlantarArtery',
         'RightVentricleApex',
         'InferiorVenaCava',
         'RightPulmonaryArtery',
         'LeftVentricleInflow',
         'GeniculateArtery',
         'CentralVenousSystem',
         'CarotidArtery',
         'CommonAtrium',
         'UmbilicalArtery',
         'NeoAorta',
         'AorticFistula',
         'LeftVentricleOutflowTract',
         'RadialArtery',
         'RightFemoralArtery',
         'SuperiorThyroidArtery',
         'IliacArtery',
         'LeftFemoralArtery',
         'TruncusArteriosusCommunis',
         'Artery',
         'AntecubitalVein',
         'CommonCarotidArtery',
         'Aorta',
         'PatentDuctusArteriosus',
         'SuperficialTemporalArtery',
         'MesentericVein',
         'Vein',
         'FistulaCoronaryToLeftVentricle',
         'PulmonaryVein',
         'CoronaryArtery',
         'DescendingAorta',
         'SubclavianArtery',
         'TibialArtery',
         'RightVentricleInflow',
         'OphthalmicArtery',
         'MesentericArtery',
         'MedialPlantarArtery',
         'VenaAnonyma',
         'PoplitealArtery',
         'PulmonaryVenousAtrium',
         'BrachialVein',
         'PulmonaryChamberInCorTriatriatum',
         'PulmonaryCapillaryWedge',
         'SuperiorVenaCava',
         'LeftVentricle',
         'AnteriorSpinalArtery',
         'HepaticArtery',
         'SplenicVein',
         'PeronealArtery',
         'RightVentricleOutflowTract',
         'SaphenousVein',
         'BrachiocephalicTrunk',
         'RenalArtery',
         'SystemicCollateralArteryToLung',
         'FacialArtery',
         'LumbarArtery',
         'LeftVentricleApex',
         'VentricularSeptalDefect',
         'AzygosVein',
         'PulmonaryArtery',
         'FemoralArtery',
         'BrachialArtery',
         'FistulaCoronaryToRightAtrium',
         'MitralValve',
         'FistulaCoronaryToRightVentricle',
         'PosteriorMedialTributary',
         'CoarctationOfAorta',
         'CommonFemoralArtery',
         'PulmonaryArteryConduit',
         'TricuspidValve',
         'RightAtrium',
         'AnteriorCommunicatingArtery',
         'FistulaCoronaryToLeftAtrium',
         'RightVentricle',
         'ThoracicAorta',
         'CommonVentricle',
         'AtrialSeptalDefect',
         'PulmonaryVeinWedge',
         'AxillaryArtery',
         'LeftPulmonaryArtery',
         'HepaticVein',
         'InternalMammaryArtery',
         'VertebralArtery',
         'PulmonaryVeinConfluence',
         'SaphenofemoralJunction',
         'InternalJugularVein',
         'UmbilicalVein',
         'FemoralVein',
         'InnominateVein',
         'NeonatalPulmonaryArteryPrimitivePA',
         'SuperficialFemoralArtery',
         'PortalVein',
         'PosteriorCommunicatingArtery',
         'JuxtaposedAppendage',
         'AorticArch',
         'RenalVein',
         'BoydPerforatingVein',
         'AbdominalAorta',
         'CoronarySinus',
         'CephalicVein',
         'LingualArtery',
         'GastricVein',
         'Baffle',
         'LacrimalArtery',
         'BasilicVein',
         'OccipitalArtery',
         'SystemicVenousAtrium',
         'HunterianPerforatingVein',
         'AscendingAorta',
         'AnomalousPulmonaryVein',
         'ProfundaFemorisArtery',
         'LeftAtrium',
         'CerebralArtery',
         'AxillaryVein',
         'GreatSaphenousVein',
         'PulmonaryArteryWedge',
         'DoddPerforatingVein']}

name_for_cid[6062] = 'InterventionalProcedureComplications'
cid_concepts[6062] = \
{'DCM': ['AbnormalDischarge', 'NoComplications'],
 'SCT': ['Pneumothorax',
         'Weal',
         'HealthcareAssociatedInfection',
         'Swelling',
         'HematomaPostoperative',
         'HemorrhagePostprocedure',
         'Rash',
         'VasovagalSyncope',
         'PersistentPainFollowingProcedure']}

name_for_cid[12222] = 'OrificeFlowProperties'
cid_concepts[12222] = \
{'LN': ['DESlope',
        'ACInterval',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'AWavePeakVelocity',
        'PeakSystolicVelocity',
        'StenosisPeakGradient',
        'StenosisPeakVelocity',
        'VolumeFlow',
        'FlowRadius',
        'TimeAveragedPeakVelocity',
        'EndDiastolicVelocity',
        'VelocityOfFlowPropagation',
        'PeakGradient',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'PeakInstantaneousFlowRate',
        'AccelerationSlope',
        'DecelerationSlope',
        'DecelerationTime',
        'AccelerationTime',
        'SystolicToDiastolicVelocityRatio',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'TimeAveragedMeanVelocity'],
 'SCT': ['RegurgitantFraction',
         'CardiovascularOrificeArea',
         'CardiovascularOrificeDiameter']}

name_for_cid[9401] = 'IEC61217DevicePositionParameters'
cid_concepts[9401] = \
{'DCM': ['IEC61217TableTopLongitudinalPosition',
         'IEC61217TableTopVerticalPosition',
         'IEC61217GantryContinuousRollAngle',
         'IEC61217GantryContinuousPitchAngle',
         'IEC61217GantryContinuousYawAngle',
         'IEC61217PatientSupportContinuousAngle',
         'IEC61217TableTopContinuousPitchAngle',
         'IEC61217TableTopContinuousRollAngle',
         'IEC61217TableTopEccentricAxisDistance',
         'IEC61217TableTopContinuousEccentricAngle',
         'IEC61217TableTopLateralPosition']}

name_for_cid[4233] = 'OphthalmicAxialLengthMeasurementsSegmentNames'
cid_concepts[4233] = \
{'DCM': ['SingleOrAnteriorLens', 'PosteriorLens'],
 'SCT': ['AnteriorChamber', 'VitreousCavity', 'Cornea']}

name_for_cid[603] = 'AnimalRoomTypes'
cid_concepts[603] = \
{'DCM': ['AnimalHousingRoom', 'PreparationRoom', 'ImagingProcedureRoom'],
 'SCT': ['InductionRoom', 'IsolationRoom', 'RecoveryRoom']}

name_for_cid[12223] = 'EchocardiographyStrokeVolumeOrigin'
cid_concepts[12223] = \
{'SCT': ['RightVentricleOutflowTract',
         'LeftVentricle',
         'MitralValve',
         'LeftVentricleOutflowTract',
         'Aorta']}

name_for_cid[4232] = 'VitreousStatus'
cid_concepts[4232] = \
{'SCT': ['PostVitrectomy',
         'GasInVitreousCavity',
         'VitreousOnly',
         'SiliconeOil']}

name_for_cid[602] = 'BiosafetyControlReasons'
cid_concepts[602] = \
{'SCT': ['BiohazardousMaterial', 'Carcinogen', 'PatientImmunocompromised'],
 'UMLS': ['TransgenicAnimal']}

name_for_cid[252] = 'SMLSizeDescriptor'
cid_concepts[252] = \
{'SCT': ['Medium', 'Small', 'Large']}

name_for_cid[3774] = 'SocialHistory'
cid_concepts[3774] = \
{'SCT': ['Nutrition',
         'AlcoholConsumption',
         'TobaccoSmokingBehavior',
         'Exercise',
         'DrugMisuseBehavior']}

name_for_cid[6126] = 'LocationinLung'
cid_concepts[6126] = \
{'DCM': ['Subpleural'],
 'SCT': ['LowerLobeOfLung',
         'UpperZoneOfLung',
         'LowerZoneOfLung',
         'UpperLobeOfLung',
         'MiddleZoneOfLung',
         'MiddleLobeOfLung']}

name_for_cid[7222] = 'ParametricMapDerivationImagePurposeofReference'
cid_concepts[7222] = \
{'DCM': ['SourceImageForImageProcessingOperation']}

name_for_cid[6063] = 'InterventionalProcedureResults'
cid_concepts[6063] = \
{'SCT': ['Benign',
         'Malignant',
         'InsufficientSample',
         'HighRiskTumor',
         'IndeterminateResult']}

name_for_cid[3261] = 'StressProtocols'
cid_concepts[3261] = \
{'SCT': ['StressTestUsingBicycleErgometer',
         'EllestadProtocol',
         'PharmacologicStressProtocol',
         'DobutamineStressProtocol',
         'NaughtonProtocol',
         'PepperProtocol',
         'AdenosineStressProtocol',
         'StressTestUsingCardiacPacing',
         'PharmacologicAndExerciseStressTest',
         'ModifiedNaughtonProtocol',
         'DipyridamoleStressProtocol',
         'RampProtocol',
         'ExerciseStressECGTest',
         'BalkeProtocol',
         'ModifiedBruceProtocol',
         'BruceProtocol']}

name_for_cid[9512] = 'TomotherapeuticRadiotherapyProcedureTechniques'
cid_concepts[9512] = \
{'DCM': ['TopographicBeam', 'HelicalBeam']}

name_for_cid[8303] = 'DisplayDeviceType'
cid_concepts[8303] = \
{'DCM': ['CRTFrontProjectionSystem',
         'DLPRearProjectionSystem',
         'OLED',
         'CRTRearProjectionSystem',
         'DLPFrontProjectionSystem',
         'CRTDisplay',
         'PlasmaDisplay',
         'LiquidCrystalDisplay']}

name_for_cid[12274] = 'CardiacUltrasoundAortaMeasurements'
cid_concepts[12274] = \
{'DCM': ['Tau', 'IndicatorMeanTransitTime', 'RRInterval'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'AorticSinotubularJunctionToAorticRootRatio',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'AorticValveCuspSeparation',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[7019] = 'SegmentationNonImageSourcePurposesofReference'
cid_concepts[7019] = \
{'DCM': ['SourceRealWorldValueMap']}

name_for_cid[7449] = 'ReaderSpecialty'
cid_concepts[7449] = \
{'DCM': ['OncologicImagingSpecialty',
         'OBGynImagingSpecialty',
         'ThoracicImagingSpecialty',
         'OncologySpecialty',
         'AbdominalImagingSpecialty',
         'VascularImagingSpecialty',
         'PediatricImagingSpecialty',
         'HeadAndNeckImagingSpecialty',
         'CardiacImagingSpecialty',
         'MusculoskeletalImagingSpecialty',
         'NeuroradiologyImagingSpecialty',
         'NeurologySpecialty']}

name_for_cid[4265] = 'OphthalmicThicknessDeviationCategories'
cid_concepts[4265] = \
{'DCM': ['PGreaterThan5Percent',
         'PLesserThan5Percent',
         'PLesserThan2Percent',
         'PLesserThan1Percent',
         'PLesserThan0Point5Percent']}

name_for_cid[12018] = 'OBGYNSummary'
cid_concepts[12018] = \
{'LN': ['NumberOfFetusesByUS', 'GestationalAgeByOvulationDate']}

name_for_cid[10025] = 'RadiationDoseReferencePoints'
cid_concepts[10025] = \
{'DCM': ['_15cmFromIsocenterTowardSource',
         '_1cmAboveTabletop',
         '_30cmInFrontOfImageInputSurface',
         '_15cmFromTableCenterline',
         'InDetectorPlane',
         '_30cmAboveTabletop',
         'EntranceExposureToA4Point2cmBreastThickness',
         'AtSurfaceOfPatient']}

name_for_cid[3689] = 'ECGGlobalWaveformDurations'
cid_concepts[3689] = \
{'MDC': ['PDurationGlobal',
         'PPTimePeriodGlobal',
         'RRTimePeriodGlobal',
         'QTDurationGlobal',
         'PQSegmentTimePeriodGlobal',
         'QRSDurationGlobal',
         'PQTimePeriodGlobal',
         'PRTimePeriodGlobal',
         'QTUTimePeriodGlobal']}

name_for_cid[3723] = 'MITypes'
cid_concepts[3723] = \
{'SCT': ['NonSTElevationMyocardialInfarction',
         'STElevationMyocardialInfarction']}

name_for_cid[3236] = 'ToleranceComparisonFindings'
cid_concepts[3236] = \
{'SCT': ['DecreasedTolerance', 'NoChange', 'IncreasedTolerance']}

name_for_cid[6034] = 'IntendedUseofCADOutput'
cid_concepts[6034] = \
{'DCM': ['PresentationRequiredRenderingDeviceIsExpectedToPresent',
         'PresentationOptionalRenderingDeviceMayPresent',
         'NotForPresentationRenderingDeviceExpectedNotToPresent']}

name_for_cid[4009] = 'DXAnatomyImaged'
cid_concepts[4009] = \
{'DCM': ['Phantom'],
 'SCT': ['BoneOfLowerLimb',
         'Ureter',
         'Sacrum',
         'SmallIntestine',
         'ParanasalSinus',
         'InternalAuditoryCanal',
         'WristJoint',
         'Humerus',
         'Ilium',
         'JawRegion',
         'Calcaneus',
         'NasalBone',
         'Esophagus',
         'Mandible',
         'OpticCanal',
         'MastoidBone',
         'NeckChestAbdomenAndPelvis',
         'Thigh',
         'ChestAndAbdomen',
         'Trachea',
         'UpperLimb',
         'AnkleJoint',
         'Maxilla',
         'Mediastinum',
         'Hand',
         'UpperUrinaryTract',
         'Bronchus',
         'Pelvis',
         'SesamoidBonesOfFoot',
         'Eye',
         'Chest',
         'Prostate',
         'ElbowJoint',
         'AbdomenAndPelvis',
         'FacialBones',
         'Colon',
         'Sternum',
         'SubmandibularGland',
         'ThoracicSpine',
         'NeckAndChest',
         'Duodenum',
         'Gallbladder',
         'BileDuct',
         'UpperArm',
         'Extremity',
         'PancreaticDuctAndBileDuctSystems',
         'Larynx',
         'Skull',
         'ThoracoLumbarSpine',
         'Urethra',
         'HipJoint',
         'Rib',
         'BoneOfUpperLimb',
         'Head',
         'AcromioclavicularJoint',
         'Clavicle',
         'ParotidGland',
         'SternoclavicularJoint',
         'Breast',
         'EsophagusStomachAndDuodenum',
         'Neck',
         'Finger',
         'Stomach',
         'Coccyx',
         'Abdomen',
         'HeadAndNeck',
         'TemporomandibularJoint',
         'ApexOfLung',
         'Ileum',
         'PelvisAndLowerExtremities',
         'Bladder',
         'Knee',
         'LargeIntestine',
         'Forearm',
         'EyeRegion',
         'Spine',
         'LumboSacralSpine',
         'Foot',
         'NeckChestAndAbdomen',
         'SacroiliacJoint',
         'Anus',
         'MuscleOfUpperLimb',
         'EntireBody',
         'Thumb',
         'LumbarSpine',
         'Zygoma',
         'SellaTurcica',
         'UterusAndFallopianTubes',
         'LowerLimb',
         'LowerLeg',
         'VertebralColumnAndCranium',
         'Jejunum',
         'ChestAbdomenAndPelvis',
         'Heart',
         'TarsalJoint',
         'Toe',
         'Shoulder',
         'Fibula',
         'OrbitalStructure',
         'MuscleOfLowerLimb',
         'CervicoThoracicSpine',
         'Femur',
         'Rectum',
         'Scapula',
         'Patella',
         'CervicalSpine']}

name_for_cid[3666] = 'P50Equations'
cid_concepts[3666] = \
{'DCM': ['IfAgeLesserThan30DayP50Equals24Point6',
         'IfAgeLesserThan18YearsP50Equals27Point2',
         'IfPremAgeLesserThan3DaysP50Equals19Point9',
         'IfAgeLesserThan1DayP50Equals21Point6',
         'IfAgeGreaterThan60YearsP50Equals29Point3',
         'IfAgeLesserThan40YearsP50Equals27Point4']}

name_for_cid[5] = 'TransducerApproach'
cid_concepts[5] = \
{'SCT': ['Peripheral',
         'Midline',
         'Ascending',
         'Right',
         'Surface',
         'Caudal',
         'Posterior',
         'Left',
         'External',
         'Superior',
         'Distal',
         'PiaMater',
         'Proximal',
         'Parasagittal',
         'Intraluminal',
         'Lateral',
         'Inferior',
         'Capsular',
         'Anterior',
         'WaterBagProsthesisForImagingProcedure',
         'Transverse',
         'GelProsthesisForImagingProcedure',
         'Cranial',
         'Central',
         'Apical',
         'Parenchyma',
         'SalineBagProsthesisForImagingProcedure',
         'MidLongitudinal',
         'SubcutaneousTissue',
         'Lumen',
         'ExternalProsthesisForSonographicProcedureStandOff',
         'DuraMater',
         'ContactWith',
         'Internal',
         'Medial',
         'Descending']}

name_for_cid[7275] = 'MRDiffusionModelSpecificMethods'
cid_concepts[7275] = \
{'DCM': ['VoxelwiseSelectionOfBValue']}

name_for_cid[12262] = 'CardiacUltrasoundPulmonaryValve'
cid_concepts[12262] = \
{'DCM': ['Tau', 'IndicatorMeanTransitTime', 'RRInterval'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'AWaveAmplitude',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'LeafletThickness',
        'HRCorrectedMeanVelocityOfCircumferentialFiberShortening',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'LeafletSeparation',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'PercentThickening',
        'ACInterval',
        'BCSlope',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'TimeFromQWaveToPulmonicValveCloses',
        'CEDistance',
        'PulmonaryArteryPressureUsingAccelTime',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'MeanVelocityOfCircumferentialFiberShorteningMeanVcfv',
        'VelocityOfFlowPropagation',
        'FractionalShortening',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'DEExcursion',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Area',
         'CardiacOutput',
         'CardiacEjectionFraction',
         'CardiovascularOrificeDiameter',
         'Length',
         'StrokeIndex',
         'Circumference',
         'BloodPressure',
         'CardiacIndex',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'StrokeVolume',
         'CardiovascularOrificeArea']}

name_for_cid[9504] = 'RTGeometricInformation'
cid_concepts[9504] = \
{'DCM': ['PatientSetupPoint',
         'ReferenceAcquisitionPoint',
         'MoveableLaserPatientSetupPoint',
         'RoomLaserPatientSetupPoint',
         'IsocentricTreatmentLocationPoint']}

name_for_cid[4273] = 'RetinalSegmentationSurfaces'
cid_concepts[4273] = \
{'DCM': ['PosteriorSurfaceOfTheRPE',
         'SurfaceOfTheCenterOfTheRPE',
         'SurfaceOfTheChoroidScleraInterface',
         'OuterSurfaceOfRNFL',
         'OuterSurfaceOfTheBM',
         'OuterSurfaceOfTheCC',
         'OuterSurfaceOfIPL',
         'OuterSurfaceOfGCL',
         'OuterSurfaceOfOPL',
         'OuterSurfaceOfINL',
         'SurfaceBetweenInnerAndOuterSegmentsOfThePhotoreceptors',
         'OuterSurfaceOfHFL',
         'AnteriorSurfaceOfTheRPE',
         'SurfaceOfTheInterdigitatingZoneBetweenRetinaAndRPE'],
 'SCT': ['ELMExternalLimitingMembrane', 'ILMInternalLimitingMembrane']}

name_for_cid[643] = 'Toxins'
cid_concepts[643] = \
{'PUBCHEM_CID': ['_6Hydroxydopamine'],
 'SCT': ['Lysophosphatidylcholine', 'Lipopolysaccharide'],
 'UMLS': ['EthidiumBromide']}

name_for_cid[3735] = 'AcuteCoronarySyndromeTimePeriod'
cid_concepts[3735] = \
{}

name_for_cid[10] = 'InterventionalDrug'
cid_concepts[10] = \
{'SCT': ['Nitroglycerin',
         'HemostaticAgent',
         'AntihemophilicFactorPreparation',
         'UrokinasePreparation',
         'Verapamil',
         'Propranolol',
         'Coagulant',
         'LidocaineHydrochloride',
         'AntiHeparinAgent',
         'Diuretic',
         'GlucagonPreparation',
         'ThrombolyticAgent',
         'Ethanol',
         'Nifedipine',
         'Anticoagulant',
         'Warfarin',
         'AntiarrhythmicDrug',
         'InotropicAgent',
         'Heparin',
         'BetaAdrenergicBlockingAgent',
         'ThrombinPreparation',
         'CentrallyActingHypotensiveAgent',
         'Thromboplastin',
         'ProtamineSulfate',
         'AlphaAdrenergicBlockingAgent',
         'Epinephrine',
         'TolazolineHydrocholoride',
         'Atropine',
         'Antihistamine',
         'Quinidine',
         'Dextran',
         'InjectableFibrinolysin',
         'StreptokinasePreparation',
         'MethyleneBlue',
         'Lidocaine',
         'CardiotonicDrug',
         'Digoxin',
         'CardiacAdrenergicBlockingAgent',
         'HypotensiveAgent',
         'AstringentDrug',
         'HumanFibrinogen']}

name_for_cid[10033] = 'CTReconstructionAlgorithm'
cid_concepts[10033] = \
{'DCM': ['IterativeReconstruction', 'FilteredBackProjection']}

name_for_cid[7263] = 'DiffusionTractographyMeasurementTypes'
cid_concepts[7263] = \
{'DCM': ['ApparentDiffusionCoefficient',
         'Trace',
         'MeanDiffusivity',
         'VolumetricDiffusionDxzComponent',
         'VolumetricDiffusionDxyComponent',
         'VolumetricDiffusionDxxComponent',
         'RelativeAnisotropy',
         'RadialDiffusivity',
         'FractionalAnisotropy',
         'AxialDiffusivity',
         'MeanKurtosis',
         'ApparentKurtosisCoefficient',
         'RadialKurtosis',
         'AxialKurtosis',
         'VolumetricDiffusionDzzComponent',
         'FractionalKurtosisAnisotropy',
         'VolumetricDiffusionDyzComponent',
         'VolumetricDiffusionDyyComponent']}

name_for_cid[3670] = 'ECGProcedureTypes'
cid_concepts[3670] = \
{'SCT': ['_15LeadECG', '_12LeadECG', '_18LeadECG']}

name_for_cid[6022] = 'Side'
cid_concepts[6022] = \
{'SCT': ['LeftBreast', 'RightBreast', 'BothBreasts']}

name_for_cid[3220] = 'StressSymptoms'
cid_concepts[3220] = \
{'DCM': ['AnginalEquivalent'],
 'SCT': ['Dizziness',
         'Nausea',
         'Claudication',
         'AtypicalAngina',
         'TypicalAngina',
         'Syncope',
         'Flushing',
         'Fatigue',
         'NoncardiacChestPain',
         'UnstableAngina',
         'ChestPain',
         'ChestPainNotPresent',
         'ChestDiscomfort',
         'Dyspnea',
         'StableAngina']}

name_for_cid[301] = 'MultienergyMaterialUnits'
cid_concepts[301] = \
{'DCM': ['ModifiedHounsfieldUnit', 'EffectiveAtomicNumber'],
 'UCUM': ['mgcm3', 'HounsfieldUnit', 'ElectronDensity', 'Percent', 'mgml']}

name_for_cid[12235] = 'MitralValveAnatomicSites'
cid_concepts[12235] = \
{'SCT': ['MedialMitralAnnulus', 'LateralMitralAnnulus', 'MitralAnnulus']}

name_for_cid[244] = 'Laterality'
cid_concepts[244] = \
{'SCT': ['Bilateral', 'Right', 'Left', 'Unilateral']}

name_for_cid[614] = 'AnesthesiaInductionCodeTypefromAnesthesiaQualityInitiative(AQI)'
cid_concepts[614] = \
{'SCT': ['IntramuscularRoute', 'PerRectum', 'IntravenousRoute', 'ByInhalation']}

name_for_cid[10064] = 'PatientModelType'
cid_concepts[10064] = \
{'DCM': ['AnthropomorphicModel', 'SimpleObjectModel', 'PatientSegmentedModel']}

name_for_cid[6130] = 'SeverityDescriptor'
cid_concepts[6130] = \
{'DCM': ['HyperAcute'],
 'SCT': ['AcuteOnset', 'Moderate', 'Mild', 'Chronic', 'Subacute', 'Severe']}

name_for_cid[3762] = 'AntiarrhythmicAgents'
cid_concepts[3762] = \
{'SCT': ['ClassIIIAntiarrhythmicAgent',
         'ClassIIAntiarrhythmicAgent',
         'ClassIVAntiarrhythmicAgent',
         'ClassIAntiarrhythmicAgent']}

name_for_cid[3627] = 'MeasurementType'
cid_concepts[3627] = \
{'SCT': ['Estimated',
         'PeakToPeak',
         'Measured',
         'Mode',
         'VisualEstimation',
         'BestValue',
         'Calculated',
         'Mean',
         'PointSourceMeasurement',
         'Median']}

name_for_cid[7023] = 'RTProcessOutput'
cid_concepts[7023] = \
{'DCM': ['RTPlanningResult',
         'RTTreatmentSimulationResult',
         'DosimetricResult',
         'RTTreatmentSessionResult',
         'PatientSetupVerificationResult',
         'RTTreatmentQAResult',
         'RTTreatmentCourseSummary',
         'RTPrescriptionResult',
         'PrePlanningResult',
         'CoordinateAlignmentImageSeries',
         'DoseCalculationImageSeries']}

name_for_cid[7473] = 'GeneralAreaCalculationMethods'
cid_concepts[7473] = \
{'DCM': ['AreaOfAClosedNURBS', 'AreaOfClosedIrregularPolygon']}

name_for_cid[3430] = 'DateTimeQualifiers'
cid_concepts[3430] = \
{'DCM': ['DatetimeUnsynchronized', 'DatetimeEstimated']}

name_for_cid[9528] = 'BrachytherapyIsotope'
cid_concepts[9528] = \
{'SCT': ['_32Phosphorus',
         '_222Radon',
         '_252Californium',
         '_198Gold',
         '_90Strontium',
         '_137Cesium',
         '_125Iodine',
         '_226Radium',
         '_90Yttrium',
         '_103Palladium',
         '_169Ytterbium',
         '_131Cesium',
         '_60Cobalt',
         '_106Ruthenium',
         '_192Iridium']}

name_for_cid[7166] = 'CommonTissueSegmentationTypes'
cid_concepts[7166] = \
{'SCT': ['Ligament',
         'Nerve',
         'Artery',
         'ConnectiveTissue',
         'LymphaticSystem',
         'Capillary',
         'Meniscus',
         'BloodVessel',
         'Tissue',
         'Cartilage',
         'Organ',
         'Vein',
         'Muscle',
         'Skin',
         'Blood',
         'Bone',
         'LymphaticVessel',
         'LymphNode',
         'BodyFat',
         'Tendon']}

name_for_cid[3826] = 'HeartRhythm'
cid_concepts[3826] = \
{'SCT': ['VentricularArrhythmia', 'NormalSinusRhythm', 'AtrialArrhythmia']}

name_for_cid[4033] = 'MRProtonSpectroscopyMetabolites'
cid_concepts[4033] = \
{'DCM': ['CreatineCholineCitrateRatio',
         'GlutamateAndGlutamine',
         'CholineCreatineRatio',
         'NAcetylaspartateCreatineRatio',
         'CreatineAndCholine',
         'NAcetylaspartateCholineRatio',
         'LipidAndLactate'],
 'SCT': ['Inositol',
         'Citrate',
         'Choline',
         'Creatine',
         'Lactate',
         'Glutamine',
         'NAcetylaspartate',
         'Lipid',
         'Tuarine']}

name_for_cid[403] = 'SecurityAlertTypeCode'
cid_concepts[403] = \
{'DCM': ['AuditRecordingStopped',
         'UseOfRestrictedFunction',
         'SoftwareConfiguration',
         'HardwareConfiguration',
         'RemoteServiceOperationStarted',
         'EmergencyOverrideStopped',
         'UserSecurityAttributesChanged',
         'SecurityRolesChanged',
         'ObjectSecurityAttributesChanged',
         'AuditRecordingStarted',
         'Login',
         'AuthorizationDecision',
         'ApplicationStop',
         'AuthenticationDecision',
         'ApplicationStart',
         'LocalServiceOperationStopped',
         'LocalServiceOperationStarted',
         'RemoteServiceOperationStopped',
         'SecurityConfiguration',
         'NetworkConfiguration',
         'EmergencyOverrideStarted',
         'NodeAuthentication',
         'Detach',
         'AccessControlDecision',
         'Attach',
         'SessionStop',
         'Logout',
         'SessionStart']}

name_for_cid[12022] = 'FetalCraniumAnatomicSites'
cid_concepts[12022] = \
{'SCT': ['Cerebellum',
         'LateralVentricle',
         'CerebralHemisphere',
         'OccipitalRegionOfScalp',
         'CisternaMagna',
         'AnteriorHornLateralVentricle',
         'PosteriorHornLateralVentricle',
         'Orbit']}

name_for_cid[3719] = 'CanadianClinicalClassification'
cid_concepts[3719] = \
{'SCT': ['AnginaClassII',
         'AnginaClassI',
         'AnginaClassIV',
         'AnginaClassIII',
         'ChestPainNotPresent']}

name_for_cid[4208] = 'MydriaticAgent'
cid_concepts[4208] = \
{'SCT': ['Tropicamide',
         'Homatropine',
         'Cyclopentolate',
         'Atropine',
         'Phenylephrine']}

name_for_cid[3467] = 'EjectionFraction'
cid_concepts[3467] = \
{'DCM': ['LeftAtrialEjectionFractionByAngiography'],
 'LN': ['RightVentricularEjectionFractionByAngiography',
        'LeftVentricularEjectionFractionByAngiography']}

name_for_cid[638] = 'ExogenousSubstance'
cid_concepts[638] = \
{'DCM': ['HumanAlphaSynucleinPreformedFibrils',
         'MouseAlphaSynucleinPreformedFibrils',
         'HumanTauPreformedFibrils',
         'MouseTauPreformedFibrils'],
 'NCIt': ['BronchioloalveolarAdenocarcinoma'],
 'PUBCHEM_CID': ['_6Hydroxydopamine'],
 'SCT': ['NonSmallCellCarcinoma',
         'SquamousCellCarcinoma',
         'AmelanoticMelanoma',
         'SpindleCellCarcinoma',
         'Lipopolysaccharide',
         'Adenocarcinoma',
         'Carcinosarcoma',
         'TumorNecrosisFactorAlpha',
         'MixedSmallCellCarcinoma',
         'LargeCellCarcinoma',
         'VascularEndothelialGrowthFactor',
         'Melanoma',
         'InterferonGamma',
         'RenalCellCarcinoma',
         'Glioblastoma',
         'Carcinoma',
         'Lysophosphatidylcholine',
         'Leukemia',
         'InfiltratingDuctalCarcinomaOfBreast',
         'Mesothelioma',
         'Cystadenocarcinoma',
         'TheilerMurineEncephalomyelitisVirus',
         'Astrocytoma',
         'AdenoAssociatedVirusGroup',
         'Osteosarcoma',
         'Sarcoma',
         'SmallCellCarcinoma'],
 'UMLS': ['EthidiumBromide']}

name_for_cid[12219] = 'PulmonaryVeinModifiers'
cid_concepts[12219] = \
{'SCT': ['RightLowerSegment',
         'RightUpperSegment',
         'LeftLowerSegment',
         'LeftUpperSegment']}

name_for_cid[3488] = 'Min/Max/Mean'
cid_concepts[3488] = \
{'SCT': ['Minimum', 'Maximum', 'Mean']}

name_for_cid[84] = 'PETUnits'
cid_concepts[84] = \
{'UCUM': ['MicromoleMinuteMilliliter',
          'Centimeter2Milliliter',
          'Centimeter2',
          'MilliliterGram',
          'StandardizedUptakeValueBodySurfaceArea',
          'StandardizedUptakeValueLeanBodyMassJanma',
          'ProportionalToCounts',
          'Counts',
          'Percent',
          'MilliliterMinuteGram',
          'StandardizedUptakeValueBodyWeight',
          'MicromoleMilliliter',
          'ProportionalToCountsPerSecond',
          'MilligramsMinuteMilliliter',
          'StandardizedUptakeValueIdealBodyWeight',
          'Centimeter',
          'BecquerelsMilliliter',
          'CountsPerSecond',
          'StandardizedUptakeValueLeanBodyMassJames',
          'StandardizedUptakeValueLeanBodyMassJames128Multiplier']}

name_for_cid[6059] = 'BreastImplantTypes'
cid_concepts[6059] = \
{'DCM': ['SalineImplant',
         'PolyurethaneImplant',
         'PercutaneousSiliconeInjection',
         'CombinationImplant',
         'PrePectoralImplant',
         'RetroPectoralImplant'],
 'SCT': ['SiliconeGelImplant', 'BreastImplantTypeNotSpecified']}

name_for_cid[8102] = 'ContainerComponentTypes'
cid_concepts[8102] = \
{'SCT': ['MicroscopeSlideCoverSlip',
         'SpecimenContainer',
         'TissueMicroarrayCassette',
         'ElectronMicroscopyGrid',
         'TissueCassette',
         'MicroscopeSlideMountingMedia',
         'SpecimenContainerLid',
         'MicroscopeSlide',
         'SpecimenVial',
         'SpecimenWell']}

name_for_cid[7062] = 'PurposeofReferencetoPredecessor3DModel'
cid_concepts[7062] = \
{'DCM': ['EditedModel', 'ComponentModel']}

name_for_cid[3471] = 'EstimatedVolumes'
cid_concepts[3471] = \
{'DCM': ['VolumeEstimatedFromTwoNonCoplanar2DRegions',
         'VolumeEstimatedFromSingle2DRegion']}

name_for_cid[92] = 'JointPositionDuringAcquisition'
cid_concepts[92] = \
{'DCM': ['NeutralMusculoskeletalPosition'],
 'SCT': ['Torsion',
         'ExternalRotation',
         'Pronation',
         'Flexion',
         'InternalRotation',
         'Abduction',
         'Extension',
         'Adduction',
         'Supination']}

name_for_cid[8114] = 'SpecimenFixatives'
cid_concepts[8114] = \
{'SCT': ['Chloroform',
         'OsmiumTetroxide',
         'CarnoyFluid',
         'HellyFluid',
         'MichelMedium',
         'BouinFluid',
         'ChromiumTrioxide',
         'PotassiumDichromate',
         'NeutralBufferedFormalin',
         'MercuricChloride',
         'FormolSublimate',
         'AceticAcid',
         'PicricAcid',
         'Formalin',
         'Methanol',
         'Formaldehyde',
         'Paraformaldehyde',
         'Ethanol',
         'ZenkerFluid']}

name_for_cid[3758] = 'CardiovascularFamilyHistory'
cid_concepts[3758] = \
{'SCT': ['NoFamilyHistoryOfCardiovascularDisease',
         'FamilyHistoryOfDiabetesMellitus',
         'FamilyHistoryOfCoronaryArteriosclerosis',
         'NoFamilyHistoryOfDiabetes',
         'FamilyHistoryUnknown',
         'FamilyHistoryOfCardiovascularDisease',
         'FamilyHistoryOfMyocardialInfarction']}

name_for_cid[3426] = 'ConsumablesParameters'
cid_concepts[3426] = \
{'DCM': ['LotIdentifier', 'UnitSerialIdentifier', 'DescriptionOfMaterial']}

name_for_cid[7465] = 'MeasurementsDerivedFromMultipleROIMeasurements'
cid_concepts[7465] = \
{'DCM': ['PercentileRankingOfMeasurement',
         'ZScoreOfMeasurement',
         'StandardDeviationOfPopulation',
         '_2SigmaDeviationOfPopulation'],
 'SCT': ['_1SigmaLowerValueOfPopulation',
         '_2SigmaLowerValueOfPopulation',
         '_5thPercentileValueOfPopulation',
         '_10thPercentileValueOfPopulation',
         '_1SigmaUpperValueOfPopulation',
         '_2SigmaUpperValueOfPopulation',
         'MeanValueOfPopulation',
         '_95thPercentileValueOfPopulation',
         '_90thPercentileValueOfPopulation',
         'MedianValueOfPopulation']}

name_for_cid[7035] = 'ActionableFindingClassification'
cid_concepts[7035] = \
{'RADLEX': ['ACRCategory2ActionableFinding',
            'ACRCategory1ActionableFinding',
            'ACRCategory3ActionableFinding']}

name_for_cid[12258] = 'CardiacUltrasoundRightVentricle'
cid_concepts[12258] = \
{'DCM': ['Tau', 'IndicatorMeanTransitTime', 'RRInterval', 'WallMass'],
 'LN': ['DESlope',
        'PeakEWavePeakAWaveByUS',
        'EWavePeakVelocity',
        'MeanGradient',
        'VelocityTimeIntegral',
        'StenosisPeakGradient',
        'JetArea',
        'EnvelopeDuration',
        'VolumeFlow',
        'TimeToPeakByUS',
        'FlowRadius',
        'HeartRateCorrectedEjectionTime',
        'HeartRate',
        'PreEjectionPeriodEjectionTimeRatio',
        'FlowArea',
        'EndDiastolicVelocity',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'EndocardialArea',
        'ROIInternalDimensionByUS',
        'PeakGradient',
        'EjectionTime',
        'PeakInstantaneousFlowRate',
        'DecelerationTime',
        'AccelerationTime',
        'EFSlope',
        'PeakReversalVelocityDuringAtrialContraction',
        'ACInterval',
        'PreEjectionPeriod',
        'AWavePeakVelocity',
        'AWaveDuration',
        'PeakSystolicVelocity',
        'StenosisPeakVelocity',
        'DpDtByUS',
        'TimeAveragedPeakVelocity',
        'ROIThicknessByUS',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'IsovolumicRelaxationTime',
        'VelocityOfFlowPropagation',
        'AliasVelocity',
        'PressureHalfTime',
        'EVelocityToAnnulusEVelocityRatio',
        'AccelerationSlope',
        'DecelerationSlope',
        'SystolicToDiastolicVelocityRatio',
        'EpicardialArea',
        'MyocardialPerformanceIndexTei',
        'TimeAveragedMeanVelocity'],
 'SCT': ['MajorAxis',
         'Volume',
         'Area',
         'CardiovascularOrificeDiameter',
         'Length',
         'Circumference',
         'BloodPressure',
         'RegurgitantFraction',
         'Radius',
         'MinorAxis',
         'Diameter',
         'CardiovascularOrificeArea']}

name_for_cid[6018] = 'ClockfaceLocationorRegion'
cid_concepts[6018] = \
{'SCT': ['_7OClockPosition',
         '_11OClockPosition',
         '_12OClockPosition',
         'ChestWall',
         '_5OClockPosition',
         '_3OClockPosition',
         '_6OClockPosition',
         '_4OClockPosition',
         'AxillaRegion',
         '_1OClockPosition',
         '_8OClockPosition',
         '_2OClockPosition',
         'CentralRegionOfBreast',
         'AxillaryTailRegion',
         'SubareolarRegion',
         '_9OClockPosition',
         '_10OClockPosition']}

name_for_cid[4025] = 'PrimaryAnatomicStructureforIntraoralRadiography(SupernumeraryDentitionDesignationofTeeth)'
cid_concepts[4025] = \
{'SCT': ['SupernumeraryPermanentMandibularLeftThirdMolarTooth',
         'SupernumeraryDeciduousMaxillaryLeftFirstMolarTooth',
         'SupernumeraryPermanentMaxillaryLeftCentralIncisorTooth',
         'SupernumeraryPermanentMaxillaryLeftSecondPremolarTooth',
         'SupernumeraryDeciduousMandibularRightCanineTooth',
         'SupernumeraryPermanentMandibularLeftLateralIncisorTooth',
         'SupernumeraryPermanentMandibularRightLateralIncisorTooth',
         'SupernumeraryPermanentMaxillaryRightFirstPremolarTooth',
         'SupernumeraryPermanentMaxillaryRightSecondMolarTooth',
         'SupernumeraryDeciduousMaxillaryRightFirstMolarTooth',
         'SupernumeraryPermanentMandibularLeftSecondPremolarTooth',
         'SupernumeraryDeciduousMaxillaryLeftSecondMolarTooth',
         'SupernumeraryPermanentMaxillaryLeftThirdMolarTooth',
         'SupernumeraryPermanentMandibularRightFirstMolarTooth',
         'SupernumeraryDeciduousMaxillaryLeftCentralIncisorTooth',
         'SupernumeraryDeciduousMandibularRightLateralIncisorTooth',
         'SupernumeraryPermanentMaxillaryRightCentralIncisorTooth',
         'SupernumeraryPermanentMaxillaryLeftFirstPremolarTooth',
         'SupernumeraryPermanentMandibularRightCentralIncisorTooth',
         'SupernumeraryPermanentMaxillaryRightCanineTooth',
         'SupernumeraryDeciduousMandibularLeftFirstMolarTooth',
         'SupernumeraryDeciduousMaxillaryRightCentralIncisorTooth',
         'SupernumeraryPermanentMandibularLeftFirstMolarTooth',
         'SupernumeraryPermanentMandibularRightSecondPremolarTooth',
         'SupernumeraryPermanentMaxillaryLeftSecondMolarTooth',
         'SupernumeraryDeciduousMandibularRightSecondMolarTooth',
         'SupernumeraryPermanentMaxillaryLeftCanineTooth',
         'SupernumeraryDeciduousMandibularLeftCanineTooth',
         'SupernumeraryPermanentMandibularLeftCentralIncisorTooth',
         'SupernumeraryPermanentMaxillaryRightFirstMolarTooth',
         'SupernumeraryDeciduousMandibularLeftSecondMolarTooth',
         'SupernumeraryDeciduousMaxillaryRightSecondMolarTooth',
         'SupernumeraryDeciduousMandibularLeftLateralIncisorTooth',
         'SupernumeraryPermanentMandibularRightFirstPremolarTooth',
         'SupernumeraryPermanentMandibularLeftCanineTooth',
         'SupernumeraryPermanentMandibularLeftSecondMolarTooth',
         'SupernumeraryDeciduousMaxillaryLeftCanineTooth',
         'SupernumeraryPermanentMaxillaryLeftFirstMolarTooth',
         'SupernumeraryDeciduousMaxillaryRightLateralIncisorTooth',
         'SupernumeraryDeciduousMandibularRightFirstMolarTooth',
         'SupernumeraryDeciduousMandibularLeftCentralIncisorTooth',
         'SupernumeraryPermanentMandibularRightThirdMolarTooth',
         'SupernumeraryPermanentMaxillaryLeftLateralIncisorTooth',
         'SupernumeraryPermanentMaxillaryRightLateralIncisorTooth',
         'SupernumeraryPermanentMandibularRightCanineTooth',
         'SupernumeraryPermanentMaxillaryRightSecondPremolarTooth',
         'SupernumeraryDeciduousMaxillaryRightCanineTooth',
         'SupernumeraryPermanentMandibularLeftFirstPremolarTooth',
         'SupernumeraryPermanentMandibularRightSecondMolarTooth',
         'SupernumeraryDeciduousMandibularRightCentralIncisorTooth',
         'SupernumeraryDeciduousMaxillaryLeftLateralIncisorTooth',
         'SupernumeraryPermanentMaxillaryRightThirdMolarTooth']}

name_for_cid[9302] = 'MediaImportPPSDiscontinuationReasons'
cid_concepts[9302] = \
{'DCM': ['IncorrectWorklistEntrySelected',
         'MediaFailure',
         'DiscontinuedForUnspecifiedReason',
         'ObjectSetIncomplete',
         'EquipmentFailure',
         'ObjectTypesNotSupported',
         'ObjectsIncorrectlyFormatted',
         'DuplicateOrder']}

name_for_cid[100] = 'QuantitativeDiagnosticImagingProcedures'
cid_concepts[100] = \
{'DCM': ['MultiparametricMRIOfProstate',
         'MultiparametricMRIOfWholeBody',
         'MultiparametricMRI'],
 'LN': ['NMUnspecifiedBodyRegion',
        'PETWholeBody',
        'SPECTBrain',
        'MRIUnspecifiedBodyRegion',
        'PETUnspecifiedBodyRegion',
        'CTUnspecifiedBodyRegion',
        'CTPerfusionHeadWithContrastIV'],
 'RADLEX': ['NMHeadPerfusionBrainPETCTAV45'],
 'SCT': ['DynamicMagneticResonanceImagingOfKnee',
         'DynamicMagneticResonanceImagingOfPelvis',
         'PETCTFDGImagingOfWholeBody',
         'PETCTMETImagingOfWholeBody',
         'MagneticResonanceImagingGuidance',
         'ImagingProcedure']}

name_for_cid[12034] = 'UltrasoundTransducerBeamSteering'
cid_concepts[12034] = \
{'DCM': ['PhasedBeamSteering', 'MechanicalBeamSteering', 'FixedBeamDirection']}

name_for_cid[10009] = 'Measured/Calculated'
cid_concepts[10009] = \
{'SCT': ['Estimated', 'Measured', 'Calculated']}

name_for_cid[3778] = 'Stages'
cid_concepts[3778] = \
{'SCT': ['Stage1', 'Stage5', 'Stage0', 'Stage4', 'Stage2', 'Stage3']}

name_for_cid[6080] = 'GynecologicalHormones'
cid_concepts[6080] = \
{'DCM': ['UnspecifiedGynecologicalHormone'],
 'SCT': ['Anastrozole',
         'Tamoxifen',
         'Raloxifene',
         'Contraceptives',
         'ProgesteroneProduct',
         'EstrogenProduct']}

name_for_cid[4052] = 'PhantomDevices'
cid_concepts[4052] = \
{'DCM': ['IECHeadDosimetryPhantom',
         'IECBodyDosimetryPhantom',
         'Phantom',
         'NEMAXR212000Phantom',
         'ACRAccreditationPhantomCT',
         'ACRAccreditationPhantomMR',
         'ACRAccreditationPhantomMammography',
         'ACRAccreditationPhantomStereotacticBreastBiopsy',
         'ACRAccreditationPhantomECT',
         'ACRAccreditationPhantomPET',
         'ACRAccreditationPhantomECTPET',
         'ACRAccreditationPhantomPETFaceplate']}

name_for_cid[12106] = 'IntracranialCerebralVessels(Unilateral)'
cid_concepts[12106] = \
{'SCT': ['BasilarArtery']}

name_for_cid[7042] = 'CMDCTECCCalciumScoringPatientSizeCategories'
cid_concepts[7042] = \
{'DCM': ['SmallLesserThan32Point0cmLateralThickness',
         'Medium32Point0To38Point0cmLateralThickness',
         'LargeGreaterThan38Point0cmLateralThickness']}

name_for_cid[3451] = 'CalibrationObjects'
cid_concepts[3451] = \
{'DCM': ['Sphere'], 'SCT': ['Catheter', 'MeasuringRuler']}

name_for_cid[6203] = 'ColonNonlesionObjectType'
cid_concepts[6203] = \
{'DCM': ['ChestTube', 'KidneyStent', 'PancreaticStent', 'Coin', 'FiducialMark'],
 'SCT': ['JejunostomyTube',
         'Suture',
         'IleostomyBag',
         'ColostomyBag',
         'VenaCavaFilter',
         'UrostomyBag',
         'UrostomySet',
         'Needle',
         'Prosthesis',
         'Jewelry',
         'RectalTube',
         'Catheter',
         'Staple',
         'UrethralCatheter',
         'ScarTissue',
         'UreteralStent',
         'Bullet',
         'Pin',
         'ColostomySet',
         'IleostomySet']}

name_for_cid[3001] = 'ECGLeads'
cid_concepts[3001] = \
{'MDC': ['FrankLeadC',
         'FrankLeadA',
         'FrankLeadM',
         'FrankLeadF',
         'ChestManubriumLeadPerV5Placement',
         'LeftArmLead',
         'CanineSixthLeftIntercostalSpaceAtCostochondralJunction',
         'RightArmLead',
         'LeftLegLead',
         'CanineOverDorsalSpinousProcessOf7thThoracicVertebra',
         'FrankLeadI',
         'FrankLeadE',
         'DerivedLeadV5',
         'DerivedLeadV6',
         'FrankLeadH',
         'DerivedLeadI',
         'DerivedLeadII',
         'DerivedLeadV1',
         'DerivedLeadV2',
         'DerivedLeadV3',
         'DerivedLeadV4',
         'EASILeadAS',
         'EASILeadAI',
         'EASIUpperSternumLead',
         'AuxiliaryBipolarLead4',
         'EASILeadES',
         'RightLegLead',
         'CanineFifthRightIntercostalSpaceNearEdgeOfSternum',
         'CanineSixthLeftIntercostalSpaceNearEdgeOfSternum',
         'LeadI',
         'UnspecifiedLead',
         'LeadV1',
         'LeadII',
         'LeadV3',
         'LeadV2',
         'LeadV5',
         'LeadV4',
         'LeadV7',
         'LeadV6',
         'DerivedLeadAvf',
         'AvrAugmentedVoltageRight',
         'ChestManubriumLeadPerV6Placement',
         'AvlAugmentedVoltageLeft',
         'DerivedLeadIII',
         'AvfAugmentedVoltageFoot',
         'DerivedLeadAvr',
         'Avr',
         'DerivedLeadAvl',
         'LeadV8',
         'LeadV9',
         'LeadV8R',
         'LeadV9R',
         'LeadIII',
         'AuxiliaryBipolarLead3',
         'LeadCR5',
         'ModifiedLimbLead',
         'AuxiliaryBipolarLead1',
         'AuxiliaryBipolarLead2',
         'ChestManubriumLeadPerV7Placement',
         'DefibrillatorLeadAnteriorLateral',
         'LeadCH5',
         'ExternalPacingLeadAnteriorPosterior',
         'AuxiliaryUnipolarLead1',
         'NegativeRightInfraclavicularFossa',
         'AuxiliaryUnipolarLead2',
         'NegativeLowRightScapulaLead',
         'AuxiliaryUnipolarLead3',
         'AuxiliaryUnipolarLead4',
         'LeadDNehbDorsal',
         'LeadANehbAnterior',
         'LeadJNehbInferior',
         'ChestLead',
         'PrecordialLead',
         'LeadVRNonaugmentedVoltageVectorOfRA',
         'LeadVLNonaugmentedVoltageVectorOfLA',
         'ChestManubriumLeadPerV2Placement',
         'LeadV7R',
         'ChestManubriumLeadPerV3Placement',
         'LeadX',
         'ChestManubriumLeadPerV4Placement',
         'LeadY',
         'LeadZ',
         'ChestLeadPerV5AndV5RPlacement',
         'ChestLeadPerV6AndV6RPlacement',
         'ChestLeadPerV7AndV8RPlacement',
         'ChestManubriumLead',
         'ChestManubriumLeadPerV1Placement',
         'ModifiedChestLeadPerV4Placement',
         'ChestLeadPerV2AndV2RPlacement',
         'ModifiedChestLeadPerV5Placement',
         'ChestLeadPerV3AndV3RPlacement',
         'ModifiedChestLeadPerV6Placement',
         'ChestLead',
         'ChestLeadPerV4AndV4RPlacement',
         'LeadV2R',
         'ChestLeadPerV1AndV1RPlacement',
         'LeadV3R',
         'LeadV4R',
         'LeadV5R',
         'LeadV6R',
         'LeadVFNonaugmentedVoltageVectorOfLL',
         'ModifiedChestLead',
         'ModifiedChestLeadPerV1Placement',
         'ModifiedChestLeadPerV2Placement',
         'ModifiedChestLeadPerV3Placement']}

name_for_cid[12014] = 'OBFetalBodyWeightEquationsandTables'
cid_concepts[12014] = \
{'LN': ['EFWByACHCHadlock1984',
        'EFWByACBPDFLHadlock1985',
        'EFWByBPDTTDHansmann1986',
        'EFWByACBPDFLHCHadlock1985',
        'EFWByACFLHCHadlock1985',
        'EFWByACBPDHadlock1984',
        'EFWByACCampbell1975',
        'EFWByACAndBPDShepard1982',
        'EFWByACFLHadlock1984',
        'EFW3ByShinozuka1996',
        'EFWByBPDAPADTADFLTokyo1987',
        'EFWByACBPDFLHadlock1984',
        'EFWByACFLHadlock1985',
        'EFWByBPDFTAFLOsaka1990',
        'EFW1ByShinozuka1996',
        'EFW2ByShinozuka1996']}

name_for_cid[9] = 'ImageGuidedTherapeuticProcedures'
cid_concepts[9] = \
{'SCT': ['Thoracentesis',
         'StentPlacement',
         'CatheterManipulation',
         'OcclusionOfCatheter',
         'UltrasonicGuidanceForAmniocentesis',
         'Amniocentesis',
         'AtherectomyByRotaryCutter',
         'PercutaneousRetrievalOfIntravascularForeignBody',
         'AmnioinfusionInjectionOfAmnion',
         'Vasoconstriction',
         'Fetocide',
         'IrrigationFollowingInsertionOfCatheter',
         'PercutaneousInsertionOfIntravascularFilter',
         'IntraArterialInfusionOfThrombolyticAgent',
         'RemovalOfCatheter',
         'AtherectomyByLaser',
         'CatheterReplacement',
         'Catheterization',
         'IntrauterineCordocentesis',
         'TranscatheterTherapyForEmbolization',
         'Biopsy',
         'Atherectomy',
         'SelectiveEmbolizationOfArtery',
         'RemovalOfForeignBody',
         'ProstaglandinInjection',
         'PercutaneousTransluminalBalloonAngioplasty',
         'FailedAttemptedProcedure',
         'IntrauterineTransfusion',
         'TranscatheterDeploymentOfDetachableBalloon',
         'BreechVersionObstetricalVersion',
         'Vasodilatation']}

name_for_cid[4005] = 'PartialViewSectionforMammography'
cid_concepts[4005] = \
{'SCT': ['Central',
         'Posterior',
         'Lateral',
         'Inferior',
         'Anterior',
         'Superior',
         'Medial']}

name_for_cid[6038] = 'MammographyQualitativeTemporalDifferenceType'
cid_concepts[6038] = \
{'SCT': ['DifferenceInMargin', 'DifferenceInSymmetry', 'DifferenceInShape']}

name_for_cid[7500] = 'NeighbourhoodGreyToneDifferenceBasedFeatures'
cid_concepts[7500] = \
{}

name_for_cid[3810] = 'VascularMorphology'
cid_concepts[3810] = \
{'SCT': ['Embolism',
         'Pseudoaneurysm',
         'Angioma',
         'ArterialDissection',
         'Stenosis',
         'Plaque',
         'Tortuosity',
         'Dilatation',
         'VascularCoiling',
         'VascularSclerosis',
         'ArteriovenousFistula',
         'Diverticulum',
         'Thrombus',
         'FibromuscularDysplasia',
         'Aneurysm',
         'Occlusion',
         'Stent']}

name_for_cid[7150] = 'SegmentationPropertyCategories'
cid_concepts[7150] = \
{'SCT': ['SpatialAndRelationalConcept',
         'AnatomicalStructure',
         'Function',
         'Substance',
         'MorphologicallyAbnormalStructure',
         'Tissue',
         'PhysicalObject',
         'BodySubstance']}

name_for_cid[3113] = 'TypesofPerfusionDefects'
cid_concepts[3113] = \
{'DCM': ['FalsePositiveDefectFinding'],
 'SCT': ['PartiallyReversibleMyocardialPerfusionDefect',
         'FixedMyocardialPerfusionDefect',
         'ReversibleMyocardialPerfusionDefect']}

name_for_cid[3406] = 'NoncoronaryTranscatheterInterventions'
cid_concepts[3406] = \
{'DCM': ['VascularIntervention',
         'MyocardialBiopsy',
         'AorticIntervention',
         'SeptalDefectIntervention',
         'ValvularIntervention']}

name_for_cid[7445] = 'DeviceParticipatingRoles'
cid_concepts[7445] = \
{'DCM': ['Recording', 'XRayReadingDevice', 'IrradiatingDevice']}

name_for_cid[7015] = 'ExportDelayReasons'
cid_concepts[7015] = \
{'DCM': ['DelayExportUntilPatientDies',
         'DelayExportUntilPatientIsDischarged',
         'DelayExportUntilExpertReviewIsAvailable',
         'DelayExportUntilClinicalInformationIsAvailable',
         'DelayExportUntilFinalReportIsAvailable',
         'DelayExportUntilHistopathologyIsAvailable',
         'DelayExportUntilConfirmationOfDiagnosisIsAvailable',
         'DelayExportUntilOtherLaboratoryResultsAreAvailable']}

name_for_cid[12002] = 'UltrasoundProtocolStageTypes'
cid_concepts[12002] = \
{'SCT': ['AfterProcedure',
         'ValsalvaManeuver',
         'CardiacStressState',
         'Hyperventilation',
         'BeforeProcedure',
         'PeakCardiacStressState',
         'CardiacPacing',
         'CardiacStressRecoveryState',
         'DuringProcedure',
         'HandGrip',
         'RestingState']}

name_for_cid[3739] = 'CathProcedureType'
cid_concepts[3739] = \
{'DCM': ['PercutaneousCoronaryIntervention'],
 'SCT': ['CatheterizationOfBothLeftAndRightHeartWithGraft',
         'CatheterizationOfLeftHeart',
         'CatheterizationOfBothLeftAndRightHeartWithoutGraft',
         'CatheterizationOfRightHeart']}

name_for_cid[4013] = 'AnatomicRegionforMammography'
cid_concepts[4013] = \
{'SCT': ['Breast']}

name_for_cid[3555] = 'Pacing'
cid_concepts[3555] = \
{'SCT': ['AtrialPacing',
         'TemporaryTranscutaneousPacing',
         'PacingWithMagnet',
         'Pacing',
         'AVSequentialPacing',
         'VentricularPacing']}

name_for_cid[9508] = 'NonspecificVolumes'
cid_concepts[9508] = \
{'DCM': ['UnclassifiedVolume', 'UnclassifiedCombination']}

name_for_cid[3806] = 'StenosisShape'
cid_concepts[3806] = \
{'SCT': ['Eccentric', 'Concentric']}

name_for_cid[7003] = 'DiagnosticImagingReportPurposesofReference'
cid_concepts[7003] = \
{'DCM': ['BestIllustrationOfFinding',
         'Baseline',
         'SourceOfMeasurement',
         'IllustrationOfROI']}

name_for_cid[12281] = 'CardiacUltrasoundTargetSiteModifiers'
cid_concepts[12281] = \
{'NCIt': ['Middle'],
 'SCT': ['Distal',
         'Proximal',
         'Lateral',
         'Posterior',
         'Inferior',
         'Anterior',
         'Medial']}

name_for_cid[7453] = 'PerformingRoles'
cid_concepts[7453] = \
{'DCM': ['Standby',
         'IrradiationAdministering',
         'IrradiationAuthorizing',
         'ImageQualityController',
         'Designator',
         'ResultsQualityController',
         'Assisting',
         'Recording',
         'Verifying',
         'Requesting',
         'EligibilityReader',
         'Performing'],
 'NCIt': ['Reader', 'Adjudicator', 'Reviewer'],
 'SCT': ['CirculatingNurse'],
 'UMLS': ['Referring']}

name_for_cid[3410] = 'NumericParametersofDrugs/Contrast'
cid_concepts[3410] = \
{'DCM': ['UndilutedDoseAdministered',
         'Concentration',
         'VolumeAdministered',
         'MassAdministered',
         'VolumeUnadministeredOrDiscarded',
         'QuantityAdministered',
         'RateOfAdministration',
         'DurationOfAdministration']}

name_for_cid[6096] = 'PregnancyStatus'
cid_concepts[6096] = \
{'SCT': ['Unknown',
         'PatientCurrentlyPregnant',
         'NotPregnant',
         'PossiblePregnancy']}

name_for_cid[4101] = 'TracerKineticModels'
cid_concepts[4101] = \
{'DCM': ['StandardToftsModel',
         'ExtendedToftsModel',
         'FirstPassLeakageProfileFPLPModel',
         'ShutterSpeedModelSSM',
         'GammaCapillaryTransitTimeGCTTModel',
         'AdiabaticTissueHomogeneityATHModel',
         'TwoCompartmentExchange2CXModel']}

name_for_cid[8122] = 'MicroscopyIlluminatorandSensorColor'
cid_concepts[8122] = \
{'SCT': ['Infrared', 'Ultraviolet', 'Red', 'FullSpectrum', 'Blue', 'Green']}

name_for_cid[10068] = 'EstimateMethodTypes'
cid_concepts[10068] = \
{'DCM': ['TabularDataAlgorithm', 'AnalyticalAlgorithm', 'EmpiricalAlgorithm'],
 'MSH': ['MonteCarloMethod']}

name_for_cid[12110] = 'LowerExtremityVeins'
cid_concepts[12110] = \
{'SCT': ['ProfundaFemorisVein',
         'CommonFemoralVein',
         'PoplitealVein',
         'LateralCalfPerforator',
         'GiacominiVein',
         'InternalIliacVein',
         'BoydPerforatingVein',
         'GastrocnemiusVein',
         'PosteriorArchVein',
         'SaphenousVein',
         'SolealVein',
         'SuperficialFemoralVein',
         'AnteriorTibialVein',
         'PosteriorTibialVein',
         'SaphenopoplitealJunction',
         'GreatSaphenousVeinOfCalf',
         'SaphenofemoralJunction',
         'HunterianPerforatingVein',
         'CockettPerforatingVein',
         'PeronealVein',
         'ThighPerforator',
         'CommonIliacVein',
         'LesserSaphenousVein',
         'GreatSaphenousVein',
         'ExternalIliacVein',
         'GreatSaphenousVeinOfThigh']}

name_for_cid[12239] = 'CardiacOutputProperties'
cid_concepts[12239] = \
{'SCT': ['CardiacOutput', 'StrokeIndex', 'CardiacIndex', 'StrokeVolume']}

name_for_cid[7111] = 'Fiducials'
cid_concepts[7111] = \
{'DCM': ['MRMarker',
         'InfraredReflectorMarker',
         'CoilMarker',
         'CylinderMarker',
         'TransponderMarker',
         'WireMarker',
         'InterHemisphericPlane',
         'VisibleReflectorMarker',
         'LeftHemisphereMostInferior',
         'Sphere',
         'LeftHemisphereMostSuperior',
         'LeftHemisphereMostPosterior',
         'LeftHemisphereMostAnterior',
         'RightHemisphereMostInferior',
         'RightHemisphereMostSuperior',
         'RightHemisphereMostPosterior',
         'RightHemisphereMostAnterior'],
 'SCT': ['VascularCalcification',
         'RadioactiveImplant',
         'PosteriorCommissure',
         'ContrastMedia',
         'Ostium',
         'Catheter',
         'RenalArtery',
         'AnteriorCommissure',
         'VesselOrigin',
         'BBShotLeadPellet',
         'AorticBifurcation',
         'CommonIliacBifurcation',
         'Vein',
         'Clip',
         'Stent',
         'FibrousPlaque',
         'CollateralBranchOfVessel']}

name_for_cid[3502] = 'HemodynamicResistanceUnits'
cid_concepts[3502] = \
{'UCUM': ['WoodU', 'PRU', 'DynScm5']}

name_for_cid[618] = 'AirwayManagementMethodCodeTypefromAnesthesiaQualityInitiative(AQI)'
cid_concepts[618] = \
{'DCM': ['NasalCannula'],
 'SCT': ['ViaTracheostomy',
         'AnestheticFaceMask',
         'ArtificialRespiration',
         'ControlledVentilation',
         'OxygenTherapy',
         'IntubationOfRespiratoryTract',
         'LaryngealMaskAirwayLMA']}

name_for_cid[6002] = 'ChangeSinceLastMammogramorPriorSurgery'
cid_concepts[6002] = \
{'SCT': ['RemovalOfImplantSincePreviousMammogram',
         'ImplantRevisedSincePreviousMammogram',
         'NoSignificantChangesInTheFinding',
         'IncreaseInSize',
         'NewFinding',
         'IncreaseInNumberOfCalcifications',
         'DecreaseInNumberOfCalcifications',
         'LessDefined',
         'FindingPartiallyRemoved',
         'MoreDefined',
         'DecreaseInSize']}

name_for_cid[3200] = 'StressTestProcedure'
cid_concepts[3200] = \
{'SCT': ['ExerciseStressTest',
         'PharmacologicStressTest',
         'PacedStressTest',
         'PharmacologicAndExerciseStressTest']}

name_for_cid[3715] = 'LesionMargin'
cid_concepts[3715] = \
{'SCT': ['Smooth', 'MultipleIrregularities', 'Ulcerative', 'Irregular']}

name_for_cid[6147] = 'ResponseCriteria'
cid_concepts[6147] = \
{'DCM': ['RECIST1Point0', 'RECIST1Point1', 'WHO'], 'NCIt': ['RANO']}

name_for_cid[7306] = 'HumanHipImplantPlanningLandmarks'
cid_concepts[7306] = \
{'DCM': ['AnatomicalAxisOfFemur',
         'AcetabularCenterOfRotation',
         'FemurHeadCenterOfRotation']}

name_for_cid[30] = 'DICOMDevices'
cid_concepts[30] = \
{'DCM': ['BoneMineralDensitometry',
         'SlideMicroscopy',
         'OphthalmicPhotography',
         'FilmDigitizer',
         'Workstation',
         'Electrocardiography',
         'GeneralMicroscopy',
         'XRayAngiography',
         'ExternalCameraPhotography',
         'IntravascularUltrasound',
         'ComputedRadiography',
         'ComputationServer',
         'MediaCreationDevice',
         'ComputedTomography',
         'OpticalSurfaceScanner',
         'Thermography',
         'Lensometry',
         'OphthalmicTomographyEnFace',
         'HemodynamicWaveform',
         'OpticalCoherenceTomography',
         '_3DManufacturingModelingSystem',
         'UltrasoundBoneDensitometry',
         'Diaphanography',
         'PositronEmissionTomography',
         'CardiacElectrophysiology',
         'LaserScan',
         'HardCopyPrintServer',
         'PanoramicXRay',
         'OphthalmicMapping',
         'OphthalmicTomographyBScanVolumeAnalysis',
         'OphthalmicVisualField',
         'DigitalRadiography',
         'OphthalmicTomography',
         'Mammography',
         'Ultrasound',
         'IntravascularOpticalCoherenceTomography',
         'DepartmentSystemScheduler',
         'ProcedureLogging',
         'MagneticResonance',
         'IntraOralRadiography',
         'ComputerAssistedDetectionDiagnosis',
         'RTImage',
         'VisualAcuity',
         'RespiratoryWaveform',
         'Endoscopy',
         'Autorefraction',
         'RadiographicImaging',
         'Radiofluoroscopy',
         'Keratometry',
         'Archive',
         'ImageCapture',
         'AutomatedSlideStainer',
         'OphthalmicAxialMeasurements',
         'NuclearMedicine',
         'RadiationTherapyDevice',
         'BiomagneticImaging',
         'SubjectiveRefraction']}

name_for_cid[10013] = 'CTAcquisitionType'
cid_concepts[10013] = \
{'DCM': ['StationaryAcquisition',
         'ConstantAngleAcquisition',
         'FreeAcquisition',
         'SequencedAcquisition'],
 'SCT': ['SpiralAcquisition', 'ConeBeamAcquisition']}

name_for_cid[4253] = 'VisualFieldStaticPerimetryFixationStrategy'
cid_concepts[4253] = \
{'DCM': ['AutomatedOptical',
         'BlindSpotMonitoring',
         'MacularFixationTesting',
         'ObservationByExaminer'],
 'SCT': ['None']}

name_for_cid[7185] = 'AbstractMultidimensionalImageModelAxisOrientation'
cid_concepts[7185] = \
{'DCM': ['LeftToRight',
         'RightToLeft',
         'WallToSeptum',
         'SeptumToWall',
         'InferiorToAnterior',
         'AnteriorToInferior',
         'BaseToApex',
         'ApexToBase',
         'PosteriorToAnterior',
         'AnteriorToPosterior',
         'FootToHead',
         'HeadToFoot']}

name_for_cid[12307] = 'CardiacPhasesandTimePoints'
cid_concepts[12307] = \
{'DCM': ['FullCardiacCycle',
         'ARWave',
         'AtrialDiastolicFillingDWave',
         'PreEjectionPeriod',
         'ElectromechanicalDelay'],
 'SCT': ['Diastole',
         'VentricularIsovolumicContraction',
         'Diastasis',
         'EndDiastole',
         'EndSystole',
         'AtrialSystoleAWave',
         'VentricularEjectionSWave',
         'DiastolicRapidInflowEWave',
         'VentricularIsovolumicRelaxation',
         'Systole']}

name_for_cid[12242] = 'AorticValveFindingSites'
cid_concepts[12242] = \
{'SCT': ['AorticValveRing']}

name_for_cid[7490] = 'ResearchAnimalSourceRegistries'
cid_concepts[7490] = \
{'DCM': ['ILCR']}

name_for_cid[9524] = 'RadiotherapyProcedureTechniques'
cid_concepts[9524] = \
{'DCM': ['StaticBeam',
         'SlidingWindowBeam',
         'SynchronizedRoboticTreatment',
         'StepAndShootBeam',
         'MultipleFixedSources',
         'ConformalArcBeam',
         'ArcBeam',
         'TopographicBeam',
         'HelicalBeam',
         'VMAT',
         'NonSynchronizedRoboticTreatment']}

name_for_cid[3083] = 'UnitsofRadioactivity'
cid_concepts[3083] = \
{'UCUM': ['Millicurie', 'Becquerel', 'Megabecquerel']}

name_for_cid[6055] = 'BreastClinicalFindingorIndicatedProblem'
cid_concepts[6055] = \
{'DCM': ['CancerElsewhere',
         'NonBloodyDischarge',
         'ImageDetectedMass',
         'DifficultPhysicalClinicalExamination'],
 'SCT': ['OEAxillaryLymphadenopathy',
         'DiscolorationOfSkinOfBreast',
         'BloodyNippleDischarge',
         'DisseminatedMalignancyOfUnknownPrimary',
         'PeauDOrangeSurfaceOfBreast',
         'DisorderOfBreastImplant',
         'SkinRetractionOfBreast',
         'RadiographicCalcificationFinding',
         'OELymphadenopathy',
         'OEBreastLumpPalpated',
         'BreastPain',
         'SkinThickeningOfBreast',
         'NippleDischargeSymptom',
         'Erythema',
         'NippleProblem',
         'BreastLump']}

name_for_cid[3607] = 'VenousSourceLocations'
cid_concepts[3607] = \
{'SCT': ['InnominateVein',
         'SubclavianVein',
         'PosteriorMedialTributary',
         'PortalVein',
         'VenaAnonyma',
         'BrachialVein',
         'InferiorVenaCava',
         'RenalVein',
         'BoydPerforatingVein',
         'SuperiorVenaCava',
         'CephalicVein',
         'CentralVenousSystem',
         'GastricVein',
         'SplenicVein',
         'SaphenousVein',
         'HepaticVein',
         'AntecubitalVein',
         'BasilicVein',
         'PulmonaryVeinConfluence',
         'SaphenofemoralJunction',
         'HunterianPerforatingVein',
         'InternalJugularVein',
         'AnomalousPulmonaryVein',
         'AzygosVein',
         'MesentericVein',
         'Vein',
         'UmbilicalVein',
         'AxillaryVein',
         'PulmonaryVein',
         'FemoralVein',
         'GreatSaphenousVein',
         'DoddPerforatingVein']}

name_for_cid[6405] = 'ChestNonlesionObjectTypeTissues'
cid_concepts[6405] = \
{'SCT': ['ScarTissue']}

name_for_cid[7701] = 'FiberTractsInBrainstem'
cid_concepts[7701] = \
{'SCT': ['CorticospinalTractInBrainstem',
         'MedialLemniscus',
         'MedialLongitudinalFasciculus',
         'InferiorCerebellarPeduncle',
         'SuperiorCerebellarPeduncle',
         'LateralLemniscus',
         'MiddleCerebellarPeduncle']}

name_for_cid[10044] = 'RadiosensitiveOrgans'
cid_concepts[10044] = \
{'SCT': ['Lung',
         'SmallIntestine',
         'Breast',
         'Stomach',
         'Heart',
         'Skin',
         'Thymus',
         'Prostate',
         'LymphNode',
         'Colon',
         'Uterus',
         'SalivaryGlands',
         'Gallbladder',
         'Testis',
         'Esophagus',
         'Bladder',
         'Brain',
         'EyeLenses',
         'OralMucosa',
         'Ovary',
         'Muscle',
         'Pancreas',
         'BoneMarrow',
         'Liver',
         'Thyroid',
         'Spleen',
         'Kidney',
         'AdrenalGland',
         'BoneSurface']}

name_for_cid[67] = 'ImagingAgentAdministrationCompletionStatus'
cid_concepts[67] = \
{'DCM': ['TerminatedDueToFlowRateAboveTerminationLimit',
         'TerminatedDueToConsumableRemoval',
         'TerminatedDueToPressureAboveTerminationLimit',
         'TerminatedDueToCriticalBatteryLevel',
         'TerminatedByScanner',
         'TerminatedDueToAirDetected',
         'TerminatedDueToRequestFromOperator',
         'TerminatedDueToInjectorCommunicationLoss',
         'TerminatedDueToExcessiveDurationPause',
         'TerminatedDueToUnspecifiedInjectorFailure'],
 'SCT': ['Complete']}

name_for_cid[6110] = 'LungAnatomyFindingorFeature'
cid_concepts[6110] = \
{'DCM': ['MidlungWindow', 'SecondaryPulmonaryLobule'],
 'SCT': ['SegmentOfLung', 'LobeOfLung']}

name_for_cid[3742] = 'MedicationAdministration,PCI'
cid_concepts[3742] = \
{'SCT': ['Contraindicated',
         'AdministeredAfterPercutaneousCoronaryIntervention',
         'AdministeredDuringPercutaneousCoronaryIntervention',
         'NotAdministered',
         'AdministeredPriorToPercutaneousCoronaryIntervention']}

name_for_cid[634] = 'PhaseofAnimalHandling'
cid_concepts[634] = \
{'DCM': ['InHomeCage',
         'DuringTransport',
         'StagingPriorToImaging',
         'PreparationForImaging'],
 'SCT': ['AnesthesiaInduction', 'ImagingProcedure'],
 'UMLS': ['AnesthesiaRecoveryPeriod']}

name_for_cid[4204] = 'OphthalmicFilter'
cid_concepts[4204] = \
{'DCM': ['NoFilter'],
 'SCT': ['BlueOpticalFilter',
         'PolarizingOpticalFilter',
         'YellowGreenOpticalFilter',
         'GreenOpticalFilter',
         'BlueGreenOpticalFilter',
         'InfraredOpticalFilter',
         'RedOpticalFilter']}

name_for_cid[1006] = 'AnatomicalReferenceBasisExtremities'
cid_concepts[1006] = \
{'SCT': ['Foot',
         'KneeJoint',
         'Toe',
         'MalleolarStructureOfTibia',
         'TibialPlateau',
         'ElbowJoint',
         'Patella',
         'WristJoint',
         'AnkleJoint',
         'Metacarpal',
         'Scaphoid',
         'Talus']}

name_for_cid[3484] = 'IVUSIndicesandRatios'
cid_concepts[3484] = \
{'DCM': ['LumenShapeIndex',
         'StentSymmetryIndex',
         'StentExpansionIndex',
         'PlaquePlusMediaEccentricityIndex',
         'RemodelingIndex',
         'LumenEccentricityIndex',
         'StentDiameterRatio',
         'EEMDiameterRatio',
         'LumenDiameterRatio']}

name_for_cid[12215] = 'EchocardiographyVenaCavae'
cid_concepts[12215] = \
{'LN': ['InferiorVenaCavaDiameter',
        'InferiorVenaCavaPercentCollapse',
        'HeartRate']}

name_for_cid[7202] = 'SourceImagePurposesofReference'
cid_concepts[7202] = \
{'DCM': ['PredecessorContainingGroupOfImagingSubjects',
         'ForProcessingPredecessor',
         'MaskImageForImageProcessingOperation',
         'SourceImageForImageProcessingOperation',
         'LossyCompressedPredecessor',
         'UncompressedPredecessor',
         'FlowImageForImageProcessing',
         'StructuralImageForImageProcessing',
         'SourceImageForMontage']}

name_for_cid[3611] = 'PressureMeasurements'
cid_concepts[3611] = \
{'DCM': ['XDescentPressure',
         'YDescentPressure',
         'VentricularEndDiastolicPressure',
         'AWavePeakPressure',
         'PressureAtDpDtMax',
         'VWavePeakPressure',
         'ZPointPressure',
         'GradientPressureAverage',
         'GradientPressurePeak',
         'CWavePressure'],
 'LN': ['IntravascularDiastolicBloodPressure',
        'IntravascularSystolicBloodPressure'],
 'SCT': ['AverageDiastolicBloodPressure',
         'MinimumDiastolicBloodPressure',
         'AverageSystolicBloodPressure',
         'MaximumSystolicBloodPressure',
         'MeanBloodPressure']}

name_for_cid[6043] = 'TypesofMammographyCADAnalysis'
cid_concepts[6043] = \
{'DCM': ['IndividualImpressionRecommendationAnalysis',
         'OverallImpressionRecommendationAnalysis'],
 'SCT': ['FocalAsymmetricDensityAnalysis',
         'BreastCompositionAnalysis',
         'SpatialCollocationAnalysis',
         'SpatialProximityAnalysis',
         'TemporalCorrelation',
         'ImageQualityAnalysis',
         'AsymmetricBreastTissueAnalysis']}

name_for_cid[3241] = 'HemodynamicMeasurementTechniques'
cid_concepts[3241] = \
{'SCT': ['ThermistorCatheterMethod',
         'TipManometerMethod',
         'FiberopticCatheterMethod',
         'DopplerCatheterMethod',
         'FluidFilledCatheterMethod',
         'HallCatheterMethod',
         'AveragedHemodynamicMeasurementMethod',
         'PulmonaryCapillaryWedgeMethod',
         'ConductanceCatheterMethod',
         'StaticCatheterMethod',
         'CompositeHemodynamicMeasurementMethod',
         'WedgeMethod',
         'ComputedHemodynamicMeasurementMethod',
         'DualCatheterMethod',
         'PullbackMethod']}

name_for_cid[3754] = 'VascularComplications'
cid_concepts[3754] = \
{'SCT': ['LossOfDistalPulse',
         'Pseudoaneurysm',
         'AVFistula',
         'ArterialDissection',
         'OcclusionOfArtery',
         'Bleeding']}

name_for_cid[6106] = 'AbnormalTextureFindingorFeature'
cid_concepts[6106] = \
{'DCM': ['MosaicPattern',
         'ReticulonodularPattern',
         'NodularPattern',
         'ReticularPattern',
         'AbnormalInterstitialPattern',
         'SmallIrregularOpacities',
         'GranularPattern',
         'HoneycombPattern',
         'MiliaryPattern']}

name_for_cid[71] = 'ImagingAgentAdministrationInjectorEventType'
cid_concepts[71] = \
{'DCM': ['TerminatedDueToUnspecifiedInjectorFailure',
         'FlowRateAboveAdjustmentLimit',
         'AirDetected',
         'SyringeDetached',
         'FlowRateAboveWarningLimit',
         'SyringeAttached',
         'PressureAboveAdjustmentLimit',
         'KeepVeinOpenEnded',
         'PressureAboveWarningLimit',
         'KeepVeinOpenStarted',
         'TerminatedDueToFlowRateAboveTerminationLimit',
         'TerminatedDueToConsumableRemoval',
         'TerminatedDueToPressureAboveTerminationLimit',
         'TerminatedDueToCriticalBatteryLevel',
         'FixedDurationPauseEnded',
         'TerminatedByScanner',
         'EquipmentFailure',
         'TerminatedDueToAirDetected',
         'TerminatedDueToRequestFromOperator',
         'TerminatedDueToInjectorCommunicationLoss',
         'TerminatedDueToExcessiveDurationPause',
         'ResourceInadequate']}

name_for_cid[1010] = 'ReferenceGeometryPlanes'
cid_concepts[1010] = \
{'DCM': ['PlaneThroughDistalExtent',
         'PlaneThroughProximalExtent',
         'PlaneThroughLateralExtent',
         'PlaneThroughMedialExtent',
         'PlaneThroughRightmostExtent',
         'PlaneThroughLeftmostExtent',
         'PlaneThroughPosteriorExtent',
         'PlaneThroughAnteriorExtent',
         'PlaneThroughCenter',
         'PlaneThroughInferiorExtent',
         'PlaneThroughSuperiorExtent']}

name_for_cid[622] = 'MedicationTypeCodeTypefromAnesthesiaQualityInitiative(AQI)'
cid_concepts[622] = \
{'SCT': ['OcularLubricant',
         'CaloricAgent',
         'Vasoconstrictor',
         'CarbapenemAntibiotic',
         'BetaLactamAntibiotic',
         'AminoglycosideAntibiotic',
         'Potassium',
         'LocalAnesthetic',
         'NasalDecongestant',
         'Anticoagulant',
         'PenicillinAntibiotic',
         'ThyroidHormone',
         'MetronidazoleAntibiotic',
         'Calcium',
         'Magnesium',
         'AntidiureticHormone',
         'Oxytocic',
         'InotropicAgent',
         'Estrogen',
         'Antiarrhythmic',
         'Barbiturate',
         'Benzodiazepine',
         'SulfonamideAntibiotic',
         'Diuretic',
         'VancomycinAntibiotic',
         'Antifungal',
         'Hypoglycemic',
         'Steroid',
         'LowMolecularWeightHeparin',
         'RespiratoryStimulant',
         'AntiHeparinAgent',
         'Anticholinergic',
         'NeuromuscularBlockingNMBNonDepolarizing',
         'DrugDiluent',
         'GeneralAnesthetic',
         'HemostaticAgent',
         'MacrolideAntibiotic',
         'Antihypertensive',
         'Antibiotic',
         'Antihistamine',
         'AdrenergicAgent',
         'QuinoloneAntibiotic',
         'NSAID',
         'Vitamin',
         'Antiemetic',
         'BloodProduct',
         'Vasodilator',
         'BenzodiazepineAntagonist',
         'CalciumChannelBlocker',
         'CephalosporinAntibiotic',
         'CholinergicAgent',
         'GastrointestinalProkinetic',
         'PlasmaExpander',
         'Anticonvulsant',
         'BetaBlocker',
         'Bronchodilator',
         'TetracyclineAntibiotic',
         'SkeletalMuscleRelaxant',
         'NarcoticOpiateAntagonist',
         'LincomycinAntibiotic',
         'NeuromuscularBlockingNMBDepolarizing',
         'LinezolidAntibiotic',
         'Analgesic'],
 'UMLS': ['H2Antagonist', 'NarcoticAnalgesic', 'Antihypoglycemic']}

name_for_cid[12203] = 'LeftVentricleOther'
cid_concepts[12203] = \
{'LN': ['LeftVentricularIsovolumicRelaxationTime', 'LeftVentricleMass'],
 'SCT': ['LeftVentricularPeakSystolicTissueVelocity',
         'RatioOfMVPeakVelocityToLVPeakTissueVelocityEWave',
         'LeftVentricularIndexOfMyocardialPerformance',
         'LVPeakDiastolicTissueVelocityDuringAtrialSystole',
         'LeftVentricularIsovolumicContractionTime',
         'LeftVentricularPeakEarlyDiastolicTissueVelocity']}

name_for_cid[3492] = 'VascularDissectionClassifications'
cid_concepts[3492] = \
{'DCM': ['MedialDissection',
         'IntraStentDissection',
         'AdventitialDissection',
         'IntimalDissection'],
 'SCT': ['IntramuralHematoma']}

name_for_cid[4029] = 'DermatologyAnatomicSites'
cid_concepts[4029] = \
{'DCM': ['SkinOfParaspinalAreaOfTheNeck',
         'SkinOfLowerParaspinalRegion',
         'SkinOfMidParaspinalRegion',
         'SkinOfUpperParaspinalRegion',
         'SkinOfParaspinalAreaOfTheSuperiorBack',
         'SkinOfInferiorPosteriorSurfaceOfThePinna',
         'SkinOfInferiorHelixOfEar',
         'SkinOfCaruncleOfEye',
         'SkinOfAnteriorHelixOfEar',
         'SkinOfLateralPartOfDorsumOfFoot',
         'SkinOfLowerEyelidMargin',
         'SkinOfLowerAntihelixOfEar',
         'SkinOfPosteriorHelixOfEar',
         'SkinOfParanasalCheek',
         'SkinOfMedialPartOfDorsumOfFoot',
         'SkinOfInfraalarGroove',
         'SkinOfSuperiorPosteriorHelixOfEar',
         'SkinOfSuperiorAntihelixOfEar',
         'SkinOfSoleOfForefoot',
         'SkinOfPosteriorLobuleOfTheEar',
         'SkinOfUpperEyelidMargin',
         'SkinOfUpperAntihelixOfEar',
         'SkinOfSuperiorPosteriorSurfaceOfThePinna',
         'SkinOfMidBack'],
 'FMA': ['SkinOfLowerOuterQuadrantOfBreast',
         'SkinOfUpperInnerQuadrantOfBreast',
         'SkinOfLowerInnerQuadrantOfBreast',
         'SkinOfDorsalPartOfFourthToe',
         'SkinOfDorsalPartOfFifthToe',
         'SkinOfUpperOuterQuadrantOfLeftBreast',
         'SkinOfVermilionProperOfLowerLip',
         'SkinOfDorsalPartOfGreatToe',
         'SkinOfDorsalPartOfSecondToe',
         'SkinOfDorsalPartOfThirdToe',
         'SkinOfVermilionProperOfUpperLip',
         'SkinOfBackOfTrunk',
         'MucosaOfOralSegmentOfHardPalate',
         'MucosaOfUvula',
         'MucosaOfPalatoglossalArch',
         'SkinOfPlantarPartOfThirdToe',
         'SkinOfPlantarPartOfFourthToe',
         'SkinOfPalmarPartOfLittleFinger',
         'SkinOfPlantarPartOfFifthToe',
         'SkinOfPlantarPartOfSecondToe',
         'SkinOfPalmarPartOfRingFinger',
         'SkinOfPalmarPartOfIndexFinger',
         'SkinOfPalmarPartOfMiddleFinger',
         'SkinOfPlantarPartOfGreatToe',
         'SkinOfPalmarPartOfThumb',
         'SkinOfDorsalPartOfLittleFinger',
         'MucosaOfPharynx',
         'SkinOfDorsalPartOfRingFinger',
         'MucosaOfDorsumOfPharyngealPartOfTongue',
         'MucosaOfDorsumOfOralPartOfTongue',
         'SkinOfDorsalPartOfIndexFinger',
         'SkinOfDorsalPartOfMiddleFinger',
         'SkinOfDorsalPartOfThumb',
         'SkinOfDorsumOfNose',
         'MucosaOfPosteriorWallOfOropharynx'],
 'SCT': ['MucosaOfFloorOfMouth',
         'SkinOfOccipitalRegion',
         'NailOfFourthToe',
         'SkinOfSupraclavicularRegionOfNeck',
         'SkinOfAnteriorSurfaceOfLowerLeg',
         'SkinOfMedialCanthus',
         'FemaleExternalUrethralOrifice',
         'SkinOfAnteriorSurfaceOfUpperArm',
         'SkinOfAnteriorSurfaceOfForearm',
         'SkinOfPosteriorSurfaceOfForearm',
         'SkinOfChin',
         'SkinOfTemporalRegion',
         'NailOfMiddleFinger',
         'SkinOfLowerEyelid',
         'MucosaOfMandibularGingiva',
         'SkinOfInguinalRegion',
         'NailOfSecondToe',
         'SkinOfLabium',
         'SkinOfPerineum',
         'MaleExternalUrethralOrifice',
         'SkinOfUpperAbdomen',
         'SkinOfLowerLip',
         'Cornea',
         'Eyelash',
         'SkinOfEyeRegion',
         'SkinOfAntitragus',
         'SkinOfScalp',
         'SkinOfMedialSurfaceOfThigh',
         'SkinOfAreola',
         'SkinOfEarLobule',
         'SkinOfNuchalRegion',
         'SkinOfPrepuceOfClitoris',
         'SkinOfPosteriorSurfaceOfElbow',
         'SkinOfBack',
         'SkinOfPoplitealFossa',
         'VulvalVestibule',
         'SkinOfPosteriorSurfaceOfUpperArm',
         'SkinOfLateralPartOfHeel',
         'SkinOfUmbilicus',
         'SkinOfSupraorbitalArea',
         'FrenulumOfLabiaMinora',
         'SkinOfLateralCanthus',
         'SkinOfBackOfUpperThoracicRegion',
         'NailOfIndexFinger',
         'SkinOfLowerChestWall',
         'SkinOfUpperExtremity',
         'SkinOfNasolabialFold',
         'SkinOfThenarRegionOfPalm',
         'SkinOfHeel',
         'SkinOfAntecubitalFossa',
         'SkinOfExternalGenitalia',
         'MucosaOfUpperLip',
         'SkinOfSubmentalArea',
         'SkinOfClitoris',
         'SkinOfPostauricularRegion',
         'SkinOfForehead',
         'MucosaOfMaxillaryGingiva',
         'SkinOfAnteriorSurfaceOfKnee',
         'SkinOfLips',
         'MucosaOfTipOfTongue',
         'SkinOfSideOfNose',
         'SkinOfLowerBack',
         'SkinOfPhiltrum',
         'Sclera',
         'SkinOfTragus',
         'SkinOfFace',
         'SkinOfCavityOfConcha',
         'SkinOfEpigastricArea',
         'SkinOfPalmOfHand',
         'SkinOfTipOfNose',
         'SkinOfLateralAspectOfAnkle',
         'SkinOfCrusOfHelix',
         'VaginalIntroitus',
         'SkinOfAnteriorPortionOfNeck',
         'SkinOfLowerAbdomen',
         'SkinOfExternalAuditoryCanal',
         'SkinOfAnteriorSurfaceOfThigh',
         'SkinOfNeck',
         'SkinOfLateralPortionOfNeck',
         'SkinOfAbdomen',
         'SkinOfLabiumMinus',
         'Tooth',
         'SkinOfDigitOfHand',
         'SubmandibularTriangle',
         'SkinOfHead',
         'SkinOfLowerExtremity',
         'SkinOfMedialPartOfHeel',
         'SkinOfRootOfPenis',
         'SkinOfPartOfDorsalSurfaceOfHand',
         'SkinOfMedialAspectOfAnkle',
         'SkinOfGlutealFold',
         'SkinOfLabiumMajus',
         'SkinOfButtock',
         'SkinOfLateralBorderOfSoleOfFoot',
         'SkinOfParietalRegion',
         'SkinOfEyebrow',
         'SkinOfHand',
         'NailOfFifthToe',
         'NailOfThirdToe',
         'SkinOfPosteriorSurfaceOfLowerLeg',
         'SkinOfAlaNasi',
         'PosteriorCommissureOfLabiumMajorum',
         'SkinOfUpperLip',
         'SkinOfPosteriorSurfaceOfThorax',
         'SkinOfToe',
         'SkinOfPosteriorSurfaceOfThigh',
         'SkinOfGlansPenis',
         'SkinOfNose',
         'SkinOfVertexOfScalp',
         'SkinOfPalmarAreaOfWrist',
         'SkinOfHelixOfEar',
         'SkinOfInfraclavicularRegion',
         'Iris',
         'SkinOfAxilla',
         'SkinOfShaftOfPenis',
         'NailOfFinger',
         'MucosaOfTongue',
         'SkinOfHypothenarRegionOfPalm',
         'SkinOfAnteriorSurfaceOfThorax',
         'OralMucosa',
         'GroinSkinCrease',
         'SkinOfAnteriorTrunk',
         'NailOfToe',
         'SkinOfUpperEyelid',
         'SkinOfEar',
         'SkinOfPreauricularRegion',
         'NailOfGreatToe',
         'NailOfLittleFinger',
         'SkinOfAnus',
         'SternalSkin',
         'Hair',
         'AnteriorTriangleOfNeck',
         'MucosaOfLowerLip',
         'Skin',
         'SkinOfNipple',
         'SkinOfPerioralRegionOfFace',
         'SkinOfCheek',
         'SkinOfJawline',
         'NailOfRingFinger',
         'NailOfThumb',
         'SkinOfFoot',
         'SkinOfIntertragalIncisure',
         'SkinOfDorsalAreaOfWrist',
         'SkinOfHypogastricRegion',
         'SkinOfScrotum',
         'SkinOfUpperTrunk',
         'Retina',
         'SkinOfPenis',
         'SkinOfMedialBorderOfSoleOfFoot']}

name_for_cid[6014] = 'MammographySingleImageFinding'
cid_concepts[6014] = \
{'DCM': ['BreastGeometry',
         'CooperLigamentChanges',
         'ImageQuality',
         'MassInTheSkin',
         'SelectedRegion',
         'NonLesion',
         'MassOnTheSkin'],
 'SCT': ['BreastComposition',
         'AxillaryAdenopathy',
         'Edema',
         'AxillaryLymphNode',
         'TrabecularThickeningOfBreast',
         'SkinRetractionOfBreast',
         'ArchitecturalDistortionOfBreast',
         'TubularDensity',
         'MammographyBreastDensity',
         'SkinLesion',
         'IndividualCalcification',
         'Nipple',
         'SkinThickeningOfBreast',
         'CalcificationCluster',
         'IntramammaryLymphNode']}

name_for_cid[26] = 'NuclearMedicineProjections'
cid_concepts[26] = \
{'SCT': ['MedialLateral',
         'SagittalObliqueAxial',
         'LeftLateralProjection',
         'VerticalLongAxis',
         'HorizontalLongAxis',
         'Sagittal',
         'RightPosteriorOblique',
         'LeftAnteriorOblique',
         'PosteriorProjection',
         'AnteriorProjection',
         'Axial',
         'LateralMedial',
         'RightAnteriorOblique',
         'LeftPosteriorOblique',
         'ShortAxis',
         'MedioLateralOblique',
         'Lateral',
         'RightLateralProjection',
         'LateroMedialOblique',
         'Coronal',
         'ObliqueAxial',
         'Transverse']}

name_for_cid[7310] = 'ImplantFixationMethod'
cid_concepts[7310] = \
{'DCM': ['Bolting', 'Wedging', 'Pinning', 'Sewing'],
 'SCT': ['RepairByNailing',
         'Anchoring',
         'Gluing',
         'UncementedComponentFixation',
         'CementedComponentFixation',
         'InternalFixationUsingStaple',
         'InternalFixationUsingInternalFixatorSystem',
         'Fusion',
         'InternalFixationUsingPlate',
         'InternalFixationUsingScrew']}

name_for_cid[6151] = 'BackgroundEchotexture'
cid_concepts[6151] = \
{'DCM': ['HomogeneousFatEchotexture',
         'HomogeneousFibroglandularEchotexture',
         'HeterogeneousEchotexture']}

name_for_cid[3703] = 'WallMotion'
cid_concepts[3703] = \
{'DCM': ['NotVisualized'],
 'SCT': ['ModerateHypokinesis',
         'HyperkineticRegion',
         'Hypokinesis',
         'SevereHypokinesis',
         'Unknown',
         'NormalWallMotion',
         'MildHypokinesis',
         'Akinesis',
         'NotEvaluated',
         'Dyskinesis']}

name_for_cid[7039] = 'PediatricSizeCategories'
cid_concepts[7039] = \
{'SCT': ['BroselowLutenPurpleZone1011kg',
         'BroselowLutenYellowZone1214kg',
         'BroselowLutenWhiteZone1518kg',
         'BroselowLutenGreenZone3036kg',
         'BroselowLutenBlueZone1923kg',
         'BroselowLutenPinkZone67kg',
         'BroselowLutenOrangeZone2429kg',
         'BroselowLutenRedZone89kg']}

name_for_cid[225] = 'MeasurementUncertaintyConcepts'
cid_concepts[225] = \
{'SCT': ['RangeOfUpperMeasurementUncertainty',
         'RangeOfMeasurementUncertainty',
         'RangeOfLowerMeasurementUncertainty']}

name_for_cid[7469] = 'GenericIntensityandSizeMeasurements'
cid_concepts[7469] = \
{'DCM': ['ApparentDiffusionCoefficient',
         'DiffusionWeighted',
         'VolumeRatio',
         'VolumeOfSphere',
         'DiffusionCoefficient',
         'WaterFatOutOfPhase',
         'FractionalOccupancySegmentation',
         'WaterFatInPhase',
         'FractionalProbabilisticSegmentation',
         'BinarySegmentation',
         'WaterFraction',
         'MRSignalIntensity',
         'XRayAttenuation',
         'VolumeOfCircumscribedSphere',
         'VolumeOfEllipsoid',
         'FatFraction',
         'VolumeEstimatedFromThreeOrMoreNonCoplanar2DRegions',
         'VolumeEstimatedFromTwoNonCoplanar2DRegions',
         'VolumeEstimatedFromSingle2DRegion',
         'PathLength',
         'VolumeOfBoundingThreeDimensionalRegion',
         'YBRICTCRComponent',
         'R2Coefficient',
         'YBRICTCBComponent',
         'ChiSquare',
         'YBRICTYComponent',
         'DW',
         'YBRPARTIALCRComponent',
         'AIC',
         'YBRPARTIALCBComponent',
         'BIC',
         'YBRPARTIALYComponent',
         'Echogenicity',
         'YBRRCTCRComponent',
         'YBRRCTCBComponent',
         'YBRRCTYComponent',
         'Distance',
         'Height',
         'RelativeLinearStoppingPower',
         'RGBRComponent',
         'UltrasoundAttenuation',
         '',
         'SpeedOfSound',
         'Vp',
         'Perfusion',
         'Elasticity',
         'EquivalentDose',
         'AbsorbedDose',
         'YBRFULLCRComponent',
         'YBRFULLCBComponent',
         'YBRFULLYComponent',
         'RGBBComponent',
         'RGBGComponent',
         'CreatineCholineCitrateRatio',
         'MagnetizationTransferRatio',
         'IAUC90BN',
         'IAUC180BN',
         'CreatineAndCholine',
         'LipidAndLactate',
         'SpatialDisplacementYComponent',
         'SpatialDisplacementXComponent',
         'NuclearMedicineTomographicActivity',
         'IAUC',
         'NuclearMedicineProjectionActivity',
         'IAUC60',
         'IAUC90',
         'IAUC180',
         'IAUCBN',
         'IAUC60BN',
         'FlowVariance',
         'FlowVelocity',
         'TissueVelocity',
         'IndexedHemodynamicResistance',
         'HemodynamicResistance',
         'SpatialDisplacementZComponent',
         'AttenuationCoefficient',
         'GlutamateAndGlutamine',
         'CholineCreatineRatio',
         'NAcetylaspartateCreatineRatio',
         'NAcetylaspartateCholineRatio',
         'Tmax',
         'R2',
         'VolumetricDiffusionDxzComponent',
         'MagneticSusceptibility',
         'VolumetricDiffusionDxyComponent',
         'VolumetricDiffusionDxxComponent',
         'RelativeRegionalBloodFlow',
         'RelativeRegionalBloodVolume',
         'Ktrans',
         'Kep',
         'Ve',
         'BloodOxygenationLevel',
         'T2WeightedDynamicContrastEnhancedMRSignalIntensity',
         'T2WeightedDynamicContrastEnhancedMRSignalIntensity',
         'AbsoluteRegionalBloodFlow',
         'T1WeightedDynamicContrastEnhancedMRSignalIntensity',
         'AbsoluteRegionalBloodVolume',
         'VolumetricDiffusionDzzComponent',
         'OxygenExtractionFraction',
         'R1',
         'VolumetricDiffusionDyzComponent',
         'R2',
         'VolumetricDiffusionDyyComponent',
         'VelocityEncoded',
         'ZScore',
         'ContrastAgentAngioMRSignalIntensity',
         'SpinTaggingPerfusionMRSignalIntensity',
         'RelativeAnisotropy',
         'FractionalAnisotropy',
         'FieldMapMRSignalIntensity',
         'T2WeightedMRSignalIntensity',
         'T2WeightedMRSignalIntensity',
         'T1WeightedMRSignalIntensity',
         'ProtonDensityWeightedMRSignalIntensity',
         'TimeOfFlightAngioMRSignalIntensity',
         'T1',
         'T2',
         'T2',
         'TimeCourseOfSignal',
         'TemperatureEncoded',
         'StudentTTest',
         'TimeToPeak',
         'SignalToNoise',
         'TemporalDerivativeThreshold',
         'MaximumSlope',
         'MaximumDifference',
         'TracerConcentration',
         'TimeOfPeakConcentration',
         'BolusArrivalTime',
         'TimeOfLeadingHalfPeakConcentration',
         'MeanTransitTime',
         'KurtosisDiffusionCoefficient',
         'GammaDistributionScaleParameter',
         'NegativeEnhancementIntegral',
         'GammaDistributionShapeParameter',
         'GammaDistributionMode',
         'DistributedDiffusionCoefficient',
         'RCoefficient',
         'AnomalousExponentParameter',
         'ProtonDensity',
         'SignalChange',
         'Suvbsa',
         'Suvibw',
         'SuvlbmJanma',
         'SuvlbmJames128',
         'MonoExponentialApparentDiffusionCoefficient',
         'SlowDiffusionCoefficient',
         'FastDiffusionCoefficient',
         'FastDiffusionCoefficientFraction',
         'StandardizedUptakeValue',
         'Suvbw',
         'Suvlbm'],
 'SCT': ['Width',
         'Volume',
         'Fat',
         'Area',
         'Lipid',
         'DiameterOfCircumscribedCircle',
         'Depth',
         'NAcetylaspartate',
         'PowerDoppler',
         'Perimeter',
         'Creatine',
         'Glutamine',
         'MinorAxis',
         'Diameter',
         'Tuarine',
         'Citrate',
         'MajorAxis',
         'Lactate',
         'LongAxis',
         'Length',
         'Circumference',
         'Inositol',
         'PerpendicularAxis',
         'Water',
         'Radius',
         'Choline',
         'AreaOfDefinedRegion',
         'ShortAxis']}

name_for_cid[7193] = 'PhysicalObjectSegmentationPropertyTypes'
cid_concepts[7193] = \
{'DCM': ['OtherMarker',
         'Mold',
         'BrachytherapyChannelShield',
         'HeadAndNeckMask',
         'BrachytherapySourceApplicator',
         'HeadMask',
         'Headframe',
         'BodyFrame',
         'BreastBoard',
         'Cast',
         'WholeBodyPod',
         'VacuumMold',
         'Portacath',
         'ChestTube',
         'CentralLine',
         'KidneyStent',
         'PancreaticStent',
         'NippleRing',
         'Coin',
         'BrachytherapyChannel',
         'RectalBalloon',
         'FiducialMark',
         'VaginalCylinder',
         'IonChamberArray',
         'DiodeArray',
         'LiquidIonChamber',
         'Diode',
         'IonChamber',
         'OSLD',
         'DiamondDetector',
         'Sphere'],
 'SCT': ['Suture',
         'Chair',
         'ForeignBody',
         'Glue',
         'InlayDentalRestoration',
         'EmbolizationCoil',
         'Jewelry',
         'Prosthesis',
         'MOSFETRadiationTherapyDosimetrySystemDosimeter',
         'Catheter',
         'EndotrachealTube',
         'MeasuringRuler',
         'Collimator',
         'RotationalAtherectomyDevice',
         'Bullet',
         'DentalProsthesis',
         'GuidingCatheter',
         'ThermoluminescentRadiationDosimeter',
         'OpaqueMarker',
         'BiteBlock',
         'CardiacPacemakerLead',
         'Clip',
         'BoneScrew',
         'CervicalCollar',
         'DetachableBalloon',
         'JejunostomyTube',
         'RadioactiveImplant',
         'CardiacPacemaker',
         'JWire',
         'VenaCavaFilter',
         'HeartValveProsthesis',
         'Needle',
         'AtherectomyDevice',
         'EmbolizationParticulate',
         'FeedingTube',
         'Laser',
         'Table',
         'Tracheotomy',
         'Staple',
         'CompressionPaddle',
         'UretericStent',
         'EmbolizationBall',
         'PercutaneousTransluminalAngioplastyBalloon',
         'BonePin',
         'BBShotLeadPellet',
         'Defibrillator',
         'Headrest',
         'CardiacPacemaker',
         'DigitalImager',
         'IDPlate',
         'Clothing',
         'LeftVentricularAssistDevice',
         'Stent',
         'Implant']}

name_for_cid[4245] = 'WideFieldOphthalmicPhotographyTransformationMethod'
cid_concepts[4245] = \
{'DCM': ['SphericalProjection', 'SurfaceContourMapping']}

name_for_cid[9532] = 'NoParameterDosimetricObjectives'
cid_concepts[9532] = \
{'DCM': ['MinimizeMeterset']}

name_for_cid[7486] = 'MixedBreeds'
cid_concepts[7486] = \
{'SCT': ['MixedBreedGoat',
         'MixedBreedCattle',
         'MixedBreedChicken',
         'MixedBreedDog',
         'MixedBreedCat',
         'MixedBreedPig',
         'MixedBreedHorse',
         'MixedBreedSheep']}

name_for_cid[12254] = 'CardiacUltrasoundHemodynamicMeasurements'
cid_concepts[12254] = \
{'DCM': ['Tau', 'IndicatorMeanTransitTime', 'RRInterval'],
 'LN': ['PreEjectionPeriod',
        'EjectionTime',
        'AWaveDuration',
        'EnvelopeDuration',
        'TimeToPeakByUS',
        'DpDtByUS',
        'HeartRateCorrectedEjectionTime',
        'FillingTime',
        'HeartRateCorrectedPreEjectionPeriod',
        'PreEjectionPeriodEjectionTimeRatio',
        'IsovolumicContractionTime',
        'ClosureToOpeningTime',
        'IsovolumicRelaxationTime'],
 'SCT': ['BloodPressure']}


